!  Copyright (C) 2019-2021, respective authors of MCFM.
!
!  This program is free software: you can redistribute it and/or modify it under
!  the terms of the GNU General Public License as published by the Free Software
!  Foundation, either version 3 of the License, or (at your option) any later
!  version.
!
!  This program is distributed in the hope that it will be useful, but WITHOUT ANY
!  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
!  PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License along with
!  this program. If not, see <http://www.gnu.org/licenses/>
 
module Beamfunctions
    use LHAPDF
    use types
    use constants
    implicit none

    public :: getbeam
    public :: getbeam2

    ! for debugging
    public :: ibar_select
    public :: set_params, set_nf

    ! legacy
    !public :: B1_qq, B2_qq, B3_qq
    !public :: B2_qq_alt

    logical, public, save :: usegrid, makegrid

    private

    !integer, parameter :: nf = 5

    !real(dp), parameter :: beta0 = 11._dp/3._dp*CA - 4._dp/3._dp*TF*nf

    ! these are used by the integrands
    integer, save :: flavor
!$omp threadprivate(flavor)
    real(dp), save :: xi, mu
!$omp threadprivate(xi,mu)
    integer, save :: flavor_i, powAs, powLperp, ih
!$omp threadprivate(flavor_i, powAs, powLperp, ih)
    integer, save :: nf
!$omp threadprivate(nf)

    integer, save :: usebeam
!$omp threadprivate(usebeam)

    real(dp), parameter :: integrationPrecision = 1e-4_dp

    contains

    subroutine set_nf(nf_in)
        implicit none
        integer, intent(in) :: nf_in

        nf = nf_in
    end subroutine

    subroutine set_params(ih_in, flavor_i_in, powAs_in, powLperp_in, nf_in)
        implicit none
        integer, intent(in) :: ih_in
        integer, intent(in) :: flavor_i_in
        integer, intent(in) :: powAs_in, powLperp_in
        integer, intent(in) :: nf_in

        ih = ih_in
        flavor_i = flavor_i_in
        powAs = powAs_in
        powLperp = powLperp_in
        nf = nf_in
    end subroutine

    function getbeam2(ih_in, flavor_i_in, powAs_in, powLperp_in, xi_in, mu_in)
        use iso_fortran_env
        use Quadpack
        implicit none

        integer, intent(in) :: ih_in
        integer, intent(in) :: flavor_i_in
        integer, intent(in) :: powAs_in, powLperp_in
        real(dp), intent(in) :: xi_in, mu_in
        real(dp) :: getbeam2

        real(dp) :: absErr
        integer(int64) :: neval, ierr
        real(dp) :: beam_contrib
        integer :: flavor_j
        real(dp) :: pdfone

        ! this one is always computed on the fly, for now

        if (flavor_i_in /= 0) then
            getbeam2 = 0._dp
        endif

        if (usegrid .and. (.not. makegrid)) then
            getbeam2 = fdist_one_beam2(ih_in, powAs_in, powLperp_in, xi_in, mu_in, flavor_i_in)
            return
        endif

        ih = ih_in
        flavor_i = flavor_i_in
        powAs = powAs_in
        powLperp = powLperp_in
        xi = xi_in
        mu = mu_in

        nf = getnumflavors(mu_in)

        getbeam2 = 0._dp

        ! perform getbeam_integrand integral
        beam_contrib = 0._dp
        call qags(getbeam2_integrand, xi, 1._dp, &
            integrationPrecision, integrationPrecision, &
            beam_contrib, absErr, neval, ierr)
        getbeam2 = beam_contrib

    end function

    function getbeam2_integrand(z)
        implicit none

        real(dp), intent(in) :: z
        real(dp) :: getbeam2_integrand

        integer :: flavor_j
        real(dp) :: pdfone, pdfz

        getbeam2_integrand = 0._dp

        do flavor_j=-nf,nf
            pdfz = fdist_one(ih, xi/z, mu, flavor_j)

            if (flavor_i == 0 .and. flavor_j == 0) then
                getbeam2_integrand = getbeam2_integrand + &
                    (-4*CA*(1-z)/z) / z * pdfz
            elseif (flavor_i == 0 .and. flavor_j /= 0) then
                getbeam2_integrand = getbeam2_integrand + &
                    (-4*CF*(1-z)/z) / z * pdfz
            endif
        enddo
    end function

    ! this gets the beam function component for a specific as and Lperp component
    function getbeam(ih_in, flavor_i_in, powAs_in, powLperp_in, xi_in, mu_in, ibeam_in)
        use iso_fortran_env
        use Quadpack
        implicit none

        integer, intent(in) :: ih_in
        integer, intent(in) :: flavor_i_in
        integer, intent(in) :: powAs_in, powLperp_in
        real(dp), intent(in) :: xi_in, mu_in
        integer, optional, intent(in) :: ibeam_in
        real(dp) :: getbeam

        real(dp) :: absErr
        integer(int64) :: neval, ierr
        real(dp) :: beam_contrib
        integer :: flavor_j
        real(dp) :: pdfone
        integer :: ibeam

        if (present(ibeam_in)) then
            ibeam = ibeam_in
        else
            ibeam = 1
        endif
        ! save globally
        usebeam = ibeam

        if (usegrid .and. (.not. makegrid)) then
            getbeam = fdist_one_beam(ih_in, powAs_in, powLperp_in, xi_in, mu_in, flavor_i_in, ibeam)
            return
        endif

        ih = ih_in
        flavor_i = flavor_i_in
        powAs = powAs_in
        powLperp = powLperp_in
        xi = xi_in
        mu = mu_in
        
        nf = getnumflavors(mu_in)

        getbeam = 0._dp


        ! perform getbeam_integrand integral
        beam_contrib = 0._dp
        call qags(getbeam_integrand, xi, 1._dp, &
            integrationPrecision, integrationPrecision, &
            beam_contrib, absErr, neval, ierr)
        getbeam = getbeam + beam_contrib

        do flavor_j=-nf,nf
             pdfone = fdist_one(ih, xi, mu, flavor_j, ibeam)

            ! delta(1-z) contribution
            getbeam = getbeam + &
                Ibar_select(flavor_j, 2, 1._dp) * pdfone

            ! boundary term for (1/(1-z))_+
            if (xi < 1._dp) then
                getbeam = getbeam + &
                    Ibar_select(flavor_j, 3, 1._dp) * pdfone * log(1._dp - xi)
            endif

            ! boundary term for (log(1-z)/(1-z))_+
            if (xi < 1._dp) then
                getbeam = getbeam + &
                    Ibar_select(flavor_j, 4, 1._dp) * pdfone * log(1._dp - xi)**2 / 2._dp
            endif
        enddo

    end function

    function getbeam_integrand(z)
        implicit none

        real(dp), intent(in) :: z
        real(dp) :: getbeam_integrand

        integer :: flavor_j
        real(dp) :: pdfone, pdfz

        getbeam_integrand = 0._dp

        do flavor_j=-nf,nf
            pdfz = fdist_one(ih, xi/z, mu, flavor_j, usebeam)
            ! O(1) contribution
            getbeam_integrand = getbeam_integrand + &
                Ibar_select(flavor_j, 1, z)/z * pdfz

            ! delta(1-z), see above

            ! (1/(1-z))_+ contribution, boundary term see above
            if (z < 1._dp) then
                pdfone = fdist_one(ih, xi, mu, flavor_j, usebeam)

                getbeam_integrand = getbeam_integrand + &
                    1._dp/(1._dp - z) * &
                    (Ibar_select(flavor_j, 3, z)/z * pdfz &
                  - Ibar_select(flavor_j, 3, 1._dp) * pdfone )

                ! (log(1-z)/(1-z))_+ contribution, boundary term see above
                getbeam_integrand = getbeam_integrand + &
                    log(1._dp-z)/(1._dp-z) * &
                    (Ibar_select(flavor_j, 4, z)/z * pdfz &
                  - Ibar_select(flavor_j, 4, 1._dp) * pdfone )
            endif

        enddo
    end function

    ! select appropriate Ibar_ functions based on i and j
    ! this has been exhaustively tested against MM code
    function Ibar_select(flavor_j, type_in, z)
        implicit none

        integer, intent(in) :: flavor_j
        integer, intent(in) :: type_in
        real(dp), intent(in) :: z
        real(dp) :: Ibar_select

        real(dp) :: Li2, Li3, ReLi3

        Ibar_select = 0._dp

        if (flavor_i /= 0) then
            if (flavor_j == flavor_i) then
                ! Ibar_qq
#include "src/Mods/include_qq.f90"
            elseif (flavor_j == 0) then
                ! Ibar_qg
#include "src/Mods/include_qg.f90"
            elseif (flavor_j == -flavor_i) then
                ! Ibar_qbq
#include "src/Mods/include_qbq.f90"
            elseif (flavor_i * flavor_j > 0) then
                ! Ibar_qpq
#include "src/Mods/include_qpq.f90"
            elseif (flavor_i * flavor_j < 0) then
                ! Ibar_qpbq
#include "src/Mods/include_qpbq.f90"
            else
                error stop "invalid flavor selection in Ibar_select"
            endif
        elseif (flavor_i == 0) then
            if (flavor_j == 0) then
                ! Ibar_gg
#include "src/Mods/include_gg.f90"
            elseif (flavor_j /= 0) then
                ! Ibar_gq
#include "src/Mods/include_gq.f90"
            else
                error stop "invalid flavor selection in Ibar_select"
            endif
        else
            error stop "invalid flavor selection in Ibar_select"
        endif

    end function

!    function B1_qq(j, xi_in, mu_in)
!        use iso_fortran_env
!        use Quadpack
!        implicit none
!        real(dp) :: B1_qq
!        integer, intent(in) :: j
!        real(dp), intent(in) :: xi_in, mu_in
!
!        real(dp) :: absErr
!        integer(int64) :: neval, ierr
!        real(dp) :: phiquark1
!
!        xi = xi_in
!        mu = mu_in
!        flavor = j
!
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!
!        call qags(B1_qq_integrand, xi, 1._dp, &
!            integrationPrecision, integrationPrecision, &
!            B1_qq, absErr, neval, ierr)
!
!        B1_qq = B1_qq - CF*pi**2/6._dp*phiquark1
!    end function
!
!    ! this is based on a direct evaluation of ((1+x^2)/(1-x))_+
!    ! and the corresponding boundary term
!    function B2_qq_alt(j, xi_in, mu_in)
!        use iso_fortran_env
!        use Quadpack
!        implicit none
!        real(dp) :: B2_qq_alt
!        integer, intent(in) :: j
!        real(dp), intent(in) :: xi_in, mu_in
!
!        real(dp) :: absErr
!        integer(int64) :: neval, ierr
!        real(dp) :: phiquark1
!
!        xi = xi_in
!        mu = mu_in
!        flavor = j
!
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!
!        call qags(B2_qq_alt_integrand, xi, 1._dp, &
!            integrationPrecision, integrationPrecision, &
!            B2_qq_alt, absErr, neval, ierr)
!
!        ! boundary term
!        if (xi < 1._dp) then
!            B2_qq_alt = B2_qq_alt - (- 0.5_dp*xi*(2._dp+xi) - 2._dp*log(1-xi))*4._dp*CF * phiquark1
!        endif
!    end function
!
!    function B3_qq(j, xi_in, mu_in)
!        use iso_fortran_env
!        use Quadpack
!        implicit none
!        real(dp) :: B3_qq
!        integer, intent(in) :: j
!        real(dp), intent(in) :: xi_in, mu_in
!
!        real(dp) :: absErr
!        integer(int64) :: neval, ierr
!        real(dp) :: phiquark1
!        real(dp) :: Li2
!
!        xi = xi_in
!        mu = mu_in
!        flavor = j
!
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!
!        B3_qq = 0._dp
!
!        call qags(B3_qq_integrand, xi, 1._dp, &
!            integrationPrecision, integrationPrecision, &
!            B3_qq, absErr, neval, ierr)
!
!        if (xi < 1._dp) then
!            B3_qq = B3_qq + (4*CF**2*phiquark1*(27 - 8*Pi**2 + 72*Log(1 - xi) + 48*Log(1 - xi)**2))/3._dp
!        endif
!
!        ! alternative version
!        !if (xi < 1._dp) then
!            !B3_qq = B3_qq + (4*CF**2*phiquark1*(-27 - 8*Pi**2 + 36*xi + 18*xi**2 + &
!                !72*Log(1 - xi) + 48*Log(1 - xi)**2 +  &
!            !48*Li2(1 - xi)))/3._dp
!        !endif
!
!
!    end function
!
!    function B3_qq_integrand(z)
!        implicit none
!        real(dp) :: B3_qq_integrand
!        real(dp), intent(in) :: z
!
!        real(dp) :: fx(-5:5)
!        real(dp) :: phigluon, phiquark, phiquark1, phiquarksum
!        real(dp) :: x
!        integer :: j
!
!        x = xi/z
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!
!        call fdist(+1, x, mu, fx)
!        phiquark = fx(flavor)
!        phigluon = fx(0)
!        phiquarksum = sum(fx(-nf:-1)) + sum(fx(1:nf))
!
!        B3_qq_integrand = &
!        (32*CF**2*phiquark1*(3 + 4*Log(1 - z)))/(-1 + z) - &
!        (16*CF*phiquarksum*TF*(-4 - 3*z + 3*z**2 + 4*z**3 -  &
!             6*z*(1 + z)*Log(z)))/(3._dp*z**2) - &
!        (16*CF**2*phiquark*(1 + 4*z + z**2 + 4*(1 + z**2)*Log(1 - z) - &
!             (1 + 3*z**2)*Log(z)))/((-1 + z)*z) + &
!        (8*phigluon*TF*(8*CA + 3*beta0*z + 6*CA*z - 3*CF*z - 6*beta0*z**2 + &
!             48*CA*z**2 + 12*CF*z**2 + 6*beta0*z**3 - 62*CA*z**3 + &
!             12*(CA + CF)*z*(1 - 2*z + 2*z**2)*Log(1 - z) -  &
!             6*z*(-2*CA + CF - 8*CA*z - 2*CF*z + 4*CF*z**2)*Log(z)))/(3._dp*z**2)
!
!        ! alternative version
!        !B3_qq_integrand = &
!        !(-16*CF*phiquarksum*TF*(-4 - 3*z + 3*z**2 + 4*z**3 - 6*z*(1 + z)*Log(z)))/(3._dp*z**2) +  & 
!        !(8*phigluon*TF*(8*CA + 3*beta0*z + 6*CA*z - 3*CF*z - 6*beta0*z**2 + 48*CA*z**2 + &
!             !12*CF*z**2 + 6*beta0*z**3 - 62*CA*z**3 + &
!             !12*(CA + CF)*z*(1 - 2*z + 2*z**2)*Log(1 - z) - &
!             !6*z*(-2*CA + CF - 8*CA*z - 2*CF*z + 4*CF*z**2)*Log(z)))/(3.*z**2) + &
!        !(16*CF**2*phiquark1*(3 + 3*z**2 + 4*Log(-2 + 1/z + z)))/(-1 + z) - &
!        !(16*CF**2*phiquark*(1 + 4*z + z**2 + 4*(-1 + z**2)*Log(1 - z) - 3*(-1 + z**2)*Log(z) + &
!             !4*Log(-2 + 1/z + z)))/((-1 + z)*z)
!
!    end function
!
!    function B2_qq_alt_integrand(z)
!        implicit none
!        real(dp) :: B2_qq_alt_integrand
!        real(dp), intent(in) :: z
!
!        real(dp) :: phiquark, phiquark1, phigluon
!        real(dp) :: x
!
!        x = xi/z
!        phiquark  = fdist_one(+1, x, mu, flavor)
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!        phigluon  = fdist_one(+1, x, mu, 0)
!
!        B2_qq_alt_integrand = 4._dp*tf*(1._dp - 2._dp*z + 2._dp*z*z)*phigluon/z
!        if (z < 1._dp) then
!            B2_qq_alt_integrand = B2_qq_alt_integrand &
!                + 4._dp*CF*(1._dp+z*z)/(1._dp-z) * (phiquark/z - phiquark1)
!        endif
!
!    end function
!
!    function B2_qq(j, xi_in, mu_in)
!        use iso_fortran_env
!        use Quadpack
!        implicit none
!        real(dp) :: B2_qq
!        integer, intent(in) :: j
!        real(dp), intent(in) :: xi_in, mu_in
!
!        real(dp) :: absErr
!        integer(int64) :: neval, ierr
!        real(dp) :: phiquark1
!
!        xi = xi_in
!        mu = mu_in
!        flavor = j
!
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!
!        call qags(B2_qq_integrand, xi, 1._dp, &
!            integrationPrecision, integrationPrecision, &
!            B2_qq, absErr, neval, ierr)
!
!        B2_qq = B2_qq + 6*cf*phiquark1
!        if (xi < 1._dp) then
!            B2_qq = B2_qq + 8._dp*cf*phiquark1*log(1._dp-xi)
!        endif
!
!    end function
!
!    ! rest piece
!    function B1_qq_integrand(z)
!        implicit none
!        real(dp) :: B1_qq_integrand
!        real(dp), intent(in) :: z
!
!        real(dp) :: phiquark, phigluon, phiquark1
!        real(dp) :: x
!
!        x = xi/z
!        phiquark  = fdist_one(+1, x, mu, flavor)
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!        phigluon  = fdist_one(+1, x, mu, 0)
!
!        B1_qq_integrand = 0._dp
!        B1_qq_integrand = B1_qq_integrand + 4*tf*(1-z)*phigluon
!        B1_qq_integrand = B1_qq_integrand + CF*(2*(1-z))/z*phiquark
!
!    end function
!
!    function B2_qq_integrand(z)
!        implicit none
!        real(dp) :: B2_qq_integrand
!        real(dp), intent(in) :: z
!
!        real(dp) :: phiquark, phiquark1, phigluon
!        real(dp) :: x
!
!        x = xi/z
!        phiquark  = fdist_one(+1, x, mu, flavor)
!        phiquark1 = fdist_one(+1, xi, mu, flavor)
!        phigluon  = fdist_one(+1, x, mu, 0)
!
!        B2_qq_integrand = 4._dp*tf*(1._dp - 2._dp*z + 2._dp*z*z)*phigluon/z
!        if (z < 1._dp) then
!            B2_qq_integrand = B2_qq_integrand + &
!                4._dp*cf*((1._dp+z*z)*phiquark/z - 2._dp*phiquark1)/(1._dp-z)
!        endif
!
!    end function

end module
