!  Copyright (C) 2019-2021, respective authors of MCFM.
!
!  This program is free software: you can redistribute it and/or modify it under
!  the terms of the GNU General Public License as published by the Free Software
!  Foundation, either version 3 of the License, or (at your option) any later
!  version.
!
!  This program is distributed in the hope that it will be useful, but WITHOUT ANY
!  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
!  PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License along with
!  this program. If not, see <http://www.gnu.org/licenses/>
 
    module qtResummation
        use types
        use constants
        use iso_fortran_env
        use qtResummation_params
        implicit none

        public :: resummation
        public :: recoilBoost
        public :: gen_res

    private

    integer, save :: nf
!$omp threadprivate(nf)

    real(dp), save :: d2
!$omp threadprivate(d2)

    real(dp), save :: beta0
!$omp threadprivate(beta0)

    real(dp), save :: Gamma0, Gamma1
!$omp threadprivate(Gamma0,Gamma1)

    real(dp), save :: gammaq0, gammaq1, gammag0, gammag1
!$omp threadprivate(gammaq0,gammaq1,gammag0,gammag1)

    !! shared parameters with Fourier integrand
    integer, save :: order, nn
!$omp threadprivate(order,nn)
    real(dp), save :: qt, q2, mu, alphasMu
!$omp threadprivate(qt,q2,mu,alphasMu)

    ! parameter that defines whether this module is set up for
    ! a quark initiated process or gluon initiated
    logical, save :: initQuark = .true.
!$omp threadprivate(initQuark)

    contains

    subroutine update_nf_parameters(nf_in)
        implicit none
        integer, intent(in) :: nf_in

        if (nf /= nf_in) then
            ! update all nf dependent constants
            nf = nf_in

            d2 = CA*(808._dp/27._dp - 28._dp*zeta3) - 224._dp/27._dp*tf*nf

            beta0 = 11._dp/3._dp*CA - 4._dp/3._dp*TF*nf

            Gamma0 = 4._dp
            Gamma1 = 4._dp*((67._dp/9._dp - pi**2/3._dp)*CA - 20._dp/9._dp*TF*nf)

            gammaq0 = -3._dp*CF
            gammaq1 = CF*nf*(130._dp/27._dp + 2._dp*pi**2/3._dp)*TF  &
                    + CF**2*(-3._dp/2._dp + 2._dp*pi**2 - 24._dp*zeta3) &
                    + CA*CF*(-961._dp/54._dp - 11._dp*pi**2/6._dp + 26._dp*zeta3)

            gammag0 = (-11*CA)/3._dp + (4*nf*TF)/3._dp
            gammag1 = (-692*CA**2)/27._dp + (11*CA**2*Pi**2)/18._dp + (256*CA*nf*TF)/27._dp + &
                4*CF*nf*TF - (2*CA*nf*Pi**2*TF)/9._dp + 2*CA**2*zeta3
        endif

    end subroutine

    subroutine recoilBoost(pt,phi,p)
        implicit none
        include 'src/Inc/mxpart.f'
        include 'src/Inc/npart.f'
        real(dp), intent(in) :: pt,phi
        real(dp), intent(inout) :: p(mxpart,4)

        real(dp) :: qtx, qty
        real(dp) :: Lxx, Lxy, Lyy
        real(dp) :: energy, rt, btx, bty, gam
        integer :: j

        real(dp) :: boostxy(4,4)

        energy = - p(1,4) - p(2,4)
        qtx = pt*cos(phi)
        qty = pt*sin(phi)
        rt = sqrt(energy**2 + pt**2)
        btx = qtx/rt
        bty = qty/rt
        gam = 1._dp/sqrt(1._dp-btx**2-bty**2)

        Lxx = (gam-1._dp)*btx**2/(btx**2 + bty**2) + 1._dp
        Lyy = (gam-1._dp)*bty**2/(btx**2 + bty**2) + 1._dp
        Lxy = (gam-1._dp)*btx*bty/(btx**2 + bty**2)

        boostxy = reshape([ Lxx, Lxy, 0._dp, -btx*gam, &
                    Lxy, Lyy, 0._dp, -bty*gam, &
                    0._dp, 0._dp, 1._dp, 0._dp, &
                    -btx*gam, -bty*gam, 0._dp, gam ],[4,4])

        p(1,:) = matmul(boostxy, p(1,:))
        p(2,:) = matmul(boostxy, p(2,:))

        do j=3,npart+2
            p(j,:) = matmul(boostxy, p(j,:))
        enddo

    end subroutine

    function qstar(x)
        use LHAPDF, only: getalphas
        implicit none
        real(dp), intent(in) :: x
        real(dp) :: qstar

        real(dp) :: Ci

        if (initQuark) then
            Ci = CF
        else
            Ci = CA
        endif

        qstar = x**2 - q2 * exp(-pi/Ci/getalphas(x))

    end function

    subroutine resummation(q2_in, qt_in, x1, x2, f1, f2, hard1, muH, order_in, muMult, res, resexp)
        use LHAPDF
        use Beamfunctions
        use newton_raphson, only: find_root
        use qtResummationHardEvolution
        use qtResummationFourier
        implicit none
        real(dp), intent(in) :: q2_in, qt_in, x1, x2, muH
        integer, intent(in) :: f1, f2
        real(dp), intent(in) :: hard1
        integer, intent(in) :: order_in
        real(dp), intent(in), optional :: muMult
        real(dp), intent(out), optional :: res, resexp

        real(dp) :: muMatch
        real(dp) :: beamFunf1(0:3), beamFunf2(0:3)
        real(dp) :: beamFunf1_grid(0:3), beamFunf2_grid(0:3)
        real(dp) :: beamFunf1_exp(0:2), beamFunf2_exp(0:2)
        real(dp) :: Mi(0:4)
        real(dp) :: alphasMuH, alphasMuMatch
        real(dp) :: mub, muc
        real(dp) :: alphas4p, alphas4m, alphas3p, alphas3m
        real(dp) :: Ci, gammai0, gammai1, gamma1rx, LM, LMU

        real(dp) :: root, muStar
        logical :: ierr

        ! defined elsewhere
        real(dp) :: GetOnePDF

        character(len=*), parameter :: prefix = "Pdfdata/BeRn2017NLO"
        integer :: j
        real(dp) :: resnew
        real(dp) :: resexp1, resexp2

        integer :: ih1,ih2
        common/density/ih1,ih2

        if (present(res) .and. present(resexp)) then
            ! because of setting nf based on mu or muMatch
            write (*,*) "res and resexp simultaneously currently unsupported"
            error stop
        endif

        !real(dp) :: test

        order = order_in
        q2 = q2_in
        qt = qt_in

        if (f1 == 0 .and. f2 == 0) then
            initQuark = .false.
        else
            initQuark = .true.
        endif

        call find_root(qstar, 1._dp, 1.0e-3_dp, 100, root, ierr)
        if (ierr .eqv. .false.) then
            write (*,*) "WARNING: could not determine qstar, using default of 1.88"
            root = 1.88_dp
        endif
        muStar = root

        ! Fourier integral is numerically unstable below mu~2GeV
        if (present(muMult)) then
            mu = max(muMult*(qt + muStar*exp(-qt/muStar)), 3.0d0)
        else
            mu = max(qt + muStar*exp(-qt/muStar), 3.0d0)
        endif

        ! alpha at resummation scale, hard scale and matching scale
        alphasMu = getalphas(mu)
        alphasMuH = getalphas(muH)

        ! fixed to muH, otherwise we need to have evaluate the hard function also at muMatch
        muMatch = muH
        alphasMuMatch = alphasMuH

        ! update all nf constants in this module, including nf itself
        if (present(resexp)) then
            call update_nf_parameters(getnumflavors(muMatch))
        else
            call update_nf_parameters(getnumflavors(mu))
        endif
        ! update nf for beamfunctions

        ! PDFs in LHAPDF typically don't expose flavor thresholds,
        ! so we use the masses instead.
        ! Since the PDFs also use grid interpolation for alphas, the
        ! thresholds will be missing or washed out, but the code below
        ! will allow us to roll our own alphas running.
        if (nf < 5) then
            mub = getquarkMass(5)
            alphas4p = getalphas(mub + 1d-6)
            alphas4m = getalphas(mub - 1d-6)
        endif
        if (nf < 4) then
            muc = getquarkMass(4)
            alphas3p = getalphas(muc + 1d-6)
            alphas3m = getalphas(muc - 1d-6)
        endif

        if (present(res)) then

        Mi(0) = fourierM(0, order, nf, initQuark, qt, q2, mu, alphasMu)
        if (order >= 4) then
            Mi(1) = fourierM(1, order, nf, initQuark, qt, q2, mu, alphasMu)
            Mi(2) = fourierM(2, order, nf, initQuark, qt, q2, mu, alphasMu)
        endif
        if (order >= 6) then
            Mi(3) = fourierM(3, order, nf, initQuark, qt, q2, mu, alphasMu)
            Mi(4) = fourierM(4, order, nf, initQuark, qt, q2, mu, alphasMu)
        endif

        resnew = 0._dp
        resnew = Mi(0) * getbeam(ih1,f1,0,0,x1,mu,1) * getbeam(ih2,f2,0,0,x2,mu,2)

        if (order >= 4) then
            ! g
            resnew = resnew + Mi(1)*alphasMu/(4._dp*pi)*( &
                getbeam(ih1,f1,1,1,x1,mu,1) * getbeam(ih2,f2,0,0,x2,mu,2) + &
                getbeam(ih1,f1,0,0,x1,mu,1) * getbeam(ih2,f2,1,1,x2,mu,2) )
            ! g^2
            resnew = resnew + Mi(0)*alphasMu/(4._dp*pi)*( &
                getbeam(ih1,f1,1,0,x1,mu,1) * getbeam(ih2,f2,0,0,x2,mu,2) + &
                getbeam(ih1,f1,0,0,x1,mu,1) * getbeam(ih2,f2,1,0,x2,mu,2) )
            resnew = resnew + Mi(2)*(alphasMu/(4._dp*pi))**2*( &
                getbeam(ih1,f1,2,2,x1,mu,1) * getbeam(ih2,f2,0,0,x2,mu,2) + &
                getbeam(ih1,f1,0,0,x1,mu,1) * getbeam(ih2,f2,2,2,x2,mu,2) + &
                getbeam(ih1,f1,1,1,x1,mu,1) * getbeam(ih2,f2,1,1,x2,mu,2) )
        endif

        if (order >= 6) then
            ! g^3
            resnew = resnew + (alphasMu/4._dp/pi)**2*Mi(1)*( &
                getbeam(ih1,f1,2,1,x1,mu)*getbeam(ih2,f2,0,0,x2,mu) + &
                getbeam(ih1,f1,1,1,x1,mu)*getbeam(ih2,f2,1,0,x2,mu) + &
                getbeam(ih1,f1,1,0,x1,mu)*getbeam(ih2,f2,1,1,x2,mu) + &
                getbeam(ih1,f1,0,0,x1,mu)*getbeam(ih2,f2,2,1,x2,mu))
            resnew = resnew + (alphasMu/4._dp/pi)**3*Mi(3)*( &
                getbeam(ih1,f1,2,2,x1,mu)*getbeam(ih2,f2,1,1,x2,mu) + &
                getbeam(ih1,f1,1,1,x1,mu)*getbeam(ih2,f2,2,2,x2,mu))
            ! g^4
            resnew = resnew + (alphasMu/4._dp/pi)**2*Mi(0)*( &
                getbeam(ih1,f1,2,0,x1,mu)*getbeam(ih2,f2,0,0,x2,mu) + &
                getbeam(ih1,f1,1,0,x1,mu)*getbeam(ih2,f2,1,0,x2,mu) + &
                getbeam(ih1,f1,0,0,x1,mu)*getbeam(ih2,f2,2,0,x2,mu))
            resnew = resnew + (alphasMu/4._dp/pi)**4*Mi(4)* &
                getbeam(ih1,f1,2,2,x1,mu)*getbeam(ih2,f2,2,2,x2,mu)
            resnew = resnew + (alphasMu/4._dp/pi)**3*Mi(2)*( &
                getbeam(ih1,f1,2,2,x1,mu)*getbeam(ih2,f2,1,0,x2,mu) + &
                getbeam(ih1,f1,2,1,x1,mu)*getbeam(ih2,f2,1,1,x2,mu) + &
                getbeam(ih1,f1,1,1,x1,mu)*getbeam(ih2,f2,2,1,x2,mu) + &
                getbeam(ih1,f1,1,0,x1,mu)*getbeam(ih2,f2,2,2,x2,mu))

            ! additional piece unique to gluon-gluon
            if (.not. initQuark) then
                resnew = resnew + (alphasMu/4._dp/pi)**2*Mi(0)*( &
                    getbeam2(ih1,f1,1,0,x1,mu)*getbeam2(ih2,f2,1,0,x2,mu) )
            endif
        endif

        res = hardEvolutionThres(q2,muH**2,alphasMuH,alphasMu,order,initQuark, &
            nf, mub, alphas4p, alphas4m, muc, alphas3p, alphas3m)*resnew

        endif

        if (present(resexp)) then

        if (initQuark) then
            Ci = CF
            gammai0 = gammaq0
            gammai1 = gammaq1
            gamma1rx = -d2*CF
        else
            Ci = CA
            gammai0 = gammag0
            gammai1 = gammag1
            gamma1rx = -d2*CA
        endif

        LM = log(q2/qt**2)
        LMU = log(muMatch**2/qt**2)

        resexp = 0._dp
        resexp1 = 0._dp
        resexp2 = 0._dp

        resexp1 = &
            (alphasMuMatch*Ci*Gamma0*LM*getbeam(ih1,f1,0,0,x1,muMatch,1)* &
               getbeam(ih2,f2,0,0,x2,muMatch,2))/(4._dp*pi*qt**2)

        if (order >= 4) then
            resexp1 = resexp1 + &
                (alphasMuMatch*gammai0*getbeam(ih1,f1,0,0,x1,muMatch,1)* &
                   getbeam(ih2,f2,0,0,x2,muMatch,2))/(2._dp*pi*qt**2) - &
                (alphasMuMatch*getbeam(ih1,f1,1,1,x1,muMatch,1)* &
                   getbeam(ih2,f2,0,0,x2,muMatch,2))/(4._dp*pi*qt**2) - &
                (alphasMuMatch*getbeam(ih1,f1,0,0,x1,muMatch,1)* &
                   getbeam(ih2,f2,1,1,x2,muMatch,2))/(4._dp*pi*qt**2)
        endif

        if (order >= 6) then
            resexp2 = -((gamma1rX*getbeam(ih1,f1,0,0,x1,muMatch)* &
             getbeam(ih2,f2,0,0,x2,muMatch))/qt**2) +  &
        (2*gammai1*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (2*gammai0*hard1*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma1*LM*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma0*hard1*LM*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (4*gammai0**2*LMU*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (4*Ci*Gamma0*gammai0*LM*LMU*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (Ci**2*Gamma0**2*LM**2*LMU*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma0*gammai0*LMU**2*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci**2*Gamma0**2*LM*LMU**2*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/(2._dp*qt**2) +  &
        (2*Ci**2*Gamma0**2*Zeta3*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (2*gammai0*LMU*Beta0*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma0*LM*LMU*Beta0*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (2*gammai0*getbeam(ih1,f1,1,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma0*LM*getbeam(ih1,f1,1,0,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (hard1*getbeam(ih1,f1,1,1,x1,muMatch)*getbeam(ih2,f2,0,0,x2,muMatch))/ &
         qt**2 + (4*gammai0*LMU*getbeam(ih1,f1,1,1,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 +  &
        (2*Ci*Gamma0*LM*LMU*getbeam(ih1,f1,1,1,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (Ci*Gamma0*LMU**2*getbeam(ih1,f1,1,1,x1,muMatch)* &
           getbeam(ih2,f2,0,0,x2,muMatch))/(2._dp*qt**2) -  &
        (getbeam(ih1,f1,2,1,x1,muMatch)*getbeam(ih2,f2,0,0,x2,muMatch))/qt**2 -  &
        (2*LMU*getbeam(ih1,f1,2,2,x1,muMatch)*getbeam(ih2,f2,0,0,x2,muMatch))/ &
         qt**2 + (2*gammai0*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,1,0,x2,muMatch))/qt**2 +  &
        (Ci*Gamma0*LM*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,1,0,x2,muMatch))/qt**2 -  &
        (getbeam(ih1,f1,1,1,x1,muMatch)*getbeam(ih2,f2,1,0,x2,muMatch))/qt**2 -  &
        (hard1*getbeam(ih1,f1,0,0,x1,muMatch)*getbeam(ih2,f2,1,1,x2,muMatch))/ &
         qt**2 + (4*gammai0*LMU*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,1,1,x2,muMatch))/qt**2 +  &
        (2*Ci*Gamma0*LM*LMU*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,1,1,x2,muMatch))/qt**2 -  &
        (Ci*Gamma0*LMU**2*getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,1,1,x2,muMatch))/(2._dp*qt**2) -  &
        (getbeam(ih1,f1,1,0,x1,muMatch)*getbeam(ih2,f2,1,1,x2,muMatch))/qt**2 -  &
        (2*LMU*getbeam(ih1,f1,1,1,x1,muMatch)*getbeam(ih2,f2,1,1,x2,muMatch))/ &
         qt**2 - (getbeam(ih1,f1,0,0,x1,muMatch)* &
           getbeam(ih2,f2,2,1,x2,muMatch))/qt**2 -  &
        (2*LMU*getbeam(ih1,f1,0,0,x1,muMatch)*getbeam(ih2,f2,2,2,x2,muMatch))/ &
         qt**2
        endif

        resexp = resexp1 + resexp2*(alphasMuMatch/4._dp/pi)**2

        endif

    end subroutine

    function gen_res(r,qt,p,wt, y_in)
      use types
      implicit none
      include 'src/Inc/constants.f'
      include 'src/Inc/mxpart.f'
      include 'src/Inc/mxdim.f'
      include 'src/Inc/breit.f'
      include 'src/Inc/limits.f'
      include 'src/Inc/energy.f'
      include 'src/Inc/x1x2.f'

      logical :: gen_res
      real(dp), intent(in) :: r(mxdim)
      real(dp), intent(in) :: qt
      real(dp), intent(out) :: p(mxpart,4)
      real(dp), intent(out) :: wt
      real(dp), optional, intent(in) :: y_in

      real(dp), parameter :: wt0 = 1._dp/2._dp/pi
      real(dp) :: Qsqmin, Qsqmax, wtbw, wt12, wt34
      real(dp) :: tau, Q(4), Qsq
      real(dp) :: y, ymin, ymax, xa, xb

      real(dp) :: massvec

      p(:,:) = 0._dp

      Qsqmin = max(wsqmin,1._dp)
      Qsqmax = min(wsqmax, sqrts**2*0.999999_dp)

      call breitw(r(1), Qsqmin, Qsqmax, mass3, width3, Qsq, wtbw)

      ! one approximation
      tau = (Qsq + qt**2)/sqrts**2

      ! "better" approximation
      !tau = ((sqrts**2 + Qsq) - sqrt((Qsq-sqrts**2)**2 - 4._dp * qt**2 * sqrts**2)) &
                !/ (2._dp * sqrts * sqrt(Qsq + qt**2))
      !tau = tau**2

      ymin = log(sqrt(tau))
      ymax = -ymin
      if (present(y_in)) then 
          y = y_in
          wt12 = 1._dp
      else
          y = ymin + r(2)*(ymax-ymin)
          wt12 = ymax-ymin
      endif
      xx(1) = min(sqrt(tau)*exp(y), 1._dp)
      xx(2) = min(sqrt(tau)*exp(-y), 1._dp)

      p(1,4) = 0.5_dp*sqrts*xx(1)
      p(1,3) = p(1,4)
      p(2,4) = 0.5_dp*sqrts*xx(2)
      p(2,3) = -p(2,4)

      wt12 = 2._dp*pi*wt12/sqrts**2

      Q = p(1,:) + p(2,:)
      call phi3m0(r(3), r(4), Q, p(3,:), p(4,:), wt34, *99)

      p(1,:) = -p(1,:)
      p(2,:) = -p(2,:)


      wt = wt0*wt12*wtbw*wt34

      ! for debugging
      !wt = ymax-ymin

      gen_res = .true.

      return

 99   continue
      wt = 0
      gen_res = .false.

    end function

end module
