!  Copyright (C) 2019-2021, respective authors of MCFM.
!
!  This program is free software: you can redistribute it and/or modify it under
!  the terms of the GNU General Public License as published by the Free Software
!  Foundation, either version 3 of the License, or (at your option) any later
!  version.
!
!  This program is distributed in the hope that it will be useful, but WITHOUT ANY
!  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
!  PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License along with
!  this program. If not, see <http://www.gnu.org/licenses/>
 
******************************************************************************
**  tdhpl:  a subroutine for the evaluation of 
**          two-dimensional harmonic polylogarithms
**          Version 1.0         09/11/2001
**  described in: 
**  T.Gehrmann and E.Remiddi: Numerical Evaluation of Two-Dimensional
**                            Harmonic Polylogarithms 
**                            (hep-ph/0111255; CERN-TH/2001/326)
**  the harmonic polylogarithms are defined in: 
**  E.Remiddi and J.Vermaseren: Harmonic Polylogarithms
**                            (hep-ph/9905237; Int.J.Mod.Phys. A15 (2000) 725)
**  and:
**  T.Gehrmann and E.Remiddi: Two-Loop Master Integrals for gamma*->3jets:
**                            the Planar Topologies 
**                            (hep-ph/0008287; Nucl.Phys. B601(2001) 248)
**  this subroutine invokes the subroutine hplog, which is described in:
**  T.Gehrmann and E.Remiddi: Numerical Evaluation of 
**                            Harmonic Polylogarithms 
**                            (hep-ph/0107173; Comp.Phys.Comm. 141(2001) 296) 
**  email:
**  Thomas.Gehrmann@cern.ch and Ettore.Remiddi@bo.infn.it
**
******************************************************************************
      subroutine tdhpl(y,z,nmax,GYZ1,GYZ2,GYZ3,GYZ4,
     $                          HZ1,HZ2,HZ3,HZ4)
*********************************************************************
***** Input:                                                    *****
***** y and z are the arguments of the 2dHPL to be evaluated    *****
***** nmax    is the maximum weight to be evaluated             *****
***** Output:                                                   *****
***** GYZ1,GYZ2,GYZ3,GYZ4 are the double precision 2dHPL of y,z *****
***** HZ1,HZ2,HZ3,HZ4     are the double precision 1dHPL of z   *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer i1,i2,i3,i4
      integer sgi1,sgi2,sgi3,sgi4
      integer iflag,nmax
      real(dp):: x,y,z
      real(dp):: zold
      real(dp):: HZ1,HZ2,HZ3,HZ4,GYZ1,GYZ2,GYZ3,GYZ4
      real(dp):: HZ1a,HZ2a,HZ3a,HZ4a,HYZ1a,HYZ2a,HYZ3a,HYZ4a
      dimension HZ1(0:1),HZ2(0:1,0:1),HZ3(0:1,0:1,0:1), 
     $          HZ4(0:1,0:1,0:1,0:1) 
      dimension GYZ1(0:3),GYZ2(0:3,0:3),GYZ3(0:3,0:3,0:3), 
     $          GYZ4(0:3,0:3,0:3,0:3) 
      common/HPL2OUT/
     $ HZ1a(0:1),HZ2a(0:1,0:1),HZ3a(0:1,0:1,0:1), 
     $           HZ4a(0:1,0:1,0:1,0:1), 
     $ HYZ1a(0:3),HYZ2a(0:3,0:3),HYZ3a(0:3,0:3,0:3), 
     $            HYZ4a(0:3,0:3,0:3,0:3)       
      data zold/ -999d0/
      save zold
!$omp threadprivate(zold,/hpl2out/)


      x=1d0-y-z
      if (x.lt.0d0.or.x.gt.1.d0) then 
         write(6,991) y,z
         stop
      endif
      if (y.lt.0d0.or.y.gt.1.d0) then 
         write(6,991) y,z
         stop
      endif
      if (z.lt.0d0.or.z.gt.1.d0) then 
         write(6,991) y,z
         stop
      endif
      if (nmax.lt.1.or.nmax.gt.4) then 
         write(6,992) nmax
         stop
      endif
      
      iflag=1
      if (z.eq.zold) iflag=2
      zold = z
*      write(6,*) iflag
      call fill2dhpl(iflag,nmax,y,z)

      do i1=0,1
         if (nmax.ge.2) then
         do i2=0,1
            if (nmax.ge.3) then
            do i3=0,1
               if (nmax.eq.4) then
               do i4=0,1
                  HZ4(i1,i2,i3,i4) = HZ4a(i1,i2,i3,i4)
               enddo
               endif
               HZ3(i1,i2,i3) = HZ3a(i1,i2,i3)
            enddo
            endif
            HZ2(i1,i2) = HZ2a(i1,i2)
         enddo
         endif
         HZ1(i1) = HZ1a(i1)
      enddo
      do i1=0,3
         if (i1.eq.1.or.i1.eq.2) then
            sgi1=-1
         else
            sgi1=1
         endif
         if (nmax.ge.2) then
         do i2=0,3
            if (i2.eq.1.or.i2.eq.2) then
               sgi2=-1
            else
               sgi2=1
            endif
            if (nmax.ge.3) then
            do i3=0,3
               if (i3.eq.1.or.i3.eq.2) then
                  sgi3=-1
               else
                  sgi3=1
               endif
               if (nmax.eq.4) then
               do i4=0,3
                  if (i4.eq.1.or.i4.eq.2) then
                     sgi4=-1
                  else
                     sgi4=1
                  endif
                  GYZ4(i1,i2,i3,i4) = sgi1*sgi2*sgi3*sgi4*
     $                                    HYZ4a(i1,i2,i3,i4)
               enddo
               endif
               GYZ3(i1,i2,i3) = sgi1*sgi2*sgi3*HYZ3a(i1,i2,i3)
            enddo
            endif
            GYZ2(i1,i2) = sgi1*sgi2*HYZ2a(i1,i2)
         enddo
         endif
         GYZ1(i1) = sgi1*HYZ1a(i1)
      enddo
      return
 991  format('Parameters out of range, y=',f9.5,', z=',f9.5)
 992  format('Weight out of range, nmax=',i3)
      end


      subroutine fill2dhpl(iflag,nmax,y,z)
*********************************************************************
***** fill2dhpl evaluates all 2dhpl writes all                  ***** 
***** 2dhpl(y) and 1dhpl(z)                                     ***** 
***** up to weight nmax into HPL2OUT                            ***** 
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      integer i1,i2,i3,i4
      real(dp):: y,z,xi
      real(dp):: HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HZE1,HZE2,HZE3,HZE4
      dimension HZE1(-1:1),HZE2(-1:1,-1:1),HZE3(-1:1,-1:1,-1:1),
     $          HZE4(-1:1,-1:1,-1:1,-1:1)
      common/HPL2OUT/
     $ HZ1(0:1),HZ2(0:1,0:1),HZ3(0:1,0:1,0:1), 
     $          HZ4(0:1,0:1,0:1,0:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      save HZE1,HZE2,HZE3,HZE4
!$omp threadprivate(HZE1,HZE2,HZE3,HZE4,/HPL2OUT/)
      if (iflag.eq.1) then
      call get1dhplog0m11(iflag,nmax,z,HZE1,HZE2,HZE3,HZE4)
      endif
      do i1=0,1
         if (nmax.ge.2) then
         do i2=0,1
            if (nmax.ge.3) then
            do i3=0,1
               if (nmax.eq.4) then
               do i4=0,1
                  HZ4(i1,i2,i3,i4) = HZE4(i1,i2,i3,i4)
               enddo
               endif
               HZ3(i1,i2,i3) = HZE3(i1,i2,i3)
            enddo
            endif
            HZ2(i1,i2) = HZE2(i1,i2)
         enddo
         endif
         HZ1(i1) = HZE1(i1)
      enddo
      xi = y/(1d0-z)
      if (xi.le.0.5d0) then
         call get2dhplat0(iflag,nmax,y,z,HYZ1,HYZ2,HYZ3,HYZ4,
     $                                   HZE1,HZE2,HZE3,HZE4)
*         write(6,*) 'call get2dhplat0'
      else
         call get2dhplat1(iflag,nmax,y,z,HYZ1,HYZ2,HYZ3,HYZ4,
     $                                   HZE1,HZE2,HZE3,HZE4)
*         write(6,*) 'call get2dhplat1'
      endif

      return
      end

      subroutine get1dhplog01(iflag,nw,z,HZ1,HZ2,HZ3,HZ4)
*********************************************************************
***** get1dhplog01 returns the 1dhpl (0,1) of z                 *****
***** up to weight nw                                           *****
***** evaluation using hplog                                    *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer n1,n2,nw,i1,i2,iflag
      parameter (n1=0) 
      parameter (n2=1) 
      complex(dp):: Hc1,Hc2,Hc3,Hc4 
      real(dp)::     HZ1,HZ2,HZ3,HZ4 
      real(dp)::     Hi1,Hi2,Hi3,Hi4 
      real(dp)::     z     
      dimension Hc1(n1:n2),Hc2(n1:n2,n1:n2),Hc3(n1:n2,n1:n2,n1:n2), 
     $          Hc4(n1:n2,n1:n2,n1:n2,n1:n2) 
      dimension HZ1(n1:n2),HZ2(n1:n2,n1:n2),HZ3(n1:n2,n1:n2,n1:n2), 
     $          HZ4(n1:n2,n1:n2,n1:n2,n1:n2) 
      dimension Hi1(n1:n2),Hi2(n1:n2,n1:n2),Hi3(n1:n2,n1:n2,n1:n2), 
     $          Hi4(n1:n2,n1:n2,n1:n2,n1:n2) 
      call hplog(z,nw,Hc1,Hc2,Hc3,Hc4, 
     $                HZ1,HZ2,HZ3,HZ4,Hi1,Hi2,Hi3,Hi4,n1,n2) 
      return 
      end


      subroutine get1dhplog0m11(iflag,nw,z,HZ1,HZ2,HZ3,HZ4)
*********************************************************************
***** get1dhplog0m11 returns the 1dhpl (-1,0,1) of z            *****
***** up to weight nw                                           *****
***** evaluation using hplog                                    *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer n1,n2,nw,i1,i2,iflag
      parameter (n1=-1) 
      parameter (n2=1) 
      complex(dp):: Hc1,Hc2,Hc3,Hc4 
      real(dp)::     HZ1,HZ2,HZ3,HZ4 
      real(dp)::     Hi1,Hi2,Hi3,Hi4 
      real(dp)::     z     
      dimension Hc1(n1:n2),Hc2(n1:n2,n1:n2),Hc3(n1:n2,n1:n2,n1:n2), 
     $          Hc4(n1:n2,n1:n2,n1:n2,n1:n2) 
      dimension HZ1(n1:n2),HZ2(n1:n2,n1:n2),HZ3(n1:n2,n1:n2,n1:n2), 
     $          HZ4(n1:n2,n1:n2,n1:n2,n1:n2) 
      dimension Hi1(n1:n2),Hi2(n1:n2,n1:n2),Hi3(n1:n2,n1:n2,n1:n2), 
     $          Hi4(n1:n2,n1:n2,n1:n2,n1:n2) 
      call hplog(z,nw,Hc1,Hc2,Hc3,Hc4, 
     $                HZ1,HZ2,HZ3,HZ4,Hi1,Hi2,Hi3,Hi4,n1,n2) 
      return 
      end


      subroutine get2dhplat1(iflag,nmax,y,z,HYZ11,HYZ12,HYZ13,HYZ14
     $                       ,HZE1,HZE2,HZE3,HZE4)
*********************************************************************
***** get2dhplat0 evaluates all 2dhpl from expansions           *****
***** for small 1-y-z=x <(1-z)/2                                *****
***** and subsequent transformation                             *****
***** up to weight nmax                                         *****
***** HZEn must be supplied (needed in cut-separation)          *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      real(dp):: x,y,z
      real(dp):: HZ1,HZ2,HZ3,HZ4,HYZ11,HYZ12,HYZ13,HYZ14
      real(dp):: HZE1,HZE2,HZE3,HZE4,HXZ01,HXZ02,HXZ03,HXZ04
      dimension HYZ11(0:3),HYZ12(0:3,0:3),HYZ13(0:3,0:3,0:3),
     $          HYZ14(0:3,0:3,0:3,0:3)
      dimension HXZ01(0:3),HXZ02(0:3,0:3),HXZ03(0:3,0:3,0:3),
     $          HXZ04(0:3,0:3,0:3,0:3)
      dimension HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1),
     $          HZ4(-1:1,-1:1,-1:1,-1:1)
      dimension HZE1(-1:1),HZE2(-1:1,-1:1),HZE3(-1:1,-1:1,-1:1),
     $          HZE4(-1:1,-1:1,-1:1,-1:1)

      x=1d0-y-z
      if (x.ge.y) then
         write(6,*) 'Error  1-y-z> y for eval. at 1'
         stop
      endif
      call get2dhplat0(iflag,nmax,x,z,HXZ01,HXZ02,HXZ03,HXZ04,
     $                                HZE1,HZE2,HZE3,HZE4)
      call swap2dhplxy(iflag,nmax,HXZ01,HXZ02,HXZ03,HXZ04,
     $                 HZE1,HZE2,HZE3,HZE4,HYZ11,HYZ12,HYZ13,HYZ14)
      return 
      end


      subroutine get2dhplat0(iflag,nmax,y,z,HYZ01,HYZ02,HYZ03,HYZ04,
     $                                      HZE1,HZE2,HZE3,HZE4)
*********************************************************************
***** get2dhplat0 evaluates all 2dhpl from expansions           *****
***** for small y<(1-z)/2                                       *****
***** up to weight nmax                                         *****
***** HZEn must be supplied (needed in cut-separation)          *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax,n
      integer i1,i2,i3,i4
      real(dp):: y,z,xi
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HYZ01,HYZ02,HYZ03,HYZ04,HZE1,HZE2,HZE3,HZE4
      dimension HYZ01(0:3),HYZ02(0:3,0:3),HYZ03(0:3,0:3,0:3),
     $          HYZ04(0:3,0:3,0:3,0:3)
      dimension HZE1(-1:1),HZE2(-1:1,-1:1),HZE3(-1:1,-1:1,-1:1),
     $          HZE4(-1:1,-1:1,-1:1,-1:1)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
!$omp threadprivate(/HPL2/)

      do i1=-1,1
         if (nmax.ge.2) then
         do i2=-1,1
            if (nmax.ge.3) then
            do i3=-1,1
               if (nmax.eq.4) then
               do i4=-1,1
                  HZ4(i1,i2,i3,i4) = HZE4(i1,i2,i3,i4)
               enddo
               endif
               HZ3(i1,i2,i3) = HZE3(i1,i2,i3)
            enddo
            endif
            HZ2(i1,i2) = HZE2(i1,i2)
         enddo
         endif
         HZ1(i1) = HZE1(i1)
      enddo

      xi = y/(1d0-z) 
      if (xi.gt.0.5d0) then
         write(6,*) 'Error y/(1-z) > 0.5 for eval. at 0'
         stop
      endif

      call fillirr2dhpl10(iflag,nmax,y,z)
      call fillirr2dhpl20(iflag,nmax,y,z)
      call fillirr2dhpl30(iflag,nmax,y,z)
      call makeexpar(y,z)

      do n=2,nmax
         call fillirr2dhpl210(iflag,n)
         call fillirr2dhpl310(iflag,n)
         if (z.le.0.5d0) then
            call fillirr2dhpl320(iflag,n)
            call fillirr2dhpl321(iflag,n)
         else
            call fillirr2dhpl320e(iflag,n)
            call fillirr2dhpl321e(iflag,n)
         endif
         call fillred2dhpl(iflag,n)
      enddo
      do i1=0,3
         if (nmax.ge.2) then
         do i2=0,3
            if (nmax.ge.3) then
            do i3=0,3
               if (nmax.eq.4) then
               do i4=0,3
                  HYZ04(i1,i2,i3,i4) = HYZ4(i1,i2,i3,i4)
               enddo
               endif
               HYZ03(i1,i2,i3) = HYZ3(i1,i2,i3)
            enddo
            endif
            HYZ02(i1,i2) = HYZ2(i1,i2)
         enddo
         endif
         HYZ01(i1) = HYZ1(i1)
      enddo
      return
      end


      subroutine fillirr2dhpl10(iflag,nmax,y,z)
*********************************************************************
***** fillirr2dhpl10 fills the irreducible 2dHPL                *****
***** with indices (1,0) up to weight n                         *****
***** by evaluating 1dHPL in y                                  *****
***** applicable for all z                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      integer i1,i2,i3,i4
      real(dp):: y,z
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HY1R,HY2R,HY3R,HY4R
      dimension HY1R(0:1),HY2R(0:1,0:1),HY3R(0:1,0:1,0:1),
     $          HY4R(0:1,0:1,0:1,0:1)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
!$omp threadprivate(/HPL2/)

      call get1dhplog01(iflag,nmax,y,HY1R,HY2R,HY3R,HY4R)
      do i1=0,1
         if (nmax.ge.2) then
         do i2=0,1
            if (nmax.ge.3) then
            do i3=0,1
               if (nmax.eq.4) then
               do i4=0,1
                  HYZ4(i1,i2,i3,i4) = HY4R(i1,i2,i3,i4)
                  HY4(i1,i2,i3,i4) = HY4R(i1,i2,i3,i4)
               enddo
               endif
               HYZ3(i1,i2,i3) = HY3R(i1,i2,i3)
               HY3(i1,i2,i3) = HY3R(i1,i2,i3)
            enddo
            endif
            HYZ2(i1,i2) = HY2R(i1,i2)
            HY2(i1,i2) = HY2R(i1,i2)
         enddo
         endif
         HYZ1(i1) = HY1R(i1)
         HY1(i1) = HY1R(i1)
      enddo
      return
      end


      subroutine fillirr2dhpl20(iflag,nmax,y,z)
*********************************************************************
***** fillirr2dhpl20 fills the irreducible 2dHPL                *****
***** with indices (2,0) up to weight n                         *****
***** by evaluating 1dHPL in xi=y/(1-z)                         *****
***** applicable for all z                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      integer i1,i2,i3,i4
      real(dp):: y,z,xi
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HXI1,HXI2,HXI3,HXI4
      dimension HXI1(0:1),HXI2(0:1,0:1),HXI3(0:1,0:1,0:1),
     $          HXI4(0:1,0:1,0:1,0:1)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
!$omp threadprivate(/HPL2/)

      xi = y/(1-z)      
      call get1dhplog01(iflag,nmax,xi,HXI1,HXI2,HXI3,HXI4)
      i1=1
      if (nmax.ge.2) then
      do i2=0,1
         if (nmax.ge.3) then
         do i3=0,1
            if (nmax.eq.4) then
            do i4=0,1
               HYZ4(2*i4,2*i3,2*i2,2*i1) = HXI4(i4,i3,i2,i1)
            enddo
            endif
            HYZ3(2*i3,2*i2,2*i1) = HXI3(i3,i2,i1)
         enddo
         endif
         HYZ2(2*i2,2*i1) = HXI2(i2,i1)
      enddo
      endif
      HYZ1(2*i1) = HXI1(i1)
      return
      end


      subroutine fillirr2dhpl30(iflag,nmax,y,z)
*********************************************************************
***** fillirr2dhpl30 fills the irreducible 2dHPL                *****
***** with indices (3,0) up to weight n                         *****
***** by evaluating 1dHPL in xi=-y/z                            *****
***** applicable for all z                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      integer i1,i2,i3,i4
      real(dp):: y,z,xi
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HXI1,HXI2,HXI3,HXI4
      dimension HXI1(0:1),HXI2(0:1,0:1),HXI3(0:1,0:1,0:1),
     $          HXI4(0:1,0:1,0:1,0:1)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
!$omp threadprivate(/HPL2/)

      xi = -y/z      
      call get1dhplog01(iflag,nmax,xi,HXI1,HXI2,HXI3,HXI4)
      i1=1
      if (nmax.ge.2) then
      do i2=0,1
         if (nmax.ge.3) then
         do i3=0,1
            if (nmax.eq.4) then
            do i4=0,1
                  HYZ4(3*i4,3*i3,3*i2,3*i1) = (-1)**(i4+i3+i2+i1)
     $                                         *HXI4(i4,i3,i2,i1)
            enddo
            endif
            HYZ3(3*i3,3*i2,3*i1) = (-1)**(i3+i2+i1)
     $                              *HXI3(i3,i2,i1)
         enddo
         endif
         HYZ2(3*i2,3*i1) = (-1)**(i2+i1)*HXI2(i2,i1)
      enddo
      endif
      HYZ1(3*i1) = (-1)**i1*HXI1(i1)
      return
      end


      subroutine fillirr2dhpl210(iflag,n)
*********************************************************************
***** fillirr2dhpl210(iflag,n) fills the irreducible 2dHPL      *****
***** with indices (2,1,0)                                      *****
***** up to weight n using the EXPANDED expressions for the     *****
***** z-dependent expansion coefficients                        *****
***** applicable for all z                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsHYZ2,CsHYZ3,CsHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      dimension CsHYZ2(0:3,0:3,22),CsHYZ3(0:3,0:3,0:3,22),
     $          CsHYZ4(0:3,0:3,0:3,0:3,22)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      common /s/
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      save CsHYZ2,CsHYZ3,CsHYZ4
!$omp threadprivate(CsHYZ2,CsHYZ3,CsHYZ4,/s/,/HPL2/)

      if (iflag.eq.1) then
      call fillcoeff2dhpl210(iflag,n,CsHYZ2,CsHYZ3,CsHYZ4)
      endif
* 2001-04-03:16:14:02.hpl 
* <- HPLexpand.210t 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      HYZ2(2,1) = 
     $  + s02 * CsHYZ2(2,1,02) 
     $  + s03 * CsHYZ2(2,1,03) 
     $  + s04 * CsHYZ2(2,1,04) 
     $  + s05 * CsHYZ2(2,1,05) 
     $  + s06 * CsHYZ2(2,1,06) 
     $  + s07 * CsHYZ2(2,1,07) 
     $  + s08 * CsHYZ2(2,1,08) 
     $  + s09 * CsHYZ2(2,1,09) 
     $  + s10 * CsHYZ2(2,1,10) 
     $  + s11 * CsHYZ2(2,1,11) 
     $  + s12 * CsHYZ2(2,1,12) 
     $  + s13 * CsHYZ2(2,1,13) 
     $  + s14 * CsHYZ2(2,1,14) 
     $  + s15 * CsHYZ2(2,1,15) 
     $  + s16 * CsHYZ2(2,1,16) 
     $  + s17 * CsHYZ2(2,1,17) 
     $  + s18 * CsHYZ2(2,1,18) 
     $  + s19 * CsHYZ2(2,1,19) 
     $  + s20 * CsHYZ2(2,1,20) 
     $  + s21 * CsHYZ2(2,1,21) 
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(0,2,1) = 
     $  + s02 * CsHYZ3(0,2,1,02) 
     $  + s03 * CsHYZ3(0,2,1,03) 
     $  + s04 * CsHYZ3(0,2,1,04) 
     $  + s05 * CsHYZ3(0,2,1,05) 
     $  + s06 * CsHYZ3(0,2,1,06) 
     $  + s07 * CsHYZ3(0,2,1,07) 
     $  + s08 * CsHYZ3(0,2,1,08) 
     $  + s09 * CsHYZ3(0,2,1,09) 
     $  + s10 * CsHYZ3(0,2,1,10) 
     $  + s11 * CsHYZ3(0,2,1,11) 
     $  + s12 * CsHYZ3(0,2,1,12) 
     $  + s13 * CsHYZ3(0,2,1,13) 
     $  + s14 * CsHYZ3(0,2,1,14) 
     $  + s15 * CsHYZ3(0,2,1,15) 
     $  + s16 * CsHYZ3(0,2,1,16) 
     $  + s17 * CsHYZ3(0,2,1,17) 
     $  + s18 * CsHYZ3(0,2,1,18) 
     $  + s19 * CsHYZ3(0,2,1,19) 
     $  + s20 * CsHYZ3(0,2,1,20) 
      HYZ3(2,0,1) = 
     $  + s02 * CsHYZ3(2,0,1,02) 
     $  + s03 * CsHYZ3(2,0,1,03) 
     $  + s04 * CsHYZ3(2,0,1,04) 
     $  + s05 * CsHYZ3(2,0,1,05) 
     $  + s06 * CsHYZ3(2,0,1,06) 
     $  + s07 * CsHYZ3(2,0,1,07) 
     $  + s08 * CsHYZ3(2,0,1,08) 
     $  + s09 * CsHYZ3(2,0,1,09) 
     $  + s10 * CsHYZ3(2,0,1,10) 
     $  + s11 * CsHYZ3(2,0,1,11) 
     $  + s12 * CsHYZ3(2,0,1,12) 
     $  + s13 * CsHYZ3(2,0,1,13) 
     $  + s14 * CsHYZ3(2,0,1,14) 
     $  + s15 * CsHYZ3(2,0,1,15) 
     $  + s16 * CsHYZ3(2,0,1,16) 
     $  + s17 * CsHYZ3(2,0,1,17) 
     $  + s18 * CsHYZ3(2,0,1,18) 
     $  + s19 * CsHYZ3(2,0,1,19) 
     $  + s20 * CsHYZ3(2,0,1,20) 
      HYZ3(2,1,1) = 
     $  + s03 * CsHYZ3(2,1,1,03) 
     $  + s04 * CsHYZ3(2,1,1,04) 
     $  + s05 * CsHYZ3(2,1,1,05) 
     $  + s06 * CsHYZ3(2,1,1,06) 
     $  + s07 * CsHYZ3(2,1,1,07) 
     $  + s08 * CsHYZ3(2,1,1,08) 
     $  + s09 * CsHYZ3(2,1,1,09) 
     $  + s10 * CsHYZ3(2,1,1,10) 
     $  + s11 * CsHYZ3(2,1,1,11) 
     $  + s12 * CsHYZ3(2,1,1,12) 
     $  + s13 * CsHYZ3(2,1,1,13) 
     $  + s14 * CsHYZ3(2,1,1,14) 
     $  + s15 * CsHYZ3(2,1,1,15) 
     $  + s16 * CsHYZ3(2,1,1,16) 
     $  + s17 * CsHYZ3(2,1,1,17) 
     $  + s18 * CsHYZ3(2,1,1,18) 
     $  + s19 * CsHYZ3(2,1,1,19) 
     $  + s20 * CsHYZ3(2,1,1,20) 
     $  + s21 * CsHYZ3(2,1,1,21) 
     $  + s22 * CsHYZ3(2,1,1,22) 
      HYZ3(2,2,1) = 
     $  + s03 * CsHYZ3(2,2,1,03) 
     $  + s04 * CsHYZ3(2,2,1,04) 
     $  + s05 * CsHYZ3(2,2,1,05) 
     $  + s06 * CsHYZ3(2,2,1,06) 
     $  + s07 * CsHYZ3(2,2,1,07) 
     $  + s08 * CsHYZ3(2,2,1,08) 
     $  + s09 * CsHYZ3(2,2,1,09) 
     $  + s10 * CsHYZ3(2,2,1,10) 
     $  + s11 * CsHYZ3(2,2,1,11) 
     $  + s12 * CsHYZ3(2,2,1,12) 
     $  + s13 * CsHYZ3(2,2,1,13) 
     $  + s14 * CsHYZ3(2,2,1,14) 
     $  + s15 * CsHYZ3(2,2,1,15) 
     $  + s16 * CsHYZ3(2,2,1,16) 
     $  + s17 * CsHYZ3(2,2,1,17) 
     $  + s18 * CsHYZ3(2,2,1,18) 
     $  + s19 * CsHYZ3(2,2,1,19) 
     $  + s20 * CsHYZ3(2,2,1,20) 
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      HYZ4(0,0,2,1) = 
     $  + s02 * CsHYZ4(0,0,2,1,02) 
     $  + s03 * CsHYZ4(0,0,2,1,03) 
     $  + s04 * CsHYZ4(0,0,2,1,04) 
     $  + s05 * CsHYZ4(0,0,2,1,05) 
     $  + s06 * CsHYZ4(0,0,2,1,06) 
     $  + s07 * CsHYZ4(0,0,2,1,07) 
     $  + s08 * CsHYZ4(0,0,2,1,08) 
     $  + s09 * CsHYZ4(0,0,2,1,09) 
     $  + s10 * CsHYZ4(0,0,2,1,10) 
     $  + s11 * CsHYZ4(0,0,2,1,11) 
     $  + s12 * CsHYZ4(0,0,2,1,12) 
     $  + s13 * CsHYZ4(0,0,2,1,13) 
     $  + s14 * CsHYZ4(0,0,2,1,14) 
     $  + s15 * CsHYZ4(0,0,2,1,15) 
     $  + s16 * CsHYZ4(0,0,2,1,16) 
     $  + s17 * CsHYZ4(0,0,2,1,17) 
     $  + s18 * CsHYZ4(0,0,2,1,18) 
      HYZ4(0,1,2,1) = 
     $  + s03 * CsHYZ4(0,1,2,1,03) 
     $  + s04 * CsHYZ4(0,1,2,1,04) 
     $  + s05 * CsHYZ4(0,1,2,1,05) 
     $  + s06 * CsHYZ4(0,1,2,1,06) 
     $  + s07 * CsHYZ4(0,1,2,1,07) 
     $  + s08 * CsHYZ4(0,1,2,1,08) 
     $  + s09 * CsHYZ4(0,1,2,1,09) 
     $  + s10 * CsHYZ4(0,1,2,1,10) 
     $  + s11 * CsHYZ4(0,1,2,1,11) 
     $  + s12 * CsHYZ4(0,1,2,1,12) 
     $  + s13 * CsHYZ4(0,1,2,1,13) 
     $  + s14 * CsHYZ4(0,1,2,1,14) 
     $  + s15 * CsHYZ4(0,1,2,1,15) 
     $  + s16 * CsHYZ4(0,1,2,1,16) 
     $  + s17 * CsHYZ4(0,1,2,1,17) 
     $  + s18 * CsHYZ4(0,1,2,1,18) 
     $  + s19 * CsHYZ4(0,1,2,1,19) 
     $  + s20 * CsHYZ4(0,1,2,1,20) 
     $  + s21 * CsHYZ4(0,1,2,1,21) 
      HYZ4(0,2,1,1) = 
     $  + s03 * CsHYZ4(0,2,1,1,03) 
     $  + s04 * CsHYZ4(0,2,1,1,04) 
     $  + s05 * CsHYZ4(0,2,1,1,05) 
     $  + s06 * CsHYZ4(0,2,1,1,06) 
     $  + s07 * CsHYZ4(0,2,1,1,07) 
     $  + s08 * CsHYZ4(0,2,1,1,08) 
     $  + s09 * CsHYZ4(0,2,1,1,09) 
     $  + s10 * CsHYZ4(0,2,1,1,10) 
     $  + s11 * CsHYZ4(0,2,1,1,11) 
     $  + s12 * CsHYZ4(0,2,1,1,12) 
     $  + s13 * CsHYZ4(0,2,1,1,13) 
     $  + s14 * CsHYZ4(0,2,1,1,14) 
     $  + s15 * CsHYZ4(0,2,1,1,15) 
     $  + s16 * CsHYZ4(0,2,1,1,16) 
     $  + s17 * CsHYZ4(0,2,1,1,17) 
     $  + s18 * CsHYZ4(0,2,1,1,18) 
     $  + s19 * CsHYZ4(0,2,1,1,19) 
     $  + s20 * CsHYZ4(0,2,1,1,20) 
      HYZ4(0,2,0,1) = 
     $  + s02 * CsHYZ4(0,2,0,1,02) 
     $  + s03 * CsHYZ4(0,2,0,1,03) 
     $  + s04 * CsHYZ4(0,2,0,1,04) 
     $  + s05 * CsHYZ4(0,2,0,1,05) 
     $  + s06 * CsHYZ4(0,2,0,1,06) 
     $  + s07 * CsHYZ4(0,2,0,1,07) 
     $  + s08 * CsHYZ4(0,2,0,1,08) 
     $  + s09 * CsHYZ4(0,2,0,1,09) 
     $  + s10 * CsHYZ4(0,2,0,1,10) 
     $  + s11 * CsHYZ4(0,2,0,1,11) 
     $  + s12 * CsHYZ4(0,2,0,1,12) 
     $  + s13 * CsHYZ4(0,2,0,1,13) 
     $  + s14 * CsHYZ4(0,2,0,1,14) 
     $  + s15 * CsHYZ4(0,2,0,1,15) 
     $  + s16 * CsHYZ4(0,2,0,1,16) 
     $  + s17 * CsHYZ4(0,2,0,1,17) 
     $  + s18 * CsHYZ4(0,2,0,1,18) 
      HYZ4(0,2,2,1) = 
     $  + s03 * CsHYZ4(0,2,2,1,03) 
     $  + s04 * CsHYZ4(0,2,2,1,04) 
     $  + s05 * CsHYZ4(0,2,2,1,05) 
     $  + s06 * CsHYZ4(0,2,2,1,06) 
     $  + s07 * CsHYZ4(0,2,2,1,07) 
     $  + s08 * CsHYZ4(0,2,2,1,08) 
     $  + s09 * CsHYZ4(0,2,2,1,09) 
     $  + s10 * CsHYZ4(0,2,2,1,10) 
     $  + s11 * CsHYZ4(0,2,2,1,11) 
     $  + s12 * CsHYZ4(0,2,2,1,12) 
     $  + s13 * CsHYZ4(0,2,2,1,13) 
     $  + s14 * CsHYZ4(0,2,2,1,14) 
     $  + s15 * CsHYZ4(0,2,2,1,15) 
     $  + s16 * CsHYZ4(0,2,2,1,16) 
     $  + s17 * CsHYZ4(0,2,2,1,17) 
     $  + s18 * CsHYZ4(0,2,2,1,18) 
     $  + s19 * CsHYZ4(0,2,2,1,19) 
      HYZ4(2,0,0,1) = 
     $  + s02 * CsHYZ4(2,0,0,1,02) 
     $  + s03 * CsHYZ4(2,0,0,1,03) 
     $  + s04 * CsHYZ4(2,0,0,1,04) 
     $  + s05 * CsHYZ4(2,0,0,1,05) 
     $  + s06 * CsHYZ4(2,0,0,1,06) 
     $  + s07 * CsHYZ4(2,0,0,1,07) 
     $  + s08 * CsHYZ4(2,0,0,1,08) 
     $  + s09 * CsHYZ4(2,0,0,1,09) 
     $  + s10 * CsHYZ4(2,0,0,1,10) 
     $  + s11 * CsHYZ4(2,0,0,1,11) 
     $  + s12 * CsHYZ4(2,0,0,1,12) 
     $  + s13 * CsHYZ4(2,0,0,1,13) 
     $  + s14 * CsHYZ4(2,0,0,1,14) 
     $  + s15 * CsHYZ4(2,0,0,1,15) 
     $  + s16 * CsHYZ4(2,0,0,1,16) 
     $  + s17 * CsHYZ4(2,0,0,1,17) 
     $  + s18 * CsHYZ4(2,0,0,1,18) 
      HYZ4(2,0,1,1) = 
     $  + s03 * CsHYZ4(2,0,1,1,03) 
     $  + s04 * CsHYZ4(2,0,1,1,04) 
     $  + s05 * CsHYZ4(2,0,1,1,05) 
     $  + s06 * CsHYZ4(2,0,1,1,06) 
     $  + s07 * CsHYZ4(2,0,1,1,07) 
     $  + s08 * CsHYZ4(2,0,1,1,08) 
     $  + s09 * CsHYZ4(2,0,1,1,09) 
     $  + s10 * CsHYZ4(2,0,1,1,10) 
     $  + s11 * CsHYZ4(2,0,1,1,11) 
     $  + s12 * CsHYZ4(2,0,1,1,12) 
     $  + s13 * CsHYZ4(2,0,1,1,13) 
     $  + s14 * CsHYZ4(2,0,1,1,14) 
     $  + s15 * CsHYZ4(2,0,1,1,15) 
     $  + s16 * CsHYZ4(2,0,1,1,16) 
     $  + s17 * CsHYZ4(2,0,1,1,17) 
     $  + s18 * CsHYZ4(2,0,1,1,18) 
     $  + s19 * CsHYZ4(2,0,1,1,19) 
     $  + s20 * CsHYZ4(2,0,1,1,20) 
      HYZ4(2,0,2,1) = 
     $  + s03 * CsHYZ4(2,0,2,1,03) 
     $  + s04 * CsHYZ4(2,0,2,1,04) 
     $  + s05 * CsHYZ4(2,0,2,1,05) 
     $  + s06 * CsHYZ4(2,0,2,1,06) 
     $  + s07 * CsHYZ4(2,0,2,1,07) 
     $  + s08 * CsHYZ4(2,0,2,1,08) 
     $  + s09 * CsHYZ4(2,0,2,1,09) 
     $  + s10 * CsHYZ4(2,0,2,1,10) 
     $  + s11 * CsHYZ4(2,0,2,1,11) 
     $  + s12 * CsHYZ4(2,0,2,1,12) 
     $  + s13 * CsHYZ4(2,0,2,1,13) 
     $  + s14 * CsHYZ4(2,0,2,1,14) 
     $  + s15 * CsHYZ4(2,0,2,1,15) 
     $  + s16 * CsHYZ4(2,0,2,1,16) 
     $  + s17 * CsHYZ4(2,0,2,1,17) 
     $  + s18 * CsHYZ4(2,0,2,1,18) 
     $  + s19 * CsHYZ4(2,0,2,1,19) 
      HYZ4(2,1,1,1) = 
     $  + s04 * CsHYZ4(2,1,1,1,04) 
     $  + s05 * CsHYZ4(2,1,1,1,05) 
     $  + s06 * CsHYZ4(2,1,1,1,06) 
     $  + s07 * CsHYZ4(2,1,1,1,07) 
     $  + s08 * CsHYZ4(2,1,1,1,08) 
     $  + s09 * CsHYZ4(2,1,1,1,09) 
     $  + s10 * CsHYZ4(2,1,1,1,10) 
     $  + s11 * CsHYZ4(2,1,1,1,11) 
     $  + s12 * CsHYZ4(2,1,1,1,12) 
     $  + s13 * CsHYZ4(2,1,1,1,13) 
     $  + s14 * CsHYZ4(2,1,1,1,14) 
     $  + s15 * CsHYZ4(2,1,1,1,15) 
     $  + s16 * CsHYZ4(2,1,1,1,16) 
     $  + s17 * CsHYZ4(2,1,1,1,17) 
     $  + s18 * CsHYZ4(2,1,1,1,18) 
     $  + s19 * CsHYZ4(2,1,1,1,19) 
     $  + s20 * CsHYZ4(2,1,1,1,20) 
     $  + s21 * CsHYZ4(2,1,1,1,21) 
     $  + s22 * CsHYZ4(2,1,1,1,22) 
      HYZ4(2,2,0,1) = 
     $  + s03 * CsHYZ4(2,2,0,1,03) 
     $  + s04 * CsHYZ4(2,2,0,1,04) 
     $  + s05 * CsHYZ4(2,2,0,1,05) 
     $  + s06 * CsHYZ4(2,2,0,1,06) 
     $  + s07 * CsHYZ4(2,2,0,1,07) 
     $  + s08 * CsHYZ4(2,2,0,1,08) 
     $  + s09 * CsHYZ4(2,2,0,1,09) 
     $  + s10 * CsHYZ4(2,2,0,1,10) 
     $  + s11 * CsHYZ4(2,2,0,1,11) 
     $  + s12 * CsHYZ4(2,2,0,1,12) 
     $  + s13 * CsHYZ4(2,2,0,1,13) 
     $  + s14 * CsHYZ4(2,2,0,1,14) 
     $  + s15 * CsHYZ4(2,2,0,1,15) 
     $  + s16 * CsHYZ4(2,2,0,1,16) 
     $  + s17 * CsHYZ4(2,2,0,1,17) 
     $  + s18 * CsHYZ4(2,2,0,1,18) 
     $  + s19 * CsHYZ4(2,2,0,1,19) 
      HYZ4(2,2,1,1) = 
     $  + s04 * CsHYZ4(2,2,1,1,04) 
     $  + s05 * CsHYZ4(2,2,1,1,05) 
     $  + s06 * CsHYZ4(2,2,1,1,06) 
     $  + s07 * CsHYZ4(2,2,1,1,07) 
     $  + s08 * CsHYZ4(2,2,1,1,08) 
     $  + s09 * CsHYZ4(2,2,1,1,09) 
     $  + s10 * CsHYZ4(2,2,1,1,10) 
     $  + s11 * CsHYZ4(2,2,1,1,11) 
     $  + s12 * CsHYZ4(2,2,1,1,12) 
     $  + s13 * CsHYZ4(2,2,1,1,13) 
     $  + s14 * CsHYZ4(2,2,1,1,14) 
     $  + s15 * CsHYZ4(2,2,1,1,15) 
     $  + s16 * CsHYZ4(2,2,1,1,16) 
     $  + s17 * CsHYZ4(2,2,1,1,17) 
     $  + s18 * CsHYZ4(2,2,1,1,18) 
     $  + s19 * CsHYZ4(2,2,1,1,19) 
     $  + s20 * CsHYZ4(2,2,1,1,20) 
     $  + s21 * CsHYZ4(2,2,1,1,21) 
      HYZ4(2,2,2,1) = 
     $  + s04 * CsHYZ4(2,2,2,1,04) 
     $  + s05 * CsHYZ4(2,2,2,1,05) 
     $  + s06 * CsHYZ4(2,2,2,1,06) 
     $  + s07 * CsHYZ4(2,2,2,1,07) 
     $  + s08 * CsHYZ4(2,2,2,1,08) 
     $  + s09 * CsHYZ4(2,2,2,1,09) 
     $  + s10 * CsHYZ4(2,2,2,1,10) 
     $  + s11 * CsHYZ4(2,2,2,1,11) 
     $  + s12 * CsHYZ4(2,2,2,1,12) 
     $  + s13 * CsHYZ4(2,2,2,1,13) 
     $  + s14 * CsHYZ4(2,2,2,1,14) 
     $  + s15 * CsHYZ4(2,2,2,1,15) 
     $  + s16 * CsHYZ4(2,2,2,1,16) 
     $  + s17 * CsHYZ4(2,2,2,1,17) 
     $  + s18 * CsHYZ4(2,2,2,1,18) 
     $  + s19 * CsHYZ4(2,2,2,1,19) 
      endif

      return
      end

      subroutine fillirr2dhpl310(iflag,n)
*********************************************************************
***** fillirr2dhpl310(iflag,n) fills the irreducible 2dHPL      ***** 
***** with indices (3,1,0)                                      *****
***** up to weight n using the EXPANDED expressions for the     *****
***** z-dependent expansion coefficients                        *****
***** applicable for all z                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      integer nmax
      real(dp):: CrHYZ2,CrHYZ3,CrHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     r01,r02,r03,r04,r05,r06,r07,r08,r09,r10,
     $     r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,r21,r22 
      dimension CrHYZ2(0:3,0:3,18),CrHYZ3(0:3,0:3,0:3,18),
     $          CrHYZ4(0:3,0:3,0:3,0:3,18)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      common /rtdhpl/
     $     r01,r02,r03,r04,r05,r06,r07,r08,r09,r10,
     $     r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,r21,r22 
      save CrHYZ2,CrHYZ3,CrHYZ4
!$omp threadprivate(CrHYZ2,CrHYZ3,CrHYZ4,/rtdhpl/,/HPL2/)
      
      if (iflag.eq.1) then 
      call fillcoeff2dhpl310(iflag,n,CrHYZ2,CrHYZ3,CrHYZ4)
      endif
* 2001-04-03:17:41:05.hpl 
* <- HPLexpand.310t 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      HYZ2(3,1) = 
     $  + r01 * CrHYZ2(3,1,01) 
     $  + r02 * CrHYZ2(3,1,02) 
     $  + r03 * CrHYZ2(3,1,03) 
     $  + r04 * CrHYZ2(3,1,04) 
     $  + r05 * CrHYZ2(3,1,05) 
     $  + r06 * CrHYZ2(3,1,06) 
     $  + r07 * CrHYZ2(3,1,07) 
     $  + r08 * CrHYZ2(3,1,08) 
     $  + r09 * CrHYZ2(3,1,09) 
     $  + r10 * CrHYZ2(3,1,10) 
     $  + r11 * CrHYZ2(3,1,11) 
     $  + r12 * CrHYZ2(3,1,12) 
     $  + r13 * CrHYZ2(3,1,13) 
     $  + r14 * CrHYZ2(3,1,14) 
     $  + r15 * CrHYZ2(3,1,15) 
     $  + r16 * CrHYZ2(3,1,16) 
     $  + r17 * CrHYZ2(3,1,17) 
     $  - HZ1( -1)*HYZ1(3)
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(0,3,1) = 
     $  + r01 * CrHYZ3(0,3,1,01) 
     $  + r02 * CrHYZ3(0,3,1,02) 
     $  + r03 * CrHYZ3(0,3,1,03) 
     $  + r04 * CrHYZ3(0,3,1,04) 
     $  + r05 * CrHYZ3(0,3,1,05) 
     $  + r06 * CrHYZ3(0,3,1,06) 
     $  + r07 * CrHYZ3(0,3,1,07) 
     $  + r08 * CrHYZ3(0,3,1,08) 
     $  + r09 * CrHYZ3(0,3,1,09) 
     $  + r10 * CrHYZ3(0,3,1,10) 
     $  + r11 * CrHYZ3(0,3,1,11) 
     $  + r12 * CrHYZ3(0,3,1,12) 
     $  + r13 * CrHYZ3(0,3,1,13) 
     $  + r14 * CrHYZ3(0,3,1,14) 
     $  + r15 * CrHYZ3(0,3,1,15) 
     $  + r16 * CrHYZ3(0,3,1,16) 
     $  + r17 * CrHYZ3(0,3,1,17) 
     $  - HZ1( -1)*HYZ2(0,3)
      HYZ3(3,0,1) = 
     $  + r01 * CrHYZ3(3,0,1,01) 
     $  + r02 * CrHYZ3(3,0,1,02) 
     $  + r03 * CrHYZ3(3,0,1,03) 
     $  + r04 * CrHYZ3(3,0,1,04) 
     $  + r05 * CrHYZ3(3,0,1,05) 
     $  + r06 * CrHYZ3(3,0,1,06) 
     $  + r07 * CrHYZ3(3,0,1,07) 
     $  + r08 * CrHYZ3(3,0,1,08) 
     $  + r09 * CrHYZ3(3,0,1,09) 
     $  + r10 * CrHYZ3(3,0,1,10) 
     $  + r11 * CrHYZ3(3,0,1,11) 
     $  + r12 * CrHYZ3(3,0,1,12) 
     $  + r13 * CrHYZ3(3,0,1,13) 
     $  + r14 * CrHYZ3(3,0,1,14) 
     $  + r15 * CrHYZ3(3,0,1,15) 
     $  + r16 * CrHYZ3(3,0,1,16) 
     $  + r17 * CrHYZ3(3,0,1,17) 
     $  - HZ2(0, -1)*HYZ1(3)
      HYZ3(3,1,1) = 
     $  + r01 * CrHYZ3(3,1,1,01) 
     $  + r02 * CrHYZ3(3,1,1,02) 
     $  + r03 * CrHYZ3(3,1,1,03) 
     $  + r04 * CrHYZ3(3,1,1,04) 
     $  + r05 * CrHYZ3(3,1,1,05) 
     $  + r06 * CrHYZ3(3,1,1,06) 
     $  + r07 * CrHYZ3(3,1,1,07) 
     $  + r08 * CrHYZ3(3,1,1,08) 
     $  + r09 * CrHYZ3(3,1,1,09) 
     $  + r10 * CrHYZ3(3,1,1,10) 
     $  + r11 * CrHYZ3(3,1,1,11) 
     $  + r12 * CrHYZ3(3,1,1,12) 
     $  + r13 * CrHYZ3(3,1,1,13) 
     $  + r14 * CrHYZ3(3,1,1,14) 
     $  + r15 * CrHYZ3(3,1,1,15) 
     $  + r16 * CrHYZ3(3,1,1,16) 
     $  + r17 * CrHYZ3(3,1,1,17) 
     $  + HZ2( -1,-1)*HYZ1(3)
      HYZ3(3,3,1) = 
     $  + r01 * CrHYZ3(3,3,1,01) 
     $  + r02 * CrHYZ3(3,3,1,02) 
     $  + r03 * CrHYZ3(3,3,1,03) 
     $  + r04 * CrHYZ3(3,3,1,04) 
     $  + r05 * CrHYZ3(3,3,1,05) 
     $  + r06 * CrHYZ3(3,3,1,06) 
     $  + r07 * CrHYZ3(3,3,1,07) 
     $  + r08 * CrHYZ3(3,3,1,08) 
     $  + r09 * CrHYZ3(3,3,1,09) 
     $  + r10 * CrHYZ3(3,3,1,10) 
     $  + r11 * CrHYZ3(3,3,1,11) 
     $  + r12 * CrHYZ3(3,3,1,12) 
     $  + r13 * CrHYZ3(3,3,1,13) 
     $  + r14 * CrHYZ3(3,3,1,14) 
     $  + r15 * CrHYZ3(3,3,1,15) 
     $  + r16 * CrHYZ3(3,3,1,16) 
     $  + r17 * CrHYZ3(3,3,1,17) 
     $  - HZ1( -1)*HYZ2(3,3)
     $  + HZ2( -1,-1)*HYZ1(3)
     $  - HZ2(0, -1)*HYZ1(3)
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      HYZ4(0,0,3,1) = 
     $  + r01 * CrHYZ4(0,0,3,1,01) 
     $  + r02 * CrHYZ4(0,0,3,1,02) 
     $  + r03 * CrHYZ4(0,0,3,1,03) 
     $  + r04 * CrHYZ4(0,0,3,1,04) 
     $  + r05 * CrHYZ4(0,0,3,1,05) 
     $  + r06 * CrHYZ4(0,0,3,1,06) 
     $  + r07 * CrHYZ4(0,0,3,1,07) 
     $  + r08 * CrHYZ4(0,0,3,1,08) 
     $  + r09 * CrHYZ4(0,0,3,1,09) 
     $  + r10 * CrHYZ4(0,0,3,1,10) 
     $  + r11 * CrHYZ4(0,0,3,1,11) 
     $  + r12 * CrHYZ4(0,0,3,1,12) 
     $  + r13 * CrHYZ4(0,0,3,1,13) 
     $  + r14 * CrHYZ4(0,0,3,1,14) 
     $  + r15 * CrHYZ4(0,0,3,1,15) 
     $  + r16 * CrHYZ4(0,0,3,1,16) 
     $  + r17 * CrHYZ4(0,0,3,1,17) 
     $  - HZ1( -1)*HYZ3(0,0,3)
      HYZ4(0,1,3,1) = 
     $  + r01 * CrHYZ4(0,1,3,1,01) 
     $  + r02 * CrHYZ4(0,1,3,1,02) 
     $  + r03 * CrHYZ4(0,1,3,1,03) 
     $  + r04 * CrHYZ4(0,1,3,1,04) 
     $  + r05 * CrHYZ4(0,1,3,1,05) 
     $  + r06 * CrHYZ4(0,1,3,1,06) 
     $  + r07 * CrHYZ4(0,1,3,1,07) 
     $  + r08 * CrHYZ4(0,1,3,1,08) 
     $  + r09 * CrHYZ4(0,1,3,1,09) 
     $  + r10 * CrHYZ4(0,1,3,1,10) 
     $  + r11 * CrHYZ4(0,1,3,1,11) 
     $  + r12 * CrHYZ4(0,1,3,1,12) 
     $  + r13 * CrHYZ4(0,1,3,1,13) 
     $  + r14 * CrHYZ4(0,1,3,1,14) 
     $  + r15 * CrHYZ4(0,1,3,1,15) 
     $  + r16 * CrHYZ4(0,1,3,1,16) 
     $  + r17 * CrHYZ4(0,1,3,1,17) 
     $  - HZ1( -1)*HYZ3(0,1,3)
     $  + 2.000000000000000d+00*HZ2(-1,-1)*HYZ2(0,1)
      HYZ4(0,3,0,1) = 
     $  + r01 * CrHYZ4(0,3,0,1,01) 
     $  + r02 * CrHYZ4(0,3,0,1,02) 
     $  + r03 * CrHYZ4(0,3,0,1,03) 
     $  + r04 * CrHYZ4(0,3,0,1,04) 
     $  + r05 * CrHYZ4(0,3,0,1,05) 
     $  + r06 * CrHYZ4(0,3,0,1,06) 
     $  + r07 * CrHYZ4(0,3,0,1,07) 
     $  + r08 * CrHYZ4(0,3,0,1,08) 
     $  + r09 * CrHYZ4(0,3,0,1,09) 
     $  + r10 * CrHYZ4(0,3,0,1,10) 
     $  + r11 * CrHYZ4(0,3,0,1,11) 
     $  + r12 * CrHYZ4(0,3,0,1,12) 
     $  + r13 * CrHYZ4(0,3,0,1,13) 
     $  + r14 * CrHYZ4(0,3,0,1,14) 
     $  + r15 * CrHYZ4(0,3,0,1,15) 
     $  + r16 * CrHYZ4(0,3,0,1,16) 
     $  + r17 * CrHYZ4(0,3,0,1,17) 
     $  - HZ2(0, -1)*HYZ2(0,3)
      HYZ4(0,3,1,1) = 
     $  + r01 * CrHYZ4(0,3,1,1,01) 
     $  + r02 * CrHYZ4(0,3,1,1,02) 
     $  + r03 * CrHYZ4(0,3,1,1,03) 
     $  + r04 * CrHYZ4(0,3,1,1,04) 
     $  + r05 * CrHYZ4(0,3,1,1,05) 
     $  + r06 * CrHYZ4(0,3,1,1,06) 
     $  + r07 * CrHYZ4(0,3,1,1,07) 
     $  + r08 * CrHYZ4(0,3,1,1,08) 
     $  + r09 * CrHYZ4(0,3,1,1,09) 
     $  + r10 * CrHYZ4(0,3,1,1,10) 
     $  + r11 * CrHYZ4(0,3,1,1,11) 
     $  + r12 * CrHYZ4(0,3,1,1,12) 
     $  + r13 * CrHYZ4(0,3,1,1,13) 
     $  + r14 * CrHYZ4(0,3,1,1,14) 
     $  + r15 * CrHYZ4(0,3,1,1,15) 
     $  + r16 * CrHYZ4(0,3,1,1,16) 
     $  + r17 * CrHYZ4(0,3,1,1,17) 
     $  + HZ2( -1,-1)*HYZ2(0,3)
      HYZ4(0,3,3,1) = 
     $  + r01 * CrHYZ4(0,3,3,1,01) 
     $  + r02 * CrHYZ4(0,3,3,1,02) 
     $  + r03 * CrHYZ4(0,3,3,1,03) 
     $  + r04 * CrHYZ4(0,3,3,1,04) 
     $  + r05 * CrHYZ4(0,3,3,1,05) 
     $  + r06 * CrHYZ4(0,3,3,1,06) 
     $  + r07 * CrHYZ4(0,3,3,1,07) 
     $  + r08 * CrHYZ4(0,3,3,1,08) 
     $  + r09 * CrHYZ4(0,3,3,1,09) 
     $  + r10 * CrHYZ4(0,3,3,1,10) 
     $  + r11 * CrHYZ4(0,3,3,1,11) 
     $  + r12 * CrHYZ4(0,3,3,1,12) 
     $  + r13 * CrHYZ4(0,3,3,1,13) 
     $  + r14 * CrHYZ4(0,3,3,1,14) 
     $  + r15 * CrHYZ4(0,3,3,1,15) 
     $  + r16 * CrHYZ4(0,3,3,1,16) 
     $  + r17 * CrHYZ4(0,3,3,1,17) 
     $  - HZ1( -1)*HYZ3(0,3,3)
     $  + HZ2( -1,-1)*HYZ2(0,3)
     $  - HZ2(0, -1)*HYZ2(0,3)
      HYZ4(3,0,0,1) = 
     $  + r01 * CrHYZ4(3,0,0,1,01) 
     $  + r02 * CrHYZ4(3,0,0,1,02) 
     $  + r03 * CrHYZ4(3,0,0,1,03) 
     $  + r04 * CrHYZ4(3,0,0,1,04) 
     $  + r05 * CrHYZ4(3,0,0,1,05) 
     $  + r06 * CrHYZ4(3,0,0,1,06) 
     $  + r07 * CrHYZ4(3,0,0,1,07) 
     $  + r08 * CrHYZ4(3,0,0,1,08) 
     $  + r09 * CrHYZ4(3,0,0,1,09) 
     $  + r10 * CrHYZ4(3,0,0,1,10) 
     $  + r11 * CrHYZ4(3,0,0,1,11) 
     $  + r12 * CrHYZ4(3,0,0,1,12) 
     $  + r13 * CrHYZ4(3,0,0,1,13) 
     $  + r14 * CrHYZ4(3,0,0,1,14) 
     $  + r15 * CrHYZ4(3,0,0,1,15) 
     $  + r16 * CrHYZ4(3,0,0,1,16) 
     $  + r17 * CrHYZ4(3,0,0,1,17) 
     $  - HZ3(0,0, -1)*HYZ1(3)
      HYZ4(3,0,1,1) = 
     $  + r01 * CrHYZ4(3,0,1,1,01) 
     $  + r02 * CrHYZ4(3,0,1,1,02) 
     $  + r03 * CrHYZ4(3,0,1,1,03) 
     $  + r04 * CrHYZ4(3,0,1,1,04) 
     $  + r05 * CrHYZ4(3,0,1,1,05) 
     $  + r06 * CrHYZ4(3,0,1,1,06) 
     $  + r07 * CrHYZ4(3,0,1,1,07) 
     $  + r08 * CrHYZ4(3,0,1,1,08) 
     $  + r09 * CrHYZ4(3,0,1,1,09) 
     $  + r10 * CrHYZ4(3,0,1,1,10) 
     $  + r11 * CrHYZ4(3,0,1,1,11) 
     $  + r12 * CrHYZ4(3,0,1,1,12) 
     $  + r13 * CrHYZ4(3,0,1,1,13) 
     $  + r14 * CrHYZ4(3,0,1,1,14) 
     $  + r15 * CrHYZ4(3,0,1,1,15) 
     $  + r16 * CrHYZ4(3,0,1,1,16) 
     $  + r17 * CrHYZ4(3,0,1,1,17) 
     $  + r18 * CrHYZ4(3,0,1,1,18) 
     $  + HZ3(0, -1,-1)*HYZ1(3)
      HYZ4(3,0,3,1) = 
     $  + r01 * CrHYZ4(3,0,3,1,01) 
     $  + r02 * CrHYZ4(3,0,3,1,02) 
     $  + r03 * CrHYZ4(3,0,3,1,03) 
     $  + r04 * CrHYZ4(3,0,3,1,04) 
     $  + r05 * CrHYZ4(3,0,3,1,05) 
     $  + r06 * CrHYZ4(3,0,3,1,06) 
     $  + r07 * CrHYZ4(3,0,3,1,07) 
     $  + r08 * CrHYZ4(3,0,3,1,08) 
     $  + r09 * CrHYZ4(3,0,3,1,09) 
     $  + r10 * CrHYZ4(3,0,3,1,10) 
     $  + r11 * CrHYZ4(3,0,3,1,11) 
     $  + r12 * CrHYZ4(3,0,3,1,12) 
     $  + r13 * CrHYZ4(3,0,3,1,13) 
     $  + r14 * CrHYZ4(3,0,3,1,14) 
     $  + r15 * CrHYZ4(3,0,3,1,15) 
     $  + r16 * CrHYZ4(3,0,3,1,16) 
     $  + r17 * CrHYZ4(3,0,3,1,17) 
     $  - HZ1( -1)*HYZ3(3,0,3)
     $  + HZ3( -1,0,-1)*HYZ1(3)
     $  - HZ3(0,0, -1)*HYZ1(3)
      HYZ4(3,1,1,1) = 
     $  + r01 * CrHYZ4(3,1,1,1,01) 
     $  + r02 * CrHYZ4(3,1,1,1,02) 
     $  + r03 * CrHYZ4(3,1,1,1,03) 
     $  + r04 * CrHYZ4(3,1,1,1,04) 
     $  + r05 * CrHYZ4(3,1,1,1,05) 
     $  + r06 * CrHYZ4(3,1,1,1,06) 
     $  + r07 * CrHYZ4(3,1,1,1,07) 
     $  + r08 * CrHYZ4(3,1,1,1,08) 
     $  + r09 * CrHYZ4(3,1,1,1,09) 
     $  + r10 * CrHYZ4(3,1,1,1,10) 
     $  + r11 * CrHYZ4(3,1,1,1,11) 
     $  + r12 * CrHYZ4(3,1,1,1,12) 
     $  + r13 * CrHYZ4(3,1,1,1,13) 
     $  + r14 * CrHYZ4(3,1,1,1,14) 
     $  + r15 * CrHYZ4(3,1,1,1,15) 
     $  + r16 * CrHYZ4(3,1,1,1,16) 
     $  + r17 * CrHYZ4(3,1,1,1,17) 
     $  - HZ3( -1,-1,-1)*HYZ1(3)
      HYZ4(3,3,0,1) = 
     $  + r01 * CrHYZ4(3,3,0,1,01) 
     $  + r02 * CrHYZ4(3,3,0,1,02) 
     $  + r03 * CrHYZ4(3,3,0,1,03) 
     $  + r04 * CrHYZ4(3,3,0,1,04) 
     $  + r05 * CrHYZ4(3,3,0,1,05) 
     $  + r06 * CrHYZ4(3,3,0,1,06) 
     $  + r07 * CrHYZ4(3,3,0,1,07) 
     $  + r08 * CrHYZ4(3,3,0,1,08) 
     $  + r09 * CrHYZ4(3,3,0,1,09) 
     $  + r10 * CrHYZ4(3,3,0,1,10) 
     $  + r11 * CrHYZ4(3,3,0,1,11) 
     $  + r12 * CrHYZ4(3,3,0,1,12) 
     $  + r13 * CrHYZ4(3,3,0,1,13) 
     $  + r14 * CrHYZ4(3,3,0,1,14) 
     $  + r15 * CrHYZ4(3,3,0,1,15) 
     $  + r16 * CrHYZ4(3,3,0,1,16) 
     $  + r17 * CrHYZ4(3,3,0,1,17) 
     $  - HZ2(0, -1)*HYZ2(3,3)
     $  + HZ3(0, -1,-1)*HYZ1(3)
     $  - HZ3(0,0, -1)*HYZ1(3)
      HYZ4(3,3,1,1) = 
     $  + r01 * CrHYZ4(3,3,1,1,01) 
     $  + r02 * CrHYZ4(3,3,1,1,02) 
     $  + r03 * CrHYZ4(3,3,1,1,03) 
     $  + r04 * CrHYZ4(3,3,1,1,04) 
     $  + r05 * CrHYZ4(3,3,1,1,05) 
     $  + r06 * CrHYZ4(3,3,1,1,06) 
     $  + r07 * CrHYZ4(3,3,1,1,07) 
     $  + r08 * CrHYZ4(3,3,1,1,08) 
     $  + r09 * CrHYZ4(3,3,1,1,09) 
     $  + r10 * CrHYZ4(3,3,1,1,10) 
     $  + r11 * CrHYZ4(3,3,1,1,11) 
     $  + r12 * CrHYZ4(3,3,1,1,12) 
     $  + r13 * CrHYZ4(3,3,1,1,13) 
     $  + r14 * CrHYZ4(3,3,1,1,14) 
     $  + r15 * CrHYZ4(3,3,1,1,15) 
     $  + r16 * CrHYZ4(3,3,1,1,16) 
     $  + r17 * CrHYZ4(3,3,1,1,17) 
     $  + HZ2( -1,-1)*HYZ2(3,3)
     $  - 2.000000000000000d+00*HZ3(-1,-1,-1)*HYZ1(3)
     $  + HZ3( -1,0,-1)*HYZ1(3)
     $  + HZ3(0, -1,-1)*HYZ1(3)
      HYZ4(3,3,3,1) = 
     $  + r01 * CrHYZ4(3,3,3,1,01) 
     $  + r02 * CrHYZ4(3,3,3,1,02) 
     $  + r03 * CrHYZ4(3,3,3,1,03) 
     $  + r04 * CrHYZ4(3,3,3,1,04) 
     $  + r05 * CrHYZ4(3,3,3,1,05) 
     $  + r06 * CrHYZ4(3,3,3,1,06) 
     $  + r07 * CrHYZ4(3,3,3,1,07) 
     $  + r08 * CrHYZ4(3,3,3,1,08) 
     $  + r09 * CrHYZ4(3,3,3,1,09) 
     $  + r10 * CrHYZ4(3,3,3,1,10) 
     $  + r11 * CrHYZ4(3,3,3,1,11) 
     $  + r12 * CrHYZ4(3,3,3,1,12) 
     $  + r13 * CrHYZ4(3,3,3,1,13) 
     $  + r14 * CrHYZ4(3,3,3,1,14) 
     $  + r15 * CrHYZ4(3,3,3,1,15) 
     $  + r16 * CrHYZ4(3,3,3,1,16) 
     $  + r17 * CrHYZ4(3,3,3,1,17) 
     $  - HZ1( -1)*HYZ3(3,3,3)
     $  + HZ2( -1,-1)*HYZ2(3,3)
     $  - HZ2(0, -1)*HYZ2(3,3)
     $  - HZ3( -1,-1,-1)*HYZ1(3)
     $  + HZ3( -1,0,-1)*HYZ1(3)
     $  + HZ3(0, -1,-1)*HYZ1(3)
     $  - HZ3(0,0, -1)*HYZ1(3)
      endif

      return
      end


      subroutine fillirr2dhpl320(iflag,n)
*********************************************************************
***** fillirr2dhpl320(iflag,n) fills the irreducible 2dHPL      ***** 
***** with indices (3,2,0)                                      *****
***** up to weight n using the EXPANDED expressions for the     *****
***** z-dependent expansion coefficients                        *****
***** applicable for z<0.5                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsHYZ2,CsHYZ3,CsHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      dimension CsHYZ2(0:3,0:3,22),CsHYZ3(0:3,0:3,0:3,22),
     $          CsHYZ4(0:3,0:3,0:3,0:3,22)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      common /s/
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      save CsHYZ2,CsHYZ3,CsHYZ4
!$omp threadprivate(CsHYZ2,CsHYZ3,CsHYZ4,/s/,/HPL2/)

      if (iflag.eq.1) then
      call fillcoeff2dhpl320(iflag,n,CsHYZ2,CsHYZ3,CsHYZ4)
      endif

* 2001-04-20:10:54:20.hpl 
* <- HPLexpand.320t 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      HYZ2(3,2) = 
     $  + s01 * CsHYZ2(3,2,01) 
     $  + s02 * CsHYZ2(3,2,02) 
     $  + s03 * CsHYZ2(3,2,03) 
     $  + s04 * CsHYZ2(3,2,04) 
     $  + s05 * CsHYZ2(3,2,05) 
     $  + s06 * CsHYZ2(3,2,06) 
     $  + s07 * CsHYZ2(3,2,07) 
     $  + s08 * CsHYZ2(3,2,08) 
     $  + s09 * CsHYZ2(3,2,09) 
     $  + s10 * CsHYZ2(3,2,10) 
     $  + s11 * CsHYZ2(3,2,11) 
     $  + s12 * CsHYZ2(3,2,12) 
     $  + s13 * CsHYZ2(3,2,13) 
     $  + s14 * CsHYZ2(3,2,14) 
     $  + s15 * CsHYZ2(3,2,15) 
     $  + s16 * CsHYZ2(3,2,16) 
     $  + s17 * CsHYZ2(3,2,17) 
     $  - HZ1(1) *HYZ1(3)
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(0,3,2) = 
     $  + s01 * CsHYZ3(0,3,2,01) 
     $  + s02 * CsHYZ3(0,3,2,02) 
     $  + s03 * CsHYZ3(0,3,2,03) 
     $  + s04 * CsHYZ3(0,3,2,04) 
     $  + s05 * CsHYZ3(0,3,2,05) 
     $  + s06 * CsHYZ3(0,3,2,06) 
     $  + s07 * CsHYZ3(0,3,2,07) 
     $  + s08 * CsHYZ3(0,3,2,08) 
     $  + s09 * CsHYZ3(0,3,2,09) 
     $  + s10 * CsHYZ3(0,3,2,10) 
     $  + s11 * CsHYZ3(0,3,2,11) 
     $  + s12 * CsHYZ3(0,3,2,12) 
     $  + s13 * CsHYZ3(0,3,2,13) 
     $  + s14 * CsHYZ3(0,3,2,14) 
     $  + s15 * CsHYZ3(0,3,2,15) 
     $  + s16 * CsHYZ3(0,3,2,16) 
     $  + s17 * CsHYZ3(0,3,2,17) 
     $  - HZ1(1) *HYZ2(0,3)
      HYZ3(3,0,2) = 
     $  + s01 * CsHYZ3(3,0,2,01) 
     $  + s02 * CsHYZ3(3,0,2,02) 
     $  + s03 * CsHYZ3(3,0,2,03) 
     $  + s04 * CsHYZ3(3,0,2,04) 
     $  + s05 * CsHYZ3(3,0,2,05) 
     $  + s06 * CsHYZ3(3,0,2,06) 
     $  + s07 * CsHYZ3(3,0,2,07) 
     $  + s08 * CsHYZ3(3,0,2,08) 
     $  + s09 * CsHYZ3(3,0,2,09) 
     $  + s10 * CsHYZ3(3,0,2,10) 
     $  + s11 * CsHYZ3(3,0,2,11) 
     $  + s12 * CsHYZ3(3,0,2,12) 
     $  + s13 * CsHYZ3(3,0,2,13) 
     $  + s14 * CsHYZ3(3,0,2,14) 
     $  + s15 * CsHYZ3(3,0,2,15) 
     $  + s16 * CsHYZ3(3,0,2,16) 
     $  + s17 * CsHYZ3(3,0,2,17) 
     $  - HZ2(0,1) *HYZ1(3)
     $  - HZ2(1,1) *HYZ1(3)
      HYZ3(3,2,2) = 
     $  + s01 * CsHYZ3(3,2,2,01) 
     $  + s02 * CsHYZ3(3,2,2,02) 
     $  + s03 * CsHYZ3(3,2,2,03) 
     $  + s04 * CsHYZ3(3,2,2,04) 
     $  + s05 * CsHYZ3(3,2,2,05) 
     $  + s06 * CsHYZ3(3,2,2,06) 
     $  + s07 * CsHYZ3(3,2,2,07) 
     $  + s08 * CsHYZ3(3,2,2,08) 
     $  + s09 * CsHYZ3(3,2,2,09) 
     $  + s10 * CsHYZ3(3,2,2,10) 
     $  + s11 * CsHYZ3(3,2,2,11) 
     $  + s12 * CsHYZ3(3,2,2,12) 
     $  + s13 * CsHYZ3(3,2,2,13) 
     $  + s14 * CsHYZ3(3,2,2,14) 
     $  + s15 * CsHYZ3(3,2,2,15) 
     $  + s16 * CsHYZ3(3,2,2,16) 
     $  + s17 * CsHYZ3(3,2,2,17) 
     $  + HZ2(1,1) *HYZ1(3)
      HYZ3(3,3,2) = 
     $  + s01 * CsHYZ3(3,3,2,01) 
     $  + s02 * CsHYZ3(3,3,2,02) 
     $  + s03 * CsHYZ3(3,3,2,03) 
     $  + s04 * CsHYZ3(3,3,2,04) 
     $  + s05 * CsHYZ3(3,3,2,05) 
     $  + s06 * CsHYZ3(3,3,2,06) 
     $  + s07 * CsHYZ3(3,3,2,07) 
     $  + s08 * CsHYZ3(3,3,2,08) 
     $  + s09 * CsHYZ3(3,3,2,09) 
     $  + s10 * CsHYZ3(3,3,2,10) 
     $  + s11 * CsHYZ3(3,3,2,11) 
     $  + s12 * CsHYZ3(3,3,2,12) 
     $  + s13 * CsHYZ3(3,3,2,13) 
     $  + s14 * CsHYZ3(3,3,2,14) 
     $  + s15 * CsHYZ3(3,3,2,15) 
     $  + s16 * CsHYZ3(3,3,2,16) 
     $  + s17 * CsHYZ3(3,3,2,17) 
     $  - HZ1(1) *HYZ2(3,3)
     $  - HZ2(0,1) *HYZ1(3)
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      HYZ4(0,0,3,2) = 
     $  + s01 * CsHYZ4(0,0,3,2,01) 
     $  + s02 * CsHYZ4(0,0,3,2,02) 
     $  + s03 * CsHYZ4(0,0,3,2,03) 
     $  + s04 * CsHYZ4(0,0,3,2,04) 
     $  + s05 * CsHYZ4(0,0,3,2,05) 
     $  + s06 * CsHYZ4(0,0,3,2,06) 
     $  + s07 * CsHYZ4(0,0,3,2,07) 
     $  + s08 * CsHYZ4(0,0,3,2,08) 
     $  + s09 * CsHYZ4(0,0,3,2,09) 
     $  + s10 * CsHYZ4(0,0,3,2,10) 
     $  + s11 * CsHYZ4(0,0,3,2,11) 
     $  + s12 * CsHYZ4(0,0,3,2,12) 
     $  + s13 * CsHYZ4(0,0,3,2,13) 
     $  + s14 * CsHYZ4(0,0,3,2,14) 
     $  + s15 * CsHYZ4(0,0,3,2,15) 
     $  + s16 * CsHYZ4(0,0,3,2,16) 
     $  + s17 * CsHYZ4(0,0,3,2,17) 
     $  - HZ1(1) *HYZ3(0,0,3)
      HYZ4(0,2,3,2) = 
     $  + s02 * CsHYZ4(0,2,3,2,02) 
     $  + s03 * CsHYZ4(0,2,3,2,03) 
     $  + s04 * CsHYZ4(0,2,3,2,04) 
     $  + s05 * CsHYZ4(0,2,3,2,05) 
     $  + s06 * CsHYZ4(0,2,3,2,06) 
     $  + s07 * CsHYZ4(0,2,3,2,07) 
     $  + s08 * CsHYZ4(0,2,3,2,08) 
     $  + s09 * CsHYZ4(0,2,3,2,09) 
     $  + s10 * CsHYZ4(0,2,3,2,10) 
     $  + s11 * CsHYZ4(0,2,3,2,11) 
     $  + s12 * CsHYZ4(0,2,3,2,12) 
     $  + s13 * CsHYZ4(0,2,3,2,13) 
     $  + s14 * CsHYZ4(0,2,3,2,14) 
     $  + s15 * CsHYZ4(0,2,3,2,15) 
     $  + s16 * CsHYZ4(0,2,3,2,16) 
     $  + s17 * CsHYZ4(0,2,3,2,17) 
     $  - HZ1(1) *HYZ3(0,2,3)
      HYZ4(0,3,0,2) = 
     $  + s01 * CsHYZ4(0,3,0,2,01) 
     $  + s02 * CsHYZ4(0,3,0,2,02) 
     $  + s03 * CsHYZ4(0,3,0,2,03) 
     $  + s04 * CsHYZ4(0,3,0,2,04) 
     $  + s05 * CsHYZ4(0,3,0,2,05) 
     $  + s06 * CsHYZ4(0,3,0,2,06) 
     $  + s07 * CsHYZ4(0,3,0,2,07) 
     $  + s08 * CsHYZ4(0,3,0,2,08) 
     $  + s09 * CsHYZ4(0,3,0,2,09) 
     $  + s10 * CsHYZ4(0,3,0,2,10) 
     $  + s11 * CsHYZ4(0,3,0,2,11) 
     $  + s12 * CsHYZ4(0,3,0,2,12) 
     $  + s13 * CsHYZ4(0,3,0,2,13) 
     $  + s14 * CsHYZ4(0,3,0,2,14) 
     $  + s15 * CsHYZ4(0,3,0,2,15) 
     $  + s16 * CsHYZ4(0,3,0,2,16) 
     $  + s17 * CsHYZ4(0,3,0,2,17) 
     $  - HZ2(0,1) *HYZ2(0,3)
     $  - HZ2(1,1) *HYZ2(0,3)
      HYZ4(0,3,2,2) = 
     $  + s01 * CsHYZ4(0,3,2,2,01) 
     $  + s02 * CsHYZ4(0,3,2,2,02) 
     $  + s03 * CsHYZ4(0,3,2,2,03) 
     $  + s04 * CsHYZ4(0,3,2,2,04) 
     $  + s05 * CsHYZ4(0,3,2,2,05) 
     $  + s06 * CsHYZ4(0,3,2,2,06) 
     $  + s07 * CsHYZ4(0,3,2,2,07) 
     $  + s08 * CsHYZ4(0,3,2,2,08) 
     $  + s09 * CsHYZ4(0,3,2,2,09) 
     $  + s10 * CsHYZ4(0,3,2,2,10) 
     $  + s11 * CsHYZ4(0,3,2,2,11) 
     $  + s12 * CsHYZ4(0,3,2,2,12) 
     $  + s13 * CsHYZ4(0,3,2,2,13) 
     $  + s14 * CsHYZ4(0,3,2,2,14) 
     $  + s15 * CsHYZ4(0,3,2,2,15) 
     $  + s16 * CsHYZ4(0,3,2,2,16) 
     $  + s17 * CsHYZ4(0,3,2,2,17) 
     $  + HZ2(1,1) *HYZ2(0,3)
      HYZ4(0,3,3,2) = 
     $  + s01 * CsHYZ4(0,3,3,2,01) 
     $  + s02 * CsHYZ4(0,3,3,2,02) 
     $  + s03 * CsHYZ4(0,3,3,2,03) 
     $  + s04 * CsHYZ4(0,3,3,2,04) 
     $  + s05 * CsHYZ4(0,3,3,2,05) 
     $  + s06 * CsHYZ4(0,3,3,2,06) 
     $  + s07 * CsHYZ4(0,3,3,2,07) 
     $  + s08 * CsHYZ4(0,3,3,2,08) 
     $  + s09 * CsHYZ4(0,3,3,2,09) 
     $  + s10 * CsHYZ4(0,3,3,2,10) 
     $  + s11 * CsHYZ4(0,3,3,2,11) 
     $  + s12 * CsHYZ4(0,3,3,2,12) 
     $  + s13 * CsHYZ4(0,3,3,2,13) 
     $  + s14 * CsHYZ4(0,3,3,2,14) 
     $  + s15 * CsHYZ4(0,3,3,2,15) 
     $  + s16 * CsHYZ4(0,3,3,2,16) 
     $  + s17 * CsHYZ4(0,3,3,2,17) 
     $  - HZ1(1) *HYZ3(0,3,3)
     $  - HZ2(0,1) *HYZ2(0,3)
      HYZ4(3,0,0,2) = 
     $  + s01 * CsHYZ4(3,0,0,2,01) 
     $  + s02 * CsHYZ4(3,0,0,2,02) 
     $  + s03 * CsHYZ4(3,0,0,2,03) 
     $  + s04 * CsHYZ4(3,0,0,2,04) 
     $  + s05 * CsHYZ4(3,0,0,2,05) 
     $  + s06 * CsHYZ4(3,0,0,2,06) 
     $  + s07 * CsHYZ4(3,0,0,2,07) 
     $  + s08 * CsHYZ4(3,0,0,2,08) 
     $  + s09 * CsHYZ4(3,0,0,2,09) 
     $  + s10 * CsHYZ4(3,0,0,2,10) 
     $  + s11 * CsHYZ4(3,0,0,2,11) 
     $  + s12 * CsHYZ4(3,0,0,2,12) 
     $  + s13 * CsHYZ4(3,0,0,2,13) 
     $  + s14 * CsHYZ4(3,0,0,2,14) 
     $  + s15 * CsHYZ4(3,0,0,2,15) 
     $  + s16 * CsHYZ4(3,0,0,2,16) 
     $  + s17 * CsHYZ4(3,0,0,2,17) 
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(0,1,1) *HYZ1(3)
     $  - HZ3(1,0,1) *HYZ1(3)
     $  - HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,0,2,2) = 
     $  + s01 * CsHYZ4(3,0,2,2,01) 
     $  + s02 * CsHYZ4(3,0,2,2,02) 
     $  + s03 * CsHYZ4(3,0,2,2,03) 
     $  + s04 * CsHYZ4(3,0,2,2,04) 
     $  + s05 * CsHYZ4(3,0,2,2,05) 
     $  + s06 * CsHYZ4(3,0,2,2,06) 
     $  + s07 * CsHYZ4(3,0,2,2,07) 
     $  + s08 * CsHYZ4(3,0,2,2,08) 
     $  + s09 * CsHYZ4(3,0,2,2,09) 
     $  + s10 * CsHYZ4(3,0,2,2,10) 
     $  + s11 * CsHYZ4(3,0,2,2,11) 
     $  + s12 * CsHYZ4(3,0,2,2,12) 
     $  + s13 * CsHYZ4(3,0,2,2,13) 
     $  + s14 * CsHYZ4(3,0,2,2,14) 
     $  + s15 * CsHYZ4(3,0,2,2,15) 
     $  + s16 * CsHYZ4(3,0,2,2,16) 
     $  + s17 * CsHYZ4(3,0,2,2,17) 
     $  + s18 * CsHYZ4(3,0,2,2,18) 
     $  + HZ3(0,1,1) *HYZ1(3)
     $  + HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,0,3,2) = 
     $  + s01 * CsHYZ4(3,0,3,2,01) 
     $  + s02 * CsHYZ4(3,0,3,2,02) 
     $  + s03 * CsHYZ4(3,0,3,2,03) 
     $  + s04 * CsHYZ4(3,0,3,2,04) 
     $  + s05 * CsHYZ4(3,0,3,2,05) 
     $  + s06 * CsHYZ4(3,0,3,2,06) 
     $  + s07 * CsHYZ4(3,0,3,2,07) 
     $  + s08 * CsHYZ4(3,0,3,2,08) 
     $  + s09 * CsHYZ4(3,0,3,2,09) 
     $  + s10 * CsHYZ4(3,0,3,2,10) 
     $  + s11 * CsHYZ4(3,0,3,2,11) 
     $  + s12 * CsHYZ4(3,0,3,2,12) 
     $  + s13 * CsHYZ4(3,0,3,2,13) 
     $  + s14 * CsHYZ4(3,0,3,2,14) 
     $  + s15 * CsHYZ4(3,0,3,2,15) 
     $  + s16 * CsHYZ4(3,0,3,2,16) 
     $  + s17 * CsHYZ4(3,0,3,2,17) 
     $  - HZ1(1) *HYZ3(3,0,3)
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(0,1,1) *HYZ1(3)
      HYZ4(3,2,2,2) = 
     $  + s01 * CsHYZ4(3,2,2,2,01) 
     $  + s02 * CsHYZ4(3,2,2,2,02) 
     $  + s03 * CsHYZ4(3,2,2,2,03) 
     $  + s04 * CsHYZ4(3,2,2,2,04) 
     $  + s05 * CsHYZ4(3,2,2,2,05) 
     $  + s06 * CsHYZ4(3,2,2,2,06) 
     $  + s07 * CsHYZ4(3,2,2,2,07) 
     $  + s08 * CsHYZ4(3,2,2,2,08) 
     $  + s09 * CsHYZ4(3,2,2,2,09) 
     $  + s10 * CsHYZ4(3,2,2,2,10) 
     $  + s11 * CsHYZ4(3,2,2,2,11) 
     $  + s12 * CsHYZ4(3,2,2,2,12) 
     $  + s13 * CsHYZ4(3,2,2,2,13) 
     $  + s14 * CsHYZ4(3,2,2,2,14) 
     $  + s15 * CsHYZ4(3,2,2,2,15) 
     $  + s16 * CsHYZ4(3,2,2,2,16) 
     $  + s17 * CsHYZ4(3,2,2,2,17) 
     $  - HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,3,0,2) = 
     $  + s01 * CsHYZ4(3,3,0,2,01) 
     $  + s02 * CsHYZ4(3,3,0,2,02) 
     $  + s03 * CsHYZ4(3,3,0,2,03) 
     $  + s04 * CsHYZ4(3,3,0,2,04) 
     $  + s05 * CsHYZ4(3,3,0,2,05) 
     $  + s06 * CsHYZ4(3,3,0,2,06) 
     $  + s07 * CsHYZ4(3,3,0,2,07) 
     $  + s08 * CsHYZ4(3,3,0,2,08) 
     $  + s09 * CsHYZ4(3,3,0,2,09) 
     $  + s10 * CsHYZ4(3,3,0,2,10) 
     $  + s11 * CsHYZ4(3,3,0,2,11) 
     $  + s12 * CsHYZ4(3,3,0,2,12) 
     $  + s13 * CsHYZ4(3,3,0,2,13) 
     $  + s14 * CsHYZ4(3,3,0,2,14) 
     $  + s15 * CsHYZ4(3,3,0,2,15) 
     $  + s16 * CsHYZ4(3,3,0,2,16) 
     $  + s17 * CsHYZ4(3,3,0,2,17) 
     $  - HZ2(0,1) *HYZ2(3,3)
     $  - HZ2(1,1) *HYZ2(3,3)
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,3,2,2) = 
     $  + s01 * CsHYZ4(3,3,2,2,01) 
     $  + s02 * CsHYZ4(3,3,2,2,02) 
     $  + s03 * CsHYZ4(3,3,2,2,03) 
     $  + s04 * CsHYZ4(3,3,2,2,04) 
     $  + s05 * CsHYZ4(3,3,2,2,05) 
     $  + s06 * CsHYZ4(3,3,2,2,06) 
     $  + s07 * CsHYZ4(3,3,2,2,07) 
     $  + s08 * CsHYZ4(3,3,2,2,08) 
     $  + s09 * CsHYZ4(3,3,2,2,09) 
     $  + s10 * CsHYZ4(3,3,2,2,10) 
     $  + s11 * CsHYZ4(3,3,2,2,11) 
     $  + s12 * CsHYZ4(3,3,2,2,12) 
     $  + s13 * CsHYZ4(3,3,2,2,13) 
     $  + s14 * CsHYZ4(3,3,2,2,14) 
     $  + s15 * CsHYZ4(3,3,2,2,15) 
     $  + s16 * CsHYZ4(3,3,2,2,16) 
     $  + s17 * CsHYZ4(3,3,2,2,17) 
     $  + HZ2(1,1) *HYZ2(3,3)
     $  + HZ3(0,1,1) *HYZ1(3)
     $  + HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,3,3,2) = 
     $  + s01 * CsHYZ4(3,3,3,2,01) 
     $  + s02 * CsHYZ4(3,3,3,2,02) 
     $  + s03 * CsHYZ4(3,3,3,2,03) 
     $  + s04 * CsHYZ4(3,3,3,2,04) 
     $  + s05 * CsHYZ4(3,3,3,2,05) 
     $  + s06 * CsHYZ4(3,3,3,2,06) 
     $  + s07 * CsHYZ4(3,3,3,2,07) 
     $  + s08 * CsHYZ4(3,3,3,2,08) 
     $  + s09 * CsHYZ4(3,3,3,2,09) 
     $  + s10 * CsHYZ4(3,3,3,2,10) 
     $  + s11 * CsHYZ4(3,3,3,2,11) 
     $  + s12 * CsHYZ4(3,3,3,2,12) 
     $  + s13 * CsHYZ4(3,3,3,2,13) 
     $  + s14 * CsHYZ4(3,3,3,2,14) 
     $  + s15 * CsHYZ4(3,3,3,2,15) 
     $  + s16 * CsHYZ4(3,3,3,2,16) 
     $  + s17 * CsHYZ4(3,3,3,2,17) 
     $  - HZ1(1) *HYZ3(3,3,3)
     $  - HZ2(0,1) *HYZ2(3,3)
     $  - HZ3(0,0,1) *HYZ1(3)
      endif
      
      return
      end


      subroutine fillirr2dhpl320e(iflag,n)
*********************************************************************
***** fillirr2dhpl320e(iflag,n) fills the irreducible 2dHPL     ***** 
***** with indices (3,2,0)                                      *****
***** up to weight n using the EXACT expressions for the        *****
***** z-dependent expansion coefficients                        *****
***** applicable for z>0.5                                      *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
      dimension CbHYZ1(3:3,22),CbHYZ2(0:3,0:3,22),
     $          CbHYZ3(0:3,0:3,0:3,22),CbHYZ4(0:3,0:3,0:3,0:3,22)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      common /b/
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
      common/aux/CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4
!$omp threadprivate(/HPL2/,/b/,/aux/)

      if (iflag.eq.1) then 
      call fillcoeff2dhplaux(iflag,n,CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4)
      call fillcoeff2dhpl320e(iflag,n,CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4)
      endif
* 2001-04-20:13:08:47.hpl 
* <- HPLexpand.320e 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      HYZ2(3,2) = 
     $  + b01 * CbHYZ2(3,2,01) 
     $  + b02 * CbHYZ2(3,2,02) 
     $  + b03 * CbHYZ2(3,2,03) 
     $  + b04 * CbHYZ2(3,2,04) 
     $  + b05 * CbHYZ2(3,2,05) 
     $  + b06 * CbHYZ2(3,2,06) 
     $  + b07 * CbHYZ2(3,2,07) 
     $  + b08 * CbHYZ2(3,2,08) 
     $  + b09 * CbHYZ2(3,2,09) 
     $  + b10 * CbHYZ2(3,2,10) 
     $  + b11 * CbHYZ2(3,2,11) 
     $  + b12 * CbHYZ2(3,2,12) 
     $  + b13 * CbHYZ2(3,2,13) 
     $  + b14 * CbHYZ2(3,2,14) 
     $  + b15 * CbHYZ2(3,2,15) 
     $  + b16 * CbHYZ2(3,2,16) 
     $  + b17 * CbHYZ2(3,2,17) 
     $  - HZ1(1) *HYZ1(3)
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(0,3,2) = 
     $  + b01 * CbHYZ3(0,3,2,01) 
     $  + b02 * CbHYZ3(0,3,2,02) 
     $  + b03 * CbHYZ3(0,3,2,03) 
     $  + b04 * CbHYZ3(0,3,2,04) 
     $  + b05 * CbHYZ3(0,3,2,05) 
     $  + b06 * CbHYZ3(0,3,2,06) 
     $  + b07 * CbHYZ3(0,3,2,07) 
     $  + b08 * CbHYZ3(0,3,2,08) 
     $  + b09 * CbHYZ3(0,3,2,09) 
     $  + b10 * CbHYZ3(0,3,2,10) 
     $  + b11 * CbHYZ3(0,3,2,11) 
     $  + b12 * CbHYZ3(0,3,2,12) 
     $  + b13 * CbHYZ3(0,3,2,13) 
     $  + b14 * CbHYZ3(0,3,2,14) 
     $  + b15 * CbHYZ3(0,3,2,15) 
     $  + b16 * CbHYZ3(0,3,2,16) 
     $  + b17 * CbHYZ3(0,3,2,17) 
     $  - HZ1(1) *HYZ2(0,3)
      HYZ3(3,0,2) = 
     $  + b01 * CbHYZ3(3,0,2,01) 
     $  + b02 * CbHYZ3(3,0,2,02) 
     $  + b03 * CbHYZ3(3,0,2,03) 
     $  + b04 * CbHYZ3(3,0,2,04) 
     $  + b05 * CbHYZ3(3,0,2,05) 
     $  + b06 * CbHYZ3(3,0,2,06) 
     $  + b07 * CbHYZ3(3,0,2,07) 
     $  + b08 * CbHYZ3(3,0,2,08) 
     $  + b09 * CbHYZ3(3,0,2,09) 
     $  + b10 * CbHYZ3(3,0,2,10) 
     $  + b11 * CbHYZ3(3,0,2,11) 
     $  + b12 * CbHYZ3(3,0,2,12) 
     $  + b13 * CbHYZ3(3,0,2,13) 
     $  + b14 * CbHYZ3(3,0,2,14) 
     $  + b15 * CbHYZ3(3,0,2,15) 
     $  + b16 * CbHYZ3(3,0,2,16) 
     $  + b17 * CbHYZ3(3,0,2,17) 
     $  - HZ2(0,1) *HYZ1(3)
     $  - HZ2(1,1) *HYZ1(3)
      HYZ3(3,2,2) = 
     $  + b01 * CbHYZ3(3,2,2,01) 
     $  + b02 * CbHYZ3(3,2,2,02) 
     $  + b03 * CbHYZ3(3,2,2,03) 
     $  + b04 * CbHYZ3(3,2,2,04) 
     $  + b05 * CbHYZ3(3,2,2,05) 
     $  + b06 * CbHYZ3(3,2,2,06) 
     $  + b07 * CbHYZ3(3,2,2,07) 
     $  + b08 * CbHYZ3(3,2,2,08) 
     $  + b09 * CbHYZ3(3,2,2,09) 
     $  + b10 * CbHYZ3(3,2,2,10) 
     $  + b11 * CbHYZ3(3,2,2,11) 
     $  + b12 * CbHYZ3(3,2,2,12) 
     $  + b13 * CbHYZ3(3,2,2,13) 
     $  + b14 * CbHYZ3(3,2,2,14) 
     $  + b15 * CbHYZ3(3,2,2,15) 
     $  + b16 * CbHYZ3(3,2,2,16) 
     $  + b17 * CbHYZ3(3,2,2,17) 
     $  + b18 * CbHYZ3(3,2,2,18) 
     $  + HZ2(1,1) *HYZ1(3)
      HYZ3(3,3,2) = 
     $  + b01 * CbHYZ3(3,3,2,01) 
     $  + b02 * CbHYZ3(3,3,2,02) 
     $  + b03 * CbHYZ3(3,3,2,03) 
     $  + b04 * CbHYZ3(3,3,2,04) 
     $  + b05 * CbHYZ3(3,3,2,05) 
     $  + b06 * CbHYZ3(3,3,2,06) 
     $  + b07 * CbHYZ3(3,3,2,07) 
     $  + b08 * CbHYZ3(3,3,2,08) 
     $  + b09 * CbHYZ3(3,3,2,09) 
     $  + b10 * CbHYZ3(3,3,2,10) 
     $  + b11 * CbHYZ3(3,3,2,11) 
     $  + b12 * CbHYZ3(3,3,2,12) 
     $  + b13 * CbHYZ3(3,3,2,13) 
     $  + b14 * CbHYZ3(3,3,2,14) 
     $  + b15 * CbHYZ3(3,3,2,15) 
     $  + b16 * CbHYZ3(3,3,2,16) 
     $  + b17 * CbHYZ3(3,3,2,17) 
     $  - HZ1(1) *HYZ2(3,3)
     $  - HZ2(0,1) *HYZ1(3)
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      HYZ4(0,0,3,2) = 
     $  + b01 * CbHYZ4(0,0,3,2,01) 
     $  + b02 * CbHYZ4(0,0,3,2,02) 
     $  + b03 * CbHYZ4(0,0,3,2,03) 
     $  + b04 * CbHYZ4(0,0,3,2,04) 
     $  + b05 * CbHYZ4(0,0,3,2,05) 
     $  + b06 * CbHYZ4(0,0,3,2,06) 
     $  + b07 * CbHYZ4(0,0,3,2,07) 
     $  + b08 * CbHYZ4(0,0,3,2,08) 
     $  + b09 * CbHYZ4(0,0,3,2,09) 
     $  + b10 * CbHYZ4(0,0,3,2,10) 
     $  + b11 * CbHYZ4(0,0,3,2,11) 
     $  + b12 * CbHYZ4(0,0,3,2,12) 
     $  + b13 * CbHYZ4(0,0,3,2,13) 
     $  + b14 * CbHYZ4(0,0,3,2,14) 
     $  + b15 * CbHYZ4(0,0,3,2,15) 
     $  + b16 * CbHYZ4(0,0,3,2,16) 
     $  + b17 * CbHYZ4(0,0,3,2,17) 
     $  - HZ1(1) *HYZ3(0,0,3)
      HYZ4(0,2,3,2) = 
     $  + b01 * CbHYZ4(0,2,3,2,01) 
     $  + b02 * CbHYZ4(0,2,3,2,02) 
     $  + b03 * CbHYZ4(0,2,3,2,03) 
     $  + b04 * CbHYZ4(0,2,3,2,04) 
     $  + b05 * CbHYZ4(0,2,3,2,05) 
     $  + b06 * CbHYZ4(0,2,3,2,06) 
     $  + b07 * CbHYZ4(0,2,3,2,07) 
     $  + b08 * CbHYZ4(0,2,3,2,08) 
     $  + b09 * CbHYZ4(0,2,3,2,09) 
     $  + b10 * CbHYZ4(0,2,3,2,10) 
     $  + b11 * CbHYZ4(0,2,3,2,11) 
     $  + b12 * CbHYZ4(0,2,3,2,12) 
     $  + b13 * CbHYZ4(0,2,3,2,13) 
     $  + b14 * CbHYZ4(0,2,3,2,14) 
     $  + b15 * CbHYZ4(0,2,3,2,15) 
     $  + b16 * CbHYZ4(0,2,3,2,16) 
     $  + b17 * CbHYZ4(0,2,3,2,17) 
     $  - HZ1(1) *HYZ3(0,2,3)
      HYZ4(0,3,0,2) = 
     $  + b01 * CbHYZ4(0,3,0,2,01) 
     $  + b02 * CbHYZ4(0,3,0,2,02) 
     $  + b03 * CbHYZ4(0,3,0,2,03) 
     $  + b04 * CbHYZ4(0,3,0,2,04) 
     $  + b05 * CbHYZ4(0,3,0,2,05) 
     $  + b06 * CbHYZ4(0,3,0,2,06) 
     $  + b07 * CbHYZ4(0,3,0,2,07) 
     $  + b08 * CbHYZ4(0,3,0,2,08) 
     $  + b09 * CbHYZ4(0,3,0,2,09) 
     $  + b10 * CbHYZ4(0,3,0,2,10) 
     $  + b11 * CbHYZ4(0,3,0,2,11) 
     $  + b12 * CbHYZ4(0,3,0,2,12) 
     $  + b13 * CbHYZ4(0,3,0,2,13) 
     $  + b14 * CbHYZ4(0,3,0,2,14) 
     $  + b15 * CbHYZ4(0,3,0,2,15) 
     $  + b16 * CbHYZ4(0,3,0,2,16) 
     $  + b17 * CbHYZ4(0,3,0,2,17) 
     $  - HZ2(0,1) *HYZ2(0,3)
     $  - HZ2(1,1) *HYZ2(0,3)
      HYZ4(0,3,2,2) = 
     $  + b01 * CbHYZ4(0,3,2,2,01) 
     $  + b02 * CbHYZ4(0,3,2,2,02) 
     $  + b03 * CbHYZ4(0,3,2,2,03) 
     $  + b04 * CbHYZ4(0,3,2,2,04) 
     $  + b05 * CbHYZ4(0,3,2,2,05) 
     $  + b06 * CbHYZ4(0,3,2,2,06) 
     $  + b07 * CbHYZ4(0,3,2,2,07) 
     $  + b08 * CbHYZ4(0,3,2,2,08) 
     $  + b09 * CbHYZ4(0,3,2,2,09) 
     $  + b10 * CbHYZ4(0,3,2,2,10) 
     $  + b11 * CbHYZ4(0,3,2,2,11) 
     $  + b12 * CbHYZ4(0,3,2,2,12) 
     $  + b13 * CbHYZ4(0,3,2,2,13) 
     $  + b14 * CbHYZ4(0,3,2,2,14) 
     $  + b15 * CbHYZ4(0,3,2,2,15) 
     $  + b16 * CbHYZ4(0,3,2,2,16) 
     $  + b17 * CbHYZ4(0,3,2,2,17) 
     $  + HZ2(1,1) *HYZ2(0,3)
      HYZ4(0,3,3,2) = 
     $  + b01 * CbHYZ4(0,3,3,2,01) 
     $  + b02 * CbHYZ4(0,3,3,2,02) 
     $  + b03 * CbHYZ4(0,3,3,2,03) 
     $  + b04 * CbHYZ4(0,3,3,2,04) 
     $  + b05 * CbHYZ4(0,3,3,2,05) 
     $  + b06 * CbHYZ4(0,3,3,2,06) 
     $  + b07 * CbHYZ4(0,3,3,2,07) 
     $  + b08 * CbHYZ4(0,3,3,2,08) 
     $  + b09 * CbHYZ4(0,3,3,2,09) 
     $  + b10 * CbHYZ4(0,3,3,2,10) 
     $  + b11 * CbHYZ4(0,3,3,2,11) 
     $  + b12 * CbHYZ4(0,3,3,2,12) 
     $  + b13 * CbHYZ4(0,3,3,2,13) 
     $  + b14 * CbHYZ4(0,3,3,2,14) 
     $  + b15 * CbHYZ4(0,3,3,2,15) 
     $  + b16 * CbHYZ4(0,3,3,2,16) 
     $  + b17 * CbHYZ4(0,3,3,2,17) 
     $  - HZ1(1) *HYZ3(0,3,3)
     $  - HZ2(0,1) *HYZ2(0,3)
      HYZ4(3,0,0,2) = 
     $  + b01 * CbHYZ4(3,0,0,2,01) 
     $  + b02 * CbHYZ4(3,0,0,2,02) 
     $  + b03 * CbHYZ4(3,0,0,2,03) 
     $  + b04 * CbHYZ4(3,0,0,2,04) 
     $  + b05 * CbHYZ4(3,0,0,2,05) 
     $  + b06 * CbHYZ4(3,0,0,2,06) 
     $  + b07 * CbHYZ4(3,0,0,2,07) 
     $  + b08 * CbHYZ4(3,0,0,2,08) 
     $  + b09 * CbHYZ4(3,0,0,2,09) 
     $  + b10 * CbHYZ4(3,0,0,2,10) 
     $  + b11 * CbHYZ4(3,0,0,2,11) 
     $  + b12 * CbHYZ4(3,0,0,2,12) 
     $  + b13 * CbHYZ4(3,0,0,2,13) 
     $  + b14 * CbHYZ4(3,0,0,2,14) 
     $  + b15 * CbHYZ4(3,0,0,2,15) 
     $  + b16 * CbHYZ4(3,0,0,2,16) 
     $  + b17 * CbHYZ4(3,0,0,2,17) 
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(0,1,1) *HYZ1(3)
     $  - HZ3(1,0,1) *HYZ1(3)
     $  - HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,0,2,2) = 
     $  + b01 * CbHYZ4(3,0,2,2,01) 
     $  + b02 * CbHYZ4(3,0,2,2,02) 
     $  + b03 * CbHYZ4(3,0,2,2,03) 
     $  + b04 * CbHYZ4(3,0,2,2,04) 
     $  + b05 * CbHYZ4(3,0,2,2,05) 
     $  + b06 * CbHYZ4(3,0,2,2,06) 
     $  + b07 * CbHYZ4(3,0,2,2,07) 
     $  + b08 * CbHYZ4(3,0,2,2,08) 
     $  + b09 * CbHYZ4(3,0,2,2,09) 
     $  + b10 * CbHYZ4(3,0,2,2,10) 
     $  + b11 * CbHYZ4(3,0,2,2,11) 
     $  + b12 * CbHYZ4(3,0,2,2,12) 
     $  + b13 * CbHYZ4(3,0,2,2,13) 
     $  + b14 * CbHYZ4(3,0,2,2,14) 
     $  + b15 * CbHYZ4(3,0,2,2,15) 
     $  + b16 * CbHYZ4(3,0,2,2,16) 
     $  + b17 * CbHYZ4(3,0,2,2,17) 
     $  + HZ3(0,1,1) *HYZ1(3)
     $  + HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,0,3,2) = 
     $  + b01 * CbHYZ4(3,0,3,2,01) 
     $  + b02 * CbHYZ4(3,0,3,2,02) 
     $  + b03 * CbHYZ4(3,0,3,2,03) 
     $  + b04 * CbHYZ4(3,0,3,2,04) 
     $  + b05 * CbHYZ4(3,0,3,2,05) 
     $  + b06 * CbHYZ4(3,0,3,2,06) 
     $  + b07 * CbHYZ4(3,0,3,2,07) 
     $  + b08 * CbHYZ4(3,0,3,2,08) 
     $  + b09 * CbHYZ4(3,0,3,2,09) 
     $  + b10 * CbHYZ4(3,0,3,2,10) 
     $  + b11 * CbHYZ4(3,0,3,2,11) 
     $  + b12 * CbHYZ4(3,0,3,2,12) 
     $  + b13 * CbHYZ4(3,0,3,2,13) 
     $  + b14 * CbHYZ4(3,0,3,2,14) 
     $  + b15 * CbHYZ4(3,0,3,2,15) 
     $  + b16 * CbHYZ4(3,0,3,2,16) 
     $  + b17 * CbHYZ4(3,0,3,2,17) 
     $  - HZ1(1) *HYZ3(3,0,3)
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(0,1,1) *HYZ1(3)
      HYZ4(3,2,2,2) = 
     $  + b01 * CbHYZ4(3,2,2,2,01) 
     $  + b02 * CbHYZ4(3,2,2,2,02) 
     $  + b03 * CbHYZ4(3,2,2,2,03) 
     $  + b04 * CbHYZ4(3,2,2,2,04) 
     $  + b05 * CbHYZ4(3,2,2,2,05) 
     $  + b06 * CbHYZ4(3,2,2,2,06) 
     $  + b07 * CbHYZ4(3,2,2,2,07) 
     $  + b08 * CbHYZ4(3,2,2,2,08) 
     $  + b09 * CbHYZ4(3,2,2,2,09) 
     $  + b10 * CbHYZ4(3,2,2,2,10) 
     $  + b11 * CbHYZ4(3,2,2,2,11) 
     $  + b12 * CbHYZ4(3,2,2,2,12) 
     $  + b13 * CbHYZ4(3,2,2,2,13) 
     $  + b14 * CbHYZ4(3,2,2,2,14) 
     $  + b15 * CbHYZ4(3,2,2,2,15) 
     $  + b16 * CbHYZ4(3,2,2,2,16) 
     $  + b17 * CbHYZ4(3,2,2,2,17) 
     $  + b18 * CbHYZ4(3,2,2,2,18) 
     $  - HZ3(1,1,1) *HYZ1(3)
      HYZ4(3,3,0,2) = 
     $  + b01 * CbHYZ4(3,3,0,2,01) 
     $  + b02 * CbHYZ4(3,3,0,2,02) 
     $  + b03 * CbHYZ4(3,3,0,2,03) 
     $  + b04 * CbHYZ4(3,3,0,2,04) 
     $  + b05 * CbHYZ4(3,3,0,2,05) 
     $  + b06 * CbHYZ4(3,3,0,2,06) 
     $  + b07 * CbHYZ4(3,3,0,2,07) 
     $  + b08 * CbHYZ4(3,3,0,2,08) 
     $  + b09 * CbHYZ4(3,3,0,2,09) 
     $  + b10 * CbHYZ4(3,3,0,2,10) 
     $  + b11 * CbHYZ4(3,3,0,2,11) 
     $  + b12 * CbHYZ4(3,3,0,2,12) 
     $  + b13 * CbHYZ4(3,3,0,2,13) 
     $  + b14 * CbHYZ4(3,3,0,2,14) 
     $  + b15 * CbHYZ4(3,3,0,2,15) 
     $  + b16 * CbHYZ4(3,3,0,2,16) 
     $  + b17 * CbHYZ4(3,3,0,2,17) 
     $  - HZ2(0,1) *HYZ2(3,3)
     $  - HZ2(1,1) *HYZ2(3,3)
     $  - HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,3,2,2) = 
     $  + b01 * CbHYZ4(3,3,2,2,01) 
     $  + b02 * CbHYZ4(3,3,2,2,02) 
     $  + b03 * CbHYZ4(3,3,2,2,03) 
     $  + b04 * CbHYZ4(3,3,2,2,04) 
     $  + b05 * CbHYZ4(3,3,2,2,05) 
     $  + b06 * CbHYZ4(3,3,2,2,06) 
     $  + b07 * CbHYZ4(3,3,2,2,07) 
     $  + b08 * CbHYZ4(3,3,2,2,08) 
     $  + b09 * CbHYZ4(3,3,2,2,09) 
     $  + b10 * CbHYZ4(3,3,2,2,10) 
     $  + b11 * CbHYZ4(3,3,2,2,11) 
     $  + b12 * CbHYZ4(3,3,2,2,12) 
     $  + b13 * CbHYZ4(3,3,2,2,13) 
     $  + b14 * CbHYZ4(3,3,2,2,14) 
     $  + b15 * CbHYZ4(3,3,2,2,15) 
     $  + b16 * CbHYZ4(3,3,2,2,16) 
     $  + b17 * CbHYZ4(3,3,2,2,17) 
     $  + b18 * CbHYZ4(3,3,2,2,18) 
     $  + HZ2(1,1) *HYZ2(3,3)
     $  + HZ3(0,1,1) *HYZ1(3)
     $  + HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,3,3,2) = 
     $  + b01 * CbHYZ4(3,3,3,2,01) 
     $  + b02 * CbHYZ4(3,3,3,2,02) 
     $  + b03 * CbHYZ4(3,3,3,2,03) 
     $  + b04 * CbHYZ4(3,3,3,2,04) 
     $  + b05 * CbHYZ4(3,3,3,2,05) 
     $  + b06 * CbHYZ4(3,3,3,2,06) 
     $  + b07 * CbHYZ4(3,3,3,2,07) 
     $  + b08 * CbHYZ4(3,3,3,2,08) 
     $  + b09 * CbHYZ4(3,3,3,2,09) 
     $  + b10 * CbHYZ4(3,3,3,2,10) 
     $  + b11 * CbHYZ4(3,3,3,2,11) 
     $  + b12 * CbHYZ4(3,3,3,2,12) 
     $  + b13 * CbHYZ4(3,3,3,2,13) 
     $  + b14 * CbHYZ4(3,3,3,2,14) 
     $  + b15 * CbHYZ4(3,3,3,2,15) 
     $  + b16 * CbHYZ4(3,3,3,2,16) 
     $  + b17 * CbHYZ4(3,3,3,2,17) 
     $  + b18 * CbHYZ4(3,3,3,2,18) 
     $  - HZ1(1) *HYZ3(3,3,3)
     $  - HZ2(0,1) *HYZ2(3,3)
     $  - HZ3(0,0,1) *HYZ1(3)
      endif
      return

      end

      subroutine fillirr2dhpl321(iflag,n)
*********************************************************************
***** fillirr2dhpl321(iflag,n) fills the irreducible 2dHPL      ***** 
***** with indices (3,2,1)                                      *****
***** up to weight n using the ONLY EXPANDED contributions      *****
***** to the z-dependent expansion coefficients                 *****
*****   applicable for z<0.5                                    *****
***** treatment of terms non-analytic in z=1:                   *****
*****        in expanded form (z<0.5: call fillcoeff2dhpl321u)  *****
***** this routine is also invoked by                           *****
*****    fillirr2dhpl321e, which handles the exact coefficients *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsHYZ2,CsHYZ3,CsHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      dimension CsHYZ2(0:3,0:3,22),CsHYZ3(0:3,0:3,0:3,22),
     $          CsHYZ4(0:3,0:3,0:3,0:3,22)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
      common /s/
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      save CsHYZ2,CsHYZ3,CsHYZ4
!$omp threadprivate(CsHYZ2,CsHYZ3,CsHYZ4,/s/,/HPL2/)
      
      if (iflag.eq.1.or.iflag.eq.-1) then
      call fillcoeff2dhpl321(iflag,n,CsHYZ2,CsHYZ3,CsHYZ4)
      if (iflag.gt.0) 
     $    call fillcoeff2dhpl321u(iflag,n,CsHYZ2,CsHYZ3,CsHYZ4)
      endif

* 2001-04-20:16:55:01.hpl 
* <- HPLexpand.321tcut 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(2,3,1) = 
     $  + s02 * CsHYZ3(2,3,1,02) 
     $  + s03 * CsHYZ3(2,3,1,03) 
     $  + s04 * CsHYZ3(2,3,1,04) 
     $  + s05 * CsHYZ3(2,3,1,05) 
     $  + s06 * CsHYZ3(2,3,1,06) 
     $  + s07 * CsHYZ3(2,3,1,07) 
     $  + s08 * CsHYZ3(2,3,1,08) 
     $  + s09 * CsHYZ3(2,3,1,09) 
     $  + s10 * CsHYZ3(2,3,1,10) 
     $  + s11 * CsHYZ3(2,3,1,11) 
     $  + s12 * CsHYZ3(2,3,1,12) 
     $  + s13 * CsHYZ3(2,3,1,13) 
     $  + s14 * CsHYZ3(2,3,1,14) 
     $  + s15 * CsHYZ3(2,3,1,15) 
     $  + s16 * CsHYZ3(2,3,1,16) 
     $  + s17 * CsHYZ3(2,3,1,17) 
     $  + s18 * CsHYZ3(2,3,1,18) 
     $  + s19 * CsHYZ3(2,3,1,19) 
     $  - HZ1( -1)*HYZ2(2,3)
      HYZ3(3,2,1) = 
     $  + s01 * CsHYZ3(3,2,1,01) 
     $  + s02 * CsHYZ3(3,2,1,02) 
     $  + s03 * CsHYZ3(3,2,1,03) 
     $  + s04 * CsHYZ3(3,2,1,04) 
     $  + s05 * CsHYZ3(3,2,1,05) 
     $  + s06 * CsHYZ3(3,2,1,06) 
     $  + s07 * CsHYZ3(3,2,1,07) 
     $  + s08 * CsHYZ3(3,2,1,08) 
     $  + s09 * CsHYZ3(3,2,1,09) 
     $  + s10 * CsHYZ3(3,2,1,10) 
     $  + s11 * CsHYZ3(3,2,1,11) 
     $  + s12 * CsHYZ3(3,2,1,12) 
     $  + s13 * CsHYZ3(3,2,1,13) 
     $  + s14 * CsHYZ3(3,2,1,14) 
     $  + s15 * CsHYZ3(3,2,1,15) 
     $  + s16 * CsHYZ3(3,2,1,16) 
     $  + s17 * CsHYZ3(3,2,1,17) 
     $  + s18 * CsHYZ3(3,2,1,18) 
     $  + s19 * CsHYZ3(3,2,1,19) 
     $  - HZ2(0, -1)*HYZ1(3)
     $  + HZ2(0,1) *HYZ1(3)
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      HYZ4(0,2,3,1) = 
     $  + s02 * CsHYZ4(0,2,3,1,02) 
     $  + s03 * CsHYZ4(0,2,3,1,03) 
     $  + s04 * CsHYZ4(0,2,3,1,04) 
     $  + s05 * CsHYZ4(0,2,3,1,05) 
     $  + s06 * CsHYZ4(0,2,3,1,06) 
     $  + s07 * CsHYZ4(0,2,3,1,07) 
     $  + s08 * CsHYZ4(0,2,3,1,08) 
     $  + s09 * CsHYZ4(0,2,3,1,09) 
     $  + s10 * CsHYZ4(0,2,3,1,10) 
     $  + s11 * CsHYZ4(0,2,3,1,11) 
     $  + s12 * CsHYZ4(0,2,3,1,12) 
     $  + s13 * CsHYZ4(0,2,3,1,13) 
     $  + s14 * CsHYZ4(0,2,3,1,14) 
     $  + s15 * CsHYZ4(0,2,3,1,15) 
     $  + s16 * CsHYZ4(0,2,3,1,16) 
     $  + s17 * CsHYZ4(0,2,3,1,17) 
     $  + s18 * CsHYZ4(0,2,3,1,18) 
     $  - HZ1( -1)*HYZ3(0,2,3)
      HYZ4(0,3,2,1) = 
     $  + s01 * CsHYZ4(0,3,2,1,01) 
     $  + s02 * CsHYZ4(0,3,2,1,02) 
     $  + s03 * CsHYZ4(0,3,2,1,03) 
     $  + s04 * CsHYZ4(0,3,2,1,04) 
     $  + s05 * CsHYZ4(0,3,2,1,05) 
     $  + s06 * CsHYZ4(0,3,2,1,06) 
     $  + s07 * CsHYZ4(0,3,2,1,07) 
     $  + s08 * CsHYZ4(0,3,2,1,08) 
     $  + s09 * CsHYZ4(0,3,2,1,09) 
     $  + s10 * CsHYZ4(0,3,2,1,10) 
     $  + s11 * CsHYZ4(0,3,2,1,11) 
     $  + s12 * CsHYZ4(0,3,2,1,12) 
     $  + s13 * CsHYZ4(0,3,2,1,13) 
     $  + s14 * CsHYZ4(0,3,2,1,14) 
     $  + s15 * CsHYZ4(0,3,2,1,15) 
     $  + s16 * CsHYZ4(0,3,2,1,16) 
     $  + s17 * CsHYZ4(0,3,2,1,17) 
     $  + s18 * CsHYZ4(0,3,2,1,18) 
     $  - HZ2(0, -1)*HYZ2(0,3)
     $  + HZ2(0,1) *HYZ2(0,3)
      HYZ4(2,0,3,1) = 
     $  + s02 * CsHYZ4(2,0,3,1,02) 
     $  + s03 * CsHYZ4(2,0,3,1,03) 
     $  + s04 * CsHYZ4(2,0,3,1,04) 
     $  + s05 * CsHYZ4(2,0,3,1,05) 
     $  + s06 * CsHYZ4(2,0,3,1,06) 
     $  + s07 * CsHYZ4(2,0,3,1,07) 
     $  + s08 * CsHYZ4(2,0,3,1,08) 
     $  + s09 * CsHYZ4(2,0,3,1,09) 
     $  + s10 * CsHYZ4(2,0,3,1,10) 
     $  + s11 * CsHYZ4(2,0,3,1,11) 
     $  + s12 * CsHYZ4(2,0,3,1,12) 
     $  + s13 * CsHYZ4(2,0,3,1,13) 
     $  + s14 * CsHYZ4(2,0,3,1,14) 
     $  + s15 * CsHYZ4(2,0,3,1,15) 
     $  + s16 * CsHYZ4(2,0,3,1,16) 
     $  + s17 * CsHYZ4(2,0,3,1,17) 
     $  + s18 * CsHYZ4(2,0,3,1,18) 
     $  - HZ1( -1)*HYZ3(2,0,3)
      HYZ4(2,2,3,1) = 
     $  + s03 * CsHYZ4(2,2,3,1,03) 
     $  + s04 * CsHYZ4(2,2,3,1,04) 
     $  + s05 * CsHYZ4(2,2,3,1,05) 
     $  + s06 * CsHYZ4(2,2,3,1,06) 
     $  + s07 * CsHYZ4(2,2,3,1,07) 
     $  + s08 * CsHYZ4(2,2,3,1,08) 
     $  + s09 * CsHYZ4(2,2,3,1,09) 
     $  + s10 * CsHYZ4(2,2,3,1,10) 
     $  + s11 * CsHYZ4(2,2,3,1,11) 
     $  + s12 * CsHYZ4(2,2,3,1,12) 
     $  + s13 * CsHYZ4(2,2,3,1,13) 
     $  + s14 * CsHYZ4(2,2,3,1,14) 
     $  + s15 * CsHYZ4(2,2,3,1,15) 
     $  + s16 * CsHYZ4(2,2,3,1,16) 
     $  + s17 * CsHYZ4(2,2,3,1,17) 
     $  + s18 * CsHYZ4(2,2,3,1,18) 
     $  - HZ1( -1)*HYZ3(2,2,3)
      HYZ4(2,3,0,1) = 
     $  + s02 * CsHYZ4(2,3,0,1,02) 
     $  + s03 * CsHYZ4(2,3,0,1,03) 
     $  + s04 * CsHYZ4(2,3,0,1,04) 
     $  + s05 * CsHYZ4(2,3,0,1,05) 
     $  + s06 * CsHYZ4(2,3,0,1,06) 
     $  + s07 * CsHYZ4(2,3,0,1,07) 
     $  + s08 * CsHYZ4(2,3,0,1,08) 
     $  + s09 * CsHYZ4(2,3,0,1,09) 
     $  + s10 * CsHYZ4(2,3,0,1,10) 
     $  + s11 * CsHYZ4(2,3,0,1,11) 
     $  + s12 * CsHYZ4(2,3,0,1,12) 
     $  + s13 * CsHYZ4(2,3,0,1,13) 
     $  + s14 * CsHYZ4(2,3,0,1,14) 
     $  + s15 * CsHYZ4(2,3,0,1,15) 
     $  + s16 * CsHYZ4(2,3,0,1,16) 
     $  + s17 * CsHYZ4(2,3,0,1,17) 
     $  + s18 * CsHYZ4(2,3,0,1,18) 
     $  - HZ2(0, -1)*HYZ2(2,3)
      HYZ4(2,3,1,1) = 
     $  + s02 * CsHYZ4(2,3,1,1,02) 
     $  + s03 * CsHYZ4(2,3,1,1,03) 
     $  + s04 * CsHYZ4(2,3,1,1,04) 
     $  + s05 * CsHYZ4(2,3,1,1,05) 
     $  + s06 * CsHYZ4(2,3,1,1,06) 
     $  + s07 * CsHYZ4(2,3,1,1,07) 
     $  + s08 * CsHYZ4(2,3,1,1,08) 
     $  + s09 * CsHYZ4(2,3,1,1,09) 
     $  + s10 * CsHYZ4(2,3,1,1,10) 
     $  + s11 * CsHYZ4(2,3,1,1,11) 
     $  + s12 * CsHYZ4(2,3,1,1,12) 
     $  + s13 * CsHYZ4(2,3,1,1,13) 
     $  + s14 * CsHYZ4(2,3,1,1,14) 
     $  + s15 * CsHYZ4(2,3,1,1,15) 
     $  + s16 * CsHYZ4(2,3,1,1,16) 
     $  + s17 * CsHYZ4(2,3,1,1,17) 
     $  + s18 * CsHYZ4(2,3,1,1,18) 
     $  + s19 * CsHYZ4(2,3,1,1,19) 
     $  + s20 * CsHYZ4(2,3,1,1,20) 
     $  + HZ2( -1,-1)*HYZ2(2,3)
      HYZ4(2,3,2,1) = 
     $  + s02 * CsHYZ4(2,3,2,1,02) 
     $  + s03 * CsHYZ4(2,3,2,1,03) 
     $  + s04 * CsHYZ4(2,3,2,1,04) 
     $  + s05 * CsHYZ4(2,3,2,1,05) 
     $  + s06 * CsHYZ4(2,3,2,1,06) 
     $  + s07 * CsHYZ4(2,3,2,1,07) 
     $  + s08 * CsHYZ4(2,3,2,1,08) 
     $  + s09 * CsHYZ4(2,3,2,1,09) 
     $  + s10 * CsHYZ4(2,3,2,1,10) 
     $  + s11 * CsHYZ4(2,3,2,1,11) 
     $  + s12 * CsHYZ4(2,3,2,1,12) 
     $  + s13 * CsHYZ4(2,3,2,1,13) 
     $  + s14 * CsHYZ4(2,3,2,1,14) 
     $  + s15 * CsHYZ4(2,3,2,1,15) 
     $  + s16 * CsHYZ4(2,3,2,1,16) 
     $  + s17 * CsHYZ4(2,3,2,1,17) 
     $  + s18 * CsHYZ4(2,3,2,1,18) 
     $  - HZ2(0, -1)*HYZ2(2,3)
     $  + HZ2(0,1) *HYZ2(2,3)
      HYZ4(2,3,3,1) = 
     $  + s02 * CsHYZ4(2,3,3,1,02) 
     $  + s03 * CsHYZ4(2,3,3,1,03) 
     $  + s04 * CsHYZ4(2,3,3,1,04) 
     $  + s05 * CsHYZ4(2,3,3,1,05) 
     $  + s06 * CsHYZ4(2,3,3,1,06) 
     $  + s07 * CsHYZ4(2,3,3,1,07) 
     $  + s08 * CsHYZ4(2,3,3,1,08) 
     $  + s09 * CsHYZ4(2,3,3,1,09) 
     $  + s10 * CsHYZ4(2,3,3,1,10) 
     $  + s11 * CsHYZ4(2,3,3,1,11) 
     $  + s12 * CsHYZ4(2,3,3,1,12) 
     $  + s13 * CsHYZ4(2,3,3,1,13) 
     $  + s14 * CsHYZ4(2,3,3,1,14) 
     $  + s15 * CsHYZ4(2,3,3,1,15) 
     $  + s16 * CsHYZ4(2,3,3,1,16) 
     $  + s17 * CsHYZ4(2,3,3,1,17) 
     $  + s18 * CsHYZ4(2,3,3,1,18) 
     $  - HZ1( -1)*HYZ3(2,3,3)
     $  + HZ2( -1,-1)*HYZ2(2,3)
     $  - HZ2(0, -1)*HYZ2(2,3)
      HYZ4(3,0,2,1) = 
     $  + s01 * CsHYZ4(3,0,2,1,01) 
     $  + s02 * CsHYZ4(3,0,2,1,02) 
     $  + s03 * CsHYZ4(3,0,2,1,03) 
     $  + s04 * CsHYZ4(3,0,2,1,04) 
     $  + s05 * CsHYZ4(3,0,2,1,05) 
     $  + s06 * CsHYZ4(3,0,2,1,06) 
     $  + s07 * CsHYZ4(3,0,2,1,07) 
     $  + s08 * CsHYZ4(3,0,2,1,08) 
     $  + s09 * CsHYZ4(3,0,2,1,09) 
     $  + s10 * CsHYZ4(3,0,2,1,10) 
     $  + s11 * CsHYZ4(3,0,2,1,11) 
     $  + s12 * CsHYZ4(3,0,2,1,12) 
     $  + s13 * CsHYZ4(3,0,2,1,13) 
     $  + s14 * CsHYZ4(3,0,2,1,14) 
     $  + s15 * CsHYZ4(3,0,2,1,15) 
     $  + s16 * CsHYZ4(3,0,2,1,16) 
     $  + s17 * CsHYZ4(3,0,2,1,17) 
     $  + s18 * CsHYZ4(3,0,2,1,18) 
     $  - 2.000000000000000d+00*HZ3(0,0,-1)*HYZ1(3)
     $  + 2.000000000000000d+00*HZ3(0,0,1)*HYZ1(3)
     $  + HZ3(0,1,1) *HYZ1(3)
     $  - 2.000000000000000d+00*HZ3(1,0,-1)*HYZ1(3)
     $  + HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,1,2,1) = 
     $  + s01 * CsHYZ4(3,1,2,1,01) 
     $  + s02 * CsHYZ4(3,1,2,1,02) 
     $  + s03 * CsHYZ4(3,1,2,1,03) 
     $  + s04 * CsHYZ4(3,1,2,1,04) 
     $  + s05 * CsHYZ4(3,1,2,1,05) 
     $  + s06 * CsHYZ4(3,1,2,1,06) 
     $  + s07 * CsHYZ4(3,1,2,1,07) 
     $  + s08 * CsHYZ4(3,1,2,1,08) 
     $  + s09 * CsHYZ4(3,1,2,1,09) 
     $  + s10 * CsHYZ4(3,1,2,1,10) 
     $  + s11 * CsHYZ4(3,1,2,1,11) 
     $  + s12 * CsHYZ4(3,1,2,1,12) 
     $  + s13 * CsHYZ4(3,1,2,1,13) 
     $  + s14 * CsHYZ4(3,1,2,1,14) 
     $  + s15 * CsHYZ4(3,1,2,1,15) 
     $  + s16 * CsHYZ4(3,1,2,1,16) 
     $  + s17 * CsHYZ4(3,1,2,1,17) 
     $  + s18 * CsHYZ4(3,1,2,1,18) 
     $  + s19 * CsHYZ4(3,1,2,1,19) 
     $  + s20 * CsHYZ4(3,1,2,1,20) 
     $  + s21 * CsHYZ4(3,1,2,1,21) 
     $  + HZ3( -1,0,-1)*HYZ1(3)
     $  - HZ3( -1,0,1)*HYZ1(3)
     $  - HZ3(0, -1,1)*HYZ1(3)
     $  - 2.000000000000000d+00*HZ3(0,0,-1)*HYZ1(3)
     $  + 2.000000000000000d+00*HZ3(0,0,1)*HYZ1(3)
     $  - HZ3(0,1, -1)*HYZ1(3)
      HYZ4(3,2,0,1) = 
     $  + s01 * CsHYZ4(3,2,0,1,01) 
     $  + s02 * CsHYZ4(3,2,0,1,02) 
     $  + s03 * CsHYZ4(3,2,0,1,03) 
     $  + s04 * CsHYZ4(3,2,0,1,04) 
     $  + s05 * CsHYZ4(3,2,0,1,05) 
     $  + s06 * CsHYZ4(3,2,0,1,06) 
     $  + s07 * CsHYZ4(3,2,0,1,07) 
     $  + s08 * CsHYZ4(3,2,0,1,08) 
     $  + s09 * CsHYZ4(3,2,0,1,09) 
     $  + s10 * CsHYZ4(3,2,0,1,10) 
     $  + s11 * CsHYZ4(3,2,0,1,11) 
     $  + s12 * CsHYZ4(3,2,0,1,12) 
     $  + s13 * CsHYZ4(3,2,0,1,13) 
     $  + s14 * CsHYZ4(3,2,0,1,14) 
     $  + s15 * CsHYZ4(3,2,0,1,15) 
     $  + s16 * CsHYZ4(3,2,0,1,16) 
     $  + s17 * CsHYZ4(3,2,0,1,17) 
     $  + s18 * CsHYZ4(3,2,0,1,18) 
     $  + 2.000000000000000d+00*HZ3(1,0,-1)*HYZ1(3)
     $  - HZ3(1,0,1) *HYZ1(3)
      HYZ4(3,2,1,1) = 
     $  + s01 * CsHYZ4(3,2,1,1,01) 
     $  + s02 * CsHYZ4(3,2,1,1,02) 
     $  + s03 * CsHYZ4(3,2,1,1,03) 
     $  + s04 * CsHYZ4(3,2,1,1,04) 
     $  + s05 * CsHYZ4(3,2,1,1,05) 
     $  + s06 * CsHYZ4(3,2,1,1,06) 
     $  + s07 * CsHYZ4(3,2,1,1,07) 
     $  + s08 * CsHYZ4(3,2,1,1,08) 
     $  + s09 * CsHYZ4(3,2,1,1,09) 
     $  + s10 * CsHYZ4(3,2,1,1,10) 
     $  + s11 * CsHYZ4(3,2,1,1,11) 
     $  + s12 * CsHYZ4(3,2,1,1,12) 
     $  + s13 * CsHYZ4(3,2,1,1,13) 
     $  + s14 * CsHYZ4(3,2,1,1,14) 
     $  + s15 * CsHYZ4(3,2,1,1,15) 
     $  + s16 * CsHYZ4(3,2,1,1,16) 
     $  + s17 * CsHYZ4(3,2,1,1,17) 
     $  + s18 * CsHYZ4(3,2,1,1,18) 
     $  + s19 * CsHYZ4(3,2,1,1,19) 
     $  + s20 * CsHYZ4(3,2,1,1,20) 
     $  + HZ3(0, -1,-1)*HYZ1(3)
     $  + HZ3(0,0, -1)*HYZ1(3)
     $  - HZ3(0,0,1) *HYZ1(3)
      HYZ4(3,2,2,1) = 
     $  + s01 * CsHYZ4(3,2,2,1,01) 
     $  + s02 * CsHYZ4(3,2,2,1,02) 
     $  + s03 * CsHYZ4(3,2,2,1,03) 
     $  + s04 * CsHYZ4(3,2,2,1,04) 
     $  + s05 * CsHYZ4(3,2,2,1,05) 
     $  + s06 * CsHYZ4(3,2,2,1,06) 
     $  + s07 * CsHYZ4(3,2,2,1,07) 
     $  + s08 * CsHYZ4(3,2,2,1,08) 
     $  + s09 * CsHYZ4(3,2,2,1,09) 
     $  + s10 * CsHYZ4(3,2,2,1,10) 
     $  + s11 * CsHYZ4(3,2,2,1,11) 
     $  + s12 * CsHYZ4(3,2,2,1,12) 
     $  + s13 * CsHYZ4(3,2,2,1,13) 
     $  + s14 * CsHYZ4(3,2,2,1,14) 
     $  + s15 * CsHYZ4(3,2,2,1,15) 
     $  + s16 * CsHYZ4(3,2,2,1,16) 
     $  + s17 * CsHYZ4(3,2,2,1,17) 
     $  + s18 * CsHYZ4(3,2,2,1,18) 
     $  - HZ3(0,0, -1)*HYZ1(3)
     $  + HZ3(0,0,1) *HYZ1(3)
     $  - HZ3(0,1,1) *HYZ1(3)
      HYZ4(3,2,3,1) = 
     $  + s01 * CsHYZ4(3,2,3,1,01) 
     $  + s02 * CsHYZ4(3,2,3,1,02) 
     $  + s03 * CsHYZ4(3,2,3,1,03) 
     $  + s04 * CsHYZ4(3,2,3,1,04) 
     $  + s05 * CsHYZ4(3,2,3,1,05) 
     $  + s06 * CsHYZ4(3,2,3,1,06) 
     $  + s07 * CsHYZ4(3,2,3,1,07) 
     $  + s08 * CsHYZ4(3,2,3,1,08) 
     $  + s09 * CsHYZ4(3,2,3,1,09) 
     $  + s10 * CsHYZ4(3,2,3,1,10) 
     $  + s11 * CsHYZ4(3,2,3,1,11) 
     $  + s12 * CsHYZ4(3,2,3,1,12) 
     $  + s13 * CsHYZ4(3,2,3,1,13) 
     $  + s14 * CsHYZ4(3,2,3,1,14) 
     $  + s15 * CsHYZ4(3,2,3,1,15) 
     $  + s16 * CsHYZ4(3,2,3,1,16) 
     $  + s17 * CsHYZ4(3,2,3,1,17) 
     $  + s18 * CsHYZ4(3,2,3,1,18) 
     $  - HZ1( -1)*HYZ3(3,2,3)
     $  + HZ3( -1,0,-1)*HYZ1(3)
     $  - HZ3( -1,0,1)*HYZ1(3)
     $  + HZ3(0, -1,1)*HYZ1(3)
     $  + HZ3(0,1, -1)*HYZ1(3)
      HYZ4(3,3,2,1) = 
     $  + s01 * CsHYZ4(3,3,2,1,01) 
     $  + s02 * CsHYZ4(3,3,2,1,02) 
     $  + s03 * CsHYZ4(3,3,2,1,03) 
     $  + s04 * CsHYZ4(3,3,2,1,04) 
     $  + s05 * CsHYZ4(3,3,2,1,05) 
     $  + s06 * CsHYZ4(3,3,2,1,06) 
     $  + s07 * CsHYZ4(3,3,2,1,07) 
     $  + s08 * CsHYZ4(3,3,2,1,08) 
     $  + s09 * CsHYZ4(3,3,2,1,09) 
     $  + s10 * CsHYZ4(3,3,2,1,10) 
     $  + s11 * CsHYZ4(3,3,2,1,11) 
     $  + s12 * CsHYZ4(3,3,2,1,12) 
     $  + s13 * CsHYZ4(3,3,2,1,13) 
     $  + s14 * CsHYZ4(3,3,2,1,14) 
     $  + s15 * CsHYZ4(3,3,2,1,15) 
     $  + s16 * CsHYZ4(3,3,2,1,16) 
     $  + s17 * CsHYZ4(3,3,2,1,17) 
     $  + s18 * CsHYZ4(3,3,2,1,18) 
     $  - HZ2(0, -1)*HYZ2(3,3)
     $  + HZ2(0,1) *HYZ2(3,3)
     $  + HZ3(0, -1,-1)*HYZ1(3)
     $  - 2.000000000000000d+00*HZ3(0,0,-1)*HYZ1(3)
     $  + 2.000000000000000d+00*HZ3(0,0,1)*HYZ1(3)
      endif

      return
      end

      subroutine fillirr2dhpl321e(iflag,n)
*********************************************************************
***** fillirr2dhpl321e(iflag,n) fills the irreducible 2dHPL     ***** 
***** with indices (3,2,1)                                      *****
***** up to weight n with EXPANDED z-dependent                  *****
***** expansion coefficients for all terms                      *****
***** which are analytic in z=1                                 *****
***** and EXACT z-dependent                                     *****
***** expansion coefficients for all terms                      *****
***** terms non-analytic in z=1,                                *****
*****       (z>0.5: call fillcoeff2dhpl321e)                    *****
*****   applicable for z>0.5                                    *****
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp)::
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
      dimension CbHYZ1(3:3,22),CbHYZ2(0:3,0:3,22),
     $          CbHYZ3(0:3,0:3,0:3,22),CbHYZ4(0:3,0:3,0:3,0:3,22)
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)
      common/aux/CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4
      common /b/
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
!$omp threadprivate(/HPL2/,/b/,/aux/)

      iflag = -iflag 
      call fillirr2dhpl321(iflag,n)
      iflag = -iflag
      if (iflag.eq.1) then
      call fillcoeff2dhplaux(iflag,n,CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4)
      call fillcoeff2dhpl321e(iflag,n,CbHYZ1,CbHYZ2,CbHYZ3,CbHYZ4)
      endif

* 2001-04-20:17:05:41.hpl 
* <- HPLexpand.321enew 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.2) return

      if (n.eq.3) then
      HYZ3(3,2,1) = HYZ3(3,2,1)  
     $  + b01 * CbHYZ3(3,2,1,01) 
     $  + b02 * CbHYZ3(3,2,1,02) 
     $  + b03 * CbHYZ3(3,2,1,03) 
     $  + b04 * CbHYZ3(3,2,1,04) 
     $  + b05 * CbHYZ3(3,2,1,05) 
     $  + b06 * CbHYZ3(3,2,1,06) 
     $  + b07 * CbHYZ3(3,2,1,07) 
     $  + b08 * CbHYZ3(3,2,1,08) 
     $  + b09 * CbHYZ3(3,2,1,09) 
     $  + b10 * CbHYZ3(3,2,1,10) 
     $  + b11 * CbHYZ3(3,2,1,11) 
     $  + b12 * CbHYZ3(3,2,1,12) 
     $  + b13 * CbHYZ3(3,2,1,13) 
     $  + b14 * CbHYZ3(3,2,1,14) 
     $  + b15 * CbHYZ3(3,2,1,15) 
     $  + b16 * CbHYZ3(3,2,1,16) 
     $  + b17 * CbHYZ3(3,2,1,17) 
     $  + b18 * CbHYZ3(3,2,1,18) 
     $  + b19 * CbHYZ3(3,2,1,19) 
     $  + b20 * CbHYZ3(3,2,1,20) 
      endif
      if (n.eq.3) return
      
      if (n.eq.4) then
      HYZ4(0,3,2,1) = HYZ4(0,3,2,1) 
     $  + b01 * CbHYZ4(0,3,2,1,01) 
     $  + b02 * CbHYZ4(0,3,2,1,02) 
     $  + b03 * CbHYZ4(0,3,2,1,03) 
     $  + b04 * CbHYZ4(0,3,2,1,04) 
     $  + b05 * CbHYZ4(0,3,2,1,05) 
     $  + b06 * CbHYZ4(0,3,2,1,06) 
     $  + b07 * CbHYZ4(0,3,2,1,07) 
     $  + b08 * CbHYZ4(0,3,2,1,08) 
     $  + b09 * CbHYZ4(0,3,2,1,09) 
     $  + b10 * CbHYZ4(0,3,2,1,10) 
     $  + b11 * CbHYZ4(0,3,2,1,11) 
     $  + b12 * CbHYZ4(0,3,2,1,12) 
     $  + b13 * CbHYZ4(0,3,2,1,13) 
     $  + b14 * CbHYZ4(0,3,2,1,14) 
     $  + b15 * CbHYZ4(0,3,2,1,15) 
     $  + b16 * CbHYZ4(0,3,2,1,16) 
     $  + b17 * CbHYZ4(0,3,2,1,17) 
     $  + b18 * CbHYZ4(0,3,2,1,18) 
     $  + b19 * CbHYZ4(0,3,2,1,19) 
      HYZ4(2,3,2,1) = HYZ4(2,3,2,1) 
     $  + b01 * CbHYZ4(2,3,2,1,01) 
     $  + b02 * CbHYZ4(2,3,2,1,02) 
     $  + b03 * CbHYZ4(2,3,2,1,03) 
     $  + b04 * CbHYZ4(2,3,2,1,04) 
     $  + b05 * CbHYZ4(2,3,2,1,05) 
     $  + b06 * CbHYZ4(2,3,2,1,06) 
     $  + b07 * CbHYZ4(2,3,2,1,07) 
     $  + b08 * CbHYZ4(2,3,2,1,08) 
     $  + b09 * CbHYZ4(2,3,2,1,09) 
     $  + b10 * CbHYZ4(2,3,2,1,10) 
     $  + b11 * CbHYZ4(2,3,2,1,11) 
     $  + b12 * CbHYZ4(2,3,2,1,12) 
     $  + b13 * CbHYZ4(2,3,2,1,13) 
     $  + b14 * CbHYZ4(2,3,2,1,14) 
     $  + b15 * CbHYZ4(2,3,2,1,15) 
     $  + b16 * CbHYZ4(2,3,2,1,16) 
     $  + b17 * CbHYZ4(2,3,2,1,17) 
     $  + b18 * CbHYZ4(2,3,2,1,18) 
     $  + b19 * CbHYZ4(2,3,2,1,19) 
      HYZ4(3,0,2,1) = HYZ4(3,0,2,1) 
     $  + b01 * CbHYZ4(3,0,2,1,01) 
     $  + b02 * CbHYZ4(3,0,2,1,02) 
     $  + b03 * CbHYZ4(3,0,2,1,03) 
     $  + b04 * CbHYZ4(3,0,2,1,04) 
     $  + b05 * CbHYZ4(3,0,2,1,05) 
     $  + b06 * CbHYZ4(3,0,2,1,06) 
     $  + b07 * CbHYZ4(3,0,2,1,07) 
     $  + b08 * CbHYZ4(3,0,2,1,08) 
     $  + b09 * CbHYZ4(3,0,2,1,09) 
     $  + b10 * CbHYZ4(3,0,2,1,10) 
     $  + b11 * CbHYZ4(3,0,2,1,11) 
     $  + b12 * CbHYZ4(3,0,2,1,12) 
     $  + b13 * CbHYZ4(3,0,2,1,13) 
     $  + b14 * CbHYZ4(3,0,2,1,14) 
     $  + b15 * CbHYZ4(3,0,2,1,15) 
     $  + b16 * CbHYZ4(3,0,2,1,16) 
     $  + b17 * CbHYZ4(3,0,2,1,17) 
     $  + b18 * CbHYZ4(3,0,2,1,18) 
     $  + b19 * CbHYZ4(3,0,2,1,19) 
      HYZ4(3,1,2,1) = HYZ4(3,1,2,1) 
     $  + b01 * CbHYZ4(3,1,2,1,01) 
     $  + b02 * CbHYZ4(3,1,2,1,02) 
     $  + b03 * CbHYZ4(3,1,2,1,03) 
     $  + b04 * CbHYZ4(3,1,2,1,04) 
     $  + b05 * CbHYZ4(3,1,2,1,05) 
     $  + b06 * CbHYZ4(3,1,2,1,06) 
     $  + b07 * CbHYZ4(3,1,2,1,07) 
     $  + b08 * CbHYZ4(3,1,2,1,08) 
     $  + b09 * CbHYZ4(3,1,2,1,09) 
     $  + b10 * CbHYZ4(3,1,2,1,10) 
     $  + b11 * CbHYZ4(3,1,2,1,11) 
     $  + b12 * CbHYZ4(3,1,2,1,12) 
     $  + b13 * CbHYZ4(3,1,2,1,13) 
     $  + b14 * CbHYZ4(3,1,2,1,14) 
     $  + b15 * CbHYZ4(3,1,2,1,15) 
     $  + b16 * CbHYZ4(3,1,2,1,16) 
     $  + b17 * CbHYZ4(3,1,2,1,17) 
     $  + b18 * CbHYZ4(3,1,2,1,18) 
     $  + b19 * CbHYZ4(3,1,2,1,19) 
     $  + b20 * CbHYZ4(3,1,2,1,20) 
     $  + b21 * CbHYZ4(3,1,2,1,21) 
     $  + b22 * CbHYZ4(3,1,2,1,22) 
      HYZ4(3,2,0,1) = HYZ4(3,2,0,1) 
     $  + b01 * CbHYZ4(3,2,0,1,01) 
     $  + b02 * CbHYZ4(3,2,0,1,02) 
     $  + b03 * CbHYZ4(3,2,0,1,03) 
     $  + b04 * CbHYZ4(3,2,0,1,04) 
     $  + b05 * CbHYZ4(3,2,0,1,05) 
     $  + b06 * CbHYZ4(3,2,0,1,06) 
     $  + b07 * CbHYZ4(3,2,0,1,07) 
     $  + b08 * CbHYZ4(3,2,0,1,08) 
     $  + b09 * CbHYZ4(3,2,0,1,09) 
     $  + b10 * CbHYZ4(3,2,0,1,10) 
     $  + b11 * CbHYZ4(3,2,0,1,11) 
     $  + b12 * CbHYZ4(3,2,0,1,12) 
     $  + b13 * CbHYZ4(3,2,0,1,13) 
     $  + b14 * CbHYZ4(3,2,0,1,14) 
     $  + b15 * CbHYZ4(3,2,0,1,15) 
     $  + b16 * CbHYZ4(3,2,0,1,16) 
     $  + b17 * CbHYZ4(3,2,0,1,17) 
     $  + b18 * CbHYZ4(3,2,0,1,18) 
     $  + b19 * CbHYZ4(3,2,0,1,19) 
      HYZ4(3,2,1,1) = HYZ4(3,2,1,1) 
     $  + b01 * CbHYZ4(3,2,1,1,01) 
     $  + b02 * CbHYZ4(3,2,1,1,02) 
     $  + b03 * CbHYZ4(3,2,1,1,03) 
     $  + b04 * CbHYZ4(3,2,1,1,04) 
     $  + b05 * CbHYZ4(3,2,1,1,05) 
     $  + b06 * CbHYZ4(3,2,1,1,06) 
     $  + b07 * CbHYZ4(3,2,1,1,07) 
     $  + b08 * CbHYZ4(3,2,1,1,08) 
     $  + b09 * CbHYZ4(3,2,1,1,09) 
     $  + b10 * CbHYZ4(3,2,1,1,10) 
     $  + b11 * CbHYZ4(3,2,1,1,11) 
     $  + b12 * CbHYZ4(3,2,1,1,12) 
     $  + b13 * CbHYZ4(3,2,1,1,13) 
     $  + b14 * CbHYZ4(3,2,1,1,14) 
     $  + b15 * CbHYZ4(3,2,1,1,15) 
     $  + b16 * CbHYZ4(3,2,1,1,16) 
     $  + b17 * CbHYZ4(3,2,1,1,17) 
     $  + b18 * CbHYZ4(3,2,1,1,18) 
     $  + b19 * CbHYZ4(3,2,1,1,19) 
     $  + b20 * CbHYZ4(3,2,1,1,20) 
     $  + b21 * CbHYZ4(3,2,1,1,21) 
      HYZ4(3,2,2,1) = HYZ4(3,2,2,1) 
     $  + b01 * CbHYZ4(3,2,2,1,01) 
     $  + b02 * CbHYZ4(3,2,2,1,02) 
     $  + b03 * CbHYZ4(3,2,2,1,03) 
     $  + b04 * CbHYZ4(3,2,2,1,04) 
     $  + b05 * CbHYZ4(3,2,2,1,05) 
     $  + b06 * CbHYZ4(3,2,2,1,06) 
     $  + b07 * CbHYZ4(3,2,2,1,07) 
     $  + b08 * CbHYZ4(3,2,2,1,08) 
     $  + b09 * CbHYZ4(3,2,2,1,09) 
     $  + b10 * CbHYZ4(3,2,2,1,10) 
     $  + b11 * CbHYZ4(3,2,2,1,11) 
     $  + b12 * CbHYZ4(3,2,2,1,12) 
     $  + b13 * CbHYZ4(3,2,2,1,13) 
     $  + b14 * CbHYZ4(3,2,2,1,14) 
     $  + b15 * CbHYZ4(3,2,2,1,15) 
     $  + b16 * CbHYZ4(3,2,2,1,16) 
     $  + b17 * CbHYZ4(3,2,2,1,17) 
     $  + b18 * CbHYZ4(3,2,2,1,18) 
     $  + b19 * CbHYZ4(3,2,2,1,19) 
      HYZ4(3,2,3,1) = HYZ4(3,2,3,1) 
     $  + b01 * CbHYZ4(3,2,3,1,01) 
     $  + b02 * CbHYZ4(3,2,3,1,02) 
     $  + b03 * CbHYZ4(3,2,3,1,03) 
     $  + b04 * CbHYZ4(3,2,3,1,04) 
     $  + b05 * CbHYZ4(3,2,3,1,05) 
     $  + b06 * CbHYZ4(3,2,3,1,06) 
     $  + b07 * CbHYZ4(3,2,3,1,07) 
     $  + b08 * CbHYZ4(3,2,3,1,08) 
     $  + b09 * CbHYZ4(3,2,3,1,09) 
     $  + b10 * CbHYZ4(3,2,3,1,10) 
     $  + b11 * CbHYZ4(3,2,3,1,11) 
     $  + b12 * CbHYZ4(3,2,3,1,12) 
     $  + b13 * CbHYZ4(3,2,3,1,13) 
     $  + b14 * CbHYZ4(3,2,3,1,14) 
     $  + b15 * CbHYZ4(3,2,3,1,15) 
     $  + b16 * CbHYZ4(3,2,3,1,16) 
     $  + b17 * CbHYZ4(3,2,3,1,17) 
     $  + b18 * CbHYZ4(3,2,3,1,18) 
     $  + b19 * CbHYZ4(3,2,3,1,19) 
      HYZ4(3,3,2,1) = HYZ4(3,3,2,1) 
     $  + b01 * CbHYZ4(3,3,2,1,01) 
     $  + b02 * CbHYZ4(3,3,2,1,02) 
     $  + b03 * CbHYZ4(3,3,2,1,03) 
     $  + b04 * CbHYZ4(3,3,2,1,04) 
     $  + b05 * CbHYZ4(3,3,2,1,05) 
     $  + b06 * CbHYZ4(3,3,2,1,06) 
     $  + b07 * CbHYZ4(3,3,2,1,07) 
     $  + b08 * CbHYZ4(3,3,2,1,08) 
     $  + b09 * CbHYZ4(3,3,2,1,09) 
     $  + b10 * CbHYZ4(3,3,2,1,10) 
     $  + b11 * CbHYZ4(3,3,2,1,11) 
     $  + b12 * CbHYZ4(3,3,2,1,12) 
     $  + b13 * CbHYZ4(3,3,2,1,13) 
     $  + b14 * CbHYZ4(3,3,2,1,14) 
     $  + b15 * CbHYZ4(3,3,2,1,15) 
     $  + b16 * CbHYZ4(3,3,2,1,16) 
     $  + b17 * CbHYZ4(3,3,2,1,17) 
     $  + b18 * CbHYZ4(3,3,2,1,18) 
     $  + b19 * CbHYZ4(3,3,2,1,19) 
      endif

      return
      end


      subroutine fillred2dhpl(iflag,n)
*********************************************************************
*****  fillred2dhpl evaluates the reducible 2dhpl              ******
*****  irreducible 2dhpl need to be present in HPL2 already    ******
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3)       
!$omp threadprivate(/HPL2/)

* 2001-04-23:14:13:53.hpl 
* <- red.out 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      HYZ2(0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)
      HYZ2(1,0) = 
     $  + HYZ1(0) *HYZ1(1)
     $  - HYZ2(0,1) 
      HYZ2(1,1) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)
      HYZ2(1,2) = 
     $  + HYZ1(1) *HYZ1(2)
     $  - HYZ2(2,1) 
      HYZ2(1,3) = 
     $  + HYZ1(1) *HYZ1(3)
     $  - HYZ2(3,1) 
      HYZ2(2,0) = 
     $  + HYZ1(0) *HYZ1(2)
     $  - HYZ2(0,2) 
      HYZ2(2,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)
      HYZ2(2,3) = 
     $  + HYZ1(2) *HYZ1(3)
     $  - HYZ2(3,2) 
      HYZ2(3,0) = 
     $  + HYZ1(0) *HYZ1(3)
     $  - HYZ2(0,3) 
      HYZ2(3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)
      endif
      if (n.eq.2) return
      
      if (n.eq.3) then
      HYZ3(0,0,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(0)*HYZ1(0)
      HYZ3(0,1,0) = 
     $  + HYZ1(0) *HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ3(0,0,1)
      HYZ3(0,1,2) = 
     $  + HYZ1(2) *HYZ2(0,1)
     $  - HYZ3(0,2,1) 
     $  - HYZ3(2,0,1) 
      HYZ3(0,1,3) = 
     $  + HYZ1(3) *HYZ2(0,1)
     $  - HYZ3(0,3,1) 
     $  - HYZ3(3,0,1) 
      HYZ3(0,2,0) = 
     $  + HYZ1(0) *HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ3(0,0,2)
      HYZ3(0,2,3) = 
     $  + HYZ1(3) *HYZ2(0,2)
     $  - HYZ3(0,3,2) 
     $  - HYZ3(3,0,2) 
      HYZ3(0,3,0) = 
     $  + HYZ1(0) *HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ3(0,0,3)
      HYZ3(1,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(1)
     $  - HYZ1(0) *HYZ2(0,1)
     $  + HYZ3(0,0,1) 
      HYZ3(1,0,1) = 
     $  + HYZ1(1) *HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ3(0,1,1)
      HYZ3(1,0,2) = 
     $  + HYZ1(1) *HYZ2(0,2)
     $  - HYZ1(2) *HYZ2(0,1)
     $  + HYZ3(2,0,1) 
      HYZ3(1,0,3) = 
     $  + HYZ1(1) *HYZ2(0,3)
     $  - HYZ1(3) *HYZ2(0,1)
     $  + HYZ3(3,0,1) 
      HYZ3(1,1,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(1)*HYZ1(1)
     $  - HYZ1(1) *HYZ2(0,1)
     $  + HYZ3(0,1,1) 
      HYZ3(1,1,1) = 
     $  + 1.666666666666666d-01*HYZ1(1)*HYZ1(1)*HYZ1(1)
      HYZ3(1,1,2) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ1(2)
     $  - HYZ1(1) *HYZ2(2,1)
     $  + HYZ3(2,1,1) 
      HYZ3(1,1,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ1(3)
     $  - HYZ1(1) *HYZ2(3,1)
     $  + HYZ3(3,1,1) 
      HYZ3(1,2,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ1(2)
     $  - HYZ1(0) *HYZ2(2,1)
     $  - HYZ1(1) *HYZ2(0,2)
     $  + HYZ3(0,2,1) 
      HYZ3(1,2,1) = 
     $  + HYZ1(1) *HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ3(2,1,1)
      HYZ3(1,2,2) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(2)*HYZ1(2)
     $  - HYZ1(2) *HYZ2(2,1)
     $  + HYZ3(2,2,1) 
      HYZ3(1,2,3) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ1(3)
     $  - HYZ1(1) *HYZ2(3,2)
     $  - HYZ1(3) *HYZ2(2,1)
     $  + HYZ3(3,2,1) 
      HYZ3(1,3,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ1(3)
     $  - HYZ1(0) *HYZ2(3,1)
     $  - HYZ1(1) *HYZ2(0,3)
     $  + HYZ3(0,3,1) 
      HYZ3(1,3,1) = 
     $  + HYZ1(1) *HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ3(3,1,1)
      HYZ3(1,3,2) = 
     $  + HYZ1(1) *HYZ2(3,2)
     $  - HYZ1(2) *HYZ2(3,1)
     $  + HYZ3(2,3,1) 
      HYZ3(1,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(3) *HYZ2(3,1)
     $  + HYZ3(3,3,1) 
      HYZ3(2,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(2)
     $  - HYZ1(0) *HYZ2(0,2)
     $  + HYZ3(0,0,2) 
      HYZ3(2,0,2) = 
     $  + HYZ1(2) *HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ3(0,2,2)
      HYZ3(2,0,3) = 
     $  + HYZ1(2) *HYZ2(0,3)
     $  - HYZ1(3) *HYZ2(0,2)
     $  + HYZ3(3,0,2) 
      HYZ3(2,1,0) = 
     $  + HYZ1(0) *HYZ2(2,1)
     $  - HYZ3(0,2,1) 
     $  - HYZ3(2,0,1) 
      HYZ3(2,1,2) = 
     $  + HYZ1(2) *HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ3(2,2,1)
      HYZ3(2,1,3) = 
     $  + HYZ1(3) *HYZ2(2,1)
     $  - HYZ3(2,3,1) 
     $  - HYZ3(3,2,1) 
      HYZ3(2,2,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(2)*HYZ1(2)
     $  - HYZ1(2) *HYZ2(0,2)
     $  + HYZ3(0,2,2) 
      HYZ3(2,2,2) = 
     $  + 1.666666666666666d-01*HYZ1(2)*HYZ1(2)*HYZ1(2)
      HYZ3(2,2,3) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ1(3)
     $  - HYZ1(2) *HYZ2(3,2)
     $  + HYZ3(3,2,2) 
      HYZ3(2,3,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ1(3)
     $  - HYZ1(0) *HYZ2(3,2)
     $  - HYZ1(2) *HYZ2(0,3)
     $  + HYZ3(0,3,2) 
      HYZ3(2,3,2) = 
     $  + HYZ1(2) *HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ3(3,2,2)
      HYZ3(2,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(3) *HYZ2(3,2)
     $  + HYZ3(3,3,2) 
      HYZ3(3,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(3)
     $  - HYZ1(0) *HYZ2(0,3)
     $  + HYZ3(0,0,3) 
      HYZ3(3,0,3) = 
     $  + HYZ1(3) *HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ3(0,3,3)
      HYZ3(3,1,0) = 
     $  + HYZ1(0) *HYZ2(3,1)
     $  - HYZ3(0,3,1) 
     $  - HYZ3(3,0,1) 
      HYZ3(3,1,2) = 
     $  + HYZ1(2) *HYZ2(3,1)
     $  - HYZ3(2,3,1) 
     $  - HYZ3(3,2,1) 
      HYZ3(3,1,3) = 
     $  + HYZ1(3) *HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ3(3,3,1)
      HYZ3(3,2,0) = 
     $  + HYZ1(0) *HYZ2(3,2)
     $  - HYZ3(0,3,2) 
     $  - HYZ3(3,0,2) 
      HYZ3(3,2,3) = 
     $  + HYZ1(3) *HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ3(3,3,2)
      HYZ3(3,3,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(3) *HYZ2(0,3)
     $  + HYZ3(0,3,3) 
      HYZ3(3,3,3) = 
     $  + 1.666666666666666d-01*HYZ1(3)*HYZ1(3)*HYZ1(3)
      endif
      if (n.eq.3) return
      
      if (n.eq.4) then
      HYZ4(0,0,0,0) = 
     $  + 4.166666666666666d-02*HYZ1(0)*HYZ1(0)*HYZ1(0)*HYZ1(0)
      HYZ4(0,0,1,0) = 
     $  + HYZ1(0) *HYZ3(0,0,1)
     $  - 3.000000000000000d+00*HYZ4(0,0,0,1)
      HYZ4(0,0,1,2) = 
     $  + HYZ1(2) *HYZ3(0,0,1)
     $  - HYZ4(0,0,2,1) 
     $  - HYZ4(0,2,0,1) 
     $  - HYZ4(2,0,0,1) 
      HYZ4(0,0,1,3) = 
     $  + HYZ1(3) *HYZ3(0,0,1)
     $  - HYZ4(0,0,3,1) 
     $  - HYZ4(0,3,0,1) 
     $  - HYZ4(3,0,0,1) 
      HYZ4(0,0,2,0) = 
     $  + HYZ1(0) *HYZ3(0,0,2)
     $  - 3.000000000000000d+00*HYZ4(0,0,0,2)
      HYZ4(0,0,2,3) = 
     $  + HYZ1(3) *HYZ3(0,0,2)
     $  - HYZ4(0,0,3,2) 
     $  - HYZ4(0,3,0,2) 
     $  - HYZ4(3,0,0,2) 
      HYZ4(0,0,3,0) = 
     $  + HYZ1(0) *HYZ3(0,0,3)
     $  - 3.000000000000000d+00*HYZ4(0,0,0,3)
      HYZ4(0,1,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,0,1)
     $  + 3.000000000000000d+00*HYZ4(0,0,0,1)
      HYZ4(0,1,0,1) = 
     $  + 5.000000000000000d-01*HYZ2(0,1)*HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ4(0,0,1,1)
      HYZ4(0,1,0,2) = 
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,0,1)
     $  + HYZ2(0,1) *HYZ2(0,2)
     $  + HYZ4(0,2,0,1) 
     $  + 2.000000000000000d+00*HYZ4(2,0,0,1)
      HYZ4(0,1,0,3) = 
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,0,1)
     $  + HYZ2(0,1) *HYZ2(0,3)
     $  + HYZ4(0,3,0,1) 
     $  + 2.000000000000000d+00*HYZ4(3,0,0,1)
      HYZ4(0,1,1,0) = 
     $  + HYZ1(0) *HYZ3(0,1,1)
     $  - 5.000000000000000d-01*HYZ2(0,1)*HYZ2(0,1)
      HYZ4(0,1,1,2) = 
     $  + HYZ1(2) *HYZ3(0,1,1)
     $  - HYZ4(0,1,2,1) 
     $  - HYZ4(0,2,1,1) 
     $  - HYZ4(2,0,1,1) 
      HYZ4(0,1,1,3) = 
     $  + HYZ1(3) *HYZ3(0,1,1)
     $  - HYZ4(0,1,3,1) 
     $  - HYZ4(0,3,1,1) 
     $  - HYZ4(3,0,1,1) 
      HYZ4(0,1,2,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(0,1)
     $  - HYZ1(0) *HYZ3(0,2,1)
     $  - HYZ1(0) *HYZ3(2,0,1)
     $  - HYZ2(0,1) *HYZ2(0,2)
     $  + 2.000000000000000d+00*HYZ4(0,0,2,1)
     $  + HYZ4(0,2,0,1) 
      HYZ4(0,1,2,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,1)
     $  - HYZ1(2) *HYZ3(0,2,1)
     $  - HYZ1(2) *HYZ3(2,0,1)
     $  + HYZ4(0,2,2,1) 
     $  + HYZ4(2,0,2,1) 
     $  + HYZ4(2,2,0,1) 
      HYZ4(0,1,2,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(0,1)
     $  - HYZ1(3) *HYZ3(0,2,1)
     $  - HYZ1(3) *HYZ3(2,0,1)
     $  - HYZ2(0,1) *HYZ2(3,2)
     $  + HYZ4(0,3,2,1) 
     $  + HYZ4(3,0,2,1) 
     $  + HYZ4(3,2,0,1) 
      HYZ4(0,1,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(0,1)
     $  - HYZ1(0) *HYZ3(0,3,1)
     $  - HYZ1(0) *HYZ3(3,0,1)
     $  - HYZ2(0,1) *HYZ2(0,3)
     $  + 2.000000000000000d+00*HYZ4(0,0,3,1)
     $  + HYZ4(0,3,0,1) 
      HYZ4(0,1,3,2) = 
     $  - HYZ1(2) *HYZ3(0,3,1)
     $  - HYZ1(2) *HYZ3(3,0,1)
     $  + HYZ2(0,1) *HYZ2(3,2)
     $  + HYZ4(0,2,3,1) 
     $  + HYZ4(2,0,3,1) 
     $  + HYZ4(2,3,0,1) 
      HYZ4(0,1,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,1)
     $  - HYZ1(3) *HYZ3(0,3,1)
     $  - HYZ1(3) *HYZ3(3,0,1)
     $  + HYZ4(0,3,3,1) 
     $  + HYZ4(3,0,3,1) 
     $  + HYZ4(3,3,0,1) 
      HYZ4(0,2,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,0,2)
     $  + 3.000000000000000d+00*HYZ4(0,0,0,2)
      HYZ4(0,2,0,2) = 
     $  + 5.000000000000000d-01*HYZ2(0,2)*HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ4(0,0,2,2)
      HYZ4(0,2,0,3) = 
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,0,2)
     $  + HYZ2(0,2) *HYZ2(0,3)
     $  + HYZ4(0,3,0,2) 
     $  + 2.000000000000000d+00*HYZ4(3,0,0,2)
      HYZ4(0,2,1,0) = 
     $  + HYZ1(0) *HYZ3(0,2,1)
     $  - 2.000000000000000d+00*HYZ4(0,0,2,1)
     $  - HYZ4(0,2,0,1) 
      HYZ4(0,2,1,2) = 
     $  + HYZ1(2) *HYZ3(0,2,1)
     $  - 2.000000000000000d+00*HYZ4(0,2,2,1)
     $  - HYZ4(2,0,2,1) 
      HYZ4(0,2,1,3) = 
     $  + HYZ1(3) *HYZ3(0,2,1)
     $  - HYZ4(0,2,3,1) 
     $  - HYZ4(0,3,2,1) 
     $  - HYZ4(3,0,2,1) 
      HYZ4(0,2,2,0) = 
     $  + HYZ1(0) *HYZ3(0,2,2)
     $  - 5.000000000000000d-01*HYZ2(0,2)*HYZ2(0,2)
      HYZ4(0,2,2,3) = 
     $  + HYZ1(3) *HYZ3(0,2,2)
     $  - HYZ4(0,2,3,2) 
     $  - HYZ4(0,3,2,2) 
     $  - HYZ4(3,0,2,2) 
      HYZ4(0,2,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(0,2)
     $  - HYZ1(0) *HYZ3(0,3,2)
     $  - HYZ1(0) *HYZ3(3,0,2)
     $  - HYZ2(0,2) *HYZ2(0,3)
     $  + 2.000000000000000d+00*HYZ4(0,0,3,2)
     $  + HYZ4(0,3,0,2) 
      HYZ4(0,2,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,2)
     $  - HYZ1(3) *HYZ3(0,3,2)
     $  - HYZ1(3) *HYZ3(3,0,2)
     $  + HYZ4(0,3,3,2) 
     $  + HYZ4(3,0,3,2) 
     $  + HYZ4(3,3,0,2) 
      HYZ4(0,3,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,0,3)
     $  + 3.000000000000000d+00*HYZ4(0,0,0,3)
      HYZ4(0,3,0,3) = 
     $  + 5.000000000000000d-01*HYZ2(0,3)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ4(0,0,3,3)
      HYZ4(0,3,1,0) = 
     $  + HYZ1(0) *HYZ3(0,3,1)
     $  - 2.000000000000000d+00*HYZ4(0,0,3,1)
     $  - HYZ4(0,3,0,1) 
      HYZ4(0,3,1,2) = 
     $  + HYZ1(2) *HYZ3(0,3,1)
     $  - HYZ4(0,2,3,1) 
     $  - HYZ4(0,3,2,1) 
     $  - HYZ4(2,0,3,1) 
      HYZ4(0,3,1,3) = 
     $  + HYZ1(3) *HYZ3(0,3,1)
     $  - 2.000000000000000d+00*HYZ4(0,3,3,1)
     $  - HYZ4(3,0,3,1) 
      HYZ4(0,3,2,0) = 
     $  + HYZ1(0) *HYZ3(0,3,2)
     $  - 2.000000000000000d+00*HYZ4(0,0,3,2)
     $  - HYZ4(0,3,0,2) 
      HYZ4(0,3,2,3) = 
     $  + HYZ1(3) *HYZ3(0,3,2)
     $  - 2.000000000000000d+00*HYZ4(0,3,3,2)
     $  - HYZ4(3,0,3,2) 
      HYZ4(0,3,3,0) = 
     $  + HYZ1(0) *HYZ3(0,3,3)
     $  - 5.000000000000000d-01*HYZ2(0,3)*HYZ2(0,3)
      HYZ4(1,0,0,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(0)*HYZ1(0)*HYZ1(1)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,1)
     $  + HYZ1(0) *HYZ3(0,0,1)
     $  - HYZ4(0,0,0,1) 
      HYZ4(1,0,0,1) = 
     $  + HYZ1(1) *HYZ3(0,0,1)
     $  - 5.000000000000000d-01*HYZ2(0,1)*HYZ2(0,1)
      HYZ4(1,0,0,2) = 
     $  + HYZ1(1) *HYZ3(0,0,2)
     $  + HYZ1(2) *HYZ3(0,0,1)
     $  - HYZ2(0,1) *HYZ2(0,2)
     $  - HYZ4(2,0,0,1) 
      HYZ4(1,0,0,3) = 
     $  + HYZ1(1) *HYZ3(0,0,3)
     $  + HYZ1(3) *HYZ3(0,0,1)
     $  - HYZ2(0,1) *HYZ2(0,3)
     $  - HYZ4(3,0,0,1) 
      HYZ4(1,0,1,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,1,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,0,1)
     $  + 5.000000000000000d-01*HYZ2(0,1)*HYZ2(0,1)
     $  + 2.000000000000000d+00*HYZ4(0,0,1,1)
      HYZ4(1,0,1,1) = 
     $  + HYZ1(1) *HYZ3(0,1,1)
     $  - 3.000000000000000d+00*HYZ4(0,1,1,1)
      HYZ4(1,0,1,2) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(0,1)
     $  - HYZ1(1) *HYZ3(0,2,1)
     $  - HYZ1(1) *HYZ3(2,0,1)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,1,1)
     $  + HYZ4(0,1,2,1) 
     $  + 2.000000000000000d+00*HYZ4(0,2,1,1)
     $  + 2.000000000000000d+00*HYZ4(2,0,1,1)
      HYZ4(1,0,1,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(0,1)
     $  - HYZ1(1) *HYZ3(0,3,1)
     $  - HYZ1(1) *HYZ3(3,0,1)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,1,1)
     $  + HYZ4(0,1,3,1) 
     $  + 2.000000000000000d+00*HYZ4(0,3,1,1)
     $  + 2.000000000000000d+00*HYZ4(3,0,1,1)
      HYZ4(1,0,2,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(0,2)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(0,1)
     $  + HYZ1(0) *HYZ3(2,0,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,0,2)
     $  + HYZ2(0,1) *HYZ2(0,2)
     $  - HYZ4(0,2,0,1) 
      HYZ4(1,0,2,1) = 
     $  + HYZ1(1) *HYZ3(0,2,1)
     $  - HYZ4(0,1,2,1) 
     $  - 2.000000000000000d+00*HYZ4(0,2,1,1)
      HYZ4(1,0,2,2) = 
     $  + HYZ1(1) *HYZ3(0,2,2)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,1)
     $  + HYZ1(2) *HYZ3(2,0,1)
     $  - HYZ4(2,2,0,1) 
      HYZ4(1,0,2,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(0,2)
     $  - HYZ1(1) *HYZ3(0,3,2)
     $  - HYZ1(1) *HYZ3(3,0,2)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(0,1)
     $  + HYZ1(3) *HYZ3(2,0,1)
     $  + HYZ2(0,1) *HYZ2(3,2)
     $  - HYZ4(3,2,0,1) 
      HYZ4(1,0,3,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(0,3)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(0,1)
     $  + HYZ1(0) *HYZ3(3,0,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,0,3)
     $  + HYZ2(0,1) *HYZ2(0,3)
     $  - HYZ4(0,3,0,1) 
      HYZ4(1,0,3,1) = 
     $  + HYZ1(1) *HYZ3(0,3,1)
     $  - HYZ4(0,1,3,1) 
     $  - 2.000000000000000d+00*HYZ4(0,3,1,1)
      HYZ4(1,0,3,2) = 
     $  + HYZ1(1) *HYZ3(0,3,2)
     $  + HYZ1(2) *HYZ3(3,0,1)
     $  - HYZ2(0,1) *HYZ2(3,2)
     $  - HYZ4(2,3,0,1) 
      HYZ4(1,0,3,3) = 
     $  + HYZ1(1) *HYZ3(0,3,3)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,1)
     $  + HYZ1(3) *HYZ3(3,0,1)
     $  - HYZ4(3,3,0,1) 
      HYZ4(1,1,0,0) = 
     $  + 2.500000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(1)*HYZ1(1)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(0,1)
     $  + HYZ1(0) *HYZ3(0,1,1)
     $  + HYZ1(1) *HYZ3(0,0,1)
     $  - HYZ4(0,0,1,1) 
      HYZ4(1,1,0,1) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,1,1)
     $  + 3.000000000000000d+00*HYZ4(0,1,1,1)
      HYZ4(1,1,0,2) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,2)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(0,1)
     $  + HYZ1(1) *HYZ3(2,0,1)
     $  + HYZ1(2) *HYZ3(0,1,1)
     $  - HYZ4(2,0,1,1) 
      HYZ4(1,1,0,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,3)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(0,1)
     $  + HYZ1(1) *HYZ3(3,0,1)
     $  + HYZ1(3) *HYZ3(0,1,1)
     $  - HYZ4(3,0,1,1) 
      HYZ4(1,1,1,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(1)*HYZ1(1)*HYZ1(1)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,1)
     $  + HYZ1(1) *HYZ3(0,1,1)
     $  - HYZ4(0,1,1,1) 
      HYZ4(1,1,1,1) = 
     $  + 4.166666666666666d-02*HYZ1(1)*HYZ1(1)*HYZ1(1)*HYZ1(1)
      HYZ4(1,1,1,2) = 
     $  + 1.666666666666666d-01*HYZ1(1)*HYZ1(1)*HYZ1(1)*HYZ1(2)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(2,1)
     $  + HYZ1(1) *HYZ3(2,1,1)
     $  - HYZ4(2,1,1,1) 
      HYZ4(1,1,1,3) = 
     $  + 1.666666666666666d-01*HYZ1(1)*HYZ1(1)*HYZ1(1)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(3,1)
     $  + HYZ1(1) *HYZ3(3,1,1)
     $  - HYZ4(3,1,1,1) 
      HYZ4(1,1,2,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(1)*HYZ1(1)*HYZ1(2)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(2,1)
     $  + HYZ1(0) *HYZ3(2,1,1)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,2)
     $  + HYZ1(1) *HYZ3(0,2,1)
     $  - HYZ4(0,2,1,1) 
      HYZ4(1,1,2,1) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(2,1,1)
     $  + 3.000000000000000d+00*HYZ4(2,1,1,1)
      HYZ4(1,1,2,2) = 
     $  + 2.500000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ1(2)*HYZ1(2)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(2,1)
     $  + HYZ1(1) *HYZ3(2,2,1)
     $  + HYZ1(2) *HYZ3(2,1,1)
     $  - HYZ4(2,2,1,1) 
      HYZ4(1,1,2,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ1(2)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(3,2)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(2,1)
     $  + HYZ1(1) *HYZ3(3,2,1)
     $  + HYZ1(3) *HYZ3(2,1,1)
     $  - HYZ4(3,2,1,1) 
      HYZ4(1,1,3,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(1)*HYZ1(1)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(3,1)
     $  + HYZ1(0) *HYZ3(3,1,1)
     $  - 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(0,3)
     $  + HYZ1(1) *HYZ3(0,3,1)
     $  - HYZ4(0,3,1,1) 
      HYZ4(1,1,3,1) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(3,1,1)
     $  + 3.000000000000000d+00*HYZ4(3,1,1,1)
      HYZ4(1,1,3,2) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ2(3,2)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(3,1)
     $  + HYZ1(1) *HYZ3(2,3,1)
     $  + HYZ1(2) *HYZ3(3,1,1)
     $  - HYZ4(2,3,1,1) 
      HYZ4(1,1,3,3) = 
     $  + 2.500000000000000d-01*HYZ1(1)*HYZ1(1)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(3,1)
     $  + HYZ1(1) *HYZ3(3,3,1)
     $  + HYZ1(3) *HYZ3(3,1,1)
     $  - HYZ4(3,3,1,1) 
      HYZ4(1,2,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(1)*HYZ1(2)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(2,1)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(0,2)
     $  + HYZ1(0) *HYZ3(0,2,1)
     $  + HYZ1(1) *HYZ3(0,0,2)
     $  - HYZ4(0,0,2,1) 
      HYZ4(1,2,0,1) = 
     $  + HYZ1(1) *HYZ3(2,0,1)
     $  - HYZ2(0,1) *HYZ2(2,1)
     $  + HYZ4(0,1,2,1) 
     $  + 2.000000000000000d+00*HYZ4(0,2,1,1)
      HYZ4(1,2,0,2) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,2,2)
     $  + HYZ1(2) *HYZ3(0,2,1)
     $  - HYZ2(0,2) *HYZ2(2,1)
     $  - HYZ4(2,0,2,1) 
      HYZ4(1,2,0,3) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(0,3)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(0,2)
     $  + HYZ1(1) *HYZ3(3,0,2)
     $  + HYZ1(3) *HYZ3(0,2,1)
     $  - HYZ2(0,3) *HYZ2(2,1)
     $  - HYZ4(3,0,2,1) 
      HYZ4(1,2,1,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(2,1,1)
     $  - HYZ1(1) *HYZ3(0,2,1)
     $  - HYZ1(1) *HYZ3(2,0,1)
     $  + HYZ2(0,1) *HYZ2(2,1)
     $  - HYZ4(0,1,2,1) 
      HYZ4(1,2,1,1) = 
     $  + HYZ1(1) *HYZ3(2,1,1)
     $  - 3.000000000000000d+00*HYZ4(2,1,1,1)
      HYZ4(1,2,1,2) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(2,2,1)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(2,1,1)
     $  + 5.000000000000000d-01*HYZ2(2,1)*HYZ2(2,1)
     $  + 2.000000000000000d+00*HYZ4(2,2,1,1)
      HYZ4(1,2,1,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(2,1)
     $  - HYZ1(1) *HYZ3(2,3,1)
     $  - HYZ1(1) *HYZ3(3,2,1)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(2,1,1)
     $  + HYZ2(2,1) *HYZ2(3,1)
     $  - HYZ4(3,1,2,1) 
      HYZ4(1,2,2,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(1)*HYZ1(2)*HYZ1(2)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(2,1)
     $  + HYZ1(0) *HYZ3(2,2,1)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(0,2)
     $  + HYZ1(1) *HYZ3(0,2,2)
     $  + HYZ2(0,2) *HYZ2(2,1)
     $  - HYZ4(0,2,2,1) 
      HYZ4(1,2,2,1) = 
     $  + HYZ1(1) *HYZ3(2,2,1)
     $  - 5.000000000000000d-01*HYZ2(2,1)*HYZ2(2,1)
      HYZ4(1,2,2,2) = 
     $  + 1.666666666666666d-01*HYZ1(1)*HYZ1(2)*HYZ1(2)*HYZ1(2)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(2,1)
     $  + HYZ1(2) *HYZ3(2,2,1)
     $  - HYZ4(2,2,2,1) 
      HYZ4(1,2,2,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(2)*HYZ1(2)*HYZ1(3)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(3,2)
     $  + HYZ1(1) *HYZ3(3,2,2)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(2,1)
     $  + HYZ1(3) *HYZ3(2,2,1)
     $  + HYZ2(2,1) *HYZ2(3,2)
     $  - HYZ4(3,2,2,1) 
      HYZ4(1,2,3,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ1(2)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(3,2)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(2,1)
     $  + HYZ1(0) *HYZ3(3,2,1)
     $  - HYZ1(1) *HYZ1(2)*HYZ2(0,3)
     $  + HYZ1(1) *HYZ3(0,3,2)
     $  + HYZ2(0,3) *HYZ2(2,1)
     $  - HYZ4(0,3,2,1) 
      HYZ4(1,2,3,1) = 
     $  + HYZ1(1) *HYZ3(2,3,1)
     $  - HYZ2(2,1) *HYZ2(3,1)
     $  + HYZ4(3,1,2,1) 
     $  + 2.000000000000000d+00*HYZ4(3,2,1,1)
      HYZ4(1,2,3,2) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(3,2,2)
     $  + HYZ1(2) *HYZ3(3,2,1)
     $  - HYZ2(2,1) *HYZ2(3,2)
     $  - HYZ4(2,3,2,1) 
      HYZ4(1,2,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(1)*HYZ1(2)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(3,2)
     $  + HYZ1(1) *HYZ3(3,3,2)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(2,1)
     $  + HYZ1(3) *HYZ3(3,2,1)
     $  - HYZ4(3,3,2,1) 
      HYZ4(1,3,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(1)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(3,1)
     $  - HYZ1(0) *HYZ1(1)*HYZ2(0,3)
     $  + HYZ1(0) *HYZ3(0,3,1)
     $  + HYZ1(1) *HYZ3(0,0,3)
     $  - HYZ4(0,0,3,1) 
      HYZ4(1,3,0,1) = 
     $  + HYZ1(1) *HYZ3(3,0,1)
     $  - HYZ2(0,1) *HYZ2(3,1)
     $  + HYZ4(0,1,3,1) 
     $  + 2.000000000000000d+00*HYZ4(0,3,1,1)
      HYZ4(1,3,0,2) = 
     $  + HYZ1(1) *HYZ3(3,0,2)
     $  + HYZ1(2) *HYZ3(0,3,1)
     $  - HYZ2(0,2) *HYZ2(3,1)
     $  - HYZ4(2,0,3,1) 
      HYZ4(1,3,0,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(0,3,3)
     $  + HYZ1(3) *HYZ3(0,3,1)
     $  - HYZ2(0,3) *HYZ2(3,1)
     $  - HYZ4(3,0,3,1) 
      HYZ4(1,3,1,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(3,1,1)
     $  - HYZ1(1) *HYZ3(0,3,1)
     $  - HYZ1(1) *HYZ3(3,0,1)
     $  + HYZ2(0,1) *HYZ2(3,1)
     $  - HYZ4(0,1,3,1) 
      HYZ4(1,3,1,1) = 
     $  + HYZ1(1) *HYZ3(3,1,1)
     $  - 3.000000000000000d+00*HYZ4(3,1,1,1)
      HYZ4(1,3,1,2) = 
     $  + HYZ1(1) *HYZ1(2)*HYZ2(3,1)
     $  - HYZ1(1) *HYZ3(2,3,1)
     $  - HYZ1(1) *HYZ3(3,2,1)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(3,1,1)
     $  + 2.000000000000000d+00*HYZ4(2,3,1,1)
     $  + HYZ4(3,1,2,1) 
     $  + 2.000000000000000d+00*HYZ4(3,2,1,1)
      HYZ4(1,3,1,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(3,3,1)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(3,1,1)
     $  + 5.000000000000000d-01*HYZ2(3,1)*HYZ2(3,1)
     $  + 2.000000000000000d+00*HYZ4(3,3,1,1)
      HYZ4(1,3,2,0) = 
     $  + HYZ1(0) *HYZ1(1)*HYZ2(3,2)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(3,1)
     $  + HYZ1(0) *HYZ3(2,3,1)
     $  - HYZ1(1) *HYZ3(0,3,2)
     $  - HYZ1(1) *HYZ3(3,0,2)
     $  + HYZ2(0,2) *HYZ2(3,1)
     $  - HYZ4(0,2,3,1) 
      HYZ4(1,3,2,1) = 
     $  + HYZ1(1) *HYZ3(3,2,1)
     $  - HYZ4(3,1,2,1) 
     $  - 2.000000000000000d+00*HYZ4(3,2,1,1)
      HYZ4(1,3,2,2) = 
     $  + HYZ1(1) *HYZ3(3,2,2)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(3,1)
     $  + HYZ1(2) *HYZ3(2,3,1)
     $  - HYZ4(2,2,3,1) 
      HYZ4(1,3,2,3) = 
     $  + HYZ1(1) *HYZ1(3)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(1)*HYZ3(3,3,2)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(3,1)
     $  + HYZ1(3) *HYZ3(2,3,1)
     $  + HYZ2(3,1) *HYZ2(3,2)
     $  - HYZ4(3,2,3,1) 
      HYZ4(1,3,3,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(1)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(3,1)
     $  + HYZ1(0) *HYZ3(3,3,1)
     $  - HYZ1(1) *HYZ1(3)*HYZ2(0,3)
     $  + HYZ1(1) *HYZ3(0,3,3)
     $  + HYZ2(0,3) *HYZ2(3,1)
     $  - HYZ4(0,3,3,1) 
      HYZ4(1,3,3,1) = 
     $  + HYZ1(1) *HYZ3(3,3,1)
     $  - 5.000000000000000d-01*HYZ2(3,1)*HYZ2(3,1)
      HYZ4(1,3,3,2) = 
     $  + HYZ1(1) *HYZ3(3,3,2)
     $  + HYZ1(2) *HYZ3(3,3,1)
     $  - HYZ2(3,1) *HYZ2(3,2)
     $  - HYZ4(2,3,3,1) 
      HYZ4(1,3,3,3) = 
     $  + 1.666666666666666d-01*HYZ1(1)*HYZ1(3)*HYZ1(3)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(3,1)
     $  + HYZ1(3) *HYZ3(3,3,1)
     $  - HYZ4(3,3,3,1) 
      HYZ4(2,0,0,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(0)*HYZ1(0)*HYZ1(2)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,2)
     $  + HYZ1(0) *HYZ3(0,0,2)
     $  - HYZ4(0,0,0,2) 
      HYZ4(2,0,0,2) = 
     $  + HYZ1(2) *HYZ3(0,0,2)
     $  - 5.000000000000000d-01*HYZ2(0,2)*HYZ2(0,2)
      HYZ4(2,0,0,3) = 
     $  + HYZ1(2) *HYZ3(0,0,3)
     $  + HYZ1(3) *HYZ3(0,0,2)
     $  - HYZ2(0,2) *HYZ2(0,3)
     $  - HYZ4(3,0,0,2) 
      HYZ4(2,0,1,0) = 
     $  + HYZ1(0) *HYZ3(2,0,1)
     $  - HYZ4(0,2,0,1) 
     $  - 2.000000000000000d+00*HYZ4(2,0,0,1)
      HYZ4(2,0,1,2) = 
     $  + HYZ1(2) *HYZ3(2,0,1)
     $  - HYZ4(2,0,2,1) 
     $  - 2.000000000000000d+00*HYZ4(2,2,0,1)
      HYZ4(2,0,1,3) = 
     $  + HYZ1(3) *HYZ3(2,0,1)
     $  - HYZ4(2,0,3,1) 
     $  - HYZ4(2,3,0,1) 
     $  - HYZ4(3,2,0,1) 
      HYZ4(2,0,2,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,2,2)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,0,2)
     $  + 5.000000000000000d-01*HYZ2(0,2)*HYZ2(0,2)
     $  + 2.000000000000000d+00*HYZ4(0,0,2,2)
      HYZ4(2,0,2,2) = 
     $  + HYZ1(2) *HYZ3(0,2,2)
     $  - 3.000000000000000d+00*HYZ4(0,2,2,2)
      HYZ4(2,0,2,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(0,2)
     $  - HYZ1(2) *HYZ3(0,3,2)
     $  - HYZ1(2) *HYZ3(3,0,2)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,2,2)
     $  + HYZ4(0,2,3,2) 
     $  + 2.000000000000000d+00*HYZ4(0,3,2,2)
     $  + 2.000000000000000d+00*HYZ4(3,0,2,2)
      HYZ4(2,0,3,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(0,3)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(0,2)
     $  + HYZ1(0) *HYZ3(3,0,2)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,0,3)
     $  + HYZ2(0,2) *HYZ2(0,3)
     $  - HYZ4(0,3,0,2) 
      HYZ4(2,0,3,2) = 
     $  + HYZ1(2) *HYZ3(0,3,2)
     $  - HYZ4(0,2,3,2) 
     $  - 2.000000000000000d+00*HYZ4(0,3,2,2)
      HYZ4(2,0,3,3) = 
     $  + HYZ1(2) *HYZ3(0,3,3)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,2)
     $  + HYZ1(3) *HYZ3(3,0,2)
     $  - HYZ4(3,3,0,2) 
      HYZ4(2,1,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(2,1)
     $  - HYZ1(0) *HYZ3(0,2,1)
     $  - HYZ1(0) *HYZ3(2,0,1)
     $  + HYZ4(0,0,2,1) 
     $  + HYZ4(0,2,0,1) 
     $  + HYZ4(2,0,0,1) 
      HYZ4(2,1,0,1) = 
     $  + HYZ2(0,1) *HYZ2(2,1)
     $  - HYZ4(0,1,2,1) 
     $  - 2.000000000000000d+00*HYZ4(0,2,1,1)
     $  - 2.000000000000000d+00*HYZ4(2,0,1,1)
      HYZ4(2,1,0,2) = 
     $  - HYZ1(2) *HYZ3(0,2,1)
     $  - HYZ1(2) *HYZ3(2,0,1)
     $  + HYZ2(0,2) *HYZ2(2,1)
     $  + HYZ4(2,0,2,1) 
     $  + 2.000000000000000d+00*HYZ4(2,2,0,1)
      HYZ4(2,1,0,3) = 
     $  - HYZ1(3) *HYZ3(0,2,1)
     $  - HYZ1(3) *HYZ3(2,0,1)
     $  + HYZ2(0,3) *HYZ2(2,1)
     $  + HYZ4(2,3,0,1) 
     $  + HYZ4(3,0,2,1) 
     $  + HYZ4(3,2,0,1) 
      HYZ4(2,1,1,0) = 
     $  + HYZ1(0) *HYZ3(2,1,1)
     $  - HYZ2(0,1) *HYZ2(2,1)
     $  + HYZ4(0,1,2,1) 
     $  + HYZ4(0,2,1,1) 
     $  + HYZ4(2,0,1,1) 
      HYZ4(2,1,1,2) = 
     $  + HYZ1(2) *HYZ3(2,1,1)
     $  - 5.000000000000000d-01*HYZ2(2,1)*HYZ2(2,1)
      HYZ4(2,1,1,3) = 
     $  + HYZ1(3) *HYZ3(2,1,1)
     $  - HYZ2(2,1) *HYZ2(3,1)
     $  + HYZ4(2,3,1,1) 
     $  + HYZ4(3,1,2,1) 
     $  + HYZ4(3,2,1,1) 
      HYZ4(2,1,2,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(2,2,1)
     $  - HYZ2(0,2) *HYZ2(2,1)
     $  + 2.000000000000000d+00*HYZ4(0,2,2,1)
     $  + HYZ4(2,0,2,1) 
      HYZ4(2,1,2,1) = 
     $  + 5.000000000000000d-01*HYZ2(2,1)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ4(2,2,1,1)
      HYZ4(2,1,2,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(2,2,1)
     $  + 3.000000000000000d+00*HYZ4(2,2,2,1)
      HYZ4(2,1,2,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(2,1)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(2,2,1)
     $  - HYZ2(2,1) *HYZ2(3,2)
     $  + HYZ4(2,3,2,1) 
     $  + 2.000000000000000d+00*HYZ4(3,2,2,1)
      HYZ4(2,1,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(2,1)
     $  - HYZ1(0) *HYZ3(2,3,1)
     $  - HYZ1(0) *HYZ3(3,2,1)
     $  - HYZ2(0,3) *HYZ2(2,1)
     $  + HYZ4(0,2,3,1) 
     $  + HYZ4(0,3,2,1) 
     $  + HYZ4(2,0,3,1) 
      HYZ4(2,1,3,1) = 
     $  + HYZ2(2,1) *HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ4(2,3,1,1)
     $  - HYZ4(3,1,2,1) 
     $  - 2.000000000000000d+00*HYZ4(3,2,1,1)
      HYZ4(2,1,3,2) = 
     $  - HYZ1(2) *HYZ3(2,3,1)
     $  - HYZ1(2) *HYZ3(3,2,1)
     $  + HYZ2(2,1) *HYZ2(3,2)
     $  + 2.000000000000000d+00*HYZ4(2,2,3,1)
     $  + HYZ4(2,3,2,1) 
      HYZ4(2,1,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(2,1)
     $  - HYZ1(3) *HYZ3(2,3,1)
     $  - HYZ1(3) *HYZ3(3,2,1)
     $  + HYZ4(2,3,3,1) 
     $  + HYZ4(3,2,3,1) 
     $  + HYZ4(3,3,2,1) 
      HYZ4(2,2,0,0) = 
     $  + 2.500000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(2)*HYZ1(2)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(0,2)
     $  + HYZ1(0) *HYZ3(0,2,2)
     $  + HYZ1(2) *HYZ3(0,0,2)
     $  - HYZ4(0,0,2,2) 
      HYZ4(2,2,0,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,2)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,2,2)
     $  + 3.000000000000000d+00*HYZ4(0,2,2,2)
      HYZ4(2,2,0,3) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,3)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(0,2)
     $  + HYZ1(2) *HYZ3(3,0,2)
     $  + HYZ1(3) *HYZ3(0,2,2)
     $  - HYZ4(3,0,2,2) 
      HYZ4(2,2,1,0) = 
     $  + HYZ1(0) *HYZ3(2,2,1)
     $  - HYZ4(0,2,2,1) 
     $  - HYZ4(2,0,2,1) 
     $  - HYZ4(2,2,0,1) 
      HYZ4(2,2,1,2) = 
     $  + HYZ1(2) *HYZ3(2,2,1)
     $  - 3.000000000000000d+00*HYZ4(2,2,2,1)
      HYZ4(2,2,1,3) = 
     $  + HYZ1(3) *HYZ3(2,2,1)
     $  - HYZ4(2,2,3,1) 
     $  - HYZ4(2,3,2,1) 
     $  - HYZ4(3,2,2,1) 
      HYZ4(2,2,2,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(2)*HYZ1(2)*HYZ1(2)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,2)
     $  + HYZ1(2) *HYZ3(0,2,2)
     $  - HYZ4(0,2,2,2) 
      HYZ4(2,2,2,2) = 
     $  + 4.166666666666666d-02*HYZ1(2)*HYZ1(2)*HYZ1(2)*HYZ1(2)
      HYZ4(2,2,2,3) = 
     $  + 1.666666666666666d-01*HYZ1(2)*HYZ1(2)*HYZ1(2)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(3,2)
     $  + HYZ1(2) *HYZ3(3,2,2)
     $  - HYZ4(3,2,2,2) 
      HYZ4(2,2,3,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(2)*HYZ1(2)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(3,2)
     $  + HYZ1(0) *HYZ3(3,2,2)
     $  - 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(0,3)
     $  + HYZ1(2) *HYZ3(0,3,2)
     $  - HYZ4(0,3,2,2) 
      HYZ4(2,2,3,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(3,2,2)
     $  + 3.000000000000000d+00*HYZ4(3,2,2,2)
      HYZ4(2,2,3,3) = 
     $  + 2.500000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(3,2)
     $  + HYZ1(2) *HYZ3(3,3,2)
     $  + HYZ1(3) *HYZ3(3,2,2)
     $  - HYZ4(3,3,2,2) 
      HYZ4(2,3,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(2)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(3,2)
     $  - HYZ1(0) *HYZ1(2)*HYZ2(0,3)
     $  + HYZ1(0) *HYZ3(0,3,2)
     $  + HYZ1(2) *HYZ3(0,0,3)
     $  - HYZ4(0,0,3,2) 
      HYZ4(2,3,0,2) = 
     $  + HYZ1(2) *HYZ3(3,0,2)
     $  - HYZ2(0,2) *HYZ2(3,2)
     $  + HYZ4(0,2,3,2) 
     $  + 2.000000000000000d+00*HYZ4(0,3,2,2)
      HYZ4(2,3,0,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(0,3,3)
     $  + HYZ1(3) *HYZ3(0,3,2)
     $  - HYZ2(0,3) *HYZ2(3,2)
     $  - HYZ4(3,0,3,2) 
      HYZ4(2,3,1,0) = 
     $  + HYZ1(0) *HYZ3(2,3,1)
     $  - HYZ4(0,2,3,1) 
     $  - HYZ4(2,0,3,1) 
     $  - HYZ4(2,3,0,1) 
      HYZ4(2,3,1,2) = 
     $  + HYZ1(2) *HYZ3(2,3,1)
     $  - 2.000000000000000d+00*HYZ4(2,2,3,1)
     $  - HYZ4(2,3,2,1) 
      HYZ4(2,3,1,3) = 
     $  + HYZ1(3) *HYZ3(2,3,1)
     $  - 2.000000000000000d+00*HYZ4(2,3,3,1)
     $  - HYZ4(3,2,3,1) 
      HYZ4(2,3,2,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(3,2,2)
     $  - HYZ1(2) *HYZ3(0,3,2)
     $  - HYZ1(2) *HYZ3(3,0,2)
     $  + HYZ2(0,2) *HYZ2(3,2)
     $  - HYZ4(0,2,3,2) 
      HYZ4(2,3,2,2) = 
     $  + HYZ1(2) *HYZ3(3,2,2)
     $  - 3.000000000000000d+00*HYZ4(3,2,2,2)
      HYZ4(2,3,2,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(3,3,2)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(3,2,2)
     $  + 5.000000000000000d-01*HYZ2(3,2)*HYZ2(3,2)
     $  + 2.000000000000000d+00*HYZ4(3,3,2,2)
      HYZ4(2,3,3,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(2)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(3,2)
     $  + HYZ1(0) *HYZ3(3,3,2)
     $  - HYZ1(2) *HYZ1(3)*HYZ2(0,3)
     $  + HYZ1(2) *HYZ3(0,3,3)
     $  + HYZ2(0,3) *HYZ2(3,2)
     $  - HYZ4(0,3,3,2) 
      HYZ4(2,3,3,2) = 
     $  + HYZ1(2) *HYZ3(3,3,2)
     $  - 5.000000000000000d-01*HYZ2(3,2)*HYZ2(3,2)
      HYZ4(2,3,3,3) = 
     $  + 1.666666666666666d-01*HYZ1(2)*HYZ1(3)*HYZ1(3)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(3,2)
     $  + HYZ1(3) *HYZ3(3,3,2)
     $  - HYZ4(3,3,3,2) 
      HYZ4(3,0,0,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(0)*HYZ1(0)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(0,3)
     $  + HYZ1(0) *HYZ3(0,0,3)
     $  - HYZ4(0,0,0,3) 
      HYZ4(3,0,0,3) = 
     $  + HYZ1(3) *HYZ3(0,0,3)
     $  - 5.000000000000000d-01*HYZ2(0,3)*HYZ2(0,3)
      HYZ4(3,0,1,0) = 
     $  + HYZ1(0) *HYZ3(3,0,1)
     $  - HYZ4(0,3,0,1) 
     $  - 2.000000000000000d+00*HYZ4(3,0,0,1)
      HYZ4(3,0,1,2) = 
     $  + HYZ1(2) *HYZ3(3,0,1)
     $  - HYZ4(2,3,0,1) 
     $  - HYZ4(3,0,2,1) 
     $  - HYZ4(3,2,0,1) 
      HYZ4(3,0,1,3) = 
     $  + HYZ1(3) *HYZ3(3,0,1)
     $  - HYZ4(3,0,3,1) 
     $  - 2.000000000000000d+00*HYZ4(3,3,0,1)
      HYZ4(3,0,2,0) = 
     $  + HYZ1(0) *HYZ3(3,0,2)
     $  - HYZ4(0,3,0,2) 
     $  - 2.000000000000000d+00*HYZ4(3,0,0,2)
      HYZ4(3,0,2,3) = 
     $  + HYZ1(3) *HYZ3(3,0,2)
     $  - HYZ4(3,0,3,2) 
     $  - 2.000000000000000d+00*HYZ4(3,3,0,2)
      HYZ4(3,0,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(0,3,3)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,0,3)
     $  + 5.000000000000000d-01*HYZ2(0,3)*HYZ2(0,3)
     $  + 2.000000000000000d+00*HYZ4(0,0,3,3)
      HYZ4(3,0,3,3) = 
     $  + HYZ1(3) *HYZ3(0,3,3)
     $  - 3.000000000000000d+00*HYZ4(0,3,3,3)
      HYZ4(3,1,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(3,1)
     $  - HYZ1(0) *HYZ3(0,3,1)
     $  - HYZ1(0) *HYZ3(3,0,1)
     $  + HYZ4(0,0,3,1) 
     $  + HYZ4(0,3,0,1) 
     $  + HYZ4(3,0,0,1) 
      HYZ4(3,1,0,1) = 
     $  + HYZ2(0,1) *HYZ2(3,1)
     $  - HYZ4(0,1,3,1) 
     $  - 2.000000000000000d+00*HYZ4(0,3,1,1)
     $  - 2.000000000000000d+00*HYZ4(3,0,1,1)
      HYZ4(3,1,0,2) = 
     $  - HYZ1(2) *HYZ3(0,3,1)
     $  - HYZ1(2) *HYZ3(3,0,1)
     $  + HYZ2(0,2) *HYZ2(3,1)
     $  + HYZ4(2,0,3,1) 
     $  + HYZ4(2,3,0,1) 
     $  + HYZ4(3,2,0,1) 
      HYZ4(3,1,0,3) = 
     $  - HYZ1(3) *HYZ3(0,3,1)
     $  - HYZ1(3) *HYZ3(3,0,1)
     $  + HYZ2(0,3) *HYZ2(3,1)
     $  + HYZ4(3,0,3,1) 
     $  + 2.000000000000000d+00*HYZ4(3,3,0,1)
      HYZ4(3,1,1,0) = 
     $  + HYZ1(0) *HYZ3(3,1,1)
     $  - HYZ2(0,1) *HYZ2(3,1)
     $  + HYZ4(0,1,3,1) 
     $  + HYZ4(0,3,1,1) 
     $  + HYZ4(3,0,1,1) 
      HYZ4(3,1,1,2) = 
     $  + HYZ1(2) *HYZ3(3,1,1)
     $  - HYZ4(2,3,1,1) 
     $  - HYZ4(3,1,2,1) 
     $  - HYZ4(3,2,1,1) 
      HYZ4(3,1,1,3) = 
     $  + HYZ1(3) *HYZ3(3,1,1)
     $  - 5.000000000000000d-01*HYZ2(3,1)*HYZ2(3,1)
      HYZ4(3,1,2,0) = 
     $  + HYZ1(0) *HYZ1(2)*HYZ2(3,1)
     $  - HYZ1(0) *HYZ3(2,3,1)
     $  - HYZ1(0) *HYZ3(3,2,1)
     $  - HYZ2(0,2) *HYZ2(3,1)
     $  + HYZ4(0,2,3,1) 
     $  + HYZ4(0,3,2,1) 
     $  + HYZ4(3,0,2,1) 
      HYZ4(3,1,2,2) = 
     $  + 5.000000000000000d-01*HYZ1(2)*HYZ1(2)*HYZ2(3,1)
     $  - HYZ1(2) *HYZ3(2,3,1)
     $  - HYZ1(2) *HYZ3(3,2,1)
     $  + HYZ4(2,2,3,1) 
     $  + HYZ4(2,3,2,1) 
     $  + HYZ4(3,2,2,1) 
      HYZ4(3,1,2,3) = 
     $  + HYZ1(2) *HYZ1(3)*HYZ2(3,1)
     $  - HYZ1(3) *HYZ3(2,3,1)
     $  - HYZ1(3) *HYZ3(3,2,1)
     $  - HYZ2(3,1) *HYZ2(3,2)
     $  + HYZ4(3,2,3,1) 
     $  + 2.000000000000000d+00*HYZ4(3,3,2,1)
      HYZ4(3,1,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(3,3,1)
     $  - HYZ2(0,3) *HYZ2(3,1)
     $  + 2.000000000000000d+00*HYZ4(0,3,3,1)
     $  + HYZ4(3,0,3,1) 
      HYZ4(3,1,3,1) = 
     $  + 5.000000000000000d-01*HYZ2(3,1)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ4(3,3,1,1)
      HYZ4(3,1,3,2) = 
     $  - 2.000000000000000d+00*HYZ1(2)*HYZ3(3,3,1)
     $  + HYZ2(3,1) *HYZ2(3,2)
     $  + 2.000000000000000d+00*HYZ4(2,3,3,1)
     $  + HYZ4(3,2,3,1) 
      HYZ4(3,1,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(3,1)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(3,3,1)
     $  + 3.000000000000000d+00*HYZ4(3,3,3,1)
      HYZ4(3,2,0,0) = 
     $  + 5.000000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ2(3,2)
     $  - HYZ1(0) *HYZ3(0,3,2)
     $  - HYZ1(0) *HYZ3(3,0,2)
     $  + HYZ4(0,0,3,2) 
     $  + HYZ4(0,3,0,2) 
     $  + HYZ4(3,0,0,2) 
      HYZ4(3,2,0,2) = 
     $  + HYZ2(0,2) *HYZ2(3,2)
     $  - HYZ4(0,2,3,2) 
     $  - 2.000000000000000d+00*HYZ4(0,3,2,2)
     $  - 2.000000000000000d+00*HYZ4(3,0,2,2)
      HYZ4(3,2,0,3) = 
     $  - HYZ1(3) *HYZ3(0,3,2)
     $  - HYZ1(3) *HYZ3(3,0,2)
     $  + HYZ2(0,3) *HYZ2(3,2)
     $  + HYZ4(3,0,3,2) 
     $  + 2.000000000000000d+00*HYZ4(3,3,0,2)
      HYZ4(3,2,1,0) = 
     $  + HYZ1(0) *HYZ3(3,2,1)
     $  - HYZ4(0,3,2,1) 
     $  - HYZ4(3,0,2,1) 
     $  - HYZ4(3,2,0,1) 
      HYZ4(3,2,1,2) = 
     $  + HYZ1(2) *HYZ3(3,2,1)
     $  - HYZ4(2,3,2,1) 
     $  - 2.000000000000000d+00*HYZ4(3,2,2,1)
      HYZ4(3,2,1,3) = 
     $  + HYZ1(3) *HYZ3(3,2,1)
     $  - HYZ4(3,2,3,1) 
     $  - 2.000000000000000d+00*HYZ4(3,3,2,1)
      HYZ4(3,2,2,0) = 
     $  + HYZ1(0) *HYZ3(3,2,2)
     $  - HYZ2(0,2) *HYZ2(3,2)
     $  + HYZ4(0,2,3,2) 
     $  + HYZ4(0,3,2,2) 
     $  + HYZ4(3,0,2,2) 
      HYZ4(3,2,2,3) = 
     $  + HYZ1(3) *HYZ3(3,2,2)
     $  - 5.000000000000000d-01*HYZ2(3,2)*HYZ2(3,2)
      HYZ4(3,2,3,0) = 
     $  + HYZ1(0) *HYZ1(3)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(0)*HYZ3(3,3,2)
     $  - HYZ2(0,3) *HYZ2(3,2)
     $  + 2.000000000000000d+00*HYZ4(0,3,3,2)
     $  + HYZ4(3,0,3,2) 
      HYZ4(3,2,3,2) = 
     $  + 5.000000000000000d-01*HYZ2(3,2)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ4(3,3,2,2)
      HYZ4(3,2,3,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(3,2)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(3,3,2)
     $  + 3.000000000000000d+00*HYZ4(3,3,3,2)
      HYZ4(3,3,0,0) = 
     $  + 2.500000000000000d-01*HYZ1(0)*HYZ1(0)*HYZ1(3)*HYZ1(3)
     $  - HYZ1(0) *HYZ1(3)*HYZ2(0,3)
     $  + HYZ1(0) *HYZ3(0,3,3)
     $  + HYZ1(3) *HYZ3(0,0,3)
     $  - HYZ4(0,0,3,3) 
      HYZ4(3,3,0,3) = 
     $  + 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,3)
     $  - 2.000000000000000d+00*HYZ1(3)*HYZ3(0,3,3)
     $  + 3.000000000000000d+00*HYZ4(0,3,3,3)
      HYZ4(3,3,1,0) = 
     $  + HYZ1(0) *HYZ3(3,3,1)
     $  - HYZ4(0,3,3,1) 
     $  - HYZ4(3,0,3,1) 
     $  - HYZ4(3,3,0,1) 
      HYZ4(3,3,1,2) = 
     $  + HYZ1(2) *HYZ3(3,3,1)
     $  - HYZ4(2,3,3,1) 
     $  - HYZ4(3,2,3,1) 
     $  - HYZ4(3,3,2,1) 
      HYZ4(3,3,1,3) = 
     $  + HYZ1(3) *HYZ3(3,3,1)
     $  - 3.000000000000000d+00*HYZ4(3,3,3,1)
      HYZ4(3,3,2,0) = 
     $  + HYZ1(0) *HYZ3(3,3,2)
     $  - HYZ4(0,3,3,2) 
     $  - HYZ4(3,0,3,2) 
     $  - HYZ4(3,3,0,2) 
      HYZ4(3,3,2,3) = 
     $  + HYZ1(3) *HYZ3(3,3,2)
     $  - 3.000000000000000d+00*HYZ4(3,3,3,2)
      HYZ4(3,3,3,0) = 
     $  + 1.666666666666666d-01*HYZ1(0)*HYZ1(3)*HYZ1(3)*HYZ1(3)
     $  - 5.000000000000000d-01*HYZ1(3)*HYZ1(3)*HYZ2(0,3)
     $  + HYZ1(3) *HYZ3(0,3,3)
     $  - HYZ4(0,3,3,3) 
      HYZ4(3,3,3,3) = 
     $  + 4.166666666666666d-02*HYZ1(3)*HYZ1(3)*HYZ1(3)*HYZ1(3)
      endif

      return
      end

      subroutine swap2dhplxy(iflag,nmax,HXZ1,HXZ2,HXZ3,HXZ4,
     $                       HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4)
*********************************************************************
*****  swap2dhpl evaluates H(..;y) in terms of H(..;x=1-y-z)   ******
*****            and H(..;z)                                   ******
***** applied for y>(1-z)/2                                    ******
***** Input: HXZn,HZn; Output: HYZn                            ******
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      integer iflag,nmax
      real(dp):: Zeta2,Zeta3,Zeta4
      real(dp):: HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      real(dp):: HXZ1,HXZ2,HXZ3,HXZ4
      dimension HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3),
     $          HYZ4(0:3,0:3,0:3,0:3)
      dimension HXZ1(0:3),HXZ2(0:3,0:3),HXZ3(0:3,0:3,0:3),
     $          HXZ4(0:3,0:3,0:3,0:3)
      dimension HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1),
     $          HZ4(-1:1,-1:1,-1:1,-1:1)
      parameter (Zeta2 = 1.6449340668482264365d0)
      parameter (Zeta3 = 1.2020569031595942854d0)
      parameter (Zeta4 = 1.0823232337111381915d0)

* 2001-04-23:17:22:16.hpl 
* <- swap.out 
* produced by form-to-fortr for gehrt@pcth62 

      HYZ1(0) = -HXZ1(2)-HZ1(1)
      HYZ1(1) = -HXZ1(3)-HZ1(0)
      HYZ1(2) = -HXZ1(0)-HZ1(1)
      HYZ1(3) = -HXZ1(1)-HZ1(0)

      if (nmax.eq.1) return
      HYZ2(0,0) = 
     $  + HXZ1(2) *HZ1(1)
     $  + HXZ2(2,2) 
     $  + HZ2(1,1) 
      HYZ2(0,1) = 
     $  + Zeta2 
     $  + HXZ1(2) *HZ1(0)
     $  + HXZ2(2,3) 
     $  + HZ2(1,0) 
      HYZ2(0,2) = 
     $  + Zeta2 
     $  + HXZ1(2) *HZ1(1)
     $  + HXZ2(2,0) 
      HYZ2(0,3) = 
     $  + Zeta2 
     $  + HXZ1(2) *HZ1(0)
     $  + HXZ2(2,1) 
     $  + HZ2(0,0) 
     $  + HZ2(1,0) 
      HYZ2(1,0) = 
     $  - Zeta2 
     $  + HXZ1(3) *HZ1(1)
     $  + HXZ2(3,2) 
     $  + HZ2(0,1) 
      HYZ2(1,1) = 
     $  + HXZ1(3) *HZ1(0)
     $  + HXZ2(3,3) 
     $  + HZ2(0,0) 
      HYZ2(1,2) = 
     $  + Zeta2 
     $  + HXZ1(3) *HZ1(1)
     $  + HXZ2(3,0) 
     $  + HZ2(0,0) 
     $  + HZ2(1,0) 
      HYZ2(1,3) = 
     $  - Zeta2 
     $  + HXZ1(3) *HZ1(0)
     $  + HXZ2(3,1) 
     $  - 2.000000000000000d+00*HZ2(-1,0)
     $  + 2.000000000000000d+00*HZ2(0,0)
      HYZ2(2,0) = 
     $  - Zeta2 
     $  + HXZ1(0) *HZ1(1)
     $  + HXZ2(0,2) 
     $  + 2.000000000000000d+00*HZ2(1,1)
      HYZ2(2,1) = 
     $  - Zeta2 
     $  + HXZ1(0) *HZ1(0)
     $  + HXZ2(0,3) 
     $  - HZ2(0,0) 
     $  + HZ2(0,1) 
      HYZ2(2,2) = 
     $  + HXZ1(0) *HZ1(1)
     $  + HXZ2(0,0) 
     $  + HZ2(1,1) 
      HYZ2(2,3) = 
     $  - Zeta2 
     $  + HXZ1(0) *HZ1(0)
     $  + HXZ2(0,1) 
     $  + HZ2(0,1) 
      HYZ2(3,0) = 
     $  - Zeta2 
     $  + HXZ1(1) *HZ1(1)
     $  + HXZ2(1,2) 
     $  - HZ2(0,0) 
     $  + HZ2(0,1) 
      HYZ2(3,1) = 
     $  + Zeta2 
     $  + HXZ1(1) *HZ1(0)
     $  + HXZ2(1,3) 
     $  + 2.000000000000000d+00*HZ2(-1,0)
      HYZ2(3,2) = 
     $  + Zeta2 
     $  + HXZ1(1) *HZ1(1)
     $  + HXZ2(1,0) 
     $  + HZ2(1,0) 
      HYZ2(3,3) = 
     $  + HXZ1(1) *HZ1(0)
     $  + HXZ2(1,1) 
     $  + HZ2(0,0) 
      if (nmax.eq.2) return
      HYZ3(0,0,0) = 
     $  - HXZ1(2) *HZ2(1,1)
     $  - HXZ2(2,2) *HZ1(1)
     $  - HXZ3(2,2,2) 
     $  - HZ3(1,1,1) 
      HYZ3(0,0,1) = 
     $  + Zeta3 
     $  - HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(1,0)
     $  - HXZ2(2,2) *HZ1(0)
     $  - HXZ3(2,2,3) 
     $  - HZ1(1) *Zeta2
     $  - HZ3(1,1,0) 
      HYZ3(0,0,2) = 
     $  + Zeta3 
     $  - HXZ1(2) *Zeta2
     $  - HXZ2(2,2) *HZ1(1)
     $  - HXZ3(2,2,0) 
      HYZ3(0,0,3) = 
     $  - HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(0,0)
     $  - HXZ1(2) *HZ2(1,0)
     $  - HXZ2(2,2) *HZ1(0)
     $  - HXZ3(2,2,1) 
     $  - HZ1(0) *Zeta2
     $  - HZ1(1) *Zeta2
     $  - HZ3(0,0,0) 
     $  - HZ3(0,1,0) 
     $  - HZ3(1,0,0) 
     $  - HZ3(1,1,0) 
      HYZ3(0,1,0) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(0,1)
     $  - HXZ2(2,3) *HZ1(1)
     $  - HXZ3(2,3,2) 
     $  + HZ1(1) *Zeta2
     $  - HZ3(1,0,1) 
      HYZ3(0,1,1) = 
     $  + Zeta3 
     $  - HXZ1(2) *HZ2(0,0)
     $  - HXZ2(2,3) *HZ1(0)
     $  - HXZ3(2,3,3) 
     $  - HZ3(1,0,0) 
      HYZ3(0,1,2) = 
     $  + Zeta3 
     $  - HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(0,0)
     $  - HXZ1(2) *HZ2(1,0)
     $  - HXZ2(2,3) *HZ1(1)
     $  - HXZ3(2,3,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta2
     $  - HZ3(0,1,0) 
     $  - HZ3(1,0,0) 
     $  - 2.000000000000000d+00*HZ3(1,1,0)
      HYZ3(0,1,3) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(2) *Zeta2
     $  + 2.000000000000000d+00*HXZ1(2)*HZ2(-1,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ2(0,0)
     $  - HXZ2(2,3) *HZ1(0)
     $  - HXZ3(2,3,1) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(0) *Zeta2
     $  + HZ1(1) *Zeta2
     $  + HZ3( -1,0,0)
     $  + 2.000000000000000d+00*HZ3(-1,1,0)
     $  - HZ3(0,1,0) 
     $  + 2.000000000000000d+00*HZ3(1,-1,0)
     $  - 2.000000000000000d+00*HZ3(1,0,0)
      HYZ3(0,2,0) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(2) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ2(1,1)
     $  - HXZ2(2,0) *HZ1(1)
     $  - HXZ3(2,0,2) 
     $  - HZ1(1) *Zeta2
      HYZ3(0,2,1) = 
     $  + Zeta3 
     $  + HXZ1(2) *Zeta2
     $  + HXZ1(2) *HZ2(0,0)
     $  - HXZ1(2) *HZ2(0,1)
     $  - HXZ2(2,0) *HZ1(0)
     $  - HXZ3(2,0,3) 
     $  + HZ1(1) *Zeta2
     $  + HZ3(0,1,0) 
     $  + HZ3(1,1,0) 
      HYZ3(0,2,2) = 
     $  + Zeta3 
     $  - HXZ1(2) *HZ2(1,1)
     $  - HXZ2(2,0) *HZ1(1)
     $  - HXZ3(2,0,0) 
      HYZ3(0,2,3) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(0,1)
     $  - HXZ2(2,0) *HZ1(0)
     $  - HXZ3(2,0,1) 
     $  - HZ1(0) *Zeta2
     $  + HZ1(1) *Zeta2
     $  + HZ3(1,0,0) 
     $  + HZ3(1,1,0) 
      HYZ3(0,3,0) = 
     $  + HXZ1(2) *Zeta2
     $  + HXZ1(2) *HZ2(0,0)
     $  - HXZ1(2) *HZ2(0,1)
     $  - HXZ2(2,1) *HZ1(1)
     $  - HXZ3(2,1,2) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta2
     $  + HZ1(1) *Zeta2
     $  + 2.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,0,1) 
     $  + HZ3(0,1,0) 
     $  + HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
      HYZ3(0,3,1) = 
     $  + Zeta3 
     $  - HXZ1(2) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ2(-1,0)
     $  - HXZ2(2,1) *HZ1(0)
     $  - HXZ3(2,1,3) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(1) *Zeta2
     $  - HZ3( -1,0,0)
     $  - 2.000000000000000d+00*HZ3(-1,1,0)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
     $  + HZ3(0,1,0) 
     $  - 2.000000000000000d+00*HZ3(1,-1,0)
      HYZ3(0,3,2) = 
     $  + Zeta3 
     $  - HXZ1(2) *Zeta2
     $  - HXZ1(2) *HZ2(1,0)
     $  - HXZ2(2,1) *HZ1(1)
     $  - HXZ3(2,1,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta2
     $  - HZ3(0,1,0) 
     $  - HZ3(1,0,0) 
     $  - 2.000000000000000d+00*HZ3(1,1,0)
      HYZ3(0,3,3) = 
     $  + Zeta3 
     $  - HXZ1(2) *HZ2(0,0)
     $  - HXZ2(2,1) *HZ1(0)
     $  - HXZ3(2,1,1) 
     $  - HZ3(0,0,0) 
     $  - HZ3(1,0,0) 
      HYZ3(1,0,0) = 
     $  + Zeta3 
     $  - HXZ1(3) *HZ2(1,1)
     $  - HXZ2(3,2) *HZ1(1)
     $  - HXZ3(3,2,2) 
     $  - HZ3(0,1,1) 
      HYZ3(1,0,1) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(1,0)
     $  - HXZ2(3,2) *HZ1(0)
     $  - HXZ3(3,2,3) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,1,0) 
      HYZ3(1,0,2) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(3) *Zeta2
     $  - HXZ2(3,2) *HZ1(1)
     $  - HXZ3(3,2,0) 
     $  - HZ1(0) *Zeta2
     $  + HZ1(1) *Zeta2
     $  + HZ3(1,0,0) 
     $  + HZ3(1,1,0) 
      HYZ3(1,0,3) = 
     $  + Zeta3 
     $  - HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(0,0)
     $  - HXZ1(3) *HZ2(1,0)
     $  - HXZ2(3,2) *HZ1(0)
     $  - HXZ3(3,2,1) 
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,1,0) 
      HYZ3(1,1,0) = 
     $  + Zeta3 
     $  + HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(0,1)
     $  - HXZ2(3,3) *HZ1(1)
     $  - HXZ3(3,3,2) 
     $  + HZ1(0) *Zeta2
     $  - HZ3(0,0,1) 
      HYZ3(1,1,1) = 
     $  - HXZ1(3) *HZ2(0,0)
     $  - HXZ2(3,3) *HZ1(0)
     $  - HXZ3(3,3,3) 
     $  - HZ3(0,0,0) 
      HYZ3(1,1,2) = 
     $  + Zeta3 
     $  - HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(0,0)
     $  - HXZ1(3) *HZ2(1,0)
     $  - HXZ2(3,3) *HZ1(1)
     $  - HXZ3(3,3,0) 
     $  - HZ3(0,0,0) 
     $  - HZ3(1,0,0) 
      HYZ3(1,1,3) = 
     $  + Zeta3 
     $  + HXZ1(3) *Zeta2
     $  + 2.000000000000000d+00*HXZ1(3)*HZ2(-1,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ2(0,0)
     $  - HXZ2(3,3) *HZ1(0)
     $  - HXZ3(3,3,1) 
     $  - HZ1( -1)*Zeta2
     $  + HZ1(0) *Zeta2
     $  - 2.000000000000000d+00*HZ3(-1,-1,0)
     $  + 3.000000000000000d+00*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HZ3(0,0,0)
      HYZ3(1,2,0) = 
     $  + Zeta3 
     $  + HXZ1(3) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ2(1,1)
     $  - HXZ2(3,0) *HZ1(1)
     $  - HXZ3(3,0,2) 
     $  + HZ1(0) *Zeta2
     $  - HZ3(0,0,1) 
     $  - HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
     $  - HZ3(1,1,0) 
      HYZ3(1,2,1) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(3) *Zeta2
     $  + HXZ1(3) *HZ2(0,0)
     $  - HXZ1(3) *HZ2(0,1)
     $  - HXZ2(3,0) *HZ1(0)
     $  - HXZ3(3,0,3) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,0,0) 
     $  - HZ3(0,1,0) 
      HYZ3(1,2,2) = 
     $  - HXZ1(3) *HZ2(1,1)
     $  - HXZ2(3,0) *HZ1(1)
     $  - HXZ3(3,0,0) 
     $  - HZ1(0) *Zeta2
     $  - HZ1(1) *Zeta2
     $  - HZ3(0,0,0) 
     $  - HZ3(0,1,0) 
     $  - HZ3(1,0,0) 
     $  - HZ3(1,1,0) 
      HYZ3(1,2,3) = 
     $  + Zeta3 
     $  + HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(0,1)
     $  - HXZ2(3,0) *HZ1(0)
     $  - HXZ3(3,0,1) 
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,1,0) 
      HYZ3(1,3,0) = 
     $  + Zeta3 
     $  + HXZ1(3) *Zeta2
     $  + HXZ1(3) *HZ2(0,0)
     $  - HXZ1(3) *HZ2(0,1)
     $  - HXZ2(3,1) *HZ1(1)
     $  - HXZ3(3,1,2) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  + HZ1(0) *Zeta2
     $  - HZ3( -1,0,0)
     $  + 2.000000000000000d+00*HZ3(-1,0,1)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HZ3(0,0,0)
     $  - 2.000000000000000d+00*HZ3(0,0,1)
      HYZ3(1,3,1) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(3) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ2(-1,0)
     $  - HXZ2(3,1) *HZ1(0)
     $  - HXZ3(3,1,3) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(0) *Zeta2
     $  + 4.000000000000000d+00*HZ3(-1,-1,0)
     $  - 2.000000000000000d+00*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
      HYZ3(1,3,2) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(3) *Zeta2
     $  - HXZ1(3) *HZ2(1,0)
     $  - HXZ2(3,1) *HZ1(1)
     $  - HXZ3(3,1,0) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(0) *Zeta2
     $  + HZ1(1) *Zeta2
     $  + HZ3( -1,0,0)
     $  + 2.000000000000000d+00*HZ3(-1,1,0)
     $  - HZ3(0,1,0) 
     $  + 2.000000000000000d+00*HZ3(1,-1,0)
     $  - 2.000000000000000d+00*HZ3(1,0,0)
      HYZ3(1,3,3) = 
     $  + Zeta3 
     $  - HXZ1(3) *HZ2(0,0)
     $  - HXZ2(3,1) *HZ1(0)
     $  - HXZ3(3,1,1) 
     $  - HZ1( -1)*Zeta2
     $  + HZ1(0) *Zeta2
     $  - 2.000000000000000d+00*HZ3(-1,-1,0)
     $  + 3.000000000000000d+00*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HZ3(0,0,0)
      HYZ3(2,0,0) = 
     $  + Zeta3 
     $  - HXZ1(0) *HZ2(1,1)
     $  - HXZ2(0,2) *HZ1(1)
     $  - HXZ3(0,2,2) 
     $  + HZ1(1) *Zeta2
     $  - 3.000000000000000d+00*HZ3(1,1,1)
      HYZ3(2,0,1) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(1,0)
     $  - HXZ2(0,2) *HZ1(0)
     $  - HXZ3(0,2,3) 
     $  + HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
     $  - HZ3(1,1,0) 
      HYZ3(2,0,2) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(0) *Zeta2
     $  - HXZ2(0,2) *HZ1(1)
     $  - HXZ3(0,2,0) 
     $  - HZ1(1) *Zeta2
      HYZ3(2,0,3) = 
     $  + Zeta3 
     $  - HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(0,0)
     $  - HXZ1(0) *HZ2(1,0)
     $  - HXZ2(0,2) *HZ1(0)
     $  - HXZ3(0,2,1) 
     $  + HZ1(0) *Zeta2
     $  - HZ3(0,0,1) 
     $  - HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
     $  - HZ3(1,1,0) 
      HYZ3(2,1,0) = 
     $  + Zeta3 
     $  + HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(0,1)
     $  - HXZ2(0,3) *HZ1(1)
     $  - HXZ3(0,3,2) 
     $  + HZ3(0,0,1) 
     $  - 2.000000000000000d+00*HZ3(0,1,1)
      HYZ3(2,1,1) = 
     $  + Zeta3 
     $  - HXZ1(0) *HZ2(0,0)
     $  - HXZ2(0,3) *HZ1(0)
     $  - HXZ3(0,3,3) 
     $  + HZ1(0) *Zeta2
     $  + 2.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,0,1) 
      HYZ3(2,1,2) = 
     $  - HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(0,0)
     $  - HXZ1(0) *HZ2(1,0)
     $  - HXZ2(0,3) *HZ1(1)
     $  - HXZ3(0,3,0) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta2
     $  + HZ1(1) *Zeta2
     $  + 2.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,0,1) 
     $  + HZ3(0,1,0) 
     $  + HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
      HYZ3(2,1,3) = 
     $  + Zeta3 
     $  + HXZ1(0) *Zeta2
     $  + 2.000000000000000d+00*HXZ1(0)*HZ2(-1,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ2(0,0)
     $  - HXZ2(0,3) *HZ1(0)
     $  - HXZ3(0,3,1) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  + HZ1(0) *Zeta2
     $  - HZ3( -1,0,0)
     $  + 2.000000000000000d+00*HZ3(-1,0,1)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HZ3(0,0,0)
     $  - 2.000000000000000d+00*HZ3(0,0,1)
      HYZ3(2,2,0) = 
     $  + Zeta3 
     $  + HXZ1(0) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ2(1,1)
     $  - HXZ2(0,0) *HZ1(1)
     $  - HXZ3(0,0,2) 
     $  + HZ1(1) *Zeta2
     $  - 3.000000000000000d+00*HZ3(1,1,1)
      HYZ3(2,2,1) = 
     $  + HXZ1(0) *Zeta2
     $  + HXZ1(0) *HZ2(0,0)
     $  - HXZ1(0) *HZ2(0,1)
     $  - HXZ2(0,0) *HZ1(0)
     $  - HXZ3(0,0,3) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,0,0) 
     $  + HZ3(0,0,1) 
     $  - HZ3(0,1,1) 
      HYZ3(2,2,2) = 
     $  - HXZ1(0) *HZ2(1,1)
     $  - HXZ2(0,0) *HZ1(1)
     $  - HXZ3(0,0,0) 
     $  - HZ3(1,1,1) 
      HYZ3(2,2,3) = 
     $  + Zeta3 
     $  + HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(0,1)
     $  - HXZ2(0,0) *HZ1(0)
     $  - HXZ3(0,0,1) 
     $  - HZ3(0,1,1) 
      HYZ3(2,3,0) = 
     $  + Zeta3 
     $  + HXZ1(0) *Zeta2
     $  + HXZ1(0) *HZ2(0,0)
     $  - HXZ1(0) *HZ2(0,1)
     $  - HXZ2(0,1) *HZ1(1)
     $  - HXZ3(0,1,2) 
     $  + HZ3(0,0,1) 
     $  - 2.000000000000000d+00*HZ3(0,1,1)
      HYZ3(2,3,1) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(0) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ2(-1,0)
     $  - HXZ2(0,1) *HZ1(0)
     $  - HXZ3(0,1,3) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  + HZ3( -1,0,0)
     $  - 2.000000000000000d+00*HZ3(-1,0,1)
      HYZ3(2,3,2) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(0) *Zeta2
     $  - HXZ1(0) *HZ2(1,0)
     $  - HXZ2(0,1) *HZ1(1)
     $  - HXZ3(0,1,0) 
     $  + HZ1(1) *Zeta2
     $  - HZ3(1,0,1) 
      HYZ3(2,3,3) = 
     $  + Zeta3 
     $  - HXZ1(0) *HZ2(0,0)
     $  - HXZ2(0,1) *HZ1(0)
     $  - HXZ3(0,1,1) 
     $  + HZ1(0) *Zeta2
     $  - HZ3(0,0,1) 
      HYZ3(3,0,0) = 
     $  - HXZ1(1) *HZ2(1,1)
     $  - HXZ2(1,2) *HZ1(1)
     $  - HXZ3(1,2,2) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,0,0) 
     $  + HZ3(0,0,1) 
     $  - HZ3(0,1,1) 
      HYZ3(3,0,1) = 
     $  + Zeta3 
     $  - HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(1,0)
     $  - HXZ2(1,2) *HZ1(0)
     $  - HXZ3(1,2,3) 
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - HZ3(0,1,0) 
      HYZ3(3,0,2) = 
     $  + Zeta3 
     $  - HXZ1(1) *Zeta2
     $  - HXZ2(1,2) *HZ1(1)
     $  - HXZ3(1,2,0) 
     $  + HZ1(1) *Zeta2
     $  + HZ3(0,1,0) 
     $  + HZ3(1,1,0) 
      HYZ3(3,0,3) = 
     $  - 2.000000000000000d+00*Zeta3
     $  - HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(0,0)
     $  - HXZ1(1) *HZ2(1,0)
     $  - HXZ2(1,2) *HZ1(0)
     $  - HXZ3(1,2,1) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,0,0) 
     $  - HZ3(0,1,0) 
      HYZ3(3,1,0) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(0,1)
     $  - HXZ2(1,3) *HZ1(1)
     $  - HXZ3(1,3,2) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  + HZ3( -1,0,0)
     $  - 2.000000000000000d+00*HZ3(-1,0,1)
      HYZ3(3,1,1) = 
     $  + Zeta3 
     $  - HXZ1(1) *HZ2(0,0)
     $  - HXZ2(1,3) *HZ1(0)
     $  - HXZ3(1,3,3) 
     $  - HZ1( -1)*Zeta2
     $  - 2.000000000000000d+00*HZ3(-1,-1,0)
     $  - HZ3( -1,0,0)
      HYZ3(3,1,2) = 
     $  + Zeta3 
     $  - HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(0,0)
     $  - HXZ1(1) *HZ2(1,0)
     $  - HXZ2(1,3) *HZ1(1)
     $  - HXZ3(1,3,0) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(1) *Zeta2
     $  - HZ3( -1,0,0)
     $  - 2.000000000000000d+00*HZ3(-1,1,0)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
     $  + HZ3(0,1,0) 
     $  - 2.000000000000000d+00*HZ3(1,-1,0)
      HYZ3(3,1,3) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(1) *Zeta2
     $  + 2.000000000000000d+00*HXZ1(1)*HZ2(-1,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ2(0,0)
     $  - HXZ2(1,3) *HZ1(0)
     $  - HXZ3(1,3,1) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta2
     $  - HZ1(0) *Zeta2
     $  + 4.000000000000000d+00*HZ3(-1,-1,0)
     $  - 2.000000000000000d+00*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HZ3(0,-1,0)
      HYZ3(3,2,0) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(1) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ2(1,1)
     $  - HXZ2(1,0) *HZ1(1)
     $  - HXZ3(1,0,2) 
     $  + HZ3(1,0,0) 
     $  - HZ3(1,0,1) 
     $  - HZ3(1,1,0) 
      HYZ3(3,2,1) = 
     $  + Zeta3 
     $  + HXZ1(1) *Zeta2
     $  + HXZ1(1) *HZ2(0,0)
     $  - HXZ1(1) *HZ2(0,1)
     $  - HXZ2(1,0) *HZ1(0)
     $  - HXZ3(1,0,3) 
     $  + 2.000000000000000d+00*HZ3(0,-1,0)
     $  - HZ3(0,1,0) 
      HYZ3(3,2,2) = 
     $  + Zeta3 
     $  - HXZ1(1) *HZ2(1,1)
     $  - HXZ2(1,0) *HZ1(1)
     $  - HXZ3(1,0,0) 
     $  - HZ1(1) *Zeta2
     $  - HZ3(1,1,0) 
      HYZ3(3,2,3) = 
     $  - 2.000000000000000d+00*Zeta3
     $  + HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(0,1)
     $  - HXZ2(1,0) *HZ1(0)
     $  - HXZ3(1,0,1) 
     $  - HZ1(0) *Zeta2
     $  - HZ3(0,1,0) 
      HYZ3(3,3,0) = 
     $  + Zeta3 
     $  + HXZ1(1) *Zeta2
     $  + HXZ1(1) *HZ2(0,0)
     $  - HXZ1(1) *HZ2(0,1)
     $  - HXZ2(1,1) *HZ1(1)
     $  - HXZ3(1,1,2) 
     $  + HZ1(0) *Zeta2
     $  + 2.000000000000000d+00*HZ3(0,0,0)
     $  - HZ3(0,0,1) 
      HYZ3(3,3,1) = 
     $  + Zeta3 
     $  - HXZ1(1) *Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ2(-1,0)
     $  - HXZ2(1,1) *HZ1(0)
     $  - HXZ3(1,1,3) 
     $  - HZ1( -1)*Zeta2
     $  - 2.000000000000000d+00*HZ3(-1,-1,0)
     $  - HZ3( -1,0,0)
      HYZ3(3,3,2) = 
     $  + Zeta3 
     $  - HXZ1(1) *Zeta2
     $  - HXZ1(1) *HZ2(1,0)
     $  - HXZ2(1,1) *HZ1(1)
     $  - HXZ3(1,1,0) 
     $  - HZ3(1,0,0) 
      HYZ3(3,3,3) = 
     $  - HXZ1(1) *HZ2(0,0)
     $  - HXZ2(1,1) *HZ1(0)
     $  - HXZ3(1,1,1) 
     $  - HZ3(0,0,0) 
      if (nmax.eq.3) return
      HYZ4(0,0,0,0) = 
     $  + HXZ1(2) *HZ3(1,1,1)
     $  + HXZ2(2,2) *HZ2(1,1)
     $  + HXZ3(2,2,2) *HZ1(1)
     $  + HXZ4(2,2,2,2) 
     $  + HZ4(1,1,1,1) 
      HYZ4(0,0,0,1) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(1,0)
     $  + HXZ3(2,2,2) *HZ1(0)
     $  + HXZ4(2,2,2,3) 
     $  - HZ1(1) *Zeta3
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(1,1,1,0) 
      HYZ4(0,0,0,2) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + HXZ2(2,2) *Zeta2
     $  + HXZ3(2,2,2) *HZ1(1)
     $  + HXZ4(2,2,2,0) 
      HYZ4(0,0,0,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(0,0)
     $  + HXZ2(2,2) *HZ2(1,0)
     $  + HXZ3(2,2,2) *HZ1(0)
     $  + HXZ4(2,2,2,1) 
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(0,0,1,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(1,0,1)
     $  - HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(0,1)
     $  + HXZ3(2,2,3) *HZ1(1)
     $  + HXZ4(2,2,3,2) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,1) *Zeta2
     $  + HZ4(1,1,0,1) 
      HYZ4(0,0,1,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ2(2,2) *HZ2(0,0)
     $  + HXZ3(2,2,3) *HZ1(0)
     $  + HXZ4(2,2,3,3) 
     $  - HZ1(1) *Zeta3
     $  + HZ4(1,1,0,0) 
      HYZ4(0,0,1,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,1,0)
     $  + HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(0,0)
     $  + HXZ2(2,2) *HZ2(1,0)
     $  + HXZ3(2,2,3) *HZ1(1)
     $  + HXZ4(2,2,3,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,0,1,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(-1,1,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,0,0)
     $  - HXZ2(2,2) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(2,2)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(2,2)*HZ2(0,0)
     $  + HXZ3(2,2,3) *HZ1(0)
     $  + HXZ4(2,2,3,1) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0,1,1,0) 
     $  - HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(0,0,2,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ2(2,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,2)*HZ2(1,1)
     $  + HXZ3(2,2,0) *HZ1(1)
     $  + HXZ4(2,2,0,2) 
     $  - HZ1(1) *Zeta3
      HYZ4(0,0,2,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(0,1,0)
     $  - HXZ1(2) *HZ3(1,1,0)
     $  - HXZ2(2,2) *Zeta2
     $  - HXZ2(2,2) *HZ2(0,0)
     $  + HXZ2(2,2) *HZ2(0,1)
     $  + HXZ3(2,2,0) *HZ1(0)
     $  + HXZ4(2,2,0,3) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(0,0,2,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ2(2,2) *HZ2(1,1)
     $  + HXZ3(2,2,0) *HZ1(1)
     $  + HXZ4(2,2,0,0) 
      HYZ4(0,0,2,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(1,0,0)
     $  - HXZ1(2) *HZ3(1,1,0)
     $  - HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(0,1)
     $  + HXZ3(2,2,0) *HZ1(0)
     $  + HXZ4(2,2,0,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,0) 
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,0,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(0,0,3,0) = 
     $  - 5.250000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,0,1)
     $  - HXZ1(2) *HZ3(0,1,0)
     $  - HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  - HXZ2(2,2) *Zeta2
     $  - HXZ2(2,2) *HZ2(0,0)
     $  + HXZ2(2,2) *HZ2(0,1)
     $  + HXZ3(2,2,1) *HZ1(1)
     $  + HXZ4(2,2,1,2) 
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
      HYZ4(0,0,3,1) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  - HXZ1(2) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,-1,0)
     $  + HXZ2(2,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,2)*HZ2(-1,0)
     $  + HXZ3(2,2,1) *HZ1(0)
     $  + HXZ4(2,2,1,3) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,0,0)
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  - HZ4(0,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - HZ4(0,1,1,0) 
     $  + HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(0,0,3,2) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,1,0)
     $  + HXZ2(2,2) *Zeta2
     $  + HXZ2(2,2) *HZ2(1,0)
     $  + HXZ3(2,2,1) *HZ1(1)
     $  + HXZ4(2,2,1,0) 
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,0,3,3) = 
     $  - 1.750000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ2(2,2) *HZ2(0,0)
     $  + HXZ3(2,2,1) *HZ1(0)
     $  + HXZ4(2,2,1,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,1,0,0) 
      HYZ4(0,1,0,0) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(0,1,1)
     $  + HXZ2(2,3) *HZ2(1,1)
     $  + HXZ3(2,3,2) *HZ1(1)
     $  + HXZ4(2,3,2,2) 
     $  - HZ1(1) *Zeta3
     $  + HZ4(1,0,1,1) 
      HYZ4(0,1,0,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(1,0)
     $  + HXZ3(2,3,2) *HZ1(0)
     $  + HXZ4(2,3,2,3) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(1,0,1,0) 
      HYZ4(0,1,0,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(1,0,0)
     $  - HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,3) *Zeta2
     $  + HXZ3(2,3,2) *HZ1(1)
     $  + HXZ4(2,3,2,0) 
     $  + 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  - 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  - HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
     $  - 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,1,0,3) = 
     $  + 5.500000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(0,0)
     $  + HXZ2(2,3) *HZ2(1,0)
     $  + HXZ3(2,3,2) *HZ1(0)
     $  + HXZ4(2,3,2,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(0,1,1,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,1)
     $  - HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(0,1)
     $  + HXZ3(2,3,3) *HZ1(1)
     $  + HXZ4(2,3,3,2) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(1,0,0,1) 
      HYZ4(0,1,1,1) = 
     $  + Zeta4 
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ2(2,3) *HZ2(0,0)
     $  + HXZ3(2,3,3) *HZ1(0)
     $  + HXZ4(2,3,3,3) 
     $  + HZ4(1,0,0,0) 
      HYZ4(0,1,1,2) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(0,0)
     $  + HXZ2(2,3) *HZ2(1,0)
     $  + HXZ3(2,3,3) *HZ1(1)
     $  + HXZ4(2,3,3,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(0,1,1,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(2)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  - HXZ2(2,3) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(2,3)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(2,3)*HZ2(0,0)
     $  + HXZ3(2,3,3) *HZ1(0)
     $  + HXZ4(2,3,3,1) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,1,0,0)
     $  + HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
      HYZ4(0,1,2,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,1)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  - HXZ2(2,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,3)*HZ2(1,1)
     $  + HXZ3(2,3,0) *HZ1(1)
     $  + HXZ4(2,3,0,2) 
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,1) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,1)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,1,2,1) = 
     $  + Zeta4 
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - HXZ2(2,3) *Zeta2
     $  - HXZ2(2,3) *HZ2(0,0)
     $  + HXZ2(2,3) *HZ2(0,1)
     $  + HXZ3(2,3,0) *HZ1(0)
     $  + HXZ4(2,3,0,3) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + HZ4(0,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
      HYZ4(0,1,2,2) = 
     $  + Zeta4 
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,3) *HZ2(1,1)
     $  + HXZ3(2,3,0) *HZ1(1)
     $  + HXZ4(2,3,0,0) 
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,1,2,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(0,1)
     $  + HXZ3(2,3,0) *HZ1(0)
     $  + HXZ4(2,3,0,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,1,0,0)
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(0,1,3,0) = 
     $  + 5.000000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,1)
     $  - HXZ2(2,3) *Zeta2
     $  - HXZ2(2,3) *HZ2(0,0)
     $  + HXZ2(2,3) *HZ2(0,1)
     $  + HXZ3(2,3,1) *HZ1(1)
     $  + HXZ4(2,3,1,2) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  - HZ4( -1,0,0,1)
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - HZ4(0,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + HZ4(0,1,0,1) 
     $  + HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,0,1)
      HYZ4(0,1,3,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + HXZ2(2,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,3)*HZ2(-1,0)
     $  + HXZ3(2,3,1) *HZ1(0)
     $  + HXZ4(2,3,1,3) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  - 4.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
      HYZ4(0,1,3,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(-1,1,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,0,0)
     $  + HXZ2(2,3) *Zeta2
     $  + HXZ2(2,3) *HZ2(1,0)
     $  + HXZ3(2,3,1) *HZ1(1)
     $  + HXZ4(2,3,1,0) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 4.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 4.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(1,1,-1,0)
     $  + 4.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(0,1,3,3) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(2)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ2(2,3) *HZ2(0,0)
     $  + HXZ3(2,3,1) *HZ1(0)
     $  + HXZ4(2,3,1,1) 
     $  - HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,1,0,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
      HYZ4(0,2,0,0) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(1,1,1)
     $  + HXZ2(2,0) *HZ2(1,1)
     $  + HXZ3(2,0,2) *HZ1(1)
     $  + HXZ4(2,0,2,2) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,1) *Zeta2
      HYZ4(0,2,0,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(1,0)
     $  + HXZ3(2,0,2) *HZ1(0)
     $  + HXZ4(2,0,2,3) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(0,2,0,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ2(2,0) *Zeta2
     $  + HXZ3(2,0,2) *HZ1(1)
     $  + HXZ4(2,0,2,0) 
      HYZ4(0,2,0,3) = 
     $  + 5.500000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,1)
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(0,0)
     $  + HXZ2(2,0) *HZ2(1,0)
     $  + HXZ3(2,0,2) *HZ1(0)
     $  + HXZ4(2,0,2,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,0,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(0,2,1,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,1,1)
     $  - HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(0,1)
     $  + HXZ3(2,0,3) *HZ1(1)
     $  + HXZ4(2,0,3,2) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,1) 
     $  - HZ4(1,1,0,1) 
      HYZ4(0,2,1,1) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,0,1)
     $  + HXZ2(2,0) *HZ2(0,0)
     $  + HXZ3(2,0,3) *HZ1(0)
     $  + HXZ4(2,0,3,3) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,0,1,0) 
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,0,0) 
      HYZ4(0,2,1,2) = 
     $  + Zeta4 
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(0)*Zeta2
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,0,1)
     $  - HXZ1(2) *HZ3(0,1,0)
     $  - HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(0,0)
     $  + HXZ2(2,0) *HZ2(1,0)
     $  + HXZ3(2,0,3) *HZ1(1)
     $  + HXZ4(2,0,3,0) 
     $  - 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  - 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
     $  - HZ4(1,1,0,0) 
     $  - 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,2,1,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,1)
     $  - HXZ2(2,0) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(2,0)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(2,0)*HZ2(0,0)
     $  + HXZ3(2,0,3) *HZ1(0)
     $  + HXZ4(2,0,3,1) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(0,2,2,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(2)*HZ3(1,1,1)
     $  - HXZ2(2,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,0)*HZ2(1,1)
     $  + HXZ3(2,0,0) *HZ1(1)
     $  + HXZ4(2,0,0,2) 
     $  - HZ1(1) *Zeta3
      HYZ4(0,2,2,1) = 
     $  + Zeta4 
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  - HXZ1(2) *HZ3(0,0,1)
     $  + HXZ1(2) *HZ3(0,1,1)
     $  - HXZ2(2,0) *Zeta2
     $  - HXZ2(2,0) *HZ2(0,0)
     $  + HXZ2(2,0) *HZ2(0,1)
     $  + HXZ3(2,0,0) *HZ1(0)
     $  + HXZ4(2,0,0,3) 
     $  + HZ1(1) *Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(0,2,2,2) = 
     $  + Zeta4 
     $  + HXZ1(2) *HZ3(1,1,1)
     $  + HXZ2(2,0) *HZ2(1,1)
     $  + HXZ3(2,0,0) *HZ1(1)
     $  + HXZ4(2,0,0,0) 
      HYZ4(0,2,2,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(0,1,1)
     $  - HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(0,1)
     $  + HXZ3(2,0,0) *HZ1(0)
     $  + HXZ4(2,0,0,1) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(0,2,3,0) = 
     $  + 5.000000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,1,1)
     $  - HXZ2(2,0) *Zeta2
     $  - HXZ2(2,0) *HZ2(0,0)
     $  + HXZ2(2,0) *HZ2(0,1)
     $  + HXZ3(2,0,1) *HZ1(1)
     $  + HXZ4(2,0,1,2) 
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,0,0)
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  - HZ4(1,1,0,1) 
      HYZ4(0,2,3,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,1)
     $  + HXZ2(2,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,0)*HZ2(-1,0)
     $  + HXZ3(2,0,1) *HZ1(0)
     $  + HXZ4(2,0,1,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(0,2,3,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ2(2,0) *Zeta2
     $  + HXZ2(2,0) *HZ2(1,0)
     $  + HXZ3(2,0,1) *HZ1(1)
     $  + HXZ4(2,0,1,0) 
     $  + 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  - 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  - HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
     $  - 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,2,3,3) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,1)
     $  + HXZ2(2,0) *HZ2(0,0)
     $  + HXZ3(2,0,1) *HZ1(0)
     $  + HXZ4(2,0,1,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,0,0,0) 
     $  - HZ4(1,1,0,0) 
      HYZ4(0,3,0,0) = 
     $  + 5.250000000000000d+00*Zeta4
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  - HXZ1(2) *HZ3(0,0,1)
     $  + HXZ1(2) *HZ3(0,1,1)
     $  + HXZ2(2,1) *HZ2(1,1)
     $  + HXZ3(2,1,2) *HZ1(1)
     $  + HXZ4(2,1,2,2) 
     $  + 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 3.000000000000000d+00*HZ4(0,0,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,0,1,1) 
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  + HZ4(1,0,0,0) 
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
      HYZ4(0,3,0,1) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(1,0)
     $  + HXZ3(2,1,2) *HZ1(0)
     $  + HXZ4(2,1,2,3) 
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(0,3,0,2) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(1)*Zeta2
     $  - HXZ1(2) *HZ3(0,1,0)
     $  - HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,1) *Zeta2
     $  + HXZ3(2,1,2) *HZ1(1)
     $  + HXZ4(2,1,2,0) 
     $  - 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  - 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
     $  - HZ4(1,1,0,0) 
     $  - 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,3,0,3) = 
     $  + 4.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  + HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(0,0)
     $  + HXZ2(2,1) *HZ2(1,0)
     $  + HXZ3(2,1,2) *HZ1(0)
     $  + HXZ4(2,1,2,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
      HYZ4(0,3,1,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  - HXZ1(2) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,1)
     $  - HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(0,1)
     $  + HXZ3(2,1,3) *HZ1(1)
     $  + HXZ4(2,1,3,2) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  + HZ4( -1,0,0,1)
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - HZ4(0,1,0,1) 
     $  - HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,1)
      HYZ4(0,3,1,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  + HXZ2(2,1) *HZ2(0,0)
     $  + HXZ3(2,1,3) *HZ1(0)
     $  + HXZ4(2,1,3,3) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
     $  - HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + HZ4(1, -1,0,0)
      HYZ4(0,3,1,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  - HXZ1(2) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(1,-1,0)
     $  + HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(0,0)
     $  + HXZ2(2,1) *HZ2(1,0)
     $  + HXZ3(2,1,3) *HZ1(1)
     $  + HXZ4(2,1,3,0) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 4.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + 4.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  + 4.000000000000000d+00*HZ4(-1,1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
     $  + 4.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(0,3,1,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ1(-1)*Zeta2
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  - HXZ2(2,1) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(2,1)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(2,1)*HZ2(0,0)
     $  + HXZ3(2,1,3) *HZ1(0)
     $  + HXZ4(2,1,3,1) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 4.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
      HYZ4(0,3,2,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  - HXZ1(2) *HZ3(1,0,0)
     $  + HXZ1(2) *HZ3(1,0,1)
     $  + HXZ1(2) *HZ3(1,1,0)
     $  - HXZ2(2,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,1)*HZ2(1,1)
     $  + HXZ3(2,1,0) *HZ1(1)
     $  + HXZ4(2,1,0,2) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  - HZ4(1,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,1)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,3,2,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,-1,0)
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - HXZ2(2,1) *Zeta2
     $  - HXZ2(2,1) *HZ2(0,0)
     $  + HXZ2(2,1) *HZ2(0,1)
     $  + HXZ3(2,1,0) *HZ1(0)
     $  + HXZ4(2,1,0,3) 
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + HZ4(0,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(0,3,2,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(1)*Zeta2
     $  + HXZ1(2) *HZ3(1,1,0)
     $  + HXZ2(2,1) *HZ2(1,1)
     $  + HXZ3(2,1,0) *HZ1(1)
     $  + HXZ4(2,1,0,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + 3.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(0,3,2,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(2)*Zeta3
     $  + HXZ1(2) *HZ1(0)*Zeta2
     $  + HXZ1(2) *HZ3(0,1,0)
     $  - HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(0,1)
     $  + HXZ3(2,1,0) *HZ1(0)
     $  + HXZ4(2,1,0,1) 
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(0,3,3,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(2) *Zeta3
     $  - HXZ1(2) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(2)*HZ3(0,0,0)
     $  + HXZ1(2) *HZ3(0,0,1)
     $  - HXZ2(2,1) *Zeta2
     $  - HXZ2(2,1) *HZ2(0,0)
     $  + HXZ2(2,1) *HZ2(0,1)
     $  + HXZ3(2,1,1) *HZ1(1)
     $  + HXZ4(2,1,1,2) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - HZ2(1,0) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - HZ4(0,1,0,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
      HYZ4(0,3,3,1) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(2)*HZ3(-1,-1,0)
     $  + HXZ1(2) *HZ3(-1,0,0)
     $  + HXZ2(2,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(2,1)*HZ2(-1,0)
     $  + HXZ3(2,1,1) *HZ1(0)
     $  + HXZ4(2,1,1,3) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  - HZ2(0, -1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + HZ4(1, -1,0,0)
      HYZ4(0,3,3,2) = 
     $  + Zeta4 
     $  - HXZ1(2) *Zeta3
     $  + HXZ1(2) *HZ3(1,0,0)
     $  + HXZ2(2,1) *Zeta2
     $  + HXZ2(2,1) *HZ2(1,0)
     $  + HXZ3(2,1,1) *HZ1(1)
     $  + HXZ4(2,1,1,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(0,3,3,3) = 
     $  + Zeta4 
     $  + HXZ1(2) *HZ3(0,0,0)
     $  + HXZ2(2,1) *HZ2(0,0)
     $  + HXZ3(2,1,1) *HZ1(0)
     $  + HXZ4(2,1,1,1) 
     $  + HZ4(0,0,0,0) 
     $  + HZ4(1,0,0,0) 
      HYZ4(1,0,0,0) = 
     $  - Zeta4 
     $  + HXZ1(3) *HZ3(1,1,1)
     $  + HXZ2(3,2) *HZ2(1,1)
     $  + HXZ3(3,2,2) *HZ1(1)
     $  + HXZ4(3,2,2,2) 
     $  + HZ4(0,1,1,1) 
      HYZ4(1,0,0,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(1,0)
     $  + HXZ3(3,2,2) *HZ1(0)
     $  + HXZ4(3,2,2,3) 
     $  - HZ1(0) *Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + HZ4(0,1,1,0) 
      HYZ4(1,0,0,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ2(3,2) *Zeta2
     $  + HXZ3(3,2,2) *HZ1(1)
     $  + HXZ4(3,2,2,0) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(1,0,0,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(0,0)
     $  + HXZ2(3,2) *HZ2(1,0)
     $  + HXZ3(3,2,2) *HZ1(0)
     $  + HXZ4(3,2,2,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(1,0,1,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(1,0,1)
     $  - HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(0,1)
     $  + HXZ3(3,2,3) *HZ1(1)
     $  + HXZ4(3,2,3,2) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0,1,0,1) 
      HYZ4(1,0,1,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ2(3,2) *HZ2(0,0)
     $  + HXZ3(3,2,3) *HZ1(0)
     $  + HXZ4(3,2,3,3) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,1,0,0) 
      HYZ4(1,0,1,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,1,0)
     $  + HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(0,0)
     $  + HXZ2(3,2) *HZ2(1,0)
     $  + HXZ3(3,2,3) *HZ1(1)
     $  + HXZ4(3,2,3,0) 
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(1,0,1,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(-1,1,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,0,0)
     $  - HXZ2(3,2) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(3,2)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(3,2)*HZ2(0,0)
     $  + HXZ3(3,2,3) *HZ1(0)
     $  + HXZ4(3,2,3,1) 
     $  - HZ1( -1)*Zeta3
     $  + 4.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
      HYZ4(1,0,2,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ2(3,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,2)*HZ2(1,1)
     $  + HXZ3(3,2,0) *HZ1(1)
     $  + HXZ4(3,2,0,2) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
     $  - HZ4(1,1,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(1,0,2,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(0,1,0)
     $  - HXZ1(3) *HZ3(1,1,0)
     $  - HXZ2(3,2) *Zeta2
     $  - HXZ2(3,2) *HZ2(0,0)
     $  + HXZ2(3,2) *HZ2(0,1)
     $  + HXZ3(3,2,0) *HZ1(0)
     $  + HXZ4(3,2,0,3) 
     $  - HZ1(0) *Zeta3
     $  - 4.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  - HZ4(1,0,0,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(1,0,2,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ2(3,2) *HZ2(1,1)
     $  + HXZ3(3,2,0) *HZ1(1)
     $  + HXZ4(3,2,0,0) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,0) 
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,0,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(1,0,2,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(1,0,0)
     $  - HXZ1(3) *HZ3(1,1,0)
     $  - HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(0,1)
     $  + HXZ3(3,2,0) *HZ1(0)
     $  + HXZ4(3,2,0,1) 
     $  + 4.000000000000000d+00*HZ1(0)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + 4.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(1,0,3,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,0,1)
     $  - HXZ1(3) *HZ3(0,1,0)
     $  - HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  - HXZ2(3,2) *Zeta2
     $  - HXZ2(3,2) *HZ2(0,0)
     $  + HXZ2(3,2) *HZ2(0,1)
     $  + HXZ3(3,2,1) *HZ1(1)
     $  + HXZ4(3,2,1,2) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 8.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
      HYZ4(1,0,3,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  - HXZ1(3) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,-1,0)
     $  + HXZ2(3,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,2)*HZ2(-1,0)
     $  + HXZ3(3,2,1) *HZ1(0)
     $  + HXZ4(3,2,1,3) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
      HYZ4(1,0,3,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,1,0)
     $  + HXZ2(3,2) *Zeta2
     $  + HXZ2(3,2) *HZ2(1,0)
     $  + HXZ3(3,2,1) *HZ1(1)
     $  + HXZ4(3,2,1,0) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,1,0,0)
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(1,0,3,3) = 
     $  - Zeta4 
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ2(3,2) *HZ2(0,0)
     $  + HXZ3(3,2,1) *HZ1(0)
     $  + HXZ4(3,2,1,1) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,1,0,0) 
      HYZ4(1,1,0,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(0,1,1)
     $  + HXZ2(3,3) *HZ2(1,1)
     $  + HXZ3(3,3,2) *HZ1(1)
     $  + HXZ4(3,3,2,2) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,0,1,1) 
      HYZ4(1,1,0,1) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(1,0)
     $  + HXZ3(3,3,2) *HZ1(0)
     $  + HXZ4(3,3,2,3) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ4(0,0,1,0) 
      HYZ4(1,1,0,2) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(1,0,0)
     $  - HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,3) *Zeta2
     $  + HXZ3(3,3,2) *HZ1(1)
     $  + HXZ4(3,3,2,0) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,0,0,0) 
     $  - HZ4(1,1,0,0) 
      HYZ4(1,1,0,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(0,0)
     $  + HXZ2(3,3) *HZ2(1,0)
     $  + HXZ3(3,3,2) *HZ1(0)
     $  + HXZ4(3,3,2,1) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,1,0) 
      HYZ4(1,1,1,0) = 
     $  - Zeta4 
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,1)
     $  - HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(0,1)
     $  + HXZ3(3,3,3) *HZ1(1)
     $  + HXZ4(3,3,3,2) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,1) 
      HYZ4(1,1,1,1) = 
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ2(3,3) *HZ2(0,0)
     $  + HXZ3(3,3,3) *HZ1(0)
     $  + HXZ4(3,3,3,3) 
     $  + HZ4(0,0,0,0) 
      HYZ4(1,1,1,2) = 
     $  + Zeta4 
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(0,0)
     $  + HXZ2(3,3) *HZ2(1,0)
     $  + HXZ3(3,3,3) *HZ1(1)
     $  + HXZ4(3,3,3,0) 
     $  + HZ4(0,0,0,0) 
     $  + HZ4(1,0,0,0) 
      HYZ4(1,1,1,3) = 
     $  - Zeta4 
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(3)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  - HXZ2(3,3) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(3,3)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(3,3)*HZ2(0,0)
     $  + HXZ3(3,3,3) *HZ1(0)
     $  + HXZ4(3,3,3,1) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
      HYZ4(1,1,2,0) = 
     $  - Zeta4 
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,1)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  - HXZ2(3,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,3)*HZ2(1,1)
     $  + HXZ3(3,3,0) *HZ1(1)
     $  + HXZ4(3,3,0,2) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,1,0,0) 
      HYZ4(1,1,2,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - HXZ2(3,3) *Zeta2
     $  - HXZ2(3,3) *HZ2(0,0)
     $  + HXZ2(3,3) *HZ2(0,1)
     $  + HXZ3(3,3,0) *HZ1(0)
     $  + HXZ4(3,3,0,3) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,1,0,0) 
      HYZ4(1,1,2,2) = 
     $  - 1.750000000000000d+00*Zeta4
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,3) *HZ2(1,1)
     $  + HXZ3(3,3,0) *HZ1(1)
     $  + HXZ4(3,3,0,0) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,1,0,0) 
      HYZ4(1,1,2,3) = 
     $  - Zeta4 
     $  - HXZ1(3) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(0,1)
     $  + HXZ3(3,3,0) *HZ1(0)
     $  + HXZ4(3,3,0,1) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,1,0,0) 
      HYZ4(1,1,3,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,1)
     $  - HXZ2(3,3) *Zeta2
     $  - HXZ2(3,3) *HZ2(0,0)
     $  + HXZ2(3,3) *HZ2(0,1)
     $  + HXZ3(3,3,1) *HZ1(1)
     $  + HXZ4(3,3,1,2) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
      HYZ4(1,1,3,1) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + HXZ2(3,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,3)*HZ2(-1,0)
     $  + HXZ3(3,3,1) *HZ1(0)
     $  + HXZ4(3,3,1,3) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 3.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  + 6.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  - 5.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(1,1,3,2) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(-1,1,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,0,0)
     $  + HXZ2(3,3) *Zeta2
     $  + HXZ2(3,3) *HZ2(1,0)
     $  + HXZ3(3,3,1) *HZ1(1)
     $  + HXZ4(3,3,1,0) 
     $  - HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,1,0,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
      HYZ4(1,1,3,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(3)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ2(3,3) *HZ2(0,0)
     $  + HXZ3(3,3,1) *HZ1(0)
     $  + HXZ4(3,3,1,1) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - 4.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + 6.000000000000000d+00*HZ4(-1,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 6.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(0,0,0,0)
      HYZ4(1,2,0,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(1,1,1)
     $  + HXZ2(3,0) *HZ2(1,1)
     $  + HXZ3(3,0,2) *HZ1(1)
     $  + HXZ4(3,0,2,2) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(0,0,1,1) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + HZ4(1,1,1,0) 
      HYZ4(1,2,0,1) = 
     $  + 5.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(1,0)
     $  + HXZ3(3,0,2) *HZ1(0)
     $  + HXZ4(3,0,2,3) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 4.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(1,2,0,2) = 
     $  + 5.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ2(3,0) *Zeta2
     $  + HXZ3(3,0,2) *HZ1(1)
     $  + HXZ4(3,0,2,0) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,0,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(1,2,0,3) = 
     $  + 2.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,1)
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(0,0)
     $  + HXZ2(3,0) *HZ2(1,0)
     $  + HXZ3(3,0,2) *HZ1(0)
     $  + HXZ4(3,0,2,1) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 4.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(1,2,1,0) = 
     $  - 3.500000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,1,1)
     $  - HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(0,1)
     $  + HXZ3(3,0,3) *HZ1(1)
     $  + HXZ4(3,0,3,2) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(1,2,1,1) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,0,1)
     $  + HXZ2(3,0) *HZ2(0,0)
     $  + HXZ3(3,0,3) *HZ1(0)
     $  + HXZ4(3,0,3,3) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
      HYZ4(1,2,1,2) = 
     $  + 4.750000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(0)*Zeta2
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,0,1)
     $  - HXZ1(3) *HZ3(0,1,0)
     $  - HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(0,0)
     $  + HXZ2(3,0) *HZ2(1,0)
     $  + HXZ3(3,0,3) *HZ1(1)
     $  + HXZ4(3,0,3,0) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
      HYZ4(1,2,1,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,1)
     $  - HXZ2(3,0) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(3,0)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(3,0)*HZ2(0,0)
     $  + HXZ3(3,0,3) *HZ1(0)
     $  + HXZ4(3,0,3,1) 
     $  - HZ1( -1)*Zeta3
     $  + HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
      HYZ4(1,2,2,0) = 
     $  - 3.500000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(3)*HZ3(1,1,1)
     $  - HXZ2(3,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,0)*HZ2(1,1)
     $  + HXZ3(3,0,0) *HZ1(1)
     $  + HXZ4(3,0,0,2) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + HZ4(1,1,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(1,2,2,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  - HXZ1(3) *HZ3(0,0,1)
     $  + HXZ1(3) *HZ3(0,1,1)
     $  - HXZ2(3,0) *Zeta2
     $  - HXZ2(3,0) *HZ2(0,0)
     $  + HXZ2(3,0) *HZ2(0,1)
     $  + HXZ3(3,0,0) *HZ1(0)
     $  + HXZ4(3,0,0,3) 
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(1,2,2,2) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + HXZ1(3) *HZ3(1,1,1)
     $  + HXZ2(3,0) *HZ2(1,1)
     $  + HXZ3(3,0,0) *HZ1(1)
     $  + HXZ4(3,0,0,0) 
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(1,2,2,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(0,1,1)
     $  - HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(0,1)
     $  + HXZ3(3,0,0) *HZ1(0)
     $  + HXZ4(3,0,0,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(1,2,3,0) = 
     $  - 2.750000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,1,1)
     $  - HXZ2(3,0) *Zeta2
     $  - HXZ2(3,0) *HZ2(0,0)
     $  + HXZ2(3,0) *HZ2(0,1)
     $  + HXZ3(3,0,1) *HZ1(1)
     $  + HXZ4(3,0,1,2) 
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
     $  - HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(1,2,3,1) = 
     $  + 5.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,1)
     $  + HXZ2(3,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,0)*HZ2(-1,0)
     $  + HXZ3(3,0,1) *HZ1(0)
     $  + HXZ4(3,0,1,3) 
     $  + HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(1,2,3,2) = 
     $  + 5.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ2(3,0) *Zeta2
     $  + HXZ2(3,0) *HZ2(1,0)
     $  + HXZ3(3,0,1) *HZ1(1)
     $  + HXZ4(3,0,1,0) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(1,2,3,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,1)
     $  + HXZ2(3,0) *HZ2(0,0)
     $  + HXZ3(3,0,1) *HZ1(0)
     $  + HXZ4(3,0,1,1) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,1,0) 
      HYZ4(1,3,0,0) = 
     $  - Zeta4 
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  - HXZ1(3) *HZ3(0,0,1)
     $  + HXZ1(3) *HZ3(0,1,1)
     $  + HXZ2(3,1) *HZ2(1,1)
     $  + HXZ3(3,1,2) *HZ1(1)
     $  + HXZ4(3,1,2,2) 
     $  + HZ1( -1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ4( -1,0,0,0)
     $  + HZ4( -1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  - 3.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
      HYZ4(1,3,0,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(1,0)
     $  + HXZ3(3,1,2) *HZ1(0)
     $  + HXZ4(3,1,2,3) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
      HYZ4(1,3,0,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(1)*Zeta2
     $  - HXZ1(3) *HZ3(0,1,0)
     $  - HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,1) *Zeta2
     $  + HXZ3(3,1,2) *HZ1(1)
     $  + HXZ4(3,1,2,0) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(1,3,0,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  + HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(0,0)
     $  + HXZ2(3,1) *HZ2(1,0)
     $  + HXZ3(3,1,2) *HZ1(0)
     $  + HXZ4(3,1,2,1) 
     $  - HZ1( -1)*Zeta3
     $  + HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
      HYZ4(1,3,1,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  - HXZ1(3) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,1)
     $  - HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(0,1)
     $  + HXZ3(3,1,3) *HZ1(1)
     $  + HXZ4(3,1,3,2) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
      HYZ4(1,3,1,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  + HXZ2(3,1) *HZ2(0,0)
     $  + HXZ3(3,1,3) *HZ1(0)
     $  + HXZ4(3,1,3,3) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 3.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - 6.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
      HYZ4(1,3,1,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  - HXZ1(3) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(1,-1,0)
     $  + HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(0,0)
     $  + HXZ2(3,1) *HZ2(1,0)
     $  + HXZ3(3,1,3) *HZ1(1)
     $  + HXZ4(3,1,3,0) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  - 4.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
      HYZ4(1,3,1,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ1(-1)*Zeta2
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  - HXZ2(3,1) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(3,1)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(3,1)*HZ2(0,0)
     $  + HXZ3(3,1,3) *HZ1(0)
     $  + HXZ4(3,1,3,1) 
     $  - 4.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 4.000000000000000d+00*HZ1(0)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 4.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + 8.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 8.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(1,3,2,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  - HXZ1(3) *HZ3(1,0,0)
     $  + HXZ1(3) *HZ3(1,0,1)
     $  + HXZ1(3) *HZ3(1,1,0)
     $  - HXZ2(3,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,1)*HZ2(1,1)
     $  + HXZ3(3,1,0) *HZ1(1)
     $  + HXZ4(3,1,0,2) 
     $  - HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,0,1)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(1,3,2,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,-1,0)
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - HXZ2(3,1) *Zeta2
     $  - HXZ2(3,1) *HZ2(0,0)
     $  + HXZ2(3,1) *HZ2(0,1)
     $  + HXZ3(3,1,0) *HZ1(0)
     $  + HXZ4(3,1,0,3) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - HZ4( -1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
      HYZ4(1,3,2,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(1)*Zeta2
     $  + HXZ1(3) *HZ3(1,1,0)
     $  + HXZ2(3,1) *HZ2(1,1)
     $  + HXZ3(3,1,0) *HZ1(1)
     $  + HXZ4(3,1,0,0) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0,1,1,0) 
     $  - HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(1,3,2,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(3)*Zeta3
     $  + HXZ1(3) *HZ1(0)*Zeta2
     $  + HXZ1(3) *HZ3(0,1,0)
     $  - HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(0,1)
     $  + HXZ3(3,1,0) *HZ1(0)
     $  + HXZ4(3,1,0,1) 
     $  - HZ1( -1)*Zeta3
     $  + 4.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
      HYZ4(1,3,3,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  - HXZ1(3) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(3)*HZ3(0,0,0)
     $  + HXZ1(3) *HZ3(0,0,1)
     $  - HXZ2(3,1) *Zeta2
     $  - HXZ2(3,1) *HZ2(0,0)
     $  + HXZ2(3,1) *HZ2(0,1)
     $  + HXZ3(3,1,1) *HZ1(1)
     $  + HXZ4(3,1,1,2) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 5.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 8.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
      HYZ4(1,3,3,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(3)*HZ3(-1,-1,0)
     $  + HXZ1(3) *HZ3(-1,0,0)
     $  + HXZ2(3,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(3,1)*HZ2(-1,0)
     $  + HXZ3(3,1,1) *HZ1(0)
     $  + HXZ4(3,1,1,3) 
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
      HYZ4(1,3,3,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(3) *Zeta3
     $  + HXZ1(3) *HZ3(1,0,0)
     $  + HXZ2(3,1) *Zeta2
     $  + HXZ2(3,1) *HZ2(1,0)
     $  + HXZ3(3,1,1) *HZ1(1)
     $  + HXZ4(3,1,1,0) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,1,0,0)
     $  + HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(1,0,0,0)
      HYZ4(1,3,3,3) = 
     $  - Zeta4 
     $  + HXZ1(3) *HZ3(0,0,0)
     $  + HXZ2(3,1) *HZ2(0,0)
     $  + HXZ3(3,1,1) *HZ1(0)
     $  + HXZ4(3,1,1,1) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
      HYZ4(2,0,0,0) = 
     $  - Zeta4 
     $  + HXZ1(0) *HZ3(1,1,1)
     $  + HXZ2(0,2) *HZ2(1,1)
     $  + HXZ3(0,2,2) *HZ1(1)
     $  + HXZ4(0,2,2,2) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  + 4.000000000000000d+00*HZ4(1,1,1,1)
      HYZ4(2,0,0,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(1,0)
     $  + HXZ3(0,2,2) *HZ1(0)
     $  + HXZ4(0,2,2,3) 
     $  + HZ1(1) *Zeta3
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,0,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ2(0,2) *Zeta2
     $  + HXZ3(0,2,2) *HZ1(1)
     $  + HXZ4(0,2,2,0) 
     $  - HZ1(1) *Zeta3
      HYZ4(2,0,0,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(0,0)
     $  + HXZ2(0,2) *HZ2(1,0)
     $  + HXZ3(0,2,2) *HZ1(0)
     $  + HXZ4(0,2,2,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ1(1) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + HZ4(1,1,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,1,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(1,0,1)
     $  - HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(0,1)
     $  + HXZ3(0,2,3) *HZ1(1)
     $  + HXZ4(0,2,3,2) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,1)
     $  + HZ4(1,1,0,1) 
      HYZ4(2,0,1,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ2(0,2) *HZ2(0,0)
     $  + HXZ3(0,2,3) *HZ1(0)
     $  + HXZ4(0,2,3,3) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,1,0,0) 
      HYZ4(2,0,1,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,1,0)
     $  + HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(0,0)
     $  + HXZ2(0,2) *HZ2(1,0)
     $  + HXZ3(0,2,3) *HZ1(1)
     $  + HXZ4(0,2,3,0) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  - HZ4(1,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,1)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,1,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(-1,1,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,0,0)
     $  - HXZ2(0,2) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(0,2)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(0,2)*HZ2(0,0)
     $  + HXZ3(0,2,3) *HZ1(0)
     $  + HXZ4(0,2,3,1) 
     $  - HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,0,1)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
      HYZ4(2,0,2,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ2(0,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,2)*HZ2(1,1)
     $  + HXZ3(0,2,0) *HZ1(1)
     $  + HXZ4(0,2,0,2) 
     $  + 4.000000000000000d+00*HZ1(1)*Zeta3
     $  + 2.000000000000000d+00*HZ2(1,1)*Zeta2
      HYZ4(2,0,2,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(0,1,0)
     $  - HXZ1(0) *HZ3(1,1,0)
     $  - HXZ2(0,2) *Zeta2
     $  - HXZ2(0,2) *HZ2(0,0)
     $  + HXZ2(0,2) *HZ2(0,1)
     $  + HXZ3(0,2,0) *HZ1(0)
     $  + HXZ4(0,2,0,3) 
     $  - 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  - HZ4(1,1,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,2,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ2(0,2) *HZ2(1,1)
     $  + HXZ3(0,2,0) *HZ1(1)
     $  + HXZ4(0,2,0,0) 
     $  - HZ1(1) *Zeta3
      HYZ4(2,0,2,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(1,0,0)
     $  - HXZ1(0) *HZ3(1,1,0)
     $  - HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(0,1)
     $  + HXZ3(0,2,0) *HZ1(0)
     $  + HXZ4(0,2,0,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,0,0)
     $  - HZ4(1,1,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,3,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,0,1)
     $  - HXZ1(0) *HZ3(0,1,0)
     $  - HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  - HXZ2(0,2) *Zeta2
     $  - HXZ2(0,2) *HZ2(0,0)
     $  + HXZ2(0,2) *HZ2(0,1)
     $  + HXZ3(0,2,1) *HZ1(1)
     $  + HXZ4(0,2,1,2) 
     $  - HZ1(1) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,1)
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
      HYZ4(2,0,3,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  - HXZ1(0) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,-1,0)
     $  + HXZ2(0,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,2)*HZ2(-1,0)
     $  + HXZ3(0,2,1) *HZ1(0)
     $  + HXZ4(0,2,1,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ1(1) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(2,0,3,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,1,0)
     $  + HXZ2(0,2) *Zeta2
     $  + HXZ2(0,2) *HZ2(1,0)
     $  + HXZ3(0,2,1) *HZ1(1)
     $  + HXZ4(0,2,1,0) 
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,1) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,1,0,1)
     $  + 3.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(2,0,3,3) = 
     $  - Zeta4 
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ2(0,2) *HZ2(0,0)
     $  + HXZ3(0,2,1) *HZ1(0)
     $  + HXZ4(0,2,1,1) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,1,0,0) 
      HYZ4(2,1,0,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(0,1,1)
     $  + HXZ2(0,3) *HZ2(1,1)
     $  + HXZ3(0,3,2) *HZ1(1)
     $  + HXZ4(0,3,2,2) 
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,0,1,1) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,1)
      HYZ4(2,1,0,1) = 
     $  + 5.000000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(1,0)
     $  + HXZ3(0,3,2) *HZ1(0)
     $  + HXZ4(0,3,2,3) 
     $  - HZ2(0,0) *Zeta2
     $  - HZ4(0,0,1,0) 
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(2,1,0,2) = 
     $  + 5.000000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(1,0,0)
     $  - HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,3) *Zeta2
     $  + HXZ3(0,3,2) *HZ1(1)
     $  + HXZ4(0,3,2,0) 
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,0,0)
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  - HZ4(1,1,0,1) 
      HYZ4(2,1,0,3) = 
     $  - 2.750000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(0,0)
     $  + HXZ2(0,3) *HZ2(1,0)
     $  + HXZ3(0,3,2) *HZ1(0)
     $  + HXZ4(0,3,2,1) 
     $  - HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
     $  - HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(2,1,1,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,1)
     $  - HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(0,1)
     $  + HXZ3(0,3,3) *HZ1(1)
     $  + HXZ4(0,3,3,2) 
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
      HYZ4(2,1,1,1) = 
     $  - Zeta4 
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ2(0,3) *HZ2(0,0)
     $  + HXZ3(0,3,3) *HZ1(0)
     $  + HXZ4(0,3,3,3) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
      HYZ4(2,1,1,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(0,0)
     $  + HXZ2(0,3) *HZ2(1,0)
     $  + HXZ3(0,3,3) *HZ1(1)
     $  + HXZ4(0,3,3,0) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - HZ2(1,0) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - HZ4(0,1,0,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
      HYZ4(2,1,1,3) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(0)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  - HXZ2(0,3) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(0,3)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(0,3)*HZ2(0,0)
     $  + HXZ3(0,3,3) *HZ1(0)
     $  + HXZ4(0,3,3,1) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 5.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 8.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
      HYZ4(2,1,2,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,1)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  - HXZ2(0,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,3)*HZ2(1,1)
     $  + HXZ3(0,3,0) *HZ1(1)
     $  + HXZ4(0,3,0,2) 
     $  - HZ1(1) *Zeta3
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,1)
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
      HYZ4(2,1,2,1) = 
     $  - 2.250000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - HXZ2(0,3) *Zeta2
     $  - HXZ2(0,3) *HZ2(0,0)
     $  + HXZ2(0,3) *HZ2(0,1)
     $  + HXZ3(0,3,0) *HZ1(0)
     $  + HXZ4(0,3,0,3) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
      HYZ4(2,1,2,2) = 
     $  - 5.250000000000000d+00*Zeta4
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,3) *HZ2(1,1)
     $  + HXZ3(0,3,0) *HZ1(1)
     $  + HXZ4(0,3,0,0) 
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
      HYZ4(2,1,2,3) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(0,1)
     $  + HXZ3(0,3,0) *HZ1(0)
     $  + HXZ4(0,3,0,1) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 8.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
      HYZ4(2,1,3,0) = 
     $  + 2.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,1)
     $  - HXZ2(0,3) *Zeta2
     $  - HXZ2(0,3) *HZ2(0,0)
     $  + HXZ2(0,3) *HZ2(0,1)
     $  + HXZ3(0,3,1) *HZ1(1)
     $  + HXZ4(0,3,1,2) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  - 4.000000000000000d+00*HZ4(-1,0,1,1)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,1)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 6.000000000000000d+00*HZ4(0,0,0,0)
     $  - 6.000000000000000d+00*HZ4(0,0,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,1,1)
      HYZ4(2,1,3,1) = 
     $  + 5.000000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + HXZ2(0,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,3)*HZ2(-1,0)
     $  + HXZ3(0,3,1) *HZ1(0)
     $  + HXZ4(0,3,1,3) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(2,1,3,2) = 
     $  + 5.000000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(-1,1,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,0,0)
     $  + HXZ2(0,3) *Zeta2
     $  + HXZ2(0,3) *HZ2(1,0)
     $  + HXZ3(0,3,1) *HZ1(1)
     $  + HXZ4(0,3,1,0) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  - HZ4( -1,0,0,1)
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - HZ4(0,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,0,0)
     $  + HZ4(0,1,0,1) 
     $  + HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,0,1)
      HYZ4(2,1,3,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(0)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ2(0,3) *HZ2(0,0)
     $  + HXZ3(0,3,1) *HZ1(0)
     $  + HXZ4(0,3,1,1) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 6.000000000000000d+00*HZ4(0,0,0,0)
     $  + 3.000000000000000d+00*HZ4(0,0,0,1)
      HYZ4(2,2,0,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(1,1,1)
     $  + HXZ2(0,0) *HZ2(1,1)
     $  + HXZ3(0,0,2) *HZ1(1)
     $  + HXZ4(0,0,2,2) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(1,1)*Zeta2
     $  + 6.000000000000000d+00*HZ4(1,1,1,1)
      HYZ4(2,2,0,1) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(1,0)
     $  + HXZ3(0,0,2) *HZ1(0)
     $  + HXZ4(0,0,2,3) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(1,0,0,0) 
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + HZ4(1,1,1,0) 
      HYZ4(2,2,0,2) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ2(0,0) *Zeta2
     $  + HXZ3(0,0,2) *HZ1(1)
     $  + HXZ4(0,0,2,0) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,1) *Zeta2
      HYZ4(2,2,0,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,1)
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(0,0)
     $  + HXZ2(0,0) *HZ2(1,0)
     $  + HXZ3(0,0,2) *HZ1(0)
     $  + HXZ4(0,0,2,1) 
     $  - HZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(0,0,1,1) 
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
     $  + HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + HZ4(1,1,1,0) 
      HYZ4(2,2,1,0) = 
     $  - Zeta4 
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,1,1)
     $  - HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(0,1)
     $  + HXZ3(0,0,3) *HZ1(1)
     $  + HXZ4(0,0,3,2) 
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,1)
     $  + 3.000000000000000d+00*HZ4(0,1,1,1)
      HYZ4(2,2,1,1) = 
     $  + 1.750000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,0,1)
     $  + HXZ2(0,0) *HZ2(0,0)
     $  + HXZ3(0,0,3) *HZ1(0)
     $  + HXZ4(0,0,3,3) 
     $  + HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + 3.000000000000000d+00*HZ4(0,0,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + HZ4(0,0,1,1) 
      HYZ4(2,2,1,2) = 
     $  + 5.250000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(0)*Zeta2
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,0,1)
     $  - HXZ1(0) *HZ3(0,1,0)
     $  - HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(0,0)
     $  + HXZ2(0,0) *HZ2(1,0)
     $  + HXZ3(0,0,3) *HZ1(1)
     $  + HXZ4(0,0,3,0) 
     $  + 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 3.000000000000000d+00*HZ4(0,0,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,0,1,1) 
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  + HZ4(1,0,0,0) 
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
      HYZ4(2,2,1,3) = 
     $  - Zeta4 
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,1)
     $  - HXZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(0,0)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(0,0)*HZ2(0,0)
     $  + HXZ3(0,0,3) *HZ1(0)
     $  + HXZ4(0,0,3,1) 
     $  + HZ1( -1)*Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ4( -1,0,0,0)
     $  + HZ4( -1,0,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,0,0)
     $  - 3.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
      HYZ4(2,2,2,0) = 
     $  - Zeta4 
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(0)*HZ3(1,1,1)
     $  - HXZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,0)*HZ2(1,1)
     $  + HXZ3(0,0,0) *HZ1(1)
     $  + HXZ4(0,0,0,2) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  + 4.000000000000000d+00*HZ4(1,1,1,1)
      HYZ4(2,2,2,1) = 
     $  - 1.750000000000000d+00*Zeta4
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  - HXZ1(0) *HZ3(0,0,1)
     $  + HXZ1(0) *HZ3(0,1,1)
     $  - HXZ2(0,0) *Zeta2
     $  - HXZ2(0,0) *HZ2(0,0)
     $  + HXZ2(0,0) *HZ2(0,1)
     $  + HXZ3(0,0,0) *HZ1(0)
     $  + HXZ4(0,0,0,3) 
     $  - HZ2(0,0) *Zeta2
     $  - HZ4(0,0,0,0) 
     $  + HZ4(0,0,0,1) 
     $  - HZ4(0,0,1,1) 
     $  + HZ4(0,1,1,1) 
      HYZ4(2,2,2,2) = 
     $  + HXZ1(0) *HZ3(1,1,1)
     $  + HXZ2(0,0) *HZ2(1,1)
     $  + HXZ3(0,0,0) *HZ1(1)
     $  + HXZ4(0,0,0,0) 
     $  + HZ4(1,1,1,1) 
      HYZ4(2,2,2,3) = 
     $  - Zeta4 
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(0,1,1)
     $  - HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(0,1)
     $  + HXZ3(0,0,0) *HZ1(0)
     $  + HXZ4(0,0,0,1) 
     $  + HZ4(0,1,1,1) 
      HYZ4(2,2,3,0) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,1,1)
     $  - HXZ2(0,0) *Zeta2
     $  - HXZ2(0,0) *HZ2(0,0)
     $  + HXZ2(0,0) *HZ2(0,1)
     $  + HXZ3(0,0,1) *HZ1(1)
     $  + HXZ4(0,0,1,2) 
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,0,1,1) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,1)
      HYZ4(2,2,3,1) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,1)
     $  + HXZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,0)*HZ2(-1,0)
     $  + HXZ3(0,0,1) *HZ1(0)
     $  + HXZ4(0,0,1,3) 
     $  - HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + HZ4( -1,0,0,0)
     $  - HZ4( -1,0,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,1)
      HYZ4(2,2,3,2) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ2(0,0) *Zeta2
     $  + HXZ2(0,0) *HZ2(1,0)
     $  + HXZ3(0,0,1) *HZ1(1)
     $  + HXZ4(0,0,1,0) 
     $  - HZ1(1) *Zeta3
     $  + HZ4(1,0,1,1) 
      HYZ4(2,2,3,3) = 
     $  - 2.500000000000000d-01*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,1)
     $  + HXZ2(0,0) *HZ2(0,0)
     $  + HXZ3(0,0,1) *HZ1(0)
     $  + HXZ4(0,0,1,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,0,1,1) 
      HYZ4(2,3,0,0) = 
     $  - Zeta4 
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  - HXZ1(0) *HZ3(0,0,1)
     $  + HXZ1(0) *HZ3(0,1,1)
     $  + HXZ2(0,1) *HZ2(1,1)
     $  + HXZ3(0,1,2) *HZ1(1)
     $  + HXZ4(0,1,2,2) 
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,1)
     $  + 3.000000000000000d+00*HZ4(0,1,1,1)
      HYZ4(2,3,0,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(1,0)
     $  + HXZ3(0,1,2) *HZ1(0)
     $  + HXZ4(0,1,2,3) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(2,3,0,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(1)*Zeta2
     $  - HXZ1(0) *HZ3(0,1,0)
     $  - HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,1) *Zeta2
     $  + HXZ3(0,1,2) *HZ1(1)
     $  + HXZ4(0,1,2,0) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(0,1,0,1) 
     $  - HZ4(1,1,0,1) 
      HYZ4(2,3,0,3) = 
     $  - 3.500000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  + HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(0,0)
     $  + HXZ2(0,1) *HZ2(1,0)
     $  + HXZ3(0,1,2) *HZ1(0)
     $  + HXZ4(0,1,2,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,1) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(2,3,1,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  - HXZ1(0) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,1)
     $  - HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(0,1)
     $  + HXZ3(0,1,3) *HZ1(1)
     $  + HXZ4(0,1,3,2) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  + 4.000000000000000d+00*HZ4(-1,0,1,1)
      HYZ4(2,3,1,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  + HXZ2(0,1) *HZ2(0,0)
     $  + HXZ3(0,1,3) *HZ1(0)
     $  + HXZ4(0,1,3,3) 
     $  + HZ1( -1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  + HZ4( -1,0,0,1)
      HYZ4(2,3,1,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  - HXZ1(0) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(1,-1,0)
     $  + HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(0,0)
     $  + HXZ2(0,1) *HZ2(1,0)
     $  + HXZ3(0,1,3) *HZ1(1)
     $  + HXZ4(0,1,3,0) 
     $  + 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  + HZ4( -1,0,0,1)
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - HZ4(0,1,0,1) 
     $  - HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,1)
      HYZ4(2,3,1,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ1(-1)*Zeta2
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  - HXZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(0,1)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(0,1)*HZ2(0,0)
     $  + HXZ3(0,1,3) *HZ1(0)
     $  + HXZ4(0,1,3,1) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
      HYZ4(2,3,2,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  - HXZ1(0) *HZ3(1,0,0)
     $  + HXZ1(0) *HZ3(1,0,1)
     $  + HXZ1(0) *HZ3(1,1,0)
     $  - HXZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,1)*HZ2(1,1)
     $  + HXZ3(0,1,0) *HZ1(1)
     $  + HXZ4(0,1,0,2) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,1)
     $  + HZ4(1,1,0,1) 
      HYZ4(2,3,2,1) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,-1,0)
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - HXZ2(0,1) *Zeta2
     $  - HXZ2(0,1) *HZ2(0,0)
     $  + HXZ2(0,1) *HZ2(0,1)
     $  + HXZ3(0,1,0) *HZ1(0)
     $  + HXZ4(0,1,0,3) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + HZ4(0,1,0,1) 
      HYZ4(2,3,2,2) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(1)*Zeta2
     $  + HXZ1(0) *HZ3(1,1,0)
     $  + HXZ2(0,1) *HZ2(1,1)
     $  + HXZ3(0,1,0) *HZ1(1)
     $  + HXZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,1) *Zeta2
     $  + HZ4(1,1,0,1) 
      HYZ4(2,3,2,3) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(0)*Zeta3
     $  + HXZ1(0) *HZ1(0)*Zeta2
     $  + HXZ1(0) *HZ3(0,1,0)
     $  - HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(0,1)
     $  + HXZ3(0,1,0) *HZ1(0)
     $  + HXZ4(0,1,0,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0,1,0,1) 
      HYZ4(2,3,3,0) = 
     $  + 1.500000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  - HXZ1(0) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(0)*HZ3(0,0,0)
     $  + HXZ1(0) *HZ3(0,0,1)
     $  - HXZ2(0,1) *Zeta2
     $  - HXZ2(0,1) *HZ2(0,0)
     $  + HXZ2(0,1) *HZ2(0,1)
     $  + HXZ3(0,1,1) *HZ1(1)
     $  + HXZ4(0,1,1,2) 
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + 2.000000000000000d+00*HZ4(0,0,1,1)
      HYZ4(2,3,3,1) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(0)*HZ3(-1,-1,0)
     $  + HXZ1(0) *HZ3(-1,0,0)
     $  + HXZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(0,1)*HZ2(-1,0)
     $  + HXZ3(0,1,1) *HZ1(0)
     $  + HXZ4(0,1,1,3) 
     $  + HZ1( -1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + HZ4( -1,0,0,1)
      HYZ4(2,3,3,2) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(0) *Zeta3
     $  + HXZ1(0) *HZ3(1,0,0)
     $  + HXZ2(0,1) *Zeta2
     $  + HXZ2(0,1) *HZ2(1,0)
     $  + HXZ3(0,1,1) *HZ1(1)
     $  + HXZ4(0,1,1,0) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,0) *Zeta2
     $  + HZ4(1,0,0,1) 
      HYZ4(2,3,3,3) = 
     $  - Zeta4 
     $  + HXZ1(0) *HZ3(0,0,0)
     $  + HXZ2(0,1) *HZ2(0,0)
     $  + HXZ3(0,1,1) *HZ1(0)
     $  + HXZ4(0,1,1,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,1) 
      HYZ4(3,0,0,0) = 
     $  - 1.750000000000000d+00*Zeta4
     $  + HXZ1(1) *HZ3(1,1,1)
     $  + HXZ2(1,2) *HZ2(1,1)
     $  + HXZ3(1,2,2) *HZ1(1)
     $  + HXZ4(1,2,2,2) 
     $  - HZ2(0,0) *Zeta2
     $  - HZ4(0,0,0,0) 
     $  + HZ4(0,0,0,1) 
     $  - HZ4(0,0,1,1) 
     $  + HZ4(0,1,1,1) 
      HYZ4(3,0,0,1) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(1,0)
     $  + HXZ3(1,2,2) *HZ1(0)
     $  + HXZ4(1,2,2,3) 
     $  + HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  - HZ4(0,0,1,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(3,0,0,2) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + HXZ2(1,2) *Zeta2
     $  + HXZ3(1,2,2) *HZ1(1)
     $  + HXZ4(1,2,2,0) 
     $  + HZ1(1) *Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,1,0) 
     $  + HZ4(1,0,1,0) 
     $  + HZ4(1,1,1,0) 
      HYZ4(3,0,0,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(0,0)
     $  + HXZ2(1,2) *HZ2(1,0)
     $  + HXZ3(1,2,2) *HZ1(0)
     $  + HXZ4(1,2,2,1) 
     $  + HZ2(0,0) *Zeta2
     $  + HZ2(0,1) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
     $  + HZ4(0,1,0,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(3,0,1,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(1,0,1)
     $  - HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(0,1)
     $  + HXZ3(1,2,3) *HZ1(1)
     $  + HXZ4(1,2,3,2) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + HZ4(0,1,0,1) 
      HYZ4(3,0,1,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ2(1,2) *HZ2(0,0)
     $  + HXZ3(1,2,3) *HZ1(0)
     $  + HXZ4(1,2,3,3) 
     $  - HZ2(0, -1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - HZ4(0, -1,0,0)
     $  + HZ4(0,1,0,0) 
      HYZ4(3,0,1,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,1,0)
     $  + HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(0,0)
     $  + HXZ2(1,2) *HZ2(1,0)
     $  + HXZ3(1,2,3) *HZ1(1)
     $  + HXZ4(1,2,3,0) 
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + HZ4(0,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(3,0,1,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(-1,1,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,0,0)
     $  - HXZ2(1,2) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(1,2)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(1,2)*HZ2(0,0)
     $  + HXZ3(1,2,3) *HZ1(0)
     $  + HXZ4(1,2,3,1) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - HZ4( -1,0,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,0,0)
      HYZ4(3,0,2,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ2(1,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,2)*HZ2(1,1)
     $  + HXZ3(1,2,0) *HZ1(1)
     $  + HXZ4(1,2,0,2) 
     $  - 3.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,1) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,0,1,0) 
     $  + HZ4(1,1,0,0) 
     $  - HZ4(1,1,0,1) 
     $  - 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(3,0,2,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(0,1,0)
     $  - HXZ1(1) *HZ3(1,1,0)
     $  - HXZ2(1,2) *Zeta2
     $  - HXZ2(1,2) *HZ2(0,0)
     $  + HXZ2(1,2) *HZ2(0,1)
     $  + HXZ3(1,2,0) *HZ1(0)
     $  + HXZ4(1,2,0,3) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  + 4.000000000000000d+00*HZ4(1,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(3,0,2,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ2(1,2) *HZ2(1,1)
     $  + HXZ3(1,2,0) *HZ1(1)
     $  + HXZ4(1,2,0,0) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(3,0,2,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(1,0,0)
     $  - HXZ1(1) *HZ3(1,1,0)
     $  - HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(0,1)
     $  + HXZ3(1,2,0) *HZ1(0)
     $  + HXZ4(1,2,0,1) 
     $  - HZ1(0) *Zeta3
     $  - 4.000000000000000d+00*HZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(0,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,1,0)
     $  - HZ4(1,0,0,0) 
     $  - 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(3,0,3,0) = 
     $  - 2.250000000000000d+00*Zeta4
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,0,1)
     $  - HXZ1(1) *HZ3(0,1,0)
     $  - HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  - HXZ2(1,2) *Zeta2
     $  - HXZ2(1,2) *HZ2(0,0)
     $  + HXZ2(1,2) *HZ2(0,1)
     $  + HXZ3(1,2,1) *HZ1(1)
     $  + HXZ4(1,2,1,2) 
     $  - 2.000000000000000d+00*HZ1(0)*Zeta3
     $  - 3.000000000000000d+00*HZ2(0,0)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
      HYZ4(3,0,3,1) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  - HXZ1(1) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,-1,0)
     $  + HXZ2(1,2) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,2)*HZ2(-1,0)
     $  + HXZ3(1,2,1) *HZ1(0)
     $  + HXZ4(1,2,1,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + HZ4( -1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
      HYZ4(3,0,3,2) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,1,0)
     $  + HXZ2(1,2) *Zeta2
     $  + HXZ2(1,2) *HZ2(1,0)
     $  + HXZ3(1,2,1) *HZ1(1)
     $  + HXZ4(1,2,1,0) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + 3.000000000000000d+00*HZ2(0,1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + HZ4(0,1,0,0) 
     $  + 3.000000000000000d+00*HZ4(0,1,1,0)
     $  + HZ4(1,0,0,0) 
     $  + HZ4(1,0,1,0) 
      HYZ4(3,0,3,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ2(1,2) *HZ2(0,0)
     $  + HXZ3(1,2,1) *HZ1(0)
     $  + HXZ4(1,2,1,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,1,0,0) 
      HYZ4(3,1,0,0) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(0,1,1)
     $  + HXZ2(1,3) *HZ2(1,1)
     $  + HXZ3(1,3,2) *HZ1(1)
     $  + HXZ4(1,3,2,2) 
     $  - HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + HZ4( -1,0,0,0)
     $  - HZ4( -1,0,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,1)
      HYZ4(3,1,0,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(1,0)
     $  + HXZ3(1,3,2) *HZ1(0)
     $  + HXZ4(1,3,2,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
      HYZ4(3,1,0,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(1,0,0)
     $  - HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,3) *Zeta2
     $  + HXZ3(1,3,2) *HZ1(1)
     $  + HXZ4(1,3,2,0) 
     $  + HZ1( -1)*Zeta3
     $  + HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  - HZ2(1,1) *Zeta2
     $  - HZ4( -1,0,1,0)
     $  - HZ4( -1,1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(1, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(3,1,0,3) = 
     $  + 5.500000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(0,0)
     $  + HXZ2(1,3) *HZ2(1,0)
     $  + HXZ3(1,3,2) *HZ1(0)
     $  + HXZ4(1,3,2,1) 
     $  + HZ1( -1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 4.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(3,1,1,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,1)
     $  - HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(0,1)
     $  + HXZ3(1,3,3) *HZ1(1)
     $  + HXZ4(1,3,3,2) 
     $  + HZ1( -1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + HZ4( -1,0,0,1)
      HYZ4(3,1,1,1) = 
     $  + Zeta4 
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ2(1,3) *HZ2(0,0)
     $  + HXZ3(1,3,3) *HZ1(0)
     $  + HXZ4(1,3,3,3) 
     $  - HZ1( -1)*Zeta3
     $  + HZ2( -1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + HZ4( -1,-1,0,0)
     $  + HZ4( -1,0,0,0)
      HYZ4(3,1,1,2) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(0,0)
     $  + HXZ2(1,3) *HZ2(1,0)
     $  + HXZ3(1,3,3) *HZ1(1)
     $  + HXZ4(1,3,3,0) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  - HZ2(0, -1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + HZ4( -1,0,0,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + HZ4(0,0,1,0) 
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + HZ4(1, -1,0,0)
      HYZ4(3,1,1,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(1)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  - HXZ2(1,3) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(1,3)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(1,3)*HZ2(0,0)
     $  + HXZ3(1,3,3) *HZ1(0)
     $  + HXZ4(1,3,3,1) 
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
      HYZ4(3,1,2,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,1)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  - HXZ2(1,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,3)*HZ2(1,1)
     $  + HXZ3(1,3,0) *HZ1(1)
     $  + HXZ4(1,3,0,2) 
     $  + HZ1( -1)*Zeta3
     $  + HZ1(1) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  - HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  + HZ4(0,1,0,0) 
     $  - HZ4(0,1,0,1) 
     $  - HZ4(0,1,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(3,1,2,1) = 
     $  + Zeta4 
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - HXZ2(1,3) *Zeta2
     $  - HXZ2(1,3) *HZ2(0,0)
     $  + HXZ2(1,3) *HZ2(0,1)
     $  + HXZ3(1,3,0) *HZ1(0)
     $  + HXZ4(1,3,0,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + HZ4( -1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
      HYZ4(3,1,2,2) = 
     $  + Zeta4 
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,3) *HZ2(1,1)
     $  + HXZ3(1,3,0) *HZ1(1)
     $  + HXZ4(1,3,0,0) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,1) *Zeta2
     $  + HZ4( -1,0,0,0)
     $  + HZ4( -1,0,1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,1,0)
     $  + HZ4(0, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  - HZ4(0,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - HZ4(0,1,1,0) 
     $  + HZ4(1, -1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  - HZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ4(1,1,-1,0)
      HYZ4(3,1,2,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(0,1)
     $  + HXZ3(1,3,0) *HZ1(0)
     $  + HXZ4(1,3,0,1) 
     $  + HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0,1) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  + 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,0,1,0)
     $  + 2.000000000000000d+00*HZ4(0,1,-1,0)
      HYZ4(3,1,3,0) = 
     $  + 5.000000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,1)
     $  - HXZ2(1,3) *Zeta2
     $  - HXZ2(1,3) *HZ2(0,0)
     $  + HXZ2(1,3) *HZ2(0,1)
     $  + HXZ3(1,3,1) *HZ1(1)
     $  + HXZ4(1,3,1,2) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,0,1)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,0,1)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 3.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(3,1,3,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + HXZ2(1,3) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,3)*HZ2(-1,0)
     $  + HXZ3(1,3,1) *HZ1(0)
     $  + HXZ4(1,3,1,3) 
     $  + 4.000000000000000d+00*HZ1(-1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 8.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + 4.000000000000000d+00*HZ4(-1,-1,0,0)
     $  + 4.000000000000000d+00*HZ4(-1,0,-1,0)
      HYZ4(3,1,3,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(-1,1,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,0,0)
     $  + HXZ2(1,3) *Zeta2
     $  + HXZ2(1,3) *HZ2(1,0)
     $  + HXZ3(1,3,1) *HZ1(1)
     $  + HXZ4(1,3,1,0) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(1,-1)*Zeta2
     $  + HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,-1,1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  - 4.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,0,0)
     $  - 4.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(1,0,-1,0)
      HYZ4(3,1,3,3) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  - 3.000000000000000d+00*HXZ1(1)*HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ2(1,3) *HZ2(0,0)
     $  + HXZ3(1,3,1) *HZ1(0)
     $  + HXZ4(1,3,1,1) 
     $  - 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 3.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - 2.000000000000000d+00*HZ2(-1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ2(0,0) *Zeta2
     $  + 6.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  - 5.000000000000000d+00*HZ4(-1,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 3.000000000000000d+00*HZ4(-1,0,0,0)
     $  - 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
      HYZ4(3,2,0,0) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(1,1,1)
     $  + HXZ2(1,0) *HZ2(1,1)
     $  + HXZ3(1,0,2) *HZ1(1)
     $  + HXZ4(1,0,2,2) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(1,0,0,0) 
     $  - HZ4(1,0,0,1) 
     $  + HZ4(1,0,1,1) 
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + HZ4(1,1,1,0) 
      HYZ4(3,2,0,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(1,0)
     $  + HXZ3(1,0,2) *HZ1(0)
     $  + HXZ4(1,0,2,3) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - 4.000000000000000d+00*HZ4(1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(3,2,0,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ2(1,0) *Zeta2
     $  + HXZ3(1,0,2) *HZ1(1)
     $  + HXZ4(1,0,2,0) 
     $  - HZ1(1) *Zeta3
     $  - HZ2(1,1) *Zeta2
     $  - HZ4(1,0,1,0) 
     $  - HZ4(1,1,1,0) 
      HYZ4(3,2,0,3) = 
     $  + 5.500000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,1)
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(0,0)
     $  + HXZ2(1,0) *HZ2(1,0)
     $  + HXZ3(1,0,2) *HZ1(0)
     $  + HXZ4(1,0,2,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + 4.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + 2.000000000000000d+00*HZ2(1,0)*Zeta2
     $  + HZ4(0,0,1,0) 
     $  + HZ4(1,0,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,0,1,0)
      HYZ4(3,2,1,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,1,1)
     $  - HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(0,1)
     $  + HXZ3(1,0,3) *HZ1(1)
     $  + HXZ4(1,0,3,2) 
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,1)
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(3,2,1,1) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,0,1)
     $  + HXZ2(1,0) *HZ2(0,0)
     $  + HXZ3(1,0,3) *HZ1(0)
     $  + HXZ4(1,0,3,3) 
     $  - HZ2(0, -1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + HZ4(0,0,1,0) 
      HYZ4(3,2,1,2) = 
     $  + Zeta4 
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(0)*Zeta2
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,0,1)
     $  - HXZ1(1) *HZ3(0,1,0)
     $  - HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(0,0)
     $  + HXZ2(1,0) *HZ2(1,0)
     $  + HXZ3(1,0,3) *HZ1(1)
     $  + HXZ4(1,0,3,0) 
     $  - HZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,1,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
     $  - 2.000000000000000d+00*HZ4(0,1,-1,0)
     $  - 2.000000000000000d+00*HZ4(1,0,-1,0)
     $  + HZ4(1,0,1,0) 
      HYZ4(3,2,1,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,1)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  - 3.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,1)
     $  - HXZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(1,0)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(1,0)*HZ2(0,0)
     $  + HXZ3(1,0,3) *HZ1(0)
     $  + HXZ4(1,0,3,1) 
     $  - HZ1( -1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - HZ2( -1,0)*Zeta2
     $  + 2.000000000000000d+00*HZ2(0,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - 2.000000000000000d+00*HZ4(-1,0,1,0)
     $  + 4.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - 2.000000000000000d+00*HZ4(0,-1,0,0)
     $  - 4.000000000000000d+00*HZ4(0,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,0,1,0)
      HYZ4(3,2,2,0) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  + 3.000000000000000d+00*HXZ1(1)*HZ3(1,1,1)
     $  - HXZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,0)*HZ2(1,1)
     $  + HXZ3(1,0,0) *HZ1(1)
     $  + HXZ4(1,0,0,2) 
     $  + HZ1(1) *Zeta3
     $  + HZ2(1,1) *Zeta2
     $  - HZ4(1,1,0,0) 
     $  + HZ4(1,1,0,1) 
     $  + 2.000000000000000d+00*HZ4(1,1,1,0)
      HYZ4(3,2,2,1) = 
     $  + Zeta4 
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  - HXZ1(1) *HZ3(0,0,1)
     $  + HXZ1(1) *HZ3(0,1,1)
     $  - HXZ2(1,0) *Zeta2
     $  - HXZ2(1,0) *HZ2(0,0)
     $  + HXZ2(1,0) *HZ2(0,1)
     $  + HXZ3(1,0,0) *HZ1(0)
     $  + HXZ4(1,0,0,3) 
     $  + HZ2(0,1) *Zeta2
     $  + 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  - HZ4(0,0,1,0) 
     $  + HZ4(0,1,1,0) 
      HYZ4(3,2,2,2) = 
     $  + Zeta4 
     $  + HXZ1(1) *HZ3(1,1,1)
     $  + HXZ2(1,0) *HZ2(1,1)
     $  + HXZ3(1,0,0) *HZ1(1)
     $  + HXZ4(1,0,0,0) 
     $  - HZ1(1) *Zeta3
     $  + HZ2(1,1) *Zeta2
     $  + HZ4(1,1,1,0) 
      HYZ4(3,2,2,3) = 
     $  - 1.250000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(0,1,1)
     $  - HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(0,1)
     $  + HXZ3(1,0,0) *HZ1(0)
     $  + HXZ4(1,0,0,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ2(0,1) *Zeta2
     $  + HZ4(0,1,1,0) 
      HYZ4(3,2,3,0) = 
     $  + 5.000000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ3(0,0,1)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,1,1)
     $  - HXZ2(1,0) *Zeta2
     $  - HXZ2(1,0) *HZ2(0,0)
     $  + HXZ2(1,0) *HZ2(0,1)
     $  + HXZ3(1,0,1) *HZ1(1)
     $  + HXZ4(1,0,1,2) 
     $  - HZ2(0,0) *Zeta2
     $  - HZ4(0,0,1,0) 
     $  - HZ4(0,1,0,0) 
     $  + HZ4(0,1,0,1) 
     $  + HZ4(0,1,1,0) 
      HYZ4(3,2,3,1) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,1)
     $  + HXZ2(1,0) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,0)*HZ2(-1,0)
     $  + HXZ3(1,0,1) *HZ1(0)
     $  + HXZ4(1,0,1,3) 
     $  + HZ1( -1)*Zeta3
     $  + HZ2( -1,0)*Zeta2
     $  - 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,1,0)
      HYZ4(3,2,3,2) = 
     $  + 7.500000000000000d-01*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ2(1,0) *Zeta2
     $  + HXZ2(1,0) *HZ2(1,0)
     $  + HXZ3(1,0,1) *HZ1(1)
     $  + HXZ4(1,0,1,0) 
     $  + 2.000000000000000d+00*HZ1(1)*Zeta3
     $  + HZ2(1,0) *Zeta2
     $  + HZ4(1,0,1,0) 
      HYZ4(3,2,3,3) = 
     $  + 3.000000000000000d+00*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,1)
     $  + HXZ2(1,0) *HZ2(0,0)
     $  + HXZ3(1,0,1) *HZ1(0)
     $  + HXZ4(1,0,1,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ4(0,0,1,0) 
      HYZ4(3,3,0,0) = 
     $  + 1.750000000000000d+00*Zeta4
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  - HXZ1(1) *HZ3(0,0,1)
     $  + HXZ1(1) *HZ3(0,1,1)
     $  + HXZ2(1,1) *HZ2(1,1)
     $  + HXZ3(1,1,2) *HZ1(1)
     $  + HXZ4(1,1,2,2) 
     $  + HZ1(0) *Zeta3
     $  + 2.000000000000000d+00*HZ2(0,0)*Zeta2
     $  + 3.000000000000000d+00*HZ4(0,0,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,0,1)
     $  + HZ4(0,0,1,1) 
      HYZ4(3,3,0,1) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(1,0)
     $  + HXZ3(1,1,2) *HZ1(0)
     $  + HXZ4(1,1,2,3) 
     $  - HZ2(0, -1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - HZ4(0, -1,0,0)
     $  - 2.000000000000000d+00*HZ4(0,0,-1,0)
     $  + HZ4(0,0,1,0) 
      HYZ4(3,3,0,2) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(1)*Zeta2
     $  - HXZ1(1) *HZ3(0,1,0)
     $  - HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,1) *Zeta2
     $  + HXZ3(1,1,2) *HZ1(1)
     $  + HXZ4(1,1,2,0) 
     $  + HZ1(1) *Zeta3
     $  - HZ2(0,1) *Zeta2
     $  - HZ4(0,0,1,0) 
     $  - HZ4(0,1,0,0) 
     $  - HZ4(0,1,1,0) 
     $  - HZ4(1,1,0,0) 
      HYZ4(3,3,0,3) = 
     $  + 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  + HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(0,0)
     $  + HXZ2(1,1) *HZ2(1,0)
     $  + HXZ3(1,1,2) *HZ1(0)
     $  + HXZ4(1,1,2,1) 
     $  + 2.000000000000000d+00*HZ1(0)*Zeta3
     $  + HZ2(0,0) *Zeta2
     $  + HZ4(0,0,0,0) 
     $  + HZ4(0,0,1,0) 
      HYZ4(3,3,1,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  - HXZ1(1) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,1)
     $  - HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(0,1)
     $  + HXZ3(1,1,3) *HZ1(1)
     $  + HXZ4(1,1,3,2) 
     $  + HZ1( -1)*Zeta3
     $  - 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  - HZ2( -1,0)*Zeta2
     $  - HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,1)
     $  - 2.000000000000000d+00*HZ4(-1,0,0,0)
     $  + HZ4( -1,0,0,1)
      HYZ4(3,3,1,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  + HXZ2(1,1) *HZ2(0,0)
     $  + HXZ3(1,1,3) *HZ1(0)
     $  + HXZ4(1,1,3,3) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + 4.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,0,0)
      HYZ4(3,3,1,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  - HXZ1(1) *HZ3(0,1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(1,-1,0)
     $  + HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(0,0)
     $  + HXZ2(1,1) *HZ2(1,0)
     $  + HXZ3(1,1,3) *HZ1(1)
     $  + HXZ4(1,1,3,0) 
     $  - 2.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(1) *Zeta3
     $  + 2.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,1)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  + HZ2(1, -1)*Zeta2
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,-1,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  - HZ4( -1,0,1,0)
     $  + 2.000000000000000d+00*HZ4(-1,1,-1,0)
     $  + HZ4( -1,1,0,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
     $  - HZ4(0,1,0,0) 
     $  + 2.000000000000000d+00*HZ4(1,-1,-1,0)
     $  + HZ4(1, -1,0,0)
      HYZ4(3,3,1,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ1(-1)*Zeta2
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  - 4.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,0,0)
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  - HXZ2(1,1) *Zeta2
     $  - 2.000000000000000d+00*HXZ2(1,1)*HZ2(-1,0)
     $  + 2.000000000000000d+00*HXZ2(1,1)*HZ2(0,0)
     $  + HXZ3(1,1,3) *HZ1(0)
     $  + HXZ4(1,1,3,1) 
     $  + 3.000000000000000d+00*HZ1(-1)*Zeta3
     $  - HZ1(0) *Zeta3
     $  - 3.000000000000000d+00*HZ2(-1,-1)*Zeta2
     $  + HZ2( -1,0)*Zeta2
     $  + HZ2(0, -1)*Zeta2
     $  - 6.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + HZ4( -1,-1,0,0)
     $  + 2.000000000000000d+00*HZ4(-1,0,-1,0)
     $  + 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  + HZ4(0, -1,0,0)
      HYZ4(3,3,2,0) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  - HXZ1(1) *HZ3(1,0,0)
     $  + HXZ1(1) *HZ3(1,0,1)
     $  + HXZ1(1) *HZ3(1,1,0)
     $  - HXZ2(1,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,1)*HZ2(1,1)
     $  + HXZ3(1,1,0) *HZ1(1)
     $  + HXZ4(1,1,0,2) 
     $  - 2.000000000000000d+00*HZ1(1)*Zeta3
     $  - HZ2(1,0) *Zeta2
     $  - 2.000000000000000d+00*HZ4(1,0,0,0)
     $  + HZ4(1,0,0,1) 
     $  + HZ4(1,1,0,0) 
      HYZ4(3,3,2,1) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,-1,0)
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - HXZ2(1,1) *Zeta2
     $  - HXZ2(1,1) *HZ2(0,0)
     $  + HXZ2(1,1) *HZ2(0,1)
     $  + HXZ3(1,1,0) *HZ1(0)
     $  + HXZ4(1,1,0,3) 
     $  - HZ2(0, -1)*Zeta2
     $  - 2.000000000000000d+00*HZ4(0,-1,-1,0)
     $  - HZ4(0, -1,0,0)
     $  + HZ4(0,1,0,0) 
      HYZ4(3,3,2,2) = 
     $  + 2.500000000000000d-01*Zeta4
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(1)*Zeta2
     $  + HXZ1(1) *HZ3(1,1,0)
     $  + HXZ2(1,1) *HZ2(1,1)
     $  + HXZ3(1,1,0) *HZ1(1)
     $  + HXZ4(1,1,0,0) 
     $  - HZ1(1) *Zeta3
     $  + HZ4(1,1,0,0) 
      HYZ4(3,3,2,3) = 
     $  - 3.000000000000000d+00*Zeta4
     $  + 2.000000000000000d+00*HXZ1(1)*Zeta3
     $  + HXZ1(1) *HZ1(0)*Zeta2
     $  + HXZ1(1) *HZ3(0,1,0)
     $  - HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(0,1)
     $  + HXZ3(1,1,0) *HZ1(0)
     $  + HXZ4(1,1,0,1) 
     $  - HZ1(0) *Zeta3
     $  + HZ4(0,1,0,0) 
      HYZ4(3,3,3,0) = 
     $  - Zeta4 
     $  - HXZ1(1) *Zeta3
     $  - HXZ1(1) *HZ1(0)*Zeta2
     $  - 2.000000000000000d+00*HXZ1(1)*HZ3(0,0,0)
     $  + HXZ1(1) *HZ3(0,0,1)
     $  - HXZ2(1,1) *Zeta2
     $  - HXZ2(1,1) *HZ2(0,0)
     $  + HXZ2(1,1) *HZ2(0,1)
     $  + HXZ3(1,1,1) *HZ1(1)
     $  + HXZ4(1,1,1,2) 
     $  - HZ1(0) *Zeta3
     $  - HZ2(0,0) *Zeta2
     $  - 3.000000000000000d+00*HZ4(0,0,0,0)
     $  + HZ4(0,0,0,1) 
      HYZ4(3,3,3,1) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ1(-1)*Zeta2
     $  + 2.000000000000000d+00*HXZ1(1)*HZ3(-1,-1,0)
     $  + HXZ1(1) *HZ3(-1,0,0)
     $  + HXZ2(1,1) *Zeta2
     $  + 2.000000000000000d+00*HXZ2(1,1)*HZ2(-1,0)
     $  + HXZ3(1,1,1) *HZ1(0)
     $  + HXZ4(1,1,1,3) 
     $  - HZ1( -1)*Zeta3
     $  + HZ2( -1,-1)*Zeta2
     $  + 2.000000000000000d+00*HZ4(-1,-1,-1,0)
     $  + HZ4( -1,-1,0,0)
     $  + HZ4( -1,0,0,0)
      HYZ4(3,3,3,2) = 
     $  + Zeta4 
     $  - HXZ1(1) *Zeta3
     $  + HXZ1(1) *HZ3(1,0,0)
     $  + HXZ2(1,1) *Zeta2
     $  + HXZ2(1,1) *HZ2(1,0)
     $  + HXZ3(1,1,1) *HZ1(1)
     $  + HXZ4(1,1,1,0) 
     $  + HZ4(1,0,0,0) 
      HYZ4(3,3,3,3) = 
     $  + HXZ1(1) *HZ3(0,0,0)
     $  + HXZ2(1,1) *HZ2(0,0)
     $  + HXZ3(1,1,1) *HZ1(0)
     $  + HXZ4(1,1,1,1) 
     $  + HZ4(0,0,0,0) 
      return
      end


      function dlogone(x)
*********************************************************************
***** evaluates log(1+x) using the expansion                    *****
***** of the logarithm around 1                                 *****
***** avoids rounding errors from dlog(1d0+x)                   *****
*********************************************************************
** 1+x = (1+ep)/(1-ep), ep = x/(2+x) 
** log(1+x) = log((1+x)/(1-x)) = 2*ep*(1+ep^2/3+ep^4/5+.....) 
** at x= -1/2, ep = -1/3 
** ep2 = 1/9, ep2^16 = 5.4 x 10^(-16) 
      implicit none
      include 'src/Inc/types.f'
      real(dp):: dlogone
      real(dp):: x,ep,e2
      ep = x/(2.d0+x) 
      e2 = ep*ep 
      dlogone =2*ep*(1+e2*(1.d0/3+e2*(1.d0/5+e2*(1.d0/ 7+e2*(1.d0/ 9 
     $              +e2*(1.d0/11+e2*(1.d0/13+e2*(1.d0/15+e2*(1.d0/17    
     $              +e2*(1.d0/19+e2*(1.d0/21+e2*(1.d0/23+e2*(1.d0/25    
     $              +e2*(1.d0/27+e2*(1.d0/29+e2*(1.d0/31+e2*(1.d0/33    
     $              )))))))))))))))))
      return
      end

      subroutine makeexpar(y,z)
*********************************************************************
*****  makeexpar(y,z) creates the expansion parameters          *****
*****  for the evaluation of the coefficients                   *****
*****  tz#n:   n-th Chebyshev polynomial in 2*z-1               *****
*****      u =  4/3*H(1;z)                                      *****
*****      v =  4/3*H(-1;z)                                     *****
*****  tu#n:   n-th Chebyshev polynomial in 2*u-1               *****
*****  tv#n:   n-th Chebyshev polynomial in 2*v-1               *****
*****  zm#n:   (1-z)^n                                          ***** 
*****      r =  4/3*H(1;y)                                      ***** 
*****      s =  4/3*H(2;y)                                      ***** 
*****  s#n:    s^n                                              ***** 
*****  r#n:    s^n                                              ***** 
*****  b#n:    (s/z)^n                                          ***** 
*********************************************************************
      implicit none
      include 'src/Inc/types.f'
      real(dp):: dlogone
      real(dp):: y,z
      real(dp):: u,v
      real(dp):: 
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
      real(dp):: 
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
      real(dp)::
     $     tv01,tv02,tv03,tv04,tv05,tv06,tv07,tv08,tv09,tv10,
     $     tv11,tv12,tv13,tv14,tv15,tv16,tv17,tv18,tv19,tv20,tv21
      real(dp):: 
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      real(dp)::
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22
      real(dp)::
     $     r01,r02,r03,r04,r05,r06,r07,r08,r09,r10,
     $     r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,r21,r22 
      real(dp)::
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
      
      common /tpz/  
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
      common /tpv/  
     $     tv01,tv02,tv03,tv04,tv05,tv06,tv07,tv08,tv09,tv10,
     $     tv11,tv12,tv13,tv14,tv15,tv16,tv17,tv18,tv19,tv20,tv21
      common /tpu/  
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
      common /zm/  
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21    
      common /s/
     $     s01,s02,s03,s04,s05,s06,s07,s08,s09,s10,
     $     s11,s12,s13,s14,s15,s16,s17,s18,s19,s20,s21,s22      
      common /rtdhpl/
     $     r01,r02,r03,r04,r05,r06,r07,r08,r09,r10,
     $     r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,r21,r22 
      common /b/
     $     b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,
     $     b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22
!$omp threadprivate(/s/,/rtdhpl/,/b/,/tpz/,/tpv/,/tpu/,/zm/)

      zm01 = 1d0-z
      zm02 = zm01*zm01 
      zm03 = zm01*zm02 
      zm04 = zm01*zm03 
      zm05 = zm01*zm04 
      zm06 = zm01*zm05 
      zm07 = zm01*zm06 
      zm08 = zm01*zm07 
      zm09 = zm01*zm08 
      zm10 = zm01*zm09 
      zm11 = zm01*zm10 
      zm12 = zm01*zm11 
      zm13 = zm01*zm12 
      zm14 = zm01*zm13 
      zm15 = zm01*zm14 
      zm16 = zm01*zm15 
      zm17 = zm01*zm16 
      zm18 = zm01*zm17 
      zm19 = zm01*zm18 
      zm20 = zm01*zm19 
      zm21 = zm01*zm20 

      tz01 = 2d0*z-1d0
      tz02 = 2d0*tz01*tz01 - 1d0
      tz03 = 2d0*tz01*tz02 - tz01
      tz04 = 2d0*tz01*tz03 - tz02
      tz05 = 2d0*tz01*tz04 - tz03
      tz06 = 2d0*tz01*tz05 - tz04
      tz07 = 2d0*tz01*tz06 - tz05
      tz08 = 2d0*tz01*tz07 - tz06
      tz09 = 2d0*tz01*tz08 - tz07
      tz10 = 2d0*tz01*tz09 - tz08
      tz11 = 2d0*tz01*tz10 - tz09
      tz12 = 2d0*tz01*tz11 - tz10
      tz13 = 2d0*tz01*tz12 - tz11
      tz14 = 2d0*tz01*tz13 - tz12
      tz15 = 2d0*tz01*tz14 - tz13
      tz16 = 2d0*tz01*tz15 - tz14
      tz17 = 2d0*tz01*tz16 - tz15
      tz18 = 2d0*tz01*tz17 - tz16
      tz19 = 2d0*tz01*tz18 - tz17
      tz20 = 2d0*tz01*tz19 - tz18
      tz21 = 2d0*tz01*tz20 - tz19

      u = -4d0/3d0*dlogone(-z)
      v =  4d0/3d0*dlogone(z)
      
      tu01 = 2d0*u-1d0 
      tu02 = 2d0*tu01*tu01 - 1d0
      tu03 = 2d0*tu01*tu02 - tu01
      tu04 = 2d0*tu01*tu03 - tu02
      tu05 = 2d0*tu01*tu04 - tu03
      tu06 = 2d0*tu01*tu05 - tu04
      tu07 = 2d0*tu01*tu06 - tu05
      tu08 = 2d0*tu01*tu07 - tu06
      tu09 = 2d0*tu01*tu08 - tu07
      tu10 = 2d0*tu01*tu09 - tu08
      tu11 = 2d0*tu01*tu10 - tu09
      tu12 = 2d0*tu01*tu11 - tu10
      tu13 = 2d0*tu01*tu12 - tu11
      tu14 = 2d0*tu01*tu13 - tu12
      tu15 = 2d0*tu01*tu14 - tu13
      tu16 = 2d0*tu01*tu15 - tu14
      tu17 = 2d0*tu01*tu16 - tu15
      tu18 = 2d0*tu01*tu17 - tu16
      tu19 = 2d0*tu01*tu18 - tu17
      tu20 = 2d0*tu01*tu19 - tu18
      tu21 = 2d0*tu01*tu20 - tu19

      tv01 = 2d0*v-1d0
      tv02 = 2d0*tv01*tv01 - 1d0
      tv03 = 2d0*tv01*tv02 - tv01
      tv04 = 2d0*tv01*tv03 - tv02
      tv05 = 2d0*tv01*tv04 - tv03
      tv06 = 2d0*tv01*tv05 - tv04
      tv07 = 2d0*tv01*tv06 - tv05
      tv08 = 2d0*tv01*tv07 - tv06
      tv09 = 2d0*tv01*tv08 - tv07
      tv10 = 2d0*tv01*tv09 - tv08
      tv11 = 2d0*tv01*tv10 - tv09
      tv12 = 2d0*tv01*tv11 - tv10
      tv13 = 2d0*tv01*tv12 - tv11
      tv14 = 2d0*tv01*tv13 - tv12
      tv15 = 2d0*tv01*tv14 - tv13
      tv16 = 2d0*tv01*tv15 - tv14
      tv17 = 2d0*tv01*tv16 - tv15
      tv18 = 2d0*tv01*tv17 - tv16
      tv19 = 2d0*tv01*tv18 - tv17
      tv20 = 2d0*tv01*tv19 - tv18
      tv21 = 2d0*tv01*tv20 - tv19

      s01 = -4d0/3d0*dlogone(-y/(1d0-z))
      b01 = s01/z
      r01 = -4d0/3d0*dlogone(-y)

      s02 = s01*s01 
      s03 = s01*s02 
      s04 = s01*s03 
      s05 = s01*s04 
      s06 = s01*s05 
      s07 = s01*s06 
      s08 = s01*s07 
      s09 = s01*s08 
      s10 = s01*s09 
      s11 = s01*s10 
      s12 = s01*s11 
      s13 = s01*s12 
      s14 = s01*s13 
      s15 = s01*s14 
      s16 = s01*s15 
      s17 = s01*s16 
      s18 = s01*s17 
      s19 = s01*s18 
      s20 = s01*s19 
      s21 = s01*s20 
      s22 = s01*s21 

      r02 = r01*r01 
      r03 = r01*r02 
      r04 = r01*r03 
      r05 = r01*r04 
      r06 = r01*r05 
      r07 = r01*r06 
      r08 = r01*r07 
      r09 = r01*r08 
      r10 = r01*r09 
      r11 = r01*r10 
      r12 = r01*r11 
      r13 = r01*r12 
      r14 = r01*r13 
      r15 = r01*r14 
      r16 = r01*r15 
      r17 = r01*r16 
      r18 = r01*r17 
      r19 = r01*r18 
      r20 = r01*r19 
      r21 = r01*r20 
      r22 = r01*r21 

      b02 = b01*b01 
      b03 = b01*b02 
      b04 = b01*b03 
      b05 = b01*b04 
      b06 = b01*b05 
      b07 = b01*b06 
      b08 = b01*b07 
      b09 = b01*b08 
      b10 = b01*b09 
      b11 = b01*b10 
      b12 = b01*b11 
      b13 = b01*b12 
      b14 = b01*b13 
      b15 = b01*b14 
      b16 = b01*b15 
      b17 = b01*b16 
      b18 = b01*b17 
      b19 = b01*b18 
      b20 = b01*b19 
      b21 = b01*b20 
      b22 = b01*b21 

      return 
      end


*********************************************************************
*********************************************************************
*********************************************************************
***** below are the subroutines to generate all                 *****
***** expansion coefficients                                    *****
*********************************************************************
*********************************************************************
*********************************************************************
      subroutine fillcoeff2dhpl210(iflag,n,CsH2YZ,CsH3YZ,CsH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsH2YZ,CsH3YZ,CsH4YZ
      real(dp):: 
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
      dimension CsH2YZ(0:3,0:3,22),CsH3YZ(0:3,0:3,0:3,22),
     $          CsH4YZ(0:3,0:3,0:3,0:3,22)
      common /tpz/  
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
!$omp threadprivate(/tpz/)
      
* 2001-04-03:16:14:02.chpl 
* <- HPLexpand.210t 
* produced by form-to-fortr for gehrt@pcth62 
      if (n.eq.2) then
      CsH2YZ(2,1,02) = 
     $  + 1.406250000000000d-01
     $  - 1.406250000000000d-01*tz01

      CsH2YZ(2,1,03) = 
     $  - 8.789062500000000d-03
     $  + 8.789062500000000d-03*tz02

      CsH2YZ(2,1,04) = 
     $  + 8.239746093750000d-04*tz01
     $  - 8.239746093750000d-04*tz03

      CsH2YZ(2,1,05) = 
     $  + 3.089904785156250d-05
     $  - 1.235961914062500d-04*tz02
     $  + 9.269714355468750d-05*tz04

      CsH2YZ(2,1,06) = 
     $  - 7.724761962890625d-06*tz01
     $  + 1.931190490722656d-05*tz03
     $  - 1.158714294433593d-05*tz05

      CsH2YZ(2,1,07) = 
     $  - 2.069132668631417d-07
     $  + 1.758762768336704d-06*tz02
     $  - 3.103699002947126d-06*tz04
     $  + 1.551849501473563d-06*tz06

      CsH2YZ(2,1,08) = 
     $  + 8.244200476578303d-08*tz01
     $  - 3.734137862920761d-07*tz03
     $  + 5.092006176710128d-07*tz05
     $  - 2.182288361447198d-07*tz07

      CsH2YZ(2,1,09) = 
     $  + 1.717541765953813d-09
     $  - 2.505590340920857d-08*tz02
     $  + 7.638009265065193d-08*tz04
     $  - 8.486676961183547d-08*tz06
     $  + 3.182503860443830d-08*tz08

      CsH2YZ(2,1,10) = 
     $  - 9.395963778453213d-10*tz01
     $  + 6.668103326644216d-09*tz03
     $  - 1.527601853013038d-08*tz05
     $  + 1.432126737199723d-08*tz07
     $  - 4.773755790665745d-09*tz09

      CsH2YZ(2,1,11) = 
     $  - 1.601584734963615d-11
     $  + 3.569984038257608d-10*tz02
     $  - 1.642915953930418d-09*tz04
     $  + 3.010720981613055d-09*tz06
     $  - 2.441125120226801d-09*tz08
     $  + 7.323375360680405d-10*tz10

      CsH2YZ(2,1,12) = 
     $  + 1.115620011955502d-11*tz01
     $  - 1.138384472402062d-10*tz03
     $  + 3.849373391468751d-10*tz05
     $  - 5.873957320545741d-10*tz07
     $  + 4.195683800389815d-10*tz09
     $  - 1.144277400106313d-10*tz11

      CsH2YZ(2,1,13) = 
     $  + 1.609067324935821d-13
     $  - 5.086608776540966d-12*tz02
     $  + 3.268560630944703d-11*tz04
     $  - 8.706428105937108d-11*tz06
     $  + 1.137675799721084d-10*tz08
     $  - 7.261760423751603d-11*tz10
     $  + 1.815440105937900d-11*tz12

      CsH2YZ(2,1,14) = 
     $  - 1.362484493716330d-13*tz01
     $  + 1.887263073092009d-12*tz03
     $  - 8.747251494562695d-12*tz05
     $  + 1.918562043928251d-11*tz07
     $  - 2.191495556453609d-11*tz09
     $  + 1.264324359492466d-11*tz11
     $  - 2.917671598828769d-12*tz13

      CsH2YZ(2,1,15) = 
     $  - 1.703105617145412d-15
     $  + 7.247547085809578d-14*tz02
     $  - 6.174755836511188d-13*tz04
     $  + 2.225445006847388d-12*tz06
     $  - 4.144174289447529d-12*tz08
     $  + 4.203878495312451d-12*tz10
     $  - 2.212567629111816d-12*tz12
     $  + 4.741216348096750d-13*tz14

      CsH2YZ(2,1,16) = 
     $  + 1.698643848236619d-15*tz01
     $  - 3.064281183188281d-14*tz03
     $  + 1.854207700276031d-13*tz05
     $  - 5.449929416796629d-13*tz07
     $  + 8.811583508051338d-13*tz09
     $  - 8.037843340132772d-13*tz11
     $  + 3.889279035548115d-13*tz13
     $  - 7.778558071096230d-14*tz15

      CsH2YZ(2,1,17) = 
     $  - 1.032651610410497d-15*tz02
     $  + 1.123932667978306d-14*tz04
     $  - 5.230207105031973d-14*tz06
     $  + 1.295445854669541d-13*tz08
     $  - 1.849858768168263d-13*tz10
     $  + 1.532833502245433d-13*tz12
     $  - 6.863433592143733d-14*tz14
     $  + 1.286893798526949d-14*tz16

      CsH2YZ(2,1,18) = 
     $  + 4.898188535411796d-16*tz03
     $  - 3.737184718969277d-15*tz05
     $  + 1.406426156289116d-14*tz07
     $  - 3.006474429066559d-14*tz09
     $  + 3.842978094648733d-14*tz11
     $  - 2.916959276661086d-14*tz13
     $  + 1.215399698608786d-14*tz15
     $  - 2.144822997544916d-15*tz17

      CsH2YZ(2,1,19) = 
     $  - 1.989017778500321d-16*tz04
     $  + 1.155944946069913d-15*tz06
     $  - 3.641768383778570d-15*tz08
     $  + 6.841877852243595d-15*tz10
     $  - 7.913805556354953d-15*tz12
     $  + 5.541263099578215d-15*tz14
     $  - 2.158933675160343d-15*tz16
     $  + 3.598222791933906d-16*tz18

      CsH2YZ(2,1,20) = 
     $  - 3.381545319998254d-16*tz07
     $  + 9.145586774312298d-16*tz09
     $  - 1.531729173737769d-15*tz11
     $  + 1.617593906909573d-15*tz13
     $  - 1.051130833093692d-15*tz15
     $  + 3.845600608879362d-16*tz17

      CsH2YZ(2,1,21) = 
     $  - 2.239290089999690d-16*tz10
     $  + 3.381941223298206d-16*tz12
     $  - 3.285328864841949d-16*tz14
     $  + 1.991471743883955d-16*tz16
      endif
      if (n.eq.2) return
      
      if (n.eq.3) then
      CsH3YZ(0,2,1,02) = 
     $  + 7.031250000000000d-02
     $  - 7.031250000000000d-02*tz01

      CsH3YZ(0,2,1,03) = 
     $  - 2.050781250000000d-02
     $  + 1.757812500000000d-02*tz01
     $  + 2.929687500000000d-03*tz02

      CsH3YZ(0,2,1,04) = 
     $  + 2.471923828125000d-03
     $  - 1.441955566406250d-03*tz01
     $  - 8.239746093750000d-04*tz02
     $  - 2.059936523437500d-04*tz03

      CsH3YZ(0,2,1,05) = 
     $  - 7.621765136718750d-05
     $  - 6.179809570312500d-05*tz01
     $  + 5.767822265625000d-05*tz02
     $  + 6.179809570312500d-05*tz03
     $  + 1.853942871093750d-05*tz04

      CsH3YZ(0,2,1,06) = 
     $  - 1.223087310791015d-05
     $  + 1.544952392578125d-05*tz01
     $  + 7.724761962890625d-06*tz02
     $  - 3.218650817871093d-06*tz03
     $  - 5.793571472167968d-06*tz04
     $  - 1.931190490722656d-06*tz05

      CsH3YZ(0,2,1,07) = 
     $  + 7.291229403748804d-07
     $  + 4.138265337262834d-07*tz01
     $  - 1.128169955039510d-06*tz02
     $  - 1.034566334315708d-06*tz03
     $  + 1.773542287398357d-07*tz04
     $  + 6.207398005894252d-07*tz05
     $  + 2.216927859247947d-07*tz06

      CsH3YZ(0,2,1,08) = 
     $  + 1.131556928157806d-07
     $  - 1.836759370884725d-07*tz01
     $  - 8.244200476578303d-08*tz02
     $  + 1.117412466555833d-07*tz03
     $  + 1.454858907631465d-07*tz04
     $  - 4.243338480591773d-09*tz05
     $  - 7.274294538157326d-08*tz06
     $  - 2.727860451808997d-08*tz07

      CsH3YZ(0,2,1,09) = 
     $  - 8.143168607992785d-09
     $  - 3.435083531907626d-09*tz01
     $  + 1.815879451377051d-08*tz02
     $  + 1.555890776216983d-08*tz03
     $  - 1.220464972513062d-08*tz04
     $  - 2.121669240295886d-08*tz05
     $  - 1.347091581140245d-09*tz06
     $  + 9.092868172696658d-09*tz07
     $  + 3.536115400493144d-09*tz08

      CsH3YZ(0,2,1,10) = 
     $  - 1.228294942328440d-09
     $  + 2.280794433318078d-09*tz01
     $  + 9.395963778453213d-10*tz02
     $  - 2.417187455908528d-09*tz03
     $  - 2.864253474399447d-09*tz04
     $  + 1.368476659990847d-09*tz05
     $  + 3.182503860443830d-09*tz06
     $  + 4.091790677713496d-10*tz07
     $  - 1.193438947666436d-09*tz08
     $  - 4.773755790665745d-10*tz09

      CsH3YZ(0,2,1,11) = 
     $  + 9.679176562964375d-11
     $  + 3.203169469927231d-11*tz01
     $  - 2.768408882537792d-10*tz02
     $  - 2.273217043174164d-10*tz03
     $  + 3.497184732122898d-10*tz04
     $  + 5.207733589817177d-10*tz05
     $  - 1.499430869903619d-10*tz06
     $  - 4.882250240453603d-10*tz07
     $  - 8.630240324034147d-11*tz08
     $  + 1.627416746817867d-10*tz09
     $  + 6.657613964254913d-11*tz10

      CsH3YZ(0,2,1,12) = 
     $  + 1.427756821861046d-11
     $  - 2.900854206195440d-11*tz01
     $  - 1.115620011955502d-11*tz02
     $  + 4.539034291285913d-11*tz03
     $  + 5.134112356032559d-11*tz04
     $  - 5.192436195297237d-11*tz05
     $  - 9.408503067540798d-11*tz06
     $  + 1.498458500139219d-11*tz07
     $  + 7.628516000708755d-11*tz08
     $  + 1.631654811262706d-11*tz09
     $  - 2.288554800212626d-11*tz10
     $  - 9.535645000885944d-12*tz11

      CsH3YZ(0,2,1,13) = 
     $  - 1.188660305247408d-12
     $  - 3.218134649871642d-13*tz01
     $  + 4.120932620799263d-12*tz02
     $  + 3.283801362698256d-12*tz03
     $  - 7.986967663325885d-12*tz04
     $  - 1.110396170615985d-11*tz05
     $  + 7.730130261730688d-12*tz06
     $  + 1.694410765542040d-11*tz07
     $  - 1.126597900499148d-12*tz08
     $  - 1.210293403958600d-11*tz09
     $  - 2.945329402640510d-12*tz10
     $  + 3.300800192614365d-12*tz11
     $  + 1.396492389183000d-12*tz12

      CsH3YZ(0,2,1,14) = 
     $  - 1.724411181646781d-13
     $  + 3.747643211886650d-13*tz01
     $  + 1.362484493716330d-13*tz02
     $  - 7.960020212545110d-13*tz03
     $  - 8.755073118601884d-13*tz04
     $  + 1.426122199924165d-12*tz05
     $  + 2.332078956947439d-12*tz06
     $  - 1.137605228677122d-12*tz07
     $  - 3.047345892110048d-12*tz08
     $  - 1.070327500099442d-14*tz09
     $  + 1.945114399219179d-12*tz10
     $  + 5.199602344233088d-13*tz11
     $  - 4.862785998047949d-13*tz12
     $  - 2.084051142020549d-13*tz13

      CsH3YZ(0,2,1,15) = 
     $  + 1.490349958471799d-14
     $  + 3.406211234290825d-15*tz01
     $  - 6.055300384844107d-14*tz02
     $  - 4.718157682730024d-14*tz03
     $  + 1.634694474831364d-13*tz04
     $  + 2.186812873640673d-13*tz05
     $  - 2.537792918939041d-13*tz06
     $  - 4.796405109820627d-13*tz07
     $  + 1.632498931241416d-13*tz08
     $  + 5.478738891134022d-13*tz09
     $  + 3.187335185993246d-14*tz10
     $  - 3.160810898731166d-13*tz11
     $  - 9.077200529689504d-14*tz12
     $  + 7.294178997071923d-14*tz13
     $  + 3.160810898731166d-14*tz14

      CsH3YZ(0,2,1,16) = 
     $  + 2.135662691942712d-15
     $  - 4.896843951656765d-15*tz01
     $  - 1.698643848236619d-15*tz02
     $  + 1.340601729068304d-14*tz03
     $  + 1.447208399182309d-14*tz04
     $  - 3.371516035770854d-14*tz05
     $  - 5.215886734798566d-14*tz06
     $  + 4.467599835448495d-14*tz07
     $  + 9.712908490892648d-14*tz08
     $  - 2.241137024496398d-14*tz09
     $  - 9.852840223388559d-14*tz10
     $  - 1.005291641042649d-14*tz11
     $  + 5.185705380730820d-14*tz12
     $  + 1.576012795998206d-14*tz13
     $  - 1.111222581585175d-14*tz14
     $  - 4.861598794435144d-15*tz15

      CsH3YZ(0,2,1,17) = 
     $  - 1.896721343193203d-16
     $  + 8.827689838328240d-16*tz02
     $  + 6.759443786444738d-16*tz03
     $  - 3.132949132942859d-15*tz04
     $  - 4.090164044726540d-15*tz05
     $  + 6.873204564130966d-15*tz06
     $  + 1.202190312528668d-14*tz07
     $  - 7.747672602503828d-15*tz08
     $  - 1.943731656187795d-14*tz09
     $  + 2.840773967248380d-15*tz10
     $  + 1.773053677970464d-14*tz11
     $  + 2.446543195992914d-15*tz12
     $  - 8.579291990179666d-15*tz13
     $  - 2.729993193513753d-15*tz14
     $  + 1.715858398035933d-15*tz15
     $  + 7.569963520746764d-16*tz16

      CsH3YZ(0,2,1,18) = 
     $  - 2.198078889245289d-16*tz03
     $  - 2.341526391621471d-16*tz04
     $  + 7.289386183564251d-16*tz05
     $  + 1.089626480214994d-15*tz06
     $  - 1.378470209397595d-15*tz07
     $  - 2.698845530561545d-15*tz08
     $  + 1.318590777076205d-15*tz09
     $  + 3.853872433683881d-15*tz10
     $  - 3.042883805319255d-16*tz11
     $  - 3.193403129677986d-15*tz12
     $  - 5.364676325361125d-16*tz13
     $  + 1.429881998363277d-15*tz14
     $  + 4.726554383478612d-16*tz15
     $  - 2.681028746931145d-16*tz16
     $  - 1.191568331969398d-16*tz17

      CsH3YZ(0,2,1,19) = 
     $  - 1.663707823051708d-16*tz06
     $  - 2.775841097939045d-16*tz07
     $  + 2.716232953034269d-16*tz08
     $  + 5.933831109999787d-16*tz09
     $  - 2.191972687817849d-16*tz10
     $  - 7.584825186806710d-16*tz11
     $  + 5.757156467094249d-16*tz13
     $  + 1.113512825901905d-16*tz14
     $  - 2.398815194622604d-16*tz15

      CsH3YZ(0,2,1,20) = 
     $  - 1.282852097295674d-16*tz10
     $  + 1.483838541816553d-16*tz12
     $  - 1.038986831170915d-16*tz14

      CsH3YZ(2,0,1,02) = 
     $  + 1.406250000000000d-01
     $  - 1.406250000000000d-01*tz01

      CsH3YZ(2,0,1,03) = 
     $  - 2.197265625000000d-02
     $  + 1.757812500000000d-02*tz01
     $  + 4.394531250000000d-03*tz02

      CsH3YZ(2,0,1,04) = 
     $  + 1.922607421875000d-03
     $  - 8.239746093750000d-04*tz01
     $  - 8.239746093750000d-04*tz02
     $  - 2.746582031250000d-04*tz03

      CsH3YZ(2,0,1,05) = 
     $  - 5.407333374023437d-05
     $  - 6.179809570312500d-05*tz01
     $  + 3.089904785156250d-05*tz02
     $  + 6.179809570312500d-05*tz03
     $  + 2.317428588867187d-05*tz04

      CsH3YZ(2,0,1,06) = 
     $  - 6.051063537597656d-06
     $  + 7.724761962890625d-06*tz01
     $  + 7.724761962890625d-06*tz02
     $  - 1.287460327148437d-06*tz03
     $  - 5.793571472167968d-06*tz04
     $  - 2.317428588867187d-06*tz05

      CsH3YZ(2,0,1,07) = 
     $  + 4.138265337262834d-07
     $  + 4.138265337262834d-07*tz01
     $  - 6.724681173052106d-07*tz02
     $  - 1.034566334315708d-06*tz03
     $  + 6.207398005894252d-07*tz05
     $  + 2.586415835789271d-07*tz06

      CsH3YZ(2,0,1,08) = 
     $  + 3.925809750751573d-08
     $  - 8.244200476578303d-08*tz01
     $  - 8.244200476578303d-08*tz02
     $  + 6.950992558683667d-08*tz03
     $  + 1.454858907631465d-07*tz04
     $  + 1.454858907631465d-08*tz05
     $  - 7.274294538157326d-08*tz06
     $  - 3.117554802067425d-08*tz07

      CsH3YZ(2,0,1,09) = 
     $  - 3.826582022676510d-09
     $  - 3.435083531907626d-09*tz01
     $  + 1.081040993865047d-08*tz02
     $  + 1.555890776216983d-08*tz03
     $  - 7.425842341035604d-09*tz04
     $  - 2.121669240295886d-08*tz05
     $  - 3.536115400493144d-09*tz06
     $  + 9.092868172696658d-09*tz07
     $  + 3.978129825554788d-09*tz08

      CsH3YZ(2,0,1,10) = 
     $  - 3.230493998021951d-10
     $  + 9.395963778453213d-10*tz01
     $  + 9.395963778453213d-10*tz02
     $  - 1.596303523651191d-09*tz03
     $  - 2.864253474399447d-09*tz04
     $  + 7.638009265065193d-10*tz05
     $  + 3.182503860443830d-09*tz06
     $  + 6.819651129522493d-10*tz07
     $  - 1.193438947666436d-09*tz08
     $  - 5.304173100739717d-10*tz09

      CsH3YZ(2,0,1,11) = 
     $  + 3.887717784064906d-11
     $  + 3.203169469927231d-11*tz01
     $  - 1.624833545632442d-10*tz02
     $  - 2.273217043174164d-10*tz03
     $  + 2.402377102445423d-10*tz04
     $  + 5.207733589817177d-10*tz05
     $  - 6.780903111741116d-11*tz06
     $  - 4.882250240453603d-10*tz07
     $  - 1.220562560113400d-10*tz08
     $  + 1.627416746817867d-10*tz09
     $  + 7.323375360680405d-11*tz10

      CsH3YZ(2,0,1,12) = 
     $  + 3.005417591300553d-12
     $  - 1.115620011955502d-11*tz01
     $  - 1.115620011955502d-11*tz02
     $  + 3.050868233369871d-11*tz03
     $  + 5.134112356032559d-11*tz04
     $  - 3.591456898111455d-11*tz05
     $  - 9.408503067540798d-11*tz06
     $  + 3.269364000303752d-12*tz07
     $  + 7.628516000708755d-11*tz08
     $  + 2.119032222419098d-11*tz09
     $  - 2.288554800212626d-11*tz10
     $  - 1.040252181914830d-11*tz11

      CsH3YZ(2,0,1,13) = 
     $  - 4.183924335858808d-13
     $  - 3.218134649871642d-13*tz01
     $  + 2.382397655776901d-12*tz02
     $  + 3.283801362698256d-12*tz03
     $  - 5.708550555338066d-12*tz04
     $  - 1.110396170615985d-11*tz05
     $  + 5.257412088095297d-12*tz06
     $  + 1.694410765542040d-11*tz07
     $  + 6.051467019793003d-13*tz08
     $  - 1.210293403958600d-11*tz09
     $  - 3.630880211875801d-12*tz10
     $  + 3.300800192614365d-12*tz11
     $  + 1.512866754948250d-12*tz12

      CsH3YZ(2,0,1,14) = 
     $  - 3.017632736344513d-14
     $  + 1.362484493716330d-13*tz01
     $  + 1.362484493716330d-13*tz02
     $  - 5.382553914495812d-13*tz03
     $  - 8.755073118601884d-13*tz04
     $  + 1.049044449424388d-12*tz05
     $  + 2.332078956947439d-12*tz06
     $  - 7.418780996568390d-13*tz07
     $  - 3.047345892110048d-12*tz08
     $  - 2.737568413715882d-13*tz09
     $  + 1.945114399219179d-12*tz10
     $  + 6.189000361151935d-13*tz11
     $  - 4.862785998047949d-13*tz12
     $  - 2.244362768329822d-13*tz13

      CsH3YZ(2,0,1,15) = 
     $  + 4.684554014858312d-15
     $  + 3.406211234290825d-15*tz01
     $  - 3.453462981190247d-14*tz02
     $  - 4.718157682730024d-14*tz03
     $  + 1.189827132923045d-13*tz04
     $  + 2.186812873640673d-13*tz05
     $  - 1.886730950045085d-13*tz06
     $  - 4.796405109820627d-13*tz07
     $  + 9.833633907163630d-14*tz08
     $  + 5.478738891134022d-13*tz09
     $  + 7.269865067081683d-14*tz10
     $  - 3.160810898731166d-13*tz11
     $  - 1.053603632910388d-13*tz12
     $  + 7.294178997071923d-14*tz13
     $  + 3.386583105783393d-14*tz14

      CsH3YZ(2,0,1,16) = 
     $  + 3.193493584406278d-16
     $  - 1.698643848236619d-15*tz01
     $  - 1.698643848236619d-15*tz02
     $  + 9.081841378469858d-15*tz03
     $  + 1.447208399182309d-14*tz04
     $  - 2.550648681206215d-14*tz05
     $  - 5.215886734798566d-14*tz06
     $  + 3.314853394167842d-14*tz07
     $  + 9.712908490892648d-14*tz08
     $  - 1.156951905930244d-14*tz09
     $  - 9.852840223388559d-14*tz10
     $  - 1.649997166596170d-14*tz11
     $  + 5.185705380730820d-14*tz12
     $  + 1.795051862560668d-14*tz13
     $  - 1.111222581585175d-14*tz14
     $  - 5.185705380730820d-15*tz15

      CsH3YZ(2,0,1,17) = 
     $  + 4.975907627614622d-16*tz02
     $  + 6.759443786444738d-16*tz03
     $  - 2.302873385962410d-15*tz04
     $  - 4.090164044726540d-15*tz05
     $  + 5.308541804447839d-15*tz06
     $  + 1.202190312528668d-14*tz07
     $  - 5.673907948743425d-15*tz08
     $  - 1.943731656187795d-14*tz09
     $  + 1.005002775992475d-15*tz10
     $  + 1.773053677970464d-14*tz11
     $  + 3.479379529350642d-15*tz12
     $  - 8.579291990179666d-15*tz13
     $  - 3.064032853635595d-15*tz14
     $  + 1.715858398035933d-15*tz15
     $  + 8.043086240793437d-16*tz16

      CsH3YZ(2,0,1,18) = 
     $  - 1.489305677024696d-16*tz03
     $  - 2.341526391621471d-16*tz04
     $  + 5.601148324641378d-16*tz05
     $  + 1.089626480214994d-15*tz06
     $  - 1.075214668800171d-15*tz07
     $  - 2.698845530561545d-15*tz08
     $  + 9.415533384636920d-16*tz09
     $  + 3.853872433683881d-15*tz10
     $  - 3.193403129677986d-15*tz12
     $  - 7.039419068865367d-16*tz13
     $  + 1.429881998363277d-15*tz14
     $  + 5.242900660665351d-16*tz15
     $  - 2.681028746931145d-16*tz16
     $  - 1.261660586791127d-16*tz17

      CsH3YZ(2,0,1,19) = 
     $  - 1.311906370812364d-16*tz06
     $  - 2.775841097939045d-16*tz07
     $  + 2.123349519026548d-16*tz08
     $  + 5.933831109999787d-16*tz09
     $  - 1.501429853243786d-16*tz10
     $  - 7.584825186806710d-16*tz11
     $  + 5.757156467094249d-16*tz13
     $  + 1.387885934031649d-16*tz14
     $  - 2.398815194622604d-16*tz15

      CsH3YZ(2,0,1,20) = 
     $  - 1.282852097295674d-16*tz10
     $  + 1.483838541816553d-16*tz12
     $  - 1.038986831170915d-16*tz14

      CsH3YZ(2,1,1,03) = 
     $  + 2.636718750000000d-02
     $  - 3.515625000000000d-02*tz01
     $  + 8.789062500000000d-03*tz02

      CsH3YZ(2,1,1,04) = 
     $  - 2.471923828125000d-03
     $  + 1.235961914062500d-03*tz01
     $  + 2.471923828125000d-03*tz02
     $  - 1.235961914062500d-03*tz03

      CsH3YZ(2,1,1,05) = 
     $  + 1.544952392578125d-05
     $  + 2.471923828125000d-04*tz01
     $  - 1.853942871093750d-04*tz02
     $  - 2.471923828125000d-04*tz03
     $  + 1.699447631835937d-04*tz04

      CsH3YZ(2,1,1,06) = 
     $  + 9.655952453613281d-06
     $  - 9.655952453613281d-06*tz01
     $  - 3.862380981445312d-05*tz02
     $  + 3.379583358764648d-05*tz03
     $  + 2.896785736083984d-05*tz04
     $  - 2.413988113403320d-05*tz05

      CsH3YZ(2,1,1,07) = 
     $  - 5.172831671578543d-08
     $  - 2.482959202357700d-06*tz01
     $  + 2.664008310862949d-06*tz02
     $  + 6.207398005894252d-06*tz03
     $  - 6.155669689178466d-06*tz04
     $  - 3.724438803536551d-06*tz05
     $  + 3.543389695031302d-06*tz06

      CsH3YZ(2,1,1,08) = 
     $  - 6.789341568946838d-08
     $  + 9.335344657301902d-08*tz01
     $  + 5.770940333604812d-07*tz02
     $  - 6.704474799335002d-07*tz03
     $  - 1.018401235342025d-06*tz04
     $  + 1.111754681915044d-06*tz05
     $  + 5.092006176710128d-07*tz06
     $  - 5.346606485545635d-07*tz07

      CsH3YZ(2,1,1,09) = 
     $  + 2.273217043174164d-10
     $  + 2.748066825526101d-08*tz01
     $  - 3.778591885098389d-08*tz02
     $  - 1.244712620973587d-07*tz03
     $  + 1.548818545415997d-07*tz04
     $  + 1.697335392236709d-07*tz05
     $  - 1.998410360621554d-07*tz06
     $  - 7.274294538157326d-08*tz07
     $  + 8.251777866722217d-08*tz08

      CsH3YZ(2,1,1,10) = 
     $  + 5.796703460094119d-10
     $  - 1.011581584212503d-09*tz01
     $  - 8.456367400607892d-09*tz02
     $  + 1.207078249925481d-08*tz03
     $  + 2.577828126959502d-08*tz04
     $  - 3.393913045459027d-08*tz05
     $  - 2.864253474399447d-08*tz06
     $  + 3.585431581346451d-08*tz07
     $  + 1.074095052899792d-08*tz08
     $  - 1.297438627391654d-08*tz09

      CsH3YZ(2,1,1,11) = 
     $  - 1.227020563076963d-12
     $  - 3.203169469927231d-10*tz01
     $  + 5.368537863609895d-10*tz02
     $  + 2.273217043174164d-09*tz03
     $  - 3.363069623305051d-09*tz04
     $  - 5.207733589817177d-09*tz05
     $  + 7.186885468045502d-09*tz06
     $  + 4.882250240453603d-09*tz07
     $  - 6.431202251264181d-09*tz08
     $  - 1.627416746817867d-09*tz09
     $  + 2.071759640725817d-09*tz10

      CsH3YZ(2,1,1,12) = 
     $  - 5.505447526437429d-12
     $  + 1.169907599367953d-11*tz01
     $  + 1.227182013151052d-10*tz02
     $  - 2.068760705599613d-10*tz03
     $  - 5.647523591635815d-10*tz04
     $  + 8.625904105282903d-10*tz05
     $  + 1.034935337429487d-09*tz06
     $  - 1.486570729371448d-09*tz07
     $  - 8.391367600779631d-10*tz08
     $  + 1.154312531273912d-09*tz09
     $  + 2.517410280233889d-10*tz10
     $  - 3.351552178644721d-10*tz11

      CsH3YZ(2,1,1,13) = 
     $  + 7.829940492180461d-15
     $  + 3.861761579845970d-12*tz01
     $  - 7.639111161820851d-12*tz02
     $  - 3.940561635237907d-11*tz03
     $  + 6.724732748674999d-11*tz04
     $  + 1.332475404739183d-10*tz05
     $  - 2.093724341153926d-10*tz06
     $  - 2.033292918650449d-10*tz07
     $  + 3.023457005636674d-10*tz08
     $  + 1.452352084750320d-10*tz09
     $  - 2.074133771827320d-10*tz10
     $  - 3.960960231137238d-11*tz11
     $  + 5.482406446903594d-11*tz12

      CsH3YZ(2,1,1,14) = 
     $  + 5.603002292187234d-14
     $  - 1.408048260547486d-13*tz01
     $  - 1.771229841831229d-12*tz02
     $  + 3.437525018808959d-12*tz03
     $  + 1.138159505418244d-11*tz04
     $  - 1.969882017709060d-11*tz05
     $  - 3.031702644031671d-11*tz06
     $  + 4.886111933422712d-11*tz07
     $  + 3.961549659743062d-11*tz08
     $  - 6.072387525095427d-11*tz09
     $  - 2.528648718984933d-11*tz10
     $  + 3.731900556206101d-11*tz11
     $  + 6.321621797462333d-12*tz12
     $  - 9.054149660997459d-12*tz13

      CsH3YZ(2,1,1,15) = 
     $  - 4.768695728007155d-14*tz01
     $  + 1.087795657345156d-13*tz02
     $  + 6.605420755822034d-13*tz03
     $  - 1.274333594005115d-12*tz04
     $  - 3.061538023096943d-12*tz05
     $  + 5.376044591004038d-12*tz06
     $  + 6.714967153748878d-12*tz07
     $  - 1.107266721841155d-11*tz08
     $  - 7.670234447587631d-12*tz09
     $  + 1.207869563013588d-11*tz10
     $  + 4.425135258223633d-12*tz11
     $  - 6.724232234646674d-12*tz12
     $  - 1.021185059590069d-12*tz13
     $  + 1.507770214897446d-12*tz14

      CsH3YZ(2,1,1,16) = 
     $  - 5.987480685276842d-16
     $  + 1.740115685686245d-15*tz01
     $  + 2.547965772354929d-14*tz02
     $  - 5.589718821909925d-14*tz03
     $  - 2.170812598773464d-13*tz04
     $  + 4.188941308458799d-13*tz05
     $  + 7.823830102197849d-13*tz06
     $  - 1.393580578250584d-12*tz07
     $  - 1.456936273633897d-12*tz08
     $  + 2.452835037220153d-12*tz09
     $  + 1.477926033508283d-12*tz10
     $  - 2.384402976367026d-12*tz11
     $  - 7.778558071096230d-13*tz12
     $  + 1.213336122874529d-12*tz13
     $  + 1.666833872377763d-13*tz14
     $  - 2.529246637895382d-13*tz15

      CsH3YZ(2,1,1,17) = 
     $  + 5.995213582011599d-16*tz01
     $  - 1.549494621610343d-15*tz02
     $  - 1.081511005831158d-14*tz03
     $  + 2.324320875762107d-14*tz04
     $  + 6.544262471562465d-14*tz05
     $  - 1.267275279080672d-13*tz06
     $  - 1.923504500045869d-13*tz07
     $  + 3.473648600030515d-13*tz08
     $  + 3.109970649900472d-13*tz09
     $  - 5.336175230714396d-13*tz10
     $  - 2.836885884752743d-13*tz11
     $  + 4.678307952338011d-13*tz12
     $  + 1.372686718428746d-13*tz13
     $  - 2.192468589381154d-13*tz14
     $  - 2.745373436857493d-14*tz15
     $  + 4.270208313478716d-14*tz16

      CsH3YZ(2,1,1,18) = 
     $  - 3.657307786870510d-16*tz02
     $  + 8.944286375451293d-16*tz03
     $  + 3.980594865756501d-15*tz04
     $  - 8.461200605295326d-15*tz05
     $  - 1.852365016365490d-14*tz06
     $  + 3.606293263600188d-14*tz07
     $  + 4.588037401954627d-14*tz08
     $  - 8.395471686863476d-14*tz09
     $  - 6.551583137262599d-14*tz10
     $  + 1.143942427095030d-13*tz11
     $  + 5.428785320452577d-14*tz12
     $  - 9.133411023169100d-14*tz13
     $  - 2.430799397217572d-14*tz14
     $  + 3.967140367944144d-14*tz15
     $  + 4.557748869782947d-15*tz16
     $  - 7.251065293144417d-15*tz17

      CsH3YZ(2,1,1,19) = 
     $  + 1.740145927054190d-16*tz03
     $  - 4.119268724376702d-16*tz04
     $  - 1.327684044896980d-15*tz05
     $  + 2.806770933893762d-15*tz06
     $  + 4.996513976290282d-15*tz07
     $  - 9.789794383637680d-15*tz08
     $  - 1.068089599799961d-14*tz09
     $  + 1.979175083848445d-14*tz10
     $  + 1.365268533625207d-14*tz11
     $  - 2.422638445062230d-14*tz12
     $  - 1.036288164076964d-14*tz13
     $  + 1.775753372352757d-14*tz14
     $  + 4.317867350320687d-15*tz15
     $  - 7.187648083851510d-15*tz16
     $  - 7.619765912330624d-16*tz17
     $  + 1.237627628101973d-15*tz18

      CsH3YZ(2,1,1,20) = 
     $  + 1.643558430458754d-16*tz05
     $  + 4.118053870374066d-16*tz06
     $  - 8.688108490758315d-16*tz07
     $  - 1.297379986721115d-15*tz08
     $  + 2.559710193613884d-15*tz09
     $  + 2.437418984861780d-15*tz10
     $  - 4.570895569576882d-15*tz11
     $  - 2.819293229451452d-15*tz12
     $  + 5.078387137440285d-15*tz13
     $  + 1.974074979224739d-15*tz14
     $  - 3.440532852144638d-15*tz15
     $  - 7.691201217758724d-16*tz16
     $  + 1.303863544157124d-15*tz17
     $  + 1.281866869626454d-16*tz18
     $  - 2.122229961096461d-16*tz19

      CsH3YZ(2,1,1,21) = 
     $  - 1.207694757142233d-16*tz07
     $  + 2.548452101841461d-16*tz08
     $  + 3.266280990825820d-16*tz09
     $  - 6.491077902818225d-16*tz10
     $  - 5.470461334777747d-16*tz11
     $  + 1.037616243918728d-15*tz12
     $  + 5.777121096105617d-16*tz13
     $  - 1.055299005738474d-15*tz14
     $  - 3.754038689620329d-16*tz15
     $  + 6.646450339736291d-16*tz16
     $  + 1.373428788885486d-16*tz17
     $  - 2.368014038216825d-16*tz18

      CsH3YZ(2,1,1,22) = 
     $  + 1.604676794021004d-16*tz11
     $  + 1.210581233339698d-16*tz12
     $  - 2.321185418455313d-16*tz13
     $  - 1.175998400483197d-16*tz14
     $  + 2.176524915001066d-16*tz15
     $  - 1.280729109145949d-16*tz17

      CsH3YZ(2,2,1,03) = 
     $  + 3.515625000000000d-02
     $  - 3.515625000000000d-02*tz01

      CsH3YZ(2,2,1,04) = 
     $  - 1.647949218750000d-03
     $  + 1.647949218750000d-03*tz02

      CsH3YZ(2,2,1,05) = 
     $  + 1.235961914062500d-04*tz01
     $  - 1.235961914062500d-04*tz03

      CsH3YZ(2,2,1,06) = 
     $  + 3.862380981445312d-06
     $  - 1.544952392578125d-05*tz02
     $  + 1.158714294433593d-05*tz04

      CsH3YZ(2,2,1,07) = 
     $  - 8.276530674525669d-07*tz01
     $  + 2.069132668631417d-06*tz03
     $  - 1.241479601178850d-06*tz05

      CsH3YZ(2,2,1,08) = 
     $  - 1.939811876841953d-08
     $  + 1.648840095315660d-07*tz02
     $  - 2.909717815262930d-07*tz04
     $  + 1.454858907631465d-07*tz06

      CsH3YZ(2,2,1,09) = 
     $  + 6.870167063815253d-09*tz01
     $  - 3.111781552433967d-08*tz03
     $  + 4.243338480591773d-08*tz05
     $  - 1.818573634539331d-08*tz07

      CsH3YZ(2,2,1,10) = 
     $  + 1.288156324465359d-10
     $  - 1.879192755690642d-09*tz02
     $  + 5.728506948798894d-09*tz04
     $  - 6.365007720887660d-09*tz06
     $  + 2.386877895332872d-09*tz08

      CsH3YZ(2,2,1,11) = 
     $  - 6.406338939854463d-11*tz01
     $  + 4.546434086348329d-10*tz03
     $  - 1.041546717963435d-09*tz05
     $  + 9.764500480907207d-10*tz07
     $  - 3.254833493635735d-10*tz09

      CsH3YZ(2,2,1,12) = 
     $  - 1.000990459352259d-12
     $  + 2.231240023911005d-11*tz02
     $  - 1.026822471206511d-10*tz04
     $  + 1.881700613508159d-10*tz06
     $  - 1.525703200141751d-10*tz08
     $  + 4.577109600425253d-11*tz10

      CsH3YZ(2,2,1,13) = 
     $  + 6.436269299743284d-13*tz01
     $  - 6.567602725396512d-12*tz03
     $  + 2.220792341231971d-11*tz05
     $  - 3.388821531084081d-11*tz07
     $  + 2.420586807917201d-11*tz09
     $  - 6.601600385228730d-12*tz11

      CsH3YZ(2,2,1,14) = 
     $  + 8.620003526441899d-15
     $  - 2.724968987432660d-13*tz02
     $  + 1.751014623720376d-12*tz04
     $  - 4.664157913894879d-12*tz06
     $  + 6.094691784220096d-12*tz08
     $  - 3.890228798438359d-12*tz10
     $  + 9.725571996095898d-13*tz12

      CsH3YZ(2,2,1,15) = 
     $  - 6.812422468581651d-15*tz01
     $  + 9.436315365460049d-14*tz03
     $  - 4.373625747281347d-13*tz05
     $  + 9.592810219641255d-13*tz07
     $  - 1.095747778226804d-12*tz09
     $  + 6.321621797462333d-13*tz11
     $  - 1.458835799414384d-13*tz13

      CsH3YZ(2,2,1,16) = 
     $  + 3.397287696473239d-15*tz02
     $  - 2.894416798364619d-14*tz04
     $  + 1.043177346959713d-13*tz06
     $  - 1.942581698178529d-13*tz08
     $  + 1.970568044677711d-13*tz10
     $  - 1.037141076146164d-13*tz12
     $  + 2.222445163170351d-14*tz14

      CsH3YZ(2,2,1,17) = 
     $  - 1.351888757288947d-15*tz03
     $  + 8.180328089453081d-15*tz05
     $  - 2.404380625057336d-14*tz07
     $  + 3.887463312375590d-14*tz09
     $  - 3.546107355940928d-14*tz11
     $  + 1.715858398035933d-14*tz13
     $  - 3.431716796071866d-15*tz15

      CsH3YZ(2,2,1,18) = 
     $  + 4.683052783242942d-16*tz04
     $  - 2.179252960429989d-15*tz06
     $  + 5.397691061123090d-15*tz08
     $  - 7.707744867367763d-15*tz10
     $  + 6.386806259355973d-15*tz12
     $  - 2.859763996726555d-15*tz14
     $  + 5.362057493862291d-16*tz16

      CsH3YZ(2,2,1,19) = 
     $  - 1.475204494329978d-16*tz05
     $  + 5.551682195878091d-16*tz07
     $  - 1.186766221999957d-15*tz09
     $  + 1.516965037361342d-15*tz11
     $  - 1.151431293418849d-15*tz13
     $  + 4.797630389245208d-16*tz15

      CsH3YZ(2,2,1,20) = 
     $  - 1.365663143916963d-16*tz08
     $  + 2.565704194591348d-16*tz10
     $  - 2.967677083633107d-16*tz12
     $  + 2.077973662341830d-16*tz14
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CsH4YZ(0,0,2,1,02) = 
     $  + 3.515625000000000d-02
     $  - 3.515625000000000d-02*tz01

      CsH4YZ(0,0,2,1,03) = 
     $  - 1.562500000000000d-02
     $  + 1.464843750000000d-02*tz01
     $  + 9.765625000000000d-04*tz02

      CsH4YZ(0,0,2,1,04) = 
     $  + 3.364562988281250d-03
     $  - 2.832412719726562d-03*tz01
     $  - 4.806518554687500d-04*tz02
     $  - 5.149841308593750d-05*tz03

      CsH4YZ(0,0,2,1,05) = 
     $  - 3.928985595703125d-04
     $  + 2.605819702148437d-04*tz01
     $  + 1.007995605468750d-04*tz02
     $  + 2.780914306640625d-05*tz03
     $  + 3.707885742187500d-06*tz04

      CsH4YZ(0,0,2,1,06) = 
     $  + 1.688718795776367d-05
     $  + 3.218650817871093d-07*tz01
     $  - 8.754730224609375d-06*tz02
     $  - 6.008148193359375d-06*tz03
     $  - 2.124309539794921d-06*tz04
     $  - 3.218650817871093d-07*tz05

      CsH4YZ(0,0,2,1,07) = 
     $  + 1.536460034006191d-06
     $  - 2.285898948202327d-06*tz01
     $  - 3.726784758943162d-07*tz02
     $  + 4.384590654957051d-07*tz03
     $  + 4.598541788040027d-07*tz04
     $  + 1.921337478014887d-07*tz05
     $  + 3.167039798925638d-08*tz06

      CsH4YZ(0,0,2,1,08) = 
     $  - 1.757569630079123d-07
     $  + 7.564761160340692d-08*tz01
     $  + 1.331022702993787d-07*tz02
     $  + 5.491923989861139d-08*tz03
     $  - 2.407445097152067d-08*tz04
     $  - 4.094316474428134d-08*tz05
     $  - 1.948471751292141d-08*tz06
     $  - 3.409825564761246d-09*tz07

      CsH4YZ(0,0,2,1,09) = 
     $  - 1.151158179783699d-08
     $  + 2.393950126133859d-08*tz01
     $  - 1.323613699297515d-09*tz02
     $  - 1.133296921450112d-08*tz03
     $  - 7.399509907985220d-09*tz04
     $  + 1.052415297765816d-09*tz05
     $  + 4.035929141908275d-09*tz06
     $  + 2.146927207442266d-09*tz07
     $  + 3.929017111659049d-10*tz08

      CsH4YZ(0,0,2,1,10) = 
     $  + 2.123361524094694d-09
     $  - 1.449744169284323d-09*tz01
     $  - 1.797862101479300d-09*tz02
     $  - 2.811843192570709d-10*tz03
     $  + 1.107814439040209d-09*tz04
     $  + 9.974497515941038d-10*tz05
     $  + 2.778376386101756d-11*tz06
     $  - 4.279331083775364d-10*tz07
     $  - 2.519482222851365d-10*tz08
     $  - 4.773755790665745d-11*tz09

      CsH4YZ(0,0,2,1,11) = 
     $  + 1.003663528693840d-10
     $  - 2.713378869405856d-10*tz01
     $  + 7.407021536386711d-11*tz02
     $  + 2.024368662311538d-10*tz03
     $  + 8.026328347884171d-11*tz04
     $  - 1.145203536440406d-10*tz05
     $  - 1.367227763715776d-10*tz06
     $  - 1.958536171495096d-11*tz07
     $  + 4.790841923024439d-11*tz08
     $  + 3.106886516652293d-11*tz09
     $  + 6.052376331140830d-12*tz10

      CsH4YZ(0,0,2,1,12) = 
     $  - 2.693621620440309d-11
     $  + 2.285157621570335d-11*tz01
     $  + 2.388274683095067d-11*tz02
     $  - 2.805755627832695d-12*tz03
     $  - 2.600813920780358d-11*tz04
     $  - 1.604736374951563d-11*tz05
     $  + 1.194088417282658d-11*tz06
     $  + 1.910307548510817d-11*tz07
     $  + 4.392175879195950d-12*tz08
     $  - 5.590713346815722d-12*tz09
     $  - 3.987633364006849d-12*tz10
     $  - 7.946370834071620d-13*tz11

      CsH4YZ(0,0,2,1,13) = 
     $  - 9.182705265908504d-13
     $  + 3.214105477666578d-12*tz01
     $  - 1.586285917259884d-12*tz02
     $  - 3.264185025161916d-12*tz03
     $  - 4.595431890036522d-13*tz04
     $  + 3.519696272368430d-12*tz05
     $  + 2.913861412148165d-12*tz06
     $  - 1.196657619687116d-12*tz07
     $  - 2.717916539858102d-12*tz08
     $  - 8.148556600855691d-13*tz09
     $  + 6.736544341462561d-13*tz10
     $  + 5.289743898420457d-13*tz11
     $  + 1.074224914756154d-13*tz12

      CsH4YZ(0,0,2,1,14) = 
     $  + 3.492716140233895d-13
     $  - 3.389040432007484d-13*tz01
     $  - 3.170153285586166d-13*tz02
     $  + 1.317399764478065d-13*tz03
     $  + 5.086210420727378d-13*tz04
     $  + 1.921166638198944d-13*tz05
     $  - 4.859748550676928d-13*tz06
     $  - 5.092576658965099d-13*tz07
     $  + 1.053896907827206d-13*tz08
     $  + 3.930370980459030d-13*tz09
     $  + 1.412037759006930d-13*tz10
     $  - 8.320165694454334d-14*tz11
     $  - 7.214023183917287d-14*tz12
     $  - 1.488607958586106d-14*tz13

      CsH4YZ(0,0,2,1,15) = 
     $  + 8.446997154042006d-15
     $  - 3.917610719915262d-14*tz01
     $  + 2.814662946028226d-14*tz02
     $  + 5.033305337384482d-14*tz03
     $  - 6.733545375847382d-15*tz04
     $  - 8.278103228586949d-14*tz05
     $  - 4.842286065755808d-14*tz06
     $  + 6.728565890963970d-14*tz07
     $  + 8.746220736610233d-14*tz08
     $  - 5.796983558230897d-15*tz09
     $  - 5.765759708082589d-14*tz10
     $  - 2.375607791687060d-14*tz11
     $  + 1.046953669152374d-14*tz12
     $  + 1.007291385309932d-14*tz13
     $  + 2.107207265820777d-15*tz14

      CsH4YZ(0,0,2,1,16) = 
     $  - 4.587148744060979d-15
     $  + 4.893728677126101d-15*tz01
     $  + 4.220308825313735d-15*tz02
     $  - 3.123195862198950d-15*tz03
     $  - 9.133774921919586d-15*tz04
     $  - 1.244863618223931d-15*tz05
     $  + 1.365136577531050d-14*tz06
     $  + 1.058105506991090d-14*tz07
     $  - 9.217608742557962d-15*tz08
     $  - 1.489661015145183d-14*tz09
     $  - 5.779091331930830d-16*tz10
     $  + 8.565068579694676d-15*tz11
     $  + 3.943890401737130d-15*tz12
     $  - 1.335127062953686d-15*tz13
     $  - 1.435329167880852d-15*tz14
     $  - 3.038499246521965d-16*tz15

      CsH4YZ(0,0,2,1,17) = 
     $  + 4.873628852193045d-16*tz01
     $  - 4.620633408821860d-16*tz02
     $  - 7.577855359897560d-16*tz03
     $  + 3.281843539121524d-16*tz04
     $  + 1.719657685968254d-15*tz05
     $  + 6.252632224164698d-16*tz06
     $  - 2.252172043902334d-15*tz07
     $  - 2.166132767451374d-15*tz08
     $  + 1.230884088353129d-15*tz09
     $  + 2.527597555609455d-15*tz10
     $  + 3.078548462451323d-16*tz11
     $  - 1.286381333794400d-15*tz12
     $  - 6.511878570762044d-16*tz13
     $  + 1.716895056594641d-16*tz14
     $  + 2.081739968205360d-16*tz15

      CsH4YZ(0,0,2,1,18) = 
     $  + 1.561706626619356d-16*tz04
     $  - 3.255707084988972d-16*tz06
     $  - 1.816842091226466d-16*tz07
     $  + 3.692431881752687d-16*tz08
     $  + 4.276276006974142d-16*tz09
     $  - 1.566342725677645d-16*tz10
     $  - 4.282828128976820d-16*tz11
     $  + 1.950615078115275d-16*tz13
     $  + 1.073746589340491d-16*tz14

      CsH4YZ(0,1,2,1,03) = 
     $  + 8.789062500000000d-03
     $  - 1.171875000000000d-02*tz01
     $  + 2.929687500000000d-03*tz02

      CsH4YZ(0,1,2,1,04) = 
     $  - 3.295898437500000d-03
     $  + 3.707885742187500d-03*tz01
     $  - 4.119873046875000d-04*tz03

      CsH4YZ(0,1,2,1,05) = 
     $  + 4.882049560546875d-04
     $  - 3.666687011718750d-04*tz01
     $  - 2.142333984375000d-04*tz02
     $  + 3.707885742187500d-05*tz03
     $  + 5.561828613281250d-05*tz04

      CsH4YZ(0,1,2,1,06) = 
     $  - 2.027750015258789d-05
     $  - 1.544952392578125d-05*tz01
     $  + 2.703666687011718d-05*tz02
     $  + 2.317428588867187d-05*tz03
     $  - 6.759166717529296d-06*tz04
     $  - 7.724761962890625d-06*tz05

      CsH4YZ(0,1,2,1,07) = 
     $  - 2.882006217022331d-06
     $  + 4.571797896404655d-06*tz01
     $  + 2.773623077236876d-06*tz02
     $  - 3.428848422303491d-06*tz03
     $  - 3.207155636378696d-06*tz04
     $  + 1.064125372439014d-06*tz05
     $  + 1.108463929623973d-06*tz06

      CsH4YZ(0,1,2,1,08) = 
     $  + 2.400517197591917d-07
     $  + 1.685211568006447d-07*tz01
     $  - 5.928550048598221d-07*tz02
     $  - 4.982891758637768d-07*tz03
     $  + 5.164749122091702d-07*tz04
     $  + 4.934396461716720d-07*tz05
     $  - 1.636716271085398d-07*tz06
     $  - 1.636716271085398d-07*tz07

      CsH4YZ(0,1,2,1,09) = 
     $  + 3.050488885492086d-08
     $  - 6.550906359085014d-08*tz01
     $  - 4.450790584087371d-08*tz02
     $  + 1.004436385950871d-07*tz03
     $  + 9.216800598161561d-08*tz04
     $  - 8.318290513541017d-08*tz05
     $  - 7.992743381432124d-08*tz06
     $  + 2.525796714637960d-08*tz07
     $  + 2.475280780345201d-08*tz08

      CsH4YZ(0,1,2,1,10) = 
     $  - 3.146511257260239d-09
     $  - 2.409610065764614d-09*tz01
     $  + 1.150247823846127d-08*tz02
     $  + 1.022947669428374d-08*tz03
     $  - 1.825393285668854d-08*tz04
     $  - 1.732191386898713d-08*tz05
     $  + 1.383631440278674d-08*tz06
     $  + 1.332105187300060d-08*tz07
     $  - 3.938348527299240d-09*tz08
     $  - 3.819004632532596d-09*tz09

      CsH4YZ(0,1,2,1,11) = 
     $  - 3.678830742752569d-10
     $  + 9.678706890021570d-10*tz01
     $  + 7.626392687613159d-10*tz02
     $  - 2.383089200260915d-09*tz03
     $  - 2.278618282016417d-09*tz04
     $  + 3.402968339796300d-09*tz05
     $  + 3.280047102781961d-09*tz06
     $  - 2.344607330480460d-09*tz07
     $  - 2.259890073422084d-09*tz08
     $  + 6.213773033304586d-10*tz09
     $  + 5.991852567829422d-10*tz10

      CsH4YZ(0,1,2,1,12) = 
     $  + 4.213443308539613d-11
     $  + 3.759364975164092d-11*tz01
     $  - 2.100949814121122d-10*tz02
     $  - 2.049205065375574d-10*tz03
     $  + 5.060329672173852d-10*tz04
     $  + 4.976981071425367d-10*tz05
     $  - 6.410940266966002d-10*tz06
     $  - 6.230802346801118d-10*tz07
     $  + 4.021923158151449d-10*tz08
     $  + 3.880654343323509d-10*tz09
     $  - 9.917070800921382d-11*tz10
     $  - 9.535645000885944d-11*tz11

      CsH4YZ(0,1,2,1,13) = 
     $  + 4.677244300487153d-12
     $  - 1.435921298427514d-11*tz01
     $  - 1.330793928830195d-11*tz02
     $  + 5.092473419775348d-11*tz03
     $  + 5.224381516595186d-11*tz04
     $  - 1.074471729365442d-10*tz05
     $  - 1.070398057739598d-10*tz06
     $  + 1.213125885491408d-10*tz07
     $  + 1.184793546972010d-10*tz08
     $  - 6.960644691080629d-11*tz09
     $  - 6.723476084648338d-11*tz10
     $  + 1.599618554882346d-11*tz11
     $  + 1.536141628101300d-11*tz12

      CsH4YZ(0,1,2,1,14) = 
     $  - 5.676322220879943d-13
     $  - 6.013793487738221d-13*tz01
     $  + 3.684658555150029d-12*tz02
     $  + 4.015736548770532d-12*tz03
     $  - 1.231866958732239d-11*tz04
     $  - 1.284212117266750d-11*tz05
     $  + 2.267162596092049d-11*tz06
     $  + 2.272642099057302d-11*tz07
     $  - 2.299451744612677d-11*tz08
     $  - 2.252742490337504d-11*tz09
     $  + 1.212959866699233d-11*tz10
     $  + 1.172962925589747d-11*tz11
     $  - 2.605063927525686d-12*tz12
     $  - 2.500861370424659d-12*tz13

      CsH4YZ(0,1,2,1,15) = 
     $  - 6.116003734569298d-14
     $  + 2.124887005325625d-13*tz01
     $  + 2.320955960959493d-13*tz02
     $  - 1.018312937897316d-12*tz03
     $  - 1.139357100638041d-12*tz04
     $  + 2.928984968821273d-12*tz05
     $  + 3.067911484118057d-12*tz06
     $  - 4.746680413646136d-12*tz07
     $  - 4.772708988332048d-12*tz08
     $  + 4.360304660789725d-12*tz09
     $  + 4.280851976947007d-12*tz10
     $  - 2.125479552601321d-12*tz11
     $  - 2.057768941507290d-12*tz12
     $  + 4.279251678282195d-13*tz13
     $  + 4.109054168350516d-13*tz14

      CsH4YZ(0,1,2,1,16) = 
     $  + 7.669467364439836d-15
     $  + 9.637988919064914d-15*tz01
     $  - 6.273378844996608d-14*tz02
     $  - 7.703180041092209d-14*tz03
     $  + 2.757277903264621d-13*tz04
     $  + 3.087784390732943d-13*tz05
     $  - 6.834649815794781d-13*tz06
     $  - 7.160959833722367d-13*tz07
     $  + 9.860967754222632d-13*tz08
     $  + 9.929319925809077d-13*tz09
     $  - 8.266376581914434d-13*tz10
     $  - 8.128370853437163d-13*tz11
     $  + 3.741828346837774d-13*tz12
     $  + 3.626788316757001d-13*tz13
     $  - 7.084043957605496d-14*tz14
     $  - 6.806238312209202d-14*tz15

      CsH4YZ(0,1,2,1,17) = 
     $  + 8.135835804593466d-16
     $  - 3.131959328578650d-15*tz01
     $  - 4.013138843713496d-15*tz02
     $  + 1.942949579637506d-14*tz03
     $  + 2.391428726112074d-14*tz04
     $  - 7.245165569930209d-14*tz05
     $  - 8.066055685341342d-14*tz06
     $  + 1.567231576890192d-13*tz07
     $  + 1.639536195238541d-13*tz08
     $  - 2.033729522925390d-13*tz09
     $  - 2.049013283155010d-13*tz10
     $  + 1.566370700960009d-13*tz11
     $  + 1.542057704599538d-13*tz12
     $  - 6.613541609546774d-14*tz13
     $  - 6.417406535175288d-14*tz14
     $  + 1.180914309236495d-14*tz15
     $  + 1.135494528112014d-14*tz16

      CsH4YZ(0,1,2,1,18) = 
     $  - 1.038670907484137d-16
     $  - 1.534415355409373d-16*tz01
     $  + 1.044228758637797d-15*tz02
     $  + 1.447494340143570d-15*tz03
     $  - 5.813557541089801d-15*tz04
     $  - 7.048292513115314d-15*tz05
     $  + 1.850174270140059d-14*tz06
     $  + 2.044211212393028d-14*tz07
     $  - 3.538227416415267d-14*tz08
     $  - 3.693095206387973d-14*tz09
     $  + 4.166730968638073d-14*tz10
     $  + 4.198518163979992d-14*tz11
     $  - 2.966217169303930d-14*tz12
     $  - 2.922963384344079d-14*tz13
     $  + 1.172957169451018d-14*tz14
     $  + 1.139404118325404d-14*tz15
     $  - 1.980982351899124d-15*tz16
     $  - 1.906509331151037d-15*tz17

      CsH4YZ(0,1,2,1,19) = 
     $  - 3.578831332227289d-16*tz03
     $  - 4.863499935587250d-16*tz04
     $  + 1.671032328215747d-15*tz05
     $  + 1.990832288610392d-15*tz06
     $  - 4.606664631356450d-15*tz07
     $  - 5.050959864780672d-15*tz08
     $  + 7.879071569579574d-15*tz09
     $  + 8.203500974740059d-15*tz10
     $  - 8.486203167485391d-15*tz11
     $  - 8.549743782776772d-15*tz12
     $  + 5.613397960823839d-15*tz13
     $  + 5.535915895429211d-15*tz14
     $  - 2.086647396849734d-15*tz15
     $  - 2.029059741055195d-15*tz16
     $  + 3.342002593127467d-16*tz17
     $  + 3.219462498046126d-16*tz18

      CsH4YZ(0,1,2,1,20) = 
     $  + 1.171669909536686d-16*tz04
     $  + 1.544120472059325d-16*tz05
     $  - 4.630755390401020d-16*tz06
     $  - 5.427193565027473d-16*tz07
     $  + 1.121970783723494d-15*tz08
     $  + 1.221411238123664d-15*tz09
     $  - 1.733470047909592d-15*tz10
     $  - 1.800379547891108d-15*tz11
     $  + 1.719151637101777d-15*tz12
     $  + 1.731549619684251d-15*tz13
     $  - 1.061616830396656d-15*tz14
     $  - 1.047668014012376d-15*tz15
     $  + 3.722110131810601d-16*tz16
     $  + 3.622960573628451d-16*tz17

      CsH4YZ(0,1,2,1,21) = 
     $  + 1.243038492078133d-16*tz07
     $  + 1.435762225576395d-16*tz08
     $  - 2.680641442327756d-16*tz09
     $  - 2.899359677370465d-16*tz10
     $  + 3.773361809300305d-16*tz11
     $  + 3.909760095420318d-16*tz12
     $  - 3.466073007417583d-16*tz13
     $  - 3.489885048905708d-16*tz14
     $  + 2.006516713942006d-16*tz15
     $  + 1.981312595950246d-16*tz16

      CsH4YZ(0,2,1,1,03) = 
     $  + 8.789062500000000d-03
     $  - 1.171875000000000d-02*tz01
     $  + 2.929687500000000d-03*tz02

      CsH4YZ(0,2,1,1,04) = 
     $  - 3.089904785156250d-03
     $  + 3.604888916015625d-03*tz01
     $  - 2.059936523437500d-04*tz02
     $  - 3.089904785156250d-04*tz03

      CsH4YZ(0,2,1,1,05) = 
     $  + 4.356765747070312d-04
     $  - 3.728485107421875d-04*tz01
     $  - 1.400756835937500d-04*tz02
     $  + 4.325866699218750d-05*tz03
     $  + 3.398895263671875d-05*tz04

      CsH4YZ(0,2,1,1,06) = 
     $  - 1.866817474365234d-05
     $  - 7.402896881103515d-06*tz01
     $  + 2.446174621582031d-05*tz02
     $  + 1.142621040344238d-05*tz03
     $  - 5.793571472167968d-06*tz04
     $  - 4.023313522338867d-06*tz05

      CsH4YZ(0,2,1,1,07) = 
     $  - 2.076522428162243d-06
     $  + 4.024955691123495d-06*tz01
     $  + 6.564569716550865d-07*tz02
     $  - 2.579026076258445d-06*tz03
     $  - 1.293207917894635d-06*tz04
     $  + 7.611452316751285d-07*tz05
     $  + 5.061985278616146d-07*tz06

      CsH4YZ(0,2,1,1,08) = 
     $  + 1.863027656716959d-07
     $  + 3.586631334785904d-09*tz01
     $  - 4.148368524121386d-07*tz02
     $  - 1.088618384008961d-07*tz03
     $  + 3.309804014861583d-07*tz04
     $  + 1.721077881354306d-07*tz05
     $  - 1.024463147457156d-07*tz06
     $  - 6.683258106932044d-08*tz07

      CsH4YZ(0,2,1,1,09) = 
     $  + 1.907313292446945d-08
     $  - 4.882870208738105d-08*tz01
     $  + 4.310693059648786d-10*tz02
     $  + 5.850530994523848d-08*tz03
     $  + 1.928361598402261d-08*tz04
     $  - 4.686194837891629d-08*tz05
     $  - 2.496609730379922d-08*tz06
     $  + 1.419497753626533d-08*tz07
     $  + 9.168642074135797d-09*tz08

      CsH4YZ(0,2,1,1,10) = 
     $  - 2.148000671045987d-09
     $  + 4.576743646923984d-10*tz01
     $  + 6.428657798096537d-09*tz02
     $  + 5.194300943652966d-10*tz03
     $  - 9.276998753193765d-09*tz04
     $  - 3.486736074721972d-09*tz05
     $  + 7.016663273264254d-09*tz06
     $  + 3.807070243055932d-09*tz07
     $  - 2.020321647121038d-09*tz08
     $  - 1.297438627391654d-09*tz09

      CsH4YZ(0,2,1,1,11) = 
     $  - 2.069613822468379d-10
     $  + 6.184457613791200d-10*tz01
     $  - 9.568216330226793d-11*tz02
     $  - 1.115512378572776d-09*tz03
     $  - 1.876813079447099d-10*tz04
     $  + 1.555674204804468d-09*tz05
     $  + 6.366485233686900d-10*tz06
     $  - 1.088449432214257d-09*tz07
     $  - 5.991852567829422d-10*tz08
     $  + 2.943616459909851d-10*tz09
     $  + 1.883417855205288d-10*tz10

      CsH4YZ(0,2,1,1,12) = 
     $  + 2.611522219254978d-11
     $  - 9.645801041019635d-12*tz01
     $  - 9.670926708799747d-11*tz02
     $  + 8.526413860668721d-12*tz03
     $  + 2.102322139750879d-10*tz04
     $  + 4.949613602095664d-11*tz05
     $  - 2.688779443249811d-10*tz06
     $  - 1.168154352469642d-10*tz07
     $  + 1.730038450160735d-10*tz08
     $  + 9.636828789506456d-11*tz09
     $  - 4.376406977073273d-11*tz10
     $  - 2.792960148870601d-11*tz11

      CsH4YZ(0,2,1,1,13) = 
     $  + 2.410018451372826d-12
     $  - 8.001571141537849d-12*tz01
     $  + 2.266757931350202d-12*tz02
     $  + 1.972477252562068d-11*tz03
     $  + 2.328716551029967d-13*tz04
     $  - 4.083435514831227d-11*tz05
     $  - 1.168557114023208d-11*tz06
     $  + 4.730440971847290d-11*tz07
     $  + 2.148436033550200d-11*tz08
     $  - 2.799362698152097d-11*tz09
     $  - 1.574634842557539d-11*tz10
     $  + 6.621046491369590d-12*tz11
     $  + 4.217235728387380d-12*tz12

      CsH4YZ(0,2,1,1,14) = 
     $  - 3.260414383073448d-13
     $  + 1.575453410005958d-13*tz01
     $  + 1.432953453218451d-12*tz02
     $  - 3.559529760334336d-13*tz03
     $  - 4.287759532212415d-12*tz04
     $  - 4.126322837464606d-13*tz05
     $  + 8.021014192600762d-12*tz06
     $  + 2.608824776076860d-12*tz07
     $  - 8.415630072716981d-12*tz08
     $  - 3.955503138042638d-12*tz09
     $  + 4.592420710162252d-12*tz10
     $  + 2.604443256530608d-12*tz11
     $  - 1.016957312744724d-12*tz12
     $  - 6.467249757855328d-13*tz13

      CsH4YZ(0,2,1,1,15) = 
     $  - 2.918158631107126d-14
     $  + 1.048806338979654d-13*tz01
     $  - 4.193098098204412d-14*tz02
     $  - 3.341058300668208d-13*tz03
     $  + 4.723538687705150d-14*tz04
     $  + 9.444852971129579d-13*tz05
     $  + 1.535237764436896d-13*tz06
     $  - 1.580844285199959d-12*tz07
     $  - 5.629433799020619d-13*tz08
     $  + 1.508286169590253d-12*tz09
     $  + 7.285456681681990d-13*tz10
     $  - 7.617461290596550d-13*tz11
     $  - 4.349974924472660d-13*tz12
     $  + 1.582747375522651d-13*tz13
     $  + 1.005180143264964d-13*tz14

      CsH4YZ(0,2,1,1,16) = 
     $  + 4.139204183440739d-15
     $  - 2.364924201195800d-15*tz01
     $  - 2.103658868975437d-14*tz02
     $  + 8.462406010983724d-15*tz03
     $  + 8.203446149375147d-14*tz04
     $  - 3.476490038236318d-15*tz05
     $  - 2.075169137865607d-13*tz06
     $  - 4.454856442406177d-14*tz07
     $  + 3.115167391105161d-13*tz08
     $  + 1.186946172498020d-13*tz09
     $  - 2.717198310170282d-13*tz10
     $  - 1.342009565392925d-13*tz11
     $  + 1.275035814149329d-13*tz12
     $  + 7.324170342884683d-14*tz13
     $  - 2.492065270929788d-14*tz14
     $  - 1.580779148684614d-14*tz15

      CsH4YZ(0,2,1,1,17) = 
     $  + 3.623920920707411d-16
     $  - 1.387338501483296d-15*tz01
     $  + 7.040948582448130d-16*tz02
     $  + 5.502497158731812d-15*tz03
     $  - 1.607737059744051d-15*tz04
     $  - 2.016383915800803d-14*tz05
     $  - 7.709045353368337d-16*tz06
     $  + 4.524207538867477d-14*tz07
     $  + 1.159630952676270d-14*tz08
     $  - 6.128003498630586d-14*tz09
     $  - 2.460641538990252d-14*tz10
     $  + 4.913513567101960d-14*tz11
     $  + 2.471974884968986d-14*tz12
     $  - 2.150590728843284d-14*tz13
     $  - 1.241625884288022d-14*tz14
     $  + 3.964294973676582d-15*tz15
     $  + 2.511887243222774d-15*tz16

      CsH4YZ(0,2,1,1,18) = 
     $  + 3.069223755843971d-16*tz02
     $  - 1.692191188015457d-16*tz03
     $  - 1.502918777975408d-15*tz04
     $  + 2.715034610342493d-16*tz05
     $  + 4.892037834468283d-15*tz06
     $  + 5.029314129138871d-16*tz07
     $  - 9.772901954493906d-15*tz08
     $  - 2.836061314541752d-15*tz09
     $  + 1.202568769580537d-14*tz10
     $  + 5.035336288053266d-15*tz11
     $  - 8.910485690020899d-15*tz12
     $  - 4.553103169460134d-15*tz13
     $  + 3.651270105910463d-15*tz14
     $  + 2.117261691806802d-15*tz15
     $  - 6.364184614312354d-16*tz16
     $  - 4.028369607302454d-16*tz17

      CsH4YZ(0,2,1,1,19) = 
     $  + 4.071279663812239d-16*tz05
     $  - 1.167406755896140d-15*tz07
     $  - 1.809226905102290d-16*tz08
     $  + 2.091594683899685d-15*tz09
     $  + 6.656125045358983d-16*tz10
     $  - 2.353727442151499d-15*tz11
     $  - 1.019847745817260d-15*tz12
     $  + 1.619507455494178d-15*tz13
     $  + 8.385916114553002d-16*tz14
     $  - 6.234633067366739d-16*tz15
     $  - 3.629023033198404d-16*tz16
     $  + 1.030090996682154d-16*tz17

      CsH4YZ(0,2,1,1,20) = 
     $  - 1.079567839865419d-16*tz06
     $  + 2.739963943948909d-16*tz08
     $  - 4.437340632700698d-16*tz10
     $  - 1.516603822047695d-16*tz11
     $  + 4.594690668475248d-16*tz12
     $  + 2.048344244408099d-16*tz13
     $  - 2.948843459975532d-16*tz14
     $  - 1.544493763574595d-16*tz15
     $  + 1.069946193969759d-16*tz16

      CsH4YZ(0,2,0,1,02) = 
     $  + 7.031250000000000d-02
     $  - 7.031250000000000d-02*tz01

      CsH4YZ(0,2,0,1,03) = 
     $  - 2.490234375000000d-02
     $  + 2.343750000000000d-02*tz01
     $  + 1.464843750000000d-03*tz02

      CsH4YZ(0,2,0,1,04) = 
     $  + 4.188537597656250d-03
     $  - 3.501892089843750d-03*tz01
     $  - 6.179809570312500d-04*tz02
     $  - 6.866455078125000d-05*tz03

      CsH4YZ(0,2,0,1,05) = 
     $  - 3.610038757324218d-04
     $  + 2.142333984375000d-04*tz01
     $  + 1.091766357421875d-04*tz02
     $  + 3.295898437500000d-05*tz03
     $  + 4.634857177734375d-06*tz04

      CsH4YZ(0,2,0,1,06) = 
     $  + 7.091760635375976d-06
     $  + 9.012222290039062d-06*tz01
     $  - 7.081031799316406d-06*tz02
     $  - 6.222724914550781d-06*tz03
     $  - 2.413988113403320d-06*tz04
     $  - 3.862380981445312d-07*tz05

      CsH4YZ(0,2,0,1,07) = 
     $  + 1.400605756409314d-06
     $  - 1.872072414476044d-06*tz01
     $  - 5.788644965814084d-07*tz02
     $  + 3.350024320641342d-07*tz03
     $  + 4.655548504420689d-07*tz04
     $  + 2.128250744878029d-07*tz05
     $  + 3.694879765413245d-08*tz06

      CsH4YZ(0,2,0,1,08) = 
     $  - 5.210165352541573d-08
     $  - 4.263544854308877d-08*tz01
     $  + 1.117412466555833d-07*tz02
     $  + 6.472775047378880d-08*tz03
     $  - 1.576097149934087d-08*tz04
     $  - 4.085728765598365d-08*tz05
     $  - 2.121669240295886d-08*tz06
     $  - 3.896943502584282d-09*tz07

      CsH4YZ(0,2,0,1,09) = 
     $  - 1.163426067974722d-08
     $  + 1.972142074789319d-08*tz01
     $  + 2.498854883015155d-09*tz02
     $  - 9.573330836636679d-09*tz03
     $  - 8.018562636737312d-09*tz04
     $  + 2.694183162280491d-10*tz05
     $  + 3.985145927539893d-09*tz06
     $  + 2.309299853383278d-09*tz07
     $  + 4.420144250616431d-10*tz08

      CsH4YZ(0,2,0,1,10) = 
     $  + 5.287998981063482d-10
     $  + 1.757954513388020d-10*tz01
     $  - 1.522297679912298d-09*tz02
     $  - 5.223347605871302d-10*tz03
     $  + 9.286091621366462d-10*tz04
     $  + 1.042042692591037d-09*tz05
     $  + 1.098721570867512d-10*tz06
     $  - 4.189214265278103d-10*tz07
     $  - 2.685237632249481d-10*tz08
     $  - 5.304173100739717d-11*tz09

      CsH4YZ(0,2,0,1,11) = 
     $  + 1.181014141358660d-10
     $  - 2.256152259585598d-10*tz01
     $  + 9.596983797973573d-12*tz02
     $  + 1.744521867223809d-10*tz03
     $  + 1.002399478180673d-10*tz04
     $  - 9.390640818434761d-11*tz05
     $  - 1.400603854855156d-10*tz06
     $  - 2.888488598248163d-11*tz07
     $  + 4.654165317604129d-11*tz08
     $  + 3.287710599632056d-11*tz09
     $  + 6.657613964254913d-12*tz10

      CsH4YZ(0,2,0,1,12) = 
     $  - 6.115217431439836d-12
     $  + 4.534056247334833d-13*tz01
     $  + 2.030907390312145d-11*tz02
     $  + 2.128001670979602d-12*tz03
     $  - 2.258686536506147d-11*tz04
     $  - 1.795284461351982d-11*tz05
     $  + 9.374194926796870d-12*tz06
     $  + 1.933508789068528d-11*tz07
     $  + 5.509483778289656d-12*tz08
     $  - 5.391759765933040d-12*tz09
     $  - 4.195683800389815d-12*tz10
     $  - 8.668768182623585d-13*tz11

      CsH4YZ(0,2,0,1,13) = 
     $  - 1.304806077115131d-12
     $  + 2.699134075481980d-12*tz01
     $  - 5.499925401650592d-13*tz02
     $  - 2.834906354609212d-12*tz03
     $  - 9.503881908000021d-13*tz04
     $  + 3.057875321257147d-12*tz05
     $  + 3.110534675119151d-12*tz06
     $  - 8.587185523480270d-13*tz07
     $  - 2.728567452669681d-12*tz08
     $  - 9.554452979191439d-13*tz09
     $  + 6.449255760954221d-13*tz10
     $  + 5.539804519073060d-13*tz11
     $  + 1.163743657652500d-13*tz12

      CsH4YZ(0,2,0,1,14) = 
     $  + 7.485022971179719d-14
     $  - 2.988208485930866d-14*tz01
     $  - 2.704474277098033d-13*tz02
     $  + 3.961957215442377d-14*tz03
     $  + 4.475920123944927d-13*tz04
     $  + 2.479506323027986d-13*tz05
     $  - 4.197413325596020d-13*tz06
     $  - 5.304173890897042d-13*tz07
     $  + 5.900367464490503d-14*tz08
     $  + 3.922309515855869d-13*tz09
     $  + 1.595667440052097d-13*tz10
     $  - 7.903701759912644d-14*tz11
     $  - 7.525740235074206d-14*tz12
     $  - 1.603116263092730d-14*tz13

      CsH4YZ(0,2,0,1,15) = 
     $  + 1.511651534198210d-14
     $  - 3.321321040372682d-14*tz01
     $  + 1.196679110521171d-14*tz02
     $  + 4.388938762872158d-14*tz03
     $  + 3.845073057582133d-15*tz04
     $  - 7.335408852337506d-14*tz05
     $  - 5.527873968117309d-14*tz06
     $  + 5.739248068413262d-14*tz07
     $  + 8.976724783085618d-14*tz08
     $  + 7.793556010339473d-16*tz09
     $  - 5.727330405640291d-14*tz10
     $  - 2.622957295916772d-14*tz11
     $  + 9.860649384930563d-15*tz12
     $  + 1.047369291887250d-14*tz13
     $  + 2.257722070522262d-15*tz14

      CsH4YZ(0,2,0,1,16) = 
     $  - 9.429026362141077d-16
     $  + 6.255185677713400d-16*tz01
     $  + 3.610503183832362d-15*tz02
     $  - 1.495455691593901d-15*tz03
     $  - 8.091117089321963d-15*tz04
     $  - 2.624278398280481d-15*tz05
     $  + 1.212543928571529d-14*tz06
     $  + 1.146221495825607d-14*tz07
     $  - 7.696197319452367d-15*tz08
     $  - 1.514303257995349d-14*tz09
     $  - 1.534752074073185d-15*tz10
     $  + 8.475799809416384d-15*tz11
     $  + 4.285805042224876d-15*tz12
     $  - 1.245193943122441d-15*tz13
     $  - 1.488244528908717d-15*tz14
     $  - 3.241065862956762d-16*tz15

      CsH4YZ(0,2,0,1,17) = 
     $  - 1.807379032678685d-16
     $  + 4.168143535262131d-16*tz01
     $  - 2.142422727095192d-16*tz02
     $  - 6.623993026321121d-16*tz03
     $  + 1.174517116005122d-16*tz04
     $  + 1.538584004814289d-15*tz05
     $  + 8.183132507778372d-16*tz06
     $  - 1.997889740678660d-15*tz07
     $  - 2.282432338441684d-15*tz08
     $  + 9.917256063164782d-16*tz09
     $  + 2.552207305147787d-15*tz10
     $  + 4.500693004113501d-16*tz11
     $  - 1.268859805931058d-15*tz12
     $  - 6.995052982296888d-16*tz13
     $  + 1.582654504791046d-16*tz14
     $  + 2.153234068123524d-16*tz15

      CsH4YZ(0,2,0,1,18) = 
     $  + 1.388774179331762d-16*tz04
     $  - 2.929735676721654d-16*tz06
     $  - 2.098568109786273d-16*tz07
     $  + 3.260397578885936d-16*tz08
     $  + 4.431747573993432d-16*tz09
     $  - 1.183890394165998d-16*tz10
     $  - 4.303158588470868d-16*tz11
     $  - 1.029819114256702d-16*tz12
     $  + 1.918529096235428d-16*tz13
     $  + 1.143338185199208d-16*tz14

      CsH4YZ(0,2,2,1,03) = 
     $  + 1.171875000000000d-02
     $  - 1.171875000000000d-02*tz01

      CsH4YZ(0,2,2,1,04) = 
     $  - 3.707885742187500d-03
     $  + 3.295898437500000d-03*tz01
     $  + 4.119873046875000d-04*tz02

      CsH4YZ(0,2,2,1,05) = 
     $  + 4.531860351562500d-04
     $  - 3.048706054687500d-04*tz01
     $  - 1.235961914062500d-04*tz02
     $  - 2.471923828125000d-05*tz03

      CsH4YZ(0,2,2,1,06) = 
     $  - 1.223087310791015d-05
     $  - 7.724761962890625d-06*tz01
     $  + 1.029968261718750d-05*tz02
     $  + 7.724761962890625d-06*tz03
     $  + 1.931190490722656d-06*tz04

      CsH4YZ(0,2,2,1,07) = 
     $  - 2.413988113403320d-06
     $  + 2.916491761499521d-06*tz01
     $  + 8.276530674525669d-07*tz02
     $  - 5.320626862195073d-07*tz03
     $  - 6.207398005894252d-07*tz04
     $  - 1.773542287398357d-07*tz05

      CsH4YZ(0,2,2,1,08) = 
     $  + 1.107309279697281d-07
     $  + 3.879623753683907d-08*tz01
     $  - 1.604386073138032d-07*tz02
     $  - 9.699059384209769d-08*tz03
     $  + 3.152194299868174d-08*tz04
     $  + 5.819435630525861d-08*tz05
     $  + 1.818573634539331d-08*tz06

      CsH4YZ(0,2,2,1,09) = 
     $  + 2.379861793347767d-08
     $  - 3.257267443197114d-08*tz01
     $  - 6.870167063815253d-09*tz02
     $  + 1.335416787437030d-08*tz03
     $  + 1.212382423026221d-08*tz04
     $  - 1.751219055482319d-09*tz05
     $  - 6.061912115131105d-09*tz06
     $  - 2.020637371710368d-09*tz07

      CsH4YZ(0,2,2,1,10) = 
     $  - 1.217686596126960d-09
     $  - 2.576312648930719d-10*tz01
     $  + 2.233814614425812d-09*tz02
     $  + 1.166918082162737d-09*tz03
     $  - 1.300280148695622d-09*tz04
     $  - 1.591251930221915d-09*tz05
     $  + 4.546434086348329d-11*tz06
     $  + 6.819651129522493d-10*tz07
     $  + 2.386877895332872d-10*tz08

      CsH4YZ(0,2,2,1,11) = 
     $  - 2.689112434027619d-10
     $  + 3.871670625185750d-10*tz01
     $  + 6.406338939854463d-11*tz02
     $  - 2.400654968321807d-10*tz03
     $  - 1.952900096181441d-10*tz04
     $  + 1.357354804705234d-10*tz05
     $  + 2.169888995757157d-10*tz06
     $  + 1.127215062730990d-11*tz07
     $  - 8.137083734089339d-11*tz08
     $  - 2.958939539668850d-11*tz09

      CsH4YZ(0,2,2,1,12) = 
     $  + 1.439058327047281d-11
     $  + 2.001980918704519d-12*tz01
     $  - 3.046293814604014d-11*tz02
     $  - 1.420760651983852d-11*tz03
     $  + 3.015887383983905d-11*tz04
     $  + 3.254833493635735d-11*tz05
     $  - 1.451032540875554d-11*tz06
     $  - 3.051406400283502d-11*tz07
     $  - 3.390451555870558d-12*tz08
     $  + 1.017135466761167d-11*tz09
     $  + 3.814258000354377d-12*tz10

      CsH4YZ(0,2,2,1,13) = 
     $  + 3.208199260696930d-12
     $  - 4.754641220989632d-12*tz01
     $  - 6.436269299743284d-13*tz02
     $  + 3.909696420735806d-12*tz03
     $  + 2.961987897711091d-12*tz04
     $  - 4.043756278219224d-12*tz05
     $  - 5.427982538965845d-12*tz06
     $  + 1.528819950832376d-12*tz07
     $  + 4.401066923485820d-12*tz08
     $  + 6.883720059811155d-13*tz09
     $  - 1.320320077045746d-12*tz10
     $  - 5.078154142483639d-13*tz11

      CsH4YZ(0,2,2,1,14) = 
     $  - 1.761535827804077d-13
     $  - 1.724000705288379d-14*tz01
     $  + 4.132664095744155d-13*tz02
     $  + 1.759179301445494d-13*tz03
     $  - 5.893688164673032d-13*tz04
     $  - 5.948550914014210d-13*tz05
     $  + 5.578355889712417d-13*tz06
     $  + 9.077200529689504d-13*tz07
     $  - 1.504259226101297d-13*tz08
     $  - 6.483714664063932d-13*tz09
     $  - 1.246220480885015d-13*tz10
     $  + 1.768285817471981d-13*tz11
     $  + 6.946837140068498d-14*tz12

      CsH4YZ(0,2,2,1,15) = 
     $  - 3.944609391516773d-14
     $  + 5.961399833887199d-14*tz01
     $  + 6.812422468581651d-15*tz02
     $  - 6.086600568496409d-14*tz03
     $  - 4.377536559300942d-14*tz04
     $  + 9.425595457553070d-14*tz05
     $  + 1.166039478473719d-13*tz06
     $  - 7.769105638542330d-14*tz07
     $  - 1.523672946055024d-13*tz08
     $  + 1.212913086651151d-14*tz09
     $  + 9.725571996095898d-14*tz10
     $  + 2.151414411257577d-14*tz11
     $  - 2.431392999023974d-14*tz12
     $  - 9.725571996095898d-15*tz13

      CsH4YZ(0,2,2,1,16) = 
     $  + 2.203284530586127d-15
     $  + 1.596661516073824d-16*tz01
     $  - 5.602195595231796d-15*tz02
     $  - 2.211636413779699d-15*tz03
     $  + 1.060491949306715d-14*tz04
     $  + 1.025068534519065d-14*tz05
     $  - 1.540682724309426d-14*tz06
     $  - 2.248314895228419d-14*tz07
     $  + 1.078287874492178d-14*tz08
     $  + 2.568158855219073d-14*tz09
     $  - 3.382451020481683d-16*tz10
     $  - 1.481630108780234d-14*tz11
     $  - 3.632843055182305d-15*tz12
     $  + 3.419146404877464d-15*tz13
     $  + 1.389028226981469d-15*tz14

      CsH4YZ(0,2,2,1,17) = 
     $  + 4.948777658582265d-16
     $  - 7.586885372772812d-16*tz01
     $  + 9.241291326846716d-16*tz03
     $  + 6.384742937569013d-16*tz04
     $  - 1.951881826743484d-15*tz05
     $  - 2.301126500646426d-15*tz06
     $  + 2.533344160400920d-15*tz07
     $  + 4.285106687158521d-15*tz08
     $  - 1.473031254134319d-15*tz09
     $  - 4.346841275024364d-15*tz10
     $  - 1.721986743832134d-16*tz11
     $  + 2.287811197381244d-15*tz12
     $  + 6.070759512120238d-16*tz13
     $  - 4.902452565816952d-16*tz14
     $  - 2.018656938865803d-16*tz15

      CsH4YZ(0,2,2,1,18) = 
     $  - 1.818017781179187d-16*tz04
     $  - 1.704235018636058d-16*tz05
     $  + 3.647578934923375d-16*tz06
     $  + 5.009126302202784d-16*tz07
     $  - 4.156666845795445d-16*tz08
     $  - 8.098881900782480d-16*tz09
     $  + 1.949029631668466d-16*tz10
     $  + 7.387723658210268d-16*tz11
     $  - 3.574704995908194d-16*tz13
     $  - 1.009996014716918d-16*tz14

      CsH4YZ(0,2,2,1,19) = 
     $  - 1.065333762063767d-16*tz08
     $  + 1.521265434348900d-16*tz10
     $  - 1.260553866978152d-16*tz12

      CsH4YZ(2,0,0,1,02) = 
     $  + 1.406250000000000d-01
     $  - 1.406250000000000d-01*tz01

      CsH4YZ(2,0,0,1,03) = 
     $  - 2.856445312500000d-02
     $  + 2.636718750000000d-02*tz01
     $  + 2.197265625000000d-03*tz02

      CsH4YZ(2,0,0,1,04) = 
     $  + 3.799438476562500d-03
     $  - 3.021240234375000d-03*tz01
     $  - 6.866455078125000d-04*tz02
     $  - 9.155273437500000d-05*tz03

      CsH4YZ(2,0,0,1,05) = 
     $  - 3.122091293334960d-04
     $  + 1.699447631835937d-04*tz01
     $  + 1.004219055175781d-04*tz02
     $  + 3.604888916015625d-05*tz03
     $  + 5.793571472167968d-06*tz04

      CsH4YZ(2,0,0,1,06) = 
     $  + 1.006579399108886d-05
     $  + 4.377365112304687d-06*tz01
     $  - 5.536079406738281d-06*tz02
     $  - 5.836486816406250d-06*tz03
     $  - 2.607107162475585d-06*tz04
     $  - 4.634857177734375d-07*tz05

      CsH4YZ(2,0,0,1,07) = 
     $  + 7.810975824083600d-07
     $  - 1.241479601178850d-06*tz01
     $  - 4.914190087999616d-07*tz02
     $  + 2.413988113403320d-07*tz03
     $  + 4.396906920841761d-07*tz04
     $  + 2.276045935494559d-07*tz05
     $  + 4.310693059648786d-08*tz06

      CsH4YZ(2,0,0,1,08) = 
     $  - 7.497086990678970d-08
     $  + 3.925809750751573d-09*tz01
     $  + 8.440490964115882d-08*tz02
     $  + 6.058063282042133d-08*tz03
     $  - 8.313479472179802d-09*tz04
     $  - 3.865767954563607d-08*tz05
     $  - 2.251567357048696d-08*tz06
     $  - 4.453649717239179d-09*tz07

      CsH4YZ(2,0,0,1,09) = 
     $  - 3.966327739715557d-09
     $  + 1.108824757726064d-08*tz01
     $  + 1.856460585258901d-09*tz02
     $  - 7.552132176767502d-09*tz03
     $  - 7.697365487859185d-09*tz04
     $  - 4.293854414884533d-10*tz05
     $  + 3.767646766001624d-09*tz06
     $  + 2.435589689115176d-09*tz07
     $  + 4.972662281943485d-10*tz08

      CsH4YZ(2,0,0,1,10) = 
     $  + 6.927277131673727d-10
     $  - 2.934975782409310d-10*tz01
     $  - 1.086345166965786d-09*tz02
     $  - 5.053277293752346d-10*tz03
     $  + 7.361434524812336d-10*tz04
     $  + 1.009005271563572d-09*tz05
     $  + 1.831202618112521d-10*tz06
     $  - 3.951067717897952d-10*tz07
     $  - 2.817841959767974d-10*tz08
     $  - 5.893525667488574d-11*tz09

      CsH4YZ(2,0,0,1,11) = 
     $  + 2.378374858097514d-11
     $  - 1.097860503805704d-10*tz01
     $  + 1.033280474170074d-11*tz02
     $  + 1.368235561213540d-10*tz03
     $  + 9.938866560923407d-11*tz04
     $  - 7.269128135786476d-11*tz05
     $  - 1.360701224422717d-10*tz06
     $  - 3.719809707012269d-11*tz07
     $  + 4.373682507073019d-11*tz08
     $  + 3.435657576615498d-11*tz09
     $  + 7.323375360680405d-12*tz10

      CsH4YZ(2,0,0,1,12) = 
     $  - 7.070325190279722d-12
     $  + 5.145364936953919d-12*tz01
     $  + 1.372888258803198d-11*tz02
     $  + 2.701760987306759d-12*tz03
     $  - 1.843613073302600d-11*tz04
     $  - 1.783421550226301d-11*tz05
     $  + 6.804753690867911d-12*tz06
     $  + 1.880631268861595d-11*tz07
     $  + 6.511208101615048d-12*tz08
     $  - 5.042868602039794d-12*tz09
     $  - 4.369059164042287d-12*tz10
     $  - 9.456838017407548d-13*tz11

      CsH4YZ(2,0,0,1,13) = 
     $  - 1.400829012319348d-13
     $  + 1.158598332158925d-12*tz01
     $  - 4.509929293154054d-13*tz02
     $  - 2.189394959035322d-12*tz03
     $  - 1.035858921116539d-12*tz04
     $  + 2.533815315406353d-12*tz05
     $  + 3.085937699182721d-12*tz06
     $  - 5.265562212027678d-13*tz07
     $  - 2.654393488227385d-12*tz08
     $  - 1.081928952023597d-12*tz09
     $  + 5.996453683249430d-13*tz10
     $  + 5.751394275009879d-13*tz11
     $  + 1.260722295790209d-13*tz12

      CsH4YZ(2,0,0,1,14) = 
     $  + 7.645204879403038d-14
     $  - 7.589579464474274d-14*tz01
     $  - 1.741963466940044d-13*tz02
     $  + 1.787141622997981d-14*tz03
     $  + 3.668442379999272d-13*tz04
     $  + 2.587679628871018d-13*tz05
     $  - 3.478805986560531d-13*tz06
     $  - 5.253549213181313d-13*tz07
     $  + 1.396744281749592d-14*tz08
     $  + 3.814269195579559d-13*tz09
     $  + 1.761484718174012d-13*tz10
     $  - 7.295724701457825d-14*tz11
     $  - 7.792926278922995d-14*tz12
     $  - 1.726432898715248d-14*tz13

      CsH4YZ(2,0,0,1,15) = 
     $  + 6.136355100326801d-16
     $  - 1.277531926400745d-14*tz01
     $  + 9.176549656802100d-15*tz02
     $  + 3.335643531760230d-14*tz03
     $  + 7.067042371950660d-15*tz04
     $  - 6.187916623371385d-14*tz05
     $  - 5.656733725468802d-14*tz06
     $  + 4.698708320834902d-14*tz07
     $  + 8.877671929610870d-14*tz08
     $  + 7.111502909314566d-15*tz09
     $  - 5.565690136003879d-14*tz10
     $  - 2.847150676346255d-14*tz11
     $  + 9.021239897172286d-15*tz12
     $  + 1.082103477587593d-14*tz13
     $  + 2.418987932702423d-15*tz14

      CsH4YZ(2,0,0,1,16) = 
     $  - 8.593535096200683d-16
     $  + 1.059945131355364d-15*tz01
     $  + 2.228616413914301d-15*tz02
     $  - 9.753215674682116d-16*tz03
     $  - 6.619972518174398d-15*tz04
     $  - 3.089097450491706d-15*tz05
     $  + 1.036997781793154d-14*tz06
     $  + 1.160431433055664d-14*tz07
     $  - 6.130927317088546d-15*tz08
     $  - 1.495633159848137d-14*tz09
     $  - 2.450906309314436d-15*tz10
     $  + 8.229028255154811d-15*tz11
     $  + 4.596852388779701d-15*tz12
     $  - 1.126564905196805d-15*tz13
     $  - 1.534545469808099d-15*tz14
     $  - 3.457136920487213d-16*tz15

      CsH4YZ(2,0,0,1,17) = 
     $  + 1.454887014682364d-16*tz01
     $  - 1.573159882028266d-16*tz02
     $  - 4.960557059371965d-16*tz03
     $  + 1.306359864039059d-15*tz05
     $  + 8.855702600789999d-16*tz06
     $  - 1.716717624180074d-15*tz07
     $  - 2.295627318225472d-15*tz08
     $  + 7.495784405352890d-16*tz09
     $  + 2.517851975070307d-15*tz10
     $  + 5.857254722732690d-16*tz11
     $  - 1.230578009504110d-15*tz12
     $  - 7.436172040938694d-16*tz13
     $  + 1.411643707567827d-16*tz14
     $  + 2.216317097463080d-16*tz15

      CsH4YZ(2,0,0,1,18) = 
     $  + 1.132463024771931d-16*tz04
     $  - 2.537863826787251d-16*tz06
     $  - 2.195900153553105d-16*tz07
     $  + 2.796162614843177d-16*tz08
     $  + 4.438024759969328d-16*tz09
     $  - 4.240890787669516d-16*tz11
     $  - 1.234567234446574d-16*tz12
     $  + 1.858357196329718d-16*tz13
     $  + 1.207088759822781d-16*tz14

      CsH4YZ(2,0,1,1,03) = 
     $  + 1.318359375000000d-02
     $  - 1.757812500000000d-02*tz01
     $  + 4.394531250000000d-03*tz02

      CsH4YZ(2,0,1,1,04) = 
     $  - 3.295898437500000d-03
     $  + 3.707885742187500d-03*tz01
     $  - 4.119873046875000d-04*tz03

      CsH4YZ(2,0,1,1,05) = 
     $  + 3.746509552001953d-04
     $  - 2.780914306640625d-04*tz01
     $  - 1.699447631835937d-04*tz02
     $  + 3.089904785156250d-05*tz03
     $  + 4.248619079589843d-05*tz04

      CsH4YZ(2,0,1,1,06) = 
     $  - 1.448392868041992d-05
     $  - 9.655952453613281d-06*tz01
     $  + 1.931190490722656d-05*tz02
     $  + 1.448392868041992d-05*tz03
     $  - 4.827976226806640d-06*tz04
     $  - 4.827976226806640d-06*tz05

      CsH4YZ(2,0,1,1,07) = 
     $  - 1.267343759536743d-06
     $  + 2.586415835789271d-06*tz01
     $  + 1.202683363642011d-06*tz02
     $  - 2.155346529824393d-06*tz03
     $  - 1.629441976547241d-06*tz04
     $  + 6.724681173052106d-07*tz05
     $  + 5.905649491718837d-07*tz06

      CsH4YZ(2,0,1,1,08) = 
     $  + 1.188134774565696d-07
     $  + 4.243338480591773d-08*tz01
     $  - 3.140070475637912d-07*tz02
     $  - 1.782202161848545d-07*tz03
     $  + 2.885470166802406d-07*tz04
     $  + 2.121669240295886d-07*tz05
     $  - 9.335344657301902d-08*tz06
     $  - 7.638009265065193d-08*tz07

      CsH4YZ(2,0,1,1,09) = 
     $  + 8.616124042808743d-09
     $  - 2.793531166389584d-08*tz01
     $  - 8.815030534086482d-09*tz02
     $  + 4.820903996005654d-08*tz03
     $  + 2.855413185898214d-08*tz04
     $  - 4.207977326586842d-08*tz05
     $  - 3.004856158180960d-08*tz06
     $  + 1.318465885041015d-08*tz07
     $  + 1.031472233340277d-08*tz08

      CsH4YZ(2,0,1,1,10) = 
     $  - 1.156499170714856d-09
     $  - 1.477591078063207d-10*tz01
     $  + 4.660094938507037d-09*tz02
     $  + 1.901924926122384d-09*tz03
     $  - 8.035822247620671d-09*tz04
     $  - 4.796487961097487d-09*tz05
     $  + 6.433204232182885d-09*tz06
     $  + 4.483920617661039d-09*tz07
     $  - 1.900977752354395d-09*tz08
     $  - 1.441598474879616d-09*tz09

      CsH4YZ(2,0,1,1,11) = 
     $  - 7.294960147640728d-11
     $  + 3.227709881188771d-10*tz01
     $  + 5.311707437530540d-11*tz02
     $  - 9.012788935948477d-10*tz03
     $  - 4.034960251634142d-10*tz04
     $  + 1.391363822493714d-09*tz05
     $  + 8.296112057067077d-10*tz06
     $  - 1.013066924894122d-09*tz07
     $  - 6.928145579310351d-10*tz08
     $  + 2.795669482926408d-10*tz09
     $  + 2.071759640725817d-10*tz10

      CsH4YZ(2,0,1,1,12) = 
     $  + 1.218746246779898d-11
     $  - 6.881809408046787d-13*tz01
     $  - 6.755272912479475d-11*tz02
     $  - 1.698252966824449d-11*tz03
     $  + 1.801702101797023d-10*tz04
     $  + 8.456855787082011d-11*tz05
     $  - 2.457804931006129d-10*tz06
     $  - 1.461496523802452d-10*tz07
     $  + 1.628324903484618d-10*tz08
     $  + 1.097204612879717d-10*tz09
     $  - 4.185694077055554d-11*tz10
     $  - 3.046865616949747d-11*tz11

      CsH4YZ(2,0,1,1,13) = 
     $  + 6.911232325695532d-13
     $  - 3.877421460830331d-12*tz01
     $  - 5.003593942772569d-14*tz02
     $  + 1.564821854511482d-11*tz03
     $  + 4.775736125243386d-12*tz04
     $  - 3.627838469193895d-11*tz05
     $  - 1.754450496970339d-11*tz06
     $  + 4.391924230888817d-11*tz07
     $  + 2.605228593294933d-11*tz08
     $  - 2.657362461945610d-11*tz09
     $  - 1.769844528674049d-11*tz10
     $  + 6.367138784245408d-12*tz11
     $  + 4.568672039086328d-12*tz12

      CsH4YZ(2,0,1,1,14) = 
     $  - 1.346930127729308d-13
     $  + 2.874478021100397d-14*tz01
     $  + 9.703897240484909d-13*tz02
     $  + 9.149475703975066d-14*tz03
     $  - 3.636158950468039d-12*tz04
     $  - 1.245482135792799d-12*tz05
     $  + 7.298072656407221d-12*tz06
     $  + 3.606334720647559d-12*tz07
     $  - 7.904034110890603d-12*tz08
     $  - 4.680432692622352d-12*tz09
     $  + 4.388646820720243d-12*tz10
     $  + 2.895813621362796d-12*tz11
     $  - 9.822231270443820d-13*tz12
     $  - 6.964730508459584d-13*tz13

      CsH4YZ(2,0,1,1,15) = 
     $  - 7.026265206467819d-15
     $  + 4.780086669714945d-14*tz01
     $  - 6.645870878647307d-15*tz02
     $  - 2.608711276913378d-13*tz03
     $  - 4.220546616277542d-14*tz04
     $  + 8.334099504901914d-13*tz05
     $  + 3.087571974239770d-13*tz06
     $  - 1.462638365744150d-12*tz07
     $  - 7.350960619432968d-13*tz08
     $  + 1.429125907828090d-12*tz09
     $  + 8.453671987904560d-13*tz10
     $  - 7.318326273140873d-13*tz11
     $  - 4.792551603359933d-13*tz12
     $  + 1.534119515542172d-13*tz13
     $  + 1.076978724926747d-13*tz14

      CsH4YZ(2,0,1,1,16) = 
     $  + 1.538454162106765d-15
     $  - 5.426195486308772d-16*tz01
     $  - 1.388119647893321d-14*tz02
     $  + 8.850458458945125d-16*tz03
     $  + 6.890839640853231d-14*tz04
     $  + 1.476414293311915d-14*tz05
     $  - 1.879094044000115d-13*tz06
     $  - 7.375126548132418d-14*tz07
     $  + 2.917822491894015d-13*tz08
     $  + 1.487065811032173d-13*tz09
     $  - 2.592401840799480d-13*tz10
     $  - 1.532658998127943d-13*tz11
     $  + 1.230278237946593d-13*tz12
     $  + 8.006565921315427d-14*tz13
     $  - 2.422613859580715d-14*tz14
     $  - 1.686164425263588d-14*tz15

      CsH4YZ(2,0,1,1,17) = 
     $  - 6.004361781449017d-16*tz01
     $  + 1.747685426321409d-16*tz02
     $  + 4.238047255062069d-15*tz03
     $  - 1.767995808147380d-14*tz05
     $  - 4.519147869154396d-15*tz06
     $  + 4.170901295681825d-14*tz07
     $  + 1.711858508740790d-14*tz08
     $  - 5.793480492153728d-14*tz09
     $  - 2.987927862124431d-14*tz10
     $  + 4.713757503038826d-14*tz11
     $  + 2.786425988131200d-14*tz12
     $  - 2.082525757699475d-14*tz13
     $  - 1.348215438612078d-14*tz14
     $  + 3.863362126733292d-15*tz15
     $  + 2.668880195924197d-15*tz16

      CsH4YZ(2,0,1,1,18) = 
     $  + 1.981447255373695d-16*tz02
     $  - 1.251857977771242d-15*tz04
     $  - 1.053652343092289d-16*tz05
     $  + 4.409670766234014d-15*tz06
     $  + 1.274106017865300d-15*tz07
     $  - 9.128070566311381d-15*tz08
     $  - 3.883357964600285d-15*tz09
     $  + 1.145203265598581d-14*tz10
     $  + 5.967446838452828d-15*tz11
     $  - 8.586681724253976d-15*tz12
     $  - 5.076455666452499d-15*tz13
     $  + 3.546298609998873d-15*tz14
     $  + 2.285651388847900d-15*tz15
     $  - 6.215238572816179d-16*tz16
     $  - 4.265332525379069d-16*tz17

      CsH4YZ(2,0,1,1,19) = 
     $  + 3.549306465234013d-16*tz05
     $  - 1.072531223794831d-15*tz07
     $  - 3.393047699020528d-16*tz08
     $  + 1.972916913602319d-15*tz09
     $  + 8.645397381936038d-16*tz10
     $  - 2.254458072072772d-15*tz11
     $  - 1.185444973202522d-15*tz12
     $  + 1.566455877920129d-15*tz13
     $  + 9.263767378716966d-16*tz14
     $  - 6.070751162429674d-16*tz15
     $  - 3.897716256321055d-16*tz16
     $  + 1.007810979394638d-16*tz17

      CsH4YZ(2,0,1,1,20) = 
     $  + 2.552216909603525d-16*tz08
     $  - 4.217685857685613d-16*tz10
     $  - 1.894735702700464d-16*tz11
     $  + 4.421649605048186d-16*tz12
     $  + 2.343758167745376d-16*tz13
     $  - 2.861126070425739d-16*tz14
     $  - 1.692745750450756d-16*tz15
     $  + 1.044083967652734d-16*tz16

      CsH4YZ(2,0,2,1,03) = 
     $  + 1.757812500000000d-02
     $  - 1.757812500000000d-02*tz01

      CsH4YZ(2,0,2,1,04) = 
     $  - 3.845214843750000d-03
     $  + 3.295898437500000d-03*tz01
     $  + 5.493164062500000d-04*tz02

      CsH4YZ(2,0,2,1,05) = 
     $  + 3.707885742187500d-04
     $  - 2.162933349609375d-04*tz01
     $  - 1.235961914062500d-04*tz02
     $  - 3.089904785156250d-05*tz03

      CsH4YZ(2,0,2,1,06) = 
     $  - 9.527206420898437d-06
     $  - 7.724761962890625d-06*tz01
     $  + 7.209777832031250d-06*tz02
     $  + 7.724761962890625d-06*tz03
     $  + 2.317428588867187d-06*tz04

      CsH4YZ(2,0,2,1,07) = 
     $  - 1.310450690133231d-06
     $  + 1.655306134905133d-06*tz01
     $  + 8.276530674525669d-07*tz02
     $  - 3.448554447719029d-07*tz03
     $  - 6.207398005894252d-07*tz04
     $  - 2.069132668631417d-07*tz05

      CsH4YZ(2,0,2,1,08) = 
     $  + 6.835527566014503d-08
     $  + 3.879623753683907d-08*tz01
     $  - 1.057659332849541d-07*tz02
     $  - 9.699059384209769d-08*tz03
     $  + 1.662695894435960d-08*tz04
     $  + 5.819435630525861d-08*tz05
     $  + 2.078369868044950d-08*tz06

      CsH4YZ(2,0,2,1,09) = 
     $  + 9.429641067981719d-09
     $  - 1.530632809070604d-08*tz01
     $  - 6.870167063815253d-09*tz02
     $  + 9.311770554631948d-09*tz03
     $  + 1.212382423026221d-08*tz04
     $  - 3.536115400493144d-10*tz05
     $  - 6.061912115131105d-09*tz06
     $  - 2.273217043174164d-09*tz07

      CsH4YZ(2,0,2,1,10) = 
     $  - 6.107376455994588d-10
     $  - 2.576312648930719d-10*tz01
     $  + 1.361909588532788d-09*tz02
     $  + 1.166918082162737d-09*tz03
     $  - 9.153487293847969d-10*tz04
     $  - 1.591251930221915d-09*tz05
     $  - 1.010318685855184d-10*tz06
     $  + 6.819651129522493d-10*tz07
     $  + 2.652086550369858d-10*tz08

      CsH4YZ(2,0,2,1,11) = 
     $  - 8.374738243148456d-11
     $  + 1.555087113625962d-10*tz01
     $  + 6.406338939854463d-11*tz02
     $  - 1.648082356301269d-10*tz03
     $  - 1.952900096181441d-10*tz04
     $  + 9.330522681755775d-11*tz05
     $  + 2.169888995757157d-10*tz06
     $  + 2.789857280259202d-11*tz07
     $  - 8.137083734089339d-11*tz08
     $  - 3.254833493635735d-11*tz09

      CsH4YZ(2,0,2,1,12) = 
     $  + 6.049485351852734d-12
     $  + 2.001980918704519d-12*tz01
     $  - 1.730255551586120d-11*tz02
     $  - 1.420760651983852d-11*tz03
     $  + 2.185740457576811d-11*tz04
     $  + 3.254833493635735d-11*tz05
     $  - 9.371442936897624d-12*tz06
     $  - 3.051406400283502d-11*tz07
     $  - 5.393900202521342d-12*tz08
     $  + 1.017135466761167d-11*tz09
     $  + 4.161008727659321d-12*tz10

      CsH4YZ(2,0,2,1,13) = 
     $  + 8.237058587659881d-13
     $  - 1.673569734343523d-12*tz01
     $  - 6.436269299743284d-13*tz02
     $  + 2.618673629588026d-12*tz03
     $  + 2.961987897711091d-12*tz04
     $  - 2.995636266517636d-12*tz05
     $  - 5.427982538965845d-12*tz06
     $  + 8.644952885418576d-13*tz07
     $  + 4.401066923485820d-12*tz08
     $  + 9.413393141900227d-13*tz09
     $  - 1.320320077045746d-12*tz10
     $  - 5.501333654357275d-13*tz11

      CsH4YZ(2,0,2,1,14) = 
     $  - 6.367823063825401d-14
     $  - 1.724000705288379d-14*tz01
     $  + 2.207642475428176d-13*tz02
     $  + 1.759179301445494d-13*tz03
     $  - 4.278732676781724d-13*tz04
     $  - 5.948550914014210d-13*tz05
     $  + 4.141141211641440d-13*tz06
     $  + 9.077200529689504d-13*tz07
     $  - 6.035345895531150d-14*tz08
     $  - 6.483714664063932d-13*tz09
     $  - 1.577855037128845d-13*tz10
     $  + 1.768285817471981d-13*tz11
     $  + 7.481209227766075d-14*tz12

      CsH4YZ(2,0,2,1,15) = 
     $  - 8.622055908233909d-15
     $  + 1.873821605943325d-14*tz01
     $  + 6.812422468581651d-15*tz02
     $  - 3.980010106272555d-14*tz03
     $  - 4.377536559300942d-14*tz04
     $  + 7.130610999620829d-14*tz05
     $  + 1.166039478473719d-13*tz06
     $  - 5.688026143385610d-14*tz07
     $  - 1.523672946055024d-13*tz08
     $  - 5.351637500497213d-16*tz09
     $  + 9.725571996095898d-14*tz10
     $  + 2.599801172116544d-14*tz11
     $  - 2.431392999023974d-14*tz12
     $  - 1.042025571010274d-14*tz13

      CsH4YZ(2,0,2,1,16) = 
     $  + 6.986015430336561d-16
     $  + 1.596661516073824d-16*tz01
     $  - 2.838422055395675d-15*tz02
     $  - 2.211636413779699d-15*tz03
     $  + 7.662630350772021d-15*tz04
     $  + 1.025068534519065d-14*tz05
     $  - 1.189590430752675d-14*tz06
     $  - 2.248314895228419d-14*tz07
     $  + 7.652338740194142d-15*tz08
     $  + 2.568158855219073d-14*tz09
     $  + 1.494063368434334d-15*tz10
     $  - 1.481630108780234d-14*tz11
     $  - 4.254937748291955d-15*tz12
     $  + 3.419146404877464d-15*tz13
     $  + 1.481630108780234d-15*tz14

      CsH4YZ(2,0,2,1,17) = 
     $  - 2.160372331613278d-16*tz01
     $  + 5.914419392948404d-16*tz03
     $  + 6.384742937569013d-16*tz04
     $  - 1.487433545193023d-15*tz05
     $  - 2.301126500646426d-15*tz06
     $  + 1.970999927403748d-15*tz07
     $  + 4.285106687158521d-15*tz08
     $  - 9.887369225719405d-16*tz09
     $  - 4.346841275024364d-15*tz10
     $  - 4.435110181070513d-16*tz11
     $  + 2.287811197381244d-15*tz12
     $  + 6.952997629403850d-16*tz13
     $  - 4.902452565816952d-16*tz14
     $  - 2.144822997544916d-16*tz15

      CsH4YZ(2,0,2,1,18) = 
     $  - 1.305395472059524d-16*tz04
     $  - 1.704235018636058d-16*tz05
     $  + 2.863835235054569d-16*tz06
     $  + 5.009126302202784d-16*tz07
     $  - 3.228196917709928d-16*tz08
     $  - 8.098881900782480d-16*tz09
     $  + 1.183655819686825d-16*tz10
     $  + 7.387723658210268d-16*tz11
     $  + 1.019392998330381d-16*tz12
     $  - 3.574704995908194d-16*tz13
     $  - 1.137497163964064d-16*tz14

      CsH4YZ(2,0,2,1,19) = 
     $  - 1.065333762063767d-16*tz08
     $  + 1.521265434348900d-16*tz10
     $  - 1.260553866978152d-16*tz12

      CsH4YZ(2,1,1,1,04) = 
     $  + 4.119873046875000d-03
     $  - 6.179809570312500d-03*tz01
     $  + 2.471923828125000d-03*tz02
     $  - 4.119873046875000d-04*tz03

      CsH4YZ(2,1,1,1,05) = 
     $  - 4.634857177734375d-04
     $  + 3.707885742187500d-04*tz01
     $  + 3.707885742187500d-04*tz02
     $  - 3.707885742187500d-04*tz03
     $  + 9.269714355468750d-05*tz04

      CsH4YZ(2,1,1,1,06) = 
     $  + 4.827976226806640d-06
     $  + 4.345178604125976d-05*tz01
     $  - 5.793571472167968d-05*tz02
     $  - 2.655386924743652d-05*tz03
     $  + 5.310773849487304d-05*tz04
     $  - 1.689791679382324d-05*tz05

      CsH4YZ(2,1,1,1,07) = 
     $  + 1.939811876841953d-06
     $  - 3.103699002947126d-06*tz01
     $  - 6.013416818210056d-06*tz02
     $  + 1.086294651031494d-05*tz03
     $  + 1.163887126105172d-06*tz04
     $  - 7.759247507367815d-06*tz05
     $  + 2.909717815262930d-06*tz06

      CsH4YZ(2,1,1,1,08) = 
     $  - 1.697335392236709d-08
     $  - 4.582805559039115d-07*tz01
     $  + 8.741277270019054d-07*tz02
     $  + 7.638009265065193d-07*tz03
     $  - 2.019829116761684d-06*tz04
     $  + 1.867068931460380d-07*tz05
     $  + 1.162674743682146d-06*tz06
     $  - 4.922272637486457d-07*tz07

      CsH4YZ(2,1,1,1,09) = 
     $  - 1.432126737199723d-08
     $  + 3.111781552433967d-08*tz01
     $  + 9.335344657301902d-08*tz02
     $  - 2.234824933111667d-07*tz03
     $  - 7.991620805114507d-08*tz04
     $  + 3.705848939716815d-07*tz05
     $  - 8.203787729144096d-08*tz06
     $  - 1.782202161848545d-07*tz07
     $  + 8.292190614156424d-08*tz08

      CsH4YZ(2,1,1,1,10) = 
     $  + 7.672107520712805d-11
     $  + 5.231240720604546d-09*tz01
     $  - 1.275274761220706d-08*tz02
     $  - 1.635769097317409d-08*tz03
     $  + 5.227262590778991d-08*tz04
     $  + 3.972446782946852d-09*tz05
     $  - 6.744634967097746d-08*tz06
     $  + 2.113949774086774d-08*tz07
     $  + 2.784975030018748d-08*tz08
     $  - 1.398549427124505d-08*tz09

      CsH4YZ(2,1,1,1,11) = 
     $  + 1.261732329006114d-10
     $  - 3.448573582542624d-10*tz01
     $  - 1.402436068574742d-09*tz02
     $  + 4.115039488382323d-09*tz03
     $  + 2.426594613558785d-09*tz04
     $  - 1.157015810951941d-08*tz05
     $  + 1.174670376553581d-09*tz06
     $  + 1.222306220913562d-08*tz07
     $  - 4.691658427991451d-09*tz08
     $  - 4.423086229744276d-09*tz09
     $  + 2.366656273553216d-09*tz10

      CsH4YZ(2,1,1,1,12) = 
     $  - 4.217883185577063d-13
     $  - 6.218602775577762d-11*tz01
     $  + 1.845434890615901d-10*tz02
     $  + 3.106859455418283d-10*tz03
     $  - 1.156055183011111d-09*tz04
     $  - 2.640256128013821d-10*tz05
     $  + 2.470491879640641d-09*tz06
     $  - 5.558198815488627d-10*tz07
     $  - 2.210725773872062d-09*tz08
     $  + 9.735068636932250d-10*tz09
     $  + 7.121673764994998d-10*tz10
     $  - 4.021612871290309d-10*tz11

      CsH4YZ(2,1,1,1,13) = 
     $  - 1.224461521187481d-12
     $  + 4.049680151658301d-12*tz01
     $  + 2.069505665737146d-11*tz02
     $  - 7.161094750152507d-11*tz03
     $  - 5.823180229222915d-11*tz04
     $  + 2.985889882597928d-10*tz05
     $  + 1.777498368274119d-12*tz06
     $  - 5.145821755516553d-10*tz07
     $  + 1.630714272407116d-10*tz08
     $  + 3.995697223640465d-10*tz09
     $  - 1.947261229185725d-10*tz10
     $  - 1.160152677223172d-10*tz11
     $  + 6.863840446563191d-11*tz12

      CsH4YZ(2,1,1,1,14) = 
     $  + 2.726497135669982d-15
     $  + 7.598717110080987d-13*tz01
     $  - 2.660047636705474d-12*tz02
     $  - 5.527724618482323d-12*tz03
     $  + 2.341648010699330d-11*tz04
     $  + 8.734987325290723d-12*tz05
     $  - 7.290647259375281d-11*tz06
     $  + 1.063760594916540d-11*tz07
     $  + 1.052810921605627d-10*tz08
     $  - 4.093251634208423d-11*tz09
     $  - 7.222430098327276d-11*tz10
     $  + 3.809319898478919d-11*tz11
     $  + 1.909052244903930d-11*tz12
     $  - 1.176542300968686d-11*tz13

      CsH4YZ(2,1,1,1,15) = 
     $  + 1.265920728633240d-14
     $  - 4.928168911916203d-14*tz01
     $  - 3.022778180358967d-13*tz02
     $  + 1.203133756583135d-12*tz03
     $  + 1.241032446856839d-12*tz04
     $  - 6.894587061981713d-12*tz05
     $  - 7.376005877508211d-13*tz06
     $  + 1.710139176697949d-11*tz07
     $  - 4.397026745141595d-12*tz08
     $  - 2.125335633783399d-11*tz09
     $  + 9.503121640257922d-12*tz10
     $  + 1.306165194672135d-11*tz11
     $  - 7.344966253309364d-12*tz12
     $  - 3.168952381349110d-12*tz13
     $  + 2.025058109836583d-12*tz14

      CsH4YZ(2,1,1,1,16) = 
     $  - 9.475725081437980d-15*tz01
     $  + 3.824281607854064d-14*tz02
     $  + 9.439832347188465d-14*tz03
     $  - 4.480079041424233d-13*tz04
     $  - 2.298576604350596d-13*tz05
     $  + 1.890015676524857d-12*tz06
     $  - 1.362193769498736d-13*tz07
     $  - 3.892734568972813d-12*tz08
     $  + 1.334307809259520d-12*tz09
     $  + 4.246416432469648d-12*tz10
     $  - 2.105566456890627d-12*tz11
     $  - 2.363987894992971d-12*tz12
     $  + 1.402318364976487d-12*tz13
     $  + 5.300754661748836d-13*tz14
     $  - 3.499052783508938d-13*tz15

      CsH4YZ(2,1,1,1,17) = 
     $  - 1.369118040902161d-16
     $  + 6.141584773010278d-16*tz01
     $  + 4.387357538355921d-15*tz02
     $  - 1.972841937144679d-14*tz03
     $  - 2.459164781219587d-14*tz04
     $  + 1.478449873573694d-13*tz05
     $  + 3.016770365482122d-14*tz06
     $  - 4.918519687943240d-13*tz07
     $  + 9.914535568839082d-14*tz08
     $  + 8.657064837247600d-13*tz09
     $  - 3.560578820180420d-13*tz10
     $  - 8.415539916589506d-13*tz11
     $  + 4.522800552547554d-13*tz12
     $  + 4.282362786615985d-13*tz13
     $  - 2.658723816680204d-13*tz14
     $  - 8.926752839630761d-14*tz15
     $  + 6.067835116602514d-14*tz16

      CsH4YZ(2,1,1,1,18) = 
     $  + 1.200243499456329d-16*tz01
     $  - 5.487793451536633d-16*tz02
     $  - 1.567346256022179d-15*tz03
     $  + 8.231969768324129d-15*tz04
     $  + 5.351483143327149d-15*tz05
     $  - 4.488266613410716d-14*tz06
     $  + 1.230250545844140d-13*tz08
     $  - 3.693223825665691d-14*tz09
     $  - 1.889895394211348d-13*tz10
     $  + 8.839946910546433d-14*tz11
     $  + 1.656900733119712d-13*tz12
     $  - 9.503771457311758d-14*tz13
     $  - 7.764992920724923d-14*tz14
     $  + 5.015316946952986d-14*tz15
     $  + 1.512365444357045d-14*tz16
     $  - 1.055787189365822d-14*tz17

      CsH4YZ(2,1,1,1,19) = 
     $  + 3.177575422857696d-16*tz03
     $  + 4.636916070786067d-16*tz04
     $  - 3.005952846618076d-15*tz05
     $  - 8.971309904243745d-16*tz06
     $  + 1.281183133121119d-14*tz07
     $  - 1.906922234373378d-15*tz08
     $  - 2.982601783490971d-14*tz09
     $  + 1.128269522253848d-14*tz10
     $  + 4.064005990995503d-14*tz11
     $  - 2.094442969420457d-14*tz12
     $  - 3.244764442441654d-14*tz13
     $  + 1.964639358652672d-14*tz14
     $  + 1.409378814927525d-14*tz15
     $  - 9.425192811334232d-15*tz16
     $  - 2.576036354143411d-15*tz17
     $  + 1.842778494921061d-15*tz18

      CsH4YZ(2,1,1,1,20) = 
     $  - 1.467489483059200d-16*tz04
     $  - 1.151977923947880d-16*tz05
     $  + 9.999121451996530d-16*tz06
     $  - 3.487614249170923d-15*tz08
     $  + 9.067367029643048d-16*tz09
     $  + 7.050811236210087d-15*tz10
     $  - 3.116438477542478d-15*tz11
     $  - 8.630649460534195d-15*tz12
     $  + 4.801178943631729d-15*tz13
     $  + 6.326121389006699d-15*tz14
     $  - 4.010500744985217d-15*tz15
     $  - 2.560599629872100d-15*tz16
     $  + 1.766242173405095d-15*tz17
     $  + 4.409048425113281d-16*tz18
     $  - 3.225716300313850d-16*tz19

      CsH4YZ(2,1,1,1,21) = 
     $  - 3.102895889556541d-16*tz07
     $  + 9.141822120049586d-16*tz09
     $  - 3.174591985174950d-16*tz10
     $  - 1.632462703420315d-15*tz11
     $  + 8.084827811911913d-16*tz12
     $  + 1.813709691942959d-15*tz13
     $  - 1.073772694364615d-15*tz14
     $  - 1.228761732908799d-15*tz15
     $  + 8.105540820276845d-16*tz16
     $  + 4.656655514846874d-16*tz17
     $  - 3.302700454696887d-16*tz18

      CsH4YZ(2,1,1,1,22) = 
     $  - 2.323510840213342d-16*tz10
     $  + 3.714194509481812d-16*tz12
     $  - 2.008082737019151d-16*tz13
     $  - 3.777490759177494d-16*tz14
     $  + 2.355812153470597d-16*tz15
     $  + 2.379127110246513d-16*tz16
     $  - 1.625018523006153d-16*tz17

      CsH4YZ(2,2,0,1,03) = 
     $  + 3.515625000000000d-02
     $  - 3.515625000000000d-02*tz01

      CsH4YZ(2,2,0,1,04) = 
     $  - 4.119873046875000d-03
     $  + 3.295898437500000d-03*tz01
     $  + 8.239746093750000d-04*tz02

      CsH4YZ(2,2,0,1,05) = 
     $  + 2.883911132812500d-04
     $  - 1.235961914062500d-04*tz01
     $  - 1.235961914062500d-04*tz02
     $  - 4.119873046875000d-05*tz03

      CsH4YZ(2,2,0,1,06) = 
     $  - 6.759166717529296d-06
     $  - 7.724761962890625d-06*tz01
     $  + 3.862380981445312d-06*tz02
     $  + 7.724761962890625d-06*tz03
     $  + 2.896785736083984d-06*tz04

      CsH4YZ(2,2,0,1,07) = 
     $  - 6.483282361711774d-07
     $  + 8.276530674525669d-07*tz01
     $  + 8.276530674525669d-07*tz02
     $  - 1.379421779087611d-07*tz03
     $  - 6.207398005894252d-07*tz04
     $  - 2.482959202357700d-07*tz05

      CsH4YZ(2,2,0,1,08) = 
     $  + 3.879623753683907d-08
     $  + 3.879623753683907d-08*tz01
     $  - 6.304388599736349d-08*tz02
     $  - 9.699059384209769d-08*tz03
     $  + 5.819435630525861d-08*tz05
     $  + 2.424764846052442d-08*tz06

      CsH4YZ(2,2,0,1,09) = 
     $  + 3.271508125626311d-09
     $  - 6.870167063815253d-09*tz01
     $  - 6.870167063815253d-09*tz02
     $  + 5.792493798903056d-09*tz03
     $  + 1.212382423026221d-08*tz04
     $  + 1.212382423026221d-09*tz05
     $  - 6.061912115131105d-09*tz06
     $  - 2.597962335056188d-09*tz07

      CsH4YZ(2,2,0,1,10) = 
     $  - 2.869936517007382d-10
     $  - 2.576312648930719d-10*tz01
     $  + 8.107807453987853d-10*tz02
     $  + 1.166918082162737d-09*tz03
     $  - 5.569381755776703d-10*tz04
     $  - 1.591251930221915d-09*tz05
     $  - 2.652086550369858d-10*tz06
     $  + 6.819651129522493d-10*tz07
     $  + 2.983597369166091d-10*tz08

      CsH4YZ(2,2,0,1,11) = 
     $  - 2.202609544105876d-11
     $  + 6.406338939854463d-11*tz01
     $  + 6.406338939854463d-11*tz02
     $  - 1.088388766125812d-10*tz03
     $  - 1.952900096181441d-10*tz04
     $  + 5.207733589817177d-11*tz05
     $  + 2.169888995757157d-10*tz06
     $  + 4.649762133765336d-11*tz07
     $  - 8.137083734089339d-11*tz08
     $  - 3.616481659595261d-11*tz09

      CsH4YZ(2,2,0,1,12) = 
     $  + 2.429823615040566d-12
     $  + 2.001980918704519d-12*tz01
     $  - 1.015520966020276d-11*tz02
     $  - 1.420760651983852d-11*tz03
     $  + 1.501485689028389d-11*tz04
     $  + 3.254833493635735d-11*tz05
     $  - 4.238064444838197d-12*tz06
     $  - 3.051406400283502d-11*tz07
     $  - 7.628516000708755d-12*tz08
     $  + 1.017135466761167d-11*tz09
     $  + 4.577109600425253d-12*tz10

      CsH4YZ(2,2,0,1,13) = 
     $  + 1.733894764211857d-13
     $  - 6.436269299743284d-13*tz01
     $  - 6.436269299743284d-13*tz02
     $  + 1.760116288482618d-12*tz03
     $  + 2.961987897711091d-12*tz04
     $  - 2.071994364295070d-12*tz05
     $  - 5.427982538965845d-12*tz06
     $  + 1.886171538636780d-13*tz07
     $  + 4.401066923485820d-12*tz08
     $  + 1.222518589857172d-12*tz09
     $  - 1.320320077045746d-12*tz10
     $  - 6.001454895662482d-13*tz11

      CsH4YZ(2,2,0,1,14) = 
     $  - 2.241388037067218d-14
     $  - 1.724000705288379d-14*tz01
     $  + 1.276284458451911d-13*tz02
     $  + 1.759179301445494d-13*tz03
     $  - 3.058152083216821d-13*tz04
     $  - 5.948550914014210d-13*tz05
     $  + 2.816470761479623d-13*tz06
     $  + 9.077200529689504d-13*tz07
     $  + 3.241857332031966d-14*tz08
     $  - 6.483714664063932d-13*tz09
     $  - 1.945114399219179d-13*tz10
     $  + 1.768285817471981d-13*tz11
     $  + 8.104643330079915d-14*tz12

      CsH4YZ(2,2,0,1,15) = 
     $  - 1.508816368172256d-15
     $  + 6.812422468581651d-15*tz01
     $  + 6.812422468581651d-15*tz02
     $  - 2.691276957247906d-14*tz03
     $  - 4.377536559300942d-14*tz04
     $  + 5.245222247121942d-14*tz05
     $  + 1.166039478473719d-13*tz06
     $  - 3.709390498284195d-14*tz07
     $  - 1.523672946055024d-13*tz08
     $  - 1.368784206857941d-14*tz09
     $  + 9.725571996095898d-14*tz10
     $  + 3.094500180575967d-14*tz11
     $  - 2.431392999023974d-14*tz12
     $  - 1.122181384164911d-14*tz13

      CsH4YZ(2,2,0,1,16) = 
     $  + 2.195884694464834d-16
     $  + 1.596661516073824d-16*tz01
     $  - 1.618810772432928d-15*tz02
     $  - 2.211636413779699d-15*tz03
     $  + 5.577314685576774d-15*tz04
     $  + 1.025068534519065d-14*tz05
     $  - 8.844051328336337d-15*tz06
     $  - 2.248314895228419d-14*tz07
     $  + 4.609515893982951d-15*tz08
     $  + 2.568158855219073d-14*tz09
     $  + 3.407749250194539d-15*tz10
     $  - 1.481630108780234d-14*tz11
     $  - 4.938767029267448d-15*tz12
     $  + 3.419146404877464d-15*tz13
     $  + 1.587460830835965d-15*tz14

      CsH4YZ(2,2,0,1,17) = 
     $  + 4.006694725795525d-16*tz03
     $  + 6.384742937569013d-16*tz04
     $  - 1.125286182885095d-15*tz05
     $  - 2.301126500646426d-15*tz06
     $  + 1.462435320956400d-15*tz07
     $  + 4.285106687158521d-15*tz08
     $  - 5.104199584986374d-16*tz09
     $  - 4.346841275024364d-15*tz10
     $  - 7.279399264394868d-16*tz11
     $  + 2.287811197381244d-15*tz12
     $  + 7.919346452473538d-16*tz13
     $  - 4.902452565816952d-16*tz14
     $  - 2.287811197381244d-16*tz15

      CsH4YZ(2,2,0,1,18) = 
     $  - 1.704235018636058d-16*tz05
     $  + 2.211892418519932d-16*tz06
     $  + 5.009126302202784d-16*tz07
     $  - 2.364128311976427d-16*tz08
     $  - 8.098881900782480d-16*tz09
     $  + 7.387723658210268d-16*tz11
     $  + 1.449741470562767d-16*tz12
     $  - 3.574704995908194d-16*tz13
     $  - 1.276680355681497d-16*tz14

      CsH4YZ(2,2,0,1,19) = 
     $  - 1.065333762063767d-16*tz08
     $  + 1.521265434348900d-16*tz10
     $  - 1.260553866978152d-16*tz12

      CsH4YZ(2,2,1,1,04) = 
     $  + 4.943847656250000d-03
     $  - 6.591796875000000d-03*tz01
     $  + 1.647949218750000d-03*tz02

      CsH4YZ(2,2,1,1,05) = 
     $  - 3.707885742187500d-04
     $  + 1.853942871093750d-04*tz01
     $  + 3.707885742187500d-04*tz02
     $  - 1.853942871093750d-04*tz03

      CsH4YZ(2,2,1,1,06) = 
     $  + 1.931190490722656d-06
     $  + 3.089904785156250d-05*tz01
     $  - 2.317428588867187d-05*tz02
     $  - 3.089904785156250d-05*tz03
     $  + 2.124309539794921d-05*tz04

      CsH4YZ(2,2,1,1,07) = 
     $  + 1.034566334315708d-06
     $  - 1.034566334315708d-06*tz01
     $  - 4.138265337262834d-06*tz02
     $  + 3.620982170104980d-06*tz03
     $  + 3.103699002947126d-06*tz04
     $  - 2.586415835789271d-06*tz05

      CsH4YZ(2,2,1,1,08) = 
     $  - 4.849529692104884d-09
     $  - 2.327774252210344d-07*tz01
     $  + 2.497507791434015d-07*tz02
     $  + 5.819435630525861d-07*tz03
     $  - 5.770940333604812d-07*tz04
     $  - 3.491661378315516d-07*tz05
     $  + 3.321927839091845d-07*tz06

      CsH4YZ(2,2,1,1,09) = 
     $  - 5.657784640789031d-09
     $  + 7.779453881084918d-09*tz01
     $  + 4.809116944670677d-08*tz02
     $  - 5.587062332779169d-08*tz03
     $  - 8.486676961183547d-08*tz04
     $  + 9.264622349292039d-08*tz05
     $  + 4.243338480591773d-08*tz06
     $  - 4.455505404621362d-08*tz07

      CsH4YZ(2,2,1,1,10) = 
     $  + 1.704912782380623d-11
     $  + 2.061050119144575d-09*tz01
     $  - 2.833943913823791d-09*tz02
     $  - 9.335344657301902d-09*tz03
     $  + 1.161613909061998d-08*tz04
     $  + 1.273001544177532d-08*tz05
     $  - 1.498807770466165d-08*tz06
     $  - 5.455720903617995d-09*tz07
     $  + 6.188833400041663d-09*tz08

      CsH4YZ(2,2,1,1,11) = 
     $  + 3.952297813700536d-11
     $  - 6.897147165085249d-11*tz01
     $  - 5.765705045869017d-10*tz02
     $  + 8.230078976764646d-10*tz03
     $  + 1.757610086563297d-09*tz04
     $  - 2.314031621903882d-09*tz05
     $  - 1.952900096181441d-09*tz06
     $  + 2.444612441827125d-09*tz07
     $  + 7.323375360680405d-10*tz08
     $  - 8.846172459488553d-10*tz09

      CsH4YZ(2,2,1,1,12) = 
     $  - 7.668878519231024d-14
     $  - 2.001980918704519d-11*tz01
     $  + 3.355336164756184d-11*tz02
     $  + 1.420760651983852d-10*tz03
     $  - 2.101918514565656d-10*tz04
     $  - 3.254833493635735d-10*tz05
     $  + 4.491803417528438d-10*tz06
     $  + 3.051406400283502d-10*tz07
     $  - 4.019501407040113d-10*tz08
     $  - 1.017135466761167d-10*tz09
     $  + 1.294849775453636d-10*tz10

      CsH4YZ(2,2,1,1,13) = 
     $  - 3.176219726790824d-13
     $  + 6.749466919430503d-13*tz01
     $  + 7.079896229717613d-12*tz02
     $  - 1.193515791692084d-11*tz03
     $  - 3.258186687482201d-11*tz04
     $  + 4.976483137663213d-11*tz05
     $  + 5.970780792862429d-11*tz06
     $  - 8.576369592527588d-11*tz07
     $  - 4.841173615834402d-11*tz08
     $  + 6.659495372734109d-11*tz09
     $  + 1.452352084750320d-11*tz10
     $  - 1.933587795371954d-11*tz11

      CsH4YZ(2,2,1,1,14) = 
     $  + 4.194610977953818d-16
     $  + 2.068800846346055d-13*tz01
     $  - 4.092380979546884d-13*tz02
     $  - 2.111015161734593d-12*tz03
     $  + 3.602535401075892d-12*tz04
     $  + 7.138261096817052d-12*tz05
     $  - 1.121638039903889d-11*tz06
     $  - 1.089264063562740d-11*tz07
     $  + 1.619709110162504d-11*tz08
     $  + 7.780457596876718d-12*tz09
     $  - 1.111143092050350d-11*tz10
     $  - 2.121942980966377d-12*tz11
     $  + 2.937003453698353d-12*tz12

      CsH4YZ(2,2,1,1,15) = 
     $  + 2.801501146093617d-15
     $  - 7.040241302737433d-15*tz01
     $  - 8.856149209156146d-14*tz02
     $  + 1.718762509404479d-13*tz03
     $  + 5.690797527091224d-13*tz04
     $  - 9.849410088545304d-13*tz05
     $  - 1.515851322015835d-12*tz06
     $  + 2.443055966711356d-12*tz07
     $  + 1.980774829871531d-12*tz08
     $  - 3.036193762547713d-12*tz09
     $  - 1.264324359492466d-12*tz10
     $  + 1.865950278103050d-12*tz11
     $  + 3.160810898731166d-13*tz12
     $  - 4.527074830498729d-13*tz13

      CsH4YZ(2,2,1,1,16) = 
     $  - 2.235326122503354d-15*tz01
     $  + 5.099042143805419d-15*tz02
     $  + 3.096290979291578d-14*tz03
     $  - 5.973438721898978d-14*tz04
     $  - 1.435095948326692d-13*tz05
     $  + 2.520020902033143d-13*tz06
     $  + 3.147640853319787d-13*tz07
     $  - 5.190312758630418d-13*tz08
     $  - 3.595422397306702d-13*tz09
     $  + 5.661888576626197d-13*tz10
     $  + 2.074282152292328d-13*tz11
     $  - 3.151983859990628d-13*tz12
     $  - 4.786804966828449d-14*tz13
     $  + 7.067672882331781d-14*tz14

      CsH4YZ(2,2,1,1,17) = 
     $  + 1.124102546627174d-15*tz02
     $  - 2.466052421430849d-15*tz03
     $  - 9.577114406353520d-15*tz04
     $  + 1.848062341967117d-14*tz05
     $  + 3.451689750969639d-14*tz06
     $  - 6.148149609929050d-14*tz07
     $  - 6.427660030737781d-14*tz08
     $  + 1.082133104655950d-13*tz09
     $  + 6.520261912536546d-14*tz10
     $  - 1.051942489573688d-13*tz11
     $  - 3.431716796071866d-14*tz12
     $  + 5.352953483269981d-14*tz13
     $  + 7.353678848725428d-15*tz14
     $  - 1.115844104953845d-14*tz15

      CsH4YZ(2,2,1,1,18) = 
     $  - 4.506295857629825d-16*tz03
     $  + 9.684670315675447d-16*tz04
     $  + 2.726776029817693d-15*tz05
     $  - 5.280313662836136d-15*tz06
     $  - 8.014602083524454d-15*tz07
     $  + 1.447353583346047d-14*tz08
     $  + 1.295821104125196d-14*tz09
     $  - 2.223406346130998d-14*tz10
     $  - 1.182035785313642d-14*tz11
     $  + 1.949294980140838d-14*tz12
     $  + 5.719527993453111d-15*tz13
     $  - 9.135285789088144d-15*tz14
     $  - 1.143905598690622d-15*tz15
     $  + 1.779253463949465d-15*tz16

      CsH4YZ(2,2,1,1,19) = 
     $  + 1.571287447009145d-16*tz04
     $  - 3.339947607353418d-16*tz05
     $  - 7.311967169863779d-16*tz06
     $  + 1.423536814579021d-15*tz07
     $  + 1.811067395508405d-15*tz08
     $  - 3.314001981656635d-15*tz09
     $  - 2.586151238393131d-15*tz10
     $  + 4.515562212217226d-15*tz11
     $  + 2.142941573862859d-15*tz12
     $  - 3.605293824935171d-15*tz13
     $  - 9.595260778490416d-16*tz14
     $  + 1.565976461030583d-15*tz15
     $  + 1.799111395966953d-16*tz16
     $  - 2.862262615714901d-16*tz17

      CsH4YZ(2,2,1,1,20) = 
     $  + 1.052539100210161d-16*tz06
     $  + 1.873692741108855d-16*tz07
     $  - 3.671172893864130d-16*tz08
     $  - 4.005335999249856d-16*tz09
     $  + 7.421906564431670d-16*tz10
     $  + 5.119757001094529d-16*tz11
     $  - 9.084894168983363d-16*tz12
     $  - 3.886080615288618d-16*tz13
     $  + 6.659075146322841d-16*tz14
     $  + 1.619200256370257d-16*tz15
     $  - 2.695368031444316d-16*tz16

      CsH4YZ(2,2,1,1,21) = 
     $  - 1.632462703420315d-16*tz11
     $  - 1.006890439089804d-16*tz12
     $  + 1.813709691942959d-16*tz13
     $  - 1.228761732908799d-16*tz15

      CsH4YZ(2,2,2,1,04) = 
     $  + 6.591796875000000d-03
     $  - 6.591796875000000d-03*tz01

      CsH4YZ(2,2,2,1,05) = 
     $  - 2.471923828125000d-04
     $  + 2.471923828125000d-04*tz02

      CsH4YZ(2,2,2,1,06) = 
     $  + 1.544952392578125d-05*tz01
     $  - 1.544952392578125d-05*tz03

      CsH4YZ(2,2,2,1,07) = 
     $  + 4.138265337262834d-07
     $  - 1.655306134905133d-06*tz02
     $  + 1.241479601178850d-06*tz04

      CsH4YZ(2,2,2,1,08) = 
     $  - 7.759247507367815d-08*tz01
     $  + 1.939811876841953d-07*tz03
     $  - 1.163887126105172d-07*tz05

      CsH4YZ(2,2,2,1,09) = 
     $  - 1.616509897368294d-09
     $  + 1.374033412763050d-08*tz02
     $  - 2.424764846052442d-08*tz04
     $  + 1.212382423026221d-08*tz06

      CsH4YZ(2,2,2,1,10) = 
     $  + 5.152625297861439d-10*tz01
     $  - 2.333836164325475d-09*tz03
     $  + 3.182503860443830d-09*tz05
     $  - 1.363930225904498d-09*tz07

      CsH4YZ(2,2,2,1,11) = 
     $  + 8.782884030445636d-12
     $  - 1.281267787970892d-10*tz02
     $  + 3.905800192362882d-10*tz04
     $  - 4.339777991514314d-10*tz06
     $  + 1.627416746817867d-10*tz08

      CsH4YZ(2,2,2,1,12) = 
     $  - 4.003961837409039d-12*tz01
     $  + 2.841521303967705d-11*tz03
     $  - 6.509666987271471d-11*tz05
     $  + 6.102812800567004d-11*tz07
     $  - 2.034270933522334d-11*tz09

      CsH4YZ(2,2,2,1,13) = 
     $  - 5.774944957801499d-14
     $  + 1.287253859948656d-12*tz02
     $  - 5.923975795422183d-12*tz04
     $  + 1.085596507793169d-11*tz06
     $  - 8.802133846971641d-12*tz08
     $  + 2.640640154091492d-12*tz10

      CsH4YZ(2,2,2,1,14) = 
     $  + 3.448001410576759d-14*tz01
     $  - 3.518358602890988d-13*tz03
     $  + 1.189710182802842d-12*tz05
     $  - 1.815440105937900d-12*tz07
     $  + 1.296742932812786d-12*tz09
     $  - 3.536571634943962d-13*tz11

      CsH4YZ(2,2,2,1,15) = 
     $  + 4.310001763220949d-16
     $  - 1.362484493716330d-14*tz02
     $  + 8.755073118601884d-14*tz04
     $  - 2.332078956947439d-13*tz06
     $  + 3.047345892110048d-13*tz08
     $  - 1.945114399219179d-13*tz10
     $  + 4.862785998047949d-14*tz12

      CsH4YZ(2,2,2,1,16) = 
     $  - 3.193323032147649d-16*tz01
     $  + 4.423272827559398d-15*tz03
     $  - 2.050137069038131d-14*tz05
     $  + 4.496629790456838d-14*tz07
     $  - 5.136317710438146d-14*tz09
     $  + 2.963260217560468d-14*tz11
     $  - 6.838292809754928d-15*tz13

      CsH4YZ(2,2,2,1,17) = 
     $  + 1.498803395502899d-16*tz02
     $  - 1.276948587513802d-15*tz04
     $  + 4.602253001292852d-15*tz06
     $  - 8.570213374317042d-15*tz08
     $  + 8.693682550048728d-15*tz10
     $  - 4.575622394762488d-15*tz12
     $  + 9.804905131633904d-16*tz14

      CsH4YZ(2,2,2,1,18) = 
     $  + 3.408470037272117d-16*tz05
     $  - 1.001825260440556d-15*tz07
     $  + 1.619776380156496d-15*tz09
     $  - 1.477544731642053d-15*tz11
     $  + 7.149409991816388d-16*tz13
     $  - 1.429881998363277d-16*tz15

      CsH4YZ(2,2,2,1,19) = 
     $  + 2.130667524127535d-16*tz08
     $  - 3.042530868697801d-16*tz10
     $  + 2.521107733956305d-16*tz12
     $  - 1.128854209234166d-16*tz14
      endif

      return
      end

      subroutine fillcoeff2dhpl310(iflag,n,CrH2YZ,CrH3YZ,CrH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CrH2YZ,CrH3YZ,CrH4YZ
      real(dp)::
     $     tv01,tv02,tv03,tv04,tv05,tv06,tv07,tv08,tv09,tv10,
     $     tv11,tv12,tv13,tv14,tv15,tv16,tv17,tv18,tv19,tv20,tv21
      dimension CrH2YZ(0:3,0:3,18),CrH3YZ(0:3,0:3,0:3,18),
     $          CrH4YZ(0:3,0:3,0:3,0:3,18)
      common /tpv/  
     $     tv01,tv02,tv03,tv04,tv05,tv06,tv07,tv08,tv09,tv10,
     $     tv11,tv12,tv13,tv14,tv15,tv16,tv17,tv18,tv19,tv20,tv21
!$omp threadprivate(/tpv/)

* 2001-04-03:17:41:05.chpl 
* <- HPLexpand.310t 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      CrH2YZ(3,1,01) = 
     $  + 6.224694552604316d-01
     $  - 1.231893849147407d-01*tv01
     $  + 4.323485407234559d-03*tv02
     $  - 2.013333506542613d-05*tv03
     $  - 2.436051393807966d-06*tv04
     $  + 2.482888902259004d-08*tv05
     $  + 1.935975648316861d-09*tv06
     $  - 2.845440438346271d-11*tv07
     $  - 1.590223275140079d-12*tv08
     $  + 3.073351448049326d-14*tv09
     $  + 1.294967440691163d-15*tv10

      CrH2YZ(3,1,02) = 
     $  - 1.232496609743965d-01
     $  + 1.727447642407785d-02*tv01
     $  - 1.205521193114873d-04*tv02
     $  - 1.946520486038176d-05*tv03
     $  + 2.478910810694149d-07*tv04
     $  + 2.320629008196622d-08*tv05
     $  - 3.978091564854309d-10*tv06
     $  - 2.541769783611239d-11*tv07
     $  + 5.525048830470887d-13*tv08
     $  + 2.587456612888669d-14*tv09
     $  - 6.983776017899723d-16*tv10

      CrH2YZ(3,1,03) = 
     $  + 1.147746444217055d-02
     $  - 3.201534089874276d-04*tv01
     $  - 7.770634776269526d-05*tv02
     $  + 1.318909176538670d-06*tv03
     $  + 1.544716788317389d-07*tv04
     $  - 3.176589164925870d-09*tv05
     $  - 2.369217147434740d-10*tv06
     $  + 5.884086928474137d-12*tv07
     $  + 3.101317087192308d-13*tv08
     $  - 9.298500730163685d-15*tv09
     $  - 3.630897196448211d-16*tv10

      CrH2YZ(3,1,04) = 
     $  - 1.581062616492846d-04
     $  - 1.547962278629246d-04*tv01
     $  + 3.940885688858334d-06*tv02
     $  + 6.164676624659284d-07*tv03
     $  - 1.584184075767777d-08*tv04
     $  - 1.419052861027290d-09*tv05
     $  + 4.110506695158243d-11*tv06
     $  + 2.477427433553500d-12*tv07
     $  - 8.354154773652079d-14*tv08
     $  - 3.626236200345914d-15*tv09
     $  + 1.449588349523810d-16*tv10

      CrH2YZ(3,1,05) = 
     $  - 6.118156113216996d-05
     $  + 6.254919982561690d-06*tv01
     $  + 1.473860025999761d-06*tv02
     $  - 5.049711961162772d-08*tv03
     $  - 5.662363918446426d-09*tv04
     $  + 1.967708129608304d-10*tv05
     $  + 1.384752567883000d-11*tv06
     $  - 5.335083953206140d-13*tv07
     $  - 2.606794191669048d-14*tv08
     $  + 1.157514139550046d-15*tv09

      CrH2YZ(3,1,06) = 
     $  + 2.034802917875601d-06
     $  + 1.950102315563474d-06*tv01
     $  - 1.003408192899235d-07*tv02
     $  - 1.504438576954017d-08*tv03
     $  + 6.534199333319257d-10*tv04
     $  + 5.525134631695897d-11*tv05
     $  - 2.482776537509031d-12*tv06
     $  - 1.387563983610612d-13*tv07
     $  + 6.929307320500746d-15*tv08
     $  + 2.726251946213482d-16*tv09

      CrH2YZ(3,1,07) = 
     $  + 5.443555560401152d-07
     $  - 1.131901715592392d-07*tv01
     $  - 2.563306824175743d-08*tv02
     $  + 1.485050486383819d-09*tv03
     $  + 1.573073631640569d-10*tv04
     $  - 8.480789807781246d-12*tv05
     $  - 5.536263165175742d-13*tv06
     $  + 3.158688972606613d-14*tv07
     $  + 1.399275349388775d-15*tv08

      CrH2YZ(3,1,08) = 
     $  - 2.719430093701721d-08
     $  - 2.532010881123733d-08*tv01
     $  + 2.206483905585203d-09*tv02
     $  + 3.129594305201052d-10*tv03
     $  - 2.109182399052637d-11*tv04
     $  - 1.655295808008590d-12*tv05
     $  + 1.101505289267443d-13*tv06
     $  + 5.583141544131623d-15*tv07
     $  - 4.035851144871539d-16*tv08

      CrH2YZ(3,1,09) = 
     $  - 5.419881787125624d-09
     $  + 1.924114756141942d-09*tv01
     $  + 4.136181196323387d-10*tv02
     $  - 3.720427104433554d-11*tv03
     $  - 3.661121060458308d-12*tv04
     $  + 2.923049394720706d-13*tv05
     $  + 1.731406898060918d-14*tv06
     $  - 1.429803955024808d-15*tv07

      CrH2YZ(3,1,10) = 
     $  + 3.627907010805181d-10
     $  + 3.250780780591458d-10*tv01
     $  - 4.406450029574053d-11*tv02
     $  - 5.816417646725183d-12*tv03
     $  + 5.806249574621167d-13*tv04
     $  + 4.137605000811091d-14*tv05
     $  - 3.984921482024515d-15*tv06
     $  - 1.777155453511350d-16*tv07

      CrH2YZ(3,1,11) = 
     $  + 5.596990294478673d-11
     $  - 3.121100462253164d-11*tv01
     $  - 6.270404313789211d-12*tv02
     $  + 8.359046833972579d-13*tv03
     $  + 7.477857347109572d-14*tv04
     $  - 8.640709348118109d-15*tv05
     $  - 4.506084216651863d-16*tv06

      CrH2YZ(3,1,12) = 
     $  - 4.791020102055574d-12
     $  - 4.081461321945775d-12*tv01
     $  + 8.216280033993986d-13*tv02
     $  + 9.880822058036520d-14*tv03
     $  - 1.427667999785929d-14*tv04
     $  - 8.965440477624312d-16*tv05
     $  + 1.245022490042256d-16*tv06

      CrH2YZ(3,1,13) = 
     $  - 5.829979718180510d-13
     $  + 4.882740329016948d-13*tv01
     $  + 8.983830927363592d-14*tv02
     $  - 1.734319995365527d-14*tv03
     $  - 1.369278947484315d-15*tv04
     $  + 2.280985118812766d-16*tv05

      CrH2YZ(3,1,14) = 
     $  + 6.248182110506193d-14
     $  + 4.978836072020910d-14*tv01
     $  - 1.454322433321751d-14*tv02
     $  - 1.547816007582853d-15*tv03
     $  + 3.223756270584355d-16*tv04

      CrH2YZ(3,1,15) = 
     $  + 6.030609540302339d-15
     $  - 7.418034720846237d-15*tv01
     $  - 1.215677111705762d-15*tv02
     $  + 3.383515897601800d-16*tv03

      CrH2YZ(3,1,16) = 
     $  - 8.037753354530532d-16
     $  - 5.855459778014826d-16*tv01
     $  + 2.469580093054529d-16*tv02

      CrH2YZ(3,1,17) = 
     $  + 1.099344577499801d-16*tv01
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      CrH3YZ(0,3,1,01) = 
     $  + 6.224694552604316d-01
     $  - 1.231893849147407d-01*tv01
     $  + 4.323485407234559d-03*tv02
     $  - 2.013333506542613d-05*tv03
     $  - 2.436051393807966d-06*tv04
     $  + 2.482888902259004d-08*tv05
     $  + 1.935975648316861d-09*tv06
     $  - 2.845440438346271d-11*tv07
     $  - 1.590223275140079d-12*tv08
     $  + 3.073351448049326d-14*tv09
     $  + 1.294967440691163d-15*tv10

      CrH3YZ(0,3,1,02) = 
     $  - 1.783378533485291d-01
     $  + 3.173524788355282d-02*tv01
     $  - 8.709295735122236d-04*tv02
     $  - 5.957602105423482d-06*tv03
     $  + 5.807051768737011d-07*tv04
     $  + 6.947728349247478d-09*tv05
     $  - 5.619000123021269d-10*tv06
     $  - 7.373648096156936d-12*tv07
     $  + 5.744193056123093d-13*tv08
     $  + 7.174749099350861d-15*tv09
     $  - 5.919951960245792d-16*tv10

      CrH3YZ(0,3,1,03) = 
     $  + 2.895811434096732d-02
     $  - 4.190861495298366d-03*tv01
     $  + 5.672135848107749d-05*tv02
     $  + 2.558203639329994d-06*tv03
     $  - 1.755912855134670d-08*tv04
     $  - 3.571697924243098d-09*tv05
     $  + 1.001859151138483d-12*tv06
     $  + 4.693974470513823d-12*tv07
     $  + 9.466887184793779d-15*tv08
     $  - 5.853609845741025d-15*tv09

      CrH3YZ(0,3,1,04) = 
     $  - 2.559870821409520d-03
     $  + 1.937505957215025d-04*tv01
     $  + 6.857471376785772d-06*tv02
     $  - 1.976386891416170d-07*tv03
     $  - 1.553720647361276d-08*tv04
     $  + 2.149907308530194d-10*tv05
     $  + 2.782585144253265d-11*tv06
     $  - 2.301399376730172d-13*tv07
     $  - 4.348556802835001d-14*tv08
     $  + 2.683929651892578d-16*tv09

      CrH3YZ(0,3,1,05) = 
     $  + 5.251365707636245d-05
     $  + 2.068645491049668d-05*tv01
     $  - 1.109285266359915d-06*tv02
     $  - 4.220019367498542d-08*tv03
     $  + 1.717943341718003d-09*tv04
     $  + 1.138203771737817d-10*tv05
     $  - 2.704670071038099d-12*tv06
     $  - 2.348445462409250d-13*tv07
     $  + 4.099278183685359d-15*tv08
     $  + 4.135959107945312d-16*tv09

      CrH3YZ(0,3,1,06) = 
     $  + 1.195485099966970d-05
     $  - 2.540481417711931d-06*tv01
     $  - 6.922205113629893d-08*tv02
     $  + 6.890500144595095d-09*tv03
     $  + 3.208806011662678d-10*tv04
     $  - 1.587564801669720d-11*tv05
     $  - 9.289966961239534d-13*tv06
     $  + 3.143475792939577d-14*tv07
     $  + 2.090995915279577d-15*tv08

      CrH3YZ(0,3,1,07) = 
     $  - 6.381177656050895d-07
     $  - 1.622317043528928d-07*tv01
     $  + 2.009660582550452d-08*tv02
     $  + 5.802212146493707d-10*tv03
     $  - 6.219556331175276d-11*tv04
     $  - 2.633474219988952d-12*tv05
     $  + 1.631468410837551d-13*tv06
     $  + 7.979869051212283d-15*tv07
     $  - 3.666853333063778d-16*tv08

      CrH3YZ(0,3,1,08) = 
     $  - 9.898198680783122d-08
     $  + 3.477905467233905d-08*tv01
     $  + 5.842575741892316d-10*tv02
     $  - 1.668264787041734d-10*tv03
     $  - 5.129038532021660d-12*tv04
     $  + 6.093869783949069d-13*tv05
     $  + 2.262189730148567d-14*tv06
     $  - 1.750557275177898d-15*tv07

      CrH3YZ(0,3,1,09) = 
     $  + 8.134799493447581d-09
     $  + 1.297350659602477d-09*tv01
     $  - 3.420259641267267d-10*tv02
     $  - 5.925819320931608d-12*tv03
     $  + 1.691245310623867d-12*tv04
     $  + 4.536460064872078d-14*tv05
     $  - 6.398177061967082d-15*tv06
     $  - 1.968231801971715d-16*tv07

      CrH3YZ(0,3,1,10) = 
     $  + 9.496569402160852d-10
     $  - 4.781109248382280d-10*tv01
     $  - 1.847552987601955d-12*tv02
     $  + 3.465574401587989d-12*tv03
     $  + 5.639642172161540d-14*tv04
     $  - 1.803832281454410d-14*tv05
     $  - 3.948477376473042d-16*tv06

      CrH3YZ(0,3,1,11) = 
     $  - 1.046883699186156d-10
     $  - 8.608754976519498d-12*tv01
     $  + 5.554125799768549d-12*tv02
     $  + 1.731536457581791d-14*tv03
     $  - 3.933408190032251d-14*tv04
     $  - 4.798020074146909d-16*tv05
     $  + 2.001829014943703d-16*tv06

      CrH3YZ(0,3,1,12) = 
     $  - 9.721613055190085d-12
     $  + 6.548789743230566d-12*tv01
     $  - 7.671618410281081d-14*tv02
     $  - 6.560684319566392d-14*tv03
     $  + 4.567136255206271d-16*tv05

      CrH3YZ(0,3,1,13) = 
     $  + 1.350831823290014d-12
     $  + 1.630602769413537d-14*tv01
     $  - 8.696423475088185d-14*tv02
     $  + 1.226428637883579d-15*tv03
     $  + 8.242674562191330d-16*tv04

      CrH3YZ(0,3,1,14) = 
     $  + 1.032683801545537d-13
     $  - 8.921747741592468d-14*tv01
     $  + 2.684821567396462d-15*tv02
     $  + 1.162943630405251d-15*tv03

      CrH3YZ(0,3,1,15) = 
     $  - 1.745942766947813d-14
     $  + 9.697059500974766d-16*tv01
     $  + 1.321665902607398d-15*tv02

      CrH3YZ(0,3,1,16) = 
     $  - 1.124698884995468d-15
     $  + 1.208455846673188d-15*tv01

      CrH3YZ(0,3,1,17) = 
     $  + 2.260711373135860d-16

      CrH3YZ(3,0,1,01) = 
     $  + 6.782092247974649d-01
     $  - 7.216113782053397d-02*tv01
     $  - 2.960951508100631d-04*tv02
     $  + 7.475412348671377d-05*tv03
     $  + 4.033350131834205d-07*tv04
     $  - 8.380850632602890d-08*tv05
     $  - 3.990598293362059d-10*tv06
     $  + 8.854815831586199d-11*tv07
     $  + 2.972134190900594d-13*tv08
     $  - 8.930056977074862d-14*tv09
     $  - 1.323707487429359d-16*tv10

      CrH3YZ(3,0,1,02) = 
     $  - 1.951869548479676d-01
     $  + 1.609331771697788d-02*tv01
     $  + 3.271357746142554d-04*tv02
     $  - 1.624330872010077d-05*tv03
     $  - 5.889559134703411d-07*tv04
     $  + 1.842232489566941d-08*tv05
     $  + 8.402595641139942d-10*tv06
     $  - 2.066493205120255d-11*tv07
     $  - 1.052990910064457d-12*tv08
     $  + 2.322566189089934d-14*tv09
     $  + 1.216093103037168d-15*tv10

      CrH3YZ(3,0,1,03) = 
     $  + 3.184139301488236d-02
     $  - 1.450158579209803d-03*tv01
     $  - 8.863477094288642d-05*tv02
     $  - 4.178214855492912d-08*tv03
     $  + 1.688723592106411d-07*tv04
     $  + 2.221779960501837d-09*tv05
     $  - 2.613290887900823d-10*tv06
     $  - 5.096453982619611d-12*tv07
     $  + 3.643259251567834d-13*tv08
     $  + 8.309801817404972d-15*tv09
     $  - 4.790049643458670d-16*tv10

      CrH3YZ(3,0,1,04) = 
     $  - 2.816738781210481d-03
     $  - 6.322313097893355d-05*tv01
     $  + 7.542512087206283d-06*tv02
     $  + 6.036939446842662d-07*tv03
     $  - 6.752385118953338d-09*tv04
     $  - 1.558977952951249d-09*tv05
     $  - 2.795300279646951d-12*tv06
     $  + 2.985424859223038d-12*tv07
     $  + 2.205629059914391d-14*tv08
     $  - 4.917647506502016d-15*tv09

      CrH3YZ(3,0,1,05) = 
     $  + 5.269557192732591d-05
     $  + 2.296876685889049d-05*tv01
     $  + 9.691720605966172d-07*tv02
     $  - 7.120668349679054d-08*tv03
     $  - 5.360552705738598d-09*tv04
     $  + 1.200095994231251d-10*tv05
     $  + 1.539618265608106d-11*tv06
     $  - 1.370461574737724d-13*tv07
     $  - 3.341290381624600d-14*tv08
     $  + 1.012185050396192d-16*tv09

      CrH3YZ(3,0,1,06) = 
     $  + 1.443872256304861d-05
     $  - 1.964806225746746d-07*tv01
     $  - 2.260360534942153d-07*tv02
     $  - 9.663014890268288d-09*tv03
     $  + 7.777008493462294d-10*tv04
     $  + 5.061008046252987d-11*tv05
     $  - 1.732405969249359d-12*tv06
     $  - 1.559698722773235d-13*tv07
     $  + 3.083984772153681d-15*tv08
     $  + 3.694662795560566d-16*tv09

      CrH3YZ(3,0,1,07) = 
     $  - 7.843162232295030d-07
     $  - 3.349358917508652d-07*tv01
     $  - 4.654137723803331d-09*tv02
     $  + 2.501942112578285d-09*tv03
     $  + 9.874469734909933d-11*tv04
     $  - 9.101202880401135d-12*tv05
     $  - 4.965382116894941d-13*tv06
     $  + 2.360625712072601d-14*tv07
     $  + 1.602245139138935d-15*tv08

      CrH3YZ(3,0,1,08) = 
     $  - 1.219948897580502d-07
     $  + 1.551457112144596d-08*tv01
     $  + 4.594018535777915d-09*tv02
     $  + 7.210301974361308d-11*tv03
     $  - 2.922645506494774d-11*tv04
     $  - 9.952483324030488d-13*tv05
     $  + 1.102376849461778d-13*tv06
     $  + 4.925778645243073d-15*tv07
     $  - 3.137571361843792d-16*tv08

      CrH3YZ(3,0,1,09) = 
     $  + 1.111278849499782d-08
     $  + 4.542188594980310d-09*tv01
     $  - 1.271455850197537d-10*tv02
     $  - 6.072350276797710d-11*tv03
     $  - 8.476397386649971d-13*tv04
     $  + 3.529302936297900d-13*tv05
     $  + 9.829937176690904d-15*tv06
     $  - 1.361925506706879d-15*tv07

      CrH3YZ(3,0,1,10) = 
     $  + 1.128314378816068d-09
     $  - 3.739793269114518d-10*tv01
     $  - 7.918694420091131d-11*tv02
     $  + 1.357912405850091d-12*tv03
     $  + 7.922423059187027d-13*tv04
     $  + 8.143490849883109d-15*tv05
     $  - 4.347482528812887d-15*tv06

      CrH3YZ(3,0,1,11) = 
     $  - 1.523035316261818d-10
     $  - 5.771008886923884d-11*tv01
     $  + 5.243937383861452d-12*tv02
     $  + 1.224090904186982d-12*tv03
     $  - 1.804822429837529d-14*tv04
     $  - 1.029269969721534d-14*tv05

      CrH3YZ(3,0,1,12) = 
     $  - 1.044134790389392d-11
     $  + 7.122484209320401d-12*tv01
     $  + 1.223585474126401d-12*tv02
     $  - 7.622457977529697d-14*tv03
     $  - 1.782079464712259d-14*tv04
     $  + 2.851618126721534d-16*tv05
     $  + 1.333748521802491d-16*tv06

      CrH3YZ(3,0,1,13) = 
     $  + 2.038380528738098d-12
     $  + 6.840436076714525d-13*tv01
     $  - 1.297847740757517d-13*tv02
     $  - 2.172210926232545d-14*tv03
     $  + 1.136539440436194d-15*tv04
     $  + 2.506299292341066d-16*tv05

      CrH3YZ(3,0,1,14) = 
     $  + 9.144121207473734d-14
     $  - 1.216330601558613d-13*tv01
     $  - 1.721747166895109d-14*tv02
     $  + 2.223516317046154d-15*tv03
     $  + 3.502545812182480d-16*tv04

      CrH3YZ(3,0,1,15) = 
     $  - 2.677719653289162d-14
     $  - 7.404519403546397d-15*tv01
     $  + 2.672416953183475d-15*tv02
     $  + 3.462273169762704d-16*tv03

      CrH3YZ(3,0,1,16) = 
     $  - 6.948232907091114d-16
     $  + 1.944903427372098d-15*tv01
     $  + 2.190567819290847d-16*tv02

      CrH3YZ(3,0,1,17) = 
     $  + 3.461428449302012d-16

      CrH3YZ(3,1,1,01) = 
     $  - 1.051640180255739d-01
     $  - 9.402033994674527d-02*tv01
     $  + 1.074023882206285d-02*tv02
     $  - 4.013233767853031d-04*tv03
     $  + 2.341932090376803d-06*tv04
     $  + 2.235429037607168d-07*tv05
     $  - 2.688036179524982d-09*tv06
     $  - 1.760134327804249d-10*tv07
     $  + 2.962886005804824d-12*tv08
     $  + 1.431994939062815d-13*tv09
     $  - 3.121094194574164d-15*tv10
     $  - 1.153473109090728d-16*tv11

      CrH3YZ(3,1,1,02) = 
     $  + 1.382028521292729d-01
     $  - 3.216360807143694d-03*tv01
     $  - 7.844002655742544d-04*tv02
     $  + 1.115324698315586d-05*tv03
     $  + 1.319448151926097d-06*tv04
     $  - 2.289825694047018d-08*tv05
     $  - 1.735622135111768d-09*tv06
     $  + 3.667342861009529d-11*tv07
     $  + 1.978721901900859d-12*tv08
     $  - 5.082080706957052d-14*tv09
     $  - 2.049665002405111d-15*tv10

      CrH3YZ(3,1,1,03) = 
     $  - 3.293442544441353d-02
     $  + 2.233908257400380d-03*tv01
     $  + 1.432264473490440d-05*tv02
     $  + 2.156891697165798d-06*tv03
     $  - 9.034059957986124d-08*tv04
     $  - 8.062325490787053d-09*tv05
     $  + 2.422241334648488d-10*tv06
     $  + 1.472464515650694d-11*tv07
     $  - 4.707848872544430d-13*tv08
     $  - 2.082869654190734d-14*tv09
     $  + 7.639824181835512d-16*tv10

      CrH3YZ(3,1,1,04) = 
     $  + 3.272193944781859d-03
     $  - 3.174338752750525d-05*tv01
     $  - 8.139473856125980d-06*tv02
     $  - 1.126173415702970d-07*tv03
     $  - 1.124530233696832d-08*tv04
     $  + 8.539756798127960d-10*tv05
     $  + 5.846251457438907d-11*tv06
     $  - 2.655386201768485d-12*tv07
     $  - 1.290302789739769d-13*tv08
     $  + 5.883146166728360d-15*tv09
     $  + 2.102239101941618d-16*tv10

      CrH3YZ(3,1,1,05) = 
     $  - 3.654673453086869d-05
     $  - 3.627829752076332d-05*tv01
     $  + 3.242523083809597d-07*tv02
     $  + 5.676497784682200d-08*tv03
     $  + 1.024798707270703d-09*tv04
     $  + 6.693802607728224d-11*tv05
     $  - 8.662138286077717d-12*tv06
     $  - 4.525006106587130d-13*tv07
     $  + 2.973316892795597d-14*tv08
     $  + 1.135124311662469d-15*tv09

      CrH3YZ(3,1,1,06) = 
     $  - 1.968358549298886d-05
     $  + 1.216899715089131d-06*tv01
     $  + 2.979834808236047d-07*tv02
     $  - 3.613833857164418d-09*tv03
     $  - 4.867736098370675d-10*tv04
     $  - 9.894106833646863d-12*tv05
     $  - 3.739329294512331d-13*tv06
     $  + 9.140612898518184d-14*tv07
     $  + 3.537376762794073d-15*tv08
     $  - 3.375374744791020d-16*tv09

      CrH3YZ(3,1,1,07) = 
     $  + 5.625887068576027d-07
     $  + 5.483781919886217d-07*tv01
     $  - 1.697385080441909d-08*tv02
     $  - 2.725789809749496d-09*tv03
     $  + 4.210429606870614d-11*tv04
     $  + 4.653846577064925d-12*tv05
     $  + 9.788239311488853d-14*tv06
     $  + 1.248472827720873d-15*tv07
     $  - 9.879251655318485d-16*tv08

      CrH3YZ(3,1,1,08) = 
     $  + 1.860893584524070d-07
     $  - 2.750093108122626d-08*tv01
     $  - 6.480145994681645d-09*tv02
     $  + 2.237217899726701d-10*tv03
     $  + 2.638643290717081d-11*tv04
     $  - 5.053593046516456d-13*tv05
     $  - 4.765129680005052d-14*tv06
     $  - 9.711668305583381d-16*tv07

      CrH3YZ(3,1,1,09) = 
     $  - 8.228924902308328d-09
     $  - 7.823356650819181d-09*tv01
     $  + 4.810433797855427d-10*tv02
     $  + 7.286214237762348d-11*tv03
     $  - 2.883650975994974d-12*tv04
     $  - 2.649655402452328d-13*tv05
     $  + 6.200910963108589d-15*tv06
     $  + 5.111636842954948d-16*tv07

      CrH3YZ(3,1,1,10) = 
     $  - 1.927709429425540d-09
     $  + 5.245443810119441d-10*tv01
     $  + 1.179274285177012d-10*tv02
     $  - 7.389249855738672d-12*tv03
     $  - 8.030853877366006d-13*tv04
     $  + 3.683490432216939d-14*tv05
     $  + 2.728327202951468d-15*tv06

      CrH3YZ(3,1,1,11) = 
     $  + 1.161103229768840d-10
     $  + 1.067676290346854d-10*tv01
     $  - 1.099862230899173d-11*tv02
     $  - 1.558761867224795d-12*tv03
     $  + 1.063633463506336d-13*tv04
     $  + 8.757357896621047d-15*tv05
     $  - 4.689714358657760d-16*tv06

      CrH3YZ(3,1,1,12) = 
     $  + 2.052060740533827d-11
     $  - 9.142220197313692d-12*tv01
     $  - 1.936133005249421d-12*tv02
     $  + 1.931265069900946d-13*tv03
     $  + 1.920279245535626d-14*tv04
     $  - 1.475375165504982d-15*tv05

      CrH3YZ(3,1,1,13) = 
     $  - 1.594893606199029d-12
     $  - 1.403475137473864d-12*tv01
     $  + 2.234152449009944d-13*tv02
     $  + 2.916072397921053d-14*tv03
     $  - 3.080647132011642d-15*tv04
     $  - 2.256800790045367d-16*tv05

      CrH3YZ(3,1,1,14) = 
     $  - 2.193912226042186d-13
     $  + 1.503367617886981d-13*tv01
     $  + 2.948728618916779d-14*tv02
     $  - 4.415729916237906d-15*tv03
     $  - 3.937408779525654d-16*tv04

      CrH3YZ(3,1,1,15) = 
     $  + 2.143333885330696d-14
     $  + 1.780003820314898d-14*tv01
     $  - 4.198468764205978d-15*tv02
     $  - 4.933232674634370d-16*tv03

      CrH3YZ(3,1,1,16) = 
     $  + 2.325738977328877d-15
     $  - 2.370557398684666d-15*tv01
     $  - 4.208861051022590d-16*tv02

      CrH3YZ(3,1,1,17) = 
     $  - 2.825196045478136d-16
     $  - 2.172481257454985d-16*tv01

      CrH3YZ(3,3,1,01) = 
     $  + 5.730452067718910d-01
     $  - 1.661814777672792d-01*tv01
     $  + 1.044414367125278d-02*tv02
     $  - 3.265692532985893d-04*tv03
     $  + 2.745267103560223d-06*tv04
     $  + 1.397343974346879d-07*tv05
     $  - 3.087096008861188d-09*tv06
     $  - 8.746527446456293d-11*tv07
     $  + 3.260099424894883d-12*tv08
     $  + 5.389892413553291d-14*tv09
     $  - 3.253464943317100d-15*tv10

      CrH3YZ(3,3,1,02) = 
     $  - 1.671604874669599d-01
     $  + 4.179849982878412d-02*tv01
     $  - 1.958019399361471d-03*tv02
     $  + 2.192514377297165d-05*tv03
     $  + 1.396120430064328d-06*tv04
     $  - 3.699305551023825d-08*tv05
     $  - 1.223544282631165d-09*tv06
     $  + 5.209659603880365d-11*tv07
     $  + 9.695598369668428d-13*tv08
     $  - 6.499477923774285d-14*tv09
     $  - 6.208070878371570d-16*tv10

      CrH3YZ(3,3,1,03) = 
     $  + 2.790939377261168d-02
     $  - 5.213948867357478d-03*tv01
     $  + 8.745444017786864d-05*tv02
     $  + 7.436197606446024d-06*tv03
     $  - 2.461349140178591d-07*tv04
     $  - 9.778020563633743d-09*tv05
     $  + 4.854560504720140d-10*tv06
     $  + 1.033369746971827d-11*tv07
     $  - 7.788458563191642d-13*tv08
     $  - 8.274105927589359d-15*tv09
     $  + 1.091503633976155d-15*tv10

      CrH3YZ(3,3,1,04) = 
     $  - 2.595844546189753d-03
     $  + 1.739272472161000d-04*tv01
     $  + 2.225977497797174d-05*tv02
     $  - 9.816331396371685d-07*tv03
     $  - 4.881784136626932d-08*tv04
     $  + 2.906516434326753d-09*tv05
     $  + 7.226145194735283d-11*tv06
     $  - 6.219868471226704d-12*tv07
     $  - 7.443031935972853d-14*tv08
     $  + 1.089839108678997d-14*tv09

      CrH3YZ(3,3,1,05) = 
     $  + 6.839873477527351d-05
     $  + 3.545976925129179d-05*tv01
     $  - 2.344328222333049d-06*tv02
     $  - 1.558707134630298d-07*tv03
     $  + 1.159131279611521d-08*tv04
     $  + 3.463789089937196d-10*tv05
     $  - 3.475294122308582d-11*tv06
     $  - 4.760603758234920d-13*tv07
     $  + 7.832220187779029d-14*tv08
     $  + 2.936604065341490d-16*tv09
     $  - 1.462176873127819d-16*tv10

      CrH3YZ(3,3,1,06) = 
     $  + 1.166462855855799d-05
     $  - 3.094999390673744d-06*tv01
     $  - 3.105890504118864d-07*tv02
     $  + 3.077157243697331d-08*tv03
     $  + 1.152376514155890d-09*tv04
     $  - 1.385950193505433d-10*tv05
     $  - 2.219849170087594d-12*tv06
     $  + 4.167455321435399d-13*tv07
     $  + 1.765911053436994d-15*tv08
     $  - 9.728812011428030d-16*tv09

      CrH3YZ(3,3,1,07) = 
     $  - 8.581070685645542d-07
     $  - 3.523325141118296d-07*tv01
     $  + 5.235694325589303d-08*tv02
     $  + 2.626400644619789d-09*tv03
     $  - 3.943237789092203d-10*tv04
     $  - 7.602816299181527d-12*tv05
     $  + 1.661990670935181d-12*tv06
     $  + 8.095146220716389d-15*tv07
     $  - 4.991454450279079d-15*tv08

      CrH3YZ(3,3,1,08) = 
     $  - 8.612281734832783d-08
     $  + 5.157326176379428d-08*tv01
     $  + 3.920622359260627d-09*tv02
     $  - 7.836814920961594d-10*tv03
     $  - 1.897860766598325d-11*tv04
     $  + 4.966065725232631d-12*tv05
     $  + 2.843308438269730d-14*tv06
     $  - 1.990628585761782d-14*tv07
     $  + 1.000736822488402d-16*tv08

      CrH3YZ(3,3,1,09) = 
     $  + 1.094375751252139d-08
     $  + 3.451334079647939d-09*tv01
     $  - 1.033934647755345d-09*tv02
     $  - 3.366357302936628d-11*tv03
     $  + 1.097400837145072d-11*tv04
     $  + 7.617393102163103d-14*tv05
     $  - 6.169324128895781d-14*tv06
     $  + 3.523718772926259d-16*tv07
     $  + 2.374253295816334d-16*tv08

      CrH3YZ(3,3,1,10) = 
     $  + 6.701453331954942d-10
     $  - 8.097366116384871d-10*tv01
     $  - 4.024296546837790d-11*tv02
     $  + 1.741110656545768d-11*tv03
     $  + 1.533221664683166d-13*tv04
     $  - 1.473068292410587d-13*tv05
     $  + 9.743041164467762d-16*tv06
     $  + 7.569496329855516d-16*tv07

      CrH3YZ(3,3,1,11) = 
     $  - 1.378608277554556d-10
     $  - 2.904250707938135d-11*tv01
     $  + 1.872802144861837d-11*tv02
     $  + 2.251041704867934d-13*tv03
     $  - 2.659129862615941d-13*tv04
     $  + 2.090110327434037d-15*tv05
     $  + 1.917612487804300d-15*tv06

      CrH3YZ(3,3,1,12) = 
     $  - 4.726165220124604d-12
     $  + 1.213460786945635d-11*tv01
     $  + 2.285052528969481d-13*tv02
     $  - 3.507397629227845d-13*tv03
     $  + 3.401082449474431d-15*tv04
     $  + 3.810885463754538d-15*tv05

      CrH3YZ(3,3,1,13) = 
     $  + 1.707888248051060d-12
     $  + 1.446541732500186d-13*tv01
     $  - 3.179490021995809d-13*tv02
     $  + 4.035556069726404d-15*tv03
     $  + 5.810778944546838d-15*tv04
     $  - 1.503915751470123d-16*tv05

      CrH3YZ(3,3,1,14) = 
     $  + 2.228876267489598d-14
     $  - 1.751329194881897d-13*tv01
     $  + 3.247761562504170d-15*tv02
     $  + 6.552224623129164d-15*tv03
     $  - 2.112865004999960d-16*tv04

      CrH3YZ(3,3,1,15) = 
     $  - 2.078864055816251d-14
     $  + 1.512391938386859d-15*tv01
     $  + 5.124830747773754d-15*tv02
     $  - 2.197475606040466d-16*tv03
     $  - 1.169489495286368d-16*tv04

      CrH3YZ(3,3,1,16) = 
     $  + 1.100992774264988d-16
     $  + 2.447356879343743d-15*tv01
     $  - 1.578994296332890d-16*tv02
     $  - 1.150584943510849d-16*tv03

      CrH3YZ(3,3,1,17) = 
     $  + 2.484133348870723d-16
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CrH4YZ(0,0,3,1,01) = 
     $  + 6.224694552604316d-01
     $  - 1.231893849147407d-01*tv01
     $  + 4.323485407234559d-03*tv02
     $  - 2.013333506542613d-05*tv03
     $  - 2.436051393807966d-06*tv04
     $  + 2.482888902259004d-08*tv05
     $  + 1.935975648316861d-09*tv06
     $  - 2.845440438346271d-11*tv07
     $  - 1.590223275140079d-12*tv08
     $  + 3.073351448049326d-14*tv09
     $  + 1.294967440691163d-15*tv10

      CrH4YZ(0,0,3,1,02) = 
     $  - 2.058819495355955d-01
     $  + 3.896563361329030d-02*tv01
     $  - 1.246118300612591d-03*tv02
     $  + 7.961992720556585d-07*tv03
     $  + 7.471122247758442d-07*tv04
     $  - 1.181552517111893d-09*tv05
     $  - 6.439454402104749d-10*tv06
     $  + 1.648376773820790d-12*tv07
     $  + 5.853765168949196d-13*tv08
     $  - 2.175159415417055d-15*tv09
     $  - 5.388039931418827d-16*tv10

      CrH4YZ(0,0,3,1,03) = 
     $  + 4.167102168733283d-02
     $  - 7.288693956503049d-03*tv01
     $  + 1.953277756707604d-04*tv02
     $  + 1.282851449223983d-06*tv03
     $  - 1.165044929879110d-07*tv04
     $  - 1.671080627425664d-09*tv05
     $  + 1.008210740930963d-10*tv06
     $  + 2.041764100365953d-12*tv07
     $  - 9.349402281400448d-14*tv08
     $  - 2.367835755574825d-15*tv09

      CrH4YZ(0,0,3,1,04) = 
     $  - 5.444687643746143d-03
     $  + 8.132283502499821d-04*tv01
     $  - 1.380946545300094d-05*tv02
     $  - 3.590569131455226d-07*tv03
     $  + 4.567005474774248d-09*tv04
     $  + 4.700130547037892d-10*tv05
     $  + 2.777727960483807d-13*tv06
     $  - 5.840050296557643d-13*tv07
     $  - 5.027436443017671d-15*tv08
     $  + 6.999532553435534d-16*tv09

      CrH4YZ(0,0,3,1,05) = 
     $  + 4.192661355013063d-04
     $  - 3.885613818191339d-05*tv01
     $  - 5.843985053910395d-07*tv02
     $  + 3.213555347157761d-08*tv03
     $  + 1.558368403229585d-09*tv04
     $  - 2.902712399325022d-11*tv05
     $  - 2.788634552570994d-12*tv06
     $  + 1.679847212362364d-14*tv07
     $  + 4.309791023762373d-15*tv08

      CrH4YZ(0,0,3,1,06) = 
     $  - 8.226764699622434d-06
     $  - 2.526999448296613d-06*tv01
     $  + 1.751561028783944d-07*tv02
     $  + 2.678224346989376d-09*tv03
     $  - 2.178079016849012d-10*tv04
     $  - 8.588966647142997d-12*tv05
     $  + 2.727533472046266d-13*tv06
     $  + 1.865900431497107d-14*tv07
     $  - 2.895082939738331d-16*tv08

      CrH4YZ(0,0,3,1,07) = 
     $  - 1.674539797583691d-06
     $  + 4.109683429933484d-07*tv01
     $  - 7.747477695119160d-10*tv02
     $  - 7.463652011683309d-10*tv03
     $  - 1.551689162499709d-11*tv04
     $  + 1.481564146352802d-12*tv05
     $  + 5.652759583342681d-14*tv06
     $  - 2.435250622989534d-15*tv07
     $  - 1.388821001822666d-16*tv08

      CrH4YZ(0,0,3,1,08) = 
     $  + 9.700321755887067d-08
     $  + 9.770677357794099d-09*tv01
     $  - 2.292021895375040d-09*tv02
     $  - 1.203495461975641d-12*tv03
     $  + 5.435151158230427d-12*tv04
     $  + 9.989769319199219d-14*tv05
     $  - 1.220501582879999d-14*tv06
     $  - 4.014707531606345d-16*tv07

      CrH4YZ(0,0,3,1,09) = 
     $  + 1.235322706065553d-08
     $  - 4.767753257819921d-09*tv01
     $  + 9.382061637146827d-11*tv02
     $  + 1.323330222380474d-11*tv03
     $  + 4.726430161200637d-15*tv04
     $  - 4.218829523258277d-14*tv05
     $  - 6.888488669751635d-16*tv06
     $  + 1.074307675892306d-16*tv07

      CrH4YZ(0,0,3,1,10) = 
     $  - 1.230028699634337d-09
     $  + 2.958212480599427d-11*tv01
     $  + 2.966617309102469d-11*tv02
     $  - 5.830412609974857d-13*tv03
     $  - 1.098763216226854d-13*tv04
     $  + 1.181873843589623d-16*tv05
     $  + 3.683396389508528d-16*tv06

      CrH4YZ(0,0,3,1,11) = 
     $  - 1.049929364628653d-10
     $  + 5.696102953228367d-11*tv01
     $  - 2.250676630736222d-12*tv02
     $  - 2.089816915097073d-13*tv03
     $  + 4.984253561005359d-15*tv04
     $  + 9.602744867404824d-16*tv05

      CrH4YZ(0,0,3,1,12) = 
     $  + 1.584308766510675d-11
     $  - 1.799726305225081d-12*tv01
     $  - 3.751452792449101d-13*tv02
     $  + 1.779741568330893d-14*tv03
     $  + 1.956081362879496d-15*tv04

      CrH4YZ(0,0,3,1,13) = 
     $  + 9.259708076176324d-13
     $  - 6.880814548927281d-13*tv01
     $  + 4.283568152153262d-14*tv02
     $  + 3.038225430142368d-15*tv03
     $  - 1.722175031350997d-16*tv04

      CrH4YZ(0,0,3,1,14) = 
     $  - 2.051957303893079d-13
     $  + 3.809224328215537d-14*tv01
     $  + 4.594285782239708d-15*tv02
     $  - 3.985738914294093d-16*tv03

      CrH4YZ(0,0,3,1,15) = 
     $  - 8.047659768387162d-15
     $  + 8.342015759527716d-15*tv01
     $  - 7.400860878745027d-16*tv02

      CrH4YZ(0,0,3,1,16) = 
     $  + 2.668450287055109d-15
     $  - 6.606312545032964d-16*tv01

      CrH4YZ(0,0,3,1,17) = 
     $  - 1.010838493672980d-16*tv01

      CrH4YZ(0,1,3,1,01) = 
     $  - 1.582031250000000d-01
     $  - 2.109375000000000d-01*tv01
     $  - 5.273437500000000d-02*tv02

      CrH4YZ(0,1,3,1,02) = 
     $  + 1.463761087988309d-01
     $  + 1.645277157848610d-02*tv01
     $  + 1.069834882635647d-02*tv02
     $  - 3.775000324767399d-06*tv03
     $  - 4.567596363389936d-07*tv04
     $  + 4.655416691735632d-09*tv05
     $  + 3.629954340594114d-10*tv06
     $  - 5.335200821899258d-12*tv07
     $  - 2.981668640887649d-13*tv08
     $  + 5.762533965092486d-15*tv09
     $  + 2.428063951295930d-16*tv10

      CrH4YZ(0,1,3,1,03) = 
     $  - 4.192098462465744d-02
     $  + 3.918143682384962d-03*tv01
     $  - 1.036683997781743d-03*tv02
     $  - 6.783503238399637d-07*tv03
     $  + 1.348474991738663d-07*tv04
     $  + 7.700033338966103d-10*tv05
     $  - 1.238996215553054d-10*tv06
     $  - 7.843412808678845d-13*tv07
     $  + 1.205837896094486d-13*tv08
     $  + 7.155803528007697d-16*tv09
     $  - 1.188997322648959d-16*tv10

      CrH4YZ(0,1,3,1,04) = 
     $  + 6.162131548126607d-03
     $  - 9.612373339117173d-04*tv01
     $  + 1.818264697599789d-05*tv02
     $  + 4.295630364537118d-07*tv03
     $  - 9.274391244021815d-09*tv04
     $  - 5.836887121894296d-10*tv05
     $  + 6.725652212294399d-12*tv06
     $  + 7.465000985428454d-13*tv07
     $  - 5.400195227282624d-15*tv08
     $  - 9.072429755653495d-16*tv09

      CrH4YZ(0,1,3,1,05) = 
     $  - 4.411068529430407d-04
     $  + 5.888475330786793d-05*tv01
     $  + 5.563285234176728d-06*tv02
     $  - 4.567470431261696d-08*tv03
     $  - 2.066518485045220d-09*tv04
     $  + 5.648894168055967d-11*tv05
     $  + 3.632498411614847d-12*tv06
     $  - 6.799487872845094d-14*tv07
     $  - 5.572540266318193d-15*tv08

      CrH4YZ(0,1,3,1,06) = 
     $  - 7.631180820016063d-08
     $  + 4.496048259464194d-06*tv01
     $  - 2.388161172990494d-07*tv02
     $  - 4.346433233591655d-09*tv03
     $  + 3.237363652786200d-10*tv04
     $  + 1.206785991980197d-11*tv05
     $  - 4.970963367294342d-13*tv06
     $  - 2.493984890588275d-14*tv07
     $  + 7.380911686914357d-16*tv08

      CrH4YZ(0,1,3,1,07) = 
     $  + 2.083388298700840d-06
     $  - 6.159972972196008d-07*tv01
     $  - 4.989346959874889d-08*tv02
     $  + 1.032600504909388d-09*tv03
     $  + 2.811578812316235d-11*tv04
     $  - 2.261582765041820d-12*tv05
     $  - 8.319855954643229d-14*tv06
     $  + 4.336190341666657d-15*tv07
     $  + 1.880370078290885d-16*tv08

      CrH4YZ(0,1,3,1,08) = 
     $  + 4.864362331324250d-09
     $  - 3.722779334809717d-08*tv01
     $  + 3.276176357910319d-09*tv02
     $  + 3.683666329509158d-11*tv03
     $  - 7.848556203860851d-12*tv04
     $  - 1.931872660933797d-13*tv05
     $  + 1.909804799338690d-14*tv06
     $  + 6.078305247097954d-16*tv07

      CrH4YZ(0,1,3,1,09) = 
     $  - 1.877955200762487d-08
     $  + 7.193656476250781d-09*tv01
     $  + 5.009461715202336d-10*tv02
     $  - 1.899596297539190d-11*tv03
     $  - 2.885888120671716d-13*tv04
     $  + 6.282087757106529d-14*tv05
     $  + 1.430937987388633d-15*tv06
     $  - 1.714222431045106d-16*tv07

      CrH4YZ(0,1,3,1,10) = 
     $  - 4.574449390158475d-11
     $  + 3.564677687662050d-10*tv01
     $  - 4.514779680373475d-11*tv02
     $  - 1.189968380843091d-13*tv03
     $  + 1.632943464552268d-13*tv04
     $  + 2.129672603975719d-15*tv05
     $  - 5.658783143807745d-16*tv06

      CrH4YZ(0,1,3,1,11) = 
     $  + 1.999784129900399d-10
     $  - 8.770500112590417d-11*tv01
     $  - 5.210651757947051d-12*tv02
     $  + 3.206949828804657d-13*tv03
     $  + 9.351377047330554d-16*tv04
     $  - 1.472763779655035d-15*tv05

      CrH4YZ(0,1,3,1,12) = 
     $  + 3.369401751773667d-13
     $  - 3.631220611176788d-12*tv01
     $  + 6.200996799776472d-13*tv02
     $  - 4.792758009871439d-15*tv03
     $  - 3.096729225885790d-15*tv04

      CrH4YZ(0,1,3,1,13) = 
     $  - 2.289476205543080d-12
     $  + 1.093972330171317d-12*tv01
     $  + 5.444511594074992d-14*tv02
     $  - 5.138661454245280d-15*tv03

      CrH4YZ(0,1,3,1,14) = 
     $  - 2.057957728605946d-15
     $  + 3.822223851405402d-14*tv01
     $  - 8.473164473108041d-15*tv02
     $  + 1.690938139112149d-16*tv03

      CrH4YZ(0,1,3,1,15) = 
     $  + 2.731818891681611d-14
     $  - 1.384609069865606d-14*tv01
     $  - 5.595180973654480d-16*tv02

      CrH4YZ(0,1,3,1,16) = 
     $  - 4.101068267229015d-16*tv01
     $  + 1.151287508624106d-16*tv02

      CrH4YZ(0,1,3,1,17) = 
     $  - 3.351099999692601d-16
     $  + 1.770887449510829d-16*tv01

      CrH4YZ(0,3,0,1,01) = 
     $  + 6.782092247974649d-01
     $  - 7.216113782053397d-02*tv01
     $  - 2.960951508100631d-04*tv02
     $  + 7.475412348671377d-05*tv03
     $  + 4.033350131834205d-07*tv04
     $  - 8.380850632602890d-08*tv05
     $  - 3.990598293362059d-10*tv06
     $  + 8.854815831586199d-11*tv07
     $  + 2.972134190900594d-13*tv08
     $  - 8.930056977074862d-14*tv09
     $  - 1.323707487429359d-16*tv10

      CrH4YZ(0,3,0,1,02) = 
     $  - 2.247577070735084d-01
     $  + 2.157687219983906d-02*tv01
     $  + 2.190857280840145d-04*tv02
     $  - 2.213805251380922d-05*tv03
     $  - 3.701032717070619d-07*tv04
     $  + 2.492525738396512d-08*tv05
     $  + 4.949535000575357d-10*tv06
     $  - 2.693524570982540d-11*tv07
     $  - 5.822229711116150d-13*tv08
     $  + 2.835668777746503d-14*tv09
     $  + 6.328660669078847d-16*tv10

      CrH4YZ(0,3,0,1,03) = 
     $  + 4.560918616508379d-02
     $  - 3.622568686138013d-03*tv01
     $  - 7.506338220581796d-05*tv02
     $  + 3.184519386640856d-06*tv03
     $  + 1.362123851683306d-07*tv04
     $  - 2.871705203135598d-09*tv05
     $  - 1.983774516109882d-10*tv06
     $  + 2.267863485879125d-12*tv07
     $  + 2.577097984836005d-13*tv08
     $  - 1.528595199895374d-15*tv09
     $  - 3.137482522773767d-16*tv10

      CrH4YZ(0,3,0,1,04) = 
     $  - 5.976662417572462d-03
     $  + 3.087401510520202d-04*tv01
     $  + 1.402876015645797d-05*tv02
     $  - 3.551071146558982d-08*tv03
     $  - 2.442170706671649d-08*tv04
     $  - 3.821497396637336d-10*tv05
     $  + 3.364756877111935d-11*tv06
     $  + 9.819816032013183d-13*tv07
     $  - 4.098122006098033d-14*tv08
     $  - 1.736280071723243d-15*tv09

      CrH4YZ(0,3,0,1,05) = 
     $  + 4.606993498425587d-04
     $  + 2.082539018864076d-06*tv01
     $  - 1.176780984146542d-06*tv02
     $  - 6.648027682795499d-08*tv03
     $  + 9.820473439804669d-10*tv04
     $  + 1.691204354869845d-10*tv05
     $  + 8.739974625954655d-13*tv06
     $  - 3.068779050005213d-13*tv07
     $  - 4.947369283002362d-15*tv08
     $  + 4.748203389237539d-16*tv09

      CrH4YZ(0,3,0,1,06) = 
     $  - 8.596871668627775d-06
     $  - 3.140935203606653d-06*tv01
     $  - 6.328019343061374d-08*tv02
     $  + 8.745969642581864d-09*tv03
     $  + 4.550347630348395d-10*tv04
     $  - 1.259439318514984d-11*tv05
     $  - 1.334376476191540d-12*tv06
     $  + 7.407582669872441d-15*tv07
     $  + 2.851742078517537d-15*tv08

      CrH4YZ(0,3,0,1,07) = 
     $  - 1.961415599006270d-06
     $  + 1.468536078743788d-07*tv01
     $  + 2.324963885610619d-08*tv02
     $  + 4.637269368946199d-10*tv03
     $  - 7.371523910270515d-11*tv04
     $  - 3.417736992985406d-12*tv05
     $  + 1.410433276772532d-13*tv06
     $  + 1.120453233190817d-14*tv07
     $  - 1.826905605386214d-16*tv08

      CrH4YZ(0,3,0,1,08) = 
     $  + 1.172480280867033d-07
     $  + 3.180085688839991d-08*tv01
     $  - 7.056664950330948d-10*tv02
     $  - 2.099968517822797d-10*tv03
     $  - 3.787493991950653d-12*tv04
     $  + 6.979466502963819d-13*tv05
     $  + 2.767584541172092d-14*tv06
     $  - 1.583904244978595d-15*tv07

      CrH4YZ(0,3,0,1,09) = 
     $  + 1.424557642710649d-08
     $  - 3.292456581248010d-09*tv01
     $  - 3.323474876726779d-10*tv02
     $  + 6.041715863001671d-12*tv03
     $  + 2.006352435690235d-12*tv04
     $  + 2.964485951191446d-14*tv05
     $  - 7.006150731030295d-15*tv06
     $  - 2.310719885737281d-16*tv07

      CrH4YZ(0,3,0,1,10) = 
     $  - 1.552593416031555d-09
     $  - 2.967780701455102d-10*tv01
     $  + 3.004857296032188d-11*tv02
     $  + 3.665294010107370d-12*tv03
     $  - 5.926422099077487d-14*tv04
     $  - 2.037985382994438d-14*tv05
     $  - 2.191002443662765d-16*tv06

      CrH4YZ(0,3,0,1,11) = 
     $  - 1.125207341453622d-10
     $  + 5.572784995847907d-11*tv01
     $  + 3.974203931613694d-12*tv02
     $  - 3.235248803639286d-13*tv03
     $  - 4.030165194868047d-14*tv04
     $  + 6.927778193462481d-16*tv05
     $  + 2.148584845696029d-16*tv06

      CrH4YZ(0,3,0,1,12) = 
     $  + 2.015089899807810d-11
     $  + 2.284321153369096d-12*tv01
     $  - 6.703033958552878d-13*tv02
     $  - 5.243547252351454d-14*tv03
     $  + 3.614245793118879d-15*tv04
     $  + 4.510474994613600d-16*tv05

      CrH4YZ(0,3,0,1,13) = 
     $  + 8.488942010073130d-13
     $  - 8.505714753523714d-13*tv01
     $  - 3.798445259218301d-14*tv02
     $  + 8.480371052228871d-15*tv03
     $  + 6.586492563025980d-16*tv04

      CrH4YZ(0,3,0,1,14) = 
     $  - 2.582590910979099d-13
     $  - 8.290523271495517d-15*tv01
     $  + 1.233886493731667d-14*tv02
     $  + 6.003774309489511d-16*tv03
     $  - 1.067079253054987d-16*tv04

      CrH4YZ(0,3,0,1,15) = 
     $  - 5.171549519263950d-15
     $  + 1.227714638386967d-14*tv01
     $  + 1.953416502743954d-16*tv02
     $  - 1.782266113169255d-16*tv03

      CrH4YZ(0,3,0,1,16) = 
     $  + 3.283709041408863d-15
     $  - 1.831488929459566d-16*tv01
     $  - 2.055088591416236d-16*tv02

      CrH4YZ(0,3,0,1,17) = 
     $  - 1.707405859566204d-16*tv01

      CrH4YZ(0,3,1,1,01) = 
     $  - 1.051640180255739d-01
     $  - 9.402033994674527d-02*tv01
     $  + 1.074023882206285d-02*tv02
     $  - 4.013233767853031d-04*tv03
     $  + 2.341932090376803d-06*tv04
     $  + 2.235429037607168d-07*tv05
     $  - 2.688036179524982d-09*tv06
     $  - 1.760134327804249d-10*tv07
     $  + 2.962886005804824d-12*tv08
     $  + 1.431994939062815d-13*tv09
     $  - 3.121094194574164d-15*tv10
     $  - 1.153473109090728d-16*tv11

      CrH4YZ(0,3,1,1,02) = 
     $  + 8.881967944443161d-02
     $  + 1.602063333644289d-02*tv01
     $  - 2.405994911923911d-03*tv02
     $  + 8.082475663882226d-05*tv03
     $  + 2.206118090173979d-07*tv04
     $  - 5.336342292536949d-08*tv05
     $  - 3.638042838949498d-10*tv06
     $  + 5.133923295137732d-11*tv07
     $  + 4.338198248620251d-13*tv08
     $  - 5.226030864221305d-14*tv09
     $  - 4.396273397199000d-16*tv10

      CrH4YZ(0,3,1,1,03) = 
     $  - 2.989668611261322d-02
     $  - 3.223866249748064d-04*tv01
     $  + 2.706404797031486d-04*tv02
     $  - 6.945869736109578d-06*tv03
     $  - 1.584518632719116d-07*tv04
     $  + 3.667698158557621d-09*tv05
     $  + 2.556935794055094d-10*tv06
     $  - 2.426173411287071d-12*tv07
     $  - 3.579734396483880d-13*tv08
     $  + 1.647194128679514d-15*tv09
     $  + 4.621018345716013d-16*tv10

      CrH4YZ(0,3,1,1,04) = 
     $  + 5.525215544999151d-03
     $  - 2.550564742218771d-04*tv01
     $  - 1.256980702012707d-05*tv02
     $  - 9.966081891801002d-08*tv03
     $  + 2.112038865675386d-08*tv04
     $  + 7.009979861933503d-10*tv05
     $  - 2.843220576457334d-11*tv06
     $  - 1.614515292340092d-12*tv07
     $  + 3.506666072451049d-14*tv08
     $  + 2.827920509639374d-15*tv09

      CrH4YZ(0,3,1,1,05) = 
     $  - 5.522412000344111d-04
     $  + 2.433149091417070d-05*tv01
     $  - 1.343805074449985d-07*tv02
     $  + 7.529271826083681d-08*tv03
     $  - 4.419579465207752d-12*tv04
     $  - 1.458921977724761d-10*tv05
     $  - 3.610011819220703d-12*tv06
     $  + 2.621673219586133d-13*tv07
     $  + 1.094988648777512d-14*tv08
     $  - 4.220660232716212d-16*tv09

      CrH4YZ(0,3,1,1,06) = 
     $  + 1.444530931093577d-05
     $  + 2.457788300475064d-06*tv01
     $  + 2.325969602204591d-08*tv02
     $  - 5.846827946079168d-09*tv03
     $  - 3.296720476247282d-10*tv04
     $  + 2.516158991356611d-12*tv05
     $  + 1.062920644932739d-12*tv06
     $  + 2.008439616118845d-14*tv07
     $  - 2.421735239231810d-15*tv08

      CrH4YZ(0,3,1,1,07) = 
     $  + 2.869284996671441d-06
     $  - 4.490803097751143d-07*tv01
     $  - 8.085775849272467d-09*tv02
     $  - 2.885140743761257d-10*tv03
     $  + 4.659511662637144d-11*tv04
     $  + 2.337219980540406d-12*tv05
     $  - 4.145673833401748d-14*tv06
     $  - 8.820930835515270d-15*tv07

      CrH4YZ(0,3,1,1,08) = 
     $  - 1.965159868538956d-07
     $  - 2.263510568892002d-08*tv01
     $  + 1.601665919224585d-09*tv02
     $  + 1.489534189483089d-10*tv03
     $  + 6.090596420717112d-14*tv04
     $  - 4.030485943967155d-13*tv05
     $  - 1.722398702890538d-14*tv06
     $  + 5.285098239000479d-16*tv07

      CrH4YZ(0,3,1,1,09) = 
     $  - 2.452369167434141d-08
     $  + 7.229632704848549d-09*tv01
     $  + 1.298147130721957d-10*tv02
     $  - 1.309337373537741d-11*tv03
     $  - 1.331200970379661d-12*tv04
     $  + 5.258272604075876d-15*tv05
     $  + 3.790342492316484d-15*tv06
     $  + 1.371062673991601d-16*tv07

      CrH4YZ(0,3,1,1,10) = 
     $  + 2.635129298902431d-09
     $  + 1.713802934419972d-10*tv01
     $  - 4.801544670227148d-11*tv02
     $  - 2.422339874002076d-12*tv03
     $  + 1.553672826621185d-13*tv04
     $  + 1.237775439152504d-14*tv05
     $  - 1.178610598568829d-16*tv06

      CrH4YZ(0,3,1,1,11) = 
     $  + 2.357395048378631d-10
     $  - 1.096894991805496d-10*tv01
     $  - 1.078510511097394d-12*tv02
     $  + 5.008399728773641d-13*tv03
     $  + 2.455529037462006d-14*tv04
     $  - 1.654280433681527d-15*tv05
     $  - 1.198754896659104d-16*tv06

      CrH4YZ(0,3,1,1,12) = 
     $  - 3.494624383769371d-11
     $  - 5.415527043404065d-13*tv01
     $  + 1.005766791320644d-12*tv02
     $  + 2.784496617974272d-14*tv03
     $  - 5.858890905002112d-15*tv04
     $  - 2.472081958268898d-16*tv05

      CrH4YZ(0,3,1,1,13) = 
     $  - 2.373396647014386d-12
     $  + 1.607515351535784d-12*tv01
     $  - 5.166481575043522d-15*tv02
     $  - 1.229506381494915d-14*tv03
     $  - 3.085463905835097d-16*tv04

      CrH4YZ(0,3,1,1,14) = 
     $  + 4.604092345602179d-13
     $  - 1.631937375593625d-14*tv01
     $  - 1.829389815125474d-14*tv02
     $  - 1.253021884176308d-16*tv03
     $  + 1.553450973034915d-16*tv04

      CrH4YZ(0,3,1,1,15) = 
     $  + 2.441012365847222d-14
     $  - 2.301175202545343d-14*tv01
     $  + 4.658853410216686d-16*tv02
     $  + 2.535905151991866d-16*tv03

      CrH4YZ(0,3,1,1,16) = 
     $  - 6.044916319359815d-15
     $  + 5.360888586909107d-16*tv01
     $  + 3.072823282034839d-16*tv02

      CrH4YZ(0,3,1,1,17) = 
     $  - 2.533787844095462d-16
     $  + 3.236921508853335d-16*tv01

      CrH4YZ(0,3,3,1,01) = 
     $  + 5.730452067718910d-01
     $  - 1.661814777672792d-01*tv01
     $  + 1.044414367125278d-02*tv02
     $  - 3.265692532985893d-04*tv03
     $  + 2.745267103560223d-06*tv04
     $  + 1.397343974346879d-07*tv05
     $  - 3.087096008861188d-09*tv06
     $  - 8.746527446456293d-11*tv07
     $  + 3.260099424894883d-12*tv08
     $  + 5.389892413553291d-14*tv09
     $  - 3.253464943317100d-15*tv10

      CrH4YZ(0,3,3,1,02) = 
     $  - 1.910262200032095d-01
     $  + 5.205827699575692d-02*tv01
     $  - 2.937286638040633d-03*tv02
     $  + 7.219430687997132d-05*tv03
     $  + 1.833226331146222d-07*tv04
     $  - 4.469672727412311d-08*tv05
     $  - 3.294163965411016d-11*tv06
     $  + 4.244803698150737d-11*tv07
     $  - 1.264887236843692d-13*tv08
     $  - 4.260343789428384d-14*tv09
     $  + 2.996211329533778d-16*tv10

      CrH4YZ(0,3,3,1,03) = 
     $  + 3.915202354671802d-02
     $  - 9.559381024497580d-03*tv01
     $  + 4.370936498427983d-04*tv02
     $  - 5.364555018929906d-06*tv03
     $  - 2.136652266041989d-07*tv04
     $  + 3.548141710818866d-09*tv05
     $  + 2.665258436811110d-10*tv06
     $  - 4.434153595119829d-12*tv07
     $  - 3.298712115465942d-13*tv08
     $  + 6.208482785139105d-15*tv09
     $  + 3.906000408990334d-16*tv10

      CrH4YZ(0,3,3,1,04) = 
     $  - 5.224378765233221d-03
     $  + 1.022115687987352d-03*tv01
     $  - 2.557944985844949d-05*tv02
     $  - 6.856165319240954d-07*tv03
     $  + 2.723147413742331d-08*tv04
     $  + 1.209806167161747d-09*tv05
     $  - 4.178455130699708d-11*tv06
     $  - 1.913244270763033d-12*tv07
     $  + 6.577124852944470d-14*tv08
     $  + 2.738637633216697d-15*tv09
     $  - 1.004436183298664d-16*tv10

      CrH4YZ(0,3,3,1,05) = 
     $  + 4.196533531610850d-04
     $  - 4.022756637803572d-05*tv01
     $  - 2.236405706003174d-06*tv02
     $  + 1.408650711188094d-07*tv03
     $  + 3.430802601251965d-09*tv04
     $  - 2.526632370843620d-10*tv05
     $  - 7.547719869839680d-12*tv06
     $  + 4.758438515944002d-13*tv07
     $  + 1.365850224852938d-14*tv08
     $  - 8.409542034026681d-16*tv09

      CrH4YZ(0,3,3,1,06) = 
     $  - 1.036765191515589d-05
     $  - 4.434675752787437d-06*tv01
     $  + 4.120697881525757d-07*tv02
     $  + 5.595660603001757d-09*tv03
     $  - 1.016038597383217d-09*tv04
     $  - 1.933123975280816d-11*tv05
     $  + 2.456241617863317d-12*tv06
     $  + 4.680297109291379d-14*tv07
     $  - 5.253318661825202d-15*tv08

      CrH4YZ(0,3,3,1,07) = 
     $  - 1.559763709172121d-06
     $  + 5.405269293742115d-07*tv01
     $  + 1.171066802120890d-08*tv02
     $  - 3.058467583054983d-09*tv03
     $  - 2.268572390565029d-11*tv04
     $  + 9.389460058012581d-12*tv05
     $  + 9.055508660796635d-14*tv06
     $  - 2.507938769261240d-14*tv07
     $  - 2.315520306203846d-16*tv08

      CrH4YZ(0,3,3,1,08) = 
     $  + 1.174212802858109d-07
     $  + 2.602417909084145d-08*tv01
     $  - 6.447276052075670d-09*tv02
     $  + 2.928232881415824d-11*tv03
     $  + 2.657256552841941d-11*tv04
     $  - 2.181515884107430d-14*tv05
     $  - 9.222819726258524d-14*tv06
     $  + 2.616326493141368d-16*tv08

      CrH4YZ(0,3,3,1,09) = 
     $  + 9.976822339927007d-09
     $  - 7.213326536742297d-09*tv01
     $  + 7.006547537281054d-11*tv02
     $  + 5.806910879069846d-11*tv03
     $  - 7.958475039365329d-13*tv04
     $  - 2.626460051313921d-13*tv05
     $  + 2.659458378603246d-15*tv06
     $  + 9.415504088662339d-16*tv07

      CrH4YZ(0,3,3,1,10) = 
     $  - 1.403966779079962d-09
     $  - 7.621942121551329d-11*tv01
     $  + 9.608177440418889d-11*tv02
     $  - 2.781486257166228d-12*tv03
     $  - 5.760838015783694d-13*tv04
     $  + 1.379864316145787d-14*tv05
     $  + 2.694415425751684d-15*tv06

      CrH4YZ(0,3,3,1,11) = 
     $  - 6.661349860900077d-11
     $  + 9.418974218092078d-11*tv01
     $  - 4.219290880338205d-12*tv02
     $  - 9.957264709177385d-13*tv03
     $  + 4.146405463437096d-14*tv04
     $  + 6.114008080004631d-15*tv05
     $  - 2.107579144182159d-16*tv06

      CrH4YZ(0,3,3,1,12) = 
     $  + 1.693407696596651d-11
     $  - 1.755467688550656d-12*tv01
     $  - 1.364672802597710d-12*tv02
     $  + 8.445929157009026d-14*tv03
     $  + 1.092581256859058d-14*tv04
     $  - 5.966684627423208d-16*tv05

      CrH4YZ(0,3,3,1,13) = 
     $  + 3.628378946478810d-13
     $  - 1.203117670383994d-12*tv01
     $  + 1.045213874742599d-13*tv02
     $  + 1.565194636057964d-14*tv03
     $  - 1.259026376127753d-15*tv04
     $  - 1.232479809297476d-16*tv05

      CrH4YZ(0,3,3,1,14) = 
     $  - 2.043072908222115d-13
     $  + 5.427174947848184d-14*tv01
     $  + 1.847137964775968d-14*tv02
     $  - 1.980974610782136d-15*tv03
     $  - 1.847227121829412d-16*tv04

      CrH4YZ(0,3,3,1,15) = 
     $  - 2.118930232916019d-16
     $  + 1.503598990959773d-14*tv01
     $  - 2.091677546575702d-15*tv02
     $  - 2.258452887937584d-16*tv03

      CrH4YZ(0,3,3,1,16) = 
     $  + 2.462755279893935d-15
     $  - 1.079937393016357d-15*tv01
     $  - 2.371159667151117d-16*tv02

      CrH4YZ(0,3,3,1,17) = 
     $  - 1.836133914620600d-16*tv01

      CrH4YZ(3,0,0,1,01) = 
     $  + 7.109364562679696d-01
     $  - 4.033936366218289d-02*tv01
     $  - 1.257269944639652d-03*tv02
     $  + 2.039473678502774d-05*tv03
     $  + 1.841808531297972d-06*tv04
     $  - 5.117457140468953d-09*tv05
     $  - 2.184697742940553d-09*tv06
     $  - 7.654674013703549d-12*tv07
     $  + 2.345506442448186d-12*tv08
     $  + 1.634472615355251d-14*tv09
     $  - 2.399688754830290d-15*tv10

      CrH4YZ(3,0,0,1,02) = 
     $  - 2.354651599405169d-01
     $  + 1.107894622777691d-02*tv01
     $  + 4.494529138813248d-04*tv02
     $  - 1.535019362465271d-06*tv03
     $  - 6.402373565675108d-07*tv04
     $  - 7.756567852743391d-09*tv05
     $  + 7.333878715197505d-10*tv06
     $  + 1.681523474151264d-11*tv07
     $  - 7.592473631621700d-13*tv08
     $  - 2.471065253937513d-14*tv09
     $  + 7.545556038208382d-16*tv10

      CrH4YZ(3,0,0,1,03) = 
     $  + 4.770193419966036d-02
     $  - 1.541611640764248d-03*tv01
     $  - 9.140098855431809d-05*tv02
     $  - 1.399217704909220d-06*tv03
     $  + 1.065018625572910d-07*tv04
     $  + 4.667113750709812d-09*tv05
     $  - 7.905087007080834d-11*tv06
     $  - 8.053700198287100d-12*tv07
     $  + 1.771576094457556d-14*tv08
     $  + 1.130052952123534d-14*tv09

      CrH4YZ(3,0,0,1,04) = 
     $  - 6.221834614257417d-03
     $  + 5.823059065058021d-05*tv01
     $  + 9.205054890476277d-06*tv02
     $  + 5.037808293314421d-07*tv03
     $  + 5.954976094178603d-10*tv04
     $  - 1.058189619400903d-09*tv05
     $  - 2.614199132538099d-11*tv06
     $  + 1.453866130945044d-12*tv07
     $  + 6.651695869218253d-14*tv08
     $  - 1.586162090245512d-15*tv09
     $  - 1.199284591384808d-16*tv10

      CrH4YZ(3,0,0,1,05) = 
     $  + 4.722087571498215d-04
     $  + 1.501736738837845d-05*tv01
     $  + 2.582407375769350d-07*tv02
     $  - 6.159196191486104d-08*tv03
     $  - 3.870953751868169d-09*tv04
     $  + 3.749082920547526d-11*tv05
     $  + 1.035146556580009d-11*tv06
     $  + 1.307480381994467d-13*tv07
     $  - 1.885358353442844d-14*tv08
     $  - 5.050981150956961d-16*tv09

      CrH4YZ(3,0,0,1,06) = 
     $  - 7.347718341774533d-06
     $  - 2.007432793707501d-06*tv01
     $  - 1.912821671018172d-07*tv02
     $  - 3.514457759969263d-09*tv03
     $  + 5.912034406629175d-10*tv04
     $  + 3.364358578889012d-11*tv05
     $  - 7.436949985189769d-13*tv06
     $  - 1.037281125624357d-13*tv07
     $  - 1.484888706441244d-16*tv08
     $  + 2.247509735772592d-16*tv09

      CrH4YZ(3,0,0,1,07) = 
     $  - 2.186164707877901d-06
     $  - 8.886263240956646d-08*tv01
     $  + 1.359483887892116d-08*tv02
     $  + 1.887376337905025d-09*tv03
     $  + 3.529105674768201d-11*tv04
     $  - 6.178638123073407d-12*tv05
     $  - 3.095888415451163d-13*tv06
     $  + 1.102402912260513d-14*tv07
     $  + 1.057261232280012d-15*tv08

      CrH4YZ(3,0,0,1,08) = 
     $  + 1.184332742186261d-07
     $  + 3.596151469569234d-08*tv01
     $  + 2.386889208171498d-09*tv02
     $  - 1.082152698237155d-10*tv03
     $  - 2.021790227620226d-11*tv04
     $  - 3.331218643748338d-13*tv05
     $  + 6.912514749083560d-14*tv06
     $  + 2.931889321770953d-15*tv07
     $  - 1.514406736875577d-16*tv08

      CrH4YZ(3,0,0,1,09) = 
     $  + 1.701714050353410d-08
     $  - 5.757215886251377d-10*tv01
     $  - 4.248518559924181d-10*tv02
     $  - 3.278898219261835d-11*tv03
     $  + 1.035633431488040d-12*tv04
     $  + 2.247868916354463d-13*tv05
     $  + 2.916811192537557d-15*tv06
     $  - 8.023805919517486d-16*tv07

      CrH4YZ(3,0,0,1,10) = 
     $  - 1.735083712204889d-09
     $  - 5.258397319666216d-10*tv01
     $  - 1.579485333250749d-11*tv02
     $  + 4.863249868355867d-12*tv03
     $  + 4.219035945562709d-13*tv04
     $  - 1.148980730144149d-14*tv05
     $  - 2.583317203115996d-15*tv06

      CrH4YZ(3,0,0,1,11) = 
     $  - 1.398831878242480d-10
     $  + 3.230411513409484d-11*tv01
     $  + 8.610234281005921d-12*tv02
     $  + 3.623749749346889d-13*tv03
     $  - 5.758675623473231d-14*tv04
     $  - 5.280373350839086d-15*tv05
     $  + 1.431960284405226d-16*tv06

      CrH4YZ(3,0,0,1,12) = 
     $  + 2.407744331955437d-11
     $  + 6.766009627246925d-12*tv01
     $  - 1.761115691808016d-13*tv02
     $  - 1.231647606817464d-13*tv03
     $  - 5.979972249466321d-15*tv04
     $  + 7.074454741562032d-16*tv05

      CrH4YZ(3,0,0,1,13) = 
     $  + 1.043495025465982d-12
     $  - 7.521463866654725d-13*tv01
     $  - 1.439651722646758d-13*tv02
     $  - 4.170055684785619d-16*tv03
     $  + 1.698966987537432d-15*tv04

      CrH4YZ(3,0,0,1,14) = 
     $  - 3.224814876773668d-13
     $  - 7.701925892347644d-14*tv01
     $  + 9.551115191379777d-15*tv02
     $  + 2.474422982834236d-15*tv03

      CrH4YZ(3,0,0,1,15) = 
     $  - 5.244694479116784d-15
     $  + 1.401460940323422d-14*tv01
     $  + 2.101954751790807d-15*tv02
     $  - 1.103324160361584d-16*tv03

      CrH4YZ(3,0,0,1,16) = 
     $  + 4.206294667094052d-15
     $  + 7.358632728897157d-16*tv01
     $  - 2.458535524813668d-16*tv02

      CrH4YZ(3,0,0,1,17) = 
     $  - 2.338603277532617d-16*tv01

      CrH4YZ(3,0,1,1,01) = 
     $  - 6.281912836661170d-02
     $  - 6.018486038956212d-02*tv01
     $  + 2.710393721264088d-03*tv02
     $  + 7.188895699974647d-05*tv03
     $  - 4.325158960784774d-06*tv04
     $  - 8.293222040753495d-08*tv05
     $  + 5.528376439317997d-09*tv06
     $  + 8.264487360086011d-11*tv07
     $  - 6.285070807594248d-12*tv08
     $  - 7.251764483596533d-14*tv09
     $  + 6.669456700387816d-15*tv10

      CrH4YZ(3,0,1,1,02) = 
     $  + 7.823324452747013d-02
     $  + 7.590679046315745d-03*tv01
     $  - 3.538946900557084d-04*tv02
     $  - 2.338178461375711d-05*tv03
     $  + 4.912816719935888d-07*tv04
     $  + 4.334183242507196d-08*tv05
     $  - 5.798979920037778d-10*tv06
     $  - 6.375447747411252d-11*tv07
     $  + 6.746347163590661d-13*tv08
     $  + 8.240609561517822d-14*tv09
     $  - 8.192116218351035d-16*tv10

      CrH4YZ(3,0,1,1,03) = 
     $  - 2.872873593430200d-02
     $  + 5.778955930298234d-04*tv01
     $  + 1.197256860435236d-05*tv02
     $  + 2.299826894249099d-06*tv03
     $  + 7.825017636596394d-08*tv04
     $  - 5.800879366434207d-09*tv05
     $  - 2.559895203179797d-10*tv06
     $  + 9.963341350074058d-12*tv07
     $  + 5.047030859637557d-13*tv08
     $  - 1.496806898109813d-14*tv09
     $  - 7.991946192338910d-16*tv10

      CrH4YZ(3,0,1,1,04) = 
     $  + 5.550426552063324d-03
     $  - 2.208812411062379d-04*tv01
     $  - 3.313487075913764d-06*tv02
     $  + 1.588558295779781d-07*tv03
     $  - 1.390560213777110d-08*tv04
     $  - 6.702985334343983d-10*tv05
     $  + 5.282835355241378d-11*tv06
     $  + 2.050593029488292d-12*tv07
     $  - 1.178955809225775d-13*tv08
     $  - 4.461912908214379d-15*tv09
     $  + 2.147170515323734d-16*tv10

      CrH4YZ(3,0,1,1,05) = 
     $  - 5.742846127158256d-04
     $  + 3.672522694195195d-06*tv01
     $  + 1.170813960214688d-06*tv02
     $  - 6.441913736054768d-09*tv03
     $  - 2.205352858105394d-09*tv04
     $  + 1.488213841068405d-10*tv05
     $  + 6.604179646442757d-12*tv06
     $  - 5.632558725254868d-13*tv07
     $  - 1.825622012632969d-14*tv08
     $  + 1.410473535104309d-15*tv09

      CrH4YZ(3,0,1,1,06) = 
     $  + 1.596773332425303d-05
     $  + 3.961837345494358d-06*tv01
     $  - 9.506844567877964d-10*tv02
     $  - 1.115581288668423d-08*tv03
     $  + 2.231077063563987d-10*tv04
     $  + 2.616537002781762d-11*tv05
     $  - 1.737249561940115d-12*tv06
     $  - 6.820399381728392d-14*tv07
     $  + 6.426510768104694d-15*tv08
     $  + 1.703874686180499d-16*tv09

      CrH4YZ(3,0,1,1,07) = 
     $  + 3.100522834282578d-06
     $  - 2.469442937798098d-07*tv01
     $  - 3.648227739691273d-08*tv02
     $  + 4.923984873746656d-10*tv03
     $  + 1.161352882607269d-10*tv04
     $  - 3.994710369514792d-12*tv05
     $  - 2.915372844140542d-13*tv06
     $  + 2.160478245008756d-14*tv07
     $  + 7.004995187439911d-16*tv08

      CrH4YZ(3,0,1,1,08) = 
     $  - 2.336788277157740d-07
     $  - 5.945281130270732d-08*tv01
     $  + 2.185784193009256d-09*tv02
     $  + 3.788599654347483d-10*tv03
     $  - 9.930056690331358d-12*tv04
     $  - 1.243471421192786d-12*tv05
     $  + 5.979835873542203d-14*tv06
     $  + 3.146967758423398d-15*tv07
     $  - 2.745151050058657d-16*tv08

      CrH4YZ(3,0,1,1,09) = 
     $  - 2.614020640528899d-08
     $  + 6.247444554305876d-09*tv01
     $  + 7.487917848555291d-10*tv02
     $  - 3.133453026602678d-11*tv03
     $  - 4.074945381189878d-12*tv04
     $  + 1.636153880601851d-13*tv05
     $  + 1.329302318754921d-14*tv06
     $  - 8.443973527557700d-16*tv07

      CrH4YZ(3,0,1,1,10) = 
     $  + 3.279036559943146d-09
     $  + 7.915775384630620d-10*tv01
     $  - 7.862965510211750d-11*tv02
     $  - 9.053395798368745d-12*tv03
     $  + 4.626204073745497d-13*tv04
     $  + 4.427715826653674d-14*tv05
     $  - 2.450175525918905d-15*tv06
     $  - 1.400261491162636d-16*tv07

      CrH4YZ(3,0,1,1,11) = 
     $  + 2.336243661937723d-10
     $  - 1.238106822820241d-10*tv01
     $  - 1.254239512931645d-11*tv02
     $  + 1.129607614727594d-12*tv03
     $  + 1.064039729203773d-13*tv04
     $  - 6.802620244450243d-15*tv05
     $  - 4.770912535061701d-16*tv06

      CrH4YZ(3,0,1,1,12) = 
     $  - 4.465150092725971d-11
     $  - 9.445227708856488d-12*tv01
     $  + 1.963571068651202d-12*tv02
     $  + 1.745054477382762d-13*tv03
     $  - 1.646776155757504d-14*tv04
     $  - 1.219101232705966d-15*tv05

      CrH4YZ(3,0,1,1,13) = 
     $  - 2.005018812987525d-12
     $  + 2.177834276145072d-12*tv01
     $  + 1.801165353737151d-13*tv02
     $  - 3.105036381859903d-14*tv03
     $  - 2.244377766034737d-15*tv04
     $  + 2.399521116462478d-16*tv05

      CrH4YZ(3,0,1,1,14) = 
     $  + 5.947708476067343d-13
     $  + 9.795468174836490d-14*tv01
     $  - 4.137754919524792d-14*tv02
     $  - 2.824613228942294d-15*tv03
     $  + 4.786299659682526d-16*tv04

      CrH4YZ(3,0,1,1,15) = 
     $  + 1.488823967937525d-14
     $  - 3.561611836912679d-14*tv01
     $  - 2.185533043188794d-15*tv02
     $  + 7.276746009660143d-16*tv03

      CrH4YZ(3,0,1,1,16) = 
     $  - 7.784579003970506d-15
     $  - 7.746359189984349d-16*tv01
     $  + 7.865840239042082d-16*tv02

      CrH4YZ(3,0,1,1,17) = 
     $  + 5.536208634795210d-16*tv01

      CrH4YZ(3,0,1,1,18) = 
     $  + 1.003476052856625d-16

      CrH4YZ(3,0,3,1,01) = 
     $  + 5.957764670434937d-01
     $  - 1.471821576586308d-01*tv01
     $  + 6.949175237582305d-03*tv02
     $  - 9.597375806006781d-05*tv03
     $  - 3.659808235898883d-06*tv04
     $  + 1.166246822930229d-07*tv05
     $  + 2.605688970901512d-09*tv06
     $  - 1.313819900829190d-10*tv07
     $  - 1.781842793951683d-12*tv08
     $  + 1.391650363759398d-13*tv09
     $  + 1.038560174363945d-15*tv10
     $  - 1.407889708473911d-16*tv11

      CrH4YZ(3,0,3,1,02) = 
     $  - 1.985043770250062d-01
     $  + 4.572188767744854d-02*tv01
     $  - 1.859332953724451d-03*tv02
     $  + 9.557466989071198d-06*tv03
     $  + 1.392332885028152d-06*tv04
     $  - 1.428320951336432d-08*tv05
     $  - 1.474415567739432d-09*tv06
     $  + 2.014000455270285d-11*tv07
     $  + 1.604169852323443d-12*tv08
     $  - 2.689925840245723d-14*tv09
     $  - 1.714566064363939d-15*tv10

      CrH4YZ(3,0,3,1,03) = 
     $  + 4.062843881407320d-02
     $  - 8.279031076890699d-03*tv01
     $  + 2.492648785137810d-04*tv02
     $  + 2.872639807197223d-06*tv03
     $  - 2.215316914726714d-07*tv04
     $  - 4.956161042696853d-09*tv05
     $  + 2.871352034098459d-10*tv06
     $  + 7.563541144158198d-12*tv07
     $  - 3.892924607419535d-13*tv08
     $  - 1.039187867701510d-14*tv09
     $  + 5.229594035948194d-16*tv10

      CrH4YZ(3,0,3,1,04) = 
     $  - 5.401118926673794d-03
     $  + 8.619440436532519d-04*tv01
     $  - 9.121867304007292d-06*tv02
     $  - 8.282430645080615d-07*tv03
     $  - 4.114714572887216d-09*tv04
     $  + 1.617773041151727d-09*tv05
     $  + 1.952564819384568d-11*tv06
     $  - 2.886278344370657d-12*tv07
     $  - 4.059539426297642d-14*tv08
     $  + 4.743935823567597d-15*tv09

      CrH4YZ(3,0,3,1,05) = 
     $  + 4.287094453061754d-04
     $  - 3.078609228014276d-05*tv01
     $  - 1.965321105015354d-06*tv02
     $  + 2.913805607543882d-08*tv03
     $  + 6.249573961534150d-09*tv04
     $  - 1.138505427952089d-11*tv05
     $  - 1.523562037268295d-11*tv06
     $  - 3.885841948462036d-14*tv07
     $  + 3.144541604503048d-14*tv08
     $  + 1.192871676074213d-16*tv09

      CrH4YZ(3,0,3,1,06) = 
     $  - 9.582191864111926d-06
     $  - 3.884953440660954d-06*tv01
     $  + 1.856468990906366d-07*tv02
     $  + 1.560120677794857d-08*tv03
     $  - 3.606372798388163d-10*tv04
     $  - 5.653182292569211d-11*tv05
     $  + 6.527286521109068d-13*tv06
     $  + 1.532407269949391d-13*tv07
     $  - 1.130783780118906d-15*tv08
     $  - 3.460365962998655d-16*tv09

      CrH4YZ(3,0,3,1,07) = 
     $  - 1.720497082357890d-06
     $  + 3.912040423949559d-07*tv01
     $  + 2.473965388902126d-08*tv02
     $  - 1.560154449225613d-09*tv03
     $  - 1.479175526629637d-10*tv04
     $  + 4.794242399219649d-12*tv05
     $  + 5.548157841327369d-13*tv06
     $  - 1.222932253103191d-14*tv07
     $  - 1.591497717214425d-15*tv08

      CrH4YZ(3,0,3,1,08) = 
     $  + 1.199877053719534d-07
     $  + 3.135709009327435d-08*tv01
     $  - 3.961585793929931d-09*tv02
     $  - 2.610237491112085d-10*tv03
     $  + 1.852583175249082d-11*tv04
     $  + 1.474150678690815d-12*tv05
     $  - 6.542667267561804d-14*tv06
     $  - 5.612120182009160d-15*tv07
     $  + 1.901544268408269d-16*tv08

      CrH4YZ(3,0,3,1,09) = 
     $  + 1.180574596928880d-08
     $  - 5.727941184267549d-09*tv01
     $  - 2.890761481088243d-10*tv02
     $  + 4.601193781117569d-11*tv03
     $  + 2.796454905516809d-12*tv04
     $  - 2.340626876131051d-13*tv05
     $  - 1.500868787837563d-14*tv06
     $  + 8.821555524469344d-16*tv07

      CrH4YZ(3,0,3,1,10) = 
     $  - 1.549701695427430d-09
     $  - 2.461441673989417d-10*tv01
     $  + 7.817339355177928d-11*tv02
     $  + 3.522069814269711d-12*tv03
     $  - 5.974356691010230d-13*tv04
     $  - 2.952824630147997d-14*tv05
     $  + 3.033474249936138d-15*tv06
     $  + 1.525566305466378d-16*tv07

      CrH4YZ(3,0,3,1,11) = 
     $  - 8.265664919292360d-11
     $  + 8.485767854689319d-11*tv01
     $  + 2.477451511243585d-12*tv02
     $  - 1.092792092964927d-12*tv03
     $  - 3.999740412838324d-14*tv04
     $  + 7.897097836152689d-15*tv05
     $  + 3.037880193834208d-16*tv06

      CrH4YZ(3,0,3,1,12) = 
     $  + 1.975392914656371d-11
     $  + 1.126235986042328d-12*tv01
     $  - 1.421476104539034d-12*tv02
     $  - 3.061080299804292d-14*tv03
     $  + 1.555801089659484d-14*tv04
     $  + 4.178277853586380d-16*tv05
     $  - 1.049273485521862d-16*tv06

      CrH4YZ(3,0,3,1,13) = 
     $  + 4.455115663994560d-13
     $  - 1.230763338399229d-12*tv01
     $  - 3.153318885421605d-16*tv02
     $  + 2.277526553305113d-14*tv03
     $  + 2.990918437126206d-16*tv04
     $  - 2.186935821612866d-16*tv05

      CrH4YZ(3,0,3,1,14) = 
     $  - 2.472904178119074d-13
     $  + 1.496537540724979d-14*tv01
     $  + 2.412636382867187d-14*tv02
     $  - 1.500754907820958d-16*tv03
     $  - 3.534954965811598d-16*tv04

      CrH4YZ(3,0,3,1,15) = 
     $  + 4.722957835714858d-16
     $  + 1.734210627520764d-14*tv01
     $  - 6.477668561032191d-16*tv02
     $  - 4.322079617750155d-16*tv03

      CrH4YZ(3,0,3,1,16) = 
     $  + 3.040840329616125d-15
     $  - 6.215873359584860d-16*tv01
     $  - 3.862752190722903d-16*tv02

      CrH4YZ(3,0,3,1,17) = 
     $  - 2.369497979986721d-16*tv01

      CrH4YZ(3,1,1,1,01) = 
     $  + 1.902177349986832d-02
     $  + 2.422677982016097d-02*tv01
     $  + 4.558824104962804d-03*tv02
     $  - 6.212458750364139d-04*tv03
     $  + 2.477599810629929d-05*tv04
     $  - 1.741456163574194d-07*tv05
     $  - 1.362442612305540d-08*tv06
     $  + 1.898187599425017d-10*tv07
     $  + 1.062152882968181d-11*tv08
     $  - 2.028852437139258d-13*tv09
     $  - 8.552622387889008d-15*tv10
     $  + 2.092886554077670d-16*tv11

      CrH4YZ(3,1,1,1,02) = 
     $  - 1.707433396571264d-02
     $  - 1.682428639866776d-02*tv01
     $  + 2.983755057068155d-04*tv02
     $  + 4.754839521600111d-05*tv03
     $  - 8.605778143791709d-07*tv04
     $  - 7.949475099434292d-08*tv05
     $  + 1.645801736400898d-09*tv06
     $  + 1.037685347604290d-10*tv07
     $  - 2.536253220080533d-12*tv08
     $  - 1.171894116057783d-13*tv09
     $  + 3.428504270356736d-15*tv10
     $  + 1.199706955553286d-16*tv11

      CrH4YZ(3,1,1,1,03) = 
     $  + 2.342935439135265d-02
     $  - 1.299879550493605d-05*tv01
     $  - 6.435483434379685d-06*tv02
     $  - 1.788297191398075d-06*tv03
     $  - 1.991358668387044d-07*tv04
     $  + 7.414841933704095d-09*tv05
     $  + 5.331962751787049d-10*tv06
     $  - 1.783936506764622d-11*tv07
     $  - 9.098348951891366d-13*tv08
     $  + 3.294031594752005d-14*tv09
     $  + 1.245180154532135d-15*tv10

      CrH4YZ(3,1,1,1,04) = 
     $  - 6.184368139552100d-03
     $  + 4.051934798378631d-04*tv01
     $  - 2.642446056393047d-06*tv02
     $  - 3.889343627302101d-07*tv03
     $  + 2.001076758537894d-08*tv04
     $  + 1.680225376498066d-09*tv05
     $  - 7.916263689288852d-11*tv06
     $  - 4.505374084745560d-12*tv07
     $  + 2.076213896156988d-13*tv08
     $  + 8.528729320874437d-15*tv09
     $  - 4.260408483641871d-16*tv10

      CrH4YZ(3,1,1,1,05) = 
     $  + 6.524431102835001d-04
     $  - 8.925766018362611d-06*tv01
     $  - 2.147667816266947d-06*tv02
     $  + 4.676319975643307d-08*tv03
     $  + 5.008937353820272d-09*tv04
     $  - 2.505589293482796d-10*tv05
     $  - 1.639942451022466d-11*tv06
     $  + 9.270677858808694d-13*tv07
     $  + 4.193860539615289d-14*tv08
     $  - 2.518722545103811d-15*tv09

      CrH4YZ(3,1,1,1,06) = 
     $  - 7.497249398649746d-06
     $  - 7.385052486746784d-06*tv01
     $  + 1.332270528737080d-07*tv02
     $  + 2.038741393541401d-08*tv03
     $  - 7.027853483692906d-10*tv04
     $  - 5.700731903313173d-11*tv05
     $  + 3.228477097043709d-12*tv06
     $  + 1.665631717743203d-13*tv07
     $  - 1.135530519130572d-14*tv08
     $  - 4.049234671982124d-16*tv09

      CrH4YZ(3,1,1,1,07) = 
     $  - 4.201576765457156d-06
     $  + 2.810462521517541d-07*tv01
     $  + 6.671444009828546d-08*tv02
     $  - 1.982545662430913d-09*tv03
     $  - 2.143681945634201d-10*tv04
     $  + 9.957249963344393d-12*tv05
     $  + 6.241109262354294d-13*tv06
     $  - 4.193797196551958d-14*tv07
     $  - 1.698508418778362d-15*tv08
     $  + 1.423190250714585d-16*tv09

      CrH4YZ(3,1,1,1,08) = 
     $  + 1.215297185491530d-07
     $  + 1.176980247663250d-07*tv01
     $  - 4.540370026285715d-09*tv02
     $  - 6.824136255576787d-10*tv03
     $  + 2.869425803138530d-11*tv04
     $  + 2.301856349776777d-12*tv05
     $  - 1.369681590623643d-13*tv06
     $  - 6.657501201748508d-15*tv07
     $  + 5.456504781647673d-16*tv08

      CrH4YZ(3,1,1,1,09) = 
     $  + 4.121016803138980d-08
     $  - 6.276980026623520d-09*tv01
     $  - 1.444802408323742d-09*tv02
     $  + 6.929218045102052d-11*tv03
     $  + 7.293461383709050d-12*tv04
     $  - 4.054296018889035d-13*tv05
     $  - 2.461530091473644d-14*tv06
     $  + 1.851166060734104d-15*tv07

      CrH4YZ(3,1,1,1,10) = 
     $  - 1.831392915411252d-09
     $  - 1.730982994073279d-09*tv01
     $  + 1.184231686570690d-10*tv02
     $  + 1.707534207547456d-11*tv03
     $  - 1.021975191270944d-12*tv04
     $  - 7.872923230007177d-14*tv05
     $  + 5.622904028718708d-15*tv06
     $  + 2.582426900103137d-16*tv07

      CrH4YZ(3,1,1,1,11) = 
     $  - 4.369163360856188d-10
     $  + 1.204160176651170d-10*tv01
     $  + 2.652566344882001d-11*tv02
     $  - 1.978125280855296d-12*tv03
     $  - 1.972447632573363d-13*tv04
     $  + 1.470804934707297d-14*tv05
     $  + 8.411398075996581d-16*tv06

      CrH4YZ(3,1,1,1,12) = 
     $  + 2.634933635857316d-11
     $  + 2.409543476018215d-11*tv01
     $  - 2.641204104315179d-12*tv02
     $  - 3.587403037339208d-13*tv03
     $  + 3.098277999901564d-14*tv04
     $  + 2.222665512616143d-15*tv05
     $  - 2.076554410955318d-16*tv06

      CrH4YZ(3,1,1,1,13) = 
     $  + 4.726999082390942d-12
     $  - 2.115040322692258d-12*tv01
     $  - 4.394441354303718d-13*tv02
     $  + 4.889374366153127d-14*tv03
     $  + 4.508594793282231d-15*tv04
     $  - 4.659403209268049d-16*tv05

      CrH4YZ(3,1,1,1,14) = 
     $  - 3.669648315581580d-13
     $  - 3.211857182975594d-13*tv01
     $  + 5.321770357963571d-14*tv02
     $  + 6.673565954782044d-15*tv03
     $  - 8.249280605878506d-16*tv04

      CrH4YZ(3,1,1,1,15) = 
     $  - 5.116025495457683d-14
     $  + 3.503049733896858d-14*tv01
     $  + 6.742501540528319d-15*tv02
     $  - 1.089902487963571d-15*tv03

      CrH4YZ(3,1,1,1,16) = 
     $  + 4.982904352755775d-15
     $  + 4.116109149528845d-15*tv01
     $  - 9.979943700502796d-16*tv02
     $  - 1.126429396702714d-16*tv03

      CrH4YZ(3,1,1,1,17) = 
     $  + 5.477485270187565d-16
     $  - 5.558129519828654d-16*tv01

      CrH4YZ(3,3,0,1,01) = 
     $  + 6.481173279013579d-01
     $  - 1.005242240517450d-01*tv01
     $  + 1.453123776624435d-03*tv02
     $  + 9.228369378477422d-05*tv03
     $  - 2.483350429486801d-06*tv04
     $  - 8.804967754800391d-08*tv05
     $  + 3.343678696377444d-09*tv06
     $  + 7.499019958715656d-11*tv07
     $  - 3.939564365146062d-12*tv08
     $  - 5.617291868241282d-14*tv09
     $  + 4.269767945557526d-15*tv10

      CrH4YZ(3,3,0,1,02) = 
     $  - 2.163734198641285d-01
     $  + 2.963673423981501d-02*tv01
     $  - 1.205418692348653d-04*tv02
     $  - 3.670629156363926d-05*tv03
     $  + 2.887495989526364d-07*tv04
     $  + 4.859112954892000d-08*tv05
     $  - 5.371222485969443d-10*tv06
     $  - 5.947987004984557d-11*tv07
     $  + 8.569232311025816d-13*tv08
     $  + 6.795749484893448d-14*tv09
     $  - 1.231110090272832d-15*tv10

      CrH4YZ(3,3,0,1,03) = 
     $  + 4.440621681102267d-02
     $  - 4.823472322786427d-03*tv01
     $  - 6.915217901598186d-05*tv02
     $  + 7.054523182210064d-06*tv03
     $  + 1.523019676586712d-07*tv04
     $  - 1.137480739235318d-08*tv05
     $  - 2.581440937464124d-10*tv06
     $  + 1.730367164491733d-11*tv07
     $  + 3.708015504473278d-13*tv08
     $  - 2.484186539505750d-14*tv09
     $  - 4.714665773896724d-16*tv10

      CrH4YZ(3,3,0,1,04) = 
     $  - 5.923011259139349d-03
     $  + 3.713388455455070d-04*tv01
     $  + 2.274591985242619d-05*tv02
     $  - 3.015664179278487d-07*tv03
     $  - 5.965664844027550d-08*tv04
     $  + 1.525461958063818d-10*tv05
     $  + 1.232449598580258d-10*tv06
     $  + 1.037952769901403d-13*tv07
     $  - 2.202108854444074d-13*tv08
     $  - 3.300140682557095d-16*tv09
     $  + 3.555566412641018d-16*tv10

      CrH4YZ(3,3,0,1,05) = 
     $  + 4.712774214795114d-04
     $  + 1.154101242710007d-05*tv01
     $  - 2.379807072500068d-06*tv02
     $  - 1.455822130691572d-07*tv03
     $  + 5.644682987089651d-09*tv04
     $  + 4.948724629587259d-10*tv05
     $  - 1.100904745713299d-11*tv06
     $  - 1.206805258355496d-12*tv07
     $  + 1.995985921297767d-14*tv08
     $  + 2.441240859399948d-15*tv09

      CrH4YZ(3,3,0,1,06) = 
     $  - 1.021721971626887d-05
     $  - 4.849693068537399d-06*tv01
     $  - 1.360392391333145d-07*tv02
     $  + 2.513079538180718d-08*tv03
     $  + 1.126468986955230d-09*tv04
     $  - 7.829148367668595d-11*tv05
     $  - 4.234926870436851d-12*tv06
     $  + 1.930034560254889d-13*tv07
     $  + 1.142507966918315d-14*tv08
     $  - 4.165983114741495d-16*tv09

      CrH4YZ(3,3,0,1,07) = 
     $  - 2.002676432716417d-06
     $  + 1.358515879576516d-07*tv01
     $  + 5.424032281138412d-08*tv02
     $  + 9.844155333756835d-10*tv03
     $  - 2.909561973764309d-10*tv04
     $  - 8.631694785369958d-12*tv05
     $  + 1.012995188675112d-12*tv06
     $  + 3.519956467852989d-14*tv07
     $  - 2.786147090662308d-15*tv08
     $  - 1.024399447421812d-16*tv09

      CrH4YZ(3,3,0,1,08) = 
     $  + 1.441892159674788d-07
     $  + 5.841539239592343d-08*tv01
     $  - 1.475851390613004d-09*tv02
     $  - 6.571384833017242d-10*tv03
     $  - 4.653267933807851d-12*tv04
     $  + 3.482564802545470d-12*tv05
     $  + 5.797131503684811d-14*tv06
     $  - 1.277057445835075d-14*tv07
     $  - 2.635011759587438d-16*tv08

      CrH4YZ(3,3,0,1,09) = 
     $  + 1.333971355439360d-08
     $  - 4.764611061896435d-09*tv01
     $  - 8.851877289882212d-10*tv02
     $  + 2.562535968587835d-11*tv03
     $  + 8.112872795869953d-12*tv04
     $  - 3.109198789564864d-14*tv05
     $  - 4.210877156108139d-14*tv06
     $  - 2.140179539225988d-16*tv07
     $  + 1.585407356336184d-16*tv08

      CrH4YZ(3,3,0,1,10) = 
     $  - 1.914175771839816d-09
     $  - 6.211630226165806d-10*tv01
     $  + 7.642318507598737d-11*tv02
     $  + 1.252946154915446d-11*tv03
     $  - 4.536732489595514d-13*tv04
     $  - 1.005087066979568d-13*tv05
     $  + 1.458968558471581d-15*tv06
     $  + 5.096926235787343d-16*tv07

      CrH4YZ(3,3,0,1,11) = 
     $  - 8.176033725007174d-11
     $  + 9.510263461246154d-11*tv01
     $  + 1.178916150497517d-11*tv02
     $  - 1.258424225440702d-12*tv03
     $  - 1.698024228130717d-13*tv04
     $  + 7.712357807561923d-15*tv05
     $  + 1.238766586512062d-15*tv06

      CrH4YZ(3,3,0,1,12) = 
     $  + 2.428226307160537d-11
     $  + 5.350176581127549d-12*tv01
     $  - 1.889491239190800d-12*tv02
     $  - 1.911913870443777d-13*tv03
     $  + 2.041743474610875d-14*tv04
     $  + 2.227678628870611d-15*tv05
     $  - 1.253998430290143d-16*tv06

      CrH4YZ(3,3,0,1,13) = 
     $  + 2.189585271036870d-13
     $  - 1.589372273030173d-12*tv01
     $  - 1.233429052983484d-13*tv02
     $  + 3.440613615758584d-14*tv03
     $  + 2.820731092234452d-15*tv04
     $  - 3.245302232739620d-16*tv05

      CrH4YZ(3,3,0,1,14) = 
     $  - 2.967177700029235d-13
     $  - 2.381459342831985d-14*tv01
     $  + 3.735819820083321d-14*tv02
     $  + 2.236200159507049d-15*tv03
     $  - 5.905265215386187d-16*tv04

      CrH4YZ(3,3,0,1,15) = 
     $  + 5.611526915583513d-15
     $  + 2.411208081636700d-14*tv01
     $  + 7.027371993863711d-16*tv02
     $  - 7.567738406994363d-16*tv03

      CrH4YZ(3,3,0,1,16) = 
     $  + 3.500496423298681d-15
     $  - 3.944406219360343d-16*tv01
     $  - 6.535955252763763d-16*tv02

      CrH4YZ(3,3,0,1,17) = 
     $  - 1.596961610132490d-16
     $  - 3.417354741485654d-16*tv01

      CrH4YZ(3,3,1,1,01) = 
     $  - 1.399355705913508d-01
     $  - 1.185740947456881d-01*tv01
     $  + 2.003448711341165d-02*tv02
     $  - 1.286971287918176d-03*tv03
     $  + 3.972522048461696d-05*tv04
     $  - 3.094813136888820d-07*tv05
     $  - 1.693008909295073d-08*tv06
     $  + 3.385550774166480d-10*tv07
     $  + 1.083063761536950d-11*tv08
     $  - 3.554678220414296d-13*tv09
     $  - 6.997539146195963d-15*tv10
     $  + 3.547665369675378d-16*tv11

      CrH4YZ(3,3,1,1,02) = 
     $  + 9.245539889013821d-02
     $  + 1.813749306687507d-02*tv01
     $  - 3.808363930545696d-03*tv02
     $  + 1.951353059711978d-04*tv03
     $  - 2.060604592595060d-06*tv04
     $  - 1.505875197285959d-07*tv05
     $  + 3.575719455605863d-09*tv06
     $  + 1.403508710605888d-10*tv07
     $  - 5.168087507524497d-12*tv08
     $  - 1.196407025456739d-13*tv09
     $  + 6.575985764533228d-15*tv10

      CrH4YZ(3,3,1,1,03) = 
     $  - 2.930869051435978d-02
     $  + 2.830264686306032d-04*tv01
     $  + 2.900337657532880d-04*tv02
     $  - 5.480054500329145d-06*tv03
     $  - 6.535781836189645d-07*tv04
     $  + 1.930245305770163d-08*tv05
     $  + 1.002620985204510d-09*tv06
     $  - 4.201456038452283d-11*tv07
     $  - 1.192017290939645d-12*tv08
     $  + 7.130872392483002d-14*tv09
     $  + 1.126010784413389d-15*tv10
     $  - 1.037487412342705d-16*tv11

      CrH4YZ(3,3,1,1,04) = 
     $  + 5.366352594331501d-03
     $  - 4.001561876564027d-04*tv01
     $  + 5.376283636816811d-08*tv02
     $  - 1.214019482243081d-06*tv03
     $  + 5.006850762671794d-08*tv04
     $  + 4.172822168304210d-09*tv05
     $  - 2.024382741099641d-10*tv06
     $  - 7.587319059693212d-12*tv07
     $  + 4.946054615464699d-13*tv08
     $  + 9.699565066629064d-15*tv09
     $  - 9.348145436017422d-16*tv10

      CrH4YZ(3,3,1,1,05) = 
     $  - 5.508876559349914d-04
     $  + 2.633435829930498d-05*tv01
     $  + 4.419683587609279d-07*tv02
     $  + 1.200570775522907d-08*tv03
     $  + 9.326193243506087d-09*tv04
     $  - 5.325682396559997d-10*tv05
     $  - 3.158000706975796d-11*tv06
     $  + 2.225030587059577d-12*tv07
     $  + 5.859728384749121d-14*tv08
     $  - 5.829359929231129d-15*tv09

      CrH4YZ(3,3,1,1,06) = 
     $  + 1.733908454832675d-05
     $  + 5.046506142102440d-06*tv01
     $  - 2.707944911598081d-07*tv02
     $  + 5.260001493551492d-09*tv03
     $  - 3.159647790795558d-10*tv04
     $  - 8.271060925567005d-11*tv05
     $  + 6.004469305405736d-12*tv06
     $  + 2.525478751024048d-13*tv07
     $  - 2.509882508508681d-14*tv08
     $  - 4.267096473674110d-16*tv09

      CrH4YZ(3,3,1,1,07) = 
     $  + 2.696031447712666d-06
     $  - 6.417882288434572d-07*tv01
     $  - 2.449556086832164d-08*tv02
     $  + 2.595221386776351d-09*tv03
     $  - 1.118391151687460d-10*tv04
     $  + 5.622940087801021d-12*tv05
     $  + 7.701580800588906d-13*tv06
     $  - 6.969413797855759d-14*tv07
     $  - 2.003225294059562d-15*tv08
     $  + 2.878081365570466d-16*tv09

      CrH4YZ(3,3,1,1,08) = 
     $  - 2.389412714929913d-07
     $  - 5.774810982276938d-08*tv01
     $  + 8.815110221836837d-09*tv02
     $  + 2.484930396414994d-11*tv03
     $  - 2.315314284661977d-11*tv04
     $  + 1.589709834517672d-12*tv05
     $  - 8.682717286272130d-14*tv06
     $  - 7.241461122745293d-15*tv07
     $  + 8.227336384642308d-16*tv08

      CrH4YZ(3,3,1,1,09) = 
     $  - 1.999149342799846d-08
     $  + 1.209203557271318d-08*tv01
     $  + 3.633611532255195d-10*tv02
     $  - 1.066965624684434d-10*tv03
     $  + 1.928667359853439d-12*tv04
     $  + 1.852101856423473d-13*tv05
     $  - 2.010110105381585d-14*tv06
     $  + 1.251644058844617d-15*tv07

      CrH4YZ(3,3,1,1,10) = 
     $  + 3.175357935408187d-09
     $  + 5.525768171335455d-10*tv01
     $  - 2.052070837284373d-10*tv02
     $  + 5.130706028923880d-13*tv03
     $  + 1.196940443870075d-12*tv04
     $  - 4.231615020512337d-14*tv05
     $  - 1.157547847919741d-15*tv06
     $  + 2.398757595342463d-16*tv07

      CrH4YZ(3,3,1,1,11) = 
     $  + 1.464019312767041d-10
     $  - 2.027124011957709d-10*tv01
     $  - 2.260454939971400d-12*tv02
     $  + 2.925549575421542d-12*tv03
     $  - 6.587598591819297d-14*tv04
     $  - 1.261925593412632d-14*tv05
     $  + 6.751149742963659d-16*tv06

      CrH4YZ(3,3,1,1,12) = 
     $  - 4.094944102261974d-11
     $  - 3.408618385381774d-12*tv01
     $  + 4.075023163295164d-12*tv02
     $  - 7.242272232979066d-14*tv03
     $  - 3.764731436065588d-14*tv04
     $  + 1.473546749064999d-15*tv05
     $  + 1.241912421368473d-16*tv06

      CrH4YZ(3,3,1,1,13) = 
     $  - 8.293688159265651d-13
     $  + 3.161674512455295d-12*tv01
     $  - 5.142222329594957d-14*tv02
     $  - 6.634056684432088d-14*tv03
     $  + 2.442741306034518d-15*tv04
     $  + 4.494343692293176d-16*tv05

      CrH4YZ(3,3,1,1,14) = 
     $  + 5.150500449994448d-13
     $  - 1.888604919453127d-14*tv01
     $  - 7.325649382422412d-14*tv02
     $  + 2.964938200031656d-15*tv03
     $  + 9.511231785781556d-16*tv04

      CrH4YZ(3,3,1,1,15) = 
     $  - 2.513790090734516d-16
     $  - 4.682024313558264d-14*tv01
     $  + 2.467033578431861d-15*tv02
     $  + 1.334144110847439d-15*tv03

      CrH4YZ(3,3,1,1,16) = 
     $  - 6.334418619592956d-15
     $  + 1.227865910752803d-15*tv01
     $  + 1.225385038249909d-15*tv02

      CrH4YZ(3,3,1,1,17) = 
     $  + 1.195840374200394d-16
     $  + 6.651338912899392d-16*tv01

      CrH4YZ(3,3,3,1,01) = 
     $  + 5.519791121767504d-01
     $  - 1.831402382280320d-01*tv01
     $  + 1.421839306380919d-02*tv02
     $  - 6.453306760967352d-04*tv03
     $  + 1.679103090961564d-05*tv04
     $  - 1.404531544719315d-07*tv05
     $  - 5.490360712835891d-09*tv06
     $  + 1.410816434604428d-10*tv07
     $  + 2.554615228135881d-12*tv08
     $  - 1.362378521739512d-13*tv09
     $  - 8.446055131372456d-16*tv10
     $  + 1.244723579511035d-16*tv11

      CrH4YZ(3,3,3,1,02) = 
     $  - 1.850769315357478d-01
     $  + 5.700783465904210d-02*tv01
     $  - 3.873386615431668d-03*tv02
     $  + 1.342624038053145d-04*tv03
     $  - 1.402558851256842d-06*tv04
     $  - 6.584347161040498d-08*tv05
     $  + 1.972693462611799d-09*tv06
     $  + 4.085694372442680d-11*tv07
     $  - 2.449545772700448d-12*tv08
     $  - 1.689989170613935d-14*tv09
     $  + 2.735583025738762d-15*tv10

      CrH4YZ(3,3,3,1,03) = 
     $  + 3.827352862596428d-02
     $  - 1.033649553290217d-02*tv01
     $  + 5.366110398724180d-04*tv02
     $  - 7.464558417727631d-06*tv03
     $  - 4.385753488401495d-07*tv04
     $  + 1.575545564222078d-08*tv05
     $  + 3.811285625733050d-10*tv06
     $  - 2.609205865728803d-11*tv07
     $  - 2.029121778090454d-13*tv08
     $  + 3.642959047914751d-14*tv09
     $  - 1.134745912033816d-16*tv10

      CrH4YZ(3,3,3,1,04) = 
     $  - 5.179405306597092d-03
     $  + 1.071470063496425d-03*tv01
     $  - 2.231508029200851d-05*tv02
     $  - 1.752016248410558d-06*tv03
     $  + 7.859496117426763d-08*tv04
     $  + 2.285146949933164d-09*tv05
     $  - 1.823170369235943d-10*tv06
     $  - 1.624425568761137d-12*tv07
     $  + 3.273736386126524d-13*tv08
     $  - 1.128167700766564d-15*tv09
     $  - 4.926861380214675d-16*tv10

      CrH4YZ(3,3,3,1,05) = 
     $  + 4.264901716419532d-04
     $  - 3.545349762197717d-05*tv01
     $  - 4.195707513233888d-06*tv02
     $  + 2.506308452364441d-07*tv03
     $  + 9.131482951443820d-09*tv04
     $  - 8.730305212197467d-10*tv05
     $  - 9.104848294935667d-12*tv06
     $  + 2.091256009169244d-12*tv07
     $  - 8.065112583962673d-15*tv08
     $  - 3.935279447271936d-15*tv09

      CrH4YZ(3,3,3,1,06) = 
     $  - 1.156865187847711d-05
     $  - 5.569962525131807d-06*tv01
     $  + 4.983613243639096d-07*tv02
     $  + 2.431415918007124d-08*tv03
     $  - 2.900366108945630d-09*tv04
     $  - 3.646202374724524d-11*tv05
     $  + 9.735628479427729d-12*tv06
     $  - 4.263054908287026d-14*tv07
     $  - 2.356621851643178d-14*tv08
     $  + 3.833910498850183d-16*tv09

      CrH4YZ(3,3,3,1,07) = 
     $  - 1.570629329339202d-06
     $  + 5.629596628496537d-07*tv01
     $  + 4.157706996834307d-08*tv02
     $  - 6.596136423385096d-09*tv03
     $  - 1.043457689209926d-10*tv04
     $  + 3.327182563438601d-11*tv05
     $  - 1.685582138439406d-13*tv06
     $  - 1.074720088444064d-13*tv07
     $  + 1.963982813661412d-15*tv08
     $  + 2.592759820676525d-16*tv09

      CrH4YZ(3,3,3,1,08) = 
     $  + 1.358342156829018d-07
     $  + 4.136788057322243d-08*tv01
     $  - 9.811400059026228d-09*tv02
     $  - 2.091893951257781d-10*tv03
     $  + 8.280457604530506d-11*tv04
     $  - 4.978572896585430d-13*tv05
     $  - 3.749880454541226d-13*tv06
     $  + 7.817348464108299d-15*tv07
     $  + 1.163983370568451d-15*tv08

      CrH4YZ(3,3,3,1,09) = 
     $  + 9.052861661552717d-09
     $  - 8.575032203468975d-09*tv01
     $  - 2.800013761045511d-10*tv02
     $  + 1.462122934432030d-10*tv03
     $  - 1.082182603542762d-12*tv04
     $  - 9.958417484783566d-13*tv05
     $  + 2.416692900892784d-14*tv06
     $  + 4.126372716388242d-15*tv07
     $  - 1.537106673423629d-16*tv08

      CrH4YZ(3,3,3,1,10) = 
     $  - 1.628269151417022d-09
     $  - 2.256750818044403d-10*tv01
     $  + 1.734745785539372d-10*tv02
     $  - 1.673980920116600d-12*tv03
     $  - 1.980173577095490d-12*tv04
     $  + 5.751124633029832d-14*tv05
     $  + 1.150992049747933d-14*tv06
     $  - 4.893828386793940d-16*tv07

      CrH4YZ(3,3,3,1,11) = 
     $  - 4.189324939849454d-11
     $  + 1.233080626047879d-10*tv01
     $  - 1.722832686287676d-12*tv02
     $  - 2.855267252628257d-12*tv03
     $  + 1.033283174672527d-13*tv04
     $  + 2.498522313871277d-14*tv05
     $  - 1.237584958246019d-15*tv06
     $  - 1.273306787868643d-16*tv07

      CrH4YZ(3,3,3,1,12) = 
     $  + 1.914438327323396d-11
     $  - 1.013237682039702d-12*tv01
     $  - 2.813920988501955d-12*tv02
     $  + 1.353174420815268d-13*tv03
     $  + 4.134626404251817d-14*tv04
     $  - 2.453647955242837d-15*tv05
     $  - 2.957745877413884d-16*tv06

      CrH4YZ(3,3,3,1,13) = 
     $  - 9.529292966191095d-14
     $  - 1.681298759173161d-12*tv01
     $  + 1.211795813046582d-13*tv02
     $  + 5.034492606034010d-14*tv03
     $  - 3.728826768781302d-15*tv04
     $  - 5.427835287081002d-16*tv05

      CrH4YZ(3,3,3,1,14) = 
     $  - 2.189935971095572d-13
     $  + 6.506207464767469d-14*tv01
     $  + 4.238387983380322d-14*tv02
     $  - 4.183400377570594d-15*tv03
     $  - 7.689139250918374d-16*tv04

      CrH4YZ(3,3,3,1,15) = 
     $  + 7.052948638552620d-15
     $  + 2.179487756569944d-14*tv01
     $  - 3.243989295711537d-15*tv02
     $  - 8.098583458423184d-16*tv03
     $  + 1.027310061311361d-16*tv04

      CrH4YZ(3,3,3,1,16) = 
     $  + 2.427050313591695d-15
     $  - 1.521375237304948d-15*tv01
     $  - 5.946187644906413d-16*tv02
     $  + 1.006194101174791d-16*tv03

      CrH4YZ(3,3,3,1,17) = 
     $  - 1.446953182050584d-16
     $  - 2.679662376734091d-16*tv01
      endif

      return
      end

      subroutine fillcoeff2dhpl320(iflag,n,CsH2YZ,CsH3YZ,CsH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsH2YZ,CsH3YZ,CsH4YZ
      real(dp):: 
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
      dimension CsH2YZ(0:3,0:3,22),CsH3YZ(0:3,0:3,0:3,22),
     $          CsH4YZ(0:3,0:3,0:3,0:3,22)

      common /tpu/  
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
!$omp threadprivate(/tpu/)
      
* 2001-08-24:11:13:02.chpl 
* <- HPLexpand.320t 
* produced by form-to-fortr for gehrt@pcth62 
      if (n.eq.2) then
      CsH2YZ(3,2,01) = 
     $  + 6.2246945526043164d-01
     $  - 1.2318938491474078d-01*tu01
     $  + 4.3234854072345596d-03*tu02
     $  - 2.0133335065426132d-05*tu03
     $  - 2.4360513938079663d-06*tu04
     $  + 2.4828889022590042d-08*tu05
     $  + 1.9359756483168612d-09*tu06
     $  - 2.8454404383462713d-11*tu07
     $  - 1.5902232751400798d-12*tu08
     $  + 3.0733514480493260d-14*tu09
     $  + 1.2949674406911631d-15*tu10

      CsH2YZ(3,2,02) = 
     $  - 1.2324966097439652d-01
     $  + 1.7274476424077856d-02*tu01
     $  - 1.2055211931148737d-04*tu02
     $  - 1.9465204860381764d-05*tu03
     $  + 2.4789108106941498d-07*tu04
     $  + 2.3206290081966222d-08*tu05
     $  - 3.9780915648543090d-10*tu06
     $  - 2.5417697836112391d-11*tu07
     $  + 5.5250488304708871d-13*tu08
     $  + 2.5874566128886696d-14*tu09
     $  - 6.9837760178997239d-16*tu10

      CsH2YZ(3,2,03) = 
     $  + 1.1477464442170556d-02
     $  - 3.2015340898742762d-04*tu01
     $  - 7.7706347762695266d-05*tu02
     $  + 1.3189091765386708d-06*tu03
     $  + 1.5447167883173893d-07*tu04
     $  - 3.1765891649258705d-09*tu05
     $  - 2.3692171474347401d-10*tu06
     $  + 5.8840869284741370d-12*tu07
     $  + 3.1013170871923085d-13*tu08
     $  - 9.2985007301636856d-15*tu09
     $  - 3.6308971964482113d-16*tu10

      CsH2YZ(3,2,04) = 
     $  - 1.5810626164928464d-04
     $  - 1.5479622786292460d-04*tu01
     $  + 3.9408856888583346d-06*tu02
     $  + 6.1646766246592846d-07*tu03
     $  - 1.5841840757677770d-08*tu04
     $  - 1.4190528610272905d-09*tu05
     $  + 4.1105066951582438d-11*tu06
     $  + 2.4774274335535008d-12*tu07
     $  - 8.3541547736520790d-14*tu08
     $  - 3.6262362003459147d-15*tu09
     $  + 1.4495883495238103d-16*tu10

      CsH2YZ(3,2,05) = 
     $  - 6.1181561132169966d-05
     $  + 6.2549199825616904d-06*tu01
     $  + 1.4738600259997613d-06*tu02
     $  - 5.0497119611627726d-08*tu03
     $  - 5.6623639184464262d-09*tu04
     $  + 1.9677081296083044d-10*tu05
     $  + 1.3847525678830008d-11*tu06
     $  - 5.3350839532061400d-13*tu07
     $  - 2.6067941916690489d-14*tu08
     $  + 1.1575141395500465d-15*tu09

      CsH2YZ(3,2,06) = 
     $  + 2.0348029178756017d-06
     $  + 1.9501023155634749d-06*tu01
     $  - 1.0034081928992352d-07*tu02
     $  - 1.5044385769540177d-08*tu03
     $  + 6.5341993333192577d-10*tu04
     $  + 5.5251346316958972d-11*tu05
     $  - 2.4827765375090312d-12*tu06
     $  - 1.3875639836106126d-13*tu07
     $  + 6.9293073205007465d-15*tu08
     $  + 2.7262519462134827d-16*tu09

      CsH2YZ(3,2,07) = 
     $  + 5.4435555604011522d-07
     $  - 1.1319017155923925d-07*tu01
     $  - 2.5633068241757439d-08*tu02
     $  + 1.4850504863838198d-09*tu03
     $  + 1.5730736316405694d-10*tu04
     $  - 8.4807898077812468d-12*tu05
     $  - 5.5362631651757420d-13*tu06
     $  + 3.1586889726066134d-14*tu07
     $  + 1.3992753493887759d-15*tu08

      CsH2YZ(3,2,08) = 
     $  - 2.7194300937017213d-08
     $  - 2.5320108811237333d-08*tu01
     $  + 2.2064839055852033d-09*tu02
     $  + 3.1295943052010529d-10*tu03
     $  - 2.1091823990526372d-11*tu04
     $  - 1.6552958080085909d-12*tu05
     $  + 1.1015052892674431d-13*tu06
     $  + 5.5831415441316239d-15*tu07
     $  - 4.0358511448715394d-16*tu08

      CsH2YZ(3,2,09) = 
     $  - 5.4198817871256241d-09
     $  + 1.9241147561419420d-09*tu01
     $  + 4.1361811963233873d-10*tu02
     $  - 3.7204271044335544d-11*tu03
     $  - 3.6611210604583089d-12*tu04
     $  + 2.9230493947207065d-13*tu05
     $  + 1.7314068980609188d-14*tu06
     $  - 1.4298039550248082d-15*tu07

      CsH2YZ(3,2,10) = 
     $  + 3.6279070108051814d-10
     $  + 3.2507807805914580d-10*tu01
     $  - 4.4064500295740536d-11*tu02
     $  - 5.8164176467251834d-12*tu03
     $  + 5.8062495746211679d-13*tu04
     $  + 4.1376050008110916d-14*tu05
     $  - 3.9849214820245158d-15*tu06
     $  - 1.7771554535113504d-16*tu07

      CsH2YZ(3,2,11) = 
     $  + 5.5969902944786730d-11
     $  - 3.1211004622531643d-11*tu01
     $  - 6.2704043137892115d-12*tu02
     $  + 8.3590468339725795d-13*tu03
     $  + 7.4778573471095725d-14*tu04
     $  - 8.6407093481181096d-15*tu05
     $  - 4.5060842166518631d-16*tu06

      CsH2YZ(3,2,12) = 
     $  - 4.7910201020555745d-12
     $  - 4.0814613219457758d-12*tu01
     $  + 8.2162800339939866d-13*tu02
     $  + 9.8808220580365202d-14*tu03
     $  - 1.4276679997859290d-14*tu04
     $  - 8.9654404776243123d-16*tu05
     $  + 1.2450224900422564d-16*tu06

      CsH2YZ(3,2,13) = 
     $  - 5.8299797181805106d-13
     $  + 4.8827403290169487d-13*tu01
     $  + 8.9838309273635923d-14*tu02
     $  - 1.7343199953655276d-14*tu03
     $  - 1.3692789474843159d-15*tu04
     $  + 2.2809851188127668d-16*tu05

      CsH2YZ(3,2,14) = 
     $  + 6.2481821105061938d-14
     $  + 4.9788360720209102d-14*tu01
     $  - 1.4543224333217515d-14*tu02
     $  - 1.5478160075828537d-15*tu03
     $  + 3.2237562705843551d-16*tu04

      CsH2YZ(3,2,15) = 
     $  + 6.0306095403023398d-15
     $  - 7.4180347208462376d-15*tu01
     $  - 1.2156771117057621d-15*tu02
     $  + 3.3835158976018000d-16*tu03

      CsH2YZ(3,2,16) = 
     $  - 8.0377533545305323d-16
     $  - 5.8554597780148261d-16*tu01
     $  + 2.4695800930545293d-16*tu02

      CsH2YZ(3,2,17) = 
     $  + 1.0993445774998011d-16*tu01
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      CsH3YZ(0,3,2,01) = 
     $  + 6.2246945526043164d-01
     $  - 1.2318938491474078d-01*tu01
     $  + 4.3234854072345596d-03*tu02
     $  - 2.0133335065426132d-05*tu03
     $  - 2.4360513938079663d-06*tu04
     $  + 2.4828889022590042d-08*tu05
     $  + 1.9359756483168612d-09*tu06
     $  - 2.8454404383462713d-11*tu07
     $  - 1.5902232751400798d-12*tu08
     $  + 3.0733514480493260d-14*tu09
     $  + 1.2949674406911631d-15*tu10

      CsH3YZ(0,3,2,02) = 
     $  - 1.7833785334852919d-01
     $  + 3.1735247883552824d-02*tu01
     $  - 8.7092957351222362d-04*tu02
     $  - 5.9576021054234824d-06*tu03
     $  + 5.8070517687370118d-07*tu04
     $  + 6.9477283492474785d-09*tu05
     $  - 5.6190001230212694d-10*tu06
     $  - 7.3736480961569369d-12*tu07
     $  + 5.7441930561230933d-13*tu08
     $  + 7.1747490993508618d-15*tu09
     $  - 5.9199519602457929d-16*tu10

      CsH3YZ(0,3,2,03) = 
     $  + 2.8958114340967328d-02
     $  - 4.1908614952983660d-03*tu01
     $  + 5.6721358481077494d-05*tu02
     $  + 2.5582036393299941d-06*tu03
     $  - 1.7559128551346701d-08*tu04
     $  - 3.5716979242430986d-09*tu05
     $  + 1.0018591511384838d-12*tu06
     $  + 4.6939744705138230d-12*tu07
     $  + 9.4668871847937798d-15*tu08
     $  - 5.8536098457410250d-15*tu09

      CsH3YZ(0,3,2,04) = 
     $  - 2.5598708214095201d-03
     $  + 1.9375059572150257d-04*tu01
     $  + 6.8574713767857722d-06*tu02
     $  - 1.9763868914161707d-07*tu03
     $  - 1.5537206473612761d-08*tu04
     $  + 2.1499073085301940d-10*tu05
     $  + 2.7825851442532654d-11*tu06
     $  - 2.3013993767301721d-13*tu07
     $  - 4.3485568028350019d-14*tu08
     $  + 2.6839296518925782d-16*tu09

      CsH3YZ(0,3,2,05) = 
     $  + 5.2513657076362452d-05
     $  + 2.0686454910496688d-05*tu01
     $  - 1.1092852663599159d-06*tu02
     $  - 4.2200193674985422d-08*tu03
     $  + 1.7179433417180033d-09*tu04
     $  + 1.1382037717378176d-10*tu05
     $  - 2.7046700710380992d-12*tu06
     $  - 2.3484454624092505d-13*tu07
     $  + 4.0992781836853592d-15*tu08
     $  + 4.1359591079453125d-16*tu09

      CsH3YZ(0,3,2,06) = 
     $  + 1.1954850999669703d-05
     $  - 2.5404814177119315d-06*tu01
     $  - 6.9222051136298937d-08*tu02
     $  + 6.8905001445950952d-09*tu03
     $  + 3.2088060116626789d-10*tu04
     $  - 1.5875648016697208d-11*tu05
     $  - 9.2899669612395349d-13*tu06
     $  + 3.1434757929395778d-14*tu07
     $  + 2.0909959152795775d-15*tu08

      CsH3YZ(0,3,2,07) = 
     $  - 6.3811776560508958d-07
     $  - 1.6223170435289280d-07*tu01
     $  + 2.0096605825504524d-08*tu02
     $  + 5.8022121464937070d-10*tu03
     $  - 6.2195563311752765d-11*tu04
     $  - 2.6334742199889520d-12*tu05
     $  + 1.6314684108375513d-13*tu06
     $  + 7.9798690512122837d-15*tu07
     $  - 3.6668533330637783d-16*tu08

      CsH3YZ(0,3,2,08) = 
     $  - 9.8981986807831222d-08
     $  + 3.4779054672339052d-08*tu01
     $  + 5.8425757418923166d-10*tu02
     $  - 1.6682647870417346d-10*tu03
     $  - 5.1290385320216609d-12*tu04
     $  + 6.0938697839490690d-13*tu05
     $  + 2.2621897301485672d-14*tu06
     $  - 1.7505572751778981d-15*tu07

      CsH3YZ(0,3,2,09) = 
     $  + 8.1347994934475814d-09
     $  + 1.2973506596024772d-09*tu01
     $  - 3.4202596412672674d-10*tu02
     $  - 5.9258193209316080d-12*tu03
     $  + 1.6912453106238675d-12*tu04
     $  + 4.5364600648720783d-14*tu05
     $  - 6.3981770619670826d-15*tu06
     $  - 1.9682318019717156d-16*tu07

      CsH3YZ(0,3,2,10) = 
     $  + 9.4965694021608521d-10
     $  - 4.7811092483822807d-10*tu01
     $  - 1.8475529876019556d-12*tu02
     $  + 3.4655744015879891d-12*tu03
     $  + 5.6396421721615408d-14*tu04
     $  - 1.8038322814544104d-14*tu05
     $  - 3.9484773764730426d-16*tu06

      CsH3YZ(0,3,2,11) = 
     $  - 1.0468836991861561d-10
     $  - 8.6087549765194981d-12*tu01
     $  + 5.5541257997685493d-12*tu02
     $  + 1.7315364575817912d-14*tu03
     $  - 3.9334081900322513d-14*tu04
     $  - 4.7980200741469096d-16*tu05
     $  + 2.0018290149437035d-16*tu06

      CsH3YZ(0,3,2,12) = 
     $  - 9.7216130551900854d-12
     $  + 6.5487897432305668d-12*tu01
     $  - 7.6716184102810819d-14*tu02
     $  - 6.5606843195663924d-14*tu03
     $  + 4.5671362552062717d-16*tu05

      CsH3YZ(0,3,2,13) = 
     $  + 1.3508318232900141d-12
     $  + 1.6306027694135379d-14*tu01
     $  - 8.6964234750881858d-14*tu02
     $  + 1.2264286378835795d-15*tu03
     $  + 8.2426745621913302d-16*tu04

      CsH3YZ(0,3,2,14) = 
     $  + 1.0326838015455377d-13
     $  - 8.9217477415924683d-14*tu01
     $  + 2.6848215673964624d-15*tu02
     $  + 1.1629436304052511d-15*tu03

      CsH3YZ(0,3,2,15) = 
     $  - 1.7459427669478130d-14
     $  + 9.6970595009747664d-16*tu01
     $  + 1.3216659026073986d-15*tu02

      CsH3YZ(0,3,2,16) = 
     $  - 1.1246988849954685d-15
     $  + 1.2084558466731886d-15*tu01

      CsH3YZ(0,3,2,17) = 
     $  + 2.2607113731358604d-16

      CsH3YZ(3,0,2,01) = 
     $  + 6.7820922479746498d-01
     $  - 7.2161137820533973d-02*tu01
     $  - 2.9609515081006315d-04*tu02
     $  + 7.4754123486713770d-05*tu03
     $  + 4.0333501318342053d-07*tu04
     $  - 8.3808506326028907d-08*tu05
     $  - 3.9905982933620598d-10*tu06
     $  + 8.8548158315861998d-11*tu07
     $  + 2.9721341909005945d-13*tu08
     $  - 8.9300569770748624d-14*tu09
     $  - 1.3237074874293598d-16*tu10

      CsH3YZ(3,0,2,02) = 
     $  - 1.9518695484796762d-01
     $  + 1.6093317716977885d-02*tu01
     $  + 3.2713577461425540d-04*tu02
     $  - 1.6243308720100778d-05*tu03
     $  - 5.8895591347034110d-07*tu04
     $  + 1.8422324895669410d-08*tu05
     $  + 8.4025956411399423d-10*tu06
     $  - 2.0664932051202552d-11*tu07
     $  - 1.0529909100644578d-12*tu08
     $  + 2.3225661890899341d-14*tu09
     $  + 1.2160931030371684d-15*tu10

      CsH3YZ(3,0,2,03) = 
     $  + 3.1841393014882363d-02
     $  - 1.4501585792098035d-03*tu01
     $  - 8.8634770942886420d-05*tu02
     $  - 4.1782148554929120d-08*tu03
     $  + 1.6887235921064110d-07*tu04
     $  + 2.2217799605018378d-09*tu05
     $  - 2.6132908879008234d-10*tu06
     $  - 5.0964539826196114d-12*tu07
     $  + 3.6432592515678342d-13*tu08
     $  + 8.3098018174049725d-15*tu09
     $  - 4.7900496434586706d-16*tu10

      CsH3YZ(3,0,2,04) = 
     $  - 2.8167387812104816d-03
     $  - 6.3223130978933551d-05*tu01
     $  + 7.5425120872062834d-06*tu02
     $  + 6.0369394468426628d-07*tu03
     $  - 6.7523851189533384d-09*tu04
     $  - 1.5589779529512490d-09*tu05
     $  - 2.7953002796469514d-12*tu06
     $  + 2.9854248592230387d-12*tu07
     $  + 2.2056290599143917d-14*tu08
     $  - 4.9176475065020168d-15*tu09

      CsH3YZ(3,0,2,05) = 
     $  + 5.2695571927325917d-05
     $  + 2.2968766858890496d-05*tu01
     $  + 9.6917206059661724d-07*tu02
     $  - 7.1206683496790548d-08*tu03
     $  - 5.3605527057385987d-09*tu04
     $  + 1.2000959942312511d-10*tu05
     $  + 1.5396182656081068d-11*tu06
     $  - 1.3704615747377242d-13*tu07
     $  - 3.3412903816246003d-14*tu08
     $  + 1.0121850503961920d-16*tu09

      CsH3YZ(3,0,2,06) = 
     $  + 1.4438722563048618d-05
     $  - 1.9648062257467462d-07*tu01
     $  - 2.2603605349421532d-07*tu02
     $  - 9.6630148902682882d-09*tu03
     $  + 7.7770084934622949d-10*tu04
     $  + 5.0610080462529873d-11*tu05
     $  - 1.7324059692493593d-12*tu06
     $  - 1.5596987227732357d-13*tu07
     $  + 3.0839847721536812d-15*tu08
     $  + 3.6946627955605668d-16*tu09

      CsH3YZ(3,0,2,07) = 
     $  - 7.8431622322950303d-07
     $  - 3.3493589175086522d-07*tu01
     $  - 4.6541377238033310d-09*tu02
     $  + 2.5019421125782858d-09*tu03
     $  + 9.8744697349099331d-11*tu04
     $  - 9.1012028804011358d-12*tu05
     $  - 4.9653821168949411d-13*tu06
     $  + 2.3606257120726018d-14*tu07
     $  + 1.6022451391389356d-15*tu08

      CsH3YZ(3,0,2,08) = 
     $  - 1.2199488975805021d-07
     $  + 1.5514571121445966d-08*tu01
     $  + 4.5940185357779152d-09*tu02
     $  + 7.2103019743613086d-11*tu03
     $  - 2.9226455064947741d-11*tu04
     $  - 9.9524833240304886d-13*tu05
     $  + 1.1023768494617780d-13*tu06
     $  + 4.9257786452430737d-15*tu07
     $  - 3.1375713618437927d-16*tu08

      CsH3YZ(3,0,2,09) = 
     $  + 1.1112788494997822d-08
     $  + 4.5421885949803102d-09*tu01
     $  - 1.2714558501975378d-10*tu02
     $  - 6.0723502767977100d-11*tu03
     $  - 8.4763973866499714d-13*tu04
     $  + 3.5293029362979004d-13*tu05
     $  + 9.8299371766909044d-15*tu06
     $  - 1.3619255067068794d-15*tu07

      CsH3YZ(3,0,2,10) = 
     $  + 1.1283143788160680d-09
     $  - 3.7397932691145185d-10*tu01
     $  - 7.9186944200911314d-11*tu02
     $  + 1.3579124058500919d-12*tu03
     $  + 7.9224230591870277d-13*tu04
     $  + 8.1434908498831090d-15*tu05
     $  - 4.3474825288128870d-15*tu06

      CsH3YZ(3,0,2,11) = 
     $  - 1.5230353162618186d-10
     $  - 5.7710088869238849d-11*tu01
     $  + 5.2439373838614524d-12*tu02
     $  + 1.2240909041869825d-12*tu03
     $  - 1.8048224298375297d-14*tu04
     $  - 1.0292699697215340d-14*tu05

      CsH3YZ(3,0,2,12) = 
     $  - 1.0441347903893927d-11
     $  + 7.1224842093204012d-12*tu01
     $  + 1.2235854741264014d-12*tu02
     $  - 7.6224579775296975d-14*tu03
     $  - 1.7820794647122598d-14*tu04
     $  + 2.8516181267215340d-16*tu05
     $  + 1.3337485218024917d-16*tu06

      CsH3YZ(3,0,2,13) = 
     $  + 2.0383805287380989d-12
     $  + 6.8404360767145257d-13*tu01
     $  - 1.2978477407575177d-13*tu02
     $  - 2.1722109262325455d-14*tu03
     $  + 1.1365394404361946d-15*tu04
     $  + 2.5062992923410666d-16*tu05

      CsH3YZ(3,0,2,14) = 
     $  + 9.1441212074737340d-14
     $  - 1.2163306015586138d-13*tu01
     $  - 1.7217471668951095d-14*tu02
     $  + 2.2235163170461543d-15*tu03
     $  + 3.5025458121824808d-16*tu04

      CsH3YZ(3,0,2,15) = 
     $  - 2.6777196532891627d-14
     $  - 7.4045194035463979d-15*tu01
     $  + 2.6724169531834751d-15*tu02
     $  + 3.4622731697627040d-16*tu03

      CsH3YZ(3,0,2,16) = 
     $  - 6.9482329070911144d-16
     $  + 1.9449034273720982d-15*tu01
     $  + 2.1905678192908470d-16*tu02

      CsH3YZ(3,0,2,17) = 
     $  + 3.4614284493020126d-16

      CsH3YZ(3,2,2,01) = 
     $  - 1.0516401802557398d-01
     $  - 9.4020339946745276d-02*tu01
     $  + 1.0740238822062851d-02*tu02
     $  - 4.0132337678530310d-04*tu03
     $  + 2.3419320903768033d-06*tu04
     $  + 2.2354290376071681d-07*tu05
     $  - 2.6880361795249822d-09*tu06
     $  - 1.7601343278042493d-10*tu07
     $  + 2.9628860058048243d-12*tu08
     $  + 1.4319949390628153d-13*tu09
     $  - 3.1210941945741643d-15*tu10
     $  - 1.1534731090907283d-16*tu11

      CsH3YZ(3,2,2,02) = 
     $  + 1.3820285212927298d-01
     $  - 3.2163608071436942d-03*tu01
     $  - 7.8440026557425447d-04*tu02
     $  + 1.1153246983155865d-05*tu03
     $  + 1.3194481519260970d-06*tu04
     $  - 2.2898256940470180d-08*tu05
     $  - 1.7356221351117680d-09*tu06
     $  + 3.6673428610095297d-11*tu07
     $  + 1.9787219019008594d-12*tu08
     $  - 5.0820807069570526d-14*tu09
     $  - 2.0496650024051116d-15*tu10

      CsH3YZ(3,2,2,03) = 
     $  - 3.2934425444413539d-02
     $  + 2.2339082574003802d-03*tu01
     $  + 1.4322644734904400d-05*tu02
     $  + 2.1568916971657980d-06*tu03
     $  - 9.0340599579861246d-08*tu04
     $  - 8.0623254907870538d-09*tu05
     $  + 2.4222413346484883d-10*tu06
     $  + 1.4724645156506940d-11*tu07
     $  - 4.7078488725444309d-13*tu08
     $  - 2.0828696541907349d-14*tu09
     $  + 7.6398241818355124d-16*tu10

      CsH3YZ(3,2,2,04) = 
     $  + 3.2721939447818592d-03
     $  - 3.1743387527505254d-05*tu01
     $  - 8.1394738561259806d-06*tu02
     $  - 1.1261734157029700d-07*tu03
     $  - 1.1245302336968326d-08*tu04
     $  + 8.5397567981279600d-10*tu05
     $  + 5.8462514574389076d-11*tu06
     $  - 2.6553862017684851d-12*tu07
     $  - 1.2903027897397695d-13*tu08
     $  + 5.8831461667283607d-15*tu09
     $  + 2.1022391019416180d-16*tu10

      CsH3YZ(3,2,2,05) = 
     $  - 3.6546734530868693d-05
     $  - 3.6278297520763327d-05*tu01
     $  + 3.2425230838095975d-07*tu02
     $  + 5.6764977846822000d-08*tu03
     $  + 1.0247987072707033d-09*tu04
     $  + 6.6938026077282246d-11*tu05
     $  - 8.6621382860777170d-12*tu06
     $  - 4.5250061065871301d-13*tu07
     $  + 2.9733168927955973d-14*tu08
     $  + 1.1351243116624697d-15*tu09

      CsH3YZ(3,2,2,06) = 
     $  - 1.9683585492988865d-05
     $  + 1.2168997150891317d-06*tu01
     $  + 2.9798348082360470d-07*tu02
     $  - 3.6138338571644185d-09*tu03
     $  - 4.8677360983706753d-10*tu04
     $  - 9.8941068336468639d-12*tu05
     $  - 3.7393292945123317d-13*tu06
     $  + 9.1406128985181846d-14*tu07
     $  + 3.5373767627940739d-15*tu08
     $  - 3.3753747447910206d-16*tu09

      CsH3YZ(3,2,2,07) = 
     $  + 5.6258870685760274d-07
     $  + 5.4837819198862178d-07*tu01
     $  - 1.6973850804419095d-08*tu02
     $  - 2.7257898097494969d-09*tu03
     $  + 4.2104296068706149d-11*tu04
     $  + 4.6538465770649250d-12*tu05
     $  + 9.7882393114888536d-14*tu06
     $  + 1.2484728277208733d-15*tu07
     $  - 9.8792516553184851d-16*tu08

      CsH3YZ(3,2,2,08) = 
     $  + 1.8608935845240703d-07
     $  - 2.7500931081226267d-08*tu01
     $  - 6.4801459946816451d-09*tu02
     $  + 2.2372178997267016d-10*tu03
     $  + 2.6386432907170813d-11*tu04
     $  - 5.0535930465164564d-13*tu05
     $  - 4.7651296800050529d-14*tu06
     $  - 9.7116683055833812d-16*tu07

      CsH3YZ(3,2,2,09) = 
     $  - 8.2289249023083281d-09
     $  - 7.8233566508191810d-09*tu01
     $  + 4.8104337978554273d-10*tu02
     $  + 7.2862142377623482d-11*tu03
     $  - 2.8836509759949748d-12*tu04
     $  - 2.6496554024523280d-13*tu05
     $  + 6.2009109631085896d-15*tu06
     $  + 5.1116368429549488d-16*tu07

      CsH3YZ(3,2,2,10) = 
     $  - 1.9277094294255402d-09
     $  + 5.2454438101194413d-10*tu01
     $  + 1.1792742851770124d-10*tu02
     $  - 7.3892498557386724d-12*tu03
     $  - 8.0308538773660062d-13*tu04
     $  + 3.6834904322169398d-14*tu05
     $  + 2.7283272029514682d-15*tu06

      CsH3YZ(3,2,2,11) = 
     $  + 1.1611032297688409d-10
     $  + 1.0676762903468549d-10*tu01
     $  - 1.0998622308991735d-11*tu02
     $  - 1.5587618672247953d-12*tu03
     $  + 1.0636334635063367d-13*tu04
     $  + 8.7573578966210478d-15*tu05
     $  - 4.6897143586577606d-16*tu06

      CsH3YZ(3,2,2,12) = 
     $  + 2.0520607405338275d-11
     $  - 9.1422201973136928d-12*tu01
     $  - 1.9361330052494214d-12*tu02
     $  + 1.9312650699009462d-13*tu03
     $  + 1.9202792455356262d-14*tu04
     $  - 1.4753751655049820d-15*tu05

      CsH3YZ(3,2,2,13) = 
     $  - 1.5948936061990290d-12
     $  - 1.4034751374738646d-12*tu01
     $  + 2.2341524490099443d-13*tu02
     $  + 2.9160723979210534d-14*tu03
     $  - 3.0806471320116422d-15*tu04
     $  - 2.2568007900453673d-16*tu05

      CsH3YZ(3,2,2,14) = 
     $  - 2.1939122260421864d-13
     $  + 1.5033676178869814d-13*tu01
     $  + 2.9487286189167796d-14*tu02
     $  - 4.4157299162379062d-15*tu03
     $  - 3.9374087795256544d-16*tu04

      CsH3YZ(3,2,2,15) = 
     $  + 2.1433338853306964d-14
     $  + 1.7800038203148985d-14*tu01
     $  - 4.1984687642059788d-15*tu02
     $  - 4.9332326746343709d-16*tu03

      CsH3YZ(3,2,2,16) = 
     $  + 2.3257389773288778d-15
     $  - 2.3705573986846661d-15*tu01
     $  - 4.2088610510225905d-16*tu02

      CsH3YZ(3,2,2,17) = 
     $  - 2.8251960454781361d-16
     $  - 2.1724812574549857d-16*tu01

      CsH3YZ(3,3,2,01) = 
     $  + 5.7304520677189100d-01
     $  - 1.6618147776727924d-01*tu01
     $  + 1.0444143671252788d-02*tu02
     $  - 3.2656925329858933d-04*tu03
     $  + 2.7452671035602238d-06*tu04
     $  + 1.3973439743468791d-07*tu05
     $  - 3.0870960088611882d-09*tu06
     $  - 8.7465274464562939d-11*tu07
     $  + 3.2600994248948837d-12*tu08
     $  + 5.3898924135532910d-14*tu09
     $  - 3.2534649433171003d-15*tu10

      CsH3YZ(3,3,2,02) = 
     $  - 1.6716048746695998d-01
     $  + 4.1798499828784126d-02*tu01
     $  - 1.9580193993614715d-03*tu02
     $  + 2.1925143772971650d-05*tu03
     $  + 1.3961204300643283d-06*tu04
     $  - 3.6993055510238257d-08*tu05
     $  - 1.2235442826311659d-09*tu06
     $  + 5.2096596038803654d-11*tu07
     $  + 9.6955983696684285d-13*tu08
     $  - 6.4994779237742853d-14*tu09
     $  - 6.2080708783715703d-16*tu10

      CsH3YZ(3,3,2,03) = 
     $  + 2.7909393772611685d-02
     $  - 5.2139488673574782d-03*tu01
     $  + 8.7454440177868647d-05*tu02
     $  + 7.4361976064460244d-06*tu03
     $  - 2.4613491401785916d-07*tu04
     $  - 9.7780205636337437d-09*tu05
     $  + 4.8545605047201409d-10*tu06
     $  + 1.0333697469718270d-11*tu07
     $  - 7.7884585631916421d-13*tu08
     $  - 8.2741059275893592d-15*tu09
     $  + 1.0915036339761555d-15*tu10

      CsH3YZ(3,3,2,04) = 
     $  - 2.5958445461897532d-03
     $  + 1.7392724721610007d-04*tu01
     $  + 2.2259774977971742d-05*tu02
     $  - 9.8163313963716858d-07*tu03
     $  - 4.8817841366269325d-08*tu04
     $  + 2.9065164343267533d-09*tu05
     $  + 7.2261451947352837d-11*tu06
     $  - 6.2198684712267047d-12*tu07
     $  - 7.4430319359728537d-14*tu08
     $  + 1.0898391086789975d-14*tu09

      CsH3YZ(3,3,2,05) = 
     $  + 6.8398734775273514d-05
     $  + 3.5459769251291792d-05*tu01
     $  - 2.3443282223330495d-06*tu02
     $  - 1.5587071346302983d-07*tu03
     $  + 1.1591312796115215d-08*tu04
     $  + 3.4637890899371961d-10*tu05
     $  - 3.4752941223085827d-11*tu06
     $  - 4.7606037582349206d-13*tu07
     $  + 7.8322201877790299d-14*tu08
     $  + 2.9366040653414903d-16*tu09
     $  - 1.4621768731278199d-16*tu10

      CsH3YZ(3,3,2,06) = 
     $  + 1.1664628558557991d-05
     $  - 3.0949993906737448d-06*tu01
     $  - 3.1058905041188648d-07*tu02
     $  + 3.0771572436973314d-08*tu03
     $  + 1.1523765141558904d-09*tu04
     $  - 1.3859501935054334d-10*tu05
     $  - 2.2198491700875947d-12*tu06
     $  + 4.1674553214353994d-13*tu07
     $  + 1.7659110534369943d-15*tu08
     $  - 9.7288120114280306d-16*tu09

      CsH3YZ(3,3,2,07) = 
     $  - 8.5810706856455421d-07
     $  - 3.5233251411182960d-07*tu01
     $  + 5.2356943255893031d-08*tu02
     $  + 2.6264006446197897d-09*tu03
     $  - 3.9432377890922038d-10*tu04
     $  - 7.6028162991815275d-12*tu05
     $  + 1.6619906709351812d-12*tu06
     $  + 8.0951462207163898d-15*tu07
     $  - 4.9914544502790794d-15*tu08

      CsH3YZ(3,3,2,08) = 
     $  - 8.6122817348327831d-08
     $  + 5.1573261763794282d-08*tu01
     $  + 3.9206223592606271d-09*tu02
     $  - 7.8368149209615945d-10*tu03
     $  - 1.8978607665983254d-11*tu04
     $  + 4.9660657252326316d-12*tu05
     $  + 2.8433084382697300d-14*tu06
     $  - 1.9906285857617822d-14*tu07
     $  + 1.0007368224884026d-16*tu08

      CsH3YZ(3,3,2,09) = 
     $  + 1.0943757512521399d-08
     $  + 3.4513340796479397d-09*tu01
     $  - 1.0339346477553451d-09*tu02
     $  - 3.3663573029366287d-11*tu03
     $  + 1.0974008371450723d-11*tu04
     $  + 7.6173931021631030d-14*tu05
     $  - 6.1693241288957812d-14*tu06
     $  + 3.5237187729262597d-16*tu07
     $  + 2.3742532958163345d-16*tu08

      CsH3YZ(3,3,2,10) = 
     $  + 6.7014533319549423d-10
     $  - 8.0973661163848719d-10*tu01
     $  - 4.0242965468377904d-11*tu02
     $  + 1.7411106565457689d-11*tu03
     $  + 1.5332216646831666d-13*tu04
     $  - 1.4730682924105874d-13*tu05
     $  + 9.7430411644677620d-16*tu06
     $  + 7.5694963298555160d-16*tu07

      CsH3YZ(3,3,2,11) = 
     $  - 1.3786082775545565d-10
     $  - 2.9042507079381356d-11*tu01
     $  + 1.8728021448618374d-11*tu02
     $  + 2.2510417048679344d-13*tu03
     $  - 2.6591298626159413d-13*tu04
     $  + 2.0901103274340378d-15*tu05
     $  + 1.9176124878043009d-15*tu06

      CsH3YZ(3,3,2,12) = 
     $  - 4.7261652201246049d-12
     $  + 1.2134607869456353d-11*tu01
     $  + 2.2850525289694815d-13*tu02
     $  - 3.5073976292278457d-13*tu03
     $  + 3.4010824494744318d-15*tu04
     $  + 3.8108854637545380d-15*tu05

      CsH3YZ(3,3,2,13) = 
     $  + 1.7078882480510609d-12
     $  + 1.4465417325001862d-13*tu01
     $  - 3.1794900219958094d-13*tu02
     $  + 4.0355560697264045d-15*tu03
     $  + 5.8107789445468384d-15*tu04
     $  - 1.5039157514701231d-16*tu05

      CsH3YZ(3,3,2,14) = 
     $  + 2.2288762674895989d-14
     $  - 1.7513291948818978d-13*tu01
     $  + 3.2477615625041706d-15*tu02
     $  + 6.5522246231291647d-15*tu03
     $  - 2.1128650049999608d-16*tu04

      CsH3YZ(3,3,2,15) = 
     $  - 2.0788640558162514d-14
     $  + 1.5123919383868597d-15*tu01
     $  + 5.1248307477737549d-15*tu02
     $  - 2.1974756060404665d-16*tu03
     $  - 1.1694894952863681d-16*tu04

      CsH3YZ(3,3,2,16) = 
     $  + 1.1009927742649883d-16
     $  + 2.4473568793437434d-15*tu01
     $  - 1.5789942963328906d-16*tu02
     $  - 1.1505849435108498d-16*tu03

      CsH3YZ(3,3,2,17) = 
     $  + 2.4841333488707231d-16
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CsH4YZ(0,0,3,2,01) = 
     $  + 6.2246945526043164d-01
     $  - 1.2318938491474078d-01*tu01
     $  + 4.3234854072345596d-03*tu02
     $  - 2.0133335065426132d-05*tu03
     $  - 2.4360513938079663d-06*tu04
     $  + 2.4828889022590042d-08*tu05
     $  + 1.9359756483168612d-09*tu06
     $  - 2.8454404383462713d-11*tu07
     $  - 1.5902232751400798d-12*tu08
     $  + 3.0733514480493260d-14*tu09
     $  + 1.2949674406911631d-15*tu10

      CsH4YZ(0,0,3,2,02) = 
     $  - 2.0588194953559553d-01
     $  + 3.8965633613290308d-02*tu01
     $  - 1.2461183006125917d-03*tu02
     $  + 7.9619927205565855d-07*tu03
     $  + 7.4711222477584428d-07*tu04
     $  - 1.1815525171118935d-09*tu05
     $  - 6.4394544021047496d-10*tu06
     $  + 1.6483767738207903d-12*tu07
     $  + 5.8537651689491964d-13*tu08
     $  - 2.1751594154170553d-15*tu09
     $  - 5.3880399314188274d-16*tu10

      CsH4YZ(0,0,3,2,03) = 
     $  + 4.1671021687332836d-02
     $  - 7.2886939565030497d-03*tu01
     $  + 1.9532777567076044d-04*tu02
     $  + 1.2828514492239833d-06*tu03
     $  - 1.1650449298791102d-07*tu04
     $  - 1.6710806274256649d-09*tu05
     $  + 1.0082107409309631d-10*tu06
     $  + 2.0417641003659532d-12*tu07
     $  - 9.3494022814004488d-14*tu08
     $  - 2.3678357555748255d-15*tu09

      CsH4YZ(0,0,3,2,04) = 
     $  - 5.4446876437461436d-03
     $  + 8.1322835024998212d-04*tu01
     $  - 1.3809465453000942d-05*tu02
     $  - 3.5905691314552265d-07*tu03
     $  + 4.5670054747742487d-09*tu04
     $  + 4.7001305470378923d-10*tu05
     $  + 2.7777279604838070d-13*tu06
     $  - 5.8400502965576432d-13*tu07
     $  - 5.0274364430176715d-15*tu08
     $  + 6.9995325534355346d-16*tu09

      CsH4YZ(0,0,3,2,05) = 
     $  + 4.1926613550130633d-04
     $  - 3.8856138181913397d-05*tu01
     $  - 5.8439850539103956d-07*tu02
     $  + 3.2135553471577610d-08*tu03
     $  + 1.5583684032295857d-09*tu04
     $  - 2.9027123993250229d-11*tu05
     $  - 2.7886345525709947d-12*tu06
     $  + 1.6798472123623649d-14*tu07
     $  + 4.3097910237623735d-15*tu08

      CsH4YZ(0,0,3,2,06) = 
     $  - 8.2267646996224340d-06
     $  - 2.5269994482966135d-06*tu01
     $  + 1.7515610287839441d-07*tu02
     $  + 2.6782243469893761d-09*tu03
     $  - 2.1780790168490120d-10*tu04
     $  - 8.5889666471429970d-12*tu05
     $  + 2.7275334720462667d-13*tu06
     $  + 1.8659004314971076d-14*tu07
     $  - 2.8950829397383311d-16*tu08

      CsH4YZ(0,0,3,2,07) = 
     $  - 1.6745397975836911d-06
     $  + 4.1096834299334841d-07*tu01
     $  - 7.7474776951191603d-10*tu02
     $  - 7.4636520116833098d-10*tu03
     $  - 1.5516891624997095d-11*tu04
     $  + 1.4815641463528029d-12*tu05
     $  + 5.6527595833426814d-14*tu06
     $  - 2.4352506229895347d-15*tu07
     $  - 1.3888210018226667d-16*tu08

      CsH4YZ(0,0,3,2,08) = 
     $  + 9.7003217558870673d-08
     $  + 9.7706773577940999d-09*tu01
     $  - 2.2920218953750402d-09*tu02
     $  - 1.2034954619756416d-12*tu03
     $  + 5.4351511582304277d-12*tu04
     $  + 9.9897693191992196d-14*tu05
     $  - 1.2205015828799991d-14*tu06
     $  - 4.0147075316063454d-16*tu07

      CsH4YZ(0,0,3,2,09) = 
     $  + 1.2353227060655537d-08
     $  - 4.7677532578199210d-09*tu01
     $  + 9.3820616371468279d-11*tu02
     $  + 1.3233302223804743d-11*tu03
     $  + 4.7264301612006379d-15*tu04
     $  - 4.2188295232582777d-14*tu05
     $  - 6.8884886697516354d-16*tu06
     $  + 1.0743076758923065d-16*tu07

      CsH4YZ(0,0,3,2,10) = 
     $  - 1.2300286996343375d-09
     $  + 2.9582124805994272d-11*tu01
     $  + 2.9666173091024695d-11*tu02
     $  - 5.8304126099748578d-13*tu03
     $  - 1.0987632162268543d-13*tu04
     $  + 1.1818738435896230d-16*tu05
     $  + 3.6833963895085286d-16*tu06

      CsH4YZ(0,0,3,2,11) = 
     $  - 1.0499293646286537d-10
     $  + 5.6961029532283679d-11*tu01
     $  - 2.2506766307362228d-12*tu02
     $  - 2.0898169150970732d-13*tu03
     $  + 4.9842535610053597d-15*tu04
     $  + 9.6027448674048249d-16*tu05

      CsH4YZ(0,0,3,2,12) = 
     $  + 1.5843087665106759d-11
     $  - 1.7997263052250811d-12*tu01
     $  - 3.7514527924491013d-13*tu02
     $  + 1.7797415683308937d-14*tu03
     $  + 1.9560813628794965d-15*tu04

      CsH4YZ(0,0,3,2,13) = 
     $  + 9.2597080761763248d-13
     $  - 6.8808145489272811d-13*tu01
     $  + 4.2835681521532623d-14*tu02
     $  + 3.0382254301423680d-15*tu03
     $  - 1.7221750313509976d-16*tu04

      CsH4YZ(0,0,3,2,14) = 
     $  - 2.0519573038930792d-13
     $  + 3.8092243282155376d-14*tu01
     $  + 4.5942857822397081d-15*tu02
     $  - 3.9857389142940939d-16*tu03

      CsH4YZ(0,0,3,2,15) = 
     $  - 8.0476597683871621d-15
     $  + 8.3420157595277165d-15*tu01
     $  - 7.4008608787450270d-16*tu02

      CsH4YZ(0,0,3,2,16) = 
     $  + 2.6684502870551095d-15
     $  - 6.6063125450329641d-16*tu01

      CsH4YZ(0,0,3,2,17) = 
     $  - 1.0108384936729807d-16*tu01

      CsH4YZ(0,2,3,2,02) = 
     $  + 1.1671302286133093d-01
     $  - 2.3098009671513896d-02*tu01
     $  + 8.1065351385647993d-04*tu02
     $  - 3.7750003247673997d-06*tu03
     $  - 4.5675963633899368d-07*tu04
     $  + 4.6554166917356328d-09*tu05
     $  + 3.6299543405941148d-10*tu06
     $  - 5.3352008218992588d-12*tu07
     $  - 2.9816686408876498d-13*tu08
     $  + 5.7625339650924862d-15*tu09
     $  + 2.4280639512959309d-16*tu10

      CsH4YZ(0,2,3,2,03) = 
     $  - 3.9449060796532443d-02
     $  + 7.2140421198849621d-03*tu01
     $  - 2.1270938840674393d-04*tu02
     $  - 6.7835032383996376d-07*tu03
     $  + 1.3484749917386633d-07*tu04
     $  + 7.7000333389661035d-10*tu05
     $  - 1.2389962155530544d-10*tu06
     $  - 7.8434128086788458d-13*tu07
     $  + 1.2058378960944863d-13*tu08
     $  + 7.1558035280076977d-16*tu09
     $  - 1.1889973226489597d-16*tu10

      CsH4YZ(0,2,3,2,04) = 
     $  + 6.1621315481266071d-03
     $  - 9.6123733391171738d-04*tu01
     $  + 1.8182646975997893d-05*tu02
     $  + 4.2956303645371186d-07*tu03
     $  - 9.2743912440218157d-09*tu04
     $  - 5.8368871218942963d-10*tu05
     $  + 6.7256522122943994d-12*tu06
     $  + 7.4650009854284547d-13*tu07
     $  - 5.4001952272826247d-15*tu08
     $  - 9.0724297556534956d-16*tu09

      CsH4YZ(0,2,3,2,05) = 
     $  - 4.5501142447624384d-04
     $  + 4.0345324596930439d-05*tu01
     $  + 9.2842805644235368d-07*tu02
     $  - 4.5674704312616965d-08*tu03
     $  - 2.0665184850452205d-09*tu04
     $  + 5.6488941680559673d-11*tu05
     $  + 3.6324984116148477d-12*tu06
     $  - 6.7994878728450943d-14*tu07
     $  - 5.5725402663181933d-15*tu08

      CsH4YZ(0,2,3,2,06) = 
     $  - 7.6311808200160633d-08
     $  + 4.4960482594641946d-06*tu01
     $  - 2.3881611729904947d-07*tu02
     $  - 4.3464332335916554d-09*tu03
     $  + 3.2373636527862008d-10*tu04
     $  + 1.2067859919801973d-11*tu05
     $  - 4.9709633672943421d-13*tu06
     $  - 2.4939848905882759d-14*tu07
     $  + 7.3809116869143571d-16*tu08

      CsH4YZ(0,2,3,2,07) = 
     $  + 2.2164039702557176d-06
     $  - 4.3864306847976506d-07*tu01
     $  - 5.5549124137899535d-09*tu02
     $  + 1.0326005049093888d-09*tu03
     $  + 2.8115788123162354d-11*tu04
     $  - 2.2615827650418204d-12*tu05
     $  - 8.3198559546432299d-14*tu06
     $  + 4.3361903416666575d-15*tu07
     $  + 1.8803700782908855d-16*tu08

      CsH4YZ(0,2,3,2,08) = 
     $  + 4.8643623313242500d-09
     $  - 3.7227793348097170d-08*tu01
     $  + 3.2761763579103191d-09*tu02
     $  + 3.6836663295091584d-11*tu03
     $  - 7.8485562038608514d-12*tu04
     $  - 1.9318726609337972d-13*tu05
     $  + 1.9098047993386904d-14*tu06
     $  + 6.0783052470979544d-16*tu07

      CsH4YZ(0,2,3,2,09) = 
     $  - 2.0234410915256339d-08
     $  + 5.2538445994088280d-09*tu01
     $  + 1.5993202309745203d-11*tu02
     $  - 1.8995962975391909d-11*tu03
     $  - 2.8858881206717163d-13*tu04
     $  + 6.2820877571065297d-14*tu05
     $  + 1.4309379873886335d-15*tu06
     $  - 1.7142224310451065d-16*tu07

      CsH4YZ(0,2,3,2,10) = 
     $  - 4.5744493901584757d-11
     $  + 3.5646776876620501d-10*tu01
     $  - 4.5147796803734759d-11*tu02
     $  - 1.1899683808430917d-13*tu03
     $  + 1.6329434645522684d-13*tu04
     $  + 2.1296726039757198d-15*tu05
     $  - 5.6587831438077450d-16*tu06

      CsH4YZ(0,2,3,2,11) = 
     $  + 2.1688663893100479d-10
     $  - 6.5160699871284362d-11*tu01
     $  + 4.2542355570790238d-13*tu02
     $  + 3.2069498288046573d-13*tu03
     $  + 9.3513770473305540d-16*tu04
     $  - 1.4727637796550357d-15*tu05

      CsH4YZ(0,2,3,2,12) = 
     $  + 3.3694017517736673d-13
     $  - 3.6312206111767886d-12*tu01
     $  + 6.2009967997764726d-13*tu02
     $  - 4.7927580098714391d-15*tu03
     $  - 3.0967292258857902d-15*tu04

      CsH4YZ(0,2,3,2,13) = 
     $  - 2.4931735196448455d-12
     $  + 8.2237591136896357d-13*tu01
     $  - 1.3453988759838569d-14*tu02
     $  - 5.1386614542452808d-15*tu03

      CsH4YZ(0,2,3,2,14) = 
     $  - 2.0579577286059469d-15
     $  + 3.8222238514054028d-14*tu01
     $  - 8.4731644731080419d-15*tu02
     $  + 1.6909381391121494d-16*tu03

      CsH4YZ(0,2,3,2,15) = 
     $  + 2.9833084301665158d-14
     $  - 1.0492896852190678d-14*tu01
     $  + 2.7878036425089874d-16*tu02

      CsH4YZ(0,2,3,2,16) = 
     $  - 4.1010682672290155d-16*tu01
     $  + 1.1512875086241065d-16*tu02

      CsH4YZ(0,2,3,2,17) = 
     $  - 3.6672586651298805d-16
     $  + 1.3493425622611239d-16*tu01

      CsH4YZ(0,3,0,2,01) = 
     $  + 6.7820922479746498d-01
     $  - 7.2161137820533973d-02*tu01
     $  - 2.9609515081006315d-04*tu02
     $  + 7.4754123486713770d-05*tu03
     $  + 4.0333501318342053d-07*tu04
     $  - 8.3808506326028907d-08*tu05
     $  - 3.9905982933620598d-10*tu06
     $  + 8.8548158315861998d-11*tu07
     $  + 2.9721341909005945d-13*tu08
     $  - 8.9300569770748624d-14*tu09
     $  - 1.3237074874293598d-16*tu10

      CsH4YZ(0,3,0,2,02) = 
     $  - 2.2475770707350849d-01
     $  + 2.1576872199839062d-02*tu01
     $  + 2.1908572808401454d-04*tu02
     $  - 2.2138052513809221d-05*tu03
     $  - 3.7010327170706190d-07*tu04
     $  + 2.4925257383965125d-08*tu05
     $  + 4.9495350005753574d-10*tu06
     $  - 2.6935245709825401d-11*tu07
     $  - 5.8222297111161508d-13*tu08
     $  + 2.8356687777465037d-14*tu09
     $  + 6.3286606690788470d-16*tu10

      CsH4YZ(0,3,0,2,03) = 
     $  + 4.5609186165083798d-02
     $  - 3.6225686861380134d-03*tu01
     $  - 7.5063382205817969d-05*tu02
     $  + 3.1845193866408569d-06*tu03
     $  + 1.3621238516833062d-07*tu04
     $  - 2.8717052031355987d-09*tu05
     $  - 1.9837745161098827d-10*tu06
     $  + 2.2678634858791256d-12*tu07
     $  + 2.5770979848360055d-13*tu08
     $  - 1.5285951998953740d-15*tu09
     $  - 3.1374825227737678d-16*tu10

      CsH4YZ(0,3,0,2,04) = 
     $  - 5.9766624175724625d-03
     $  + 3.0874015105202028d-04*tu01
     $  + 1.4028760156457978d-05*tu02
     $  - 3.5510711465589820d-08*tu03
     $  - 2.4421707066716498d-08*tu04
     $  - 3.8214973966373364d-10*tu05
     $  + 3.3647568771119351d-11*tu06
     $  + 9.8198160320131833d-13*tu07
     $  - 4.0981220060980332d-14*tu08
     $  - 1.7362800717232437d-15*tu09

      CsH4YZ(0,3,0,2,05) = 
     $  + 4.6069934984255876d-04
     $  + 2.0825390188640761d-06*tu01
     $  - 1.1767809841465422d-06*tu02
     $  - 6.6480276827954993d-08*tu03
     $  + 9.8204734398046696d-10*tu04
     $  + 1.6912043548698456d-10*tu05
     $  + 8.7399746259546557d-13*tu06
     $  - 3.0687790500052130d-13*tu07
     $  - 4.9473692830023621d-15*tu08
     $  + 4.7482033892375392d-16*tu09

      CsH4YZ(0,3,0,2,06) = 
     $  - 8.5968716686277756d-06
     $  - 3.1409352036066530d-06*tu01
     $  - 6.3280193430613744d-08*tu02
     $  + 8.7459696425818644d-09*tu03
     $  + 4.5503476303483954d-10*tu04
     $  - 1.2594393185149843d-11*tu05
     $  - 1.3343764761915405d-12*tu06
     $  + 7.4075826698724417d-15*tu07
     $  + 2.8517420785175372d-15*tu08

      CsH4YZ(0,3,0,2,07) = 
     $  - 1.9614155990062708d-06
     $  + 1.4685360787437887d-07*tu01
     $  + 2.3249638856106198d-08*tu02
     $  + 4.6372693689461996d-10*tu03
     $  - 7.3715239102705157d-11*tu04
     $  - 3.4177369929854061d-12*tu05
     $  + 1.4104332767725329d-13*tu06
     $  + 1.1204532331908173d-14*tu07
     $  - 1.8269056053862140d-16*tu08

      CsH4YZ(0,3,0,2,08) = 
     $  + 1.1724802808670338d-07
     $  + 3.1800856888399919d-08*tu01
     $  - 7.0566649503309487d-10*tu02
     $  - 2.0999685178227973d-10*tu03
     $  - 3.7874939919506539d-12*tu04
     $  + 6.9794665029638190d-13*tu05
     $  + 2.7675845411720929d-14*tu06
     $  - 1.5839042449785955d-15*tu07

      CsH4YZ(0,3,0,2,09) = 
     $  + 1.4245576427106495d-08
     $  - 3.2924565812480104d-09*tu01
     $  - 3.3234748767267790d-10*tu02
     $  + 6.0417158630016713d-12*tu03
     $  + 2.0063524356902353d-12*tu04
     $  + 2.9644859511914467d-14*tu05
     $  - 7.0061507310302959d-15*tu06
     $  - 2.3107198857372813d-16*tu07

      CsH4YZ(0,3,0,2,10) = 
     $  - 1.5525934160315550d-09
     $  - 2.9677807014551027d-10*tu01
     $  + 3.0048572960321888d-11*tu02
     $  + 3.6652940101073708d-12*tu03
     $  - 5.9264220990774877d-14*tu04
     $  - 2.0379853829944386d-14*tu05
     $  - 2.1910024436627651d-16*tu06

      CsH4YZ(0,3,0,2,11) = 
     $  - 1.1252073414536224d-10
     $  + 5.5727849958479077d-11*tu01
     $  + 3.9742039316136943d-12*tu02
     $  - 3.2352488036392862d-13*tu03
     $  - 4.0301651948680475d-14*tu04
     $  + 6.9277781934624816d-16*tu05
     $  + 2.1485848456960290d-16*tu06

      CsH4YZ(0,3,0,2,12) = 
     $  + 2.0150898998078103d-11
     $  + 2.2843211533690969d-12*tu01
     $  - 6.7030339585528783d-13*tu02
     $  - 5.2435472523514542d-14*tu03
     $  + 3.6142457931188797d-15*tu04
     $  + 4.5104749946136008d-16*tu05

      CsH4YZ(0,3,0,2,13) = 
     $  + 8.4889420100731307d-13
     $  - 8.5057147535237144d-13*tu01
     $  - 3.7984452592183018d-14*tu02
     $  + 8.4803710522288719d-15*tu03
     $  + 6.5864925630259802d-16*tu04

      CsH4YZ(0,3,0,2,14) = 
     $  - 2.5825909109790994d-13
     $  - 8.2905232714955170d-15*tu01
     $  + 1.2338864937316674d-14*tu02
     $  + 6.0037743094895113d-16*tu03
     $  - 1.0670792530549872d-16*tu04

      CsH4YZ(0,3,0,2,15) = 
     $  - 5.1715495192639500d-15
     $  + 1.2277146383869676d-14*tu01
     $  + 1.9534165027439547d-16*tu02
     $  - 1.7822661131692553d-16*tu03

      CsH4YZ(0,3,0,2,16) = 
     $  + 3.2837090414088631d-15
     $  - 1.8314889294595663d-16*tu01
     $  - 2.0550885914162369d-16*tu02

      CsH4YZ(0,3,0,2,17) = 
     $  - 1.7074058595662049d-16*tu01

      CsH4YZ(0,3,2,2,01) = 
     $  - 1.0516401802557398d-01
     $  - 9.4020339946745276d-02*tu01
     $  + 1.0740238822062851d-02*tu02
     $  - 4.0132337678530310d-04*tu03
     $  + 2.3419320903768033d-06*tu04
     $  + 2.2354290376071681d-07*tu05
     $  - 2.6880361795249822d-09*tu06
     $  - 1.7601343278042493d-10*tu07
     $  + 2.9628860058048243d-12*tu08
     $  + 1.4319949390628153d-13*tu09
     $  - 3.1210941945741643d-15*tu10
     $  - 1.1534731090907283d-16*tu11

      CsH4YZ(0,3,2,2,02) = 
     $  + 8.8819679444431613d-02
     $  + 1.6020633336442892d-02*tu01
     $  - 2.4059949119239119d-03*tu02
     $  + 8.0824756638822264d-05*tu03
     $  + 2.2061180901739792d-07*tu04
     $  - 5.3363422925369493d-08*tu05
     $  - 3.6380428389494986d-10*tu06
     $  + 5.1339232951377324d-11*tu07
     $  + 4.3381982486202517d-13*tu08
     $  - 5.2260308642213050d-14*tu09
     $  - 4.3962733971990000d-16*tu10

      CsH4YZ(0,3,2,2,03) = 
     $  - 2.9896686112613229d-02
     $  - 3.2238662497480642d-04*tu01
     $  + 2.7064047970314866d-04*tu02
     $  - 6.9458697361095781d-06*tu03
     $  - 1.5845186327191166d-07*tu04
     $  + 3.6676981585576215d-09*tu05
     $  + 2.5569357940550944d-10*tu06
     $  - 2.4261734112870715d-12*tu07
     $  - 3.5797343964838808d-13*tu08
     $  + 1.6471941286795149d-15*tu09
     $  + 4.6210183457160138d-16*tu10

      CsH4YZ(0,3,2,2,04) = 
     $  + 5.5252155449991518d-03
     $  - 2.5505647422187712d-04*tu01
     $  - 1.2569807020127077d-05*tu02
     $  - 9.9660818918010026d-08*tu03
     $  + 2.1120388656753860d-08*tu04
     $  + 7.0099798619335037d-10*tu05
     $  - 2.8432205764573341d-11*tu06
     $  - 1.6145152923400927d-12*tu07
     $  + 3.5066660724510498d-14*tu08
     $  + 2.8279205096393745d-15*tu09

      CsH4YZ(0,3,2,2,05) = 
     $  - 5.5224120003441114d-04
     $  + 2.4331490914170702d-05*tu01
     $  - 1.3438050744499858d-07*tu02
     $  + 7.5292718260836812d-08*tu03
     $  - 4.4195794652077528d-12*tu04
     $  - 1.4589219777247613d-10*tu05
     $  - 3.6100118192207034d-12*tu06
     $  + 2.6216732195861338d-13*tu07
     $  + 1.0949886487775122d-14*tu08
     $  - 4.2206602327162128d-16*tu09

      CsH4YZ(0,3,2,2,06) = 
     $  + 1.4445309310935771d-05
     $  + 2.4577883004750648d-06*tu01
     $  + 2.3259696022045916d-08*tu02
     $  - 5.8468279460791680d-09*tu03
     $  - 3.2967204762472828d-10*tu04
     $  + 2.5161589913566110d-12*tu05
     $  + 1.0629206449327395d-12*tu06
     $  + 2.0084396161188450d-14*tu07
     $  - 2.4217352392318103d-15*tu08

      CsH4YZ(0,3,2,2,07) = 
     $  + 2.8692849966714411d-06
     $  - 4.4908030977511433d-07*tu01
     $  - 8.0857758492724676d-09*tu02
     $  - 2.8851407437612570d-10*tu03
     $  + 4.6595116626371449d-11*tu04
     $  + 2.3372199805404067d-12*tu05
     $  - 4.1456738334017481d-14*tu06
     $  - 8.8209308355152702d-15*tu07

      CsH4YZ(0,3,2,2,08) = 
     $  - 1.9651598685389568d-07
     $  - 2.2635105688920027d-08*tu01
     $  + 1.6016659192245859d-09*tu02
     $  + 1.4895341894830891d-10*tu03
     $  + 6.0905964207171123d-14*tu04
     $  - 4.0304859439671550d-13*tu05
     $  - 1.7223987028905382d-14*tu06
     $  + 5.2850982390004793d-16*tu07

      CsH4YZ(0,3,2,2,09) = 
     $  - 2.4523691674341416d-08
     $  + 7.2296327048485497d-09*tu01
     $  + 1.2981471307219570d-10*tu02
     $  - 1.3093373735377413d-11*tu03
     $  - 1.3312009703796610d-12*tu04
     $  + 5.2582726040758769d-15*tu05
     $  + 3.7903424923164843d-15*tu06
     $  + 1.3710626739916013d-16*tu07

      CsH4YZ(0,3,2,2,10) = 
     $  + 2.6351292989024316d-09
     $  + 1.7138029344199722d-10*tu01
     $  - 4.8015446702271485d-11*tu02
     $  - 2.4223398740020763d-12*tu03
     $  + 1.5536728266211853d-13*tu04
     $  + 1.2377754391525048d-14*tu05
     $  - 1.1786105985688290d-16*tu06

      CsH4YZ(0,3,2,2,11) = 
     $  + 2.3573950483786316d-10
     $  - 1.0968949918054963d-10*tu01
     $  - 1.0785105110973947d-12*tu02
     $  + 5.0083997287736412d-13*tu03
     $  + 2.4555290374620061d-14*tu04
     $  - 1.6542804336815275d-15*tu05
     $  - 1.1987548966591043d-16*tu06

      CsH4YZ(0,3,2,2,12) = 
     $  - 3.4946243837693714d-11
     $  - 5.4155270434040659d-13*tu01
     $  + 1.0057667913206446d-12*tu02
     $  + 2.7844966179742722d-14*tu03
     $  - 5.8588909050021123d-15*tu04
     $  - 2.4720819582688982d-16*tu05

      CsH4YZ(0,3,2,2,13) = 
     $  - 2.3733966470143867d-12
     $  + 1.6075153515357840d-12*tu01
     $  - 5.1664815750435223d-15*tu02
     $  - 1.2295063814949150d-14*tu03
     $  - 3.0854639058350974d-16*tu04

      CsH4YZ(0,3,2,2,14) = 
     $  + 4.6040923456021797d-13
     $  - 1.6319373755936253d-14*tu01
     $  - 1.8293898151254747d-14*tu02
     $  - 1.2530218841763089d-16*tu03
     $  + 1.5534509730349152d-16*tu04

      CsH4YZ(0,3,2,2,15) = 
     $  + 2.4410123658472221d-14
     $  - 2.3011752025453435d-14*tu01
     $  + 4.6588534102166864d-16*tu02
     $  + 2.5359051519918664d-16*tu03

      CsH4YZ(0,3,2,2,16) = 
     $  - 6.0449163193598150d-15
     $  + 5.3608885869091070d-16*tu01
     $  + 3.0728232820348391d-16*tu02

      CsH4YZ(0,3,2,2,17) = 
     $  - 2.5337878440954621d-16
     $  + 3.2369215088533354d-16*tu01

      CsH4YZ(0,3,3,2,01) = 
     $  + 5.7304520677189100d-01
     $  - 1.6618147776727924d-01*tu01
     $  + 1.0444143671252788d-02*tu02
     $  - 3.2656925329858933d-04*tu03
     $  + 2.7452671035602238d-06*tu04
     $  + 1.3973439743468791d-07*tu05
     $  - 3.0870960088611882d-09*tu06
     $  - 8.7465274464562939d-11*tu07
     $  + 3.2600994248948837d-12*tu08
     $  + 5.3898924135532910d-14*tu09
     $  - 3.2534649433171003d-15*tu10

      CsH4YZ(0,3,3,2,02) = 
     $  - 1.9102622000320955d-01
     $  + 5.2058276995756922d-02*tu01
     $  - 2.9372866380406336d-03*tu02
     $  + 7.2194306879971325d-05*tu03
     $  + 1.8332263311462221d-07*tu04
     $  - 4.4696727274123112d-08*tu05
     $  - 3.2941639654110160d-11*tu06
     $  + 4.2448036981507378d-11*tu07
     $  - 1.2648872368436927d-13*tu08
     $  - 4.2603437894283847d-14*tu09
     $  + 2.9962113295337780d-16*tu10

      CsH4YZ(0,3,3,2,03) = 
     $  + 3.9152023546718023d-02
     $  - 9.5593810244975801d-03*tu01
     $  + 4.3709364984279831d-04*tu02
     $  - 5.3645550189299065d-06*tu03
     $  - 2.1366522660419893d-07*tu04
     $  + 3.5481417108188662d-09*tu05
     $  + 2.6652584368111103d-10*tu06
     $  - 4.4341535951198291d-12*tu07
     $  - 3.2987121154659420d-13*tu08
     $  + 6.2084827851391052d-15*tu09
     $  + 3.9060004089903344d-16*tu10

      CsH4YZ(0,3,3,2,04) = 
     $  - 5.2243787652332211d-03
     $  + 1.0221156879873525d-03*tu01
     $  - 2.5579449858449494d-05*tu02
     $  - 6.8561653192409540d-07*tu03
     $  + 2.7231474137423313d-08*tu04
     $  + 1.2098061671617472d-09*tu05
     $  - 4.1784551306997088d-11*tu06
     $  - 1.9132442707630337d-12*tu07
     $  + 6.5771248529444700d-14*tu08
     $  + 2.7386376332166972d-15*tu09
     $  - 1.0044361832986640d-16*tu10

      CsH4YZ(0,3,3,2,05) = 
     $  + 4.1965335316108501d-04
     $  - 4.0227566378035727d-05*tu01
     $  - 2.2364057060031741d-06*tu02
     $  + 1.4086507111880948d-07*tu03
     $  + 3.4308026012519651d-09*tu04
     $  - 2.5266323708436204d-10*tu05
     $  - 7.5477198698396808d-12*tu06
     $  + 4.7584385159440020d-13*tu07
     $  + 1.3658502248529383d-14*tu08
     $  - 8.4095420340266818d-16*tu09

      CsH4YZ(0,3,3,2,06) = 
     $  - 1.0367651915155893d-05
     $  - 4.4346757527874375d-06*tu01
     $  + 4.1206978815257570d-07*tu02
     $  + 5.5956606030017572d-09*tu03
     $  - 1.0160385973832172d-09*tu04
     $  - 1.9331239752808161d-11*tu05
     $  + 2.4562416178633173d-12*tu06
     $  + 4.6802971092913799d-14*tu07
     $  - 5.2533186618252021d-15*tu08

      CsH4YZ(0,3,3,2,07) = 
     $  - 1.5597637091721213d-06
     $  + 5.4052692937421151d-07*tu01
     $  + 1.1710668021208909d-08*tu02
     $  - 3.0584675830549835d-09*tu03
     $  - 2.2685723905650299d-11*tu04
     $  + 9.3894600580125818d-12*tu05
     $  + 9.0555086607966354d-14*tu06
     $  - 2.5079387692612402d-14*tu07
     $  - 2.3155203062038468d-16*tu08

      CsH4YZ(0,3,3,2,08) = 
     $  + 1.1742128028581092d-07
     $  + 2.6024179090841452d-08*tu01
     $  - 6.4472760520756700d-09*tu02
     $  + 2.9282328814158240d-11*tu03
     $  + 2.6572565528419413d-11*tu04
     $  - 2.1815158841074305d-14*tu05
     $  - 9.2228197262585248d-14*tu06
     $  + 2.6163264931413689d-16*tu08

      CsH4YZ(0,3,3,2,09) = 
     $  + 9.9768223399270073d-09
     $  - 7.2133265367422976d-09*tu01
     $  + 7.0065475372810545d-11*tu02
     $  + 5.8069108790698463d-11*tu03
     $  - 7.9584750393653294d-13*tu04
     $  - 2.6264600513139213d-13*tu05
     $  + 2.6594583786032467d-15*tu06
     $  + 9.4155040886623393d-16*tu07

      CsH4YZ(0,3,3,2,10) = 
     $  - 1.4039667790799629d-09
     $  - 7.6219421215513295d-11*tu01
     $  + 9.6081774404188897d-11*tu02
     $  - 2.7814862571662289d-12*tu03
     $  - 5.7608380157836940d-13*tu04
     $  + 1.3798643161457875d-14*tu05
     $  + 2.6944154257516846d-15*tu06

      CsH4YZ(0,3,3,2,11) = 
     $  - 6.6613498609000772d-11
     $  + 9.4189742180920781d-11*tu01
     $  - 4.2192908803382056d-12*tu02
     $  - 9.9572647091773857d-13*tu03
     $  + 4.1464054634370969d-14*tu04
     $  + 6.1140080800046319d-15*tu05
     $  - 2.1075791441821592d-16*tu06

      CsH4YZ(0,3,3,2,12) = 
     $  + 1.6934076965966517d-11
     $  - 1.7554676885506566d-12*tu01
     $  - 1.3646728025977109d-12*tu02
     $  + 8.4459291570090261d-14*tu03
     $  + 1.0925812568590581d-14*tu04
     $  - 5.9666846274232082d-16*tu05

      CsH4YZ(0,3,3,2,13) = 
     $  + 3.6283789464788109d-13
     $  - 1.2031176703839942d-12*tu01
     $  + 1.0452138747425998d-13*tu02
     $  + 1.5651946360579649d-14*tu03
     $  - 1.2590263761277537d-15*tu04
     $  - 1.2324798092974765d-16*tu05

      CsH4YZ(0,3,3,2,14) = 
     $  - 2.0430729082221150d-13
     $  + 5.4271749478481846d-14*tu01
     $  + 1.8471379647759686d-14*tu02
     $  - 1.9809746107821369d-15*tu03
     $  - 1.8472271218294126d-16*tu04

      CsH4YZ(0,3,3,2,15) = 
     $  - 2.1189302329160193d-16
     $  + 1.5035989909597733d-14*tu01
     $  - 2.0916775465757022d-15*tu02
     $  - 2.2584528879375847d-16*tu03

      CsH4YZ(0,3,3,2,16) = 
     $  + 2.4627552798939352d-15
     $  - 1.0799373930163577d-15*tu01
     $  - 2.3711596671511178d-16*tu02

      CsH4YZ(0,3,3,2,17) = 
     $  - 1.8361339146206005d-16*tu01

      CsH4YZ(3,0,0,2,01) = 
     $  + 7.1093645626796962d-01
     $  - 4.0339363662182896d-02*tu01
     $  - 1.2572699446396526d-03*tu02
     $  + 2.0394736785027746d-05*tu03
     $  + 1.8418085312979725d-06*tu04
     $  - 5.1174571404689533d-09*tu05
     $  - 2.1846977429405536d-09*tu06
     $  - 7.6546740137035490d-12*tu07
     $  + 2.3455064424481868d-12*tu08
     $  + 1.6344726153552514d-14*tu09
     $  - 2.3996887548302905d-15*tu10

      CsH4YZ(3,0,0,2,02) = 
     $  - 2.3546515994051698d-01
     $  + 1.1078946227776910d-02*tu01
     $  + 4.4945291388132485d-04*tu02
     $  - 1.5350193624652712d-06*tu03
     $  - 6.4023735656751080d-07*tu04
     $  - 7.7565678527433912d-09*tu05
     $  + 7.3338787151975058d-10*tu06
     $  + 1.6815234741512641d-11*tu07
     $  - 7.5924736316217008d-13*tu08
     $  - 2.4710652539375134d-14*tu09
     $  + 7.5455560382083827d-16*tu10

      CsH4YZ(3,0,0,2,03) = 
     $  + 4.7701934199660361d-02
     $  - 1.5416116407642488d-03*tu01
     $  - 9.1400988554318091d-05*tu02
     $  - 1.3992177049092205d-06*tu03
     $  + 1.0650186255729101d-07*tu04
     $  + 4.6671137507098125d-09*tu05
     $  - 7.9050870070808343d-11*tu06
     $  - 8.0537001982871003d-12*tu07
     $  + 1.7715760944575567d-14*tu08
     $  + 1.1300529521235344d-14*tu09

      CsH4YZ(3,0,0,2,04) = 
     $  - 6.2218346142574179d-03
     $  + 5.8230590650580214d-05*tu01
     $  + 9.2050548904762778d-06*tu02
     $  + 5.0378082933144216d-07*tu03
     $  + 5.9549760941786034d-10*tu04
     $  - 1.0581896194009032d-09*tu05
     $  - 2.6141991325380992d-11*tu06
     $  + 1.4538661309450449d-12*tu07
     $  + 6.6516958692182535d-14*tu08
     $  - 1.5861620902455120d-15*tu09
     $  - 1.1992845913848088d-16*tu10

      CsH4YZ(3,0,0,2,05) = 
     $  + 4.7220875714982151d-04
     $  + 1.5017367388378450d-05*tu01
     $  + 2.5824073757693508d-07*tu02
     $  - 6.1591961914861043d-08*tu03
     $  - 3.8709537518681692d-09*tu04
     $  + 3.7490829205475268d-11*tu05
     $  + 1.0351465565800093d-11*tu06
     $  + 1.3074803819944674d-13*tu07
     $  - 1.8853583534428448d-14*tu08
     $  - 5.0509811509569619d-16*tu09

      CsH4YZ(3,0,0,2,06) = 
     $  - 7.3477183417745331d-06
     $  - 2.0074327937075010d-06*tu01
     $  - 1.9128216710181724d-07*tu02
     $  - 3.5144577599692631d-09*tu03
     $  + 5.9120344066291751d-10*tu04
     $  + 3.3643585788890120d-11*tu05
     $  - 7.4369499851897699d-13*tu06
     $  - 1.0372811256243571d-13*tu07
     $  - 1.4848887064412449d-16*tu08
     $  + 2.2475097357725920d-16*tu09

      CsH4YZ(3,0,0,2,07) = 
     $  - 2.1861647078779019d-06
     $  - 8.8862632409566468d-08*tu01
     $  + 1.3594838878921161d-08*tu02
     $  + 1.8873763379050257d-09*tu03
     $  + 3.5291056747682014d-11*tu04
     $  - 6.1786381230734077d-12*tu05
     $  - 3.0958884154511634d-13*tu06
     $  + 1.1024029122605138d-14*tu07
     $  + 1.0572612322800124d-15*tu08

      CsH4YZ(3,0,0,2,08) = 
     $  + 1.1843327421862613d-07
     $  + 3.5961514695692341d-08*tu01
     $  + 2.3868892081714986d-09*tu02
     $  - 1.0821526982371554d-10*tu03
     $  - 2.0217902276202264d-11*tu04
     $  - 3.3312186437483384d-13*tu05
     $  + 6.9125147490835608d-14*tu06
     $  + 2.9318893217709537d-15*tu07
     $  - 1.5144067368755776d-16*tu08

      CsH4YZ(3,0,0,2,09) = 
     $  + 1.7017140503534106d-08
     $  - 5.7572158862513773d-10*tu01
     $  - 4.2485185599241813d-10*tu02
     $  - 3.2788982192618357d-11*tu03
     $  + 1.0356334314880408d-12*tu04
     $  + 2.2478689163544639d-13*tu05
     $  + 2.9168111925375570d-15*tu06
     $  - 8.0238059195174863d-16*tu07

      CsH4YZ(3,0,0,2,10) = 
     $  - 1.7350837122048892d-09
     $  - 5.2583973196662165d-10*tu01
     $  - 1.5794853332507491d-11*tu02
     $  + 4.8632498683558672d-12*tu03
     $  + 4.2190359455627092d-13*tu04
     $  - 1.1489807301441493d-14*tu05
     $  - 2.5833172031159964d-15*tu06

      CsH4YZ(3,0,0,2,11) = 
     $  - 1.3988318782424805d-10
     $  + 3.2304115134094845d-11*tu01
     $  + 8.6102342810059216d-12*tu02
     $  + 3.6237497493468893d-13*tu03
     $  - 5.7586756234732313d-14*tu04
     $  - 5.2803733508390864d-15*tu05
     $  + 1.4319602844052265d-16*tu06

      CsH4YZ(3,0,0,2,12) = 
     $  + 2.4077443319554375d-11
     $  + 6.7660096272469255d-12*tu01
     $  - 1.7611156918080160d-13*tu02
     $  - 1.2316476068174643d-13*tu03
     $  - 5.9799722494663210d-15*tu04
     $  + 7.0744547415620324d-16*tu05

      CsH4YZ(3,0,0,2,13) = 
     $  + 1.0434950254659825d-12
     $  - 7.5214638666547250d-13*tu01
     $  - 1.4396517226467583d-13*tu02
     $  - 4.1700556847856196d-16*tu03
     $  + 1.6989669875374329d-15*tu04

      CsH4YZ(3,0,0,2,14) = 
     $  - 3.2248148767736684d-13
     $  - 7.7019258923476445d-14*tu01
     $  + 9.5511151913797775d-15*tu02
     $  + 2.4744229828342362d-15*tu03

      CsH4YZ(3,0,0,2,15) = 
     $  - 5.2446944791167849d-15
     $  + 1.4014609403234221d-14*tu01
     $  + 2.1019547517908071d-15*tu02
     $  - 1.1033241603615846d-16*tu03

      CsH4YZ(3,0,0,2,16) = 
     $  + 4.2062946670940523d-15
     $  + 7.3586327288971576d-16*tu01
     $  - 2.4585355248136683d-16*tu02

      CsH4YZ(3,0,0,2,17) = 
     $  - 2.3386032775326175d-16*tu01

      CsH4YZ(3,0,2,2,01) = 
     $  - 6.2819128366611704d-02
     $  - 6.0184860389562120d-02*tu01
     $  + 2.7103937212640883d-03*tu02
     $  + 7.1888956999746477d-05*tu03
     $  - 4.3251589607847741d-06*tu04
     $  - 8.2932220407534956d-08*tu05
     $  + 5.5283764393179977d-09*tu06
     $  + 8.2644873600860110d-11*tu07
     $  - 6.2850708075942488d-12*tu08
     $  - 7.2517644835965336d-14*tu09
     $  + 6.6694567003878169d-15*tu10

      CsH4YZ(3,0,2,2,02) = 
     $  + 7.8233244527470134d-02
     $  + 7.5906790463157458d-03*tu01
     $  - 3.5389469005570845d-04*tu02
     $  - 2.3381784613757111d-05*tu03
     $  + 4.9128167199358886d-07*tu04
     $  + 4.3341832425071969d-08*tu05
     $  - 5.7989799200377789d-10*tu06
     $  - 6.3754477474112524d-11*tu07
     $  + 6.7463471635906614d-13*tu08
     $  + 8.2406095615178222d-14*tu09
     $  - 8.1921162183510358d-16*tu10

      CsH4YZ(3,0,2,2,03) = 
     $  - 2.8728735934302002d-02
     $  + 5.7789559302982347d-04*tu01
     $  + 1.1972568604352366d-05*tu02
     $  + 2.2998268942490993d-06*tu03
     $  + 7.8250176365963941d-08*tu04
     $  - 5.8008793664342075d-09*tu05
     $  - 2.5598952031797979d-10*tu06
     $  + 9.9633413500740585d-12*tu07
     $  + 5.0470308596375571d-13*tu08
     $  - 1.4968068981098132d-14*tu09
     $  - 7.9919461923389109d-16*tu10

      CsH4YZ(3,0,2,2,04) = 
     $  + 5.5504265520633248d-03
     $  - 2.2088124110623792d-04*tu01
     $  - 3.3134870759137642d-06*tu02
     $  + 1.5885582957797815d-07*tu03
     $  - 1.3905602137771104d-08*tu04
     $  - 6.7029853343439834d-10*tu05
     $  + 5.2828353552413789d-11*tu06
     $  + 2.0505930294882922d-12*tu07
     $  - 1.1789558092257759d-13*tu08
     $  - 4.4619129082143792d-15*tu09
     $  + 2.1471705153237348d-16*tu10

      CsH4YZ(3,0,2,2,05) = 
     $  - 5.7428461271582563d-04
     $  + 3.6725226941951958d-06*tu01
     $  + 1.1708139602146884d-06*tu02
     $  - 6.4419137360547681d-09*tu03
     $  - 2.2053528581053943d-09*tu04
     $  + 1.4882138410684056d-10*tu05
     $  + 6.6041796464427573d-12*tu06
     $  - 5.6325587252548688d-13*tu07
     $  - 1.8256220126329695d-14*tu08
     $  + 1.4104735351043094d-15*tu09

      CsH4YZ(3,0,2,2,06) = 
     $  + 1.5967733324253039d-05
     $  + 3.9618373454943580d-06*tu01
     $  - 9.5068445678779649d-10*tu02
     $  - 1.1155812886684237d-08*tu03
     $  + 2.2310770635639879d-10*tu04
     $  + 2.6165370027817628d-11*tu05
     $  - 1.7372495619401155d-12*tu06
     $  - 6.8203993817283920d-14*tu07
     $  + 6.4265107681046942d-15*tu08
     $  + 1.7038746861804995d-16*tu09

      CsH4YZ(3,0,2,2,07) = 
     $  + 3.1005228342825788d-06
     $  - 2.4694429377980986d-07*tu01
     $  - 3.6482277396912737d-08*tu02
     $  + 4.9239848737466563d-10*tu03
     $  + 1.1613528826072699d-10*tu04
     $  - 3.9947103695147922d-12*tu05
     $  - 2.9153728441405420d-13*tu06
     $  + 2.1604782450087560d-14*tu07
     $  + 7.0049951874399118d-16*tu08

      CsH4YZ(3,0,2,2,08) = 
     $  - 2.3367882771577403d-07
     $  - 5.9452811302707324d-08*tu01
     $  + 2.1857841930092563d-09*tu02
     $  + 3.7885996543474834d-10*tu03
     $  - 9.9300566903313586d-12*tu04
     $  - 1.2434714211927864d-12*tu05
     $  + 5.9798358735422033d-14*tu06
     $  + 3.1469677584233985d-15*tu07
     $  - 2.7451510500586571d-16*tu08

      CsH4YZ(3,0,2,2,09) = 
     $  - 2.6140206405288992d-08
     $  + 6.2474445543058767d-09*tu01
     $  + 7.4879178485552917d-10*tu02
     $  - 3.1334530266026783d-11*tu03
     $  - 4.0749453811898781d-12*tu04
     $  + 1.6361538806018511d-13*tu05
     $  + 1.3293023187549219d-14*tu06
     $  - 8.4439735275577008d-16*tu07

      CsH4YZ(3,0,2,2,10) = 
     $  + 3.2790365599431460d-09
     $  + 7.9157753846306200d-10*tu01
     $  - 7.8629655102117505d-11*tu02
     $  - 9.0533957983687454d-12*tu03
     $  + 4.6262040737454978d-13*tu04
     $  + 4.4277158266536743d-14*tu05
     $  - 2.4501755259189053d-15*tu06
     $  - 1.4002614911626366d-16*tu07

      CsH4YZ(3,0,2,2,11) = 
     $  + 2.3362436619377230d-10
     $  - 1.2381068228202411d-10*tu01
     $  - 1.2542395129316453d-11*tu02
     $  + 1.1296076147275948d-12*tu03
     $  + 1.0640397292037736d-13*tu04
     $  - 6.8026202444502432d-15*tu05
     $  - 4.7709125350617014d-16*tu06

      CsH4YZ(3,0,2,2,12) = 
     $  - 4.4651500927259717d-11
     $  - 9.4452277088564889d-12*tu01
     $  + 1.9635710686512026d-12*tu02
     $  + 1.7450544773827628d-13*tu03
     $  - 1.6467761557575044d-14*tu04
     $  - 1.2191012327059663d-15*tu05

      CsH4YZ(3,0,2,2,13) = 
     $  - 2.0050188129875258d-12
     $  + 2.1778342761450722d-12*tu01
     $  + 1.8011653537371513d-13*tu02
     $  - 3.1050363818599038d-14*tu03
     $  - 2.2443777660347379d-15*tu04
     $  + 2.3995211164624782d-16*tu05

      CsH4YZ(3,0,2,2,14) = 
     $  + 5.9477084760673434d-13
     $  + 9.7954681748364900d-14*tu01
     $  - 4.1377549195247926d-14*tu02
     $  - 2.8246132289422942d-15*tu03
     $  + 4.7862996596825261d-16*tu04

      CsH4YZ(3,0,2,2,15) = 
     $  + 1.4888239679375253d-14
     $  - 3.5616118369126795d-14*tu01
     $  - 2.1855330431887940d-15*tu02
     $  + 7.2767460096601433d-16*tu03

      CsH4YZ(3,0,2,2,16) = 
     $  - 7.7845790039705062d-15
     $  - 7.7463591899843494d-16*tu01
     $  + 7.8658402390420829d-16*tu02

      CsH4YZ(3,0,2,2,17) = 
     $  + 5.5362086347952104d-16*tu01

      CsH4YZ(3,0,2,2,18) = 
     $  + 1.0034760528566250d-16

      CsH4YZ(3,0,3,2,01) = 
     $  + 5.9577646704349378d-01
     $  - 1.4718215765863089d-01*tu01
     $  + 6.9491752375823056d-03*tu02
     $  - 9.5973758060067810d-05*tu03
     $  - 3.6598082358988839d-06*tu04
     $  + 1.1662468229302294d-07*tu05
     $  + 2.6056889709015126d-09*tu06
     $  - 1.3138199008291900d-10*tu07
     $  - 1.7818427939516830d-12*tu08
     $  + 1.3916503637593986d-13*tu09
     $  + 1.0385601743639457d-15*tu10
     $  - 1.4078897084739112d-16*tu11

      CsH4YZ(3,0,3,2,02) = 
     $  - 1.9850437702500622d-01
     $  + 4.5721887677448541d-02*tu01
     $  - 1.8593329537244513d-03*tu02
     $  + 9.5574669890711989d-06*tu03
     $  + 1.3923328850281522d-06*tu04
     $  - 1.4283209513364324d-08*tu05
     $  - 1.4744155677394324d-09*tu06
     $  + 2.0140004552702855d-11*tu07
     $  + 1.6041698523234431d-12*tu08
     $  - 2.6899258402457230d-14*tu09
     $  - 1.7145660643639399d-15*tu10

      CsH4YZ(3,0,3,2,03) = 
     $  + 4.0628438814073208d-02
     $  - 8.2790310768906997d-03*tu01
     $  + 2.4926487851378104d-04*tu02
     $  + 2.8726398071972233d-06*tu03
     $  - 2.2153169147267140d-07*tu04
     $  - 4.9561610426968530d-09*tu05
     $  + 2.8713520340984597d-10*tu06
     $  + 7.5635411441581985d-12*tu07
     $  - 3.8929246074195351d-13*tu08
     $  - 1.0391878677015100d-14*tu09
     $  + 5.2295940359481946d-16*tu10

      CsH4YZ(3,0,3,2,04) = 
     $  - 5.4011189266737947d-03
     $  + 8.6194404365325194d-04*tu01
     $  - 9.1218673040072921d-06*tu02
     $  - 8.2824306450806151d-07*tu03
     $  - 4.1147145728872160d-09*tu04
     $  + 1.6177730411517273d-09*tu05
     $  + 1.9525648193845688d-11*tu06
     $  - 2.8862783443706574d-12*tu07
     $  - 4.0595394262976428d-14*tu08
     $  + 4.7439358235675970d-15*tu09

      CsH4YZ(3,0,3,2,05) = 
     $  + 4.2870944530617545d-04
     $  - 3.0786092280142767d-05*tu01
     $  - 1.9653211050153543d-06*tu02
     $  + 2.9138056075438822d-08*tu03
     $  + 6.2495739615341501d-09*tu04
     $  - 1.1385054279520894d-11*tu05
     $  - 1.5235620372682954d-11*tu06
     $  - 3.8858419484620365d-14*tu07
     $  + 3.1445416045030487d-14*tu08
     $  + 1.1928716760742132d-16*tu09

      CsH4YZ(3,0,3,2,06) = 
     $  - 9.5821918641119266d-06
     $  - 3.8849534406609542d-06*tu01
     $  + 1.8564689909063668d-07*tu02
     $  + 1.5601206777948570d-08*tu03
     $  - 3.6063727983881630d-10*tu04
     $  - 5.6531822925692112d-11*tu05
     $  + 6.5272865211090685d-13*tu06
     $  + 1.5324072699493917d-13*tu07
     $  - 1.1307837801189067d-15*tu08
     $  - 3.4603659629986559d-16*tu09

      CsH4YZ(3,0,3,2,07) = 
     $  - 1.7204970823578908d-06
     $  + 3.9120404239495594d-07*tu01
     $  + 2.4739653889021261d-08*tu02
     $  - 1.5601544492256132d-09*tu03
     $  - 1.4791755266296371d-10*tu04
     $  + 4.7942423992196499d-12*tu05
     $  + 5.5481578413273696d-13*tu06
     $  - 1.2229322531031912d-14*tu07
     $  - 1.5914977172144259d-15*tu08

      CsH4YZ(3,0,3,2,08) = 
     $  + 1.1998770537195345d-07
     $  + 3.1357090093274351d-08*tu01
     $  - 3.9615857939299310d-09*tu02
     $  - 2.6102374911120856d-10*tu03
     $  + 1.8525831752490823d-11*tu04
     $  + 1.4741506786908154d-12*tu05
     $  - 6.5426672675618041d-14*tu06
     $  - 5.6121201820091609d-15*tu07
     $  + 1.9015442684082694d-16*tu08

      CsH4YZ(3,0,3,2,09) = 
     $  + 1.1805745969288801d-08
     $  - 5.7279411842675498d-09*tu01
     $  - 2.8907614810882439d-10*tu02
     $  + 4.6011937811175699d-11*tu03
     $  + 2.7964549055168092d-12*tu04
     $  - 2.3406268761310517d-13*tu05
     $  - 1.5008687878375635d-14*tu06
     $  + 8.8215555244693444d-16*tu07

      CsH4YZ(3,0,3,2,10) = 
     $  - 1.5497016954274300d-09
     $  - 2.4614416739894176d-10*tu01
     $  + 7.8173393551779284d-11*tu02
     $  + 3.5220698142697119d-12*tu03
     $  - 5.9743566910102309d-13*tu04
     $  - 2.9528246301479972d-14*tu05
     $  + 3.0334742499361388d-15*tu06
     $  + 1.5255663054663785d-16*tu07

      CsH4YZ(3,0,3,2,11) = 
     $  - 8.2656649192923608d-11
     $  + 8.4857678546893198d-11*tu01
     $  + 2.4774515112435851d-12*tu02
     $  - 1.0927920929649270d-12*tu03
     $  - 3.9997404128383246d-14*tu04
     $  + 7.8970978361526893d-15*tu05
     $  + 3.0378801938342080d-16*tu06

      CsH4YZ(3,0,3,2,12) = 
     $  + 1.9753929146563718d-11
     $  + 1.1262359860423280d-12*tu01
     $  - 1.4214761045390346d-12*tu02
     $  - 3.0610802998042922d-14*tu03
     $  + 1.5558010896594849d-14*tu04
     $  + 4.1782778535863802d-16*tu05
     $  - 1.0492734855218627d-16*tu06

      CsH4YZ(3,0,3,2,13) = 
     $  + 4.4551156639945609d-13
     $  - 1.2307633383992293d-12*tu01
     $  - 3.1533188854216059d-16*tu02
     $  + 2.2775265533051134d-14*tu03
     $  + 2.9909184371262062d-16*tu04
     $  - 2.1869358216128666d-16*tu05

      CsH4YZ(3,0,3,2,14) = 
     $  - 2.4729041781190742d-13
     $  + 1.4965375407249794d-14*tu01
     $  + 2.4126363828671870d-14*tu02
     $  - 1.5007549078209582d-16*tu03
     $  - 3.5349549658115980d-16*tu04

      CsH4YZ(3,0,3,2,15) = 
     $  + 4.7229578357148589d-16
     $  + 1.7342106275207646d-14*tu01
     $  - 6.4776685610321911d-16*tu02
     $  - 4.3220796177501550d-16*tu03

      CsH4YZ(3,0,3,2,16) = 
     $  + 3.0408403296161259d-15
     $  - 6.2158733595848606d-16*tu01
     $  - 3.8627521907229030d-16*tu02

      CsH4YZ(3,0,3,2,17) = 
     $  - 2.3694979799867213d-16*tu01

      CsH4YZ(3,2,2,2,01) = 
     $  + 1.9021773499868327d-02
     $  + 2.4226779820160979d-02*tu01
     $  + 4.5588241049628047d-03*tu02
     $  - 6.2124587503641390d-04*tu03
     $  + 2.4775998106299298d-05*tu04
     $  - 1.7414561635741949d-07*tu05
     $  - 1.3624426123055401d-08*tu06
     $  + 1.8981875994250170d-10*tu07
     $  + 1.0621528829681813d-11*tu08
     $  - 2.0288524371392582d-13*tu09
     $  - 8.5526223878890083d-15*tu10
     $  + 2.0928865540776705d-16*tu11

      CsH4YZ(3,2,2,2,02) = 
     $  - 1.7074333965712641d-02
     $  - 1.6824286398667769d-02*tu01
     $  + 2.9837550570681555d-04*tu02
     $  + 4.7548395216001114d-05*tu03
     $  - 8.6057781437917096d-07*tu04
     $  - 7.9494750994342926d-08*tu05
     $  + 1.6458017364008982d-09*tu06
     $  + 1.0376853476042902d-10*tu07
     $  - 2.5362532200805337d-12*tu08
     $  - 1.1718941160577836d-13*tu09
     $  + 3.4285042703567363d-15*tu10
     $  + 1.1997069555532863d-16*tu11

      CsH4YZ(3,2,2,2,03) = 
     $  + 2.3429354391352657d-02
     $  - 1.2998795504936051d-05*tu01
     $  - 6.4354834343796855d-06*tu02
     $  - 1.7882971913980751d-06*tu03
     $  - 1.9913586683870448d-07*tu04
     $  + 7.4148419337040953d-09*tu05
     $  + 5.3319627517870499d-10*tu06
     $  - 1.7839365067646226d-11*tu07
     $  - 9.0983489518913669d-13*tu08
     $  + 3.2940315947520054d-14*tu09
     $  + 1.2451801545321357d-15*tu10

      CsH4YZ(3,2,2,2,04) = 
     $  - 6.1843681395521004d-03
     $  + 4.0519347983786311d-04*tu01
     $  - 2.6424460563930472d-06*tu02
     $  - 3.8893436273021014d-07*tu03
     $  + 2.0010767585378945d-08*tu04
     $  + 1.6802253764980667d-09*tu05
     $  - 7.9162636892888524d-11*tu06
     $  - 4.5053740847455601d-12*tu07
     $  + 2.0762138961569880d-13*tu08
     $  + 8.5287293208744376d-15*tu09
     $  - 4.2604084836418716d-16*tu10

      CsH4YZ(3,2,2,2,05) = 
     $  + 6.5244311028350012d-04
     $  - 8.9257660183626114d-06*tu01
     $  - 2.1476678162669471d-06*tu02
     $  + 4.6763199756433079d-08*tu03
     $  + 5.0089373538202721d-09*tu04
     $  - 2.5055892934827960d-10*tu05
     $  - 1.6399424510224669d-11*tu06
     $  + 9.2706778588086940d-13*tu07
     $  + 4.1938605396152897d-14*tu08
     $  - 2.5187225451038113d-15*tu09

      CsH4YZ(3,2,2,2,06) = 
     $  - 7.4972493986497465d-06
     $  - 7.3850524867467841d-06*tu01
     $  + 1.3322705287370800d-07*tu02
     $  + 2.0387413935414017d-08*tu03
     $  - 7.0278534836929067d-10*tu04
     $  - 5.7007319033131736d-11*tu05
     $  + 3.2284770970437090d-12*tu06
     $  + 1.6656317177432030d-13*tu07
     $  - 1.1355305191305728d-14*tu08
     $  - 4.0492346719821244d-16*tu09

      CsH4YZ(3,2,2,2,07) = 
     $  - 4.2015767654571566d-06
     $  + 2.8104625215175418d-07*tu01
     $  + 6.6714440098285460d-08*tu02
     $  - 1.9825456624309130d-09*tu03
     $  - 2.1436819456342017d-10*tu04
     $  + 9.9572499633443937d-12*tu05
     $  + 6.2411092623542948d-13*tu06
     $  - 4.1937971965519589d-14*tu07
     $  - 1.6985084187783625d-15*tu08
     $  + 1.4231902507145852d-16*tu09

      CsH4YZ(3,2,2,2,08) = 
     $  + 1.2152971854915309d-07
     $  + 1.1769802476632508d-07*tu01
     $  - 4.5403700262857157d-09*tu02
     $  - 6.8241362555767873d-10*tu03
     $  + 2.8694258031385300d-11*tu04
     $  + 2.3018563497767778d-12*tu05
     $  - 1.3696815906236438d-13*tu06
     $  - 6.6575012017485083d-15*tu07
     $  + 5.4565047816476736d-16*tu08

      CsH4YZ(3,2,2,2,09) = 
     $  + 4.1210168031389808d-08
     $  - 6.2769800266235204d-09*tu01
     $  - 1.4448024083237420d-09*tu02
     $  + 6.9292180451020526d-11*tu03
     $  + 7.2934613837090501d-12*tu04
     $  - 4.0542960188890351d-13*tu05
     $  - 2.4615300914736444d-14*tu06
     $  + 1.8511660607341043d-15*tu07

      CsH4YZ(3,2,2,2,10) = 
     $  - 1.8313929154112520d-09
     $  - 1.7309829940732794d-09*tu01
     $  + 1.1842316865706901d-10*tu02
     $  + 1.7075342075474562d-11*tu03
     $  - 1.0219751912709445d-12*tu04
     $  - 7.8729232300071774d-14*tu05
     $  + 5.6229040287187088d-15*tu06
     $  + 2.5824269001031371d-16*tu07

      CsH4YZ(3,2,2,2,11) = 
     $  - 4.3691633608561884d-10
     $  + 1.2041601766511702d-10*tu01
     $  + 2.6525663448820019d-11*tu02
     $  - 1.9781252808552967d-12*tu03
     $  - 1.9724476325733632d-13*tu04
     $  + 1.4708049347072976d-14*tu05
     $  + 8.4113980759965814d-16*tu06

      CsH4YZ(3,2,2,2,12) = 
     $  + 2.6349336358573161d-11
     $  + 2.4095434760182151d-11*tu01
     $  - 2.6412041043151791d-12*tu02
     $  - 3.5874030373392082d-13*tu03
     $  + 3.0982779999015645d-14*tu04
     $  + 2.2226655126161434d-15*tu05
     $  - 2.0765544109553182d-16*tu06

      CsH4YZ(3,2,2,2,13) = 
     $  + 4.7269990823909425d-12
     $  - 2.1150403226922586d-12*tu01
     $  - 4.3944413543037183d-13*tu02
     $  + 4.8893743661531278d-14*tu03
     $  + 4.5085947932822316d-15*tu04
     $  - 4.6594032092680496d-16*tu05

      CsH4YZ(3,2,2,2,14) = 
     $  - 3.6696483155815800d-13
     $  - 3.2118571829755946d-13*tu01
     $  + 5.3217703579635713d-14*tu02
     $  + 6.6735659547820442d-15*tu03
     $  - 8.2492806058785068d-16*tu04

      CsH4YZ(3,2,2,2,15) = 
     $  - 5.1160254954576831d-14
     $  + 3.5030497338968588d-14*tu01
     $  + 6.7425015405283193d-15*tu02
     $  - 1.0899024879635718d-15*tu03

      CsH4YZ(3,2,2,2,16) = 
     $  + 4.9829043527557753d-15
     $  + 4.1161091495288455d-15*tu01
     $  - 9.9799437005027968d-16*tu02
     $  - 1.1264293967027148d-16*tu03

      CsH4YZ(3,2,2,2,17) = 
     $  + 5.4774852701875656d-16
     $  - 5.5581295198286545d-16*tu01

      CsH4YZ(3,3,0,2,01) = 
     $  + 6.4811732790135792d-01
     $  - 1.0052422405174501d-01*tu01
     $  + 1.4531237766244357d-03*tu02
     $  + 9.2283693784774223d-05*tu03
     $  - 2.4833504294868015d-06*tu04
     $  - 8.8049677548003910d-08*tu05
     $  + 3.3436786963774441d-09*tu06
     $  + 7.4990199587156561d-11*tu07
     $  - 3.9395643651460620d-12*tu08
     $  - 5.6172918682412821d-14*tu09
     $  + 4.2697679455575263d-15*tu10

      CsH4YZ(3,3,0,2,02) = 
     $  - 2.1637341986412859d-01
     $  + 2.9636734239815010d-02*tu01
     $  - 1.2054186923486531d-04*tu02
     $  - 3.6706291563639268d-05*tu03
     $  + 2.8874959895263647d-07*tu04
     $  + 4.8591129548920007d-08*tu05
     $  - 5.3712224859694430d-10*tu06
     $  - 5.9479870049845579d-11*tu07
     $  + 8.5692323110258163d-13*tu08
     $  + 6.7957494848934480d-14*tu09
     $  - 1.2311100902728327d-15*tu10

      CsH4YZ(3,3,0,2,03) = 
     $  + 4.4406216811022673d-02
     $  - 4.8234723227864270d-03*tu01
     $  - 6.9152179015981868d-05*tu02
     $  + 7.0545231822100643d-06*tu03
     $  + 1.5230196765867121d-07*tu04
     $  - 1.1374807392353180d-08*tu05
     $  - 2.5814409374641246d-10*tu06
     $  + 1.7303671644917332d-11*tu07
     $  + 3.7080155044732784d-13*tu08
     $  - 2.4841865395057500d-14*tu09
     $  - 4.7146657738967245d-16*tu10

      CsH4YZ(3,3,0,2,04) = 
     $  - 5.9230112591393493d-03
     $  + 3.7133884554550707d-04*tu01
     $  + 2.2745919852426198d-05*tu02
     $  - 3.0156641792784877d-07*tu03
     $  - 5.9656648440275500d-08*tu04
     $  + 1.5254619580638184d-10*tu05
     $  + 1.2324495985802584d-10*tu06
     $  + 1.0379527699014036d-13*tu07
     $  - 2.2021088544440744d-13*tu08
     $  - 3.3001406825570955d-16*tu09
     $  + 3.5555664126410180d-16*tu10

      CsH4YZ(3,3,0,2,05) = 
     $  + 4.7127742147951146d-04
     $  + 1.1541012427100078d-05*tu01
     $  - 2.3798070725000684d-06*tu02
     $  - 1.4558221306915728d-07*tu03
     $  + 5.6446829870896518d-09*tu04
     $  + 4.9487246295872598d-10*tu05
     $  - 1.1009047457132991d-11*tu06
     $  - 1.2068052583554968d-12*tu07
     $  + 1.9959859212977671d-14*tu08
     $  + 2.4412408593999483d-15*tu09

      CsH4YZ(3,3,0,2,06) = 
     $  - 1.0217219716268878d-05
     $  - 4.8496930685373993d-06*tu01
     $  - 1.3603923913331452d-07*tu02
     $  + 2.5130795381807183d-08*tu03
     $  + 1.1264689869552300d-09*tu04
     $  - 7.8291483676685958d-11*tu05
     $  - 4.2349268704368519d-12*tu06
     $  + 1.9300345602548895d-13*tu07
     $  + 1.1425079669183159d-14*tu08
     $  - 4.1659831147414958d-16*tu09

      CsH4YZ(3,3,0,2,07) = 
     $  - 2.0026764327164173d-06
     $  + 1.3585158795765163d-07*tu01
     $  + 5.4240322811384124d-08*tu02
     $  + 9.8441553337568355d-10*tu03
     $  - 2.9095619737643093d-10*tu04
     $  - 8.6316947853699588d-12*tu05
     $  + 1.0129951886751126d-12*tu06
     $  + 3.5199564678529898d-14*tu07
     $  - 2.7861470906623088d-15*tu08
     $  - 1.0243994474218121d-16*tu09

      CsH4YZ(3,3,0,2,08) = 
     $  + 1.4418921596747882d-07
     $  + 5.8415392395923433d-08*tu01
     $  - 1.4758513906130049d-09*tu02
     $  - 6.5713848330172422d-10*tu03
     $  - 4.6532679338078519d-12*tu04
     $  + 3.4825648025454701d-12*tu05
     $  + 5.7971315036848118d-14*tu06
     $  - 1.2770574458350757d-14*tu07
     $  - 2.6350117595874385d-16*tu08

      CsH4YZ(3,3,0,2,09) = 
     $  + 1.3339713554393601d-08
     $  - 4.7646110618964350d-09*tu01
     $  - 8.8518772898822121d-10*tu02
     $  + 2.5625359685878351d-11*tu03
     $  + 8.1128727958699530d-12*tu04
     $  - 3.1091987895648649d-14*tu05
     $  - 4.2108771561081399d-14*tu06
     $  - 2.1401795392259881d-16*tu07
     $  + 1.5854073563361847d-16*tu08

      CsH4YZ(3,3,0,2,10) = 
     $  - 1.9141757718398162d-09
     $  - 6.2116302261658066d-10*tu01
     $  + 7.6423185075987377d-11*tu02
     $  + 1.2529461549154467d-11*tu03
     $  - 4.5367324895955142d-13*tu04
     $  - 1.0050870669795685d-13*tu05
     $  + 1.4589685584715816d-15*tu06
     $  + 5.0969262357873436d-16*tu07

      CsH4YZ(3,3,0,2,11) = 
     $  - 8.1760337250071743d-11
     $  + 9.5102634612461540d-11*tu01
     $  + 1.1789161504975178d-11*tu02
     $  - 1.2584242254407021d-12*tu03
     $  - 1.6980242281307176d-13*tu04
     $  + 7.7123578075619239d-15*tu05
     $  + 1.2387665865120622d-15*tu06

      CsH4YZ(3,3,0,2,12) = 
     $  + 2.4282263071605379d-11
     $  + 5.3501765811275493d-12*tu01
     $  - 1.8894912391908008d-12*tu02
     $  - 1.9119138704437772d-13*tu03
     $  + 2.0417434746108752d-14*tu04
     $  + 2.2276786288706110d-15*tu05
     $  - 1.2539984302901431d-16*tu06

      CsH4YZ(3,3,0,2,13) = 
     $  + 2.1895852710368700d-13
     $  - 1.5893722730301736d-12*tu01
     $  - 1.2334290529834842d-13*tu02
     $  + 3.4406136157585842d-14*tu03
     $  + 2.8207310922344527d-15*tu04
     $  - 3.2453022327396200d-16*tu05

      CsH4YZ(3,3,0,2,14) = 
     $  - 2.9671777000292353d-13
     $  - 2.3814593428319857d-14*tu01
     $  + 3.7358198200833217d-14*tu02
     $  + 2.2362001595070493d-15*tu03
     $  - 5.9052652153861873d-16*tu04

      CsH4YZ(3,3,0,2,15) = 
     $  + 5.6115269155835133d-15
     $  + 2.4112080816367005d-14*tu01
     $  + 7.0273719938637112d-16*tu02
     $  - 7.5677384069943631d-16*tu03

      CsH4YZ(3,3,0,2,16) = 
     $  + 3.5004964232986819d-15
     $  - 3.9444062193603430d-16*tu01
     $  - 6.5359552527637636d-16*tu02

      CsH4YZ(3,3,0,2,17) = 
     $  - 1.5969616101324908d-16
     $  - 3.4173547414856547d-16*tu01

      CsH4YZ(3,3,2,2,01) = 
     $  - 1.3993557059135089d-01
     $  - 1.1857409474568816d-01*tu01
     $  + 2.0034487113411656d-02*tu02
     $  - 1.2869712879181768d-03*tu03
     $  + 3.9725220484616966d-05*tu04
     $  - 3.0948131368888205d-07*tu05
     $  - 1.6930089092950739d-08*tu06
     $  + 3.3855507741664806d-10*tu07
     $  + 1.0830637615369507d-11*tu08
     $  - 3.5546782204142963d-13*tu09
     $  - 6.9975391461959633d-15*tu10
     $  + 3.5476653696753782d-16*tu11

      CsH4YZ(3,3,2,2,02) = 
     $  + 9.2455398890138215d-02
     $  + 1.8137493066875074d-02*tu01
     $  - 3.8083639305456960d-03*tu02
     $  + 1.9513530597119782d-04*tu03
     $  - 2.0606045925950606d-06*tu04
     $  - 1.5058751972859595d-07*tu05
     $  + 3.5757194556058639d-09*tu06
     $  + 1.4035087106058887d-10*tu07
     $  - 5.1680875075244978d-12*tu08
     $  - 1.1964070254567397d-13*tu09
     $  + 6.5759857645332280d-15*tu10

      CsH4YZ(3,3,2,2,03) = 
     $  - 2.9308690514359785d-02
     $  + 2.8302646863060321d-04*tu01
     $  + 2.9003376575328800d-04*tu02
     $  - 5.4800545003291459d-06*tu03
     $  - 6.5357818361896457d-07*tu04
     $  + 1.9302453057701636d-08*tu05
     $  + 1.0026209852045106d-09*tu06
     $  - 4.2014560384522831d-11*tu07
     $  - 1.1920172909396454d-12*tu08
     $  + 7.1308723924830029d-14*tu09
     $  + 1.1260107844133891d-15*tu10
     $  - 1.0374874123427053d-16*tu11

      CsH4YZ(3,3,2,2,04) = 
     $  + 5.3663525943315015d-03
     $  - 4.0015618765640272d-04*tu01
     $  + 5.3762836368168117d-08*tu02
     $  - 1.2140194822430819d-06*tu03
     $  + 5.0068507626717941d-08*tu04
     $  + 4.1728221683042101d-09*tu05
     $  - 2.0243827410996411d-10*tu06
     $  - 7.5873190596932125d-12*tu07
     $  + 4.9460546154646995d-13*tu08
     $  + 9.6995650666290640d-15*tu09
     $  - 9.3481454360174225d-16*tu10

      CsH4YZ(3,3,2,2,05) = 
     $  - 5.5088765593499147d-04
     $  + 2.6334358299304982d-05*tu01
     $  + 4.4196835876092795d-07*tu02
     $  + 1.2005707755229074d-08*tu03
     $  + 9.3261932435060874d-09*tu04
     $  - 5.3256823965599972d-10*tu05
     $  - 3.1580007069757967d-11*tu06
     $  + 2.2250305870595778d-12*tu07
     $  + 5.8597283847491211d-14*tu08
     $  - 5.8293599292311295d-15*tu09

      CsH4YZ(3,3,2,2,06) = 
     $  + 1.7339084548326751d-05
     $  + 5.0465061421024402d-06*tu01
     $  - 2.7079449115980810d-07*tu02
     $  + 5.2600014935514925d-09*tu03
     $  - 3.1596477907955581d-10*tu04
     $  - 8.2710609255670051d-11*tu05
     $  + 6.0044693054057364d-12*tu06
     $  + 2.5254787510240480d-13*tu07
     $  - 2.5098825085086818d-14*tu08
     $  - 4.2670964736741109d-16*tu09

      CsH4YZ(3,3,2,2,07) = 
     $  + 2.6960314477126668d-06
     $  - 6.4178822884345723d-07*tu01
     $  - 2.4495560868321642d-08*tu02
     $  + 2.5952213867763513d-09*tu03
     $  - 1.1183911516874604d-10*tu04
     $  + 5.6229400878010214d-12*tu05
     $  + 7.7015808005889060d-13*tu06
     $  - 6.9694137978557590d-14*tu07
     $  - 2.0032252940595626d-15*tu08
     $  + 2.8780813655704669d-16*tu09

      CsH4YZ(3,3,2,2,08) = 
     $  - 2.3894127149299132d-07
     $  - 5.7748109822769385d-08*tu01
     $  + 8.8151102218368378d-09*tu02
     $  + 2.4849303964149942d-11*tu03
     $  - 2.3153142846619774d-11*tu04
     $  + 1.5897098345176727d-12*tu05
     $  - 8.6827172862721304d-14*tu06
     $  - 7.2414611227452936d-15*tu07
     $  + 8.2273363846423087d-16*tu08

      CsH4YZ(3,3,2,2,09) = 
     $  - 1.9991493427998465d-08
     $  + 1.2092035572713181d-08*tu01
     $  + 3.6336115322551950d-10*tu02
     $  - 1.0669656246844348d-10*tu03
     $  + 1.9286673598534391d-12*tu04
     $  + 1.8521018564234738d-13*tu05
     $  - 2.0101101053815857d-14*tu06
     $  + 1.2516440588446175d-15*tu07

      CsH4YZ(3,3,2,2,10) = 
     $  + 3.1753579354081870d-09
     $  + 5.5257681713354550d-10*tu01
     $  - 2.0520708372843731d-10*tu02
     $  + 5.1307060289238800d-13*tu03
     $  + 1.1969404438700754d-12*tu04
     $  - 4.2316150205123371d-14*tu05
     $  - 1.1575478479197413d-15*tu06
     $  + 2.3987575953424636d-16*tu07

      CsH4YZ(3,3,2,2,11) = 
     $  + 1.4640193127670416d-10
     $  - 2.0271240119577096d-10*tu01
     $  - 2.2604549399714009d-12*tu02
     $  + 2.9255495754215429d-12*tu03
     $  - 6.5875985918192970d-14*tu04
     $  - 1.2619255934126323d-14*tu05
     $  + 6.7511497429636593d-16*tu06

      CsH4YZ(3,3,2,2,12) = 
     $  - 4.0949441022619740d-11
     $  - 3.4086183853817749d-12*tu01
     $  + 4.0750231632951646d-12*tu02
     $  - 7.2422722329790665d-14*tu03
     $  - 3.7647314360655885d-14*tu04
     $  + 1.4735467490649998d-15*tu05
     $  + 1.2419124213684732d-16*tu06

      CsH4YZ(3,3,2,2,13) = 
     $  - 8.2936881592656512d-13
     $  + 3.1616745124552952d-12*tu01
     $  - 5.1422223295949577d-14*tu02
     $  - 6.6340566844320884d-14*tu03
     $  + 2.4427413060345184d-15*tu04
     $  + 4.4943436922931767d-16*tu05

      CsH4YZ(3,3,2,2,14) = 
     $  + 5.1505004499944484d-13
     $  - 1.8886049194531276d-14*tu01
     $  - 7.3256493824224128d-14*tu02
     $  + 2.9649382000316566d-15*tu03
     $  + 9.5112317857815563d-16*tu04

      CsH4YZ(3,3,2,2,15) = 
     $  - 2.5137900907345167d-16
     $  - 4.6820243135582647d-14*tu01
     $  + 2.4670335784318610d-15*tu02
     $  + 1.3341441108474398d-15*tu03

      CsH4YZ(3,3,2,2,16) = 
     $  - 6.3344186195929565d-15
     $  + 1.2278659107528031d-15*tu01
     $  + 1.2253850382499094d-15*tu02

      CsH4YZ(3,3,2,2,17) = 
     $  + 1.1958403742003941d-16
     $  + 6.6513389128993921d-16*tu01

      CsH4YZ(3,3,3,2,01) = 
     $  + 5.5197911217675041d-01
     $  - 1.8314023822803204d-01*tu01
     $  + 1.4218393063809198d-02*tu02
     $  - 6.4533067609673523d-04*tu03
     $  + 1.6791030909615640d-05*tu04
     $  - 1.4045315447193150d-07*tu05
     $  - 5.4903607128358914d-09*tu06
     $  + 1.4108164346044281d-10*tu07
     $  + 2.5546152281358811d-12*tu08
     $  - 1.3623785217395129d-13*tu09
     $  - 8.4460551313724565d-16*tu10
     $  + 1.2447235795110355d-16*tu11

      CsH4YZ(3,3,3,2,02) = 
     $  - 1.8507693153574787d-01
     $  + 5.7007834659042109d-02*tu01
     $  - 3.8733866154316684d-03*tu02
     $  + 1.3426240380531455d-04*tu03
     $  - 1.4025588512568421d-06*tu04
     $  - 6.5843471610404989d-08*tu05
     $  + 1.9726934626117993d-09*tu06
     $  + 4.0856943724426801d-11*tu07
     $  - 2.4495457727004487d-12*tu08
     $  - 1.6899891706139356d-14*tu09
     $  + 2.7355830257387624d-15*tu10

      CsH4YZ(3,3,3,2,03) = 
     $  + 3.8273528625964281d-02
     $  - 1.0336495532902176d-02*tu01
     $  + 5.3661103987241802d-04*tu02
     $  - 7.4645584177276319d-06*tu03
     $  - 4.3857534884014958d-07*tu04
     $  + 1.5755455642220782d-08*tu05
     $  + 3.8112856257330508d-10*tu06
     $  - 2.6092058657288038d-11*tu07
     $  - 2.0291217780904545d-13*tu08
     $  + 3.6429590479147518d-14*tu09
     $  - 1.1347459120338164d-16*tu10

      CsH4YZ(3,3,3,2,04) = 
     $  - 5.1794053065970926d-03
     $  + 1.0714700634964255d-03*tu01
     $  - 2.2315080292008516d-05*tu02
     $  - 1.7520162484105583d-06*tu03
     $  + 7.8594961174267638d-08*tu04
     $  + 2.2851469499331641d-09*tu05
     $  - 1.8231703692359437d-10*tu06
     $  - 1.6244255687611379d-12*tu07
     $  + 3.2737363861265242d-13*tu08
     $  - 1.1281677007665640d-15*tu09
     $  - 4.9268613802146752d-16*tu10

      CsH4YZ(3,3,3,2,05) = 
     $  + 4.2649017164195329d-04
     $  - 3.5453497621977176d-05*tu01
     $  - 4.1957075132338883d-06*tu02
     $  + 2.5063084523644415d-07*tu03
     $  + 9.1314829514438200d-09*tu04
     $  - 8.7303052121974670d-10*tu05
     $  - 9.1048482949356672d-12*tu06
     $  + 2.0912560091692445d-12*tu07
     $  - 8.0651125839626730d-15*tu08
     $  - 3.9352794472719367d-15*tu09

      CsH4YZ(3,3,3,2,06) = 
     $  - 1.1568651878477111d-05
     $  - 5.5699625251318077d-06*tu01
     $  + 4.9836132436390967d-07*tu02
     $  + 2.4314159180071248d-08*tu03
     $  - 2.9003661089456306d-09*tu04
     $  - 3.6462023747245240d-11*tu05
     $  + 9.7356284794277294d-12*tu06
     $  - 4.2630549082870269d-14*tu07
     $  - 2.3566218516431789d-14*tu08
     $  + 3.8339104988501835d-16*tu09

      CsH4YZ(3,3,3,2,07) = 
     $  - 1.5706293293392023d-06
     $  + 5.6295966284965374d-07*tu01
     $  + 4.1577069968343070d-08*tu02
     $  - 6.5961364233850961d-09*tu03
     $  - 1.0434576892099268d-10*tu04
     $  + 3.3271825634386015d-11*tu05
     $  - 1.6855821384394062d-13*tu06
     $  - 1.0747200884440647d-13*tu07
     $  + 1.9639828136614125d-15*tu08
     $  + 2.5927598206765254d-16*tu09

      CsH4YZ(3,3,3,2,08) = 
     $  + 1.3583421568290180d-07
     $  + 4.1367880573222437d-08*tu01
     $  - 9.8114000590262289d-09*tu02
     $  - 2.0918939512577813d-10*tu03
     $  + 8.2804576045305061d-11*tu04
     $  - 4.9785728965854305d-13*tu05
     $  - 3.7498804545412263d-13*tu06
     $  + 7.8173484641082993d-15*tu07
     $  + 1.1639833705684513d-15*tu08

      CsH4YZ(3,3,3,2,09) = 
     $  + 9.0528616615527175d-09
     $  - 8.5750322034689759d-09*tu01
     $  - 2.8000137610455115d-10*tu02
     $  + 1.4621229344320304d-10*tu03
     $  - 1.0821826035427623d-12*tu04
     $  - 9.9584174847835663d-13*tu05
     $  + 2.4166929008927847d-14*tu06
     $  + 4.1263727163882429d-15*tu07
     $  - 1.5371066734236296d-16*tu08

      CsH4YZ(3,3,3,2,10) = 
     $  - 1.6282691514170228d-09
     $  - 2.2567508180444039d-10*tu01
     $  + 1.7347457855393724d-10*tu02
     $  - 1.6739809201166001d-12*tu03
     $  - 1.9801735770954904d-12*tu04
     $  + 5.7511246330298329d-14*tu05
     $  + 1.1509920497479335d-14*tu06
     $  - 4.8938283867939404d-16*tu07

      CsH4YZ(3,3,3,2,11) = 
     $  - 4.1893249398494547d-11
     $  + 1.2330806260478792d-10*tu01
     $  - 1.7228326862876767d-12*tu02
     $  - 2.8552672526282572d-12*tu03
     $  + 1.0332831746725276d-13*tu04
     $  + 2.4985223138712779d-14*tu05
     $  - 1.2375849582460195d-15*tu06
     $  - 1.2733067878686433d-16*tu07

      CsH4YZ(3,3,3,2,12) = 
     $  + 1.9144383273233960d-11
     $  - 1.0132376820397026d-12*tu01
     $  - 2.8139209885019550d-12*tu02
     $  + 1.3531744208152688d-13*tu03
     $  + 4.1346264042518176d-14*tu04
     $  - 2.4536479552428371d-15*tu05
     $  - 2.9577458774138841d-16*tu06

      CsH4YZ(3,3,3,2,13) = 
     $  - 9.5292929661910951d-14
     $  - 1.6812987591731615d-12*tu01
     $  + 1.2117958130465824d-13*tu02
     $  + 5.0344926060340105d-14*tu03
     $  - 3.7288267687813026d-15*tu04
     $  - 5.4278352870810025d-16*tu05

      CsH4YZ(3,3,3,2,14) = 
     $  - 2.1899359710955729d-13
     $  + 6.5062074647674696d-14*tu01
     $  + 4.2383879833803225d-14*tu02
     $  - 4.1834003775705941d-15*tu03
     $  - 7.6891392509183742d-16*tu04

      CsH4YZ(3,3,3,2,15) = 
     $  + 7.0529486385526201d-15
     $  + 2.1794877565699445d-14*tu01
     $  - 3.2439892957115371d-15*tu02
     $  - 8.0985834584231849d-16*tu03
     $  + 1.0273100613113615d-16*tu04

      CsH4YZ(3,3,3,2,16) = 
     $  + 2.4270503135916956d-15
     $  - 1.5213752373049481d-15*tu01
     $  - 5.9461876449064131d-16*tu02
     $  + 1.0061941011747911d-16*tu03

      CsH4YZ(3,3,3,2,17) = 
     $  - 1.4469531820505842d-16
     $  - 2.6796623767340911d-16*tu01


      endif

      return
      end

      subroutine fillcoeff2dhplaux(iflag,n,CbH1YZ,CbH2YZ,CbH3YZ,
     $                             CbH4YZ)
      implicit none
      include 'src/Inc/types.f'
      real(dp):: CbH1YZ,CbH2YZ,CbH3YZ,CbH4YZ
      real(dp):: 
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      integer iflag,n
      dimension CbH1YZ(3:3,22),CbH2YZ(0:3,0:3,22),
     $          CbH3YZ(0:3,0:3,0:3,22),CbH4YZ(0:3,0:3,0:3,0:3,22)
      common /zm/  
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
!$omp threadprivate(/zm/)

* 2001-04-20:13:11:57.chpl 
* <- HPLexpand.320aux 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.2) then
      CbH1YZ(3,01) = 
     $  + 7.500000000000000d-01*zm01

      CbH1YZ(3,02) = 
     $  - 2.812500000000000d-01*zm01

      CbH1YZ(3,03) = 
     $  + 7.031250000000000d-02*zm01
     $  + 7.031250000000000d-02*zm02

      CbH1YZ(3,04) = 
     $  - 1.318359375000000d-02*zm01
     $  - 5.273437500000000d-02*zm02
     $  - 1.318359375000000d-02*zm03

      CbH1YZ(3,05) = 
     $  + 1.977539062500000d-03*zm01
     $  + 2.175292968750000d-02*zm02
     $  + 2.175292968750000d-02*zm03
     $  + 1.977539062500000d-03*zm04

      CbH1YZ(3,06) = 
     $  - 2.471923828125000d-04*zm01
     $  - 6.427001953125000d-03*zm02
     $  - 1.631469726562500d-02*zm03
     $  - 6.427001953125000d-03*zm04
     $  - 2.471923828125000d-04*zm05

      CbH1YZ(3,07) = 
     $  + 2.648489815848214d-05*zm01
     $  + 1.509639195033482d-03*zm02
     $  + 7.998439243861607d-03*zm03
     $  + 7.998439243861607d-03*zm04
     $  + 1.509639195033482d-03*zm05
     $  + 2.648489815848214d-05*zm06

      CbH1YZ(3,08) = 
     $  - 2.482959202357700d-06*zm01
     $  - 2.979551042829241d-04*zm02
     $  - 2.957204410008021d-03*zm03
     $  - 5.998829432896205d-03*zm04
     $  - 2.957204410008021d-03*zm05
     $  - 2.979551042829241d-04*zm06
     $  - 2.482959202357700d-06*zm07

      CbH1YZ(3,09) = 
     $  + 2.069132668631417d-07*zm01
     $  + 5.110757691519601d-05*zm02
     $  + 8.882786546434674d-04*zm03
     $  + 3.231778315135410d-03*zm04
     $  + 3.231778315135410d-03*zm05
     $  + 8.882786546434674d-04*zm06
     $  + 5.110757691519601d-05*zm07
     $  + 2.069132668631417d-07*zm08

      CbH1YZ(3,10) = 
     $  - 1.551849501473563d-08*zm01
     $  - 7.790284497397286d-06*zm02
     $  - 2.266941751752580d-04*zm03
     $  - 1.369258889130183d-03*zm04
     $  - 2.423833736351558d-03*zm05
     $  - 1.369258889130183d-03*zm06
     $  - 2.266941751752580d-04*zm07
     $  - 7.790284497397286d-06*zm08
     $  - 1.551849501473563d-08*zm09

      CbH1YZ(3,11) = 
     $  + 1.058079205550156d-09*zm01
     $  + 1.071834235222308d-06*zm02
     $  + 5.061850919351949d-05*zm03
     $  + 4.816291897327868d-04*zm04
     $  + 1.386458319309469d-03*zm05
     $  + 1.386458319309469d-03*zm06
     $  + 4.816291897327868d-04*zm07
     $  + 5.061850919351949d-05*zm08
     $  + 1.071834235222308d-06*zm09
     $  + 1.058079205550156d-09*zm10

      CbH1YZ(3,12) = 
     $  - 6.612995034688478d-11*zm01
     $  - 1.346405789062574d-07*zm02
     $  - 1.009387723109745d-05*zm03
     $  - 1.457165520299564d-04*zm04
     $  - 6.439809952923036d-04*zm05
     $  - 1.039843739482102d-03*zm06
     $  - 6.439809952923036d-04*zm07
     $  - 1.457165520299564d-04*zm08
     $  - 1.009387723109745d-05*zm09
     $  - 1.346405789062574d-07*zm10
     $  - 6.612995034688478d-11*zm11

      CbH1YZ(3,13) = 
     $  + 3.815189443089507d-12*zm01
     $  + 1.557741849613445d-08*zm02
     $  + 1.824694470135861d-06*zm03
     $  + 3.886794826152132d-05*zm04
     $  + 2.530175418866059d-04*zm05
     $  + 6.200151579195426d-04*zm06
     $  + 6.200151579195426d-04*zm07
     $  + 2.530175418866059d-04*zm08
     $  + 3.886794826152132d-05*zm09
     $  + 1.824694470135861d-06*zm10
     $  + 1.557741849613445d-08*zm11
     $  + 3.815189443089507d-12*zm12

      CbH1YZ(3,14) = 
     $  - 2.043851487369378d-13*zm01
     $  - 1.671461746370677d-09*zm02
     $  - 3.024340185999141d-07*zm03
     $  - 9.306360950755924d-06*zm04
     $  - 8.651245948857437d-05*zm05
     $  - 3.077266758541127d-04*zm06
     $  - 4.650113684396570d-04*zm07
     $  - 3.077266758541127d-04*zm08
     $  - 8.651245948857437d-05*zm09
     $  - 9.306360950755924d-06*zm10
     $  - 3.024340185999141d-07*zm11
     $  - 1.671461746370677d-09*zm12
     $  - 2.043851487369378d-13*zm13

      CbH1YZ(3,15) = 
     $  + 1.021925743684689d-14*zm01
     $  + 1.672790249837468d-10*zm02
     $  + 4.636797983780952d-08*zm03
     $  + 2.027610900381137d-06*zm04
     $  + 2.628129534752155d-05*zm05
     $  + 1.312486095260922d-04*zm06
     $  + 2.858446492955250d-04*zm07
     $  + 2.858446492955250d-04*zm08
     $  + 1.312486095260922d-04*zm09
     $  + 2.628129534752155d-05*zm10
     $  + 2.027610900381137d-06*zm11
     $  + 4.636797983780952d-08*zm12
     $  + 1.672790249837468d-10*zm13
     $  + 1.021925743684689d-14*zm14

      CbH1YZ(3,16) = 
     $  - 4.790276923521981d-16*zm01
     $  - 1.568911497991919d-11*zm02
     $  - 6.622432820541435d-09*zm03
     $  - 4.062590324802311d-07*zm04
     $  - 7.205165467584388d-06*zm05
     $  - 4.923302862336418d-05*zm06
     $  - 1.491632826939143d-04*zm07
     $  - 2.143834869716437d-04*zm08
     $  - 1.491632826939143d-04*zm09
     $  - 4.923302862336418d-05*zm10
     $  - 7.205165467584388d-06*zm11
     $  - 4.062590324802311d-07*zm12
     $  - 6.622432820541435d-09*zm13
     $  - 1.568911497991919d-11*zm14
     $  - 4.790276923521981d-16*zm15

      CbH1YZ(3,17) = 
     $  + 2.113357466259697d-17*zm01
     $  + 1.384650678318691d-12*zm02
     $  + 8.861887972651400d-10*zm03
     $  + 7.549093043770426d-08*zm04
     $  + 1.804453046809619d-06*zm05
     $  + 1.652889670074764d-05*zm06
     $  + 6.778558522460480d-05*zm07
     $  + 1.348913576478696d-04*zm08
     $  + 1.348913576478696d-04*zm09
     $  + 6.778558522460480d-05*zm10
     $  + 1.652889670074764d-05*zm11
     $  + 1.804453046809619d-06*zm12
     $  + 7.549093043770426d-08*zm13
     $  + 8.861887972651400d-10*zm14
     $  + 1.384650678318691d-12*zm15
     $  + 2.113357466259697d-17*zm16

      CbH1YZ(3,18) = 
     $  - 8.805656109415407d-19*zm01
     $  - 1.154016455763326d-13*zm02
     $  - 1.116390063320916d-10*zm03
     $  - 1.309876520468870d-08*zm04
     $  - 4.168186387390939d-07*zm05
     $  - 5.034450698591722d-06*zm06
     $  - 2.734654001168573d-05*zm07
     $  - 7.320777972620855d-05*zm08
     $  - 1.011685182359022d-04*zm09
     $  - 7.320777972620855d-05*zm10
     $  - 2.734654001168573d-05*zm11
     $  - 5.034450698591722d-06*zm12
     $  - 4.168186387390939d-07*zm13
     $  - 1.309876520468870d-08*zm14
     $  - 1.116390063320916d-10*zm15
     $  - 1.154016455763326d-13*zm16
     $  - 8.805656109415407d-19*zm17

      CbH1YZ(3,19) = 
     $  + 3.475916885295555d-20*zm01
     $  + 9.111247135580974d-15*zm02
     $  + 1.329329389442748d-11*zm03
     $  + 2.134328128173797d-09*zm04
     $  + 8.950562789057019d-08*zm05
     $  + 1.406263677440469d-06*zm06
     $  + 9.941020597298717d-06*zm07
     $  + 3.499240176071887d-05*zm08
     $  + 6.483925505467917d-05*zm09
     $  + 6.483925505467917d-05*zm10
     $  + 3.499240176071887d-05*zm11
     $  + 9.941020597298717d-06*zm12
     $  + 1.406263677440469d-06*zm13
     $  + 8.950562789057019d-08*zm14
     $  + 2.134328128173797d-09*zm15
     $  + 1.329329389442748d-11*zm16
     $  + 9.111247135580974d-15*zm17
     $  + 3.475916885295555d-20*zm18

      CbH1YZ(3,20) = 
     $  - 1.303468831985833d-21*zm01
     $  - 6.833669976075488d-16*zm02
     $  - 1.501303983172025d-12*zm03
     $  - 3.281251955627261d-10*zm04
     $  - 1.798286480157967d-08*zm05
     $  - 3.633997820666549d-07*zm06
     $  - 3.295071449543142d-06*zm07
     $  - 1.497117979700008d-05*zm08
     $  - 3.631761430725076d-05*zm09
     $  - 4.862944129100938d-05*zm10
     $  - 3.631761430725076d-05*zm11
     $  - 1.497117979700008d-05*zm12
     $  - 3.295071449543142d-06*zm13
     $  - 3.633997820666549d-07*zm14
     $  - 1.798286480157967d-08*zm15
     $  - 3.281251955627261d-10*zm16
     $  - 1.501303983172025d-12*zm17
     $  - 6.833669976075488d-16*zm18
     $  - 1.303468831985833d-21*zm19

      CbH1YZ(3,21) = 
     $  + 4.655245828520833d-23*zm01
     $  + 4.881281289724662d-17*zm02
     $  + 1.612933055526075d-13*zm03
     $  + 4.778653392731532d-11*zm04
     $  + 3.398725969175070d-09*zm05
     $  + 8.750505944370088d-08*zm06
     $  + 1.005467753419113d-06*zm07
     $  + 5.807334543573626d-06*zm08
     $  + 1.808973879747349d-05*zm09
     $  + 3.163529179606615d-05*zm10
     $  + 3.163529179606615d-05*zm11
     $  + 1.808973879747349d-05*zm12
     $  + 5.807334543573626d-06*zm13
     $  + 1.005467753419113d-06*zm14
     $  + 8.750505944370088d-08*zm15
     $  + 3.398725969175070d-09*zm16
     $  + 4.778653392731532d-11*zm17
     $  + 1.612933055526075d-13*zm18
     $  + 4.881281289724662d-17*zm19
     $  + 4.655245828520833d-23*zm20

      CbH1YZ(3,22) = 
     $  - 1.5870156233593749d-24*zm01
     $  - 3.3281780742156460d-18*zm02
     $  - 1.6527523639870582d-14*zm03
     $  - 6.6153209730411898d-12*zm04
     $  - 6.0702275872635392d-10*zm05
     $  - 1.9752612687579764d-08*zm06
     $  - 2.8468807382690809d-07*zm07
     $  - 2.0637008487428384d-06*zm08
     $  - 8.1239658492176797d-06*zm09
     $  - 1.8185106256716254d-05*zm10
     $  - 2.3726468847049612d-05*zm11
     $  - 1.8185106256716254d-05*zm12
     $  - 8.1239658492176797d-06*zm13
     $  - 2.0637008487428384d-06*zm14
     $  - 2.8468807382690809d-07*zm15
     $  - 1.9752612687579764d-08*zm16
     $  - 6.0702275872635392d-10*zm17
     $  - 6.6153209730411898d-12*zm18
     $  - 1.6527523639870582d-14*zm19
     $  - 3.3281780742156460d-18*zm20
     $  - 1.5870156233593749d-24*zm21
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      CbH2YZ(0,3,01) = 
     $  + 7.500000000000000d-01*zm01

      CbH2YZ(0,3,02) = 
     $  - 2.812500000000000d-01*zm01
     $  + 1.406250000000000d-01*zm02

      CbH2YZ(0,3,03) = 
     $  + 7.031250000000000d-02*zm01
     $  - 3.515625000000000d-02*zm02
     $  + 1.171875000000000d-02*zm03

      CbH2YZ(0,3,04) = 
     $  - 1.318359375000000d-02*zm01
     $  - 6.591796875000000d-03*zm02

      CbH2YZ(0,3,05) = 
     $  + 1.977539062500000d-03*zm01
     $  + 6.921386718750000d-03*zm02
     $  + 3.295898437500000d-04*zm03
     $  + 3.295898437500000d-04*zm04
     $  - 6.591796875000000d-05*zm05

      CbH2YZ(0,3,06) = 
     $  - 2.471923828125000d-04*zm01
     $  - 2.595520019531250d-03*zm02
     $  - 1.977539062500000d-03*zm03
     $  - 1.235961914062500d-04*zm04

      CbH2YZ(0,3,07) = 
     $  + 2.648489815848214d-05*zm01
     $  + 6.753649030412946d-04*zm02
     $  + 1.540538242885044d-03*zm03
     $  + 4.590715680803571d-04*zm04
     $  + 2.648489815848214d-05*zm05
     $  - 4.414149693080357d-06*zm06
     $  + 6.305928132971938d-07*zm07

      CbH2YZ(0,3,08) = 
     $  - 2.482959202357700d-06*zm01
     $  - 1.402871949332101d-04*zm02
     $  - 7.051604134695870d-04*zm03
     $  - 6.182568413870675d-04*zm04
     $  - 9.683540889195033d-05*zm05
     $  - 1.241479601178850d-06*zm06

      CbH2YZ(0,3,09) = 
     $  + 2.069132668631417d-07*zm01
     $  + 2.472613539014543d-05*zm02
     $  + 2.383985689708164d-04*zm03
     $  + 4.503467253276279d-04*zm04
     $  + 1.960089376994541d-04*zm05
     $  + 1.737381730760846d-05*zm06
     $  - 1.448392868041992d-07*zm07
     $  + 6.207398005894252d-08*zm08
     $  - 6.897108895438058d-09*zm09

      CbH2YZ(0,3,10) = 
     $  - 1.551849501473563d-08*zm01
     $  - 3.825309021132332d-06*zm02
     $  - 6.536390100206647d-05*zm03
     $  - 2.279278955289295d-04*zm04
     $  - 2.112532726355961d-04*zm05
     $  - 5.216542099203382d-05*zm06
     $  - 2.576070172446114d-06*zm07
     $  - 7.759247507367815d-09*zm08

      CbH2YZ(0,3,11) = 
     $  + 1.058079205550156d-09*zm01
     $  + 5.306267215834035d-07*zm02
     $  + 1.528095623308962d-05*zm03
     $  + 8.984538496701748d-05*zm04
     $  + 1.515081249080695d-04*zm05
     $  + 7.956826164350881d-05*zm06
     $  + 1.196969735932040d-05*zm07
     $  + 3.424649695297340d-07*zm08
     $  + 4.937702959234064d-09*zm09
     $  - 8.817326712917971d-10*zm10
     $  + 8.015751557198156d-11*zm11

      CbH2YZ(0,3,12) = 
     $  - 6.612995034688478d-11*zm01
     $  - 6.695657472622084d-08*zm02
     $  - 3.141437161278415d-06*zm03
     $  - 2.936090439461268d-05*zm04
     $  - 8.181875202708043d-05*zm05
     $  - 7.785207921542324d-05*zm06
     $  - 2.526693142853774d-05*zm07
     $  - 2.422736860908471d-06*zm08
     $  - 4.476997638484100d-08*zm09
     $  - 3.306497517344239d-11*zm10

      CbH2YZ(0,3,13) = 
     $  + 3.815189443089507d-12*zm01
     $  + 7.765818111408691d-09*zm02
     $  + 5.797568236367886d-07*zm03
     $  + 8.267595006288359d-06*zm04
     $  + 3.572183736600214d-05*zm05
     $  + 5.570674316525425d-05*zm06
     $  + 3.286685082325184d-05*zm07
     $  + 6.977054618386863d-06*zm08
     $  + 4.425194632874455d-07*zm09
     $  + 5.461661698607948d-09*zm10
     $  - 7.341514542630808d-11*zm11
     $  + 1.255379002464213d-11*zm12
     $  - 9.656761557417030d-13*zm13

      CbH2YZ(0,3,14) = 
     $  - 2.043851487369378d-13*zm01
     $  - 8.344023697185488d-10*zm02
     $  - 9.747373005443051d-08*zm03
     $  - 2.058537480039297d-06*zm04
     $  - 1.318541693763628d-05*zm05
     $  - 3.150965390256436d-05*zm06
     $  - 3.041916245987744d-05*zm07
     $  - 1.184906732937132d-05*zm08
     $  - 1.713117279149604d-06*zm09
     $  - 7.407745550079013d-08*zm10
     $  - 5.567451451594187d-10*zm11
     $  - 1.021925743684689d-13*zm12

      CbH2YZ(0,3,15) = 
     $  + 1.021925743684689d-14*zm01
     $  + 8.356797768981547d-11*zm02
     $  + 1.509386537594730d-08*zm03
     $  + 4.616211289674424d-07*zm04
     $  + 4.240692585775937d-06*zm05
     $  + 1.480694727805548d-05*zm06
     $  + 2.179744625614652d-05*zm07
     $  + 1.393313490579410d-05*zm08
     $  + 3.746296131725950d-06*zm09
     $  + 3.803518557165851d-07*zm10
     $  + 1.144105652711015d-08*zm11
     $  + 5.031281078074287d-11*zm12
     $  + 1.256968664732167d-12*zm13
     $  - 1.788370051448206d-13*zm14
     $  + 1.192246700965470d-14*zm15

      CbH2YZ(0,3,16) = 
     $  - 4.790276923521981d-16*zm01
     $  - 7.840964782266955d-12*zm02
     $  - 2.170886437863232d-09*zm03
     $  - 9.450937719700228d-08*zm04
     $  - 1.214210588244072d-06*zm05
     $  - 5.979452025446564d-06*zm06
     $  - 1.276696606277838d-05*zm07
     $  - 1.243509905082979d-05*zm08
     $  - 5.520276698586223d-06*zm09
     $  - 1.059109173326062d-06*zm10
     $  - 7.731912978436515d-08*zm11
     $  - 1.638615296533780d-09*zm12
     $  - 5.228587262024242d-12*zm13
     $  - 2.395138461760990d-16*zm14

      CbH2YZ(0,3,17) = 
     $  + 2.113357466259697d-17*zm01
     $  + 6.921562705620449d-13*zm02
     $  + 2.919354844022364d-10*zm03
     $  + 1.785095841401823d-08*zm04
     $  + 3.144781174854765d-07*zm05
     $  + 2.125859881820321d-06*zm06
     $  + 6.343018074940180d-06*zm07
     $  + 8.932647112308483d-06*zm08
     $  + 6.055281515232591d-06*zm09
     $  + 1.934333310274407d-06*zm10
     $  + 2.716875935297089d-07*zm11
     $  + 1.452729046928051d-08*zm12
     $  + 2.195752377924366d-10*zm13
     $  + 5.634175782423916d-13*zm14
     $  - 2.037347042723223d-14*zm15
     $  + 2.548004651820442d-15*zm16
     $  - 1.498826265776730d-16*zm17

      CbH2YZ(0,3,18) = 
     $  - 8.805656109415407d-19*zm01
     $  - 5.769333798047333d-14*zm02
     $  - 3.690530430813470d-11*zm03
     $  - 3.136296410016672d-09*zm04
     $  - 7.458209779977210d-08*zm05
     $  - 6.774805711991141d-07*zm06
     $  - 2.745253308185195d-06*zm07
     $  - 5.376249167021425d-06*zm08
     $  - 5.267336285076444d-06*zm09
     $  - 2.580175316052056d-06*zm10
     $  - 6.095579530244810d-07*zm11
     $  - 6.396208561836291d-08*zm12
     $  - 2.542009617608957d-09*zm13
     $  - 2.776550833170861d-11*zm14
     $  - 3.846486701714838d-14*zm15
     $  - 4.402828054707703d-19*zm16

      CbH2YZ(0,3,19) = 
     $  + 3.475916885295555d-20*zm01
     $  + 4.555310735270810d-15*zm02
     $  + 4.405284537309295d-12*zm03
     $  + 5.159608938942122d-10*zm04
     $  + 1.635324289643140d-08*zm05
     $  + 1.962197128150716d-07*zm06
     $  + 1.055737761528969d-06*zm07
     $  + 2.790443577796405d-06*zm08
     $  + 3.793819522102079d-06*zm09
     $  + 2.690105983365838d-06*zm10
     $  + 9.801316282205759d-07*zm11
     $  + 1.749972976278424d-07*zm12
     $  + 1.394481492658244d-08*zm13
     $  + 4.169098807911084d-10*zm14
     $  + 3.318581614550364d-12*zm15
     $  + 1.185464764127082d-15*zm16
     $  + 3.267560495999839d-16*zm17
     $  - 3.630429666839526d-17*zm18
     $  + 1.910752456231329d-18*zm19

      CbH2YZ(0,3,20) = 
     $  - 1.303468831985833d-21*zm01
     $  - 3.416711158498705d-16*zm02
     $  - 4.983846343622425d-13*zm03
     $  - 7.991316419464200d-11*zm04
     $  - 3.340850834893080d-09*zm05
     $  - 5.221450325720978d-08*zm06
     $  - 3.662954862774578d-07*zm07
     $  - 1.276167309424141d-06*zm08
     $  - 2.333734066017895d-06*zm09
     $  - 2.295836656481253d-06*zm10
     $  - 1.214477067494384d-06*zm11
     $  - 3.367405157958857d-07*zm12
     $  - 4.624210178277355d-08*zm13
     $  - 2.836076399088574d-09*zm14
     $  - 6.442709380321523d-11*zm15
     $  - 3.743579091656232d-13*zm16
     $  - 2.277850887960203d-16*zm17
     $  - 6.517344159929166d-22*zm18

      CbH2YZ(0,3,21) = 
     $  + 4.655245828520833d-23*zm01
     $  + 2.440594092404045d-17*zm02
     $  + 5.360986422501693d-14*zm03
     $  + 1.170538909281625d-11*zm04
     $  + 6.399468709194389d-10*zm05
     $  + 1.287765158483164d-08*zm06
     $  + 1.160432828066626d-07*zm07
     $  + 5.228410730350437d-07*zm08
     $  + 1.254756094224214d-06*zm09
     $  + 1.657821866537558d-06*zm10
     $  + 1.218113751286637d-06*zm11
     $  + 4.923834403839267d-07*zm12
     $  + 1.058371984706373d-07*zm13
     $  + 1.134072611814387d-08*zm14
     $  + 5.413317744462503d-10*zm15
     $  + 9.420957656098064d-12*zm16
     $  + 4.010268510735302d-14*zm17
     $  + 4.903139573308315d-17*zm18
     $  - 5.172703910631728d-18*zm19
     $  + 5.172727186860871d-19*zm20
     $  - 2.463203422314700d-20*zm21

      CbH2YZ(3,3,02) = 
     $  + 2.812500000000000d-01*zm02

      CbH2YZ(3,3,03) = 
     $  - 2.109375000000000d-01*zm02

      CbH2YZ(3,3,04) = 
     $  + 9.228515625000000d-02*zm02
     $  + 5.273437500000000d-02*zm03

      CbH2YZ(3,3,05) = 
     $  - 2.966308593750000d-02*zm02
     $  - 5.932617187500000d-02*zm03
     $  - 9.887695312500000d-03*zm04

      CbH2YZ(3,3,06) = 
     $  + 7.662963867187500d-03*zm02
     $  + 3.609008789062500d-02*zm03
     $  + 2.249450683593750d-02*zm04
     $  + 1.483154296875000d-03*zm05

      CbH2YZ(3,3,07) = 
     $  - 1.668548583984375d-03*zm02
     $  - 1.557312011718750d-02*zm03
     $  - 2.298889160156250d-02*zm04
     $  - 6.303405761718750d-03*zm05
     $  - 1.853942871093750d-04*zm06

      CbH2YZ(3,3,08) = 
     $  + 3.153358186994280d-04*zm02
     $  + 5.303600856236049d-03*zm03
     $  + 1.521060807364327d-02*zm04
     $  + 1.017020089285714d-02*zm05
     $  + 1.427701541355678d-03*zm06
     $  + 1.986367361886160d-05*zm07

      CbH2YZ(3,3,09) = 
     $  - 5.276288305010114d-05*zm02
     $  - 1.508397715432303d-03*zm03
     $  - 7.526470082146780d-03*zm04
     $  - 9.807688849312918d-03*zm05
     $  - 3.502834694726126d-03*zm06
     $  - 2.743669918605259d-04*zm07
     $  - 1.862219401768275d-06*zm08

      CbH2YZ(3,3,10) = 
     $  + 7.929950952529907d-06*zm02
     $  + 3.709230678422110d-04*zm03
     $  + 3.003340895686830d-03*zm04
     $  + 6.742692972932543d-03*zm05
     $  + 4.760965641055788d-03*zm06
     $  + 9.988013761384146d-04*zm07
     $  + 4.610544868877955d-05*zm08
     $  + 1.551849501473563d-07*zm09

      CbH2YZ(3,3,11) = 
     $  - 1.082415027277810d-06*zm02
     $  - 8.072721106665475d-05*zm03
     $  - 1.011580856783049d-03*zm04
     $  - 3.620643349630492d-03*zm05
     $  - 4.411574485046522d-03*zm06
     $  - 1.868504392249243d-03*zm07
     $  - 2.449051290750503d-04*zm08
     $  - 6.913489529064723d-06*zm09
     $  - 1.163887126105172d-08*zm10

      CbH2YZ(3,3,12) = 
     $  + 1.353680083600731d-07*zm02
     $  + 1.581220016754292d-05*zm03
     $  + 2.964224765536847d-04*zm04
     $  + 1.604119495960412d-03*zm05
     $  + 3.098735332960722d-03*zm06
     $  + 2.282741343142924d-03*zm07
     $  + 6.196804869581352d-04*zm08
     $  + 5.297220638626581d-05*zm09
     $  + 9.384501253726420d-07*zm10
     $  + 7.935594041626174d-10*zm11

      CbH2YZ(3,3,13) = 
     $  - 1.562320076945153d-08*zm02
     $  - 2.822591605680910d-06*zm03
     $  - 7.719770682425322d-05*zm04
     $  - 6.079438756304708d-04*zm05
     $  - 1.757609392198818d-03*zm06
     $  - 2.054506450341167d-03*zm07
     $  - 9.816422080772312d-04*zm08
     $  - 1.789147129233974d-04*zm09
     $  - 1.029201909482430d-05*zm10
     $  - 1.165540374863844d-07*zm11
     $  - 4.959746276016359d-11*zm12

      CbH2YZ(3,3,14) = 
     $  + 1.674118753304258d-09*zm02
     $  + 4.636718273572945d-07*zm03
     $  + 1.815221988342631d-05*zm04
     $  + 2.021446439910082d-04*zm05
     $  + 8.390477625066045d-04*zm06
     $  + 1.462758538698184d-03*zm07
     $  + 1.114295974531307d-03*zm08
     $  + 3.627568392098217d-04*zm09
     $  + 4.593466022696541d-05*zm10
     $  + 1.820509616087089d-06*zm11
     $  + 1.335432123332278d-08*zm12
     $  + 2.861392082317130d-12*zm13

      CbH2YZ(3,3,15) = 
     $  - 1.674220945878626d-10*zm02
     $  - 7.063881844289526d-08*zm03
     $  - 3.900585182661770d-06*zm04
     $  - 6.007758898700302d-05*zm05
     $  - 3.470050415223638d-04*zm06
     $  - 8.629709273397120d-04*zm07
     $  - 9.809344467788703d-04*zm08
     $  - 5.129157037965175d-04*zm09
     $  - 1.179821628903668d-04*zm10
     $  - 1.065353038177877d-05*zm11
     $  - 2.962107360830528d-07*zm12
     $  - 1.420865115504318d-09*zm13
     $  - 1.532888615527034d-13*zm14

      CbH2YZ(3,3,16) = 
     $  + 1.569630039530447d-11*zm02
     $  + 1.004344788671773d-08*zm03
     $  + 7.732675561781079d-07*zm04
     $  + 1.618696841459417d-05*zm05
     $  + 1.269884734852375d-04*zm06
     $  + 4.357073659971258d-04*zm07
     $  + 7.048634832301961d-04*zm08
     $  + 5.516543958242004d-04*zm09
     $  + 2.057139591191656d-04*zm10
     $  + 3.437723174994948d-05*zm11
     $  + 2.259199746585603d-06*zm12
     $  + 4.469397349633707d-08*zm13
     $  + 1.411479046900369d-10*zm14
     $  + 7.664443077635170d-15*zm15

      CbH2YZ(3,3,17) = 
     $  - 1.384988815513292d-12*zm02
     $  - 1.339666807970620d-09*zm03
     $  - 1.425113594144251d-07*zm04
     $  - 3.997954931452429d-06*zm05
     $  - 4.178791142380348d-05*zm06
     $  - 1.927530584759832d-04*zm07
     $  - 4.283575342283654d-04*zm08
     $  - 4.772726874366624d-04*zm09
     $  - 2.677007490288958d-04*zm10
     $  - 7.330880879063833d-05*zm11
     $  - 9.097134757780547d-06*zm12
     $  - 4.422388091591924d-07*zm13
     $  - 6.294724587141757d-09*zm14
     $  - 1.315146577968342d-11*zm15
     $  - 3.592707692641486d-16*zm16

      CbH2YZ(3,3,18) = 
     $  + 1.154166151917186d-13*zm02
     $  + 1.683816627842866d-10*zm03
     $  + 2.457029007360643d-08*zm04
     $  + 9.132463858369741d-07*zm05
     $  + 1.252114086529415d-05*zm06
     $  + 7.606113882060286d-05*zm07
     $  + 2.259240184921400d-04*zm08
     $  + 3.452318063597133d-04*zm09
     $  + 2.762533478095886d-04*zm10
     $  + 1.145036552134957d-04*zm11
     $  + 2.356447360574767d-05*zm12
     $  + 2.209967973115910d-06*zm13
     $  + 8.052384630393577d-08*zm14
     $  + 8.319847727244027d-10*zm15
     $  + 1.153888773749740d-12*zm16
     $  + 1.585018099694773d-17*zm17

      CbH2YZ(3,3,19) = 
     $  - 9.111872800620328d-15*zm02
     $  - 2.001738378321725d-11*zm03
     $  - 3.983625956941663d-09*zm04
     $  - 1.943413719244756d-07*zm05
     $  - 3.450626322883605d-06*zm06
     $  - 2.714668866419769d-05*zm07
     $  - 1.054504426963988d-04*zm08
     $  - 2.147184561270406d-04*zm09
     $  - 2.356890100093108d-04*zm10
     $  - 1.398463724822909d-04*zm11
     $  - 4.388075564901820d-05*zm12
     $  - 6.913850367773076d-06*zm13
     $  - 4.971312770752795d-07*zm14
     $  - 1.372397060570912d-08*zm15
     $  - 1.036600872981037d-10*zm16
     $  - 9.566244655866163d-14*zm17
     $  - 6.604242082061555d-19*zm18

      CbH2YZ(3,3,20) = 
     $  + 6.833917635153565d-16*zm02
     $  + 2.258106166289920d-12*zm03
     $  + 6.100528223322208d-10*zm04
     $  + 3.875983414142538d-08*zm05
     $  + 8.817766039550573d-07*zm06
     $  + 8.860920994661670d-06*zm07
     $  + 4.422393098020732d-05*zm08
     $  + 1.172780016211676d-04*zm09
     $  + 1.713342718656822d-04*zm10
     $  + 1.396631415916379d-04*zm11
     $  + 6.301246685277244d-05*zm12
     $  + 1.526198868455535d-05*zm13
     $  + 1.869345141117481d-06*zm14
     $  + 1.042883089632226d-07*zm15
     $  + 2.200828948041748d-09*zm16
     $  + 1.222124315175885d-11*zm17
     $  + 7.516801045824448d-15*zm18
     $  + 2.606937663971666d-20*zm19

      CbH2YZ(3,3,21) = 
     $  - 4.881374394641232d-17*zm02
     $  - 2.424036756289516d-13*zm03
     $  - 8.857501421924943d-11*zm04
     $  - 7.281719323571613d-09*zm05
     $  - 2.103585714519037d-07*zm06
     $  - 2.669097114288184d-06*zm07
     $  - 1.686708900792155d-05*zm08
     $  - 5.718429879107132d-05*zm09
     $  - 1.085613696713185d-04*zm10
     $  - 1.177952399091374d-04*zm11
     $  - 7.319412467375933d-05*zm12
     $  - 2.562417026851425d-05*zm13
     $  - 4.867850864895549d-06*zm14
     $  - 4.694212594007101d-07*zm15
     $  - 2.052277402808432d-08*zm16
     $  - 3.335429329039150d-10*zm17
     $  - 1.367869137759765d-12*zm18
     $  - 5.613380145504499d-16*zm19
     $  - 9.776016239893749d-22*zm20

      CbH2YZ(2,3,02) = 
     $  + 2.812500000000000d-01*zm01
     $  - 2.812500000000000d-01*zm02

      CbH2YZ(2,3,03) = 
     $  - 7.031250000000000d-02*zm01
     $  + 7.031250000000000d-02*zm02

      CbH2YZ(2,3,04) = 
     $  + 1.318359375000000d-02*zm01
     $  - 1.318359375000000d-02*zm03

      CbH2YZ(2,3,05) = 
     $  - 1.977539062500000d-03*zm01
     $  - 5.932617187500000d-03*zm02
     $  + 5.932617187500000d-03*zm03
     $  + 1.977539062500000d-03*zm04

      CbH2YZ(2,3,06) = 
     $  + 2.471923828125000d-04*zm01
     $  + 2.471923828125000d-03*zm02
     $  - 2.471923828125000d-03*zm04
     $  - 2.471923828125000d-04*zm05

      CbH2YZ(2,3,07) = 
     $  - 2.648489815848214d-05*zm01
     $  - 6.621224539620535d-04*zm02
     $  - 1.059395926339285d-03*zm03
     $  + 1.059395926339285d-03*zm04
     $  + 6.621224539620535d-04*zm05
     $  + 2.648489815848214d-05*zm06

      CbH2YZ(2,3,08) = 
     $  + 2.482959202357700d-06*zm01
     $  + 1.390457153320312d-04*zm02
     $  + 6.083250045776367d-04*zm03
     $  - 6.083250045776367d-04*zm05
     $  - 1.390457153320312d-04*zm06
     $  - 2.482959202357700d-06*zm07

      CbH2YZ(2,3,09) = 
     $  - 2.069132668631417d-07*zm01
     $  - 2.462267875671386d-05*zm02
     $  - 2.216041088104248d-04*zm03
     $  - 2.534687519073486d-04*zm04
     $  + 2.534687519073486d-04*zm05
     $  + 2.216041088104248d-04*zm06
     $  + 2.462267875671386d-05*zm07
     $  + 2.069132668631417d-07*zm08

      CbH2YZ(2,3,10) = 
     $  + 1.551849501473563d-08*zm01
     $  + 3.817549773624965d-06*zm02
     $  + 6.278783082962036d-05*zm03
     $  + 1.757624745368957d-04*zm04
     $  - 1.757624745368957d-04*zm06
     $  - 6.278783082962036d-05*zm07
     $  - 3.817549773624965d-06*zm08
     $  - 1.551849501473563d-08*zm09

      CbH2YZ(2,3,11) = 
     $  - 1.058079205550156d-09*zm01
     $  - 5.300976819806284d-07*zm02
     $  - 1.492526527349050d-05*zm03
     $  - 7.790213958783583d-05*zm04
     $  - 7.190283049236644d-05*zm05
     $  + 7.190283049236644d-05*zm06
     $  + 7.790213958783583d-05*zm07
     $  + 1.492526527349050d-05*zm08
     $  + 5.300976819806284d-07*zm09
     $  + 1.058079205550156d-09*zm10

      CbH2YZ(2,3,12) = 
     $  + 6.612995034688478d-11*zm01
     $  + 6.692350975104740d-08*zm02
     $  + 3.096667184893574d-06*zm03
     $  + 2.693816753370421d-05*zm04
     $  + 5.655182059854269d-05*zm05
     $  - 5.655182059854269d-05*zm07
     $  - 2.693816753370421d-05*zm08
     $  - 3.096667184893574d-06*zm09
     $  - 6.692350975104740d-08*zm10
     $  - 6.612995034688478d-11*zm11

      CbH2YZ(2,3,13) = 
     $  - 3.815189443089507d-12*zm01
     $  - 7.763910516687146d-09*zm02
     $  - 5.745713453187228d-07*zm03
     $  - 7.824385084549558d-06*zm04
     $  - 2.874602557282772d-05*zm05
     $  - 2.283823524171916d-05*zm06
     $  + 2.283823524171916d-05*zm07
     $  + 2.874602557282772d-05*zm08
     $  + 7.824385084549558d-06*zm09
     $  + 5.745713453187228d-07*zm10
     $  + 7.763910516687146d-09*zm11
     $  + 3.815189443089507d-12*zm12

      CbH2YZ(2,3,14) = 
     $  + 2.043851487369378d-13*zm01
     $  + 8.343001771441804d-10*zm02
     $  + 9.691698490927109d-08*zm03
     $  + 1.984460024538506d-06*zm04
     $  + 1.147229965848667d-05*zm05
     $  + 1.966058657319303d-05*zm06
     $  - 1.966058657319303d-05*zm08
     $  - 1.147229965848667d-05*zm09
     $  - 1.984460024538506d-06*zm10
     $  - 9.691698490927109d-08*zm11
     $  - 8.343001771441804d-10*zm12
     $  - 2.043851487369378d-13*zm13

      CbH2YZ(2,3,15) = 
     $  - 1.021925743684689d-14*zm01
     $  - 8.356286806109705d-11*zm02
     $  - 1.503812784267717d-08*zm03
     $  - 4.501963466078004d-07*zm04
     $  - 3.860304926890922d-06*zm05
     $  - 1.106071081827691d-05*zm06
     $  - 7.864234629277215d-06*zm07
     $  + 7.864234629277215d-06*zm08
     $  + 1.106071081827691d-05*zm09
     $  + 3.860304926890922d-06*zm10
     $  + 4.501963466078004d-07*zm11
     $  + 1.503812784267717d-08*zm12
     $  + 8.356286806109705d-11*zm13
     $  + 1.021925743684689d-14*zm14

      CbH2YZ(2,3,16) = 
     $  + 4.790276923521981d-16*zm01
     $  + 7.840725268420779d-12*zm02
     $  + 2.165657850601208d-09*zm03
     $  + 9.287076190046850d-08*zm04
     $  + 1.136891458459707d-06*zm05
     $  + 4.920342852120502d-06*zm06
     $  + 7.246689364192160d-06*zm07
     $  - 7.246689364192160d-06*zm09
     $  - 4.920342852120502d-06*zm10
     $  - 1.136891458459707d-06*zm11
     $  - 9.287076190046850d-08*zm12
     $  - 2.165657850601208d-09*zm13
     $  - 7.840725268420779d-12*zm14
     $  - 4.790276923521981d-16*zm15

      CbH2YZ(2,3,17) = 
     $  - 2.113357466259697d-17*zm01
     $  - 6.921457037747136d-13*zm02
     $  - 2.914739870100668d-10*zm03
     $  - 1.763102645557454d-08*zm04
     $  - 2.999517544898893d-07*zm05
     $  - 1.854170433343226d-06*zm06
     $  - 4.408687679583094d-06*zm07
     $  - 2.877361953429240d-06*zm08
     $  + 2.877361953429240d-06*zm09
     $  + 4.408687679583094d-06*zm10
     $  + 1.854170433343226d-06*zm11
     $  + 2.999517544898893d-07*zm12
     $  + 1.763102645557454d-08*zm13
     $  + 2.914739870100668d-10*zm14
     $  + 6.921457037747136d-13*zm15
     $  + 2.113357466259697d-17*zm16

      CbH2YZ(2,3,18) = 
     $  + 8.805656109415407d-19*zm01
     $  + 5.769289769766786d-14*zm02
     $  + 3.686683944111755d-11*zm03
     $  + 3.108530901684963d-09*zm04
     $  + 7.204008818216315d-08*zm05
     $  + 6.135184855807512d-07*zm06
     $  + 2.135695355160714d-06*zm07
     $  + 2.796073850969369d-06*zm08
     $  - 2.796073850969369d-06*zm10
     $  - 2.135695355160714d-06*zm11
     $  - 6.135184855807512d-07*zm12
     $  - 7.204008818216315d-08*zm13
     $  - 3.108530901684963d-09*zm14
     $  - 3.686683944111755d-11*zm15
     $  - 5.769289769766786d-14*zm16
     $  - 8.805656109415407d-19*zm17

      CbH2YZ(2,3,19) = 
     $  - 3.475916885295555d-20*zm01
     $  - 4.555293355686384d-15*zm02
     $  - 4.402247553415080d-12*zm03
     $  - 5.126497183561822d-10*zm04
     $  - 1.593631079741073d-08*zm05
     $  - 1.822749497310247d-07*zm06
     $  - 8.807403676221322d-07*zm07
     $  - 1.810312093994321d-06*zm08
     $  - 1.103713362224751d-06*zm09
     $  + 1.103713362224751d-06*zm10
     $  + 1.810312093994321d-06*zm11
     $  + 8.807403676221322d-07*zm12
     $  + 1.822749497310247d-07*zm13
     $  + 1.593631079741073d-08*zm14
     $  + 5.126497183561822d-10*zm15
     $  + 4.402247553415080d-12*zm16
     $  + 4.555293355686384d-15*zm17
     $  + 3.475916885295555d-20*zm18

      CbH2YZ(2,3,20) = 
     $  + 1.303468831985833d-21*zm01
     $  + 3.416704641154545d-16*zm02
     $  + 4.981568492734465d-13*zm03
     $  + 7.953880628547638d-11*zm04
     $  + 3.276423741089865d-09*zm05
     $  + 4.937842685812121d-08*zm06
     $  + 3.200533844946843d-07*zm07
     $  + 9.394267936282560d-07*zm08
     $  + 1.119256998523511d-06*zm09
     $  - 1.119256998523511d-06*zm11
     $  - 9.394267936282560d-07*zm12
     $  - 3.200533844946843d-07*zm13
     $  - 4.937842685812121d-08*zm14
     $  - 3.276423741089865d-09*zm15
     $  - 7.953880628547638d-11*zm16
     $  - 4.981568492734465d-13*zm17
     $  - 3.416704641154545d-16*zm18
     $  - 1.303468831985833d-21*zm19

      CbH2YZ(2,3,21) = 
     $  - 4.655245828520833d-23*zm01
     $  - 2.440591764781131d-17*zm02
     $  - 5.359359343480063d-14*zm03
     $  - 1.166513898498407d-11*zm04
     $  - 6.305264145006052d-10*zm05
     $  - 1.233631847375268d-08*zm06
     $  - 1.047025595527316d-07*zm07
     $  - 4.170038695520337d-07*zm08
     $  - 7.623726610803812d-07*zm09
     $  - 4.397081065628079d-07*zm10
     $  + 4.397081065628079d-07*zm11
     $  + 7.623726610803812d-07*zm12
     $  + 4.170038695520337d-07*zm13
     $  + 1.047025595527316d-07*zm14
     $  + 1.233631847375268d-08*zm15
     $  + 6.305264145006052d-10*zm16
     $  + 1.166513898498407d-11*zm17
     $  + 5.359359343480063d-14*zm18
     $  + 2.440591764781131d-17*zm19
     $  + 4.655245828520833d-23*zm20
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CbH3YZ(0,0,3,01) = 
     $  + 7.500000000000000d-01*zm01

      CbH3YZ(0,0,3,02) = 
     $  - 2.812500000000000d-01*zm01
     $  + 2.109375000000000d-01*zm02

      CbH3YZ(0,0,3,03) = 
     $  + 7.031250000000000d-02*zm01
     $  - 8.789062500000000d-02*zm02
     $  + 3.320312500000000d-02*zm03

      CbH3YZ(0,0,3,04) = 
     $  - 1.318359375000000d-02*zm01
     $  + 1.647949218750000d-02*zm02
     $  - 1.098632812500000d-02*zm03
     $  + 2.746582031250000d-03*zm04

      CbH3YZ(0,0,3,05) = 
     $  + 1.977539062500000d-03*zm01
     $  - 4.943847656250000d-04*zm02
     $  + 6.042480468750000d-04*zm03
     $  - 2.197265625000000d-04*zm04
     $  + 3.076171875000000d-05*zm05

      CbH3YZ(0,0,3,06) = 
     $  - 2.471923828125000d-04*zm01
     $  - 6.797790527343750d-04*zm02
     $  + 2.471923828125000d-04*zm03
     $  - 2.162933349609375d-04*zm04
     $  + 8.651733398437500d-05*zm05
     $  - 1.441955566406250d-05*zm06

      CbH3YZ(0,0,3,07) = 
     $  + 2.648489815848214d-05*zm01
     $  + 2.582277570452008d-04*zm02
     $  + 8.313315255301339d-05*zm03
     $  + 3.163473946707589d-05*zm04
     $  - 1.368386404854910d-05*zm05
     $  + 3.825596400669642d-06*zm06
     $  - 4.564290839103498d-07*zm07

      CbH3YZ(0,0,3,08) = 
     $  - 2.482959202357700d-06*zm01
     $  - 6.145324025835309d-05*zm02
     $  - 1.121469906398228d-04*zm03
     $  - 1.438047204698835d-05*zm04
     $  - 7.862704140799386d-06*zm05
     $  + 3.724438803536551d-06*zm06
     $  - 1.064125372439014d-06*zm07
     $  + 1.330156715548768d-07*zm08

      CbH3YZ(0,0,3,09) = 
     $  + 2.069132668631417d-07*zm01
     $  + 1.153541462762015d-05*zm02
     $  + 5.255022219249180d-05*zm03
     $  + 3.376134804316929d-05*zm04
     $  + 5.440439496721540d-06*zm05
     $  - 7.313234465462820d-07*zm06
     $  + 2.927330075478067d-07*zm07
     $  - 6.542400437958386d-08*zm08
     $  + 6.502988387127311d-09*zm09

      CbH3YZ(0,0,3,10) = 
     $  - 1.551849501473563d-08*zm01
     $  - 1.842821282999856d-06*zm02
     $  - 1.687377691268920d-05*zm03
     $  - 2.745286428502627d-05*zm04
     $  - 9.307217385087694d-06*zm05
     $  - 9.382223444325583d-07*zm06
     $  + 1.681170293263026d-07*zm07
     $  - 6.401379193578447d-08*zm08
     $  + 1.422528709684099d-08*zm09
     $  - 1.422528709684099d-09*zm10

      CbH3YZ(0,0,3,11) = 
     $  + 1.058079205550156d-09*zm01
     $  + 2.600229647639509d-07*zm02
     $  + 4.313583183404687d-06*zm03
     $  + 1.383417987494499d-05*zm04
     $  + 1.093377315669090d-05*zm05
     $  + 2.327603783893894d-06*zm06
     $  + 4.738263426441872d-08*zm07
     $  + 1.911680405900739d-08*zm08
     $  - 5.823634357532012d-09*zm09
     $  + 1.076553604377222d-09*zm10
     $  - 9.058146261865829d-11*zm11

      CbH3YZ(0,0,3,12) = 
     $  - 6.612995034688478d-11*zm01
     $  - 3.311457263620255d-08*zm02
     $  - 9.367638116387964d-07*zm03
     $  - 5.232507522465878d-06*zm04
     $  - 7.991738369470679d-06*zm05
     $  - 3.651651771521413d-06*zm06
     $  - 4.795743999156084d-07*zm07
     $  - 3.108107666303585d-09*zm08
     $  - 3.593060635514073d-09*zm09
     $  + 1.074611693136877d-09*zm10
     $  - 1.953839442067050d-10*zm11
     $  + 1.628199535055875d-11*zm12

      CbH3YZ(0,0,3,13) = 
     $  + 3.815189443089507d-12*zm01
     $  + 3.860017919045808d-09*zm02
     $  + 1.790988755090949d-07*zm03
     $  + 1.619151562799352d-06*zm04
     $  + 4.229894660161594d-06*zm05
     $  + 3.644597647326917d-06*zm06
     $  + 1.050666755994774d-06*zm07
     $  + 8.413176030227721d-08*zm08
     $  + 2.428962419562163d-09*zm09
     $  - 4.254187979640704d-10*zm10
     $  + 1.093492044059912d-10*zm11
     $  - 1.717871823227835d-11*zm12
     $  + 1.247157082810511d-12*zm13

      CbH3YZ(0,0,3,14) = 
     $  - 2.043851487369378d-13*zm01
     $  - 4.158726813924843d-10*zm02
     $  - 3.082775262590690d-08*zm03
     $  - 4.298580502885802d-07*zm04
     $  - 1.777367286950095d-06*zm05
     $  - 2.585558053335583d-06*zm06
     $  - 1.390671147598968d-06*zm07
     $  - 2.642961620223181d-07*zm08
     $  - 1.414892300174395d-08*zm09
     $  - 3.332840492070333d-10*zm10
     $  + 7.058100469715588d-11*zm11
     $  - 1.765376722215300d-11*zm12
     $  + 2.715964188023539d-12*zm13
     $  - 1.939974420016814d-13*zm14

      CbH3YZ(0,0,3,15) = 
     $  + 1.021925743684689d-14*zm01
     $  + 4.171245404284980d-11*zm02
     $  + 4.850534491130087d-09*zm03
     $  + 1.008536787684703d-07*zm04
     $  + 6.262604238645527d-07*zm05
     $  + 1.424057173234992d-06*zm06
     $  + 1.282990242433084d-06*zm07
     $  + 4.586516207911783d-07*zm08
     $  + 5.952608735418918d-08*zm09
     $  + 2.319533159145009d-09*zm10
     $  - 1.423720623771747d-11*zm11
     $  + 8.938469644301065d-12*zm12
     $  - 1.966033866782386d-12*zm13
     $  + 2.680879091728537d-13*zm14
     $  - 1.707769614421327d-14*zm15

      CbH3YZ(0,0,3,16) = 
     $  - 4.790276923521981d-16*zm01
     $  - 3.916889683440836d-12*zm02
     $  - 7.053499940983536d-10*zm03
     $  - 2.133495681843092d-08*zm04
     $  - 1.916382212845802d-07*zm05
     $  - 6.452385985526970d-07*zm06
     $  - 9.020828681787734d-07*zm07
     $  - 5.395441546526037d-07*zm08
     $  - 1.337692734850437d-07*zm09
     $  - 1.227242589307559d-08*zm10
     $  - 3.349613114465054d-10*zm11
     $  + 3.015938391562258d-12*zm12
     $  - 1.330180063713328d-12*zm13
     $  + 2.850214769495579d-13*zm14
     $  - 3.800286359327438d-14*zm15
     $  + 2.375178974579649d-15*zm16

      CbH3YZ(0,0,3,17) = 
     $  + 2.113357466259697d-17*zm01
     $  + 3.459090666837216d-13*zm02
     $  + 9.558228280066021d-11*zm03
     $  + 4.128201613702248d-09*zm04
     $  + 5.216229930146946d-08*zm05
     $  + 2.499853817004480d-07*zm06
     $  + 5.133112680336074d-07*zm07
     $  + 4.749418039965511d-07*zm08
     $  + 1.978672532515145d-07*zm09
     $  + 3.513952842622906d-08*zm10
     $  + 2.337354049645954d-09*zm11
     $  + 4.193051428373237d-11*zm12
     $  + 7.632820287519057d-13*zm13
     $  - 1.778364669118022d-13*zm14
     $  + 3.420906202054496d-14*zm15
     $  - 4.116882461829343d-15*zm16
     $  + 2.333529314853923d-16*zm17

      CbH3YZ(0,0,3,18) = 
     $  - 8.805656109415407d-19*zm01
     $  - 2.883918418254366d-14*zm02
     $  - 1.214795912040466d-11*zm03
     $  - 7.385192558026505d-10*zm04
     $  - 1.284856564370700d-08*zm05
     $  - 8.507486963848718d-08*zm06
     $  - 2.463364103604784d-07*zm07
     $  - 3.333185816320202d-07*zm08
     $  - 2.149056076395823d-07*zm09
     $  - 6.460248472958145d-08*zm10
     $  - 8.430734107984487d-09*zm11
     $  - 4.122455718726253d-10*zm12
     $  - 5.272014336400449d-12*zm13
     $  - 1.003883321218835d-13*zm14
     $  + 2.420589743135904d-14*zm15
     $  - 4.538633286055162d-15*zm16
     $  + 5.339568571829603d-16*zm17
     $  - 2.966426984349779d-17*zm18

      CbH3YZ(0,0,3,19) = 
     $  + 3.475916885295555d-20*zm01
     $  + 2.277342535115728d-15*zm02
     $  + 1.455523238070775d-12*zm03
     $  + 1.231681305212699d-10*zm04
     $  + 2.901144187722470d-09*zm05
     $  + 2.593394903115950d-08*zm06
     $  + 1.026566320176994d-07*zm07
     $  + 1.948204991980015d-07*zm08
     $  + 1.834215589915625d-07*zm09
     $  + 8.558903934502130d-08*zm10
     $  + 1.907547946503492d-08*zm11
     $  + 1.866121825630255d-09*zm12
     $  + 6.784324209005049d-11*zm13
     $  + 7.036020179146802d-13*zm14
     $  - 1.329523166820248d-14*zm15
     $  + 3.397899464808564d-15*zm16
     $  - 5.804083732250417d-16*zm17
     $  + 6.247291387676045d-17*zm18
     $  - 3.187482180027848d-18*zm19

      CbH3YZ(0,0,3,20) = 
     $  - 1.303468831985833d-21*zm01
     $  - 1.708231749710314d-16*zm02
     $  - 1.651032724042546d-13*zm03
     $  - 1.927660214094242d-11*zm04
     $  - 6.064850401276003d-10*zm05
     $  - 7.186204609215964d-09*zm06
     $  - 3.795551739263348d-08*zm07
     $  - 9.784319791498831d-08*zm08
     $  - 1.288461878931150d-07*zm09
     $  - 8.785285896569876d-08*zm10
     $  - 3.054077071249059d-08*zm11
     $  - 5.156131615289197d-09*zm12
     $  - 3.840806815738446d-10*zm13
     $  - 1.053654486226155d-11*zm14
     $  - 8.052164197639489d-14*zm15
     $  + 1.765643794771948d-15*zm16
     $  - 4.288447216402244d-16*zm17
     $  + 7.147408406590318d-17*zm18
     $  - 7.523587796410861d-18*zm19
     $  + 3.761793898205430d-19*zm20

      CbH3YZ(0,0,3,21) = 
     $  + 4.655245828520833d-23*zm01
     $  + 1.220250493743737d-17*zm02
     $  + 1.779267221040187d-14*zm03
     $  + 2.846280248257255d-12*zm04
     $  + 1.183120213398131d-10*zm05
     $  + 1.830776540565772d-09*zm06
     $  + 1.265451924259657d-08*zm07
     $  + 4.320972545483645d-08*zm08
     $  + 7.700329592348225d-08*zm09
     $  + 7.337823154557710d-08*zm10
     $  + 3.735938220775357d-08*zm11
     $  + 9.899134858865919d-09*zm12
     $  + 1.288075749295697d-09*zm13
     $  + 7.400858026987819d-11*zm14
     $  + 1.551447503806861d-12*zm15
     $  + 7.017039578555797d-15*zm16
     $  + 2.951463067396491d-16*zm17
     $  - 6.286410173475074d-17*zm18
     $  + 9.653663225980027d-18*zm19
     $  - 9.395026866636983d-19*zm20
     $  + 4.356526916383577d-20*zm21

      CbH3YZ(0,2,3,02) = 
     $  + 1.406250000000000d-01*zm01
     $  - 1.406250000000000d-01*zm02

      CbH3YZ(0,2,3,03) = 
     $  - 5.859375000000000d-02*zm01
     $  + 9.375000000000000d-02*zm02
     $  - 3.515625000000000d-02*zm03

      CbH3YZ(0,2,3,04) = 
     $  + 1.318359375000000d-02*zm01
     $  - 2.307128906250000d-02*zm02
     $  + 1.318359375000000d-02*zm03
     $  - 3.295898437500000d-03*zm04

      CbH3YZ(0,2,3,05) = 
     $  - 2.043457031250000d-03*zm01
     $  + 1.779785156250000d-03*zm02
     $  + 1.977539062500000d-04*zm03
     $  + 6.591796875000000d-05*zm04

      CbH3YZ(0,2,3,06) = 
     $  + 2.471923828125000d-04*zm01
     $  + 5.561828613281250d-04*zm02
     $  - 9.475708007812500d-04*zm03
     $  + 2.471923828125000d-04*zm04
     $  - 1.235961914062500d-04*zm05
     $  + 2.059936523437500d-05*zm06

      CbH3YZ(0,2,3,07) = 
     $  - 2.585430534518494d-05*zm01
     $  - 2.490841612523915d-04*zm02
     $  + 1.311633051658163d-04*zm03
     $  + 1.336856764190051d-04*zm04
     $  + 1.072007782605229d-05*zm05
     $  - 6.305928132971938d-07*zm06

      CbH3YZ(0,2,3,08) = 
     $  + 2.482959202357700d-06*zm01
     $  + 6.021176065717424d-05*zm02
     $  + 6.786755153111049d-05*zm03
     $  - 1.065603324345179d-04*zm04
     $  - 1.903602055140904d-05*zm05
     $  - 6.414311272757393d-06*zm06
     $  + 1.655306134905133d-06*zm07
     $  - 2.069132668631417d-07*zm08

      CbH3YZ(0,2,3,09) = 
     $  - 2.138103757585797d-07*zm01
     $  - 1.137333256857735d-05*zm02
     $  - 4.476913384028843d-05*zm03
     $  + 1.431150095803397d-05*zm04
     $  + 3.638224942343575d-05*zm05
     $  + 5.662526403154645d-06*zm06
     $  - 6.897108895438058d-09*zm07
     $  + 6.897108895438058d-09*zm08

      CbH3YZ(0,2,3,10) = 
     $  + 1.551849501473563d-08*zm01
     $  + 1.835062035492488d-06*zm02
     $  + 1.561936523233141d-05*zm03
     $  + 1.330710947513580d-05*zm04
     $  - 1.930345594882965d-05*zm05
     $  - 1.042842864990234d-05*zm06
     $  - 1.125090888568333d-06*zm07
     $  + 1.008702175957815d-07*zm08
     $  - 2.327774252210344d-08*zm09
     $  + 2.327774252210344d-09*zm10

      CbH3YZ(0,2,3,11) = 
     $  - 9.779216899781750d-10*zm01
     $  - 2.603355790746817d-07*zm02
     $  - 4.134300212159207d-06*zm03
     $  - 1.037581325668089d-05*zm04
     $  + 2.198880967170598d-06*zm05
     $  + 9.784924114890475d-06*zm06
     $  + 2.672996640275754d-06*zm07
     $  + 1.135992310686122d-07*zm08
     $  + 1.106173714893345d-09*zm09
     $  - 8.015751557198156d-11*zm10

      CbH3YZ(0,2,3,12) = 
     $  + 6.612995034688478d-11*zm01
     $  + 3.308150766102911d-08*zm02
     $  + 9.145551699806343d-07*zm03
     $  + 4.497117675877139d-06*zm04
     $  + 3.296181245090125d-06*zm05
     $  - 4.506656921214678d-06*zm06
     $  - 3.649315179942490d-06*zm07
     $  - 5.747023334896022d-07*zm08
     $  - 8.795283396135676d-09*zm09
     $  - 1.835106122126052d-09*zm10
     $  + 3.306497517344239d-10*zm11
     $  - 2.755414597786866d-11*zm12

      CbH3YZ(0,2,3,13) = 
     $  - 4.780865598831210d-12*zm01
     $  - 3.846039372377492d-09*zm02
     $  - 1.765866317231251d-07*zm03
     $  - 1.480423920509944d-06*zm04
     $  - 2.826223506715128d-06*zm05
     $  + 4.376875132802250d-07*zm06
     $  + 2.824917521251083d-06*zm07
     $  + 1.117743522205215d-06*zm08
     $  + 1.049527465400997d-07*zm09
     $  + 1.793727387979372d-09*zm10
     $  - 1.111715458599891d-11*zm11
     $  + 9.656761557417030d-13*zm12

      CbH3YZ(0,2,3,14) = 
     $  + 2.043851487369378d-13*zm01
     $  + 4.157704888181158d-10*zm02
     $  + 3.055002727296485d-08*zm03
     $  + 4.062533624955632d-07*zm04
     $  + 1.412869166528270d-06*zm05
     $  + 9.347357443622748d-07*zm06
     $  - 1.224053913940730d-06*zm07
     $  - 1.250214716615126d-06*zm08
     $  - 2.927401945034401d-07*zm09
     $  - 1.753388511316135d-08*zm10
     $  - 3.077427184532073d-10*zm11
     $  + 3.054876689788098d-11*zm12
     $  - 4.707671259240802d-12*zm13
     $  + 3.362622328029144d-13*zm14

      CbH3YZ(0,2,3,15) = 
     $  + 1.703209572807815d-15*zm01
     $  - 4.188022018577137d-11*zm02
     $  - 4.821534225602150d-09*zm03
     $  - 9.716856940168968d-08*zm04
     $  - 5.420401838444881d-07*zm05
     $  - 8.576481095005158d-07*zm06
     $  + 1.040562399087118d-07*zm07
     $  + 8.789730887509401d-07*zm08
     $  + 4.482337637651976d-07*zm09
     $  + 6.764820214707963d-08*zm10
     $  + 2.791417420227894d-09*zm11
     $  + 1.739998899580464d-11*zm12
     $  + 1.754305859992050d-13*zm13
     $  - 1.192246700965470d-14*zm14

      CbH3YZ(0,2,3,16) = 
     $  + 4.790276923521981d-16*zm01
     $  + 3.916650169594660d-12*zm02
     $  + 7.027374569022135d-10*zm03
     $  + 2.080072595485256d-08*zm04
     $  + 1.740184999177678d-07*zm05
     $  + 4.751729896845077d-07*zm06
     $  + 2.908263057838202d-07*zm07
     $  - 3.686629705003720d-07*zm08
     $  - 4.406131764711768d-07*zm09
     $  - 1.376875289961923d-07*zm10
     $  - 1.415135177859478d-08*zm11
     $  - 4.103129662381216d-10*zm12
     $  + 6.050119754408262d-13*zm13
     $  - 5.030988338928961d-13*zm14
     $  + 6.706387692930774d-14*zm15
     $  - 4.191492308081733d-15*zm16

      CbH3YZ(0,2,3,17) = 
     $  - 1.710162012402700d-16*zm01
     $  - 3.434254365578587d-13*zm02
     $  - 9.537077687350703d-11*zm03
     $  - 4.055932533300207d-09*zm04
     $  - 4.878007857184137d-08*zm05
     $  - 2.043366489998425d-07*zm06
     $  - 2.813509870098538d-07*zm07
     $  + 2.798546845594465d-08*zm08
     $  + 2.917214152584153d-07*zm09
     $  + 1.787180526537513d-07*zm10
     $  + 3.741967575010157d-08*zm11
     $  + 2.720341958933790d-09*zm12
     $  + 5.423676639429751d-11*zm13
     $  + 1.729625205899318d-13*zm14
     $  - 2.467779944865956d-15*zm15
     $  + 1.498826265776730d-16*zm16

      CbH3YZ(0,2,3,18) = 
     $  + 8.805656109415407d-19*zm01
     $  + 2.883874389973819d-14*zm02
     $  + 1.212873035591947d-11*zm03
     $  + 7.293650375924363d-10*zm04
     $  + 1.224693817269226d-08*zm05
     $  + 7.394804619239042d-08*zm06
     $  + 1.686280260967709d-07*zm07
     $  + 9.701094647146872d-08*zm08
     $  - 1.198913718510742d-07*zm09
     $  - 1.608138622364719d-07*zm10
     $  - 6.219149683052722d-08*zm11
     $  - 9.187949132702297d-09*zm12
     $  - 4.840293234261673d-10*zm13
     $  - 6.721249439029440d-12*zm14
     $  - 5.613638057158056d-14*zm15
     $  + 8.121544686274924d-15*zm16
     $  - 9.555017444326658d-16*zm17
     $  + 5.308343024625921d-17*zm18

      CbH3YZ(0,2,3,19) = 
     $  + 1.875993287378374d-18*zm01
     $  - 2.312674075971581d-15*zm02
     $  - 1.453695676204164d-12*zm03
     $  - 1.220745825141916d-10*zm04
     $  - 2.801231171749559d-09*zm05
     $  - 2.344154262206265d-08*zm06
     $  - 7.925399182636396d-08*zm07
     $  - 9.795725132875790d-08*zm08
     $  + 8.239329414350550d-09*zm09
     $  + 1.021320068081245d-07*zm10
     $  + 7.188967912578468d-08*zm11
     $  + 1.916806276399698d-08*zm12
     $  + 2.067395356328916d-09*zm13
     $  + 8.024864752015707d-11*zm14
     $  + 8.246853203555872d-13*zm15
     $  + 7.030494902209597d-16*zm16
     $  + 3.535761023249284d-17*zm17
     $  - 1.910752456231329d-18*zm18

      CbH3YZ(0,2,3,20) = 
     $  + 1.303468831985833d-21*zm01
     $  + 1.708225232366154d-16*zm02
     $  + 1.649893859427112d-13*zm03
     $  + 1.915248999981341d-11*zm04
     $  + 5.908995661402847d-10*zm05
     $  + 6.668359821702073d-09*zm06
     $  + 3.153199221663129d-08*zm07
     $  + 6.249449054130963d-08*zm08
     $  + 3.418143001089311d-08*zm09
     $  - 4.138931396564662d-08*zm10
     $  - 6.067876812094572d-08*zm11
     $  - 2.763563655083224d-08*zm12
     $  - 5.339486277419556d-09*zm13
     $  - 4.306872458391493d-10*zm14
     $  - 1.250172201669754d-11*zm15
     $  - 9.650786316011954d-14*zm16
     $  + 7.000781508029954d-16*zm17
     $  - 1.293343827483661d-16*zm18
     $  + 1.361411125064822d-17*zm19
     $  - 6.807055625324113d-19*zm20

      CbH3YZ(0,2,3,21) = 
     $  - 2.467858668143221d-20*zm01
     $  - 1.169752495963372d-17*zm02
     $  - 1.778945934568223d-14*zm03
     $  - 2.832884049940860d-12*zm04
     $  - 1.160165200800705d-10*zm05
     $  - 1.730239981898968d-09*zm06
     $  - 1.103029040428273d-08*zm07
     $  - 3.154671521302623d-08*zm08
     $  - 3.575641427024467d-08*zm09
     $  + 2.601120616968608d-09*zm10
     $  + 3.737234361601393d-08*zm11
     $  + 2.933390695255699d-08*zm12
     $  + 9.426236315398668d-09*zm13
     $  + 1.363327631548340d-09*zm14
     $  + 8.373987057642027d-11*zm15
     $  + 1.842058122770978d-12*zm16
     $  + 1.000371293361236d-14*zm17
     $  + 1.034581677755609d-17*zm18
     $  - 5.049450634599423d-19*zm19
     $  + 2.463203422314700d-20*zm20

      CbH3YZ(0,3,3,02) = 
     $  + 1.406250000000000d-01*zm02

      CbH3YZ(0,3,3,03) = 
     $  - 1.054687500000000d-01*zm02
     $  + 3.515625000000000d-02*zm03

      CbH3YZ(0,3,3,04) = 
     $  + 4.614257812500000d-02*zm02
     $  - 1.318359375000000d-02*zm03
     $  + 3.295898437500000d-03*zm04

      CbH3YZ(0,3,3,05) = 
     $  - 1.483154296875000d-02*zm02
     $  - 4.943847656250000d-03*zm03

      CbH3YZ(0,3,3,06) = 
     $  + 3.831481933593750d-03*zm02
     $  + 6.921386718750000d-03*zm03
     $  + 4.325866699218750d-04*zm04
     $  + 1.235961914062500d-04*zm05
     $  - 2.059936523437500d-05*zm06

      CbH3YZ(0,3,3,07) = 
     $  - 8.342742919921875d-04*zm02
     $  - 3.800582885742187d-03*zm03
     $  - 1.946640014648437d-03*zm04
     $  - 9.269714355468750d-05*zm05

      CbH3YZ(0,3,3,08) = 
     $  + 1.576679093497140d-04*zm02
     $  + 1.452531133379255d-03*zm03
     $  + 1.986988101686750d-03*zm04
     $  + 4.444496972220284d-04*zm05
     $  + 1.572540828159877d-05*zm06
     $  - 1.655306134905133d-06*zm07
     $  + 2.069132668631417d-07*zm08

      CbH3YZ(0,3,3,09) = 
     $  - 2.638144152505057d-05*zm02
     $  - 4.412425415856497d-04*zm03
     $  - 1.219753708158220d-03*zm04
     $  - 7.417840617043631d-04*zm05
     $  - 8.928307465144566d-05*zm06
     $  - 9.311097008841378d-07*zm07

      CbH3YZ(0,3,3,10) = 
     $  + 3.964975476264953d-06*zm02
     $  + 1.130677546773638d-04*zm03
     $  + 5.529666532363210d-04*zm04
     $  + 6.849786107029233d-04*zm05
     $  + 2.226787645901952d-04*zm06
     $  + 1.544090253966195d-05*zm07
     $  - 2.715736627578735d-08*zm08
     $  + 2.327774252210344d-08*zm09
     $  - 2.327774252210344d-09*zm10

      CbH3YZ(0,3,3,11) = 
     $  - 5.412075136389051d-07*zm02
     $  - 2.528544781463486d-05*zm03
     $  - 2.023243185664926d-04*zm04
     $  - 4.408746239330087d-04*zm05
     $  - 2.943877902414117d-04*zm06
     $  - 5.665220586316926d-05*zm07
     $  - 2.287038202796663d-06*zm08
     $  - 5.819435630525861d-09*zm09

      CbH3YZ(0,3,3,12) = 
     $  + 6.768400418003658d-08*zm02
     $  + 5.045120041914187d-06*zm03
     $  + 6.275409904411425d-05*zm04
     $  + 2.204173407214996d-04*zm05
     $  + 2.593023246517042d-04*zm06
     $  + 1.038467707475284d-04*zm07
     $  + 1.255582915256162d-05*zm08
     $  + 3.054321973354785d-07*zm09
     $  + 2.215353336620640d-09*zm10
     $  - 3.306497517344239d-10*zm11
     $  + 2.755414597786866d-11*zm12

      CbH3YZ(0,3,3,13) = 
     $  - 7.811600384725765d-09*zm02
     $  - 9.122213338163088d-07*zm03
     $  - 1.701887337152253d-05*zm04
     $  - 9.095480305735360d-05*zm05
     $  - 1.716618279566649d-04*zm06
     $  - 1.218390935206447d-04*zm07
     $  - 3.132595586917036d-05*zm08
     $  - 2.476103730838407d-06*zm09
     $  - 3.876041714706784d-08*zm10
     $  - 2.479873138008179d-11*zm11

      CbH3YZ(0,3,3,14) = 
     $  + 8.370593766521290d-10*zm02
     $  + 1.512090382791563d-07*zm03
     $  + 4.122230115588898d-06*zm04
     $  + 3.218446856702385d-05*zm05
     $  + 9.156459090056496d-05*zm06
     $  + 1.043202587848092d-04*zm07
     $  + 4.800677037970526d-05*zm08
     $  + 8.301801881287795d-06*zm09
     $  + 4.425650820526436d-07*zm10
     $  + 4.568117079683221d-09*zm11
     $  - 2.916916714390665d-11*zm12
     $  + 4.707671259240802d-12*zm13
     $  - 3.362622328029144d-13*zm14

      CbH3YZ(0,3,3,15) = 
     $  - 8.371104729393133d-11*zm02
     $  - 2.318352494269138d-08*zm03
     $  - 9.055957208373685d-07*zm04
     $  - 1.002320721155839d-05*zm05
     $  - 4.112882823446331d-05*zm06
     $  - 7.040163903279175d-05*zm07
     $  - 5.221516681456703d-05*zm08
     $  - 1.637883734383870d-05*zm09
     $  - 1.970913882733465d-06*zm10
     $  - 7.263281528285899d-08*zm11
     $  - 4.732895793014086d-10*zm12
     $  - 7.664443077635170d-14*zm13

      CbH3YZ(0,3,3,16) = 
     $  + 7.848150197652238d-12*zm02
     $  + 3.311191261316869d-09*zm03
     $  + 1.825555174452471d-07*zm04
     $  + 2.799260441050241d-06*zm05
     $  + 1.603276810561415d-05*zm06
     $  + 3.933995499156919d-05*zm07
     $  + 4.385048603825917d-05*zm08
     $  + 2.231672305756966d-05*zm09
     $  + 4.949689771617795d-06*zm10
     $  + 4.253721018402469d-07*zm11
     $  + 1.102826978203077d-08*zm12
     $  + 4.468418217030539d-11*zm13
     $  + 5.068112985086256d-13*zm14
     $  - 6.706387692930774d-14*zm15
     $  + 4.191492308081733d-15*zm16

      CbH3YZ(0,3,3,17) = 
     $  - 6.924944077566464d-13*zm02
     $  - 4.430931306180902d-10*zm03
     $  - 3.407701389644297d-08*zm04
     $  - 7.109907501597341d-07*zm05
     $  - 5.542670403647779d-06*zm06
     $  - 1.882624057655109d-05*zm07
     $  - 3.001189105785680d-05*zm08
     $  - 2.301840754621680d-05*zm09
     $  - 8.355348865916141d-06*zm10
     $  - 1.347396975384121d-06*zm11
     $  - 8.439607545631805d-08*zm12
     $  - 1.558340913661706d-09*zm13
     $  - 4.382923749637981d-12*zm14
     $  - 1.796353846320743d-16*zm15

      CbH3YZ(0,3,3,18) = 
     $  + 5.770830759585934d-14*zm02
     $  + 5.581944328758430d-11*zm03
     $  + 5.933249606073168d-09*zm04
     $  + 1.660361782703899d-07*zm05
     $  + 1.727111757963180d-06*zm06
     $  + 7.905113960720671d-06*zm07
     $  + 1.737097061552658d-05*zm08
     $  + 1.905801113382750d-05*zm09
     $  + 1.047312476051411d-05*zm10
     $  + 2.792605193580257d-06*zm11
     $  + 3.346864375571559d-07*zm12
     $  + 1.552687290561347d-08*zm13
     $  + 2.063915554191699d-10*zm14
     $  + 4.279034031815357d-13*zm15
     $  - 8.113839737179185d-15*zm16
     $  + 9.555017444326658d-16*zm17
     $  - 5.308343024625921d-17*zm18

      CbH3YZ(0,3,3,19) = 
     $  - 4.555936400310164d-15*zm02
     $  - 6.646644288137327d-12*zm03
     $  - 9.693199120828665d-10*zm04
     $  - 3.596031464864652d-08*zm05
     $  - 4.911969863004257d-07*zm06
     $  - 2.965875406041737d-06*zm07
     $  - 8.732492227987250d-06*zm08
     $  - 1.318456017990898d-05*zm09
     $  - 1.038434082102210d-05*zm10
     $  - 4.217027735735639d-06*zm11
     $  - 8.453778135944247d-07*zm12
     $  - 7.663120558554645d-08*zm13
     $  - 2.667431683000058d-09*zm14
     $  - 2.578747938058875d-11*zm15
     $  - 3.188561098429729d-14*zm16
     $  - 3.302121041030777d-19*zm17

      CbH3YZ(0,3,3,20) = 
     $  + 3.416958817576782d-16*zm02
     $  + 7.506518801394275d-13*zm03
     $  + 1.493229350924626d-10*zm04
     $  + 7.274133435989196d-09*zm05
     $  + 1.287774337358699d-07*zm06
     $  + 1.008290988908498d-06*zm07
     $  + 3.889518515549604d-06*zm08
     $  + 7.844864381619160d-06*zm09
     $  + 8.504076366787152d-06*zm10
     $  + 4.965670720342399d-06*zm11
     $  + 1.526785864140904d-06*zm12
     $  + 2.344385978021634d-07*zm13
     $  + 1.630535403588125d-08*zm14
     $  + 4.304123165290116d-10*zm15
     $  + 3.047960337293150d-12*zm16
     $  + 1.729517799191281d-15*zm17
     $  + 1.293470915694780d-16*zm18
     $  - 1.361411125064822d-17*zm19
     $  + 6.807055625324113d-19*zm20

      CbH3YZ(0,3,3,21) = 
     $  - 2.440687197320616d-17*zm02
     $  - 8.064664835382025d-14*zm03
     $  - 2.178084363722016d-11*zm04
     $  - 1.382288996347774d-09*zm05
     $  - 3.137365791838989d-08*zm06
     $  - 3.140703207874765d-07*zm07
     $  - 1.558763064612109d-06*zm08
     $  - 4.102264327901544d-06*zm09
     $  - 5.933419773649997d-06*zm10
     $  - 4.775238399907954d-06*zm11
     $  - 2.120145056223315d-06*zm12
     $  - 5.032972894234168d-07*zm13
     $  - 6.010518210772960d-08*zm14
     $  - 3.245665643106863d-09*zm15
     $  - 6.554876059019661d-11*zm16
     $  - 3.411252913489387d-13*zm17
     $  - 1.871095757783406d-16*zm18
     $  - 4.888008119946874d-22*zm19

      CbH3YZ(3,0,3,02) = 
     $  + 2.812500000000000d-01*zm02

      CbH3YZ(3,0,3,03) = 
     $  - 2.109375000000000d-01*zm02
     $  + 3.515625000000000d-02*zm03

      CbH3YZ(3,0,3,04) = 
     $  + 9.228515625000000d-02*zm02
     $  + 1.318359375000000d-02*zm03
     $  + 2.197265625000000d-03*zm04

      CbH3YZ(3,0,3,05) = 
     $  - 2.966308593750000d-02*zm02
     $  - 3.460693359375000d-02*zm03
     $  - 3.295898437500000d-03*zm04

      CbH3YZ(3,0,3,06) = 
     $  + 7.662963867187500d-03*zm02
     $  + 2.496643066406250d-02*zm03
     $  + 1.034088134765625d-02*zm04
     $  + 4.531860351562500d-04*zm05
     $  - 8.239746093750000d-06*zm06

      CbH3YZ(3,0,3,07) = 
     $  - 1.668548583984375d-03*zm02
     $  - 1.158714294433593d-02*zm03
     $  - 1.266860961914062d-02*zm04
     $  - 2.471923828125000d-03*zm05
     $  - 4.325866699218750d-05*zm06

      CbH3YZ(3,0,3,08) = 
     $  + 3.153358186994280d-04*zm02
     $  + 4.104331561497279d-03*zm03
     $  + 9.240332671574183d-03*zm04
     $  + 4.764798709324428d-03*zm05
     $  + 4.903016771589006d-04*zm06
     $  + 3.641673496791294d-06*zm07
     $  + 5.911807624661192d-08*zm08

      CbH3YZ(3,0,3,09) = 
     $  - 5.276288305010114d-05*zm02
     $  - 1.195441399301801d-03*zm03
     $  - 4.849012408937726d-03*zm04
     $  - 5.116965089525495d-03*zm05
     $  - 1.444420133318219d-03*zm06
     $  - 8.394471236637660d-05*zm07
     $  - 3.428848422303491d-07*zm08

      CbH3YZ(3,0,3,10) = 
     $  + 7.929950952529907d-06*zm02
     $  + 2.985292885984693d-04*zm03
     $  + 2.010532245039939d-03*zm04
     $  + 3.771017566323280d-03*zm05
     $  + 2.199779257178306d-03*zm06
     $  + 3.703685402870178d-04*zm07
     $  + 1.278657481378438d-05*zm08
     $  + 3.037191167169687d-08*zm09
     $  - 5.172831671578543d-10*zm10

      CbH3YZ(3,0,3,11) = 
     $  - 1.082415027277810d-06*zm02
     $  - 6.564905334796224d-05*zm03
     $  - 6.952634932739394d-04*zm04
     $  - 2.123605172548975d-03*zm05
     $  - 2.199887498681034d-03*zm06
     $  - 7.802633339805262d-04*zm07
     $  - 8.290733792343918d-05*zm08
     $  - 1.762790187281005d-06*zm09
     $  - 1.718119090917159d-09*zm10

      CbH3YZ(3,0,3,12) = 
     $  + 1.353680083600731d-07*zm02
     $  + 1.295122012568565d-05*zm03
     $  + 2.075954597570340d-04*zm04
     $  + 9.734197312893418d-04*zm05
     $  + 1.630779009518007d-03*zm06
     $  + 1.033803068307022d-03*zm07
     $  + 2.372676085418086d-04*zm08
     $  + 1.655809940158602d-05*zm09
     $  + 2.217312042769010d-07*zm10
     $  + 5.227415122658511d-11*zm11
     $  + 5.009844723248847d-12*zm12

      CbH3YZ(3,0,3,13) = 
     $  - 1.562320076945153d-08*zm02
     $  - 2.323517136656763d-06*zm03
     $  - 5.481255330695686d-05*zm04
     $  - 3.783013407056304d-04*zm05
     $  - 9.623325535477223d-04*zm06
     $  - 9.864892854503440d-04*zm07
     $  - 4.091782581518146d-04*zm08
     $  - 6.349131147937023d-05*zm09
     $  - 2.996329628302610d-06*zm10
     $  - 2.561945129956545d-08*zm11
     $  - 6.484170456090651d-12*zm12

      CbH3YZ(3,0,3,14) = 
     $  + 1.674118753304258d-09*zm02
     $  + 3.830449519578035d-07*zm03
     $  + 1.302134686168027d-05*zm04
     $  + 1.282011041944849d-04*zm05
     $  + 4.733639938016453d-04*zm06
     $  + 7.337924794366992d-04*zm07
     $  + 4.942086765396339d-04*zm08
     $  + 1.405869722513460d-04*zm09
     $  + 1.523416401770327d-05*zm10
     $  + 4.969427416554460d-07*zm11
     $  + 2.745176651741661d-09*zm12
     $  + 1.019890740039256d-12*zm13
     $  - 5.173265120044837d-14*zm14

      CbH3YZ(3,0,3,15) = 
     $  - 1.674220945878626d-10*zm02
     $  - 5.850293416413902d-08*zm03
     $  - 2.819817782415569d-06*zm04
     $  - 3.866387635346334d-05*zm05
     $  - 2.003589658748936d-04*zm06
     $  - 4.477484480186310d-04*zm07
     $  - 4.560317965416921d-04*zm08
     $  - 2.121191938414644d-04*zm09
     $  - 4.284948093564562d-05*zm10
     $  - 3.323855184226189d-06*zm11
     $  - 7.621635449661987d-08*zm12
     $  - 2.751857037839101d-10*zm13
     $  - 1.478168116164896d-14*zm14

      CbH3YZ(3,0,3,16) = 
     $  + 1.569630039530447d-11*zm02
     $  + 8.332915204675738d-09*zm03
     $  + 5.623312596097449d-07*zm04
     $  + 1.053745305482134d-05*zm05
     $  + 7.467343365795802d-05*zm06
     $  + 2.321379738882429d-04*zm07
     $  + 3.399393470033780d-04*zm08
     $  + 2.397763338408196d-04*zm09
     $  + 7.991982508767041d-05*zm10
     $  + 1.177453891689528d-05*zm11
     $  + 6.666308459714675d-07*zm12
     $  + 1.088797455233583d-08*zm13
     $  + 2.586728061416322d-11*zm14
     $  - 7.690093311711172d-15*zm15
     $  + 5.588656410775645d-16*zm16

      CbH3YZ(3,0,3,17) = 
     $  - 1.384988815513292d-12*zm02
     $  - 1.112926534603400d-09*zm03
     $  - 1.041098745237188d-07*zm04
     $  - 2.626312489703853d-06*zm05
     $  - 2.493382924142766d-05*zm06
     $  - 1.048976296686497d-04*zm07
     $  - 2.127142723713524d-04*zm08
     $  - 2.157325093961131d-04*zm09
     $  - 1.095508616673982d-04*zm10
     $  - 2.691608769640573d-05*zm11
     $  - 2.953658111731694d-06*zm12
     $  - 1.239576030160045d-07*zm13
     $  - 1.457262978214756d-09*zm14
     $  - 2.277328647463101d-12*zm15
     $  - 5.522457112631734d-17*zm16

      CbH3YZ(3,0,3,18) = 
     $  + 1.154166151917186d-13*zm02
     $  + 1.400102746797276d-10*zm03
     $  + 1.801301687765522d-08*zm04
     $  + 6.042850739892346d-07*zm05
     $  + 7.559716810104070d-06*zm06
     $  + 4.211172455075303d-05*zm07
     $  + 1.148763955721773d-04*zm08
     $  + 1.610397367924791d-04*zm09
     $  + 1.178093322232550d-04*zm10
     $  + 4.436947206843856d-05*zm11
     $  + 8.217171296364352d-06*zm12
     $  + 6.830944453473255d-07*zm13
     $  + 2.151881954153405d-08*zm14
     $  + 1.836046784139357d-10*zm15
     $  + 1.889653090682079d-13*zm16
     $  + 1.084678842894483d-16*zm17
     $  - 6.245109440736378d-18*zm18

      CbH3YZ(3,0,3,19) = 
     $  - 9.111872800620328d-15*zm02
     $  - 1.665533617974595d-11*zm03
     $  - 2.928518168491773d-09*zm04
     $  - 1.293451540517116d-07*zm05
     $  - 2.103496931063851d-06*zm06
     $  - 1.524379555846259d-05*zm07
     $  - 5.467047583989030d-05*zm08
     $  - 1.027696514218858d-04*zm09
     $  - 1.039230903809186d-04*zm10
     $  - 5.657048786269407d-05*zm11
     $  - 1.617639149954473d-05*zm12
     $  - 2.299232275788936d-06*zm13
     $  - 1.468135938132518d-07*zm14
     $  - 3.507517652000406d-09*zm15
     $  - 2.186352414686510d-11*zm16
     $  - 1.499111891258455d-14*zm17
     $  + 1.556875912028877d-19*zm18

      CbH3YZ(3,0,3,20) = 
     $  + 6.833917635153565d-16*zm02
     $  + 1.879704963284388d-12*zm03
     $  + 4.494361251517627d-10*zm04
     $  + 2.591885638799359d-08*zm05
     $  + 5.418062619751314d-07*zm06
     $  + 5.034309327220625d-06*zm07
     $  + 2.330044091933259d-05*zm08
     $  + 5.734097027300858d-05*zm09
     $  + 7.765204640813050d-05*zm10
     $  + 5.851024822129756d-05*zm11
     $  + 2.428727746966162d-05*zm12
     $  + 5.373740901738643d-06*zm13
     $  + 5.949273293642146d-07*zm14
     $  + 2.951616163976461d-08*zm15
     $  + 5.393702090987256d-10*zm16
     $  + 2.469247759757751d-12*zm17
     $  + 1.136482181179779d-15*zm18
     $  - 1.367249409734930d-18*zm19
     $  + 7.165321710867487d-20*zm20

      CbH3YZ(3,0,3,21) = 
     $  - 4.881374394641232d-17*zm02
     $  - 2.018484861682960d-13*zm03
     $  - 6.536420962918742d-11*zm04
     $  - 4.888041896663958d-09*zm05
     $  - 1.301057595894852d-07*zm06
     $  - 1.531345266480479d-06*zm07
     $  - 9.007708717671656d-06*zm08
     $  - 2.846250681423187d-05*zm09
     $  - 5.034303111251490d-05*zm10
     $  - 5.080103682757355d-05*zm11
     $  - 2.925907288196828d-05*zm12
     $  - 9.446199714131830d-06*zm13
     $  - 1.642550393751867d-06*zm14
     $  - 1.433979456136016d-07*zm15
     $  - 5.581255665650210d-09*zm16
     $  - 7.856544307928392d-11*zm17
     $  - 2.653009908436887d-13*zm18
     $  - 8.024954554675037d-17*zm19
     $  - 2.350533301377378d-21*zm20

      CbH3YZ(3,3,3,03) = 
     $  + 7.031250000000000d-02*zm03

      CbH3YZ(3,3,3,04) = 
     $  - 7.910156250000000d-02*zm03

      CbH3YZ(3,3,3,05) = 
     $  + 4.943847656250000d-02*zm03
     $  + 1.977539062500000d-02*zm04

      CbH3YZ(3,3,3,06) = 
     $  - 2.224731445312500d-02*zm03
     $  - 2.966308593750000d-02*zm04
     $  - 3.707885742187500d-03*zm05

      CbH3YZ(3,3,3,07) = 
     $  + 7.971954345703125d-03*zm03
     $  + 2.373046875000000d-02*zm04
     $  + 1.075286865234375d-02*zm05
     $  + 5.561828613281250d-04*zm06

      CbH3YZ(3,3,3,08) = 
     $  - 2.398538589477539d-03*zm03
     $  - 1.334838867187500d-02*zm04
     $  - 1.386981010437011d-02*zm05
     $  - 2.919960021972656d-03*zm06
     $  - 6.952285766601562d-05*zm07

      CbH3YZ(3,3,3,09) = 
     $  + 6.259126322610037d-04*zm03
     $  + 5.890820707593645d-03*zm04
     $  + 1.149610110691615d-02*zm05
     $  + 5.774949278150285d-03*zm06
     $  + 6.461901324135916d-04*zm07
     $  + 7.448877607073102d-06*zm08

      CbH3YZ(3,3,3,10) = 
     $  - 1.447875584874834d-04*zm03
     $  - 2.162036725452968d-03*zm04
     $  - 7.084580936602183d-03*zm05
     $  - 6.783134170940944d-03*zm06
     $  - 1.901326009205409d-03*zm07
     $  - 1.219753708158220d-04*zm08
     $  - 6.983322756631033d-07*zm09

      CbH3YZ(3,3,3,11) = 
     $  + 3.015631543738501d-05*zm03
     $  + 6.840397417545318d-04*zm04
     $  + 3.504440858960151d-03*zm05
     $  + 5.649845637381076d-03*zm06
     $  + 3.082008026540279d-03*zm07
     $  + 5.235397070646286d-04*zm08
     $  + 2.020508050918579d-05*zm09
     $  + 5.819435630525861d-08*zm10

      CbH3YZ(3,3,3,12) = 
     $  - 5.721960083714553d-06*zm03
     $  - 1.911335438489913d-04*zm04
     $  - 1.457628958991595d-03*zm05
     $  - 3.659147645479866d-03*zm06
     $  - 3.389678678608366d-03*zm07
     $  - 1.149053384682961d-03*zm08
     $  - 1.248356234282255d-04*zm09
     $  - 2.994099631905555d-06*zm10
     $  - 4.364576722894396d-09*zm11

      CbH3YZ(3,3,3,13) = 
     $  + 9.981489380482922d-07*zm03
     $  + 4.799100094136189d-05*zm04
     $  + 5.257866705175150d-04*zm05
     $  + 1.947831235487352d-03*zm06
     $  + 2.814309920544143d-03*zm07
     $  + 1.639820873864333d-03*zm08
     $  + 3.657354598006161d-04*zm09
     $  + 2.638961959557680d-05*zm10
     $  + 4.023842153732072d-07*zm11
     $  + 2.975847765609815d-10*zm12

      CbH3YZ(3,3,3,14) = 
     $  - 1.612537507989818d-07*zm03
     $  - 1.096971882597918d-05*zm04
     $  - 1.681098250652188d-04*zm05
     $  - 8.839941778229307d-04*zm06
     $  - 1.879960579522462d-03*zm07
     $  - 1.717442837385037d-03*zm08
     $  - 6.681624490502381d-04*zm09
     $  - 1.020938972186587d-04*zm10
     $  - 5.029666299142499d-06*zm11
     $  - 4.954786529740342d-08*zm12
     $  - 1.859904853506134d-11*zm13

      CbH3YZ(3,3,3,15) = 
     $  + 2.427176855751249d-08*zm03
     $  + 2.305816919503141d-06*zm04
     $  + 4.841992667346562d-05*zm05
     $  + 3.509549068257781d-04*zm06
     $  + 1.053536262087364d-03*zm07
     $  + 1.418101264721785d-03*zm08
     $  + 8.716207520146837d-04*zm09
     $  + 2.362254028323799d-04*zm10
     $  + 2.548182891960839d-05*zm11
     $  + 8.752041448541715d-07*zm12
     $  + 5.634591972954271d-09*zm13
     $  + 1.073022030868923d-12*zm14

      CbH3YZ(3,3,3,16) = 
     $  - 3.421065364084000d-09*zm03
     $  - 4.493047368349504d-07*zm04
     $  - 1.272019709364958d-05*zm05
     $  - 1.242190451567028d-04*zm06
     $  - 5.100165486359023d-04*zm07
     $  - 9.673158347724803d-04*zm08
     $  - 8.790107844357902d-04*zm09
     $  - 3.799169176972696d-04*zm10
     $  - 7.403481071099808d-05*zm11
     $  - 5.769529490552925d-06*zm12
     $  - 1.403940406075299d-07*zm13
     $  - 5.955502204614856d-10*zm14
     $  - 5.748332308226377d-14*zm15

      CbH3YZ(3,3,3,17) = 
     $  + 4.534805467344399d-10*zm03
     $  + 8.169424610115933d-08*zm04
     $  + 3.077907788225533d-06*zm05
     $  + 3.976867973757162d-05*zm06
     $  + 2.179100631839815d-04*zm07
     $  + 5.631345433191215d-04*zm08
     $  + 7.215214950587713d-04*zm09
     $  + 4.634073116403277d-04*zm10
     $  + 1.455705169025189d-04*zm11
     $  + 2.090230761139301d-05*zm12
     $  + 1.198342951618905d-06*zm13
     $  + 2.092119698847299d-08*zm14
     $  + 5.881370274084892d-11*zm15
     $  + 2.874166154113188d-15*zm16

      CbH3YZ(3,3,3,18) = 
     $  - 5.674277620911805d-11*zm03
     $  - 1.393605745278708d-08*zm04
     $  - 6.914198124940485d-07*zm05
     $  - 1.164770528398285d-05*zm06
     $  - 8.352557628437343d-05*zm07
     $  - 2.865387515361988d-04*zm08
     $  - 4.995942449845584d-04*zm09
     $  - 4.534799068462546d-04*zm10
     $  - 2.130344840182874d-04*zm11
     $  - 4.989831673093617d-05*zm12
     $  - 5.382797132741308d-06*zm13
     $  - 2.303544738947265d-07*zm14
     $  - 2.914188378903061d-09*zm15
     $  - 5.451035746660294d-12*zm16
     $  - 1.347265384740557d-16*zm17

      CbH3YZ(3,3,3,19) = 
     $  + 6.724095206942600d-12*zm03
     $  + 2.240674465147476d-09*zm04
     $  + 1.451358219822184d-07*zm05
     $  + 3.149523775743029d-06*zm06
     $  + 2.909102519461983d-05*zm07
     $  + 1.297560193515118d-04*zm08
     $  + 2.995128844229666d-04*zm09
     $  + 3.701202268296033d-04*zm10
     $  + 2.466107339518731d-04*zm11
     $  + 8.702071753076612d-05*zm12
     $  + 1.551042381777638d-05*zm13
     $  + 1.276930905447972d-06*zm14
     $  + 4.127583819341004d-08*zm15
     $  + 3.813855190018480d-10*zm16
     $  + 4.759835770351732d-13*zm17
     $  + 5.943817873855399d-18*zm18

      CbH3YZ(3,3,3,20) = 
     $  - 7.568024060110658d-13*zm03
     $  - 3.408862787881577d-10*zm04
     $  - 2.862273198169673d-08*zm05
     $  - 7.921269575300142d-07*zm06
     $  - 9.301185441370569d-06*zm07
     $  - 5.303576796793990d-05*zm08
     $  - 1.585643480763648d-04*zm09
     $  - 2.591643588244777d-04*zm10
     $  - 2.354808421854879d-04*zm11
     $  - 1.183867820424926d-04*zm12
     $  - 3.204979829983027d-05*zm13
     $  - 4.419503473151362d-06*zm14
     $  - 2.812846266456108d-07*zm15
     $  - 6.934855938126708d-09*zm16
     $  - 4.709656369174672d-11*zm17
     $  - 3.929012210065265d-14*zm18
     $  - 2.476590780773083d-19*zm19

      CbH3YZ(3,3,3,21) = 
     $  + 8.111037892131116d-14*zm03
     $  + 4.923817365075389d-11*zm04
     $  + 5.327781899693800d-09*zm05
     $  + 1.864593771088200d-07*zm06
     $  + 2.752851860677472d-06*zm07
     $  + 1.978794555271422d-05*zm08
     $  + 7.527615283152749d-05*zm09
     $  + 1.590403363823556d-04*zm10
     $  + 1.911881115768152d-04*zm11
     $  + 1.314154323489621d-04*zm12
     $  + 5.095550361037525d-05*zm13
     $  + 1.076727233598165d-05*zm14
     $  + 1.164486977846779d-06*zm15
     $  + 5.791675490004658d-08*zm16
     $  + 1.097889081546724d-09*zm17
     $  + 5.507791300886098d-12*zm18
     $  + 3.075062527486186d-15*zm19
     $  + 9.776016239893749d-21*zm20
      endif

      return
      end


      subroutine fillcoeff2dhpl320e(iflag,n,CbH1YZ,CbH2YZ,CbH3YZ,
     $                              CbH4YZ)
      implicit none
      include 'src/Inc/types.f'
      real(dp):: CbH1YZ,CbH2YZ,CbH3YZ,CbH4YZ
      real(dp):: 
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      integer iflag,n
      dimension CbH1YZ(3:3,22),CbH2YZ(0:3,0:3,22),
     $          CbH3YZ(0:3,0:3,0:3,22),CbH4YZ(0:3,0:3,0:3,0:3,22)
      common /zm/  
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3) 
!$omp threadprivate(/zm/,/HPL2/)

* 2001-04-20:13:08:47.chpl 
* <- HPLexpand.320e 
* produced by form-to-fortr for gehrt@pcth62 

      if (n.eq.2) then
      CbH2YZ(3,2,01) = 
     $  + HZ1(1)*CbH1YZ(3,1)

      CbH2YZ(3,2,02) = 
     $  + 2.812500000000000d-01*zm01
     $  - 2.812500000000000d-01*zm02
     $  + HZ1(1)*CbH1YZ(3,2)

      CbH2YZ(3,2,03) = 
     $  - 1.406250000000000d-01*zm01
     $  + 1.406250000000000d-01*zm02
     $  + HZ1(1)*CbH1YZ(3,3)

      CbH2YZ(3,2,04) = 
     $  + 3.955078125000000d-02*zm01
     $  - 3.955078125000000d-02*zm03
     $  + HZ1(1)*CbH1YZ(3,4)

      CbH2YZ(3,2,05) = 
     $  - 7.910156250000000d-03*zm01
     $  - 2.373046875000000d-02*zm02
     $  + 2.373046875000000d-02*zm03
     $  + 7.910156250000000d-03*zm04
     $  + HZ1(1)*CbH1YZ(3,5)

      CbH2YZ(3,2,06) = 
     $  + 1.235961914062500d-03*zm01
     $  + 1.235961914062500d-02*zm02
     $  - 1.235961914062500d-02*zm04
     $  - 1.235961914062500d-03*zm05
     $  + HZ1(1)*CbH1YZ(3,6)

      CbH2YZ(3,2,07) = 
     $  - 1.589093889508928d-04*zm01
     $  - 3.972734723772321d-03*zm02
     $  - 6.356375558035714d-03*zm03
     $  + 6.356375558035714d-03*zm04
     $  + 3.972734723772321d-03*zm05
     $  + 1.589093889508928d-04*zm06
     $  + HZ1(1)*CbH1YZ(3,7)

      CbH2YZ(3,2,08) = 
     $  + 1.738071441650390d-05*zm01
     $  + 9.733200073242187d-04*zm02
     $  + 4.258275032043457d-03*zm03
     $  - 4.258275032043457d-03*zm05
     $  - 9.733200073242187d-04*zm06
     $  - 1.738071441650390d-05*zm07
     $  + HZ1(1)*CbH1YZ(3,8)

      CbH2YZ(3,2,09) = 
     $  - 1.655306134905133d-06*zm01
     $  - 1.969814300537109d-04*zm02
     $  - 1.772832870483398d-03*zm03
     $  - 2.027750015258789d-03*zm04
     $  + 2.027750015258789d-03*zm05
     $  + 1.772832870483398d-03*zm06
     $  + 1.969814300537109d-04*zm07
     $  + 1.655306134905133d-06*zm08
     $  + HZ1(1)*CbH1YZ(3,9)

      CbH2YZ(3,2,10) = 
     $  + 1.396664551326206d-07*zm01
     $  + 3.435794796262468d-05*zm02
     $  + 5.650904774665832d-04*zm03
     $  + 1.581862270832061d-03*zm04
     $  - 1.581862270832061d-03*zm06
     $  - 5.650904774665832d-04*zm07
     $  - 3.435794796262468d-05*zm08
     $  - 1.396664551326206d-07*zm09
     $  + HZ1(1)*CbH1YZ(3,10)

      CbH2YZ(3,2,11) = 
     $  - 1.058079205550156d-08*zm01
     $  - 5.300976819806284d-06*zm02
     $  - 1.492526527349050d-04*zm03
     $  - 7.790213958783583d-04*zm04
     $  - 7.190283049236644d-04*zm05
     $  + 7.190283049236644d-04*zm06
     $  + 7.790213958783583d-04*zm07
     $  + 1.492526527349050d-04*zm08
     $  + 5.300976819806284d-06*zm09
     $  + 1.058079205550156d-08*zm10
     $  + HZ1(1)*CbH1YZ(3,11)

      CbH2YZ(3,2,12) = 
     $  + 7.274294538157326d-10*zm01
     $  + 7.361586072615214d-07*zm02
     $  + 3.406333903382931d-05*zm03
     $  + 2.963198428707463d-04*zm04
     $  + 6.220700265839695d-04*zm05
     $  - 6.220700265839695d-04*zm07
     $  - 2.963198428707463d-04*zm08
     $  - 3.406333903382931d-05*zm09
     $  - 7.361586072615214d-07*zm10
     $  - 7.274294538157326d-10*zm11
     $  + HZ1(1)*CbH1YZ(3,12)

      CbH2YZ(3,2,13) = 
     $  - 4.578227331707408d-11*zm01
     $  - 9.316692620024576d-08*zm02
     $  - 6.894856143824674d-06*zm03
     $  - 9.389262101459470d-05*zm04
     $  - 3.449523068739326d-04*zm05
     $  - 2.740588229006299d-04*zm06
     $  + 2.740588229006299d-04*zm07
     $  + 3.449523068739326d-04*zm08
     $  + 9.389262101459470d-05*zm09
     $  + 6.894856143824674d-06*zm10
     $  + 9.316692620024576d-08*zm11
     $  + 4.578227331707408d-11*zm12
     $  + HZ1(1)*CbH1YZ(3,13)

      CbH2YZ(3,2,14) = 
     $  + 2.657006933580192d-12*zm01
     $  + 1.084590230287434d-08*zm02
     $  + 1.259920803820524d-06*zm03
     $  + 2.579798031900059d-05*zm04
     $  + 1.491398955603267d-04*zm05
     $  + 2.555876254515095d-04*zm06
     $  - 2.555876254515095d-04*zm08
     $  - 1.491398955603267d-04*zm09
     $  - 2.579798031900059d-05*zm10
     $  - 1.259920803820524d-06*zm11
     $  - 1.084590230287434d-08*zm12
     $  - 2.657006933580192d-12*zm13
     $  + HZ1(1)*CbH1YZ(3,14)

      CbH2YZ(3,2,15) = 
     $  - 1.430696041158565d-13*zm01
     $  - 1.169880152855358d-09*zm02
     $  - 2.105337897974804d-07*zm03
     $  - 6.302748852509206d-06*zm04
     $  - 5.404426897647291d-05*zm05
     $  - 1.548499514558768d-04*zm06
     $  - 1.100992848098810d-04*zm07
     $  + 1.100992848098810d-04*zm08
     $  + 1.548499514558768d-04*zm09
     $  + 5.404426897647291d-05*zm10
     $  + 6.302748852509206d-06*zm11
     $  + 2.105337897974804d-07*zm12
     $  + 1.169880152855358d-09*zm13
     $  + 1.430696041158565d-13*zm14
     $  + HZ1(1)*CbH1YZ(3,15)

      CbH2YZ(3,2,16) = 
     $  + 7.185415385282972d-15*zm01
     $  + 1.176108790263116d-10*zm02
     $  + 3.248486775901812d-08*zm03
     $  + 1.393061428507027d-06*zm04
     $  + 1.705337187689560d-05*zm05
     $  + 7.380514278180754d-05*zm06
     $  + 1.087003404628824d-04*zm07
     $  - 1.087003404628824d-04*zm09
     $  - 7.380514278180754d-05*zm10
     $  - 1.705337187689560d-05*zm11
     $  - 1.393061428507027d-06*zm12
     $  - 3.248486775901812d-08*zm13
     $  - 1.176108790263116d-10*zm14
     $  - 7.185415385282972d-15*zm15
     $  + HZ1(1)*CbH1YZ(3,16)

      CbH2YZ(3,2,17) = 
     $  - 3.381371946015516d-16*zm01
     $  - 1.107433126039541d-11*zm02
     $  - 4.663583792161070d-09*zm03
     $  - 2.820964232891927d-07*zm04
     $  - 4.799228071838228d-06*zm05
     $  - 2.966672693349161d-05*zm06
     $  - 7.053900287332951d-05*zm07
     $  - 4.603779125486784d-05*zm08
     $  + 4.603779125486784d-05*zm09
     $  + 7.053900287332951d-05*zm10
     $  + 2.966672693349161d-05*zm11
     $  + 4.799228071838228d-06*zm12
     $  + 2.820964232891927d-07*zm13
     $  + 4.663583792161070d-09*zm14
     $  + 1.107433126039541d-11*zm15
     $  + 3.381371946015516d-16*zm16
     $  + HZ1(1)*CbH1YZ(3,17)
      endif
      if (n.eq.2) return

      if (n.eq.3) then
      CbH3YZ(0,3,2,01) = 
     $  + HZ1(1)*CbH2YZ(0,3,1)

      CbH3YZ(0,3,2,02) = 
     $  + 1.406250000000000d-01*zm01
     $  - 1.406250000000000d-01*zm02
     $  + HZ1(1)*CbH2YZ(0,3,2)

      CbH3YZ(0,3,2,03) = 
     $  - 8.203125000000000d-02*zm01
     $  + 1.171875000000000d-01*zm02
     $  - 3.515625000000000d-02*zm03
     $  + HZ1(1)*CbH2YZ(0,3,3)

      CbH3YZ(0,3,2,04) = 
     $  + 2.636718750000000d-02*zm01
     $  - 3.625488281250000d-02*zm02
     $  + 1.318359375000000d-02*zm03
     $  - 3.295898437500000d-03*zm04
     $  + HZ1(1)*CbH2YZ(0,3,4)

      CbH3YZ(0,3,2,05) = 
     $  - 5.866699218750000d-03*zm01
     $  + 2.175292968750000d-03*zm02
     $  + 3.757324218750000d-03*zm03
     $  - 6.591796875000000d-05*zm04
     $  + HZ1(1)*CbH2YZ(0,3,5)

      CbH3YZ(0,3,2,06) = 
     $  + 9.887695312500000d-04*zm01
     $  + 2.533721923828125d-03*zm02
     $  - 3.172302246093750d-03*zm03
     $  - 2.471923828125000d-04*zm04
     $  - 1.235961914062500d-04*zm05
     $  + 2.059936523437500d-05*zm06
     $  + HZ1(1)*CbH2YZ(0,3,6)

      CbH3YZ(0,3,2,07) = 
     $  - 1.330550836057079d-04*zm01
     $  - 1.260555033781090d-03*zm02
     $  + 2.661101672114158d-04*zm03
     $  + 1.058134740712691d-03*zm04
     $  + 6.873461664939413d-05*zm05
     $  + 6.305928132971938d-07*zm06
     $  + HZ1(1)*CbH2YZ(0,3,7)

      CbH3YZ(0,3,2,08) = 
     $  + 1.489775521414620d-05*zm01
     $  + 3.656157425471714d-04*zm02
     $  + 4.999024527413504d-04*zm03
     $  - 6.031521729060581d-04*zm04
     $  - 2.648489815848214d-04*zm05
     $  - 1.386318887983049d-05*zm06
     $  + 1.655306134905133d-06*zm07
     $  - 2.069132668631417d-07*zm08
     $  + HZ1(1)*CbH2YZ(0,3,8)

      CbH3YZ(0,3,2,09) = 
     $  - 1.441495759146554d-06*zm01
     $  - 8.049615791865757d-05*zm02
     $  - 3.318130118506295d-04*zm03
     $  + 4.362421376364571d-05*zm04
     $  + 3.112320389066423d-04*zm05
     $  + 5.806675979069301d-05*zm06
     $  + 8.345501763480050d-07*zm07
     $  - 6.897108895438058d-09*zm08
     $  + HZ1(1)*CbH2YZ(0,3,9)

      CbH3YZ(0,3,2,10) = 
     $  + 1.241479601178850d-07*zm01
     $  + 1.471541289772306d-05*zm02
     $  + 1.286095274346215d-04*zm03
     $  + 1.297578960657119d-04*zm04
     $  - 1.523745507001876d-04*zm05
     $  - 1.101502776145935d-04*zm06
     $  - 1.070000231266021d-05*zm07
     $  + 3.879623753683907d-08*zm08
     $  - 2.327774252210344d-08*zm09
     $  + 2.327774252210344d-09*zm10
     $  + HZ1(1)*CbH2YZ(0,3,10)

      CbH3YZ(0,3,2,11) = 
     $  - 9.602870365523391d-09*zm01
     $  - 2.337248870550952d-06*zm02
     $  - 3.782383068393225d-05*zm03
     $  - 1.004632739207257d-04*zm04
     $  + 9.170180096465834d-06*zm05
     $  + 9.868406564208383d-05*zm06
     $  + 3.113792437307950d-05*zm07
     $  + 1.637521854116896d-06*zm08
     $  + 4.184222312857437d-09*zm09
     $  + 8.015751557198156d-11*zm10
     $  + HZ1(1)*CbH2YZ(0,3,11)

      CbH3YZ(0,3,2,12) = 
     $  + 6.612995034688478d-10*zm01
     $  + 3.309969339737450d-07*zm02
     $  + 9.226296369179889d-06*zm03
     $  + 4.676592707869826d-05*zm04
     $  + 3.909695246438314d-05*zm05
     $  - 4.495199907317080d-05*zm06
     $  - 4.282469776543703d-05*zm07
     $  - 7.419019934491484d-06*zm08
     $  - 2.232547123710830d-07*zm09
     $  - 2.165755873860476d-09*zm10
     $  + 3.306497517344239d-10*zm11
     $  - 2.755414597786866d-11*zm12
     $  + HZ1(1)*CbH2YZ(0,3,12)

      CbH3YZ(0,3,2,13) = 
     $  - 4.100140771824287d-11*zm01
     $  - 4.246273008784294d-08*zm02
     $  - 1.951899928197624d-06*zm03
     $  - 1.667151493333685d-05*zm04
     $  - 3.349074793114715d-05*zm05
     $  + 2.308470587097829d-06*zm06
     $  + 3.358018479274656d-05*zm07
     $  + 1.469746810153812d-05*zm08
     $  + 1.541332019668567d-06*zm09
     $  + 2.917798051102124d-08*zm10
     $  + 3.400829124453595d-11*zm11
     $  - 9.656761557417030d-13*zm12
     $  + HZ1(1)*CbH2YZ(0,3,13)

      CbH3YZ(0,3,2,14) = 
     $  + 2.452621784843254d-12*zm01
     $  + 4.989910117550785d-09*zm02
     $  + 3.678008515750676d-07*zm03
     $  + 4.947776800422495d-06*zm04
     $  + 1.770704961970097d-05*zm05
     $  + 1.298332365154544d-05*zm06
     $  - 1.468229968138238d-05*zm07
     $  - 1.678017906891584d-05*zm08
     $  - 4.258096788512225d-06*zm09
     $  - 2.868456552790648d-07*zm10
     $  - 3.547042940784935d-09*zm11
     $  + 2.932245600545935d-11*zm12
     $  - 4.707671259240802d-12*zm13
     $  + 3.362622328029144d-13*zm14
     $  + HZ1(1)*CbH2YZ(0,3,14)

      CbH3YZ(0,3,2,15) = 
     $  - 1.447728136886643d-13*zm01
     $  - 5.420583690130969d-10*zm02
     $  - 6.282599515369622d-08*zm03
     $  - 1.275576055587716d-06*zm04
     $  - 7.246775436473400d-06*zm05
     $  - 1.196931776594914d-05*zm06
     $  + 6.592877699721294d-07*zm07
     $  + 1.212009350260334d-05*zm08
     $  + 6.646931271390005d-06*zm09
     $  + 1.079679674407090d-06*zm10
     $  + 4.867307982871360d-08*zm11
     $  + 3.722500778137305d-10*zm12
     $  - 1.038957839412767d-13*zm13
     $  + 1.192246700965470d-14*zm14
     $  + HZ1(1)*CbH2YZ(0,3,15)

      CbH3YZ(0,3,2,16) = 
     $  + 6.706387692930774d-15*zm01
     $  + 5.483489872817156d-11*zm02
     $  + 9.851377901247586d-09*zm03
     $  + 2.931637562579174d-07*zm04
     $  + 2.483141118150705d-06*zm05
     $  + 6.954224840824548d-06*zm06
     $  + 4.624368288186444d-06*zm07
     $  - 5.160930947716175d-06*zm08
     $  - 6.722007836545493d-06*zm09
     $  - 2.228992040072892d-06*zm10
     $  - 2.452390539015048d-07*zm11
     $  - 7.598803709305995d-09*zm12
     $  - 3.509740096326085d-11*zm13
     $  - 5.064520277393614d-13*zm14
     $  + 6.706387692930774d-14*zm15
     $  - 4.191492308081733d-15*zm16
     $  + HZ1(1)*CbH2YZ(0,3,16)

      CbH3YZ(0,3,2,17) = 
     $  - 1.671209933612815d-16*zm01
     $  - 5.191035096083037d-12*zm02
     $  - 1.431484851183645d-09*zm03
     $  - 6.112417859079794d-08*zm04
     $  - 7.415972468731492d-07*zm05
     $  - 3.159363189025446d-06*zm06
     $  - 4.509835392283194d-06*zm07
     $  + 2.062735988018859d-07*zm08
     $  + 4.589329656913514d-06*zm09
     $  + 2.970341376941656d-06*zm10
     $  + 6.556085902793314d-07*zm11
     $  + 5.070119767953541d-08*zm12
     $  + 1.098742087209294d-09*zm13
     $  + 3.517635889888703d-12*zm14
     $  + 2.636848542166732d-15*zm15
     $  - 1.498826265776730d-16*zm16
     $  + HZ1(1)*CbH2YZ(0,3,17)

      CbH3YZ(3,0,2,01) = 
     $  + HZ2(0,1)*CbH1YZ(3,1)
     $  + HZ2(1,1)*CbH1YZ(3,1)

      CbH3YZ(3,0,2,02) = 
     $  + 2.812500000000000d-01*zm01
     $  - 2.812500000000000d-01*zm02
     $  + HZ2(0,1)*CbH1YZ(3,2)
     $  + HZ2(1,1)*CbH1YZ(3,2)

      CbH3YZ(3,0,2,03) = 
     $  - 1.757812500000000d-01*zm01
     $  + 2.109375000000000d-01*zm02
     $  - 3.515625000000000d-02*zm03
     $  + HZ2(0,1)*CbH1YZ(3,3)
     $  + HZ2(1,1)*CbH1YZ(3,3)

      CbH3YZ(3,0,2,04) = 
     $  + 6.152343750000000d-02*zm01
     $  - 4.614257812500000d-02*zm02
     $  - 1.318359375000000d-02*zm03
     $  - 2.197265625000000d-03*zm04
     $  + HZ2(0,1)*CbH1YZ(3,4)
     $  + HZ2(1,1)*CbH1YZ(3,4)

      CbH3YZ(3,0,2,05) = 
     $  - 1.516113281250000d-02*zm01
     $  - 1.384277343750000d-02*zm02
     $  + 2.570800781250000d-02*zm03
     $  + 3.295898437500000d-03*zm04
     $  + HZ2(0,1)*CbH1YZ(3,5)
     $  + HZ2(1,1)*CbH1YZ(3,5)

      CbH3YZ(3,0,2,06) = 
     $  + 2.875671386718750d-03*zm01
     $  + 1.404876708984375d-02*zm02
     $  - 7.498168945312500d-03*zm03
     $  - 8.981323242187500d-03*zm04
     $  - 4.531860351562500d-04*zm05
     $  + 8.239746093750000d-06*zm06
     $  + HZ2(0,1)*CbH1YZ(3,6)
     $  + HZ2(1,1)*CbH1YZ(3,6)

      CbH3YZ(3,0,2,07) = 
     $  - 4.405321393694196d-04*zm01
     $  - 5.875233241489955d-03*zm02
     $  - 3.610774448939732d-03*zm03
     $  + 7.583509172712053d-03*zm04
     $  + 2.299771990094866d-03*zm05
     $  + 4.325866699218750d-05*zm06
     $  + HZ2(0,1)*CbH1YZ(3,7)
     $  + HZ2(1,1)*CbH1YZ(3,7)

      CbH3YZ(3,0,2,08) = 
     $  + 5.625676135627590d-05*zm01
     $  + 1.721435614994594d-03*zm02
     $  + 4.366697583879743d-03*zm03
     $  - 2.087754862649100d-03*zm04
     $  - 3.581254822867257d-03*zm05
     $  - 4.716794831412179d-04*zm06
     $  - 3.641673496791294d-06*zm07
     $  - 5.911807624661192d-08*zm08
     $  + HZ2(0,1)*CbH1YZ(3,8)
     $  + HZ2(1,1)*CbH1YZ(3,8)

      CbH3YZ(3,0,2,09) = 
     $  - 6.136456314398317d-06*zm01
     $  - 4.008323805672781d-04*zm02
     $  - 2.307248456137520d-03*zm03
     $  - 1.123539039066859d-03*zm04
     $  + 2.542964049748011d-03*zm05
     $  + 1.212263447897774d-03*zm06
     $  + 8.218594959803989d-05*zm07
     $  + 3.428848422303491d-07*zm08
     $  + HZ2(0,1)*CbH1YZ(3,9)
     $  + HZ2(1,1)*CbH1YZ(3,9)

      CbH3YZ(3,0,2,10) = 
     $  + 5.838649005306010d-07*zm01
     $  + 7.858499367626345d-05*zm02
     $  + 8.625117455210004d-04*zm03
     $  + 1.527448219912392d-03*zm04
     $  - 6.844308078289031d-04*zm05
     $  - 1.441305054085595d-03*zm06
     $  - 3.307239583560398d-04*zm07
     $  - 1.263914911114439d-05*zm08
     $  - 3.037191167169687d-08*zm09
     $  + 5.172831671578543d-10*zm10
     $  + HZ2(0,1)*CbH1YZ(3,10)
     $  + HZ2(1,1)*CbH1YZ(3,10)

      CbH3YZ(3,0,2,11) = 
     $  - 4.933168334067420d-08*zm01
     $  - 1.341700359681320d-05*zm02
     $  - 2.578162952345244d-04*zm03
     $  - 9.360843660769524d-04*zm04
     $  - 3.915364963861254d-04*zm05
     $  + 9.308077250788738d-04*zm06
     $  + 5.894576183387211d-04*zm07
     $  + 7.688475108544769d-05*zm08
     $  + 1.751680355622728d-06*zm09
     $  + 1.718119090917159d-09*zm10
     $  + HZ2(0,1)*CbH1YZ(3,11)
     $  + HZ2(1,1)*CbH1YZ(3,11)

      CbH3YZ(3,0,2,12) = 
     $  + 3.749539556984234d-09*zm01
     $  + 2.039213694635509d-06*zm02
     $  + 6.521618621707211d-05*zm03
     $  + 4.116172497545184d-04*zm04
     $  + 5.768329875858186d-04*zm05
     $  - 2.471682943657717d-04*zm06
     $  - 5.974516448044738d-04*zm07
     $  - 1.951367241720943d-04*zm08
     $  - 1.573169545541778d-05*zm09
     $  - 2.209707098479118d-07*zm10
     $  - 5.227415122658511d-11*zm11
     $  - 5.009844723248847d-12*zm12
     $  + HZ2(0,1)*CbH1YZ(3,12)
     $  + HZ2(1,1)*CbH1YZ(3,12)

      CbH3YZ(3,0,2,13) = 
     $  - 2.582866736999981d-10*zm01
     $  - 2.801440922111172d-07*zm02
     $  - 1.442907892104235d-05*zm03
     $  - 1.459462221832193d-04*zm04
     $  - 3.922599786086082d-04*zm05
     $  - 1.473935277061211d-04*zm06
     $  + 3.627254599851875d-04*zm07
     $  + 2.795002996309540d-04*zm08
     $  + 5.516512470120962d-05*zm09
     $  + 2.892747234922730d-06*zm10
     $  + 2.557176143152683d-08*zm11
     $  + 6.484170456090651d-12*zm12
     $  + HZ2(0,1)*CbH1YZ(3,13)
     $  + HZ2(1,1)*CbH1YZ(3,13)

      CbH3YZ(3,0,2,14) = 
     $  + 1.623767522471413d-11*zm01
     $  + 3.517323774090562d-08*zm02
     $  + 2.853407282064314d-06*zm03
     $  + 4.403403081113979d-05*zm04
     $  + 1.939161165593301d-04*zm05
     $  + 2.301696996873983d-04*zm06
     $  - 9.543128603327024d-05*zm07
     $  - 2.548556682178075d-04*zm08
     $  - 1.064936509449355d-04*zm09
     $  - 1.374011447354138d-05*zm10
     $  - 4.849807601837021d-07*zm11
     $  - 2.742417452233713d-09*zm12
     $  - 1.019890740039256d-12*zm13
     $  + 5.173265120044837d-14*zm14
     $  + HZ2(0,1)*CbH1YZ(3,14)
     $  + HZ2(1,1)*CbH1YZ(3,14)

      CbH3YZ(3,0,2,15) = 
     $  - 9.447341079147163d-13*zm01
     $  - 4.070173378130343d-09*zm02
     $  - 5.121110751452167d-07*zm03
     $  - 1.166607918331336d-05*zm04
     $  - 7.795303932341878d-05*zm05
     $  - 1.690925436304935d-04*zm06
     $  - 5.866692976331474d-05*zm07
     $  + 1.481225986713430d-04*zm08
     $  + 1.318358381685796d-04*zm09
     $  + 3.478336898895461d-05*zm10
     $  + 3.077758353000338d-06*zm11
     $  + 7.493485961403926d-08*zm12
     $  + 2.750375245510758d-10*zm13
     $  + 1.478168116164896d-14*zm14
     $  + HZ2(0,1)*CbH1YZ(3,15)
     $  + HZ2(1,1)*CbH1YZ(3,15)

      CbH3YZ(3,0,2,16) = 
     $  + 5.154966464251066d-14*zm01
     $  + 4.370284851385175d-10*zm02
     $  + 8.436471192224349d-08*zm03
     $  + 2.770893424817942d-06*zm04
     $  + 2.681889154144599d-05*zm05
     $  + 9.135635487514957d-05*zm06
     $  + 9.574690994528782d-05*zm07
     $  - 3.870678341227419d-05*zm08
     $  - 1.114795597945977d-04*zm09
     $  - 5.591686160388019d-05*zm10
     $  - 1.003476093173988d-05*zm11
     $  - 6.291000839688175d-07*zm12
     $  - 1.075989947304598d-08*zm13
     $  - 2.585985568493176d-11*zm14
     $  + 7.690093311711172d-15*zm15
     $  - 5.588656410775645d-16*zm16
     $  + HZ2(0,1)*CbH1YZ(3,16)
     $  + HZ2(1,1)*CbH1YZ(3,16)

      CbH3YZ(3,0,2,17) = 
     $  - 2.591253530637954d-15*zm01
     $  - 4.378416592380813d-11*zm02
     $  - 1.286900304342740d-08*zm03
     $  - 5.988254286515277d-07*zm04
     $  - 8.135230177144451d-06*zm05
     $  - 4.045372407302695d-05*zm06
     $  - 7.466877612491439d-05*zm07
     $  - 2.435593774453221d-05*zm08
     $  + 6.272076379025541d-05*zm09
     $  + 6.235282400218208d-05*zm10
     $  + 2.042368728305803d-05*zm11
     $  + 2.608061783595541d-06*zm12
     $  + 1.186219366228185d-07*zm13
     $  + 1.445265321077352d-09*zm14
     $  + 2.276979943481168d-12*zm15
     $  + 5.522457112631734d-17*zm16
     $  + HZ2(0,1)*CbH1YZ(3,17)
     $  + HZ2(1,1)*CbH1YZ(3,17)

      CbH3YZ(3,2,2,01) = 
     $  - HZ2(1,1)*CbH1YZ(3,1)

      CbH3YZ(3,2,2,02) = 
     $  - HZ2(1,1)*CbH1YZ(3,2)

      CbH3YZ(3,2,2,03) = 
     $  + 7.031250000000000d-02*zm01
     $  - 1.406250000000000d-01*zm02
     $  + 7.031250000000000d-02*zm03
     $  - HZ2(1,1)*CbH1YZ(3,3)

      CbH3YZ(3,2,2,04) = 
     $  - 3.955078125000000d-02*zm01
     $  + 7.910156250000000d-02*zm02
     $  - 3.955078125000000d-02*zm03
     $  - HZ2(1,1)*CbH1YZ(3,4)

      CbH3YZ(3,2,2,05) = 
     $  + 1.186523437500000d-02*zm01
     $  - 1.186523437500000d-02*zm02
     $  - 1.186523437500000d-02*zm03
     $  + 1.186523437500000d-02*zm04
     $  - HZ2(1,1)*CbH1YZ(3,5)

      CbH3YZ(3,2,2,06) = 
     $  - 2.471923828125000d-03*zm01
     $  - 4.943847656250000d-03*zm02
     $  + 1.483154296875000d-02*zm03
     $  - 4.943847656250000d-03*zm04
     $  - 2.471923828125000d-03*zm05
     $  - HZ2(1,1)*CbH1YZ(3,6)

      CbH3YZ(3,2,2,07) = 
     $  + 3.972734723772321d-04*zm01
     $  + 3.575461251395089d-03*zm02
     $  - 3.972734723772321d-03*zm03
     $  - 3.972734723772321d-03*zm04
     $  + 3.575461251395089d-03*zm05
     $  + 3.972734723772321d-04*zm06
     $  - HZ2(1,1)*CbH1YZ(3,7)

      CbH3YZ(3,2,2,08) = 
     $  - 5.214214324951171d-05*zm01
     $  - 1.251411437988281d-03*zm02
     $  - 7.821321487426757d-04*zm03
     $  + 4.171371459960937d-03*zm04
     $  - 7.821321487426757d-04*zm05
     $  - 1.251411437988281d-03*zm06
     $  - 5.214214324951171d-05*zm07
     $  - HZ2(1,1)*CbH1YZ(3,8)

      CbH3YZ(3,2,2,09) = 
     $  + 5.793571472167968d-06*zm01
     $  + 3.186464309692382d-04*zm02
     $  + 1.094985008239746d-03*zm03
     $  - 1.419425010681152d-03*zm04
     $  - 1.419425010681152d-03*zm05
     $  + 1.094985008239746d-03*zm06
     $  + 3.186464309692382d-04*zm07
     $  + 5.793571472167968d-06*zm08
     $  - HZ2(1,1)*CbH1YZ(3,9)

      CbH3YZ(3,2,2,10) = 
     $  - 5.586658205304827d-07*zm01
     $  - 6.592256682259695d-05*zm02
     $  - 5.318498611450195d-04*zm03
     $  - 8.603453636169433d-05*zm04
     $  + 1.368731260299682d-03*zm05
     $  - 8.603453636169433d-05*zm06
     $  - 5.318498611450195d-04*zm07
     $  - 6.592256682259695d-05*zm08
     $  - 5.586658205304827d-07*zm09
     $  - HZ2(1,1)*CbH1YZ(3,10)

      CbH3YZ(3,2,2,11) = 
     $  + 4.761356424975704d-08*zm01
     $  + 1.166532324119047d-05*zm02
     $  + 1.809315441490767d-04*zm03
     $  + 3.466267477382313d-04*zm04
     $  - 5.392712286927483d-04*zm05
     $  - 5.392712286927483d-04*zm06
     $  + 3.466267477382313d-04*zm07
     $  + 1.809315441490767d-04*zm08
     $  + 1.166532324119047d-05*zm09
     $  + 4.761356424975704d-08*zm10
     $  - HZ2(1,1)*CbH1YZ(3,11)

      CbH3YZ(3,2,2,12) = 
     $  - 3.637147269078663d-09*zm01
     $  - 1.818573634539331d-06*zm02
     $  - 4.948338859581521d-05*zm03
     $  - 2.164830054555620d-04*zm04
     $  + 2.062262501567602d-05*zm05
     $  + 4.943319596350193d-04*zm06
     $  + 2.062262501567602d-05*zm07
     $  - 2.164830054555620d-04*zm08
     $  - 4.948338859581521d-05*zm09
     $  - 1.818573634539331d-06*zm10
     $  - 3.637147269078663d-09*zm11
     $  - HZ2(1,1)*CbH1YZ(3,12)

      CbH3YZ(3,2,2,13) = 
     $  + 2.518025032439074d-10*zm01
     $  + 2.545723307795904d-07*zm02
     $  + 1.153633168611962d-05*zm03
     $  + 9.078109748200973d-05*zm04
     $  + 1.127596789776542d-04*zm05
     $  - 2.153319322790663d-04*zm06
     $  - 2.153319322790663d-04*zm07
     $  + 1.127596789776542d-04*zm08
     $  + 9.078109748200973d-05*zm09
     $  + 1.153633168611962d-05*zm10
     $  + 2.545723307795904d-07*zm11
     $  + 2.518025032439074d-10*zm12
     $  - HZ2(1,1)*CbH1YZ(3,13)

      CbH3YZ(3,2,2,14) = 
     $  - 1.594204160148115d-11*zm01
     $  - 3.242611261741266d-08*zm02
     $  - 2.368445352565649d-06*zm03
     $  - 3.029386455321456d-05*zm04
     $  - 8.742256918316231d-05*zm05
     $  + 2.468612388356075d-05*zm06
     $  + 1.908623945200815d-04*zm07
     $  + 2.468612388356075d-05*zm08
     $  - 8.742256918316231d-05*zm09
     $  - 3.029386455321456d-05*zm10
     $  - 2.368445352565649d-06*zm11
     $  - 3.242611261741266d-08*zm12
     $  - 1.594204160148115d-11*zm13
     $  - HZ2(1,1)*CbH1YZ(3,14)

      CbH3YZ(3,2,2,15) = 
     $  + 9.299524267530673d-13*zm01
     $  + 3.795135853579267d-09*zm02
     $  + 4.371762155311774d-07*zm03
     $  + 8.588320830313023d-06*zm04
     $  + 4.316967033446417d-05*zm05
     $  + 3.725670546191395d-05*zm06
     $  - 8.945566890802833d-05*zm07
     $  - 8.945566890802833d-05*zm08
     $  + 3.725670546191395d-05*zm09
     $  + 4.316967033446417d-05*zm10
     $  + 8.588320830313023d-06*zm11
     $  + 4.371762155311774d-07*zm12
     $  + 3.795135853579267d-09*zm13
     $  + 9.299524267530673d-13*zm14
     $  - HZ2(1,1)*CbH1YZ(3,15)

      CbH3YZ(3,2,2,16) = 
     $  - 5.029790769698080d-14*zm01
     $  - 4.112356933305150d-10*zm02
     $  - 7.360449948443849d-08*zm03
     $  - 2.141794357984591d-06*zm04
     $  - 1.678412816858099d-05*zm05
     $  - 3.543949774666543d-05*zm06
     $  + 1.573265624273290d-05*zm07
     $  + 7.741355963194759d-05*zm08
     $  + 1.573265624273290d-05*zm09
     $  - 3.543949774666543d-05*zm10
     $  - 1.678412816858099d-05*zm11
     $  - 2.141794357984591d-06*zm12
     $  - 7.360449948443849d-08*zm13
     $  - 4.112356933305150d-10*zm14
     $  - 5.029790769698080d-14*zm15
     $  - HZ2(1,1)*CbH1YZ(3,16)

      CbH3YZ(3,2,2,17) = 
     $  + 2.536028959511637d-15*zm01
     $  + 4.150718598032696d-11*zm02
     $  + 1.142373772235005d-08*zm03
     $  + 4.802034920287092d-07*zm04
     $  + 5.527168393548910d-06*zm05
     $  + 2.003003678996891d-05*zm06
     $  + 1.231595212273230d-05*zm07
     $  - 3.836482604572320d-05*zm08
     $  - 3.836482604572320d-05*zm09
     $  + 1.231595212273230d-05*zm10
     $  + 2.003003678996891d-05*zm11
     $  + 5.527168393548910d-06*zm12
     $  + 4.802034920287092d-07*zm13
     $  + 1.142373772235005d-08*zm14
     $  + 4.150718598032696d-11*zm15
     $  + 2.536028959511637d-15*zm16
     $  - HZ2(1,1)*CbH1YZ(3,17)

      CbH3YZ(3,2,2,18) = 
     $  - 1.197569230880495d-16*zm01
     $  - 3.922039231133622d-12*zm02
     $  - 1.647763767402322d-09*zm03
     $  - 9.825746398853205d-08*zm04
     $  - 1.599817458861116d-06*zm05
     $  - 8.807239180168909d-06*zm06
     $  - 1.447559772869258d-05*zm07
     $  + 8.677512448205176d-06*zm08
     $  + 3.261010213886472d-05*zm09
     $  + 8.677512448205176d-06*zm10
     $  - 1.447559772869258d-05*zm11
     $  - 8.807239180168909d-06*zm12
     $  - 1.599817458861116d-06*zm13
     $  - 9.825746398853205d-08*zm14
     $  - 1.647763767402322d-09*zm15
     $  - 3.922039231133622d-12*zm16
     $  - 1.197569230880495d-16*zm17
     $  - HZ2(1,1)*CbH1YZ(3,18)

      CbH3YZ(3,3,2,01) = 
     $  + HZ1(1)*CbH2YZ(3,3,1)
     $  + HZ2(0,1)*CbH1YZ(3,1)

      CbH3YZ(3,3,2,02) = 
     $  + HZ1(1)*CbH2YZ(3,3,2)
     $  + HZ2(0,1)*CbH1YZ(3,2)

      CbH3YZ(3,3,2,03) = 
     $  + 7.031250000000000d-02*zm02
     $  - 7.031250000000000d-02*zm03
     $  + HZ1(1)*CbH2YZ(3,3,3)
     $  + HZ2(0,1)*CbH1YZ(3,3)

      CbH3YZ(3,3,2,04) = 
     $  - 6.591796875000000d-02*zm02
     $  + 6.591796875000000d-02*zm03
     $  + HZ1(1)*CbH2YZ(3,3,4)
     $  + HZ2(0,1)*CbH1YZ(3,4)

      CbH3YZ(3,3,2,05) = 
     $  + 3.361816406250000d-02*zm02
     $  - 1.582031250000000d-02*zm03
     $  - 1.779785156250000d-02*zm04
     $  + HZ1(1)*CbH2YZ(3,3,5)
     $  + HZ2(0,1)*CbH1YZ(3,5)

      CbH3YZ(3,3,2,06) = 
     $  - 1.211242675781250d-02*zm02
     $  - 1.038208007812500d-02*zm03
     $  + 1.903381347656250d-02*zm04
     $  + 3.460693359375000d-03*zm05
     $  + HZ1(1)*CbH2YZ(3,3,6)
     $  + HZ2(0,1)*CbH1YZ(3,6)

      CbH3YZ(3,3,2,07) = 
     $  + 3.416551862444196d-03*zm02
     $  + 1.157390049525669d-02*zm03
     $  - 6.276920863560267d-03*zm04
     $  - 8.183833530970982d-03*zm05
     $  - 5.296979631696428d-04*zm06
     $  + HZ1(1)*CbH2YZ(3,3,7)
     $  + HZ2(0,1)*CbH1YZ(3,7)

      CbH3YZ(3,3,2,08) = 
     $  - 7.970299039568219d-04*zm02
     $  - 6.160221781049455d-03*zm03
     $  - 2.681595938546316d-03*zm04
     $  + 7.165820258004324d-03*zm05
     $  + 2.405987467084612d-03*zm06
     $  + 6.703989846365792d-05*zm07
     $  + HZ1(1)*CbH2YZ(3,3,8)
     $  + HZ2(0,1)*CbH1YZ(3,8)

      CbH3YZ(3,3,2,09) = 
     $  + 1.591163022177559d-04*zm02
     $  + 2.354259150368826d-03*zm03
     $  + 4.286208323069981d-03*zm04
     $  - 2.511927059718540d-03*zm05
     $  - 3.722990410668509d-03*zm06
     $  - 5.574243409293038d-04*zm07
     $  - 7.241964340209960d-06*zm08
     $  + HZ1(1)*CbH2YZ(3,3,9)
     $  + HZ2(0,1)*CbH1YZ(3,9)

      CbH3YZ(3,3,2,10) = 
     $  - 2.782466156142098d-05*zm02
     $  - 7.224169799259730d-04*zm03
     $  - 2.796634542090552d-03*zm04
     $  - 8.364468812942504d-04*zm05
     $  + 2.901787864310400d-03*zm06
     $  + 1.372455699103219d-03*zm07
     $  + 1.083966876779283d-04*zm08
     $  + 6.828137806483677d-07*zm09
     $  + HZ1(1)*CbH2YZ(3,3,10)
     $  + HZ2(0,1)*CbH1YZ(3,10)

      CbH3YZ(3,3,2,11) = 
     $  + 4.334950505138991d-06*zm02
     $  + 1.876937283517478d-04*zm03
     $  + 1.286991467433316d-03*zm04
     $  + 1.684185936563200d-03*zm05
     $  - 1.037060462131902d-03*zm06
     $  - 1.702943564719193d-03*zm07
     $  - 4.048115813306399d-04*zm08
     $  - 1.833333839456756d-05*zm09
     $  - 5.713627709970845d-08*zm10
     $  + HZ1(1)*CbH2YZ(3,3,11)
     $  + HZ2(0,1)*CbH1YZ(3,11)

      CbH3YZ(3,3,2,12) = 
     $  - 6.095197523472371d-07*zm02
     $  - 4.260878347755446d-05*zm03
     $  - 4.737730838914776d-04*zm04
     $  - 1.252433277981041d-03*zm05
     $  - 2.921052488752386d-04*zm06
     $  + 1.228061084780696d-03*zm07
     $  + 7.296166357935055d-04*zm08
     $  + 1.010891518180633d-04*zm09
     $  + 2.758743138620992d-06*zm10
     $  + 4.298446772547511d-09*zm11
     $  + HZ1(1)*CbH2YZ(3,3,12)
     $  + HZ2(0,1)*CbH1YZ(3,12)

      CbH3YZ(3,3,2,13) = 
     $  + 7.813889498391619d-08*zm02
     $  + 8.630675775883765d-06*zm03
     $  + 1.476450959567332d-04*zm04
     $  + 6.553873892563705d-04*zm05
     $  + 6.931315543914825d-04*zm06
     $  - 4.419109667325718d-04*zm07
     $  - 7.869403480636337d-04*zm08
     $  - 2.535563763524112d-04*zm09
     $  - 2.208972940968998d-05*zm10
     $  - 3.751399475601050d-07*zm11
     $  - 2.937695871178920d-10*zm12
     $  + HZ1(1)*CbH2YZ(3,3,13)
     $  + HZ2(0,1)*CbH1YZ(3,13)

      CbH3YZ(3,3,2,14) = 
     $  - 9.208981646640210d-09*zm02
     $  - 1.583492334502812d-06*zm03
     $  - 4.024619232880408d-05*zm04
     $  - 2.751690524302858d-04*zm05
     $  - 5.652618191232968d-04*zm06
     $  - 1.101129199154316d-04*zm07
     $  + 5.367778300268805d-04*zm08
     $  + 3.760412476416269d-04*zm09
     $  + 7.520065142866704d-05*zm10
     $  + 4.316313895155484d-06*zm11
     $  + 4.662372697442405d-08*zm12
     $  + 1.839466338632440d-11*zm13
     $  + HZ1(1)*CbH2YZ(3,3,14)
     $  + HZ2(0,1)*CbH1YZ(3,14)

      CbH3YZ(3,3,2,15) = 
     $  + 1.004604102329233d-09*zm02
     $  + 2.661491245936055d-07*zm03
     $  + 9.804586652540814d-06*zm04
     $  + 9.811498047735260d-05*zm05
     $  + 3.252497752563974d-04*zm06
     $  + 2.958297884292988d-04*zm07
     $  - 1.935947382486950d-04*zm08
     $  - 3.678288672128705d-04*zm09
     $  - 1.474517465738830d-04*zm10
     $  - 1.961981045861284d-05*zm11
     $  - 7.657790852006654d-07*zm12
     $  - 5.341902220705540d-09*zm13
     $  - 1.062802773432076d-12*zm14
     $  + HZ1(1)*CbH2YZ(3,3,15)
     $  + HZ2(0,1)*CbH1YZ(3,15)

      CbH3YZ(3,3,2,16) = 
     $  - 1.020295452771717d-10*zm02
     $  - 4.134274797465956d-08*zm03
     $  - 2.167597615758065d-06*zm04
     $  - 3.068448559542436d-05*zm05
     $  - 1.509176207272158d-04*zm06
     $  - 2.584697991623445d-04*zm07
     $  - 4.386768379143697d-05*zm08
     $  + 2.406394554205806d-04*zm09
     $  + 1.910823848401033d-04*zm10
     $  + 4.970649751981616d-05*zm11
     $  + 4.594964554807269d-06*zm12
     $  + 1.247611807236898d-07*zm13
     $  + 5.680966643850888d-10*zm14
     $  + 5.700429538991158d-14*zm15
     $  + HZ1(1)*CbH2YZ(3,3,16)
     $  + HZ2(0,1)*CbH1YZ(3,16)

      CbH3YZ(3,3,2,17) = 
     $  + 9.695090777190351d-12*zm02
     $  + 5.977253344596243d-09*zm03
     $  + 4.399001600732747d-07*zm04
     $  + 8.596204421675060d-06*zm05
     $  + 5.972808401398264d-05*zm06
     $  + 1.600228100259025d-04*zm07
     $  + 1.299849608834853d-04*zm08
     $  - 8.682453158204671d-05*zm09
     $  - 1.738782561639764d-04*zm10
     $  - 8.226187540269767d-05*zm11
     $  - 1.481426886441758d-05*zm12
     $  - 9.801290886055725d-07*zm13
     $  - 1.882895828224005d-08*zm14
     $  - 5.639067500505818d-11*zm15
     $  - 2.853032579450591d-15*zm16
     $  + HZ1(1)*CbH2YZ(3,3,17)
     $  + HZ2(0,1)*CbH1YZ(3,17)
      endif
      if (n.eq.3) return
      
      if (n.eq.4) then
      CbH4YZ(0,0,3,2,01) = 
     $  + HZ1(1)*CbH3YZ(0,0,3,1)

      CbH4YZ(0,0,3,2,02) = 
     $  + 7.031250000000000d-02*zm01
     $  - 7.031250000000000d-02*zm02
     $  + HZ1(1)*CbH3YZ(0,0,3,2)

      CbH4YZ(0,0,3,2,03) = 
     $  - 4.492187500000000d-02*zm01
     $  + 7.421875000000000d-02*zm02
     $  - 2.929687500000000d-02*zm03
     $  + HZ1(1)*CbH3YZ(0,0,3,3)

      CbH4YZ(0,0,3,2,04) = 
     $  + 1.593017578125000d-02*zm01
     $  - 3.268432617187500d-02*zm02
     $  + 2.252197265625000d-02*zm03
     $  - 5.767822265625000d-03*zm04
     $  + HZ1(1)*CbH3YZ(0,0,3,4)

      CbH4YZ(0,0,3,2,05) = 
     $  - 3.919921875000000d-03*zm01
     $  + 7.768432617187500d-03*zm02
     $  - 6.252319335937500d-03*zm03
     $  + 2.980590820312500d-03*zm04
     $  - 5.767822265625000d-04*zm05
     $  + HZ1(1)*CbH3YZ(0,0,3,5)

      CbH4YZ(0,0,3,2,06) = 
     $  + 7.271575927734375d-04*zm01
     $  - 7.240676879882812d-04*zm02
     $  + 1.448822021484375d-04*zm03
     $  - 2.142333984375000d-04*zm04
     $  + 7.827758789062500d-05*zm05
     $  - 1.201629638671875d-05*zm06
     $  + HZ1(1)*CbH3YZ(0,0,3,6)

      CbH4YZ(0,0,3,2,07) = 
     $  - 1.061137563633154d-04*zm01
     $  - 1.976623201509258d-04*zm02
     $  + 3.629707247453250d-04*zm03
     $  - 1.153053973228521d-04*zm04
     $  + 7.706144460783755d-05*zm05
     $  - 2.426130778587941d-05*zm06
     $  + 3.310612269810267d-06*zm07
     $  + HZ1(1)*CbH3YZ(0,0,3,7)

      CbH4YZ(0,0,3,2,08) = 
     $  + 1.254781168334338d-05*zm01
     $  + 1.088733271676666d-04*zm02
     $  - 6.676894061419428d-05*zm03
     $  - 4.474745721233134d-05*zm04
     $  - 1.400605756409314d-05*zm05
     $  + 5.382208191618627d-06*zm06
     $  - 1.453319374395876d-06*zm07
     $  + 1.724277223859514d-07*zm08
     $  + HZ1(1)*CbH3YZ(0,0,3,8)

      CbH4YZ(0,0,3,2,09) = 
     $  - 1.241085480670539d-06*zm01
     $  - 2.959022290852605d-05*zm02
     $  - 2.796691716933736d-05*zm03
     $  + 4.829577341371653d-05*zm04
     $  + 8.122577350966784d-06*zm05
     $  + 3.370677330056015d-06*zm06
     $  - 1.257096626320663d-06*zm07
     $  + 2.979058392193852d-07*zm08
     $  - 3.161174910409109d-08*zm09
     $  + HZ1(1)*CbH3YZ(0,0,3,9)

      CbH4YZ(0,0,3,2,10) = 
     $  + 1.072069363934653d-07*zm01
     $  + 5.968736484646797d-06*zm02
     $  + 2.192155537860734d-05*zm03
     $  - 8.525343877928597d-06*zm04
     $  - 1.667511431234223d-05*zm05
     $  - 3.007872296231133d-06*zm06
     $  + 3.055850309985024d-07*zm07
     $  - 1.174232789448329d-07*zm08
     $  + 2.508823360715593d-08*zm09
     $  - 2.418298806462969d-09*zm10
     $  + HZ1(1)*CbH3YZ(0,0,3,10)

      CbH4YZ(0,0,3,2,11) = 
     $  - 8.454209697354576d-09*zm01
     $  - 9.973691420539057d-07*zm02
     $  - 8.262988735454113d-06*zm03
     $  - 6.213428321721156d-06*zm04
     $  + 9.958732702584799d-06*zm05
     $  + 5.047275236341366d-06*zm06
     $  + 5.233108110675728d-07*zm07
     $  - 6.423488787787037d-08*zm08
     $  + 2.076604320688975d-08*zm09
     $  - 3.953372138032449d-09*zm10
     $  + 3.438757418038009d-10*zm11
     $  + HZ1(1)*CbH3YZ(0,0,3,11)

      CbH4YZ(0,0,3,2,12) = 
     $  + 6.114515484725218d-10*zm01
     $  + 1.455135701552424d-07*zm02
     $  + 2.289528680106002d-06*zm03
     $  + 5.465210801985009d-06*zm04
     $  - 1.475181042160640d-06*zm05
     $  - 5.049756753402448d-06*zm06
     $  - 1.334147901387517d-06*zm07
     $  - 3.558868445277900d-08*zm08
     $  - 8.147259981183438d-09*zm09
     $  + 2.336165742121786d-09*zm10
     $  - 4.124605160650776d-10*zm11
     $  + 3.343236378648064d-11*zm12
     $  + HZ1(1)*CbH3YZ(0,0,3,12)

      CbH4YZ(0,0,3,2,13) = 
     $  - 3.843337535796387d-11*zm01
     $  - 1.906455690593406d-08*zm02
     $  - 5.223494076315867d-07*zm03
     $  - 2.501902870701262d-06*zm04
     $  - 1.672754142957553d-06*zm05
     $  + 2.517801968181526d-06*zm06
     $  + 1.908500000369948d-06*zm07
     $  + 2.852591531475047d-07*zm08
     $  + 3.525373875861267d-09*zm09
     $  + 1.311382110425945d-09*zm10
     $  - 3.383688148497314d-10*zm11
     $  + 5.387685694705595d-11*zm12
     $  - 3.974155669884903d-12*zm13
     $  + HZ1(1)*CbH3YZ(0,0,3,13)

      CbH4YZ(0,0,3,2,14) = 
     $  + 2.054239194104635d-12*zm01
     $  + 2.275126337985850d-09*zm02
     $  + 1.033566006025002d-07*zm03
     $  + 8.529736278003017d-07*zm04
     $  + 1.565921982719399d-06*zm05
     $  - 3.196294294059575d-07*zm06
     $  - 1.561565702102347d-06*zm07
     $  - 5.902722961752256d-07*zm08
     $  - 5.187759408429203d-08*zm09
     $  - 1.325366302618035d-09*zm10
     $  + 1.785541712861555d-10*zm11
     $  - 4.366331340697533d-11*zm12
     $  + 6.564224595594877d-12*zm13
     $  - 4.587114146281174d-13*zm14
     $  + HZ1(1)*CbH3YZ(0,0,3,14)

      CbH4YZ(0,0,3,2,15) = 
     $  - 1.174758601076041d-13*zm01
     $  - 2.494838188141703d-10*zm02
     $  - 1.825736343797654d-08*zm03
     $  - 2.401230801016009d-07*zm04
     $  - 8.159982998408942d-07*zm05
     $  - 5.019315504152133d-07*zm06
     $  + 7.193843238251918d-07*zm07
     $  + 6.925712994231521d-07*zm08
     $  + 1.555462818850709d-07*zm09
     $  + 8.910995503241853d-09*zm10
     $  + 1.668742062402440d-10*zm11
     $  - 2.461517720850650d-11*zm12
     $  + 5.431344928931679d-12*zm13
     $  - 7.435574078600612d-13*zm14
     $  + 4.763714964707954d-14*zm15
     $  + HZ1(1)*CbH3YZ(0,0,3,15)

      CbH4YZ(0,0,3,2,16) = 
     $  + 8.602538975158225d-15*zm01
     $  + 2.537572356210121d-11*zm02
     $  + 2.931656562813402d-09*zm03
     $  + 5.860532100419622d-08*zm04
     $  + 3.219156648919870d-07*zm05
     $  + 4.932569624938021d-07*zm06
     $  - 8.071808370634651d-08*zm07
     $  - 5.089945816100445d-07*zm08
     $  - 2.493606371049082d-07*zm09
     $  - 3.619979047940363d-08*zm10
     $  - 1.462166000929591d-09*zm11
     $  + 3.302396951588872d-12*zm12
     $  - 3.717713960858228d-12*zm13
     $  + 7.808422264095433d-13*zm14
     $  - 1.021748959655368d-13*zm15
     $  + 6.272411414502175d-15*zm16
     $  + HZ1(1)*CbH3YZ(0,0,3,16)

      CbH4YZ(0,0,3,2,17) = 
     $  - 3.793403501840769d-16*zm01
     $  - 2.417555358767890d-12*zm02
     $  - 4.333389610222774d-10*zm03
     $  - 1.275616378139776d-08*zm04
     $  - 1.055467150814375d-07*zm05
     $  - 2.824963469435460d-07*zm06
     $  - 1.627217967924155d-07*zm07
     $  + 2.243908301170995d-07*zm08
     $  + 2.547479958116463d-07*zm09
     $  + 7.696353584329644d-08*zm10
     $  + 7.641556977441978d-09*zm11
     $  + 2.132766879501156d-10*zm12
     $  - 7.886594412535162d-13*zm13
     $  + 4.490382240124496d-13*zm14
     $  - 8.607415253152677d-14*zm15
     $  + 1.033802978975718d-14*zm16
     $  - 5.855761312761245d-16*zm17
     $  + HZ1(1)*CbH3YZ(0,0,3,17)

      CbH4YZ(0,2,3,2,01) = 
     $  + HZ1(1)*CbH3YZ(0,2,3,1)

      CbH4YZ(0,2,3,2,02) = 
     $  + HZ1(1)*CbH3YZ(0,2,3,2)

      CbH4YZ(0,2,3,2,03) = 
     $  + 2.343750000000000d-02*zm01
     $  - 4.687500000000000d-02*zm02
     $  + 2.343750000000000d-02*zm03
     $  + HZ1(1)*CbH3YZ(0,2,3,3)

      CbH4YZ(0,2,3,2,04) = 
     $  - 1.318359375000000d-02*zm01
     $  + 3.295898437500000d-02*zm02
     $  - 2.636718750000000d-02*zm03
     $  + 6.591796875000000d-03*zm04
     $  + HZ1(1)*CbH3YZ(0,2,3,4)

      CbH4YZ(0,2,3,2,05) = 
     $  + 3.823242187500000d-03*zm01
     $  - 9.755859375000000d-03*zm02
     $  + 8.701171875000000d-03*zm03
     $  - 3.427734375000000d-03*zm04
     $  + 6.591796875000000d-04*zm05
     $  + HZ1(1)*CbH3YZ(0,2,3,5)

      CbH4YZ(0,2,3,2,06) = 
     $  - 7.415771484375000d-04*zm01
     $  + 1.235961914062500d-03*zm02
     $  - 2.471923828125000d-04*zm03
     $  - 2.471923828125000d-04*zm04
     $  + HZ1(1)*CbH3YZ(0,2,3,6)

      CbH4YZ(0,2,3,2,07) = 
     $  + 1.072007782605229d-04*zm01
     $  + 1.305327123525191d-04*zm02
     $  - 5.996937654456313d-04*zm03
     $  + 4.023182148836096d-04*zm04
     $  - 6.810402383609693d-05*zm05
     $  + 3.216023347815688d-05*zm06
     $  - 4.414149693080357d-06*zm07
     $  + HZ1(1)*CbH3YZ(0,2,3,7)

      CbH4YZ(0,2,3,2,08) = 
     $  - 1.241479601178850d-05*zm01
     $  - 1.030428068978445d-04*zm02
     $  + 1.638753073556082d-04*zm03
     $  + 2.731255122593470d-05*zm04
     $  - 7.200581686837332d-05*zm05
     $  - 3.724438803536551d-06*zm06
     $  + HZ1(1)*CbH3YZ(0,2,3,8)

      CbH4YZ(0,2,3,2,09) = 
     $  + 1.227685383387974d-06*zm01
     $  + 2.903682844979422d-05*zm02
     $  + 4.887751170567103d-06*zm03
     $  - 8.642077445983886d-05*zm04
     $  + 3.731335912431989d-05*zm05
     $  + 1.219868659973144d-05*zm06
     $  + 2.138103757585797d-06*zm07
     $  - 4.276207515171595d-07*zm08
     $  + 4.598072596958705d-08*zm09
     $  + HZ1(1)*CbH3YZ(0,2,3,9)

      CbH4YZ(0,2,3,2,10) = 
     $  - 1.086294651031494d-07*zm01
     $  - 5.897028105599539d-06*zm02
     $  - 1.764452883175441d-05*zm03
     $  + 3.001276935849870d-05*zm04
     $  + 1.388905303818838d-05*zm05
     $  - 1.713241849626813d-05*zm06
     $  - 3.088180507932390d-06*zm07
     $  - 3.103699002947126d-08*zm08
     $  + HZ1(1)*CbH3YZ(0,2,3,10)

      CbH4YZ(0,2,3,2,11) = 
     $  + 8.624948675545216d-09*zm01
     $  + 9.910194280226368d-07*zm02
     $  + 7.564480776031014d-06*zm03
     $  - 1.029479003994073d-06*zm04
     $  - 1.797048135307631d-05*zm05
     $  + 4.315840953426631d-06*zm06
     $  + 5.550715575322350d-06*zm07
     $  + 5.922678510582573d-07*zm08
     $  - 2.843988652493905d-08*zm09
     $  + 5.979750661669824d-09*zm10
     $  - 5.290396027750783d-10*zm11
     $  + HZ1(1)*CbH3YZ(0,2,3,11)

      CbH4YZ(0,2,3,2,12) = 
     $  - 5.951695531219631d-10*zm01
     $  - 1.453866958376262d-07*zm02
     $  - 2.186652938170092d-06*zm03
     $  - 3.569297940022759d-06*zm04
     $  + 7.004484340742036d-06*zm05
     $  + 4.535787164342480d-06*zm06
     $  - 4.069240364645208d-06*zm07
     $  - 1.498107895158328d-06*zm08
     $  - 7.082517682151360d-08*zm09
     $  - 1.653248758672119d-10*zm10
     $  + HZ1(1)*CbH3YZ(0,2,3,12)

      CbH4YZ(0,2,3,2,13) = 
     $  + 3.622054211941166d-11*zm01
     $  + 1.907291183395654d-08*zm02
     $  + 5.089648585285040d-07*zm03
     $  + 2.081532282766843d-06*zm04
     $  - 6.520714266592944d-07*zm05
     $  - 4.553122680829213d-06*zm06
     $  + 4.267677954466586d-07*zm07
     $  + 1.826866989935328d-06*zm08
     $  + 3.343344840586444d-07*zm09
     $  + 7.165708377084779d-09*zm10
     $  + 5.310911405415454d-10*zm11
     $  - 8.459379024508939d-11*zm12
     $  + 6.358649071815845d-12*zm13
     $  + HZ1(1)*CbH3YZ(0,2,3,13)

      CbH4YZ(0,2,3,2,14) = 
     $  - 2.248236636106316d-12*zm01
     $  - 2.271332157913590d-09*zm02
     $  - 1.018068995928023d-07*zm03
     $  - 7.710991795260008d-07*zm04
     $  - 8.083330439971524d-07*zm05
     $  + 1.920098323660650d-06*zm06
     $  + 1.449435297905660d-06*zm07
     $  - 1.009291062560069d-06*zm08
     $  - 6.120957092146126d-07*zm09
     $  - 6.355662777698188d-08*zm10
     $  - 1.076905348694925d-09*zm11
     $  - 6.131554462108136d-13*zm12
     $  + HZ1(1)*CbH3YZ(0,2,3,14)

      CbH4YZ(0,2,3,2,15) = 
     $  + 1.464760232614721d-13*zm01
     $  + 2.489905042392017d-10*zm02
     $  + 1.808913540805237d-08*zm03
     $  + 2.258333500903261d-07*zm04
     $  + 6.084321830898595d-07*zm05
     $  - 2.795645711798137d-07*zm06
     $  - 1.307260006556245d-06*zm07
     $  - 2.847610391737798d-08*zm08
     $  + 5.824172926276938d-07*zm09
     $  + 1.692162279598633d-07*zm10
     $  + 1.091533364110985d-08*zm11
     $  + 1.554564796023634d-10*zm12
     $  - 8.556924893786465d-12*zm13
     $  + 1.200762748829510d-12*zm14
     $  - 7.846118765401337d-14*zm15
     $  + HZ1(1)*CbH3YZ(0,2,3,15)

      CbH4YZ(0,2,3,2,16) = 
     $  - 6.227360000578576d-15*zm01
     $  - 2.540930539928382d-11*zm02
     $  - 2.913997306732274d-09*zm03
     $  - 5.633310621779999d-08*zm04
     $  - 2.721524626632545d-07*zm05
     $  - 1.960420239352170d-07*zm06
     $  + 5.865614440127982d-07*zm07
     $  + 4.808446204379054d-07*zm08
     $  - 2.619589200716367d-07*zm09
     $  - 2.356377708496158d-07*zm10
     $  - 4.058092580509955d-08*zm11
     $  - 1.749548529528702d-09*zm12
     $  - 1.189186246264331d-11*zm13
     $  - 1.676596923232693d-15*zm14
     $  + HZ1(1)*CbH3YZ(0,2,3,16)

      CbH4YZ(0,2,3,2,17) = 
     $  - 3.895207878988462d-18*zm01
     $  + 2.423837939029560d-12*zm02
     $  + 4.316362985217232d-10*zm03
     $  + 1.242230595097086d-08*zm04
     $  + 9.485178897623776d-08*zm05
     $  + 1.873313173349071d-07*zm06
     $  - 1.124552322051772d-07*zm07
     $  - 4.102775266959387d-07*zm08
     $  - 4.742470789396770d-08*zm09
     $  + 1.885382580518935d-07*zm10
     $  + 7.760605485412915d-08*zm11
     $  + 8.714310806987779d-09*zm12
     $  + 2.588304896036717d-10*zm13
     $  + 4.170492165008183d-13*zm14
     $  + 1.392280727539522d-13*zm15
     $  - 1.706573448548228d-14*zm16
     $  + 9.862334842545256d-16*zm17
     $  + HZ1(1)*CbH3YZ(0,2,3,17)

      CbH4YZ(0,3,0,2,01) = 
     $  + HZ2(0,1)*CbH2YZ(0,3,1)
     $  + HZ2(1,1)*CbH2YZ(0,3,1)

      CbH4YZ(0,3,0,2,02) = 
     $  + 1.406250000000000d-01*zm01
     $  - 1.406250000000000d-01*zm02
     $  + HZ2(0,1)*CbH2YZ(0,3,2)
     $  + HZ2(1,1)*CbH2YZ(0,3,2)

      CbH4YZ(0,3,0,2,03) = 
     $  - 9.375000000000000d-02*zm01
     $  + 1.406250000000000d-01*zm02
     $  - 4.687500000000000d-02*zm03
     $  + HZ2(0,1)*CbH2YZ(0,3,3)
     $  + HZ2(1,1)*CbH2YZ(0,3,3)

      CbH4YZ(0,3,0,2,04) = 
     $  + 3.515625000000000d-02*zm01
     $  - 5.767822265625000d-02*zm02
     $  + 2.966308593750000d-02*zm03
     $  - 7.141113281250000d-03*zm04
     $  + HZ2(0,1)*CbH2YZ(0,3,4)
     $  + HZ2(1,1)*CbH2YZ(0,3,4)

      CbH4YZ(0,3,0,2,05) = 
     $  - 9.294433593750000d-03*zm01
     $  + 1.057983398437500d-02*zm02
     $  - 3.262939453125000d-03*zm03
     $  + 2.471923828125000d-03*zm04
     $  - 4.943847656250000d-04*zm05
     $  + HZ2(0,1)*CbH2YZ(0,3,5)
     $  + HZ2(1,1)*CbH2YZ(0,3,5)

      CbH4YZ(0,3,0,2,06) = 
     $  + 1.886901855468750d-03*zm01
     $  + 1.040267944335937d-03*zm02
     $  - 2.828979492187500d-03*zm03
     $  - 6.179809570312500d-05*zm04
     $  - 4.119873046875000d-05*zm05
     $  + 4.806518554687500d-06*zm06
     $  + HZ2(0,1)*CbH2YZ(0,3,6)
     $  + HZ2(1,1)*CbH2YZ(0,3,6)

      CbH4YZ(0,3,0,2,07) = 
     $  - 3.074770557637117d-04*zm01
     $  - 1.384907936563297d-03*zm02
     $  + 1.015885022221779d-03*zm03
     $  + 6.154585857780612d-04*zm04
     $  + 7.693232322225765d-05*zm05
     $  - 1.853942871093750d-05*zm06
     $  + 2.648489815848214d-06*zm07
     $  + HZ2(0,1)*CbH2YZ(0,3,7)
     $  + HZ2(1,1)*CbH2YZ(0,3,7)

      CbH4YZ(0,3,0,2,08) = 
     $  + 4.135900614212970d-05*zm01
     $  + 5.331534147262573d-04*zm02
     $  + 2.226248809269496d-04*zm03
     $  - 6.494317735944475d-04*zm04
     $  - 1.437357493809291d-04*zm05
     $  - 4.434841019766671d-06*zm06
     $  + 5.103860582624162d-07*zm07
     $  - 4.532385845573580d-08*zm08
     $  + HZ2(0,1)*CbH2YZ(0,3,8)
     $  + HZ2(1,1)*CbH2YZ(0,3,8)

      CbH4YZ(0,3,0,2,09) = 
     $  - 4.694960555251763d-06*zm01
     $  - 1.413882691033032d-04*zm02
     $  - 3.285725201879228d-04*zm03
     $  + 2.010679670742579d-04*zm04
     $  + 2.452784350940159d-04*zm05
     $  + 2.884715795516967d-05*zm06
     $  - 7.414392062595912d-07*zm07
     $  + 2.290825454556212d-07*zm08
     $  - 2.545361616173569d-08*zm09
     $  + HZ2(0,1)*CbH2YZ(0,3,9)
     $  + HZ2(1,1)*CbH2YZ(0,3,9)

      CbH4YZ(0,3,0,2,10) = 
     $  + 4.597169404127159d-07*zm01
     $  + 2.986606415741297d-05*zm02
     $  + 1.649912349423583d-04*zm03
     $  + 6.068791981254305d-05*zm04
     $  - 1.792367552007947d-04*zm05
     $  - 7.249516674450465d-05*zm06
     $  - 4.291898437908717d-06*zm07
     $  + 2.471874563061461d-08*zm08
     $  - 6.281295601202517d-09*zm09
     $  + 4.470804516150026d-10*zm10
     $  + HZ2(0,1)*CbH2YZ(0,3,10)
     $  + HZ2(1,1)*CbH2YZ(0,3,10)

      CbH4YZ(0,3,0,2,11) = 
     $  - 3.972881297515081d-08*zm01
     $  - 5.341233298255372d-06*zm02
     $  - 5.730712162209795d-05*zm03
     $  - 9.429102979486075d-05*zm04
     $  + 5.186610641628679d-05*zm05
     $  + 8.682117015651153d-05*zm06
     $  + 1.774483474044056d-05*zm07
     $  + 5.334862836960657d-07*zm08
     $  + 1.633747580240821d-08*zm09
     $  - 3.103699002947126d-09*zm10
     $  + 2.821544548133751d-10*zm11
     $  + HZ2(0,1)*CbH2YZ(0,3,11)
     $  + HZ2(1,1)*CbH2YZ(0,3,11)

      CbH4YZ(0,3,0,2,12) = 
     $  + 3.088240053515386d-09*zm01
     $  + 8.371230600365473d-07*zm02
     $  + 1.587288641584225d-05*zm03
     $  + 5.549883623330997d-05*zm04
     $  + 1.874906905099113d-05*zm05
     $  - 5.557662977492315d-05*zm06
     $  - 3.159530976992826d-05*zm07
     $  - 3.717644423495197d-06*zm08
     $  - 7.087367211843465d-08*zm09
     $  - 6.183937618747381d-10*zm10
     $  + 7.762396552622428d-11*zm11
     $  - 4.589972022633706d-12*zm12
     $  + HZ2(0,1)*CbH2YZ(0,3,12)
     $  + HZ2(1,1)*CbH2YZ(0,3,12)

      CbH4YZ(0,3,0,2,13) = 
     $  - 2.172852659817552d-10*zm01
     $  - 1.175369694657466d-07*zm02
     $  - 3.728090776240504d-06*zm03
     $  - 2.299844987186936d-05*zm04
     $  - 3.035663636612736d-05*zm05
     $  + 1.552518210596665d-05*zm06
     $  + 3.149557149295347d-05*zm07
     $  + 9.478675321461882d-06*zm08
     $  + 6.922684526924050d-07*zm09
     $  + 9.449544364208876d-09*zm10
     $  - 2.555345411222116d-10*zm11
     $  + 4.320991073802131d-11*zm12
     $  - 3.323839287540100d-12*zm13
     $  + HZ2(0,1)*CbH2YZ(0,3,13)
     $  + HZ2(1,1)*CbH2YZ(0,3,13)

      CbH4YZ(0,3,0,2,14) = 
     $  + 1.378505343987087d-11*zm01
     $  + 1.500206096431825d-08*zm02
     $  + 7.685272329724758d-07*zm03
     $  + 7.658113612005016d-06*zm04
     $  + 1.992558616391581d-05*zm05
     $  + 6.309350093435096d-06*zm06
     $  - 1.874625529990980d-05*zm07
     $  - 1.335646275619037d-05*zm08
     $  - 2.455197513253452d-06*zm09
     $  - 1.177281251995053d-07*zm10
     $  - 9.564742222632628d-10*zm11
     $  + 8.140229879217984d-12*zm12
     $  - 9.686676337872633d-13*zm13
     $  + 4.886700372748551d-14*zm14
     $  + HZ2(0,1)*CbH2YZ(0,3,14)
     $  + HZ2(1,1)*CbH2YZ(0,3,14)

      CbH4YZ(0,3,0,2,15) = 
     $  - 7.999612942260519d-13*zm01
     $  - 1.758457775499041d-09*zm02
     $  - 1.421417096366776d-07*zm03
     $  - 2.171200653021378d-06*zm04
     $  - 9.364611340742044d-06*zm05
     $  - 1.057951874285399d-05*zm06
     $  + 5.127064450342725d-06*zm07
     $  + 1.187324869574973d-05*zm08
     $  + 4.675129558549046d-06*zm09
     $  + 5.652911963253242d-07*zm10
     $  + 1.842084468590943d-08*zm11
     $  + 7.326923271564991d-11*zm12
     $  + 4.256440168312876d-12*zm13
     $  - 6.078586516052684d-13*zm14
     $  + 4.052391010701789d-14*zm15
     $  + HZ2(0,1)*CbH2YZ(0,3,15)
     $  + HZ2(1,1)*CbH2YZ(0,3,15)

      CbH4YZ(0,3,0,2,16) = 
     $  + 4.484327694957989d-14*zm01
     $  + 1.907604686280511d-10*zm02
     $  + 2.394756248673439d-08*zm03
     $  + 5.416028390581193d-07*zm04
     $  + 3.567303270919572d-06*zm05
     $  + 7.526965675701622d-06*zm06
     $  + 2.264697842869308d-06*zm07
     $  - 6.741016631297724d-06*zm08
     $  - 5.647713211963389d-06*zm09
     $  - 1.415387708006357d-06*zm10
     $  - 1.179268754363852d-07*zm11
     $  - 2.655575256465435d-09*zm12
     $  - 7.884695863152803d-12*zm13
     $  - 1.213797191238569d-13*zm14
     $  + 1.222571030867449d-14*zm15
     $  - 5.370677276043950d-16*zm16
     $  + HZ2(0,1)*CbH2YZ(0,3,16)
     $  + HZ2(1,1)*CbH2YZ(0,3,16)

      CbH4YZ(0,3,0,2,17) = 
     $  - 2.424132537276673d-15*zm01
     $  - 1.927717235356433d-11*zm02
     $  - 3.716120202313431d-09*zm03
     $  - 1.214188918070854d-07*zm04
     $  - 1.163037467355838d-06*zm05
     $  - 3.889651879288574d-06*zm06
     $  - 3.910395722922411d-06*zm07
     $  + 1.817718235736251d-06*zm08
     $  + 4.641330001301737d-06*zm09
     $  + 2.225184261482652d-06*zm10
     $  + 3.810717147636486d-07*zm11
     $  + 2.257752477784284d-08*zm12
     $  + 3.565823574149975d-10*zm13
     $  + 1.101303966248109d-12*zm14
     $  - 6.863792025368843d-14*zm15
     $  + 8.592559231567554d-15*zm16
     $  - 5.054446606804443d-16*zm17
     $  + HZ2(0,1)*CbH2YZ(0,3,17)
     $  + HZ2(1,1)*CbH2YZ(0,3,17)

      CbH4YZ(0,3,2,2,01) = 
     $  - HZ2(1,1)*CbH2YZ(0,3,1)

      CbH4YZ(0,3,2,2,02) = 
     $  - HZ2(1,1)*CbH2YZ(0,3,2)

      CbH4YZ(0,3,2,2,03) = 
     $  + 2.343750000000000d-02*zm01
     $  - 4.687500000000000d-02*zm02
     $  + 2.343750000000000d-02*zm03
     $  - HZ2(1,1)*CbH2YZ(0,3,3)

      CbH4YZ(0,3,2,2,04) = 
     $  - 1.647949218750000d-02*zm01
     $  + 3.955078125000000d-02*zm02
     $  - 2.966308593750000d-02*zm03
     $  + 6.591796875000000d-03*zm04
     $  - HZ2(1,1)*CbH2YZ(0,3,4)

      CbH4YZ(0,3,2,2,05) = 
     $  + 5.998535156250000d-03*zm01
     $  - 1.390869140625000d-02*zm02
     $  + 1.048095703125000d-02*zm03
     $  - 3.229980468750000d-03*zm04
     $  + 6.591796875000000d-04*zm05
     $  - HZ2(1,1)*CbH2YZ(0,3,5)

      CbH4YZ(0,3,2,2,06) = 
     $  - 1.462554931640625d-03*zm01
     $  + 1.895141601562500d-03*zm02
     $  + 6.179809570312500d-04*zm03
     $  - 1.071166992187500d-03*zm04
     $  + 2.059936523437500d-05*zm05
     $  - HZ2(1,1)*CbH2YZ(0,3,6)

      CbH4YZ(0,3,2,2,07) = 
     $  + 2.642183887715242d-04*zm01
     $  + 4.313254842952806d-04*zm02
     $  - 1.534232314752072d-03*zm03
     $  + 7.390547771843112d-04*zm04
     $  + 7.377935915577168d-05*zm05
     $  + 3.026845503826530d-05*zm06
     $  - 4.414149693080357d-06*zm07
     $  - HZ2(1,1)*CbH2YZ(0,3,7)

      CbH4YZ(0,3,2,2,08) = 
     $  - 3.745130130222865d-05*zm01
     $  - 3.184395177023751d-04*zm02
     $  + 4.189993653978620d-04*zm03
     $  + 2.462267875671386d-04*zm04
     $  - 2.886440072740827d-04*zm05
     $  - 2.048441341945103d-05*zm06
     $  - 2.069132668631417d-07*zm07
     $  - HZ2(1,1)*CbH2YZ(0,3,8)

      CbH4YZ(0,3,2,2,09) = 
     $  + 4.352075713021414d-06*zm01
     $  + 1.023875815527779d-04*zm02
     $  + 4.231836114610944d-05*zm03
     $  - 3.244744879858834d-04*zm04
     $  + 9.411105087825230d-05*zm05
     $  + 7.756258760179792d-05*zm06
     $  + 4.103779792785644d-06*zm07
     $  - 4.069294248308454d-07*zm08
     $  + 4.598072596958705d-08*zm09
     $  - HZ2(1,1)*CbH2YZ(0,3,9)

      CbH4YZ(0,3,2,2,10) = 
     $  - 4.321900861603873d-07*zm01
     $  - 2.373243442603519d-05*zm02
     $  - 7.599717378616333d-05*zm03
     $  + 1.117036789655685d-04*zm04
     $  + 8.337311446666717d-05*zm05
     $  - 7.687707245349884d-05*zm06
     $  - 1.777178049087524d-05*zm07
     $  - 2.684699637549264d-07*zm08
     $  + 2.327774252210344d-09*zm09
     $  - HZ2(1,1)*CbH2YZ(0,3,10)

      CbH4YZ(0,3,2,2,11) = 
     $  + 3.801069388423365d-08*zm01
     $  + 4.478785151081355d-06*zm02
     $  + 3.504529865865442d-05*zm03
     $  + 9.062288080505947d-07*zm04
     $  - 8.719627526638217d-05*zm05
     $  + 1.223671807519379d-05*zm06
     $  + 3.112814515617971d-05*zm07
     $  + 3.364820125674413d-06*zm08
     $  - 6.941640848533603d-09*zm09
     $  + 5.739278114953879d-09*zm10
     $  - 5.290396027750783d-10*zm11
     $  - HZ2(1,1)*CbH2YZ(0,3,11)

      CbH4YZ(0,3,2,2,12) = 
     $  - 3.003401911587684d-09*zm01
     $  - 7.274349646449282d-07*zm02
     $  - 1.110518602926477d-05*zm03
     $  - 1.974765964263638d-05*zm04
     $  + 3.418425764803859d-05*zm05
     $  + 2.850250457413494d-05*zm06
     $  - 2.118436587889763d-05*zm07
     $  - 9.391709418314230d-06*zm08
     $  - 5.260472225218817d-07*zm09
     $  - 1.328109836133269d-09*zm10
     $  - 2.755414597786866d-11*zm11
     $  - HZ2(1,1)*CbH2YZ(0,3,12)

      CbH4YZ(0,3,2,2,13) = 
     $  + 2.108010955256645d-10*zm01
     $  + 1.048104944765788d-07*zm02
     $  + 2.824659682922490d-06*zm03
     $  + 1.195372141191140d-05*zm04
     $  - 2.328609345471741d-06*zm05
     $  - 2.681730649950791d-05*zm06
     $  + 5.229896318768195d-07*zm07
     $  + 1.134237806189810d-05*zm08
     $  + 2.323674565627739d-06*zm09
     $  + 7.294793844511157d-08*zm10
     $  + 5.985948385890006d-10*zm11
     $  - 8.169676177786428d-11*zm12
     $  + 6.358649071815845d-12*zm13
     $  - HZ2(1,1)*CbH2YZ(0,3,13)

      CbH4YZ(0,3,2,2,14) = 
     $  - 1.315315758383498d-11*zm01
     $  - 1.363618587215805d-08*zm02
     $  - 6.139898602596835d-07*zm03
     $  - 4.740030635240839d-06*zm04
     $  - 5.452530682596082d-06*zm05
     $  + 1.150575014362998d-05*zm06
     $  + 1.016453997622842d-05*zm07
     $  - 6.070586173694334d-06*zm08
     $  - 4.275106673900843d-06*zm09
     $  - 4.947753247467259d-07*zm10
     $  - 9.609894795039070d-09*zm11
     $  - 1.187185735377699d-11*zm12
     $  + 3.362622328029144d-13*zm13
     $  - HZ2(1,1)*CbH2YZ(0,3,14)

      CbH4YZ(0,3,2,2,15) = 
     $  + 7.851796130644030d-13*zm01
     $  + 1.621114871398478d-09*zm02
     $  + 1.179392618061047d-07*zm03
     $  + 1.490074907051436d-06*zm04
     $  + 4.161529395202537d-06*zm05
     $  - 1.513521803264526d-06*zm06
     $  - 9.029338163246617d-06*zm07
     $  - 7.200197336780222d-07*zm08
     $  + 4.091520151827418d-06*zm09
     $  + 1.305329782651674d-06*zm10
     $  + 9.369684749784303d-08*zm11
     $  + 1.173999649075456d-09*zm12
     $  - 7.632082095751821d-12*zm13
     $  + 1.164995347800545d-12*zm14
     $  - 7.846118765401337d-14*zm15
     $  - HZ2(1,1)*CbH2YZ(0,3,15)

      CbH4YZ(0,3,2,2,16) = 
     $  - 4.778301231213176d-14*zm01
     $  - 1.778118459642125d-10*zm02
     $  - 2.043969787249902d-08*zm03
     $  - 3.981908900124354d-07*zm04
     $  - 1.963173551010995d-06*zm05
     $  - 1.562247775738706d-06*zm06
     $  + 4.146463411715922d-06*zm07
     $  + 3.788767205159209d-06*zm08
     $  - 1.793179136875121d-06*zm09
     $  - 1.839418004139497d-06*zm10
     $  - 3.421727930039108d-07*zm11
     $  - 1.610598619453638d-08*zm12
     $  - 1.249597626116098d-10*zm13
     $  + 4.155565231155318d-14*zm14
     $  - 4.191492308081733d-15*zm15
     $  - HZ2(1,1)*CbH2YZ(0,3,16)

      CbH4YZ(0,3,2,2,17) = 
     $  + 2.368907966150355d-15*zm01
     $  + 1.813920773888944d-11*zm02
     $  + 3.241785263393049d-09*zm03
     $  + 9.378145115039805d-08*zm04
     $  + 7.255072920622768d-07*zm05
     $  + 1.484363890511549d-06*zm06
     $  - 7.660746965932127d-07*zm07
     $  - 3.248496261088249d-06*zm08
     $  - 5.272559597574811d-07*zm09
     $  + 1.491501152082029d-06*zm10
     $  + 6.613450853393076d-07*zm11
     $  + 7.951587407610570d-08*zm12
     $  + 2.540128158142658d-09*zm13
     $  + 1.200004560428658d-11*zm14
     $  + 1.328033417413777d-13*zm15
     $  - 1.661608660574926d-14*zm16
     $  + 9.862334842545256d-16*zm17
     $  - HZ2(1,1)*CbH2YZ(0,3,17)

      CbH4YZ(0,3,3,2,01) = 
     $  + HZ1(1)*CbH3YZ(0,3,3,1)
     $  + HZ2(0,1)*CbH2YZ(0,3,1)

      CbH4YZ(0,3,3,2,02) = 
     $  + HZ1(1)*CbH3YZ(0,3,3,2)
     $  + HZ2(0,1)*CbH2YZ(0,3,2)

      CbH4YZ(0,3,3,2,03) = 
     $  + 2.343750000000000d-02*zm02
     $  - 2.343750000000000d-02*zm03
     $  + HZ1(1)*CbH3YZ(0,3,3,3)
     $  + HZ2(0,1)*CbH2YZ(0,3,3)

      CbH4YZ(0,3,3,2,04) = 
     $  - 2.307128906250000d-02*zm02
     $  + 2.966308593750000d-02*zm03
     $  - 6.591796875000000d-03*zm04
     $  + HZ1(1)*CbH3YZ(0,3,3,4)
     $  + HZ2(0,1)*CbH2YZ(0,3,4)

      CbH4YZ(0,3,3,2,05) = 
     $  + 1.232666015625000d-02*zm02
     $  - 1.502929687500000d-02*zm03
     $  + 3.361816406250000d-03*zm04
     $  - 6.591796875000000d-04*zm05
     $  + HZ1(1)*CbH3YZ(0,3,3,5)
     $  + HZ2(0,1)*CbH2YZ(0,3,5)

      CbH4YZ(0,3,3,2,06) = 
     $  - 4.634857177734375d-03*zm02
     $  + 2.904510498046875d-03*zm03
     $  + 1.750946044921875d-03*zm04
     $  - 2.059936523437500d-05*zm05
     $  + HZ1(1)*CbH3YZ(0,3,3,6)
     $  + HZ2(0,1)*CbH2YZ(0,3,6)

      CbH4YZ(0,3,3,2,07) = 
     $  + 1.357666327028858d-03*zm02
     $  + 1.026605100047831d-03*zm03
     $  - 2.198877139967315d-03*zm04
     $  - 1.582787961375956d-04*zm05
     $  - 3.152964066485969d-05*zm06
     $  + 4.414149693080357d-06*zm07
     $  + HZ1(1)*CbH3YZ(0,3,3,7)
     $  + HZ2(0,1)*CbH2YZ(0,3,7)

      CbH4YZ(0,3,3,2,08) = 
     $  - 3.271298749106270d-04*zm02
     $  - 1.089398350034441d-03*zm03
     $  + 6.999875817980085d-04*zm04
     $  + 6.852967398507254d-04*zm05
     $  + 3.103699002947126d-05*zm06
     $  + 2.069132668631417d-07*zm07
     $  + HZ1(1)*CbH3YZ(0,3,3,8)
     $  + HZ2(0,1)*CbH2YZ(0,3,8)

      CbH4YZ(0,3,3,2,09) = 
     $  + 6.708817822592599d-05*zm02
     $  + 5.167589868818010d-04*zm03
     $  + 1.847390617643083d-04*zm04
     $  - 6.002553871699741d-04*zm05
     $  - 1.634499856403895d-04*zm06
     $  - 5.255596978323800d-06*zm07
     $  + 4.207236426217215d-07*zm08
     $  - 4.598072596958705d-08*zm09
     $  + HZ1(1)*CbH3YZ(0,3,3,9)
     $  + HZ2(0,1)*CbH2YZ(0,3,9)

      CbH4YZ(0,3,3,2,10) = 
     $  - 1.199191702263695d-05*zm02
     $  - 1.773647591471672d-04*zm03
     $  - 3.109479642340115d-04*zm04
     $  + 2.062989931021417d-04*zm05
     $  + 2.600612672311919d-04*zm06
     $  + 3.358900653464453d-05*zm07
     $  + 3.577013100896562d-07*zm08
     $  - 2.327774252210344d-09*zm09
     $  + HZ1(1)*CbH3YZ(0,3,3,10)
     $  + HZ2(0,1)*CbH2YZ(0,3,10)

      CbH4YZ(0,3,3,2,11) = 
     $  + 1.901672930932804d-06*zm02
     $  + 4.939056018096919d-05*zm03
     $  + 1.881629864794093d-04*zm04
     $  + 4.579159995230902d-05*zm05
     $  - 1.957489013751042d-04*zm06
     $  - 8.359250558678768d-05*zm07
     $  - 5.901789462024515d-06*zm08
     $  + 1.747433839469198d-09*zm09
     $  - 5.899593146097843d-09*zm10
     $  + 5.290396027750783d-10*zm11
     $  + HZ1(1)*CbH3YZ(0,3,3,11)
     $  + HZ2(0,1)*CbH2YZ(0,3,11)

      CbH4YZ(0,3,3,2,12) = 
     $  - 2.712485238353346d-07*zm02
     $  - 1.174995304514690d-05*zm03
     $  - 7.988520596303384d-05*zm04
     $  - 1.004908631356166d-04*zm05
     $  + 6.851980408998327d-05*zm06
     $  + 1.009198150585417d-04*zm07
     $  + 2.206891458390836d-05*zm08
     $  + 8.869018290772453d-07*zm09
     $  + 1.807551976148184d-09*zm10
     $  + 2.755414597786866d-11*zm11
     $  + HZ1(1)*CbH3YZ(0,3,3,12)
     $  + HZ2(0,1)*CbH2YZ(0,3,12)

      CbH4YZ(0,3,3,2,13) = 
     $  + 3.518414800345433d-08*zm02
     $  + 2.460567714834009d-06*zm03
     $  + 2.722116667472113d-05*zm04
     $  + 7.062649488627721d-05*zm05
     $  + 1.348388638885010d-05*zm06
     $  - 6.982627251508573d-05*zm07
     $  - 3.888122001429894d-05*zm08
     $  - 4.999765804902522d-06*zm09
     $  - 1.194654896057128d-07*zm10
     $  - 6.532582580272592d-10*zm11
     $  + 8.362811408934768d-11*zm12
     $  - 6.358649071815845d-12*zm13
     $  + HZ1(1)*CbH3YZ(0,3,3,13)
     $  + HZ2(0,1)*CbH2YZ(0,3,13)

      CbH4YZ(0,3,3,2,14) = 
     $  - 4.187187445886462d-09*zm02
     $  - 4.626213387890738d-07*zm03
     $  - 7.888534344097399d-06*zm04
     $  - 3.463384946765987d-05*zm05
     $  - 3.531401686675296d-05*zm06
     $  + 2.459593703589869d-05*zm07
     $  + 4.037188525648350d-05*zm08
     $  + 1.232057185258048d-05*zm09
     $  + 9.996652155041713d-07*zm10
     $  + 1.513652031317305d-08*zm11
     $  + 1.366022740522519d-11*zm12
     $  - 3.362622328029144d-13*zm13
     $  + HZ1(1)*CbH3YZ(0,3,3,14)
     $  + HZ2(0,1)*CbH2YZ(0,3,14)

      CbH4YZ(0,3,3,2,15) = 
     $  + 4.605189139244956d-10*zm02
     $  + 7.920154106206757d-08*zm03
     $  + 2.008645489763755d-06*zm04
     $  + 1.363556893096359d-05*zm05
     $  + 2.750587339338833d-05*zm06
     $  + 4.424776408060642d-06*zm07
     $  - 2.643322063311594d-05*zm08
     $  - 1.767556520143710d-05*zm09
     $  - 3.363736317554520d-06*zm10
     $  - 1.802297707087927d-07*zm11
     $  - 1.781129707554200d-09*zm12
     $  + 7.880750693381763d-12*zm13
     $  - 1.188840281819855d-12*zm14
     $  + 7.846118765401337d-14*zm15
     $  + HZ1(1)*CbH3YZ(0,3,3,15)
     $  + HZ2(0,1)*CbH2YZ(0,3,15)

      CbH4YZ(0,3,3,2,16) = 
     $  - 4.709405073360621d-11*zm02
     $  - 1.247830768399934d-08*zm03
     $  - 4.589966668958478d-07*zm04
     $  - 4.570596885065423d-06*zm05
     $  - 1.498178140751383d-05*zm06
     $  - 1.318557472029653d-05*zm07
     $  + 9.336081431216923d-06*zm08
     $  + 1.661977465976354d-05*zm09
     $  + 6.409893677235218d-06*zm10
     $  + 8.138008933363815d-07*zm11
     $  + 2.973794249738752d-08*zm12
     $  + 1.865099110450207d-10*zm13
     $  - 3.664561846494315d-14*zm14
     $  + 4.191492308081733d-15*zm15
     $  + HZ1(1)*CbH3YZ(0,3,3,16)
     $  + HZ2(0,1)*CbH2YZ(0,3,16)

      CbH4YZ(0,3,3,2,17) = 
     $  + 4.501381745213533d-12*zm02
     $  + 1.824174339101263d-09*zm03
     $  + 9.553982085280394d-08*zm04
     $  + 1.347679438583223d-06*zm05
     $  + 6.577641390728178d-06*zm06
     $  + 1.107779253485925d-05*zm07
     $  + 1.566402555733786d-06*zm08
     $  - 1.045943829675696d-05*zm09
     $  - 8.012902076653357d-06*zm10
     $  - 2.012177304775801d-06*zm11
     $  - 1.778177271761920d-07*zm12
     $  - 4.531070450883955d-09*zm13
     $  - 1.781861257153253d-11*zm14
     $  - 1.379819377397294d-13*zm15
     $  + 1.691585185890460d-14*zm16
     $  - 9.862334842545256d-16*zm17
     $  + HZ1(1)*CbH3YZ(0,3,3,17)
     $  + HZ2(0,1)*CbH2YZ(0,3,17)

      CbH4YZ(3,0,0,2,01) = 
     $  + HZ3(0,0,1)*CbH1YZ(3,1)
     $  + HZ3(0,1,1)*CbH1YZ(3,1)
     $  + HZ3(1,0,1)*CbH1YZ(3,1)
     $  + HZ3(1,1,1)*CbH1YZ(3,1)

      CbH4YZ(3,0,0,2,02) = 
     $  + 2.812500000000000d-01*zm01
     $  - 2.812500000000000d-01*zm02
     $  + HZ3(0,0,1)*CbH1YZ(3,2)
     $  + HZ3(0,1,1)*CbH1YZ(3,2)
     $  + HZ3(1,0,1)*CbH1YZ(3,2)
     $  + HZ3(1,1,1)*CbH1YZ(3,2)

      CbH4YZ(3,0,0,2,03) = 
     $  - 1.933593750000000d-01*zm01
     $  + 2.460937500000000d-01*zm02
     $  - 5.273437500000000d-02*zm03
     $  + HZ3(0,0,1)*CbH1YZ(3,3)
     $  + HZ3(0,1,1)*CbH1YZ(3,3)
     $  + HZ3(1,0,1)*CbH1YZ(3,3)
     $  + HZ3(1,1,1)*CbH1YZ(3,3)

      CbH4YZ(3,0,0,2,04) = 
     $  + 7.543945312500000d-02*zm01
     $  - 7.800292968750000d-02*zm02
     $  + 8.789062500000000d-03*zm03
     $  - 6.225585937500000d-03*zm04
     $  + HZ3(0,0,1)*CbH1YZ(3,4)
     $  + HZ3(0,1,1)*CbH1YZ(3,4)
     $  + HZ3(1,0,1)*CbH1YZ(3,4)
     $  + HZ3(1,1,1)*CbH1YZ(3,4)

      CbH4YZ(3,0,0,2,05) = 
     $  - 2.095642089843750d-02*zm01
     $  - 1.977539062500000d-03*zm02
     $  + 1.895141601562500d-02*zm03
     $  + 4.394531250000000d-03*zm04
     $  - 4.119873046875000d-04*zm05
     $  + HZ3(0,0,1)*CbH1YZ(3,5)
     $  + HZ3(0,1,1)*CbH1YZ(3,5)
     $  + HZ3(1,0,1)*CbH1YZ(3,5)
     $  + HZ3(1,1,1)*CbH1YZ(3,5)

      CbH4YZ(3,0,0,2,06) = 
     $  + 4.518539428710937d-03*zm01
     $  + 1.268371582031250d-02*zm02
     $  - 9.540252685546875d-03*zm03
     $  - 7.385559082031250d-03*zm04
     $  - 2.725982666015625d-04*zm05
     $  - 3.845214843750000d-06*zm06
     $  + HZ3(0,0,1)*CbH1YZ(3,6)
     $  + HZ3(0,1,1)*CbH1YZ(3,6)
     $  + HZ3(1,0,1)*CbH1YZ(3,6)
     $  + HZ3(1,1,1)*CbH1YZ(3,6)

      CbH4YZ(3,0,0,2,07) = 
     $  - 7.906919206891741d-04*zm01
     $  - 6.653079986572265d-03*zm02
     $  - 1.542818886893136d-03*zm03
     $  + 7.163134983607700d-03*zm04
     $  + 1.800384521484375d-03*zm05
     $  + 2.152633666992187d-05*zm06
     $  + 1.544952392578125d-06*zm07
     $  + HZ3(0,0,1)*CbH1YZ(3,7)
     $  + HZ3(0,1,1)*CbH1YZ(3,7)
     $  + HZ3(1,0,1)*CbH1YZ(3,7)
     $  + HZ3(1,1,1)*CbH1YZ(3,7)

      CbH4YZ(3,0,0,2,08) = 
     $  + 1.153844654038765d-04*zm01
     $  + 2.243319350845959d-03*zm02
     $  + 3.890160565473595d-03*zm03
     $  - 2.817074863278135d-03*zm04
     $  - 3.075663240588441d-03*zm05
     $  - 3.528375674267204d-04*zm06
     $  - 3.331500656750737d-06*zm07
     $  + 4.279022661659529d-08*zm08
     $  + HZ3(0,0,1)*CbH1YZ(3,8)
     $  + HZ3(0,1,1)*CbH1YZ(3,8)
     $  + HZ3(1,0,1)*CbH1YZ(3,8)
     $  + HZ3(1,1,1)*CbH1YZ(3,8)

      CbH4YZ(3,0,0,2,09) = 
     $  - 1.431452020264227d-05*zm01
     $  - 5.845584540951008d-04*zm02
     $  - 2.458401652014985d-03*zm03
     $  - 5.072749694999383d-04*zm04
     $  + 2.518213281826097d-03*zm05
     $  + 9.852619572561614d-04*zm06
     $  + 6.074273951199589d-05*zm07
     $  + 3.427018577086334d-07*zm08
     $  - 1.108463929623973d-08*zm09
     $  + HZ3(0,0,1)*CbH1YZ(3,9)
     $  + HZ3(0,1,1)*CbH1YZ(3,9)
     $  + HZ3(1,0,1)*CbH1YZ(3,9)
     $  + HZ3(1,1,1)*CbH1YZ(3,9)

      CbH4YZ(3,0,0,2,10) = 
     $  + 1.534310962907079d-06*zm01
     $  + 1.263305192612350d-04*zm02
     $  + 1.033094752382258d-03*zm03
     $  + 1.342687693359900d-03*zm04
     $  - 9.447627176131520d-04*zm05
     $  - 1.289813077145693d-03*zm06
     $  - 2.596968560194482d-04*zm07
     $  - 9.360306473237084d-06*zm08
     $  - 1.383099064187475d-08*zm09
     $  - 4.877241290345483d-10*zm10
     $  + HZ3(0,0,1)*CbH1YZ(3,10)
     $  + HZ3(0,1,1)*CbH1YZ(3,10)
     $  + HZ3(1,0,1)*CbH1YZ(3,10)
     $  + HZ3(1,1,1)*CbH1YZ(3,10)

      CbH4YZ(3,0,0,2,11) = 
     $  - 1.443240910041815d-07*zm01
     $  - 2.352743564672687d-05*zm02
     $  - 3.390256461112992d-04*zm03
     $  - 9.634113384186115d-04*zm04
     $  - 1.805170249933435d-04*zm05
     $  + 9.479753609857278d-04*zm06
     $  + 4.982831584161807d-04*zm07
     $  + 5.907555446660616d-05*zm08
     $  + 1.291328015853990d-06*zm09
     $  + 2.703860228332764d-10*zm10
     $  + 9.699059384209769d-11*zm11
     $  + HZ3(0,0,1)*CbH1YZ(3,11)
     $  + HZ3(0,1,1)*CbH1YZ(3,11)
     $  + HZ3(1,0,1)*CbH1YZ(3,11)
     $  + HZ3(1,1,1)*CbH1YZ(3,11)

      CbH4YZ(3,0,0,2,12) = 
     $  + 1.209782455513981d-08*zm01
     $  + 3.869757012274848d-06*zm02
     $  + 9.291469827553249d-05*zm03
     $  + 4.688718646614452d-04*zm04
     $  + 5.023731936500287d-04*zm05
     $  - 3.459697432748405d-04*zm06
     $  - 5.510848988010222d-04*zm07
     $  - 1.589119955375222d-04*zm08
     $  - 1.191345261496620d-05*zm09
     $  - 1.613810019114502d-07*zm10
     $  - 1.458549152438495d-10*zm11
     $  + 5.661341413666143d-12*zm12
     $  + HZ3(0,0,1)*CbH1YZ(3,12)
     $  + HZ3(0,1,1)*CbH1YZ(3,12)
     $  + HZ3(1,0,1)*CbH1YZ(3,12)
     $  + HZ3(1,1,1)*CbH1YZ(3,12)

      CbH4YZ(3,0,0,2,13) = 
     $  - 9.152103523543046d-10*zm01
     $  - 5.718235618469732d-07*zm02
     $  - 2.207647648719872d-05*zm03
     $  - 1.801923121639104d-04*zm04
     $  - 3.945970345797244d-04*zm05
     $  - 6.886227151125393d-05*zm06
     $  + 3.767797702786712d-04*zm07
     $  + 2.435694275664058d-04*zm08
     $  + 4.376496297418292d-05*zm09
     $  + 2.168003112279123d-06*zm10
     $  + 1.865446559703880d-08*zm11
     $  + 1.605649656979671d-11*zm12
     $  - 9.393458856091589d-13*zm13
     $  + HZ3(0,0,1)*CbH1YZ(3,13)
     $  + HZ3(0,1,1)*CbH1YZ(3,13)
     $  + HZ3(1,0,1)*CbH1YZ(3,13)
     $  + HZ3(1,1,1)*CbH1YZ(3,13)

      CbH4YZ(3,0,0,2,14) = 
     $  + 6.293272409663106d-11*zm01
     $  + 7.686201037988780d-08*zm02
     $  + 4.657964187921840d-06*zm03
     $  + 5.825785066199586d-05*zm04
     $  + 2.135880456803431d-04*zm05
     $  + 1.991433122314432d-04*zm06
     $  - 1.349006684590475d-04*zm07
     $  - 2.404680258190597d-04*zm08
     $  - 8.929796379694925d-05*zm09
     $  - 1.069465296277285d-05*zm10
     $  - 3.607726892630700d-07*zm11
     $  - 2.014573477276632d-09*zm12
     $  + 6.625736796245896d-13*zm13
     $  - 6.681198657913456d-14*zm14
     $  + HZ3(0,0,1)*CbH1YZ(3,14)
     $  + HZ3(0,1,1)*CbH1YZ(3,14)
     $  + HZ3(1,0,1)*CbH1YZ(3,14)
     $  + HZ3(1,1,1)*CbH1YZ(3,14)

      CbH4YZ(3,0,0,2,15) = 
     $  - 3.948820093967175d-12*zm01
     $  - 9.485904993906272d-09*zm02
     $  - 8.874813158310633d-07*zm03
     $  - 1.641515550706661d-05*zm04
     $  - 9.222173286947825d-05*zm05
     $  - 1.673661186291012d-04*zm06
     $  - 2.767049750153122d-05*zm07
     $  + 1.561451944092198d-04*zm08
     $  + 1.176492031303077d-04*zm09
     $  + 2.835880373515686d-05*zm10
     $  + 2.361614319230415d-06*zm11
     $  + 5.545823759499373d-08*zm12
     $  + 2.019603577340168d-10*zm13
     $  - 1.247451353370347d-13*zm14
     $  + 9.699872100084071d-15*zm15
     $  + HZ3(0,0,1)*CbH1YZ(3,15)
     $  + HZ3(0,1,1)*CbH1YZ(3,15)
     $  + HZ3(1,0,1)*CbH1YZ(3,15)
     $  + HZ3(1,1,1)*CbH1YZ(3,15)

      CbH4YZ(3,0,0,2,16) = 
     $  + 2.285848362064537d-13*zm01
     $  + 1.082696391220340d-09*zm02
     $  + 1.545876559875512d-07*zm03
     $  + 4.123897723683016d-06*zm04
     $  + 3.373258082519473d-05*zm05
     $  + 9.822687353363462d-05*zm06
     $  + 8.243697161540978d-05*zm07
     $  - 5.512125300177551d-05*zm08
     $  - 1.070482095516029d-04*zm09
     $  - 4.801249404602347d-05*zm10
     $  - 8.008574335530035d-06*zm11
     $  - 4.775094223415808d-07*zm12
     $  - 7.935140873991568d-09*zm13
     $  - 1.876855814809380d-11*zm14
     $  - 1.298054945284806d-14*zm15
     $  + 8.005170067599970d-16*zm16
     $  + HZ3(0,0,1)*CbH1YZ(3,16)
     $  + HZ3(0,1,1)*CbH1YZ(3,16)
     $  + HZ3(1,0,1)*CbH1YZ(3,16)
     $  + HZ3(1,1,1)*CbH1YZ(3,16)

      CbH4YZ(3,0,0,2,17) = 
     $  - 1.246366234515635d-14*zm01
     $  - 1.149622957614133d-10*zm02
     $  - 2.485084034044671d-08*zm03
     $  - 9.386524639979976d-07*zm04
     $  - 1.080743702975477d-05*zm05
     $  - 4.640188124026401d-05*zm06
     $  - 7.301799112985010d-05*zm07
     $  - 1.157011367383068d-05*zm08
     $  + 6.687721055314744d-05*zm09
     $  + 5.670791012194338d-05*zm10
     $  + 1.703841281815931d-05*zm11
     $  + 2.047151259022046d-06*zm12
     $  + 8.929191606068496d-08*zm13
     $  + 1.063031134939355d-09*zm14
     $  + 1.651725572105252d-12*zm15
     $  + 1.708853231428238d-15*zm16
     $  - 1.047873077020433d-16*zm17
     $  + HZ3(0,0,1)*CbH1YZ(3,17)
     $  + HZ3(0,1,1)*CbH1YZ(3,17)
     $  + HZ3(1,0,1)*CbH1YZ(3,17)
     $  + HZ3(1,1,1)*CbH1YZ(3,17)

      CbH4YZ(3,0,2,2,01) = 
     $  - HZ3(0,1,1)*CbH1YZ(3,1)
     $  - HZ3(1,1,1)*CbH1YZ(3,1)

      CbH4YZ(3,0,2,2,02) = 
     $  - HZ3(0,1,1)*CbH1YZ(3,2)
     $  - HZ3(1,1,1)*CbH1YZ(3,2)

      CbH4YZ(3,0,2,2,03) = 
     $  + 3.515625000000000d-02*zm01
     $  - 7.031250000000000d-02*zm02
     $  + 3.515625000000000d-02*zm03
     $  - HZ3(0,1,1)*CbH1YZ(3,3)
     $  - HZ3(1,1,1)*CbH1YZ(3,3)

      CbH4YZ(3,0,2,2,04) = 
     $  - 2.636718750000000d-02*zm01
     $  + 5.932617187500000d-02*zm02
     $  - 3.955078125000000d-02*zm03
     $  + 6.591796875000000d-03*zm04
     $  - HZ3(0,1,1)*CbH1YZ(3,4)
     $  - HZ3(1,1,1)*CbH1YZ(3,4)

      CbH4YZ(3,0,2,2,05) = 
     $  + 1.038208007812500d-02*zm01
     $  - 1.977539062500000d-02*zm02
     $  + 8.898925781250000d-03*zm03
     $  + 4.943847656250000d-04*zm05
     $  - HZ3(0,1,1)*CbH1YZ(3,5)
     $  - HZ3(1,1,1)*CbH1YZ(3,5)

      CbH4YZ(3,0,2,2,06) = 
     $  - 2.780914306640625d-03*zm01
     $  + 1.235961914062500d-03*zm02
     $  + 5.561828613281250d-03*zm03
     $  - 3.707885742187500d-03*zm04
     $  - 3.089904785156250d-04*zm05
     $  - HZ3(0,1,1)*CbH1YZ(3,6)
     $  - HZ3(1,1,1)*CbH1YZ(3,6)

      CbH4YZ(3,0,2,2,07) = 
     $  + 5.605970110212053d-04*zm01
     $  + 1.767866952078683d-03*zm02
     $  - 3.939628601074218d-03*zm03
     $  + 3.752027239118303d-04*zm04
     $  + 1.191820417131696d-03*zm05
     $  + 4.634857177734375d-05*zm06
     $  - 2.207074846540178d-06*zm07
     $  - HZ3(0,1,1)*CbH1YZ(3,7)
     $  - HZ3(1,1,1)*CbH1YZ(3,7)

      CbH4YZ(3,0,2,2,08) = 
     $  - 8.980035781860351d-05*zm01
     $  - 9.820103645324707d-04*zm02
     $  + 6.517767906188964d-04*zm03
     $  + 1.709103584289550d-03*zm04
     $  - 9.993910789489746d-04*zm05
     $  - 2.867817878723144d-04*zm06
     $  - 2.896785736083984d-06*zm07
     $  - HZ3(0,1,1)*CbH1YZ(3,8)
     $  - HZ3(1,1,1)*CbH1YZ(3,8)

      CbH4YZ(3,0,2,2,09) = 
     $  + 1.184578452791486d-05*zm01
     $  + 3.201982804707118d-04*zm02
     $  + 3.937214612960815d-04*zm03
     $  - 1.189854741096496d-03*zm04
     $  - 8.690357208251953d-05*zm05
     $  + 4.965573549270629d-04*zm06
     $  + 5.431473255157470d-05*zm07
     $  + 1.034566334315708d-07*zm08
     $  + 1.724277223859514d-08*zm09
     $  - HZ3(0,1,1)*CbH1YZ(3,9)
     $  - HZ3(1,1,1)*CbH1YZ(3,9)

      CbH4YZ(3,0,2,2,10) = 
     $  - 1.322951700006212d-06*zm01
     $  - 7.903569511004856d-05*zm02
     $  - 3.388153016567230d-04*zm03
     $  + 2.862386405467987d-04*zm04
     $  + 5.852412432432174d-04*zm05
     $  - 2.795036137104034d-04*zm06
     $  - 1.641391217708587d-04*zm07
     $  - 8.643801723207746d-06*zm08
     $  - 1.939811876841953d-08*zm09
     $  - HZ3(0,1,1)*CbH1YZ(3,10)
     $  - HZ3(1,1,1)*CbH1YZ(3,10)

      CbH4YZ(3,0,2,2,11) = 
     $  + 1.281333917921239d-07*zm01
     $  + 1.608571364217764d-05*zm02
     $  + 1.486688575332427d-04*zm03
     $  + 8.952725581921540d-05*zm04
     $  - 4.099676012992858d-04*zm05
     $  - 7.926568626002831d-05*zm06
     $  + 1.907965456220236d-04*zm07
     $  + 4.280247810212048d-05*zm08
     $  + 1.221552442807655d-06*zm09
     $  + 2.909717815262930d-09*zm10
     $  - 1.587118808325234d-10*zm11
     $  - HZ3(0,1,1)*CbH1YZ(3,11)
     $  - HZ3(1,1,1)*CbH1YZ(3,11)

      CbH4YZ(3,0,2,2,12) = 
     $  - 1.098418475261756d-08*zm01
     $  - 2.819880277716687d-06*zm02
     $  - 4.823184622052524d-05*zm03
     $  - 1.183804144550647d-04*zm04
     $  + 1.226511527784168d-04*zm05
     $  + 2.171669346280395d-04*zm06
     $  - 8.036713348701596d-05*zm07
     $  - 8.040859696588345d-05*zm08
     $  - 9.441670595801302d-06*zm09
     $  - 1.574884767511061d-07*zm10
     $  - 7.274294538157326d-11*zm11
     $  - HZ3(0,1,1)*CbH1YZ(3,12)
     $  - HZ3(1,1,1)*CbH1YZ(3,12)

      CbH4YZ(3,0,2,2,13) = 
     $  + 8.451280481350439d-10*zm01
     $  + 4.372401676442172d-07*zm02
     $  + 1.282958934278032d-05*zm03
     $  + 6.485107023164036d-05*zm04
     $  + 1.806799419432786d-05*zm05
     $  - 1.532257765232186d-04*zm06
     $  - 4.390550990445682d-05*zm07
     $  + 7.278263225389009d-05*zm08
     $  + 2.631383371868191d-05*zm09
     $  + 1.829562016000336d-06*zm10
     $  + 1.853266423875158d-08*zm11
     $  - 1.487923882804907d-11*zm12
     $  + 1.589662267953961d-12*zm13
     $  - HZ3(0,1,1)*CbH1YZ(3,13)
     $  - HZ3(1,1,1)*CbH1YZ(3,13)

      CbH4YZ(3,0,2,2,14) = 
     $  - 5.876413668101525d-11*zm01
     $  - 6.102826085601672d-08*zm02
     $  - 2.937575440636446d-06*zm03
     $  - 2.563292527694863d-05*zm04
     $  - 4.214781868039589d-05*zm05
     $  + 5.304335559984402d-05*zm06
     $  + 8.545987163961399d-05*zm07
     $  - 2.312076844189468d-05*zm08
     $  - 3.704564797604986d-05*zm09
     $  - 7.236818931474141d-06*zm10
     $  - 3.185955902896536d-07*zm11
     $  - 1.989566791864848d-09*zm12
     $  - 3.099841422510224d-13*zm13
     $  - HZ3(0,1,1)*CbH1YZ(3,14)
     $  - HZ3(1,1,1)*CbH1YZ(3,14)

      CbH4YZ(3,0,2,2,15) = 
     $  + 3.718495802484674d-12*zm01
     $  + 7.762563788877927d-09*zm02
     $  + 5.956740102442001d-07*zm03
     $  + 8.256950895373137d-06*zm04
     $  + 2.772452893780665d-05*zm05
     $  + 1.819239191896411d-06*zm06
     $  - 6.067146752958084d-05*zm07
     $  - 2.189689618868591d-05*zm08
     $  + 2.797785009117470d-05*zm09
     $  + 1.439454834725976d-05*zm10
     $  + 1.740996798657668d-06*zm11
     $  + 5.061050422207460d-08*zm12
     $  + 1.984252777997687d-10*zm13
     $  + 2.508827700745912d-13*zm14
     $  - 1.681311164014572d-14*zm15
     $  - HZ3(0,1,1)*CbH1YZ(3,15)
     $  - HZ3(1,1,1)*CbH1YZ(3,15)

      CbH4YZ(3,0,2,2,16) = 
     $  - 2.178960107455191d-13*zm01
     $  - 9.080776112142144d-10*zm02
     $  - 1.089896051024620d-07*zm03
     $  - 2.285822462565748d-06*zm04
     $  - 1.277900507215321d-05*zm05
     $  - 1.525394203084495d-05*zm06
     $  + 2.333356822912624d-05*zm07
     $  + 3.515637087441386d-05*zm08
     $  - 6.364644513828978d-06*zm09
     $  - 1.663979317284890d-05*zm10
     $  - 4.674001282117791d-06*zm11
     $  - 3.753979434762537d-07*zm12
     $  - 7.415914553025003d-09*zm13
     $  - 1.881060313159382d-11*zm14
     $  + 6.158927473099690d-17*zm15
     $  - HZ3(0,1,1)*CbH1YZ(3,16)
     $  - HZ3(1,1,1)*CbH1YZ(3,16)

      CbH4YZ(3,0,2,2,17) = 
     $  + 1.201700341546683d-14*zm01
     $  + 9.840848347249400d-11*zm02
     $  + 1.823324577917209d-08*zm03
     $  + 5.603800026769368d-07*zm04
     $  + 4.769516060675826d-06*zm05
     $  + 1.180289062133744d-05*zm06
     $  - 1.170360881148383d-06*zm07
     $  - 2.508823626353235d-05*zm08
     $  - 1.057507802951810d-05*zm09
     $  + 1.087014234358611d-05*zm10
     $  + 7.413394185637083d-06*zm11
     $  + 1.324259902140449d-06*zm12
     $  + 7.374753490124025d-08*zm13
     $  + 1.011175612965675d-09*zm14
     $  + 1.684127620973535d-12*zm15
     $  - 2.961417626648767d-15*zm16
     $  + 1.849187782977235d-16*zm17
     $  - HZ3(0,1,1)*CbH1YZ(3,17)
     $  - HZ3(1,1,1)*CbH1YZ(3,17)

      CbH4YZ(3,0,3,2,01) = 
     $  + HZ1(1)*CbH3YZ(3,0,3,1)
     $  + HZ3(0,0,1)*CbH1YZ(3,1)
     $  + HZ3(0,1,1)*CbH1YZ(3,1)

      CbH4YZ(3,0,3,2,02) = 
     $  + HZ1(1)*CbH3YZ(3,0,3,2)
     $  + HZ3(0,0,1)*CbH1YZ(3,2)
     $  + HZ3(0,1,1)*CbH1YZ(3,2)

      CbH4YZ(3,0,3,2,03) = 
     $  + 3.515625000000000d-02*zm02
     $  - 3.515625000000000d-02*zm03
     $  + HZ1(1)*CbH3YZ(3,0,3,3)
     $  + HZ3(0,0,1)*CbH1YZ(3,3)
     $  + HZ3(0,1,1)*CbH1YZ(3,3)

      CbH4YZ(3,0,3,2,04) = 
     $  - 3.515625000000000d-02*zm02
     $  + 4.174804687500000d-02*zm03
     $  - 6.591796875000000d-03*zm04
     $  + HZ1(1)*CbH3YZ(3,0,3,4)
     $  + HZ3(0,0,1)*CbH1YZ(3,4)
     $  + HZ3(0,1,1)*CbH1YZ(3,4)

      CbH4YZ(3,0,3,2,05) = 
     $  + 1.911621093750000d-02*zm02
     $  - 1.862182617187500d-02*zm03
     $  - 4.943847656250000d-04*zm05
     $  + HZ1(1)*CbH3YZ(3,0,3,5)
     $  + HZ3(0,0,1)*CbH1YZ(3,5)
     $  + HZ3(0,1,1)*CbH1YZ(3,5)

      CbH4YZ(3,0,3,2,06) = 
     $  - 7.325134277343750d-03*zm02
     $  + 1.198883056640625d-03*zm03
     $  + 5.825500488281250d-03*zm04
     $  + 3.007507324218750d-04*zm05
     $  + HZ1(1)*CbH3YZ(3,0,3,6)
     $  + HZ3(0,0,1)*CbH1YZ(3,6)
     $  + HZ3(0,1,1)*CbH1YZ(3,6)

      CbH4YZ(3,0,3,2,07) = 
     $  + 2.188535417829241d-03*zm02
     $  + 3.374176025390625d-03*zm03
     $  - 3.985535757882254d-03*zm04
     $  - 1.533034188406808d-03*zm05
     $  - 4.634857177734375d-05*zm06
     $  + 2.207074846540178d-06*zm07
     $  + HZ1(1)*CbH3YZ(3,0,3,7)
     $  + HZ3(0,0,1)*CbH1YZ(3,7)
     $  + HZ3(0,1,1)*CbH1YZ(3,7)

      CbH4YZ(3,0,3,2,08) = 
     $  - 5.379508466136698d-04*zm02
     $  - 2.472105123558822d-03*zm03
     $  + 4.913303316855917d-04*zm04
     $  + 2.183727147627849d-03*zm05
     $  + 3.320425870467205d-04*zm06
     $  + 2.955903812330596d-06*zm07
     $  + HZ1(1)*CbH3YZ(3,0,3,8)
     $  + HZ3(0,0,1)*CbH1YZ(3,8)
     $  + HZ3(0,1,1)*CbH1YZ(3,8)

      CbH4YZ(3,0,3,2,09) = 
     $  + 1.124898754820531d-04*zm02
     $  + 1.089766360059076d-03*zm03
     $  + 9.720341891658549d-04*zm04
     $  - 1.322907361449027d-03*zm05
     $  - 7.917940130039137d-04*zm06
     $  - 5.946835084837310d-05*zm07
     $  - 1.034566334315708d-07*zm08
     $  - 1.724277223859514d-08*zm09
     $  + HZ1(1)*CbH3YZ(3,0,3,9)
     $  + HZ3(0,0,1)*CbH1YZ(3,9)
     $  + HZ3(0,1,1)*CbH1YZ(3,9)

      CbH4YZ(3,0,3,2,10) = 
     $  - 2.047857550942167d-05*zm02
     $  - 3.659302876616010d-04*zm03
     $  - 9.078652861775184d-04*zm04
     $  + 1.990182325243949d-04*zm05
     $  + 8.612404851889123d-04*zm06
     $  + 2.248454420846335d-04*zm07
     $  + 9.151108714998984d-06*zm08
     $  + 1.888083560126168d-08*zm09
     $  + HZ1(1)*CbH3YZ(3,0,3,10)
     $  + HZ3(0,0,1)*CbH1YZ(3,10)
     $  + HZ3(0,1,1)*CbH1YZ(3,10)

      CbH4YZ(3,0,3,2,11) = 
     $  + 3.301867161201855d-06*zm02
     $  + 1.014783561768801d-04*zm03
     $  + 4.889368565929345d-04*zm04
     $  + 3.282883330827133d-04*zm05
     $  - 4.910260994844345d-04*zm06
     $  - 3.761683619842071d-04*zm07
     $  - 5.354340036843725d-05*zm08
     $  - 1.264800170716416d-06*zm09
     $  - 2.909717815262930d-09*zm10
     $  + 1.587118808325234d-10*zm11
     $  + HZ1(1)*CbH3YZ(3,0,3,11)
     $  + HZ3(0,0,1)*CbH1YZ(3,11)
     $  + HZ3(0,1,1)*CbH1YZ(3,11)

      CbH4YZ(3,0,3,2,12) = 
     $  - 4.779012835441481d-07*zm02
     $  - 2.422742228599246d-05*zm03
     $  - 1.985747656976352d-04*zm04
     $  - 3.535450155426823d-04*zm05
     $  + 8.130318429755374d-05*zm06
     $  + 3.547380960359588d-04*zm07
     $  + 1.295035595609857d-04*zm08
     $  + 1.111942007166168d-05*zm09
     $  + 1.607670909041049d-07*zm10
     $  + 7.775279010482211d-11*zm11
     $  + HZ1(1)*CbH3YZ(3,0,3,12)
     $  + HZ3(0,0,1)*CbH1YZ(3,12)
     $  + HZ3(0,1,1)*CbH1YZ(3,12)

      CbH4YZ(3,0,3,2,13) = 
     $  + 6.277398088816466d-08*zm02
     $  + 5.109745099043231d-06*zm03
     $  + 6.640385845084208d-05*zm04
     $  + 2.167670857010181d-04*zm05
     $  + 1.215357257377317d-04*zm06
     $  - 1.941776206128126d-04*zm07
     $  - 1.760554309265686d-04*zm08
     $  - 3.756235521874141d-05*zm09
     $  - 2.065030302593567d-06*zm10
     $  - 1.876519838371073d-08*zm11
     $  + 1.487923882804907d-11*zm12
     $  - 1.589662267953961d-12*zm13
     $  + HZ1(1)*CbH3YZ(3,0,3,13)
     $  + HZ3(0,0,1)*CbH1YZ(3,13)
     $  + HZ3(0,1,1)*CbH1YZ(3,13)

      CbH4YZ(3,0,3,2,14) = 
     $  - 7.551100568560665d-09*zm02
     $  - 9.691050889641184d-07*zm03
     $  - 1.911025403417442d-05*zm04
     $  - 1.004830840023996d-04*zm05
     $  - 1.440328404749553d-04*zm06
     $  + 3.396634984908015d-05*zm07
     $  + 1.511611565997596d-04*zm08
     $  + 6.960120357560061d-05*zm09
     $  + 9.523441771630066d-06*zm10
     $  + 3.486776256746173d-07*zm11
     $  + 2.005021065552425d-09*zm12
     $  + 2.582514910505740d-13*zm13
     $  + HZ1(1)*CbH3YZ(3,0,3,14)
     $  + HZ3(0,0,1)*CbH1YZ(3,14)
     $  + HZ3(0,1,1)*CbH1YZ(3,14)

      CbH4YZ(3,0,3,2,15) = 
     $  + 8.381267418492859d-10*zm02
     $  + 1.674557989648113d-07*zm03
     $  + 4.862693379726812d-06*zm04
     $  + 3.848057318379150d-05*zm05
     $  + 9.687493107694302d-05*zm06
     $  + 4.788701922276414d-05*zm07
     $  - 8.029442591434078d-05*zm08
     $  - 8.244619934591798d-05*zm09
     $  - 2.331788448357615d-05*zm10
     $  - 2.160658671336328d-06*zm11
     $  - 5.414284887856341d-08*zm12
     $  - 1.992908126625479d-10*zm13
     $  - 2.508827700745912d-13*zm14
     $  + 1.681311164014572d-14*zm15
     $  + HZ1(1)*CbH3YZ(3,0,3,15)
     $  + HZ3(0,0,1)*CbH1YZ(3,15)
     $  + HZ3(0,1,1)*CbH1YZ(3,15)

      CbH4YZ(3,0,3,2,16) = 
     $  - 8.638479454650977d-11*zm02
     $  - 2.662830194994884d-08*zm03
     $  - 1.114325291003666d-06*zm04
     $  - 1.271934602527849d-05*zm05
     $  - 4.954894424690897d-05*zm06
     $  - 6.081327049364199d-05*zm07
     $  + 1.454541199419557d-05*zm08
     $  + 6.618688467399072d-05*zm09
     $  + 3.618453969937025d-05*zm10
     $  + 6.852122073724326d-06*zm11
     $  + 4.458235513334072d-07*zm12
     $  + 7.799900927667834d-09*zm13
     $  + 1.884953840033772d-11*zm14
     $  + 4.972763663465676d-16*zm15
     $  + HZ1(1)*CbH3YZ(3,0,3,16)
     $  + HZ3(0,0,1)*CbH1YZ(3,16)
     $  + HZ3(0,1,1)*CbH1YZ(3,16)

      CbH4YZ(3,0,3,2,17) = 
     $  + 8.312693215207696d-12*zm02
     $  + 3.927764042266225d-09*zm03
     $  + 2.330715044006504d-07*zm04
     $  + 3.726442312283296d-06*zm05
     $  + 2.104441332085405d-05*zm06
     $  + 4.386373879943574d-05*zm07
     $  + 1.974860161531972d-05*zm08
     $  - 3.435275376694771d-05*zm09
     $  - 3.883467116941475d-05*zm10
     $  - 1.354138224768339d-05*zm11
     $  - 1.805685846075351d-06*zm12
     $  - 8.465887537733518d-08*zm13
     $  - 1.050039772952081d-09*zm14
     $  - 1.686533955725875d-12*zm15
     $  + 2.961417626648767d-15*zm16
     $  - 1.849187782977235d-16*zm17
     $  + HZ1(1)*CbH3YZ(3,0,3,17)
     $  + HZ3(0,0,1)*CbH1YZ(3,17)
     $  + HZ3(0,1,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,2,2,01) = 
     $  + HZ3(1,1,1)*CbH1YZ(3,1)

      CbH4YZ(3,2,2,2,02) = 
     $  + HZ3(1,1,1)*CbH1YZ(3,2)

      CbH4YZ(3,2,2,2,03) = 
     $  + HZ3(1,1,1)*CbH1YZ(3,3)

      CbH4YZ(3,2,2,2,04) = 
     $  + 1.318359375000000d-02*zm01
     $  - 3.955078125000000d-02*zm02
     $  + 3.955078125000000d-02*zm03
     $  - 1.318359375000000d-02*zm04
     $  + HZ3(1,1,1)*CbH1YZ(3,4)

      CbH4YZ(3,2,2,2,05) = 
     $  - 7.910156250000000d-03*zm01
     $  + 2.373046875000000d-02*zm02
     $  - 2.373046875000000d-02*zm03
     $  + 7.910156250000000d-03*zm04
     $  + HZ3(1,1,1)*CbH1YZ(3,5)

      CbH4YZ(3,2,2,2,06) = 
     $  + 2.471923828125000d-03*zm01
     $  - 4.943847656250000d-03*zm02
     $  + 4.943847656250000d-03*zm04
     $  - 2.471923828125000d-03*zm05
     $  + HZ3(1,1,1)*CbH1YZ(3,6)

      CbH4YZ(3,2,2,2,07) = 
     $  - 5.296979631696428d-04*zm01
     $  - 5.296979631696428d-04*zm02
     $  + 4.237583705357142d-03*zm03
     $  - 4.237583705357142d-03*zm04
     $  + 5.296979631696428d-04*zm05
     $  + 5.296979631696428d-04*zm06
     $  + HZ3(1,1,1)*CbH1YZ(3,7)

      CbH4YZ(3,2,2,2,08) = 
     $  + 8.690357208251953d-05*zm01
     $  + 6.952285766601562d-04*zm02
     $  - 1.651167869567871d-03*zm03
     $  + 1.651167869567871d-03*zm05
     $  - 6.952285766601562d-04*zm06
     $  - 8.690357208251953d-05*zm07
     $  + HZ3(1,1,1)*CbH1YZ(3,8)

      CbH4YZ(3,2,2,2,09) = 
     $  - 1.158714294433593d-05*zm01
     $  - 2.665042877197265d-04*zm02
     $  + 1.042842864990234d-04*zm03
     $  + 1.100778579711914d-03*zm04
     $  - 1.100778579711914d-03*zm05
     $  - 1.042842864990234d-04*zm06
     $  + 2.665042877197265d-04*zm07
     $  + 1.158714294433593d-05*zm08
     $  + HZ3(1,1,1)*CbH1YZ(3,9)

      CbH4YZ(3,2,2,2,10) = 
     $  + 1.303553581237792d-06*zm01
     $  + 7.039189338684082d-05*zm02
     $  + 1.746761798858642d-04*zm03
     $  - 5.657422542572021d-04*zm04
     $  + 5.657422542572021d-04*zm06
     $  - 1.746761798858642d-04*zm07
     $  - 7.039189338684082d-05*zm08
     $  - 1.303553581237792d-06*zm09
     $  + HZ3(1,1,1)*CbH1YZ(3,10)

      CbH4YZ(3,2,2,2,11) = 
     $  - 1.269695046660187d-07*zm01
     $  - 1.485543204592419d-05*zm02
     $  - 1.058925668914596d-04*zm03
     $  + 1.013216647234829d-04*zm04
     $  + 3.306285901503129d-04*zm05
     $  - 3.306285901503129d-04*zm06
     $  - 1.013216647234829d-04*zm07
     $  + 1.058925668914596d-04*zm08
     $  + 1.485543204592419d-05*zm09
     $  + 1.269695046660187d-07*zm10
     $  + HZ3(1,1,1)*CbH1YZ(3,11)

      CbH4YZ(3,2,2,2,12) = 
     $  + 1.091144180723599d-08*zm01
     $  + 2.662391800965581d-06*zm02
     $  + 3.879017562472394d-05*zm03
     $  + 3.797181748918124d-05*zm04
     $  - 2.030182862654328d-04*zm05
     $  + 2.030182862654328d-04*zm07
     $  - 3.797181748918124d-05*zm08
     $  - 3.879017562472394d-05*zm09
     $  - 2.662391800965581d-06*zm10
     $  - 1.091144180723599d-08*zm11
     $  + HZ3(1,1,1)*CbH1YZ(3,12)

      CbH4YZ(3,2,2,2,13) = 
     $  - 8.393416774796915d-10*zm01
     $  - 4.188314970623660d-07*zm02
     $  - 1.099957268337135d-05*zm03
     $  - 3.853837312148003d-05*zm04
     $  + 5.471668395490109d-05*zm05
     $  + 1.093175387583099d-04*zm06
     $  - 1.093175387583099d-04*zm07
     $  - 5.471668395490109d-05*zm08
     $  + 3.853837312148003d-05*zm09
     $  + 1.099957268337135d-05*zm10
     $  + 4.188314970623660d-07*zm11
     $  + 8.393416774796915d-10*zm12
     $  + HZ3(1,1,1)*CbH1YZ(3,13)

      CbH4YZ(3,2,2,2,14) = 
     $  + 5.845415253876423d-11*zm01
     $  + 5.903869406415187d-08*zm02
     $  + 2.618979850346792d-06*zm03
     $  + 1.839610634547449d-05*zm04
     $  + 5.102170704346036d-06*zm05
     $  - 7.616412404173871d-05*zm06
     $  + 7.616412404173871d-05*zm08
     $  - 5.102170704346036d-06*zm09
     $  - 1.839610634547449d-05*zm10
     $  - 2.618979850346792d-06*zm11
     $  - 5.903869406415187d-08*zm12
     $  - 5.845415253876423d-11*zm13
     $  + HZ3(1,1,1)*CbH1YZ(3,14)

      CbH4YZ(3,2,2,2,15) = 
     $  - 3.719809707012269d-12*zm01
     $  - 7.562373134355943d-09*zm02
     $  - 5.450711559879218d-07*zm03
     $  - 6.515931146818080d-06*zm04
     $  - 1.333003108032114d-05*zm05
     $  + 2.615869504890205d-05*zm06
     $  + 3.877446314852152d-05*zm07
     $  - 3.877446314852152d-05*zm08
     $  - 2.615869504890205d-05*zm09
     $  + 1.333003108032114d-05*zm10
     $  + 6.515931146818080d-06*zm11
     $  + 5.450711559879218d-07*zm12
     $  + 7.562373134355943d-09*zm13
     $  + 3.719809707012269d-12*zm14
     $  + HZ3(1,1,1)*CbH1YZ(3,15)

      CbH4YZ(3,2,2,2,16) = 
     $  + 2.179576000202501d-13*zm01
     $  + 8.892670080826206d-10*zm02
     $  + 1.015736905494370d-07*zm03
     $  + 1.910424519089495d-06*zm04
     $  + 8.105003790035425d-06*zm05
     $  - 1.385851142003957d-06*zm06
     $  - 2.969821274295522d-05*zm07
     $  + 2.969821274295522d-05*zm09
     $  + 1.385851142003957d-06*zm10
     $  - 8.105003790035425d-06*zm11
     $  - 1.910424519089495d-06*zm12
     $  - 1.015736905494370d-07*zm13
     $  - 8.892670080826206d-10*zm14
     $  - 2.179576000202501d-13*zm15
     $  + HZ3(1,1,1)*CbH1YZ(3,16)

      CbH4YZ(3,2,2,2,17) = 
     $  - 1.183480181105430d-14*zm01
     $  - 9.674950480536896d-11*zm02
     $  - 1.722194442143717d-08*zm03
     $  - 4.866329078823889d-07*zm04
     $  - 3.445255014257976d-06*zm05
     $  - 4.389498724255163d-06*zm06
     $  + 1.204050682103490d-05*zm07
     $  + 1.451315373863875d-05*zm08
     $  - 1.451315373863875d-05*zm09
     $  - 1.204050682103490d-05*zm10
     $  + 4.389498724255163d-06*zm11
     $  + 3.445255014257976d-06*zm12
     $  + 4.866329078823889d-07*zm13
     $  + 1.722194442143717d-08*zm14
     $  + 9.674950480536896d-11*zm15
     $  + 1.183480181105430d-14*zm16
     $  + HZ3(1,1,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,2,2,18) = 
     $  + 5.987846154402476d-16*zm01
     $  + 9.799709016295093d-12*zm02
     $  + 2.687471098865074d-09*zm03
     $  + 1.106841086556681d-07*zm04
     $  + 1.191644490636714d-06*zm05
     $  + 3.424288371376946d-06*zm06
     $  - 1.821381101986422d-06*zm07
     $  - 1.196629484532977d-05*zm08
     $  + 1.196629484532977d-05*zm10
     $  + 1.821381101986422d-06*zm11
     $  - 3.424288371376946d-06*zm12
     $  - 1.191644490636714d-06*zm13
     $  - 1.106841086556681d-07*zm14
     $  - 2.687471098865074d-09*zm15
     $  - 9.799709016295093d-12*zm16
     $  - 5.987846154402476d-16*zm17
     $  + HZ3(1,1,1)*CbH1YZ(3,18)

      CbH4YZ(3,3,0,2,01) = 
     $  + HZ2(0,1)*CbH2YZ(3,3,1)
     $  + HZ2(1,1)*CbH2YZ(3,3,1)
     $  + HZ3(0,0,1)*CbH1YZ(3,1)
     $  + HZ3(1,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,3,0,2,02) = 
     $  + HZ2(0,1)*CbH2YZ(3,3,2)
     $  + HZ2(1,1)*CbH2YZ(3,3,2)
     $  + HZ3(0,0,1)*CbH1YZ(3,2)
     $  + HZ3(1,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,3,0,2,03) = 
     $  + 7.031250000000000d-02*zm02
     $  - 7.031250000000000d-02*zm03
     $  + HZ2(0,1)*CbH2YZ(3,3,3)
     $  + HZ2(1,1)*CbH2YZ(3,3,3)
     $  + HZ3(0,0,1)*CbH1YZ(3,3)
     $  + HZ3(1,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,3,0,2,04) = 
     $  - 7.250976562500000d-02*zm02
     $  + 7.910156250000000d-02*zm03
     $  - 6.591796875000000d-03*zm04
     $  + HZ2(0,1)*CbH2YZ(3,3,4)
     $  + HZ2(1,1)*CbH2YZ(3,3,4)
     $  + HZ3(0,0,1)*CbH1YZ(3,4)
     $  + HZ3(1,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,3,0,2,05) = 
     $  + 4.086914062500000d-02*zm02
     $  - 3.065185546875000d-02*zm03
     $  - 9.887695312500000d-03*zm04
     $  - 3.295898437500000d-04*zm05
     $  + HZ2(0,1)*CbH2YZ(3,3,5)
     $  + HZ2(1,1)*CbH2YZ(3,3,5)
     $  + HZ3(0,0,1)*CbH1YZ(3,5)
     $  + HZ3(1,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,3,0,2,06) = 
     $  - 1.631469726562500d-02*zm02
     $  - 3.584289550781250d-03*zm03
     $  + 1.804504394531250d-02*zm04
     $  + 1.853942871093750d-03*zm05
     $  + HZ2(0,1)*CbH2YZ(3,3,6)
     $  + HZ2(1,1)*CbH2YZ(3,3,6)
     $  + HZ3(0,0,1)*CbH1YZ(3,6)
     $  + HZ3(1,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,3,0,2,07) = 
     $  + 5.101874215262276d-03*zm02
     $  + 1.076169695172991d-02*zm03
     $  - 9.424209594726562d-03*zm04
     $  - 6.193052019391741d-03*zm05
     $  - 2.471923828125000d-04*zm06
     $  + 8.828299386160714d-07*zm07
     $  + HZ2(0,1)*CbH2YZ(3,3,7)
     $  + HZ2(1,1)*CbH2YZ(3,3,7)
     $  + HZ3(0,0,1)*CbH1YZ(3,7)
     $  + HZ3(1,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,3,0,2,08) = 
     $  - 1.318203040531703d-03*zm02
     $  - 7.054087093898228d-03*zm03
     $  - 2.516065325055803d-04*zm04
     $  + 7.031740461077008d-03*zm05
     $  + 1.565505777086530d-03*zm06
     $  + 2.665042877197265d-05*zm07
     $  + HZ2(0,1)*CbH2YZ(3,3,8)
     $  + HZ2(1,1)*CbH2YZ(3,3,8)
     $  + HZ3(0,0,1)*CbH1YZ(3,8)
     $  + HZ3(1,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,3,0,2,09) = 
     $  + 2.908077288647087d-04*zm02
     $  + 3.079428076744079d-03*zm03
     $  + 3.610429593494960d-03*zm04
     $  - 3.617154274668012d-03*zm05
     $  - 3.039624861308506d-03*zm06
     $  - 3.213569947651454d-04*zm07
     $  - 2.524341855730329d-06*zm08
     $  - 4.926506353884327d-09*zm09
     $  + HZ2(0,1)*CbH2YZ(3,3,9)
     $  + HZ2(1,1)*CbH2YZ(3,3,9)
     $  + HZ3(0,0,1)*CbH1YZ(3,9)
     $  + HZ3(1,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,3,0,2,10) = 
     $  - 5.601511621961788d-05*zm02
     $  - 1.052177239741597d-03*zm03
     $  - 2.963380771023886d-03*zm04
     $  + 1.066508569887706d-04*zm05
     $  + 2.922924054520470d-03*zm06
     $  + 9.853546002081462d-04*zm07
     $  + 5.642524787357875d-05*zm08
     $  + 2.183673941359228d-07*zm09
     $  + HZ2(0,1)*CbH2YZ(3,3,10)
     $  + HZ2(1,1)*CbH2YZ(3,3,10)
     $  + HZ3(0,0,1)*CbH1YZ(3,10)
     $  + HZ3(1,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,3,0,2,11) = 
     $  + 9.575369925080955d-06*zm02
     $  + 3.003855424489072d-04*zm03
     $  + 1.559857503402155d-03*zm04
     $  + 1.337108552891325d-03*zm05
     $  - 1.476614814832226d-03*zm06
     $  - 1.459291307231435d-03*zm07
     $  - 2.622672565668434d-04*zm08
     $  - 8.736665808148198d-06*zm09
     $  - 1.695949812324679d-08*zm10
     $  + 3.526930685167188d-11*zm11
     $  + HZ2(0,1)*CbH2YZ(3,3,11)
     $  + HZ2(1,1)*CbH2YZ(3,3,11)
     $  + HZ3(0,0,1)*CbH1YZ(3,11)
     $  + HZ3(1,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,3,0,2,12) = 
     $  - 1.470938877415098d-06*zm02
     $  - 7.427785553881321d-05*zm03
     $  - 6.367076322417326d-04*zm04
     $  - 1.264050974431991d-03*zm05
     $  + 9.266009608694872d-05*zm06
     $  + 1.264491532161202d-03*zm07
     $  + 5.580982800706156d-04*zm08
     $  + 6.004061571586745d-05*zm09
     $  + 1.215733951890706d-06*zm10
     $  + 1.143103427424722d-09*zm11
     $  + HZ2(0,1)*CbH2YZ(3,3,12)
     $  + HZ2(1,1)*CbH2YZ(3,3,12)
     $  + HZ3(0,0,1)*CbH1YZ(3,12)
     $  + HZ3(1,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,3,0,2,13) = 
     $  + 2.051046373116010d-07*zm02
     $  + 1.628024726323734d-05*zm03
     $  + 2.165954133995076d-04*zm04
     $  + 7.544980648747719d-04*zm05
     $  + 5.285385277478831d-04*zm06
     $  - 6.278440777382643d-04*zm07
     $  - 6.989026258828222d-04*zm08
     $  - 1.770568002528131d-04*zm09
     $  - 1.215978607272393d-05*zm10
     $  - 1.539987229675121d-07*zm11
     $  - 6.896409107603699d-11*zm12
     $  - 2.890295032643565d-13*zm13
     $  + HZ2(0,1)*CbH2YZ(3,3,13)
     $  + HZ2(1,1)*CbH2YZ(3,3,13)
     $  + HZ3(0,0,1)*CbH1YZ(3,13)
     $  + HZ3(1,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,3,0,2,14) = 
     $  - 2.617534587218669d-08*zm02
     $  - 3.214698892941135d-06*zm03
     $  - 6.381740898793819d-05*zm04
     $  - 3.494562137361974d-04*zm05
     $  - 5.513896724188695d-04*zm06
     $  + 5.301730119423849d-05*zm07
     $  + 5.629259615111242d-04*zm08
     $  + 3.011293396967568d-04*zm09
     $  + 4.859356692050311d-05*zm10
     $  + 2.220065965355211d-06*zm11
     $  + 1.792985068106686d-08*zm12
     $  + 4.243159557565943d-12*zm13
     $  + HZ2(0,1)*CbH2YZ(3,3,14)
     $  + HZ2(1,1)*CbH2YZ(3,3,14)
     $  + HZ3(0,0,1)*CbH1YZ(3,14)
     $  + HZ3(1,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,3,0,2,15) = 
     $  + 3.078795553311915d-09*zm02
     $  + 5.788433200766681d-07*zm03
     $  + 1.668654616924868d-05*zm04
     $  + 1.352329120900897d-04*zm05
     $  + 3.606310946910810d-04*zm06
     $  + 2.190139445297058d-04*zm07
     $  - 2.753368958254406d-04*zm08
     $  - 3.359308371098495d-04*zm09
     $  - 1.086638335031209d-04*zm10
     $  - 1.184257108348083d-05*zm11
     $  - 3.703480295698281d-07*zm12
     $  - 1.933783727600106d-09*zm13
     $  - 2.631525148802600d-13*zm14
     $  + 2.586632560022418d-15*zm15
     $  + HZ2(0,1)*CbH2YZ(3,3,15)
     $  + HZ2(1,1)*CbH2YZ(3,3,15)
     $  + HZ3(0,0,1)*CbH1YZ(3,15)
     $  + HZ3(1,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,3,0,2,16) = 
     $  - 3.357721848309834d-10*zm02
     $  - 9.594251251850603d-08*zm03
     $  - 3.937921980262536d-06*zm04
     $  - 4.544520131122734d-05*zm05
     $  - 1.837180548575249d-04*zm06
     $  - 2.457984279369123d-04*zm07
     $  + 2.744584775865276d-05*zm08
     $  + 2.562509658117630d-04*zm09
     $  + 1.585083254265360d-04*zm10
     $  + 3.412456072881849d-05*zm11
     $  + 2.608968662201356d-06*zm12
     $  + 5.702156432826705d-08*zm13
     $  + 1.944071211437364d-10*zm14
     $  + 1.120943942069872d-14*zm15
     $  + HZ2(0,1)*CbH2YZ(3,3,16)
     $  + HZ2(1,1)*CbH2YZ(3,3,16)
     $  + HZ3(0,0,1)*CbH1YZ(3,16)
     $  + HZ3(1,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,3,0,2,17) = 
     $  + 3.413053520310799d-11*zm02
     $  + 1.474888785013014d-08*zm03
     $  + 8.493045402119573d-07*zm04
     $  + 1.358543158850865d-05*zm05
     $  + 7.852685540458732d-05*zm06
     $  + 1.723939278604910d-04*zm07
     $  + 9.406913219557560d-05*zm08
     $  - 1.237575755827034d-04*zm09
     $  - 1.624147923345177d-04*zm10
     $  - 6.317838881029041d-05*zm11
     $  - 9.554124400458347d-06*zm12
     $  - 5.263683639602384d-07*zm13
     $  - 8.166811271393142d-09*zm14
     $  - 1.830437840554875d-11*zm15
     $  - 1.552652695141565d-16*zm16
     $  - 2.465583710636314d-17*zm17
     $  + HZ2(0,1)*CbH2YZ(3,3,17)
     $  + HZ2(1,1)*CbH2YZ(3,3,17)
     $  + HZ3(0,0,1)*CbH1YZ(3,17)
     $  + HZ3(1,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,3,2,2,01) = 
     $  - HZ2(1,1)*CbH2YZ(3,3,1)
     $  - HZ3(0,1,1)*CbH1YZ(3,1)
     $  - HZ3(1,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,3,2,2,02) = 
     $  - HZ2(1,1)*CbH2YZ(3,3,2)
     $  - HZ3(0,1,1)*CbH1YZ(3,2)
     $  - HZ3(1,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,3,2,2,03) = 
     $  - HZ2(1,1)*CbH2YZ(3,3,3)
     $  - HZ3(0,1,1)*CbH1YZ(3,3)
     $  - HZ3(1,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,3,2,2,04) = 
     $  + 1.318359375000000d-02*zm02
     $  - 2.636718750000000d-02*zm03
     $  + 1.318359375000000d-02*zm04
     $  - HZ2(1,1)*CbH2YZ(3,3,4)
     $  - HZ3(0,1,1)*CbH1YZ(3,4)
     $  - HZ3(1,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,3,2,2,05) = 
     $  - 1.384277343750000d-02*zm02
     $  + 2.768554687500000d-02*zm03
     $  - 1.384277343750000d-02*zm04
     $  - HZ2(1,1)*CbH2YZ(3,3,5)
     $  - HZ3(0,1,1)*CbH1YZ(3,5)
     $  - HZ3(1,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,3,2,2,06) = 
     $  + 7.662963867187500d-03*zm02
     $  - 1.137084960937500d-02*zm03
     $  - 2.471923828125000d-04*zm04
     $  + 3.955078125000000d-03*zm05
     $  - HZ2(1,1)*CbH2YZ(3,3,6)
     $  - HZ3(0,1,1)*CbH1YZ(3,6)
     $  - HZ3(1,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,3,2,2,07) = 
     $  - 2.939823695591517d-03*zm02
     $  + 5.561828613281250d-04*zm03
     $  + 6.912558419363839d-03*zm04
     $  - 3.734370640345982d-03*zm05
     $  - 7.945469447544642d-04*zm06
     $  - HZ2(1,1)*CbH2YZ(3,3,7)
     $  - HZ3(0,1,1)*CbH1YZ(3,7)
     $  - HZ3(1,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,3,2,2,08) = 
     $  + 8.715186800275530d-04*zm02
     $  + 1.986367361886160d-03*zm03
     $  - 4.474292482648577d-03*zm04
     $  - 3.724438803536551d-04*zm05
     $  + 1.864702360970633d-03*zm06
     $  + 1.241479601178850d-04*zm07
     $  - HZ2(1,1)*CbH2YZ(3,3,8)
     $  - HZ3(0,1,1)*CbH1YZ(3,8)
     $  - HZ3(1,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,3,2,2,09) = 
     $  - 2.116722720009940d-04*zm02
     $  - 1.379283836909702d-03*zm03
     $  + 9.536632469722202d-04*zm04
     $  + 2.467233794076102d-03*zm05
     $  - 1.235892942973545d-03*zm06
     $  - 5.781156676156180d-04*zm07
     $  - 1.593232154846191d-05*zm08
     $  - HZ2(1,1)*CbH2YZ(3,3,9)
     $  - HZ3(0,1,1)*CbH1YZ(3,9)
     $  - HZ3(1,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,3,2,2,10) = 
     $  + 4.368456346648080d-05*zm02
     $  + 5.861025197165352d-04*zm03
     $  + 4.794749404702867d-04*zm04
     $  - 1.792696544102260d-03*zm05
     $  - 2.748480652059827d-04*zm06
     $  + 8.190040928976876d-04*zm07
     $  + 1.375404213156018d-04*zm08
     $  + 1.738071441650390d-06*zm09
     $  - HZ2(1,1)*CbH2YZ(3,3,10)
     $  - HZ3(0,1,1)*CbH1YZ(3,10)
     $  - HZ3(1,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,3,2,2,11) = 
     $  - 7.854121942798812d-06*zm02
     $  - 1.910711171758639d-04*zm03
     $  - 5.519205655951004d-04*zm04
     $  + 5.486468685331282d-04*zm05
     $  + 9.784640357285350d-04*zm06
     $  - 4.202024014545725d-04*zm07
     $  - 3.287314541347615d-04*zm08
     $  - 2.716618360250027d-05*zm09
     $  - 1.650603560658244d-07*zm10
     $  - HZ2(1,1)*CbH2YZ(3,3,11)
     $  - HZ3(0,1,1)*CbH1YZ(3,11)
     $  - HZ3(1,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,3,2,2,12) = 
     $  + 1.252699649421038d-06*zm02
     $  + 5.170370167871187d-05*zm03
     $  + 3.022218086793051d-04*zm04
     $  + 9.483775535187163d-05*zm05
     $  - 7.427444890165677d-04*zm06
     $  - 1.635491544404974d-04*zm07
     $  + 3.505495763928085d-04*zm08
     $  + 1.010730161101787d-04*zm09
     $  + 4.641198305195415d-06*zm10
     $  + 1.388728957284580d-08*zm11
     $  - HZ2(1,1)*CbH2YZ(3,3,12)
     $  - HZ3(0,1,1)*CbH1YZ(3,12)
     $  - HZ3(1,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,3,2,2,13) = 
     $  - 1.797068683378450d-07*zm02
     $  - 1.209446338012807d-05*zm03
     $  - 1.214645105907432d-04*zm04
     $  - 2.117667369394247d-04*zm05
     $  + 2.742075389851215d-04*zm06
     $  + 4.108544624330090d-04*zm07
     $  - 1.433708819957984d-04*zm08
     $  - 1.696374769925933d-04*zm09
     $  - 2.584392923434226d-05*zm10
     $  - 7.032462396658026d-07*zm11
     $  - 1.049177096849614d-09*zm12
     $  - HZ2(1,1)*CbH2YZ(3,3,13)
     $  - HZ3(0,1,1)*CbH1YZ(3,13)
     $  - HZ3(1,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,3,2,2,14) = 
     $  + 2.344031955319319d-08*zm02
     $  + 2.507823352125019d-06*zm03
     $  + 3.995448730420196d-05*zm04
     $  + 1.424869592846491d-04*zm05
     $  + 4.641044289631111d-06*zm06
     $  - 3.177040998777304d-04*zm07
     $  - 8.909941606943191d-05*zm08
     $  + 1.487665539976904d-04*zm09
     $  + 6.259584896017057d-05*zm10
     $  + 5.731183167936456d-06*zm11
     $  + 9.610332763214934d-08*zm12
     $  + 7.194357235540213d-11*zm13
     $  - HZ2(1,1)*CbH2YZ(3,3,14)
     $  - HZ3(0,1,1)*CbH1YZ(3,14)
     $  - HZ3(1,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,3,2,2,15) = 
     $  - 2.804481037651329d-09*zm02
     $  - 4.689518621934774d-07*zm03
     $  - 1.130333530422421d-05*zm04
     $  - 6.734512647833735d-05*zm05
     $  - 8.001608329963200d-05*zm06
     $  + 1.320974622017464d-04*zm07
     $  + 1.791807388820940d-04*zm08
     $  - 4.825195359739379d-05*zm09
     $  - 8.353159279015588d-05*zm10
     $  - 1.921553741107474d-05*zm11
     $  - 1.130823980178811d-06*zm12
     $  - 1.198736270079783d-08*zm13
     $  - 4.516911787086327d-12*zm14
     $  - HZ2(1,1)*CbH2YZ(3,3,15)
     $  - HZ3(0,1,1)*CbH1YZ(3,15)
     $  - HZ3(1,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,3,2,2,16) = 
     $  + 3.100109145764950d-10*zm02
     $  + 8.008398565130511d-08*zm03
     $  + 2.830275051148397d-06*zm04
     $  + 2.575184599513898d-05*zm05
     $  + 6.471949497000986d-05*zm06
     $  - 1.096675652543606d-05*zm07
     $  - 1.396711000944711d-04*zm08
     $  - 4.660461181698233d-05*zm09
     $  + 6.305647359960511d-05*zm10
     $  + 3.547785054318149d-05*zm11
     $  + 5.122784474055791d-06*zm12
     $  + 2.019724143106296d-07*zm13
     $  + 1.377131324275639d-09*zm14
     $  + 2.615491200243001d-13*zm15
     $  - HZ2(1,1)*CbH2YZ(3,3,16)
     $  - HZ3(0,1,1)*CbH1YZ(3,16)
     $  - HZ3(1,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,3,2,2,17) = 
     $  - 3.185520769544831d-11*zm02
     $  - 1.261027183179567d-08*zm03
     $  - 6.391961861401747d-07*zm04
     $  - 8.437033172168028d-06*zm05
     $  - 3.493325342198733d-05*zm06
     $  - 2.991087496134595d-05*zm07
     $  + 6.311787021421771d-05*zm08
     $  + 8.035224027885122d-05*zm09
     $  - 1.561277311136700d-05*zm10
     $  - 4.014578315704033d-05*zm11
     $  - 1.252827350159937d-05*zm12
     $  - 1.217072764250511d-06*zm13
     $  - 3.306133083225231d-08*zm14
     $  - 1.467452446518239d-10*zm15
     $  - 1.405382715062698d-14*zm16
     $  - HZ2(1,1)*CbH2YZ(3,3,17)
     $  - HZ3(0,1,1)*CbH1YZ(3,17)
     $  - HZ3(1,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,3,2,2,18) = 
     $  + 3.058562757003624d-12*zm02
     $  + 1.844841744359908d-09*zm03
     $  + 1.319688107056048d-07*zm04
     $  + 2.443792522594341d-06*zm05
     $  + 1.508232492459366d-05*zm06
     $  + 2.891402659930009d-05*zm07
     $  - 9.828002551779143d-06*zm08
     $  - 6.283502028763135d-05*zm09
     $  - 2.390599648746122d-05*zm10
     $  + 2.677122530284548d-05*zm11
     $  + 1.911089947915477d-05*zm12
     $  + 3.845727217461064d-06*zm13
     $  + 2.621854091240379d-07*zm14
     $  + 5.006572448907054d-09*zm15
     $  + 1.458763218793902d-11*zm16
     $  + 7.044524887532325d-16*zm17
     $  - HZ2(1,1)*CbH2YZ(3,3,18)
     $  - HZ3(0,1,1)*CbH1YZ(3,18)
     $  - HZ3(1,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,3,3,2,01) = 
     $  + HZ1(1)*CbH3YZ(3,3,3,1)
     $  + HZ2(0,1)*CbH2YZ(3,3,1)
     $  + HZ3(0,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,3,3,2,02) = 
     $  + HZ1(1)*CbH3YZ(3,3,3,2)
     $  + HZ2(0,1)*CbH2YZ(3,3,2)
     $  + HZ3(0,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,3,3,2,03) = 
     $  + HZ1(1)*CbH3YZ(3,3,3,3)
     $  + HZ2(0,1)*CbH2YZ(3,3,3)
     $  + HZ3(0,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,3,3,2,04) = 
     $  + 1.318359375000000d-02*zm03
     $  - 1.318359375000000d-02*zm04
     $  + HZ1(1)*CbH3YZ(3,3,3,4)
     $  + HZ2(0,1)*CbH2YZ(3,3,4)
     $  + HZ3(0,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,3,3,2,05) = 
     $  - 1.779785156250000d-02*zm03
     $  + 1.779785156250000d-02*zm04
     $  + HZ1(1)*CbH3YZ(3,3,3,5)
     $  + HZ2(0,1)*CbH2YZ(3,3,5)
     $  + HZ3(0,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,3,3,2,06) = 
     $  + 1.285400390625000d-02*zm03
     $  - 8.157348632812500d-03*zm04
     $  - 4.696655273437500d-03*zm05
     $  + HZ1(1)*CbH3YZ(3,3,3,6)
     $  + HZ2(0,1)*CbH2YZ(3,3,6)
     $  + HZ3(0,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,3,3,2,07) = 
     $  - 6.515284946986607d-03*zm03
     $  - 1.430184500558035d-03*zm04
     $  + 7.044982910156250d-03*zm05
     $  + 9.004865373883928d-04*zm06
     $  + HZ1(1)*CbH3YZ(3,3,3,7)
     $  + HZ2(0,1)*CbH2YZ(3,3,7)
     $  + HZ3(0,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,3,3,2,08) = 
     $  + 2.579794611249651d-03*zm03
     $  + 4.456911768232073d-03*zm04
     $  - 4.116746357509068d-03*zm05
     $  - 2.783397265842982d-03*zm06
     $  - 1.365627561296735d-04*zm07
     $  + HZ1(1)*CbH3YZ(3,3,3,8)
     $  + HZ2(0,1)*CbH2YZ(3,3,8)
     $  + HZ3(0,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,3,3,2,09) = 
     $  - 8.448268686022077d-04*zm03
     $  - 3.473039184297834d-03*zm04
     $  + 1.862219401768275d-04*zm05
     $  + 3.333372729165213d-03*zm06
     $  + 7.810975824083600d-04*zm07
     $  + 1.717380114964076d-05*zm08
     $  + HZ1(1)*CbH3YZ(3,3,3,9)
     $  + HZ2(0,1)*CbH2YZ(3,3,9)
     $  + HZ3(0,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,3,3,2,10) = 
     $  + 2.370294928550720d-04*zm03
     $  + 1.816951951810291d-03*zm04
     $  + 1.652952496494565d-03*zm05
     $  - 1.984970697334834d-03*zm06
     $  - 1.544742030756814d-03*zm07
     $  - 1.753745121615273d-04*zm08
     $  - 1.846700906753540d-06*zm09
     $  + HZ1(1)*CbH3YZ(3,3,3,10)
     $  + HZ2(0,1)*CbH2YZ(3,3,10)
     $  + HZ3(0,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,3,3,2,11) = 
     $  - 5.835518434450223d-05*zm03
     $  - 7.461997789221924d-04*zm04
     $  - 1.642433073032986d-03*zm05
     $  + 2.690801227634603d-04*zm06
     $  + 1.611469443161766d-03*zm07
     $  + 5.329481473603805d-04*zm08
     $  + 3.331679802436333d-05*zm09
     $  + 1.735249897102256d-07*zm10
     $  + HZ1(1)*CbH3YZ(3,3,3,11)
     $  + HZ2(0,1)*CbH2YZ(3,3,11)
     $  + HZ3(0,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,3,3,2,12) = 
     $  + 1.282636677943073d-05*zm03
     $  + 2.565952510476209d-04*zm04
     $  + 1.015997808426618d-03*zm05
     $  + 6.490951550023799d-04*zm06
     $  - 9.495952704244039d-04*zm07
     $  - 8.292457044378600d-04*zm08
     $  - 1.501268906784909d-04*zm09
     $  - 5.532233256169340d-06*zm10
     $  - 1.448245912596776d-08*zm11
     $  + HZ1(1)*CbH3YZ(3,3,3,12)
     $  + HZ2(0,1)*CbH2YZ(3,3,12)
     $  + HZ3(0,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,3,3,2,13) = 
     $  - 2.550061178192697d-06*zm03
     $  - 7.657095513292137d-05*zm04
     $  - 4.845719114801921d-04*zm05
     $  - 7.620042851243584d-04*zm06
     $  + 1.743498997977719d-04*zm07
     $  + 7.859173431663637d-04*zm08
     $  + 3.284043804872650d-04*zm09
     $  + 3.620449202270112d-05*zm10
     $  + 8.200101125715570d-07*zm11
     $  + 1.087328991280509d-09*zm12
     $  + HZ1(1)*CbH3YZ(3,3,3,13)
     $  + HZ2(0,1)*CbH2YZ(3,3,13)
     $  + HZ3(0,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,3,3,2,14) = 
     $  + 4.633035253192705d-07*zm03
     $  + 2.029164023126389d-05*zm04
     $  + 1.923103833842896d-04*zm05
     $  + 5.287845839762960d-04*zm06
     $  + 2.667888954826551d-04*zm07
     $  - 4.557818137927152d-04*zm08
     $  - 4.371059761221922d-04*zm09
     $  - 1.079468722704166d-04*zm10
     $  - 7.693972643924683d-06*zm11
     $  - 1.100975787661674d-07*zm12
     $  - 7.419180899150845d-11*zm13
     $  + HZ1(1)*CbH3YZ(3,3,3,14)
     $  + HZ2(0,1)*CbH2YZ(3,3,14)
     $  + HZ3(0,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,3,3,2,15) = 
     $  - 7.755821633783968d-08*zm03
     $  - 4.853568238937388d-06*zm04
     $  - 6.608042127366633d-05*zm05
     $  - 2.809692653537047d-04*zm06
     $  - 3.546882670944242d-04*zm07
     $  + 9.706249920502802d-05*zm08
     $  + 3.858962528095410d-04*zm09
     $  + 1.915533577573353d-04*zm10
     $  + 3.067427095421771d-05*zm11
     $  + 1.469160365540657d-06*zm12
     $  + 1.353444586490464d-08*zm13
     $  + 4.639542876328489d-12*zm14
     $  + HZ1(1)*CbH3YZ(3,3,3,15)
     $  + HZ2(0,1)*CbH2YZ(3,3,15)
     $  + HZ3(0,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,3,3,2,16) = 
     $  + 1.204436332364210d-08*zm03
     $  + 1.060617970441371d-06*zm04
     $  + 2.015181875393463d-05*zm05
     $  + 1.241791305474002d-04*zm06
     $  + 2.678494974893390d-04*zm07
     $  + 1.136986474570121d-04*zm08
     $  - 2.202648882388203d-04*zm09
     $  - 2.281081145443548d-04*zm10
     $  - 7.059034421230515d-05*zm11
     $  - 7.731264045849046d-06*zm12
     $  - 2.556093021443883d-07*zm13
     $  - 1.535970200671935d-09*zm14
     $  - 2.677764800248787d-13*zm15
     $  + HZ1(1)*CbH3YZ(3,3,3,16)
     $  + HZ2(0,1)*CbH2YZ(3,3,16)
     $  + HZ3(0,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,3,3,2,17) = 
     $  - 1.745036264779866d-09*zm03
     $  - 2.137470121960455d-07*zm04
     $  - 5.548120666402245d-06*zm05
     $  - 4.756957224885597d-05*zm06
     $  - 1.545236157961135d-04*zm07
     $  - 1.666559092586790d-04*zm08
     $  + 5.122499403661660d-05*zm09
     $  + 1.906777601406093d-04*zm10
     $  + 1.081334373788716d-04*zm11
     $  + 2.267612518377174d-05*zm12
     $  + 1.759284377700821d-06*zm13
     $  + 4.094691086601355d-08*zm14
     $  + 1.619757258389403d-10*zm15
     $  + 1.434969719590334d-14*zm16
     $  + HZ1(1)*CbH3YZ(3,3,3,17)
     $  + HZ2(0,1)*CbH2YZ(3,3,17)
     $  + HZ3(0,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,3,3,2,18) = 
     $  + 2.370245558780849d-10*zm03
     $  + 4.002747175189688d-08*zm04
     $  + 1.396480334695174d-06*zm05
     $  + 1.619523555997120d-05*zm06
     $  + 7.429466619819253d-05*zm07
     $  + 1.343211630069045d-04*zm08
     $  + 4.988604953075883d-05*zm09
     $  - 1.072788223798945d-04*zm10
     $  - 1.184187052870692d-04*zm11
     $  - 4.356593065706736d-05*zm12
     $  - 6.498136978697646d-06*zm13
     $  - 3.661590007315795d-07*zm14
     $  - 6.088864657139871d-09*zm15
     $  - 1.595799480865458d-11*zm16
     $  - 7.176609729173556d-16*zm17
     $  + HZ1(1)*CbH3YZ(3,3,3,18)
     $  + HZ2(0,1)*CbH2YZ(3,3,18)
     $  + HZ3(0,0,1)*CbH1YZ(3,18)
      endif

      return
      end

      subroutine fillcoeff2dhpl321(iflag,n,CsH2YZ,CsH3YZ,CsH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsH2YZ,CsH3YZ,CsH4YZ
      real(dp):: 
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
      dimension CsH2YZ(0:3,0:3,22),CsH3YZ(0:3,0:3,0:3,22),
     $          CsH4YZ(0:3,0:3,0:3,0:3,22)
      common /tpz/  
     $     tz01,tz02,tz03,tz04,tz05,tz06,tz07,tz08,tz09,tz10,
     $     tz11,tz12,tz13,tz14,tz15,tz16,tz17,tz18,tz19,tz20,tz21
!$omp threadprivate(/tpz/)
      
* 2001-04-20:16:55:01.chpl 
* <- HPLexpand.321tcut 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.2) return

      if (n.eq.3) then
      CsH3YZ(2,3,1,02) = 
     $  + 1.271177752013422d-01
     $  - 1.389905854659051d-01*tz01
     $  + 1.328962491341869d-02*tz02
     $  - 1.603854525936234d-03*tz03
     $  + 2.131576791845240d-04*tz04
     $  - 2.989851065051721d-05*tz05
     $  + 4.341920903187438d-06*tz06
     $  - 6.460820904724618d-07*tz07
     $  + 9.788928873545235d-08*tz08
     $  - 1.503968816575284d-08*tz09
     $  + 2.336496513438171d-09*tz10
     $  - 3.662899844140558d-10*tz11
     $  + 5.785701524942945d-11*tz12
     $  - 9.197097805005088d-12*tz13
     $  + 1.469982525756198d-12*tz14
     $  - 2.360603790468797d-13*tz15
     $  + 3.806507400973038d-14*tz16
     $  - 6.160397630037750d-15*tz17
     $  + 1.000227479759130d-15*tz18
     $  - 1.628265045514259d-16*tz19

      CsH3YZ(2,3,1,03) = 
     $  - 1.976122852752233d-02
     $  + 1.787343185370242d-02*tz01
     $  + 2.141178813544960d-03*tz02
     $  - 2.891643942110493d-04*tz03
     $  + 4.100450081337254d-05*tz04
     $  - 6.002286600219772d-06*tz05
     $  + 8.985830375842590d-07*tz06
     $  - 1.368003781361765d-07*tz07
     $  + 2.109980916059613d-08*tz08
     $  - 3.288528920461157d-09*tz09
     $  + 5.169368461205368d-10*tz10
     $  - 8.184128521088087d-11*tz11
     $  + 1.303570875867757d-11*tz12
     $  - 2.087147548272499d-12*tz13
     $  + 3.356841888994552d-13*tz14
     $  - 5.420334703899632d-14*tz15
     $  + 8.782934020648240d-15*tz16
     $  - 1.427585838687177d-15*tz17
     $  + 2.326907728784685d-16*tz18

      CsH3YZ(2,3,1,04) = 
     $  + 1.734064230275385d-03
     $  - 9.567669570189350d-04*tz01
     $  - 6.560583334593310d-04*tz02
     $  - 1.387908301458499d-04*tz03
     $  + 2.016008993064897d-05*tz04
     $  - 3.003165782605045d-06*tz05
     $  + 4.556422818627776d-07*tz06
     $  - 7.010350684453921d-08*tz07
     $  + 1.090569099417403d-08*tz08
     $  - 1.711830522596629d-09*tz09
     $  + 2.707049628121143d-10*tz10
     $  - 4.307776847415096d-11*tz11
     $  + 6.891874053660811d-12*tz12
     $  - 1.107733126313849d-12*tz13
     $  + 1.787693847558027d-13*tz14
     $  - 2.895364565271600d-14*tz15
     $  + 4.704275163137335d-15*tz16
     $  - 7.665021033143601d-16*tz17
     $  + 1.252124247320243d-16*tz18

      CsH3YZ(2,3,1,05) = 
     $  - 5.310807945725451d-05
     $  - 4.446844648639702d-05*tz01
     $  + 4.135292293789848d-05*tz02
     $  + 4.624905057607084d-05*tz03
     $  + 1.148147894817547d-05*tz04
     $  - 1.738057692917934d-06*tz05
     $  + 2.670021859694367d-07*tz06
     $  - 4.149108696079809d-08*tz07
     $  + 6.507486687813282d-09*tz08
     $  - 1.028452041753583d-09*tz09
     $  + 1.635820766703014d-10*tz10
     $  - 2.616119468214108d-11*tz11
     $  + 4.203635501964770d-12*tz12
     $  - 6.782299071074073d-13*tz13
     $  + 1.098245958777668d-13*tz14
     $  - 1.784088059458531d-14*tz15
     $  + 2.906536333821858d-15*tz16
     $  - 4.747424077703165d-16*tz17

      CsH3YZ(2,3,1,06) = 
     $  - 4.654008869779762d-06
     $  + 7.590711223948209d-06*tz01
     $  + 4.591616143679665d-06*tz02
     $  - 2.315233030109710d-06*tz03
     $  - 4.236948446408059d-06*tz04
     $  - 1.127468190180616d-06*tz05
     $  + 1.750224432095715d-07*tz06
     $  - 2.742995655507783d-08*tz07
     $  + 4.332625521690110d-09*tz08
     $  - 6.888351073478770d-10*tz09
     $  + 1.101261671545744d-10*tz10
     $  - 1.769055676517290d-11*tz11
     $  + 2.853642062014215d-12*tz12
     $  - 4.620039252083520d-13*tz13
     $  + 7.504109418383888d-14*tz14
     $  - 1.222387648351377d-14*tz15
     $  + 1.996385406283144d-15*tz16
     $  - 3.268151282232528d-16*tz17

      CsH3YZ(2,3,1,07) = 
     $  + 3.457210326332692d-07
     $  + 2.042846851707213d-07*tz01
     $  - 6.107112634763809d-07*tz02
     $  - 6.342875146331062d-07*tz03
     $  + 1.378680086057512d-07*tz04
     $  + 4.500764988947557d-07*tz05
     $  + 1.239313269512578d-07*tz06
     $  - 1.956358993358094d-08*tz07
     $  + 3.108962377607537d-09*tz08
     $  - 4.968635160492970d-10*tz09
     $  + 7.979328031552290d-11*tz10
     $  - 1.286845349365465d-11*tz11
     $  + 2.083016520317602d-12*tz12
     $  - 3.382832863021905d-13*tz13
     $  + 5.509797133106017d-14*tz14
     $  - 8.997628545487400d-15*tz15
     $  + 1.472801619940899d-15*tz16
     $  - 2.415988401784877d-16*tz17

      CsH3YZ(2,3,1,08) = 
     $  + 2.762561092403343d-08
     $  - 6.880743191357026d-08*tz01
     $  - 3.957757118963818d-08*tz02
     $  + 6.043042776932315d-08*tz03
     $  + 9.181703296229490d-08*tz04
     $  - 6.059390642244313d-09*tz05
     $  - 5.272258324972236d-08*tz06
     $  - 1.473458978434748d-08*tz07
     $  + 2.353969050136647d-09*tz08
     $  - 3.779239901191859d-10*tz09
     $  + 6.093458400715164d-11*tz10
     $  - 9.861623417550949d-12*tz11
     $  + 1.601286363495895d-12*tz12
     $  - 2.607764255046168d-13*tz13
     $  + 4.258081535084940d-14*tz14
     $  - 6.969371391184073d-15*tz15
     $  + 1.143162085666570d-15*tz16
     $  - 1.878793330455423d-16*tz17

      CsH3YZ(2,3,1,09) = 
     $  - 2.762438850193176d-09
     $  - 1.265133505379206d-09*tz01
     $  + 7.795234124018670d-09*tz02
     $  + 8.598216873027958d-09*tz03
     $  - 6.618031712186089d-09*tz04
     $  - 1.364586125731291d-08*tz05
     $  - 3.207830986398424d-10*tz06
     $  + 6.620080707254018d-09*tz07
     $  + 1.856286900013618d-09*tz08
     $  - 2.992274938621527d-10*tz09
     $  + 4.841752587288997d-11*tz10
     $  - 7.860582639433086d-12*tz11
     $  + 1.279962803036909d-12*tz12
     $  - 2.089756338943597d-13*tz13
     $  + 3.420072701823231d-14*tz14
     $  - 5.609426720182514d-15*tz15
     $  + 9.218510454338425d-16*tz16
     $  - 1.517699192054417d-16*tz17

      CsH3YZ(2,3,1,10) = 
     $  - 2.194946551636724d-10
     $  + 6.740616611656362d-10*tz01
     $  + 4.172309447359918d-10*tz02
     $  - 1.081433497600964d-09*tz03
     $  - 1.696299607206996d-09*tz04
     $  + 7.480758528833931d-10*tz05
     $  + 2.074620967553779d-09*tz06
     $  + 1.690869649519303d-10*tz07
     $  - 8.754730566595765d-10*tz08
     $  - 2.445100298435820d-10*tz09
     $  + 3.968981717562602d-11*tz10
     $  - 6.461956868342892d-12*tz11
     $  + 1.054907162679701d-12*tz12
     $  - 1.726288746274860d-13*tz13
     $  + 2.831146483882659d-14*tz14
     $  - 4.652381058844980d-15*tz15
     $  + 7.659105880534854d-16*tz16
     $  - 1.262998116258898d-16*tz17

      CsH3YZ(2,3,1,11) = 
     $  + 2.455027744907508d-11
     $  + 1.088147743459175d-11*tz01
     $  - 9.810807718328471d-11*tz02
     $  - 1.237415650009422d-10*tz03
     $  + 1.616997248889283d-10*tz04
     $  + 3.195975183074412d-10*tz05
     $  - 8.335325730910178d-11*tz06
     $  - 3.216649619255360d-10*tz07
     $  - 3.906973835788549d-11*tz08
     $  + 1.205239580968122d-10*tz09
     $  + 3.336523133570562d-11*tz10
     $  - 5.446223654881573d-12*tz11
     $  + 8.911557755142924d-13*tz12
     $  - 1.461395560640429d-13*tz13
     $  + 2.401326403527847d-14*tz14
     $  - 3.953015836303016d-15*tz15
     $  + 6.518299353685222d-16*tz16
     $  - 1.076484150292498d-16*tz17

      CsH3YZ(2,3,1,12) = 
     $  + 2.034882006879706d-12
     $  - 6.971834825896436d-12*tz01
     $  - 5.211006254001285d-12*tz02
     $  + 1.760495133100337d-11*tz03
     $  + 3.026353658299051d-11*tz04
     $  - 2.459294908165086d-11*tz05
     $  - 5.896083411853576d-11*tz06
     $  + 8.653422419089466d-12*tz07
     $  + 5.072137193726880d-11*tz08
     $  + 7.616848501118708d-12*tz09
     $  - 1.713271303519263d-11*tz10
     $  - 4.685597982526580d-12*tz11
     $  + 7.683232931459630d-13*tz12
     $  - 1.262401403292978d-13*tz13
     $  + 2.078017814098095d-14*tz14
     $  - 3.426359108476997d-15*tz15
     $  + 5.658354133454476d-16*tz16

      CsH3YZ(2,3,1,13) = 
     $  - 2.327423523022032d-13
     $  - 1.254678369679285d-13*tz01
     $  + 1.247538238446469d-12*tz02
     $  + 1.873988797607272d-12*tz03
     $  - 3.308932972482367d-12*tz04
     $  - 6.786893872321197d-12*tz05
     $  + 3.725730020952518d-12*tz06
     $  + 1.076697908672182d-11*tz07
     $  - 7.333703835560073d-13*tz08
     $  - 8.114589672997722d-12*tz09
     $  - 1.390893504451503d-12*tz10
     $  + 2.500006424093784d-12*tz11
     $  + 6.738695646434945d-13*tz12
     $  - 1.109179229829213d-13*tz13
     $  + 1.828787034246538d-14*tz14
     $  - 3.019959043956685d-15*tz15
     $  + 4.994173723430136d-16*tz16

      CsH3YZ(2,3,1,14) = 
     $  - 2.065973613391058d-14
     $  + 7.499692713853229d-14*tz01
     $  + 7.233366057419485d-14*tz02
     $  - 2.766232612805440d-13*tz03
     $  - 5.332570178574311d-13*tz04
     $  + 6.175878229756244d-13*tz05
     $  + 1.453554792253312d-12*tz06
     $  - 5.558692018899579d-13*tz07
     $  - 1.956314846474754d-12*tz08
     $  + 2.104196771772570d-14*tz09
     $  + 1.314511977091169d-12*tz10
     $  + 2.463859642886052d-13*tz11
     $  - 3.728222835530159d-13*tz12
     $  - 9.887767233818509d-14*tz13
     $  + 1.632739114010337d-14*tz14
     $  - 2.699991901360980d-15*tz15
     $  + 4.470837260677367d-16*tz16

      CsH3YZ(2,3,1,15) = 
     $  + 2.301607275204872d-15
     $  + 1.698316468505297d-15*tz01
     $  - 1.607658725312652d-14*tz02
     $  - 2.920678354920734d-14*tz03
     $  + 6.242533108546047d-14*tz04
     $  + 1.366465031842407d-13*tz05
     $  - 1.132200594329590d-13*tz06
     $  - 3.026094387392614d-13*tz07
     $  + 8.077739639811645d-14*tz08
     $  + 3.546170171384606d-13*tz09
     $  + 1.160407488923818d-14*tz10
     $  - 2.152612609964063d-13*tz11
     $  - 4.298755150314754d-14*tz12
     $  + 5.662864787578391d-14*tz13
     $  + 1.475913703114543d-14*tz14
     $  - 2.443838469586114d-15*tz15
     $  + 4.051587310198266d-16*tz16

      CsH3YZ(2,3,1,16) = 
     $  + 2.219454216053685d-16
     $  - 8.297694432915468d-16*tz01
     $  - 1.055560765297037d-15*tz02
     $  + 4.269229466242350d-15*tz03
     $  + 9.300567655364718d-15*tz04
     $  - 1.368100065196384d-14*tz05
     $  - 3.299516419299647d-14*tz06
     $  + 2.034769026286363d-14*tz07
     $  + 6.180132190313791d-14*tz08
     $  - 1.125039828555977d-14*tz09
     $  - 6.422334389225392d-14*tz10
     $  - 4.278980780282256d-15*tz11
     $  + 3.558554363143079d-14*tz12
     $  + 7.443117198323458d-15*tz13
     $  - 8.737531479779891d-15*tz14
     $  - 2.235900110719257d-15*tz15
     $  + 3.711055459698154d-16*tz16

      CsH3YZ(2,3,1,17) = 
     $  + 2.093736345750262d-16*tz02
     $  + 4.603950338254971d-16*tz03
     $  - 1.122133074268705d-15*tz04
     $  - 2.652994305520726d-15*tz05
     $  + 2.902933796713004d-15*tz06
     $  + 7.658530841093721d-15*tz07
     $  - 3.581994044710902d-15*tz08
     $  - 1.244825543723970d-14*tz09
     $  + 1.457476808304778d-15*tz10
     $  + 1.163056696520624d-14*tz11
     $  + 1.089038995747963d-15*tz12
     $  - 5.931850022876990d-15*tz13
     $  - 1.284171472711814d-15*tz14
     $  + 1.366584075078284d-15*tz15
     $  + 3.431266984572026d-16*tz16

      CsH3YZ(2,3,1,18) = 
     $  - 1.604052394616694d-16*tz04
     $  + 2.810086932387939d-16*tz05
     $  + 7.081700937595654d-16*tz06
     $  - 5.984390590045156d-16*tz07
     $  - 1.727335793245568d-15*tz08
     $  + 6.166900665111921d-16*tz09
     $  + 2.481451441374517d-15*tz10
     $  - 1.635246936682817d-16*tz11
     $  - 2.107139197957544d-15*tz12
     $  - 2.434029107385481d-16*tz13
     $  + 9.960898008798297d-16*tz14
     $  + 2.212863114679326d-16*tz15
     $  - 2.162864626563659d-16*tz16

      CsH3YZ(2,3,1,19) = 
     $  - 1.804191812181675d-16*tz07
     $  + 1.202609210047088d-16*tz08
     $  + 3.811213428955940d-16*tz09
     $  - 1.035026844007183d-16*tz10
     $  - 4.906849200253468d-16*tz11
     $  + 3.820248537142516d-16*tz13
     $  - 1.683626359957024d-16*tz15

      CsH3YZ(3,2,1,01) = 
     $  + 3.554862660343892d-01
     $  - 3.733420351546513d-01*tz01
     $  + 1.934227531208478d-02*tz02
     $  - 1.638210073282368d-03*tz03
     $  + 1.690199663409635d-04*tz04
     $  - 1.943887306576524d-05*tz05
     $  + 2.396283431709894d-06*tz06
     $  - 3.100486089562243d-07*tz07
     $  + 4.157875397123377d-08*tz08
     $  - 5.731962476383159d-09*tz09
     $  + 8.077450794108086d-10*tz10
     $  - 1.158835878395434d-10*tz11
     $  + 1.687465154171362d-11*tz12
     $  - 2.488341416020160d-12*tz13
     $  + 3.709048703380455d-13*tz14
     $  - 5.580416728733533d-14*tz15
     $  + 8.464722906819654d-15*tz16
     $  - 1.293265096589482d-15*tz17
     $  + 1.988119921479958d-16*tz18

      CsH3YZ(3,2,1,02) = 
     $  + 3.132985740608352d-01
     $  - 3.482654717325442d-02*tz01
     $  + 3.068998956815397d-03*tz02
     $  - 3.248637085035370d-04*tz03
     $  + 3.804357011522906d-05*tz04
     $  - 4.753381238366491d-06*tz05
     $  + 6.215019601517675d-07*tz06
     $  - 8.404782950194623d-08*tz07
     $  + 1.166661228645154d-08*tz08
     $  - 1.653526027869315d-09*tz09
     $  + 2.383834838515699d-10*tz10
     $  - 3.485855901452921d-11*tz11
     $  + 5.159015925959223d-12*tz12
     $  - 7.714502182456119d-13*tz13
     $  + 1.163966228673412d-13*tz14
     $  - 1.770040676180833d-14*tz15
     $  + 2.710428592945471d-15*tz16
     $  - 4.176168694181151d-16*tz17

      CsH3YZ(3,2,1,03) = 
     $  + 1.497539931662031d-02
     $  + 1.873845348219286d-02*tz01
     $  + 1.603651931876759d-03*tz02
     $  - 1.808585016089716d-04*tz03
     $  + 2.212126582914580d-05*tz04
     $  - 2.853657657639492d-06*tz05
     $  + 3.823893625281194d-07*tz06
     $  - 5.273127461153364d-08*tz07
     $  + 7.437074057417598d-09*tz08
     $  - 1.068130815460257d-09*tz09
     $  + 1.557255578780793d-10*tz10
     $  - 2.299167282291959d-11*tz11
     $  + 3.431251229087295d-12*tz12
     $  - 5.168558627777046d-13*tz13
     $  + 7.848895581423717d-14*tz14
     $  - 1.200467223251848d-14*tz15
     $  + 1.847755848655218d-15*tz16
     $  - 2.860247934041468d-16*tz17

      CsH3YZ(3,2,1,04) = 
     $  + 4.844661879509661d-03
     $  - 1.811696648677867d-03*tz01
     $  - 7.389569460804916d-04*tz02
     $  - 1.090235144734803d-04*tz03
     $  + 1.390886670854463d-05*tz04
     $  - 1.850791632352381d-06*tz05
     $  + 2.540189948667470d-07*tz06
     $  - 3.570595444688376d-08*tz07
     $  + 5.115478815045284d-09*tz08
     $  - 7.443953621599992d-10*tz09
     $  + 1.097432463520912d-10*tz10
     $  - 1.635892136607006d-11*tz11
     $  + 2.461864441569941d-12*tz12
     $  - 3.735682130009110d-13*tz13
     $  + 5.709992312397762d-14*tz14
     $  - 8.784171350736040d-15*tz15
     $  + 1.359134600110806d-15*tz16
     $  - 2.113832229598786d-16*tz17

      CsH3YZ(3,2,1,05) = 
     $  - 7.224474437631126d-05
     $  - 6.551729612094564d-05*tz01
     $  + 7.791065416654982d-05*tz02
     $  + 5.146466179033682d-05*tz03
     $  + 9.528726835983097d-06*tz04
     $  - 1.301592686699214d-06*tz05
     $  + 1.823521276494225d-07*tz06
     $  - 2.606224832221865d-08*tz07
     $  + 3.785694771871028d-09*tz08
     $  - 5.573351694309316d-10*tz09
     $  + 8.298870648458445d-11*tz10
     $  - 1.247811469560446d-11*tz11
     $  + 1.892110940723987d-12*tz12
     $  - 2.890405859112540d-13*tz13
     $  + 4.444402017294374d-14*tz14
     $  - 6.873853288607695d-15*tz15
     $  + 1.068701265625091d-15*tz16
     $  - 1.669413509966584d-16*tz17

      CsH3YZ(3,2,1,06) = 
     $  - 2.173983150767964d-05
     $  + 1.922974558874880d-05*tz01
     $  + 4.600011335722878d-06*tz02
     $  - 4.848612167220907d-06*tz03
     $  - 4.630953367435928d-06*tz04
     $  - 9.710290259447900d-07*tz05
     $  + 1.384224742337900d-07*tz06
     $  - 2.006799377258914d-08*tz07
     $  + 2.950094187628140d-09*tz08
     $  - 4.387720921711982d-10*tz09
     $  + 6.591301150858358d-11*tz10
     $  - 9.987271395770439d-12*tz11
     $  + 1.524737355524542d-12*tz12
     $  - 2.343314081028755d-13*tz13
     $  + 3.622711338037648d-14*tz14
     $  - 5.630366921938753d-15*tz15
     $  + 8.792445467377233d-16*tz16
     $  - 1.378994365579413d-16*tz17

      CsH3YZ(3,2,1,07) = 
     $  + 5.974610264647137d-07
     $  + 3.782504514835450d-07*tz01
     $  - 1.293544809650030d-06*tz02
     $  - 6.185535015229340d-07*tz03
     $  + 3.559725719640509d-07*tz04
     $  + 4.849496859241115d-07*tz05
     $  + 1.094545705780703d-07*tz06
     $  - 1.606544630920525d-08*tz07
     $  + 2.386567104053800d-09*tz08
     $  - 3.581724346365604d-10*tz09
     $  + 5.422937818959109d-11*tz10
     $  - 8.273904348925137d-12*tz11
     $  + 1.270927186481558d-12*tz12
     $  - 1.963973289616485d-13*tz13
     $  + 3.051262123663158d-14*tz14
     $  - 4.763426657225863d-15*tz15
     $  + 7.468855610795764d-16*tz16
     $  - 1.175753791124346d-16*tz17

      CsH3YZ(3,2,1,08) = 
     $  + 1.826380341862755d-07
     $  - 2.174014373211668d-07*tz01
     $  - 1.972433105401984d-08*tz02
     $  + 1.181990855343109d-07*tz03
     $  + 9.073413251531738d-08*tz04
     $  - 2.764204178222171d-08*tz05
     $  - 5.615833649015793d-08*tz06
     $  - 1.325321358197057d-08*tz07
     $  + 1.987058401093795d-09*tz08
     $  - 3.006139462840280d-10*tz09
     $  + 4.583581987887476d-11*tz10
     $  - 7.036947179186433d-12*tz11
     $  + 1.086942727920693d-12*tz12
     $  - 1.688062518677949d-13*tz13
     $  + 2.634454770842915d-14*tz14
     $  - 4.129617164621381d-15*tz15
     $  + 6.499325373807175d-16*tz16
     $  - 1.026642587238836d-16*tz17

      CsH3YZ(3,2,1,09) = 
     $  - 5.681269714556042d-09
     $  - 2.494502819314143d-09*tz01
     $  + 1.786477999312606d-08*tz02
     $  + 6.174115779676108d-09*tz03
     $  - 1.270014353926092d-08*tz04
     $  - 1.365269845390952d-08*tz05
     $  + 2.034569067360027d-09*tz06
     $  + 6.985822119513993d-09*tz07
     $  + 1.693002926284112d-09*tz08
     $  - 2.579677622871023d-10*tz09
     $  + 3.958292773562193d-11*tz10
     $  - 6.111296638996401d-12*tz11
     $  + 9.487456574894906d-13*tz12
     $  - 1.480170872769248d-13*tz13
     $  + 2.319589072125166d-14*tz14
     $  - 3.649797904329112d-15*tz15
     $  + 5.764056583204743d-16*tz16

      CsH3YZ(3,2,1,10) = 
     $  - 1.833708822630889d-09
     $  + 2.524248236672288d-09*tz01
     $  - 7.124351703825869d-11*tz02
     $  - 2.047331166632404d-09*tz03
     $  - 1.440055877714952d-09*tz04
     $  + 1.477746521110466d-09*tz05
     $  + 2.092331632162529d-09*tz06
     $  - 1.071703697520154d-10*tz07
     $  - 9.167712362115102d-10*tz08
     $  - 2.254489616914390d-10*tz09
     $  + 3.479141982699619d-11*tz10
     $  - 5.399079214674849d-12*tz11
     $  + 8.420519654539030d-13*tz12
     $  - 1.319217233797353d-13*tz13
     $  + 2.075244960280340d-14*tz14
     $  - 3.276730104227735d-15*tz15
     $  + 5.191475943142633d-16*tz16

      CsH3YZ(3,2,1,11) = 
     $  + 5.765537829593847d-11
     $  + 1.857800026087320d-11*tz01
     $  - 2.329522593203549d-10*tz02
     $  - 6.019753322694131d-11*tz03
     $  + 2.813946164266066d-10*tz04
     $  + 2.924143110875258d-10*tz05
     $  - 1.783754460133764d-10*tz06
     $  - 3.259510658255583d-10*tz07
     $  - 4.826381607817551d-12*tz08
     $  + 1.254081155398551d-10*tz09
     $  + 3.103618785331038d-11*tz10
     $  - 4.838847854071059d-12*tz11
     $  + 7.578712726715271d-13*tz12
     $  - 1.191906660234697d-13*tz13
     $  + 1.881575737810281d-14*tz14
     $  - 2.980537257541775d-15*tz15
     $  + 4.736270400280916d-16*tz16

      CsH3YZ(3,2,1,12) = 
     $  + 2.001785019517400d-11
     $  - 2.987679532277385d-11*tz01
     $  + 3.401285249982219d-12*tz02
     $  + 3.168119133540651d-11*tz03
     $  + 2.254664626614791d-11*tz04
     $  - 4.162304509363436d-11*tz05
     $  - 5.605140181454898d-11*tz06
     $  + 2.172492677941365d-11*tz07
     $  + 5.153284268558836d-11*tz08
     $  + 3.182571180501642d-12*tz09
     $  - 1.773221493637191d-11*tz10
     $  - 4.390237614243251d-12*tz11
     $  + 6.902824321315390d-13*tz12
     $  - 1.089464805301084d-13*tz13
     $  + 1.725461206220580d-14*tz14
     $  - 2.741442133766078d-15*tz15
     $  + 4.368411671629664d-16*tz16

      CsH3YZ(3,2,1,13) = 
     $  - 6.084937539497460d-13
     $  - 1.666568933516797d-13*tz01
     $  + 2.975746494378420d-12*tz02
     $  + 6.755527721719750d-13*tz03
     $  - 5.325608141460106d-12*tz04
     $  - 5.809306358448907d-12*tz05
     $  + 6.329325117447014d-12*tz06
     $  + 1.045945704111761d-11*tz07
     $  - 2.603040481784844d-12*tz08
     $  - 8.255285137553077d-12*tz09
     $  - 7.959867168938734d-13*tz10
     $  + 2.575861116446505d-12*tz11
     $  + 6.352395372894382d-13*tz12
     $  - 1.005883481120497d-13*tz13
     $  + 1.597902175807466d-14*tz14
     $  - 2.545860244191133d-15*tz15
     $  + 4.067274215212837d-16*tz16

      CsH3YZ(3,2,1,14) = 
     $  - 2.294650855573236d-13
     $  + 3.593879176328110d-13*tz01
     $  - 5.999534728596552d-14*tz02
     $  - 4.681918497095840d-13*tz03
     $  - 3.713295128515105d-13*tz04
     $  + 9.536942879607460d-13*tz05
     $  + 1.324690741905475d-12*tz06
     $  - 9.701061420912583d-13*tz07
     $  - 1.925514815324824d-12*tz08
     $  + 2.964150778618700d-13*tz09
     $  + 1.337965178611579d-12*tz10
     $  + 1.642034628842093d-13*tz11
     $  - 3.826646085044344d-13*tz12
     $  - 9.369214484240701d-14*tz13
     $  + 1.492534435218378d-14*tz14
     $  - 2.384171939848240d-15*tz15
     $  + 3.818170618362644d-16*tz16

      CsH3YZ(3,2,1,15) = 
     $  + 6.596064919482268d-15
     $  + 1.872367762607578d-15*tz01
     $  - 3.780640453152969d-14*tz02
     $  - 9.813025009976064d-15*tz03
     $  + 9.392839885029718d-14*tz04
     $  + 1.131976338173847d-13*tz05
     $  - 1.727382748311374d-13*tz06
     $  - 2.852046229775337d-13*tz07
     $  + 1.483105877851705d-13*tz08
     $  + 3.519601673559537d-13*tz09
     $  - 2.989496875217270d-14*tz10
     $  - 2.190975043818894d-13*tz11
     $  - 3.135233015607516d-14*tz12
     $  + 5.793285537748058d-14*tz13
     $  + 1.404741676652725d-14*tz14
     $  - 2.249388879885784d-15*tz15
     $  + 3.610478551939327d-16*tz16

      CsH3YZ(3,2,1,16) = 
     $  + 2.716435037196124d-15
     $  - 4.383300731495686d-15*tz01
     $  + 8.372510523148074d-16*tz02
     $  + 6.791655950564317d-15*tz03
     $  + 6.441911345685765d-15*tz04
     $  - 1.977773342359407d-14*tz05
     $  - 2.941325316357838d-14*tz06
     $  + 3.114779504658621d-14*tz07
     $  + 5.942755164687057d-14*tz08
     $  - 2.244228802332646d-14*tz09
     $  - 6.409393682466816d-14*tz10
     $  + 2.092123703211953d-15*tz11
     $  + 3.620743287834315d-14*tz12
     $  + 5.761043727115101d-15*tz13
     $  - 8.913438264170570d-15*tz14
     $  - 2.136322897121989d-15*tz15
     $  + 3.436284423055493d-16*tz16

      CsH3YZ(3,2,1,17) = 
     $  + 4.807622448202002d-16*tz02
     $  + 1.731213506948204d-16*tz03
     $  - 1.597620593406510d-15*tz04
     $  - 2.187844554037192d-15*tz05
     $  + 4.137431103432104d-15*tz06
     $  + 7.094558284011358d-15*tz07
     $  - 5.562377367039428d-15*tz08
     $  - 1.212562086513067d-14*tz09
     $  + 3.334447055002373d-15*tz10
     $  + 1.165028812971602d-14*tz11
     $  - 6.032291905177956d-15*tz13
     $  - 1.036584740865100d-15*tz14
     $  + 1.390666878364939d-15*tz15
     $  + 3.289596167163652d-16*tz16

      CsH3YZ(3,2,1,18) = 
     $  - 1.148900051583887d-16*tz04
     $  + 3.869384735278297d-16*tz05
     $  + 6.273894315365522d-16*tz06
     $  - 8.504410593335896d-16*tz07
     $  - 1.637277404107711d-15*tz08
     $  + 9.815641777300728d-16*tz09
     $  + 2.438364382631724d-15*tz10
     $  - 4.812075415994241d-16*tz11
     $  - 2.116070704565359d-15*tz12
     $  + 1.012299987573162d-15*tz14
     $  + 1.842691446333386d-16*tz15
     $  - 2.196254105987111d-16*tz16

      CsH3YZ(3,2,1,19) = 
     $  - 1.659493992989305d-16*tz07
     $  + 1.715735429848104d-16*tz08
     $  + 3.666862520384490d-16*tz09
     $  - 1.708822271992645d-16*tz10
     $  - 4.851392498414569d-16*tz11
     $  + 3.843119527615326d-16*tz13
     $  - 1.709812428343503d-16*tz15
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CsH4YZ(0,2,3,1,02) = 
     $  + 6.355888760067112d-02
     $  - 6.949529273295256d-02*tz01
     $  + 6.644812456709349d-03*tz02
     $  - 8.019272629681172d-04*tz03
     $  + 1.065788395922620d-04*tz04
     $  - 1.494925532525860d-05*tz05
     $  + 2.170960451593719d-06*tz06
     $  - 3.230410452362309d-07*tz07
     $  + 4.894464436772617d-08*tz08
     $  - 7.519844082876420d-09*tz09
     $  + 1.168248256719085d-09*tz10
     $  - 1.831449922070279d-10*tz11
     $  + 2.892850762471472d-11*tz12
     $  - 4.598548902502544d-12*tz13
     $  + 7.349912628780990d-13*tz14
     $  - 1.180301895234398d-13*tz15
     $  + 1.903253700486519d-14*tz16
     $  - 3.080198815018875d-15*tz17
     $  + 5.001137398795654d-16*tz18

      CsH4YZ(0,2,3,1,03) = 
     $  - 2.247679807600856d-02
     $  + 2.333163380113895d-02*tz01
     $  - 9.474768429956838d-04*tz02
     $  + 1.040936843383462d-04*tz03
     $  - 1.297654296027464d-05*tz04
     $  + 1.736551631241394d-06*tz05
     $  - 2.432124337036687d-07*tz06
     $  + 3.516013526366554d-08*tz07
     $  - 5.202891371724552d-09*tz08
     $  + 7.837847138987193d-10*tz09
     $  - 1.197497821309543d-10*tz10
     $  + 1.850581964813002d-11*tz11
     $  - 2.886890644139832d-12*tz12
     $  + 4.539213762014694d-13*tz13
     $  - 7.185307620608518d-14*tz14
     $  + 1.143976503452786d-14*tz15
     $  - 1.830478583060870d-15*tz16
     $  + 2.941877575256598d-16*tz17

      CsH4YZ(0,2,3,1,04) = 
     $  + 3.775792660164795d-03
     $  - 3.543621898967912d-03*tz01
     $  - 2.090123051805474d-04*tz02
     $  - 2.638371580499186d-05*tz03
     $  + 3.693792084352204d-06*tz04
     $  - 5.384502485664063d-07*tz05
     $  + 8.055029627639418d-08*tz06
     $  - 1.227211575859241d-08*tz07
     $  + 1.895455742102367d-09*tz08
     $  - 2.959043900483398d-10*tz09
     $  + 4.659422989208015d-11*tz10
     $  - 7.389282384869877d-12*tz11
     $  + 1.178882710506358d-12*tz12
     $  - 1.890416888303189d-13*tz13
     $  + 3.044830662624644d-14*tz14
     $  - 4.923180195228718d-15*tz15
     $  + 7.987387191016422d-16*tz16
     $  - 1.299815131786720d-16*tz17

      CsH4YZ(0,2,3,1,05) = 
     $  - 3.259379506076267d-04
     $  + 2.304272561076009d-04*tz01
     $  + 7.754851097401353d-05*tz02
     $  + 1.694820618042433d-05*tz03
     $  + 1.168706239961789d-06*tz04
     $  - 1.786455417652597d-07*tz05
     $  + 2.765148203153272d-08*tz06
     $  - 4.322957923836466d-09*tz07
     $  + 6.813812238815034d-10*tz08
     $  - 1.081330777755413d-10*tz09
     $  + 1.725982605689935d-11*tz10
     $  - 2.768668339413285d-12*tz11
     $  + 4.460463174290600d-13*tz12
     $  - 7.213299409909741d-14*tz13
     $  + 1.170425615406774d-14*tz14
     $  - 1.904796724676146d-15*tz15
     $  + 3.108283766609533d-16*tz16

      CsH4YZ(0,2,3,1,06) = 
     $  + 6.780579694262007d-06
     $  + 6.749629112442690d-06*tz01
     $  - 7.918105256203018d-06*tz02
     $  - 4.243271379292481d-06*tz03
     $  - 1.281861940785012d-06*tz04
     $  - 1.005551595762924d-07*tz05
     $  + 1.572446412065683d-08*tz06
     $  - 2.478829672402982d-09*tz07
     $  + 3.934174208787048d-10*tz08
     $  - 6.279973490639631d-11*tz09
     $  + 1.007423380660297d-11*tz10
     $  - 1.623068640189113d-12*tz11
     $  + 2.624849829663776d-13*tz12
     $  - 4.259182781657510d-14*tz13
     $  + 6.931782674670288d-15*tz14
     $  - 1.131161068163674d-15*tz15
     $  + 1.850363422344611d-16*tz16

      CsH4YZ(0,2,3,1,07) = 
     $  + 1.183666680642959d-06
     $  - 1.797317990777681d-06*tz01
     $  - 1.907281107769127d-07*tz02
     $  + 3.612747591551655d-07*tz03
     $  + 3.209854922404281d-07*tz04
     $  + 1.134347627800852d-07*tz05
     $  + 1.005982187014963d-08*tz06
     $  - 1.594590510430684d-09*tz07
     $  + 2.542848383417835d-10*tz08
     $  - 4.075912655899498d-11*tz09
     $  + 6.562386852005967d-12*tz10
     $  - 1.060690752859942d-12*tz11
     $  + 1.720310826123759d-13*tz12
     $  - 2.798665808378591d-14*tz13
     $  + 4.565438785721640d-15*tz14
     $  - 7.465873864804518d-16*tz15
     $  + 1.223630456993819d-16*tz16

      CsH4YZ(0,2,3,1,08) = 
     $  - 4.175756698035706d-08
     $  - 2.339737974647490d-08*tz01
     $  + 9.639932114336804d-08*tz02
     $  + 3.016427786017843d-08*tz03
     $  - 2.076191208395968d-08*tz04
     $  - 2.831804590290377d-08*tz05
     $  - 1.139591659215156d-08*tz06
     $  - 1.082127263126461d-09*tz07
     $  + 1.733729463133936d-10*tz08
     $  - 2.790319556786659d-11*tz09
     $  + 4.508632169938333d-12*tz10
     $  - 7.310529167999182d-13*tz11
     $  + 1.189039379777169d-13*tz12
     $  - 1.939299189588455d-14*tz13
     $  + 3.170839783500080d-15*tz14
     $  - 5.196115853891867d-16*tz15

      CsH4YZ(0,2,3,1,09) = 
     $  - 9.987024353480294d-09
     $  + 1.764999638818857d-08*tz01
     $  - 1.284551107194544d-09*tz02
     $  - 7.313510795047159d-09*tz03
     $  - 4.376788292608920d-09*tz04
     $  + 1.161357149823877d-09*tz05
     $  + 2.794157807320928d-09*tz06
     $  + 1.249548538106686d-09*tz07
     $  + 1.240606184354108d-10*tz08
     $  - 2.004043034952028d-11*tz09
     $  + 3.248724741572547d-12*tz10
     $  - 5.282964063331590d-13*tz11
     $  + 8.615013691974648d-14*tz12
     $  - 1.408396848932705d-14*tz13
     $  + 2.307705110847597d-15*tz14
     $  - 3.789035640993217d-16*tz15

      CsH4YZ(0,2,3,1,10) = 
     $  + 3.841539629812876d-10
     $  + 4.598916009673395d-11*tz01
     $  - 1.134015336862607d-09*tz02
     $  - 1.139766621674040d-10*tz03
     $  + 7.052336752543200d-10*tz04
     $  + 6.061508635863032d-10*tz05
     $  - 3.510482826665945d-11*tz06
     $  - 2.992432139583298d-10*tz07
     $  - 1.463086246800258d-10*tz08
     $  - 1.497210261966651d-11*tz09
     $  + 2.434255817018785d-12*tz10
     $  - 3.968951009036392d-13*tz11
     $  + 6.487631393130271d-14*tz12
     $  - 1.062896186307564d-14*tz13
     $  + 1.745011878777833d-15*tz14
     $  - 2.870294858328049d-16*tz15

      CsH4YZ(0,2,3,1,11) = 
     $  + 1.044016413113160d-10
     $  - 1.938171725167871d-10*tz01
     $  + 4.048905086082537d-11*tz02
     $  + 1.145191029681542d-10*tz03
     $  + 4.465328964915215d-11*tz04
     $  - 7.346371014184482d-11*tz05
     $  - 8.448525074100881d-11*tz06
     $  - 6.035625883429988d-12*tz07
     $  + 3.408331061710109d-11*tz08
     $  + 1.803649981266312d-11*tz09
     $  + 1.883361454073370d-12*tz10
     $  - 3.078133402928257d-13*tz11
     $  + 5.042473792548584d-14*tz12
     $  - 8.277691368876320d-15*tz13
     $  + 1.361456990657564d-15*tz14
     $  - 2.243131266961378d-16*tz15

      CsH4YZ(0,2,3,1,12) = 
     $  - 4.255577618312845d-12
     $  + 9.367497546607334d-13*tz01
     $  + 1.378971247740639d-11*tz02
     $  - 1.438781096150822d-12*tz03
     $  - 1.471598704637498d-11*tz04
     $  - 9.427957323937832d-12*tz05
     $  + 7.809257726826330d-12*tz06
     $  + 1.196020823537490d-11*tz07
     $  + 1.943748126138924d-12*tz08
     $  - 4.073241142225563d-12*tz09
     $  - 2.317530334015848d-12*tz10
     $  - 2.451612991770192d-13*tz11
     $  + 4.024177092001099d-14*tz12
     $  - 6.618150929183610d-15*tz13
     $  + 1.090334844674737d-15*tz14
     $  - 1.799208956074178d-16*tz15

      CsH4YZ(0,2,3,1,13) = 
     $  - 1.187151426765550d-12
     $  + 2.265427715278634d-12*tz01
     $  - 7.171135419369005d-13*tz02
     $  - 1.708281403248381d-12*tz03
     $  - 3.277234320822663d-13*tz04
     $  + 2.015311266749063d-12*tz05
     $  + 1.752168293323708d-12*tz06
     $  - 8.136295010582023d-13*tz07
     $  - 1.721788139551858d-12*tz08
     $  - 3.994379930704486d-13*tz09
     $  + 5.059268125085645d-13*tz10
     $  + 3.080918949045720d-13*tz11
     $  + 3.284389986695982d-14*tz12
     $  - 5.410634796969482d-15*tz13
     $  + 8.927853287598269d-16*tz14
     $  - 1.475342411444161d-16*tz15

      CsH4YZ(0,2,3,1,14) = 
     $  + 5.122415971833344d-14
     $  - 2.375218435866437d-14*tz01
     $  - 1.735579853247182d-13*tz02
     $  + 5.229042911105371d-14*tz03
     $  + 2.702027113309743d-13*tz04
     $  + 1.239826130216426d-13*tz05
     $  - 2.816717873822706d-13*tz06
     $  - 3.104775168858978d-13*tz07
     $  + 7.817854856960679d-14*tz08
     $  + 2.518810386870912d-13*tz09
     $  + 7.256493564217538d-14*tz10
     $  - 6.485505244468874d-14*tz11
     $  - 4.214018603354051d-14*tz12
     $  - 4.509041596969292d-15*tz13
     $  + 7.450937024903483d-16*tz14
     $  - 1.232929506862969d-16*tz15

      CsH4YZ(0,2,3,1,15) = 
     $  + 1.412088631428636d-14
     $  - 2.752625691682757d-14*tz01
     $  + 1.117278503636219d-14*tz02
     $  + 2.527089486390068d-14*tz03
     $  - 4.486929691001846d-14*tz05
     $  - 3.059173326282022d-14*tz06
     $  + 3.949571588902815d-14*tz07
     $  + 5.385755100519534d-14*tz08
     $  - 5.888725148596144d-15*tz09
     $  - 3.739466424480864d-14*tz10
     $  - 1.254133701947825d-14*tz11
     $  + 8.534945569271665d-15*tz12
     $  + 5.904768406306004d-15*tz13
     $  + 6.322167849513038d-16*tz14
     $  - 1.047457441339728d-16*tz15

      CsH4YZ(0,2,3,1,16) = 
     $  - 6.417404414548403d-16
     $  + 4.107011109853147d-16*tz01
     $  + 2.245839511861019d-15*tz02
     $  - 1.101814573341858d-15*tz03
     $  - 4.703463404676720d-15*tz04
     $  - 1.346536484055158d-15*tz05
     $  + 7.517580246918099d-15*tz06
     $  + 6.634409151825224d-15*tz07
     $  - 5.489040306559032d-15*tz08
     $  - 9.250078429902388d-15*tz09
     $  + 5.625950459197345d-15*tz11
     $  + 2.118641676929007d-15*tz12
     $  - 1.148314111503737d-15*tz13
     $  - 8.447394786693846d-16*tz14

      CsH4YZ(0,2,3,1,17) = 
     $  - 1.729426639188727d-16
     $  + 3.433409410422585d-16*tz01
     $  - 1.658121982871874d-16*tz02
     $  - 3.742269426176570d-16*tz03
     $  + 9.077888653080050d-16*tz05
     $  + 4.711212201765964d-16*tz06
     $  - 1.255980382222887d-15*tz07
     $  - 1.354635284573259d-15*tz08
     $  + 7.466383285620750d-16*tz09
     $  + 1.581697594075364d-15*tz10
     $  + 1.185157298546347d-16*tz11
     $  - 8.565632574234529d-16*tz12
     $  - 3.541554412829667d-16*tz13
     $  + 1.574316386439670d-16*tz14
     $  + 1.230473843071199d-16*tz15

      CsH4YZ(0,2,3,1,18) = 
     $  - 1.749661066583483d-16*tz06
     $  - 1.251252350565325d-16*tz07
     $  + 2.081241535258123d-16*tz08
     $  + 2.673783059231662d-16*tz09
     $  - 2.700496296490099d-16*tz11
     $  + 1.318144481507315d-16*tz13

      CsH4YZ(0,3,2,1,01) = 
     $  + 3.554862660343892d-01
     $  - 3.733420351546513d-01*tz01
     $  + 1.934227531208478d-02*tz02
     $  - 1.638210073282368d-03*tz03
     $  + 1.690199663409635d-04*tz04
     $  - 1.943887306576524d-05*tz05
     $  + 2.396283431709894d-06*tz06
     $  - 3.100486089562243d-07*tz07
     $  + 4.157875397123377d-08*tz08
     $  - 5.731962476383159d-09*tz09
     $  + 8.077450794108086d-10*tz10
     $  - 1.158835878395434d-10*tz11
     $  + 1.687465154171362d-11*tz12
     $  - 2.488341416020160d-12*tz13
     $  + 3.709048703380455d-13*tz14
     $  - 5.580416728733533d-14*tz15
     $  + 8.464722906819654d-15*tz16
     $  - 1.293265096589482d-15*tz17
     $  + 1.988119921479958d-16*tz18

      CsH4YZ(0,3,2,1,02) = 
     $  + 8.999561214896964d-02
     $  + 5.258835800486991d-02*tz01
     $  - 2.092177142608197d-03*tz02
     $  + 1.447325344886755d-04*tz03
     $  - 1.266945863131610d-05*tz04
     $  + 1.268098080647732d-06*tz05
     $  - 1.385521633696937d-07*tz06
     $  + 1.611019942831483d-08*tz07
     $  - 1.962710226351749d-09*tz08
     $  + 2.479799503829128d-10*tz09
     $  - 3.226046043335726d-11*tz10
     $  + 4.298893208135490d-12*tz11
     $  - 5.844891683862864d-13*tz12
     $  + 8.083890151261940d-14*tz13
     $  - 1.134631547644591d-14*tz14
     $  + 1.613072576188251d-15*tz15
     $  - 2.318789239110999d-16*tz16

      CsH4YZ(0,3,2,1,03) = 
     $  - 2.861604907861030d-02
     $  + 4.766000258096330d-03*tz01
     $  + 4.531488261083199d-04*tz02
     $  - 4.527523603508540d-05*tz03
     $  + 5.259245986055858d-06*tz04
     $  - 6.607789560699253d-07*tz05
     $  + 8.721730444420263d-08*tz06
     $  - 1.191562236436599d-08*tz07
     $  + 1.670366180799951d-09*tz08
     $  - 2.389147653532340d-10*tz09
     $  + 3.473160067704075d-11*tz10
     $  - 5.117252113573899d-12*tz11
     $  + 7.625398492901376d-13*tz12
     $  - 1.147343335316783d-13*tz13
     $  + 1.740879601202670d-14*tz14
     $  - 2.660934006696508d-15*tz15
     $  + 4.093763375192792d-16*tz16

      CsH4YZ(0,3,2,1,04) = 
     $  + 3.478689448719673d-03
     $  - 2.617777775811623d-03*tz01
     $  - 2.991167736083886d-04*tz02
     $  - 1.410739067655481d-05*tz03
     $  + 1.849171092941581d-06*tz04
     $  - 2.508711890714956d-07*tz05
     $  + 3.493897207520409d-08*tz06
     $  - 4.967867118860787d-09*tz07
     $  + 7.183621236102751d-10*tz08
     $  - 1.053388347246902d-10*tz09
     $  + 1.563009698833844d-11*tz10
     $  - 2.342759747531981d-12*tz11
     $  + 3.542435255478861d-13*tz12
     $  - 5.397724265677567d-14*tz13
     $  + 8.280664097636325d-15*tz14
     $  - 1.278025120838631d-15*tz15
     $  + 1.983193742898546d-16*tz16

      CsH4YZ(0,3,2,1,05) = 
     $  - 2.686481313458501d-04
     $  + 3.312600556307233d-04*tz01
     $  + 8.433813398459026d-05*tz02
     $  + 1.691813079821636d-05*tz03
     $  + 1.055111960724827d-06*tz04
     $  - 1.465537107507355d-07*tz05
     $  + 2.079329034008946d-08*tz06
     $  - 3.001608414385867d-09*tz07
     $  + 4.395462298608464d-10*tz08
     $  - 6.514732235130418d-11*tz09
     $  + 9.755931705746954d-12*tz10
     $  - 1.474065684934508d-12*tz11
     $  + 2.244672116290869d-13*tz12
     $  - 3.441731928169735d-14*tz13
     $  + 5.309544700179345d-15*tz14
     $  - 8.235927712793582d-16*tz15
     $  + 1.283838994111182d-16*tz16

      CsH4YZ(0,3,2,1,06) = 
     $  + 1.579423930106241d-05
     $  - 4.303318964070840d-06*tz01
     $  - 1.010062866774177d-05*tz02
     $  - 4.852595867940104d-06*tz03
     $  - 1.261494361623308d-06*tz04
     $  - 9.459978965992882d-08*tz05
     $  + 1.361240762850885d-08*tz06
     $  - 1.988572030852862d-09*tz07
     $  + 2.941866314007180d-10*tz08
     $  - 4.398971484025401d-11*tz09
     $  + 6.638617147115287d-12*tz10
     $  - 1.009914017725173d-12*tz11
     $  + 1.547214171301751d-13*tz12
     $  - 2.385219434693633d-14*tz13
     $  + 3.697668975383421d-15*tz14
     $  - 5.761056008842913d-16*tz15

      CsH4YZ(0,3,2,1,07) = 
     $  + 1.249522206009479d-07
     $  - 2.905145168605603d-06*tz01
     $  + 6.088942687400495d-09*tz02
     $  + 5.259884538569898d-07*tz03
     $  + 3.615018097669127d-07*tz04
     $  + 1.127447220160410d-07*tz05
     $  + 9.419994275524115d-09*tz06
     $  - 1.391466916065610d-09*tz07
     $  + 2.078160991163934d-10*tz08
     $  - 3.133176119374273d-11*tz09
     $  + 4.762630631077773d-12*tz10
     $  - 7.291667804311519d-13*tz11
     $  + 1.123478315354630d-13*tz12
     $  - 1.740847508645311d-14*tz13
     $  + 2.711208274358953d-15*tz14
     $  - 4.241840347775459d-16*tz15

      CsH4YZ(0,3,2,1,08) = 
     $  - 1.681921478085629d-07
     $  + 1.416664914546838d-07*tz01
     $  + 1.279725171908334d-07*tz02
     $  + 2.110958516856873d-08*tz03
     $  - 3.321498866569693d-08*tz04
     $  - 3.177872495126667d-08*tz05
     $  - 1.135290231851001d-08*tz06
     $  - 1.019270249637885d-09*tz07
     $  + 1.535252566497838d-10*tz08
     $  - 2.331866626482068d-11*tz09
     $  + 3.567831434191082d-12*tz10
     $  - 5.494253206815729d-13*tz11
     $  + 8.509569825827942d-14*tz12
     $  - 1.324773754766001d-14*tz13
     $  + 2.072006625112438d-15*tz14
     $  - 3.254357011652109d-16*tz15

      CsH4YZ(0,3,2,1,09) = 
     $  + 4.922211959964284d-09
     $  + 2.963765106534366d-08*tz01
     $  - 6.863790262637054d-09*tz02
     $  - 1.007671164855515d-08*tz03
     $  - 3.790050139781567d-09*tz04
     $  + 2.222431809329631d-09*tz05
     $  + 3.127393918968897d-09*tz06
     $  + 1.245986429233739d-09*tz07
     $  + 1.175670059271737d-10*tz08
     $  - 1.797641719152932d-11*tz09
     $  + 2.766764156878374d-12*tz10
     $  - 4.283247648013963d-13*tz11
     $  + 6.665601671401939d-14*tz12
     $  - 1.042185249717001d-14*tz13
     $  + 1.636429404309280d-15*tz14
     $  - 2.579455271364880d-16*tz15

      CsH4YZ(0,3,2,1,10) = 
     $  + 2.099472647885692d-09
     $  - 2.296194778705250d-09*tz01
     $  - 1.431976546485265d-09*tz02
     $  + 2.693921279765700d-10*tz03
     $  + 9.689459745412272d-10*tz04
     $  + 5.718009992679003d-10*tz05
     $  - 1.362440381588643d-10*tz06
     $  - 3.339482300966074d-10*tz07
     $  - 1.459771257398334d-10*tz08
     $  - 1.426137540984259d-11*tz09
     $  + 2.206803647168205d-12*tz10
     $  - 3.432904322650760d-13*tz11
     $  + 5.365667960676269d-14*tz12
     $  - 8.422725538558231d-15*tz13
     $  + 1.327329756785722d-15*tz14
     $  - 2.099196034978783d-16*tz15

      CsH4YZ(0,3,2,1,11) = 
     $  - 9.366242236201148d-11
     $  - 3.260530422572337d-10*tz01
     $  + 1.407442298451161d-10*tz02
     $  + 1.440662200270089d-10*tz03
     $  + 1.128295930947810d-11*tz04
     $  - 1.020246150219100d-10*tz05
     $  - 8.315319346132047d-11*tz06
     $  + 4.418986806348405d-12*tz07
     $  + 3.793593399549701d-11*tz08
     $  + 1.800122170655258d-11*tz09
     $  + 1.801954579981176d-12*tz10
     $  - 2.815541681894284d-13*tz11
     $  + 4.418423612078538d-14*tz12
     $  - 6.961242264302481d-15*tz13
     $  + 1.100703891317205d-15*tz14
     $  - 1.746166538674897d-16*tz15

      CsH4YZ(0,3,2,1,12) = 
     $  - 2.736765347832271d-11
     $  + 3.326647128097052d-11*tz01
     $  + 1.566069390835303d-11*tz02
     $  - 9.462630458263831d-12*tz03
     $  - 1.822620222459400d-11*tz04
     $  - 6.285757680976352d-12*tz05
     $  + 1.119923240814794d-11*tz06
     $  + 1.205348975669459d-11*tz07
     $  + 7.927348562718403d-13*tz08
     $  - 4.523650768465360d-12*tz09
     $  - 2.313308428394637d-12*tz10
     $  - 2.354763798119637d-13*tz11
     $  + 3.709025054849934d-14*tz12
     $  - 5.863414098771581d-15*tz13
     $  + 9.300087961290255d-16*tz14
     $  - 1.479620788411412d-16*tz15

      CsH4YZ(0,3,2,1,13) = 
     $  + 1.402712321877300d-12
     $  + 3.763988640457636d-12*tz01
     $  - 2.274618553751801d-12*tz02
     $  - 1.909622882511605d-12*tz03
     $  + 4.861929087247159d-13*tz04
     $  + 2.497297072562645d-12*tz05
     $  + 1.440205347873354d-12*tz06
     $  - 1.240672425561674d-12*tz07
     $  - 1.760334431741711d-12*tz08
     $  - 2.660724796100161d-13*tz09
     $  + 5.608717551150189d-13*tz10
     $  + 3.075397827939748d-13*tz11
     $  + 3.165378747549513d-14*tz12
     $  - 5.019735213398729d-15*tz13
     $  + 7.985008641298189d-16*tz14
     $  - 1.273804734559459d-16*tz15

      CsH4YZ(0,3,2,1,14) = 
     $  + 3.620424800779537d-13
     $  - 4.647091209776580d-13*tz01
     $  - 1.715776048405323d-13*tz02
     $  + 1.883902981830108d-13*tz03
     $  + 3.034643371741958d-13*tz04
     $  + 3.194904817888820d-14*tz05
     $  - 3.526705283236692d-13*tz06
     $  - 2.788094350589691d-13*tz07
     $  + 1.343258500548860d-13*tz08
     $  + 2.599050343428551d-13*tz09
     $  + 5.644556150158441d-14*tz10
     $  - 7.180034302715692d-14*tz11
     $  - 4.206330111594710d-14*tz12
     $  - 4.358669538796444d-15*tz13
     $  + 6.952180220902654d-16*tz14
     $  - 1.111824468439032d-16*tz15

      CsH4YZ(0,3,2,1,15) = 
     $  - 1.966223214698831d-14
     $  - 4.497076134850153d-14*tz01
     $  + 3.387471428897737d-14*tz02
     $  + 2.497019566121993d-14*tz03
     $  - 1.556383582089801d-14*tz04
     $  - 5.114858870040194d-14*tz05
     $  - 1.941990053195502d-14*tz06
     $  + 5.030183777598939d-14*tz07
     $  + 5.068354159540461d-14*tz08
     $  - 1.351748208694367d-14*tz09
     $  - 3.882361103113926d-14*tz10
     $  - 1.052261045532388d-14*tz11
     $  + 9.439745475343151d-15*tz12
     $  + 5.893590811043417d-15*tz13
     $  + 6.127528944973409d-16*tz14

      CsH4YZ(0,3,2,1,16) = 
     $  - 4.822681106908909d-15
     $  + 6.398000519328085d-15*tz01
     $  + 1.901848144157147d-15*tz02
     $  - 3.198911820974626d-15*tz03
     $  - 4.859295618464430d-15*tz04
     $  + 6.651466699932135d-16*tz05
     $  + 8.700720099789868d-15*tz06
     $  + 5.220116186823548d-15*tz07
     $  - 7.165653556070910d-15*tz08
     $  - 8.952124306984426d-15*tz09
     $  + 1.114207409211870d-15*tz10
     $  + 5.865551891572315d-15*tz11
     $  + 1.858042061181766d-15*tz12
     $  - 1.269279666368176d-15*tz13
     $  - 8.430661154218655d-16*tz14

      CsH4YZ(0,3,2,1,17) = 
     $  + 2.687881769544216d-16
     $  + 5.513784094366046d-16*tz01
     $  - 4.865226553437010d-16*tz02
     $  - 3.307114195934907d-16*tz03
     $  + 3.353471197179598d-16*tz04
     $  + 9.723619011258243d-16*tz05
     $  + 1.880197513456943d-16*tz06
     $  - 1.474488733358870d-15*tz07
     $  - 1.171618874316586d-15*tz08
     $  + 1.010307035947374d-15*tz09
     $  + 1.558682078495424d-15*tz10
     $  - 8.956744678340664d-16*tz12
     $  - 3.196211200118481d-16*tz13
     $  + 1.739713478846900d-16*tz14
     $  + 1.227916170343975d-16*tz15

      CsH4YZ(0,3,2,1,18) = 
     $  - 1.920210794831602d-16*tz06
     $  + 2.478842277648392d-16*tz08
     $  + 2.435266479977565d-16*tz09
     $  - 1.395344879746456d-16*tz10
     $  - 2.694194776308517d-16*tz11
     $  + 1.381222701436798d-16*tz13

      CsH4YZ(2,0,3,1,02) = 
     $  + 1.271177752013422d-01
     $  - 1.389905854659051d-01*tz01
     $  + 1.328962491341869d-02*tz02
     $  - 1.603854525936234d-03*tz03
     $  + 2.131576791845240d-04*tz04
     $  - 2.989851065051721d-05*tz05
     $  + 4.341920903187438d-06*tz06
     $  - 6.460820904724618d-07*tz07
     $  + 9.788928873545235d-08*tz08
     $  - 1.503968816575284d-08*tz09
     $  + 2.336496513438171d-09*tz10
     $  - 3.662899844140558d-10*tz11
     $  + 5.785701524942945d-11*tz12
     $  - 9.197097805005088d-12*tz13
     $  + 1.469982525756198d-12*tz14
     $  - 2.360603790468797d-13*tz15
     $  + 3.806507400973038d-14*tz16
     $  - 6.160397630037750d-15*tz17
     $  + 1.000227479759130d-15*tz18
     $  - 1.628265045514259d-16*tz19

      CsH4YZ(2,0,3,1,03) = 
     $  - 2.577033616392895d-02
     $  + 2.631053911008935d-02*tz01
     $  - 5.906137074048570d-04*tz02
     $  + 5.589961863650465d-05*tz03
     $  - 6.142459491379228d-06*tz04
     $  + 7.361705312047573d-07*tz05
     $  - 9.344859410629896d-08*tz06
     $  + 1.236007224096114d-08*tz07
     $  - 1.686256511632184d-09*tz08
     $  + 2.356965604798763d-10*tz09
     $  - 3.359364111815481d-11*tz10
     $  + 4.865605437174922d-12*tz11
     $  - 7.142725254123743d-13*tz12
     $  + 1.060634416307745d-13*tz13
     $  - 1.590571972772368d-14*tz14
     $  + 2.405862907348874d-15*tz15
     $  - 3.666655249006465d-16*tz16

      CsH4YZ(2,0,3,1,04) = 
     $  + 3.423744203307500d-03
     $  - 3.080420504243298d-03*tz01
     $  - 3.155964302868668d-04*tz02
     $  - 3.168456157520761d-05*tz03
     $  + 4.541152394258403d-06*tz04
     $  - 6.719230065549105d-07*tz05
     $  + 1.015598579035485d-07*tz06
     $  - 1.559031649639596d-08*tz07
     $  + 2.421883290826145d-09*tz08
     $  - 3.798081517005869d-10*tz09
     $  + 6.002603728622220d-11*tz10
     $  - 9.548276172765052d-12*tz11
     $  + 1.527201581170203d-12*tz12
     $  - 2.454266193890052d-13*tz13
     $  + 3.960363468534586d-14*tz14
     $  - 6.413873143969845d-15*tz15
     $  + 1.042068363495899d-15*tz16
     $  - 1.697926352125337d-16*tz17

      CsH4YZ(2,0,3,1,05) = 
     $  - 2.822779752191090d-04
     $  + 1.863129778761660d-04*tz01
     $  + 7.459776927666245d-05*tz02
     $  + 1.993838775816003d-05*tz03
     $  + 1.646675888589221d-06*tz04
     $  - 2.514805671918877d-07*tz05
     $  + 3.889553733566654d-08*tz06
     $  - 6.076886385615958d-09*tz07
     $  + 9.573028805521186d-10*tz08
     $  - 1.518481902019873d-10*tz09
     $  + 2.422735915742193d-11*tz10
     $  - 3.884912557183208d-12*tz11
     $  + 6.256756502326937d-13*tz12
     $  - 1.011527329406137d-13*tz13
     $  + 1.640872624747912d-14*tz14
     $  - 2.669796718015739d-15*tz15
     $  + 4.355703222216680d-16*tz16

      CsH4YZ(2,0,3,1,06) = 
     $  + 9.502201060991793d-06
     $  + 2.389616477392558d-06*tz01
     $  - 6.155943518888855d-06*tz02
     $  - 4.174067029873387d-06*tz03
     $  - 1.445226415358381d-06*tz04
     $  - 1.347588854408123d-07*tz05
     $  + 2.103741156650618d-08*tz06
     $  - 3.312017107492273d-09*tz07
     $  + 5.250999048072214d-10*tz08
     $  - 8.374709375419693d-11*tz09
     $  + 1.342480761016193d-11*tz10
     $  - 2.161541578257797d-12*tz11
     $  + 3.493803764706973d-13*tz12
     $  - 5.666529279823379d-14*tz13
     $  + 9.218424191605103d-15*tz14
     $  - 1.503754864643036d-15*tz15
     $  + 2.459049108225894d-16*tz16

      CsH4YZ(2,0,3,1,07) = 
     $  + 6.308752846052343d-07
     $  - 1.181785858349756d-06*tz01
     $  - 1.842111194388650d-07*tz02
     $  + 2.854787031930055d-07*tz03
     $  + 3.127410244064883d-07*tz04
     $  + 1.259022576000730d-07*tz05
     $  + 1.274078031909620d-08*tz06
     $  - 2.018377350746678d-09*tz07
     $  + 3.217072447783029d-10*tz08
     $  - 5.154460964188733d-11*tz09
     $  + 8.295883971265748d-12*tz10
     $  - 1.340453697461035d-12*tz11
     $  + 2.173443410636550d-13*tz12
     $  - 3.534958239539065d-14*tz13
     $  + 5.765260112645644d-15*tz14
     $  - 9.426050644196638d-16*tz15
     $  + 1.544615390161399d-16*tz16

      CsH4YZ(2,0,3,1,08) = 
     $  - 6.334465884469774d-08
     $  + 1.953447072656643d-08*tz01
     $  + 6.942052012064645d-08*tz02
     $  + 3.075675066479116d-08*tz03
     $  - 1.521317097227660d-08*tz04
     $  - 2.753748746803941d-08*tz05
     $  - 1.247544931703332d-08*tz06
     $  - 1.323595468469119d-09*tz07
     $  + 2.119867712350288d-10*tz08
     $  - 3.410764323477714d-11*tz09
     $  + 5.509713396930351d-12*tz10
     $  - 8.931670499172259d-13*tz11
     $  + 1.452415319373765d-13*tz12
     $  - 2.368423643570950d-14*tz13
     $  + 3.871834215162264d-15*tz14
     $  - 6.343894243638491d-16*tz15
     $  + 1.041581198492749d-16*tz16

      CsH4YZ(2,0,3,1,09) = 
     $  - 3.146147146506767d-09
     $  + 9.380000143070324d-09*tz01
     $  - 8.966282430181141d-10*tz02
     $  - 5.816111540236944d-09*tz03
     $  - 4.364990319043104d-09*tz04
     $  + 6.494411981613718d-10*tz05
     $  + 2.713539718227456d-09*tz06
     $  + 1.353527003906564d-09*tz07
     $  + 1.479288486650443d-10*tz08
     $  - 2.389037635208299d-11*tz09
     $  + 3.872025624536962d-12*tz10
     $  - 6.295388412092653d-13*tz11
     $  + 1.026429103958965d-13*tz12
     $  - 1.677773660349979d-14*tz13
     $  + 2.748717804028748d-15*tz14
     $  - 4.512584736651106d-16*tz15

      CsH4YZ(2,0,3,1,10) = 
     $  + 5.504542919829252d-10
     $  - 3.813783578722688d-10*tz01
     $  - 7.414427416444254d-10*tz02
     $  - 1.563248127294132d-10*tz03
     $  + 5.712801899953053d-10*tz04
     $  + 6.018925943678060d-10*tz05
     $  + 1.784896696524914d-11*tz06
     $  - 2.900732473922559d-10*tz07
     $  - 1.572015981752673d-10*tz08
     $  - 1.750167276612694d-11*tz09
     $  + 2.845056726598764d-12*tz10
     $  - 4.638057708153089d-13*tz11
     $  + 7.580342164175896d-14*tz12
     $  - 1.241770376102402d-14*tz13
     $  + 2.038456394876274d-15*tz14
     $  - 3.352635140774914d-16*tz15

      CsH4YZ(2,0,3,1,11) = 
     $  + 2.007311339096594d-11
     $  - 8.469709743029676d-11*tz01
     $  + 2.917033000109198d-11*tz02
     $  + 8.823244042137777d-11*tz03
     $  + 4.774022312035464d-11*tz04
     $  - 5.891571607107609d-11*tz05
     $  - 8.368767958734966d-11*tz06
     $  - 1.200500186289791d-11*tz07
     $  + 3.297273122698603d-11*tz08
     $  + 1.925398673926419d-11*tz09
     $  + 2.166964560244883d-12*tz10
     $  - 3.541224459413017d-13*tz11
     $  + 5.800462455423487d-14*tz12
     $  - 9.521064367899089d-15*tz13
     $  + 1.565817924581651d-15*tz14
     $  - 2.579623736930815d-16*tz15

      CsH4YZ(2,0,3,1,12) = 
     $  - 5.427270429989283d-12
     $  + 5.206919493127577d-12*tz01
     $  + 8.238397697585030d-12*tz02
     $  - 2.822760925633647d-13*tz03
     $  - 1.195133977466670d-11*tz04
     $  - 9.699868376928223d-12*tz05
     $  + 6.061920855718403d-12*tz06
     $  + 1.182441586438521d-11*tz07
     $  + 2.661477914958419d-12*tz08
     $  - 3.932336182292294d-12*tz09
     $  - 2.460832971751679d-12*tz10
     $  - 2.784582339462252d-13*tz11
     $  + 4.570317237760214d-14*tz12
     $  - 7.515718260356136d-15*tz13
     $  + 1.238115840799396d-15*tz14
     $  - 2.042928599974667d-16*tz15

      CsH4YZ(2,0,3,1,13) = 
     $  - 1.414993880671516d-13
     $  + 8.347018373487781d-13*tz01
     $  - 4.902044120266286d-13*tz02
     $  - 1.275397934812566d-12*tz03
     $  - 4.395995119357815d-13*tz04
     $  + 1.669501496810989d-12*tz05
     $  + 1.775938671070478d-12*tz06
     $  - 5.884641823562721d-13*tz07
     $  - 1.699625740571757d-12*tz08
     $  - 4.901083717750230d-13*tz09
     $  + 4.873799160099815d-13*tz10
     $  + 3.256899686182854d-13*tz11
     $  + 3.690624015590789d-14*tz12
     $  - 6.079439505441330d-15*tz13
     $  + 1.003078746548599d-15*tz14
     $  - 1.657507921554142d-16*tz15

      CsH4YZ(2,0,3,1,14) = 
     $  + 5.750289620480649d-14
     $  - 6.636139085899758d-14*tz01
     $  - 9.619505390609530d-14*tz02
     $  + 2.770781980484233d-14*tz03
     $  + 2.179009385197690d-13*tz04
     $  + 1.366638931920029d-13*tz05
     $  - 2.344117803450367d-13*tz06
     $  - 3.122371067827586d-13*tz07
     $  + 4.765697542957449d-14*tz08
     $  + 2.483227298980437d-13*tz09
     $  + 8.447813095356756d-14*tz10
     $  - 6.233975273662294d-14*tz11
     $  - 4.437915430130317d-14*tz12
     $  - 5.021021378696884d-15*tz13
     $  + 8.296510163718580d-16*tz14
     $  - 1.372781089386311d-16*tz15

      CsH4YZ(2,0,3,1,15) = 
     $  + 1.028549079445705d-15
     $  - 8.769509359079197d-15*tz01
     $  + 7.218403474455014d-15*tz02
     $  + 1.839811637184518d-14*tz03
     $  + 2.728310926337420d-15*tz04
     $  - 3.746104792269234d-14*tz05
     $  - 3.211160208348041d-14*tz06
     $  + 3.265230832136939d-14*tz07
     $  + 5.390877108652791d-14*tz08
     $  - 1.591533281135251d-15*tz09
     $  - 3.682604829351042d-14*tz10
     $  - 1.415755132295588d-14*tz11
     $  + 8.185246390242436d-15*tz12
     $  + 6.198298102041038d-15*tz13
     $  + 6.985623715050905d-16*tz14
     $  - 1.157329496331900d-16*tz15

      CsH4YZ(2,0,3,1,16) = 
     $  - 6.383431298609857d-16
     $  + 8.320935581193185d-16*tz01
     $  + 1.174600065284050d-15*tz02
     $  - 6.385033887520959d-16*tz03
     $  - 3.765173268685577d-15*tz04
     $  - 1.717820663051001d-15*tz05
     $  + 6.383246862574890d-15*tz06
     $  + 6.820156203465336d-15*tz07
     $  - 4.457976336191398d-15*tz08
     $  - 9.230126592680309d-15*tz09
     $  - 5.731387426528211d-16*tz10
     $  + 5.535029687729026d-15*tz11
     $  + 2.343826541445331d-15*tz12
     $  - 1.098661922594555d-15*tz13
     $  - 8.842240355976557d-16*tz14

      CsH4YZ(2,0,3,1,17) = 
     $  - 1.016597371374760d-16*tz02
     $  - 2.674400495350602d-16*tz03
     $  + 7.593320097343644d-16*tz05
     $  + 5.230480628182668d-16*tz06
     $  - 1.073851230683993d-15*tz07
     $  - 1.377215653091686d-15*tz08
     $  + 5.867711127702661d-16*tz09
     $  + 1.574804987221201d-15*tz10
     $  + 2.110624365066110d-16*tz11
     $  - 8.419787852679701d-16*tz12
     $  - 3.862449712763378d-16*tz13
     $  + 1.502530154321053d-16*tz14
     $  + 1.284785778786337d-16*tz15

      CsH4YZ(2,0,3,1,18) = 
     $  - 1.498124085474826d-16*tz06
     $  - 1.326112583713095d-16*tz07
     $  + 1.779693775696729d-16*tz08
     $  + 2.700249239306910d-16*tz09
     $  - 2.684283178751369d-16*tz11
     $  + 1.294646516806165d-16*tz13

      CsH4YZ(2,2,3,1,03) = 
     $  + 3.177944380033556d-02
     $  - 3.474764636647628d-02*tz01
     $  + 3.322406228354674d-03*tz02
     $  - 4.009636314840586d-04*tz03
     $  + 5.328941979613100d-05*tz04
     $  - 7.474627662629303d-06*tz05
     $  + 1.085480225796859d-06*tz06
     $  - 1.615205226181154d-07*tz07
     $  + 2.447232218386308d-08*tz08
     $  - 3.759922041438210d-09*tz09
     $  + 5.841241283595429d-10*tz10
     $  - 9.157249610351396d-11*tz11
     $  + 1.446425381235736d-11*tz12
     $  - 2.299274451251272d-12*tz13
     $  + 3.674956314390495d-13*tz14
     $  - 5.901509476171994d-14*tz15
     $  + 9.516268502432596d-15*tz16
     $  - 1.540099407509437d-15*tz17
     $  + 2.500568699397827d-16*tz18

      CsH4YZ(2,2,3,1,04) = 
     $  - 3.705230348910438d-03
     $  + 3.351268472569205d-03*tz01
     $  + 4.014710275396801d-04*tz02
     $  - 5.421832391457174d-05*tz03
     $  + 7.688343902507344d-06*tz04
     $  - 1.125428737541207d-06*tz05
     $  + 1.684843195470409d-07*tz06
     $  - 2.565007090053310d-08*tz07
     $  + 3.956214217603914d-09*tz08
     $  - 6.165991725864669d-10*tz09
     $  + 9.692565863939943d-11*tz10
     $  - 1.534524097704016d-11*tz11
     $  + 2.444195383568390d-12*tz12
     $  - 3.913401653010936d-13*tz13
     $  + 6.294077603815669d-14*tz14
     $  - 1.016312756981181d-14*tz15
     $  + 1.646789690430252d-15*tz16
     $  - 2.676723447538456d-16*tz17

      CsH4YZ(2,2,3,1,05) = 
     $  + 2.601096345413078d-04
     $  - 1.435150435528402d-04*tz01
     $  - 9.840875001889966d-05*tz02
     $  - 2.081862452187749d-05*tz03
     $  + 3.024013489597342d-06*tz04
     $  - 4.504748673907568d-07*tz05
     $  + 6.834634227941333d-08*tz06
     $  - 1.051552602668088d-08*tz07
     $  + 1.635853649122697d-09*tz08
     $  - 2.567745783894943d-10*tz09
     $  + 4.060574441826180d-11*tz10
     $  - 6.461665271122645d-12*tz11
     $  + 1.033781104284627d-12*tz12
     $  - 1.661599689470774d-13*tz13
     $  + 2.681540364678691d-14*tz14
     $  - 4.343046847907400d-15*tz15
     $  + 7.056367492498618d-16*tz16
     $  - 1.149753154971540d-16*tz17

      CsH4YZ(2,2,3,1,06) = 
     $  - 6.638509932156814d-06
     $  - 5.558555810799617d-06*tz01
     $  + 5.169115367237310d-06*tz02
     $  + 5.781131322008866d-06*tz03
     $  + 1.435184868521933d-06*tz04
     $  - 2.172572116147312d-07*tz05
     $  + 3.337527324617958d-08*tz06
     $  - 5.186385870088880d-09*tz07
     $  + 8.134358359766602d-10*tz08
     $  - 1.285565052078695d-10*tz09
     $  + 2.044775958378767d-11*tz10
     $  - 3.270149323295559d-12*tz11
     $  + 5.254544377455963d-13*tz12
     $  - 8.477872547736432d-14*tz13
     $  + 1.372807448472085d-14*tz14
     $  - 2.230095728699164d-15*tz15
     $  + 3.633170417277322d-16*tz16

      CsH4YZ(2,2,3,1,07) = 
     $  - 4.986438074764031d-07
     $  + 8.132904882801714d-07*tz01
     $  + 4.919588725371070d-07*tz02
     $  - 2.480606817974627d-07*tz03
     $  - 4.539587621151491d-07*tz04
     $  - 1.208001632336312d-07*tz05
     $  + 1.875240462959695d-08*tz06
     $  - 2.938923916609014d-09*tz07
     $  + 4.642098773239404d-10*tz08
     $  - 7.380376149482562d-11*tz09
     $  + 1.179923219513297d-11*tz10
     $  - 1.895416789153354d-12*tz11
     $  + 3.057473637872373d-13*tz12
     $  - 4.950041288481396d-14*tz13
     $  + 8.040117233982737d-15*tz14
     $  - 1.309692526000873d-15*tz15
     $  + 2.138984363874797d-16*tz16

      CsH4YZ(2,2,3,1,08) = 
     $  + 3.241134680936899d-08
     $  + 1.915168923475911d-08*tz01
     $  - 5.725418095091071d-08*tz02
     $  - 5.946445449684969d-08*tz03
     $  + 1.292512580678917d-08*tz04
     $  + 4.219467177138744d-08*tz05
     $  + 1.161856190168042d-08*tz06
     $  - 1.834086556269002d-09*tz07
     $  + 2.914652229007066d-10*tz08
     $  - 4.658095462523766d-11*tz09
     $  + 7.480620029580272d-12*tz10
     $  - 1.206417510397107d-12*tz11
     $  + 1.952827987797752d-13*tz12
     $  - 3.171405309443968d-14*tz13
     $  + 5.165434812286891d-15*tz14
     $  - 8.435221245942467d-16*tz15
     $  + 1.380751518694592d-16*tz16

      CsH4YZ(2,2,3,1,09) = 
     $  + 2.302134243669453d-09
     $  - 5.733952659461416d-09*tz01
     $  - 3.298130932469848d-09*tz02
     $  + 5.035868980779726d-09*tz03
     $  + 7.651419413524575d-09*tz04
     $  - 5.049492201841785d-10*tz05
     $  - 4.393548604143530d-09*tz06
     $  - 1.227882482026028d-09*tz07
     $  + 1.961640875113872d-10*tz08
     $  - 3.149366584021647d-11*tz09
     $  + 5.077882000595970d-12*tz10
     $  - 8.218019482403174d-13*tz11
     $  + 1.334405302913246d-13*tz12
     $  - 2.173136531706326d-14*tz13
     $  + 3.548401279237450d-15*tz14
     $  - 5.807770881674041d-16*tz15

      CsH4YZ(2,2,3,1,10) = 
     $  - 2.071829137646524d-10
     $  - 9.488501290141323d-11*tz01
     $  + 5.846425593010706d-10*tz02
     $  + 6.448662654791419d-10*tz03
     $  - 4.963523784142903d-10*tz04
     $  - 1.023439594296386d-09*tz05
     $  - 2.405873239832882d-11*tz06
     $  + 4.965060530461929d-10*tz07
     $  + 1.392215175006701d-10*tz08
     $  - 2.244206203743207d-11*tz09
     $  + 3.631314440100310d-12*tz10
     $  - 5.895436956014270d-13*tz11
     $  + 9.599720983977563d-14*tz12
     $  - 1.567317000123389d-14*tz13
     $  + 2.565054107239679d-15*tz14
     $  - 4.207041808547099d-16*tz15

      CsH4YZ(2,2,3,1,11) = 
     $  - 1.496554467037538d-11
     $  + 4.595874962647156d-11*tz01
     $  + 2.844756441356664d-11*tz02
     $  - 7.373410210760129d-11*tz03
     $  - 1.156567914007310d-10*tz04
     $  + 5.100517178908757d-11*tz05
     $  + 1.414514296056801d-10*tz06
     $  + 1.152865670289672d-11*tz07
     $  - 5.969134477251128d-11*tz08
     $  - 1.667113839673048d-11*tz09
     $  + 2.706123898059132d-12*tz10
     $  - 4.405879665042054d-13*tz11
     $  + 7.192548806910771d-14*tz12
     $  - 1.177014861034103d-14*tz13
     $  + 1.930326828978068d-15*tz14
     $  - 3.172056523088139d-16*tz15

      CsH4YZ(2,2,3,1,12) = 
     $  + 1.534392340468948d-12
     $  + 6.800923408725854d-13*tz01
     $  - 6.131754824152563d-12*tz02
     $  - 7.733847811337672d-12*tz03
     $  + 1.010623280535835d-11*tz04
     $  + 1.997484489545839d-11*tz05
     $  - 5.209578582022721d-12*tz06
     $  - 2.010406011906715d-11*tz07
     $  - 2.441858647578045d-12*tz08
     $  + 7.532747382382075d-12*tz09
     $  + 2.085326958262304d-12*tz10
     $  - 3.403889770231429d-13*tz11
     $  + 5.569723573744622d-14*tz12
     $  - 9.133720736697842d-15*tz13
     $  + 1.500828751374761d-15*tz14
     $  - 2.470618038746669d-16*tz15

      CsH4YZ(2,2,3,1,13) = 
     $  + 1.173970387791684d-13
     $  - 4.022212389796579d-13*tz01
     $  - 3.006349763515634d-13*tz02
     $  + 1.015670270080814d-12*tz03
     $  + 1.745973264242183d-12*tz04
     $  - 1.418823984477572d-12*tz05
     $  - 3.401586583926178d-12*tz06
     $  + 4.992359098245444d-13*tz07
     $  + 2.926232996211274d-12*tz08
     $  + 4.394335684454487d-13*tz09
     $  - 9.884257522073011d-13*tz10
     $  - 2.703229593961859d-13*tz11
     $  + 4.432634364797809d-14*tz12
     $  - 7.283083795847869d-15*tz13
     $  + 1.198856228808979d-15*tz14
     $  - 1.976732048949464d-16*tz15

      CsH4YZ(2,2,3,1,14) = 
     $  - 1.246834036730390d-14
     $  - 6.721490461795104d-15*tz01
     $  + 6.683240549974147d-14*tz02
     $  + 1.003922578260748d-13*tz03
     $  - 1.772642665158983d-13*tz04
     $  - 3.635835994768357d-13*tz05
     $  + 1.995926795581782d-13*tz06
     $  + 5.768024519240714d-13*tz07
     $  - 3.928769925900046d-14*tz08
     $  - 4.347101601689024d-13*tz09
     $  - 7.451215217017091d-14*tz10
     $  + 1.339289165112580d-13*tz11
     $  + 3.610015509418814d-14*tz12
     $  - 5.942030580247394d-15*tz13
     $  + 9.797071728006322d-16*tz14
     $  - 1.617824000748629d-16*tz15

      CsH4YZ(2,2,3,1,15) = 
     $  - 1.032986861676482d-15
     $  + 3.749847032536742d-15*tz01
     $  + 3.616682918311478d-15*tz02
     $  - 1.383116238249067d-14*tz03
     $  - 2.666285100461069d-14*tz04
     $  + 3.087939184265325d-14*tz05
     $  + 7.267773949857901d-14*tz06
     $  - 2.779345938080094d-14*tz07
     $  - 9.781574244137369d-14*tz08
     $  + 1.052099128863110d-15*tz09
     $  + 6.572559873183251d-14*tz10
     $  + 1.231929899962167d-14*tz11
     $  - 1.864111430759594d-14*tz12
     $  - 4.943882770134196d-15*tz13
     $  + 8.163694166323243d-16*tz14
     $  - 1.349986542068729d-16*tz15

      CsH4YZ(2,2,3,1,16) = 
     $  + 1.078877940432727d-16
     $  - 7.535901218270896d-16*tz02
     $  - 1.369067396966927d-15*tz03
     $  + 2.926187299148385d-15*tz04
     $  + 6.405305429195617d-15*tz05
     $  - 5.307190383408464d-15*tz06
     $  - 1.418481683154185d-14*tz07
     $  + 3.786440355640221d-15*tz08
     $  + 1.662267331272579d-14*tz09
     $  + 5.439409055620654d-16*tz10
     $  - 1.009037093880288d-14*tz11
     $  - 2.015041587749896d-15*tz12
     $  + 2.654468592161712d-15*tz13
     $  + 6.918344283844812d-16*tz14
     $  - 1.145541249459140d-16*tz15

      CsH4YZ(2,2,3,1,17) = 
     $  + 1.883483588048095d-16*tz03
     $  + 4.103190785038813d-16*tz04
     $  - 6.035735581748754d-16*tz05
     $  - 1.455669093015461d-15*tz06
     $  + 8.976922174792779d-16*tz07
     $  + 2.726528820361550d-15*tz08
     $  - 4.963411008335192d-16*tz09
     $  - 2.833382909675777d-15*tz10
     $  - 1.887785638359819d-16*tz11
     $  + 1.569950358080909d-15*tz12
     $  + 3.283728175730937d-16*tz13
     $  - 3.854794339607276d-16*tz14

      CsH4YZ(2,2,3,1,18) = 
     $  - 1.105414293966969d-16*tz05
     $  + 1.209555006935562d-16*tz06
     $  + 3.191054517122384d-16*tz07
     $  - 1.492498283399349d-16*tz08
     $  - 5.186773098849876d-16*tz09
     $  + 4.846069568835934d-16*tz11
     $  - 2.471604176198746d-16*tz13

      CsH4YZ(2,3,0,1,02) = 
     $  + 1.333073497628959d-01
     $  - 1.400032631829942d-01*tz01
     $  + 7.253353242031792d-03*tz02
     $  - 6.143287774808880d-04*tz03
     $  + 6.338248737786133d-05*tz04
     $  - 7.289577399661965d-06*tz05
     $  + 8.986062868912105d-07*tz06
     $  - 1.162682283585841d-07*tz07
     $  + 1.559203273921266d-08*tz08
     $  - 2.149485928643684d-09*tz09
     $  + 3.029044047790532d-10*tz10
     $  - 4.345634543982877d-11*tz11
     $  + 6.327994328142610d-12*tz12
     $  - 9.331280310075600d-13*tz13
     $  + 1.390893263767670d-13*tz14
     $  - 2.092656273275074d-14*tz15
     $  + 3.174271090057370d-15*tz16
     $  - 4.849744112210558d-16*tz17

      CsH4YZ(2,3,0,1,03) = 
     $  - 2.714410648479118d-02
     $  + 2.644961320668639d-02*tz01
     $  + 7.672497392038493d-04*tz02
     $  - 8.121592712588424d-05*tz03
     $  + 9.510892528807266d-06*tz04
     $  - 1.188345309591607d-06*tz05
     $  + 1.553754900379418d-07*tz06
     $  - 2.101195737547093d-08*tz07
     $  + 2.916653071612885d-09*tz08
     $  - 4.133815069510640d-10*tz09
     $  + 5.959587096289248d-11*tz10
     $  - 8.714639736443247d-12*tz11
     $  + 1.289753981489805d-12*tz12
     $  - 1.928625360241860d-13*tz13
     $  + 2.909915571683530d-14*tz14
     $  - 4.425081093544337d-15*tz15
     $  + 6.776071482363678d-16*tz16
     $  - 1.043801514938981d-16*tz17

      CsH4YZ(2,3,0,1,04) = 
     $  + 3.631861981241309d-03
     $  - 3.078336847088837d-03*tz01
     $  - 5.232898721481075d-04*tz02
     $  - 3.391096905168217d-05*tz03
     $  + 4.147737342964838d-06*tz04
     $  - 5.350608108073969d-07*tz05
     $  + 7.169800547402239d-08*tz06
     $  - 9.887113989654414d-09*tz07
     $  + 1.394451385765799d-09*tz08
     $  - 2.002745278903196d-10*tz09
     $  + 2.919854210213987d-11*tz10
     $  - 4.310938645337076d-12*tz11
     $  + 6.433596054538678d-13*tz12
     $  - 9.691046460769989d-14*tz13
     $  + 1.471667921516946d-14*tz14
     $  - 2.250865306797525d-15*tz15
     $  + 3.464542216228534d-16*tz16

      CsH4YZ(2,3,0,1,05) = 
     $  - 3.032689797923007d-04
     $  + 1.814315378545698d-04*tz01
     $  + 9.515011043167624d-05*tz02
     $  + 2.484520329772795d-05*tz03
     $  + 2.086330006281695d-06*tz04
     $  - 2.776187448528524d-07*tz05
     $  + 3.810284923001206d-08*tz06
     $  - 5.355893167027676d-09*tz07
     $  + 7.673218222567926d-10*tz08
     $  - 1.116593043189113d-10*tz09
     $  + 1.646148695281368d-11*tz10
     $  - 2.453838199532845d-12*tz11
     $  + 3.692796662354912d-13*tz12
     $  - 5.603522615069514d-14*tz13
     $  + 8.564988468596643d-15*tz14
     $  - 1.317619258786501d-15*tz15
     $  + 2.038701900166209d-16*tz16

      CsH4YZ(2,3,0,1,06) = 
     $  + 1.060317694197476d-05
     $  + 3.397480929217735d-06*tz01
     $  - 6.998152482110958d-06*tz02
     $  - 5.154060220543831d-06*tz03
     $  - 1.705694881586097d-06*tz04
     $  - 1.626990858373985d-07*tz05
     $  + 2.279401595617781d-08*tz06
     $  - 3.257781040274069d-09*tz07
     $  + 4.732118464838786d-10*tz08
     $  - 6.966689617547030d-11*tz09
     $  + 1.037358831057305d-11*tz10
     $  - 1.559764333361441d-12*tz11
     $  + 2.365138675904984d-13*tz12
     $  - 3.613006936829165d-14*tz13
     $  + 5.555502521617968d-15*tz14
     $  - 8.592273603925119d-16*tz15
     $  + 1.335876582031364d-16*tz16

      CsH4YZ(2,3,0,1,07) = 
     $  + 6.640775990830124d-07
     $  - 1.277870820883217d-06*tz01
     $  - 2.796178388901826d-07*tz02
     $  + 3.633357778424050d-07*tz03
     $  + 3.728621457427744d-07*tz04
     $  + 1.442570960274019d-07*tz05
     $  + 1.483097938219179d-08*tz06
     $  - 2.150142189917925d-09*tz07
     $  + 3.160815201030150d-10*tz08
     $  - 4.701129558733528d-11*tz09
     $  + 7.062108375919669d-12*tz10
     $  - 1.070064789829614d-12*tz11
     $  + 1.633647166633438d-13*tz12
     $  - 2.510693380617918d-14*tz13
     $  + 3.881476433611766d-15*tz14
     $  - 6.032505140327408d-16*tz15

      CsH4YZ(2,3,0,1,08) = 
     $  - 7.330882055839667d-08
     $  + 1.994248481184838d-08*tz01
     $  + 8.661334689687232d-08*tz02
     $  + 3.512157932064040d-08*tz03
     $  - 2.094230392994493d-08*tz04
     $  - 3.212844201829097d-08*tz05
     $  - 1.398628246883032d-08*tz06
     $  - 1.506135591486225d-09*tz07
     $  + 2.237406660050437d-10*tz08
     $  - 3.357866574533820d-11*tz09
     $  + 5.084004205274165d-12*tz10
     $  - 7.756785307678884d-13*tz11
     $  + 1.191494237326461d-13*tz12
     $  - 1.841224749385312d-14*tz13
     $  + 2.860558240934211d-15*tz14
     $  - 4.465689198911255d-16*tz15

      CsH4YZ(2,3,0,1,09) = 
     $  - 2.815795672733771d-09
     $  + 1.036842724805347d-08*tz01
     $  - 1.308845775761078d-09*tz02
     $  - 7.396697277156876d-09*tz03
     $  - 4.701204511855426d-09*tz04
     $  + 1.102713801889874d-09*tz05
     $  + 3.114025630988736d-09*tz06
     $  + 1.493527869893360d-09*tz07
     $  + 1.655882000911496d-10*tz08
     $  - 2.505116218889195d-11*tz09
     $  + 3.819651656572896d-12*tz10
     $  - 5.864122634064593d-13*tz11
     $  + 9.057856066005775d-14*tz12
     $  - 1.406718601024001d-14*tz13
     $  + 2.195378975702429d-15*tz14
     $  - 3.441329354856017d-16*tz15

      CsH4YZ(2,3,0,1,10) = 
     $  + 6.216683935579888d-10
     $  - 4.625799673940817d-10*tz01
     $  - 9.065129328941287d-10*tz02
     $  - 7.331229028169577d-11*tz03
     $  + 7.176542707046777d-10*tz04
     $  + 6.403672368532468d-10*tz05
     $  - 2.060480895174383d-11*tz06
     $  - 3.285197322256401d-10*tz07
     $  - 1.713845174453006d-10*tz08
     $  - 1.934758217036450d-11*tz09
     $  + 2.968719580171645d-12*tz10
     $  - 4.583472466901807d-13*tz11
     $  + 7.115592431171179d-14*tz12
     $  - 1.110128021439258d-14*tz13
     $  + 1.739691804093874d-15*tz14
     $  - 2.737333635171543d-16*tz15

      CsH4YZ(2,3,0,1,11) = 
     $  + 1.345120333701100d-11
     $  - 9.066769122002986d-11*tz01
     $  + 4.553213193169149d-11*tz02
     $  + 1.043790595491688d-10*tz03
     $  + 3.885688514819440d-11*tz04
     $  - 7.402394758379528d-11*tz05
     $  - 8.872204554576931d-11*tz06
     $  - 8.598671256692169d-12*tz07
     $  + 3.694611589717582d-11*tz08
     $  + 2.079329648159597d-11*tz09
     $  + 2.372142260931559d-12*tz10
     $  - 3.681190363404066d-13*tz11
     $  + 5.741263400822066d-14*tz12
     $  - 8.994661853624523d-15*tz13
     $  + 1.414939745645686d-15*tz14
     $  - 2.234121894365647d-16*tz15

      CsH4YZ(2,3,0,1,12) = 
     $  - 5.862510462321587d-12
     $  + 6.527264193844958d-12*tz01
     $  + 9.238378475182674d-12*tz02
     $  - 2.661256320644280d-12*tz03
     $  - 1.372584769995003d-11*tz04
     $  - 8.873088729156703d-12*tz05
     $  + 7.753878655549794d-12*tz06
     $  + 1.251543847863457d-11*tz07
     $  + 2.354204868276673d-12*tz08
     $  - 4.367618379073199d-12*tz09
     $  - 2.637347859593354d-12*tz10
     $  - 3.024279900129946d-13*tz11
     $  + 4.736695454197044d-14*tz12
     $  - 7.449415692063697d-15*tz13
     $  + 1.175984836131425d-15*tz14
     $  - 1.862825403706270d-16*tz15

      CsH4YZ(2,3,0,1,13) = 
     $  - 5.397621505792017d-14
     $  + 8.379159299997979d-13*tz01
     $  - 7.537169829441143d-13*tz02
     $  - 1.356681189836979d-12*tz03
     $  - 1.505213394499244d-13*tz04
     $  + 1.899534725532490d-12*tz05
     $  + 1.707030688469714d-12*tz06
     $  - 7.887423591298215d-13*tz07
     $  - 1.796700276896192d-12*tz08
     $  - 4.627690433695452d-13*tz09
     $  + 5.373658726969428d-13*tz10
     $  + 3.468625510289464d-13*tz11
     $  + 3.982398646912725d-14*tz12
     $  - 6.285373072267458d-15*tz13
     $  + 9.954583882041807d-16*tz14
     $  - 1.581592292468168d-16*tz15

      CsH4YZ(2,3,0,1,14) = 
     $  + 5.918496073628377d-14
     $  - 8.285413200702108d-14*tz01
     $  - 9.535903910790065d-14*tz02
     $  + 6.723073239863192d-14*tz03
     $  + 2.282046416250682d-13*tz04
     $  + 1.018459120081715d-13*tz05
     $  - 2.669998929760985d-13*tz06
     $  - 3.079657720795156d-13*tz07
     $  + 7.230313315694649d-14*tz08
     $  + 2.621733509118692d-13*tz09
     $  + 8.220954995045071d-14*tz10
     $  - 6.830745157703873d-14*tz11
     $  - 4.701574390789728d-14*tz12
     $  - 5.388660803162089d-15*tz13
     $  + 8.560189096017991d-16*tz14
     $  - 1.363845892905507d-16*tz15

      CsH4YZ(2,3,0,1,15) = 
     $  - 8.117896506030777d-15*tz01
     $  + 1.062000686702608d-14*tz02
     $  + 1.751570926793600d-14*tz03
     $  - 2.721030006986731d-15*tz04
     $  - 3.941303455780126d-14*tz05
     $  - 2.786182705127572d-14*tz06
     $  + 3.746722124952826d-14*tz07
     $  + 5.397306093909954d-14*tz08
     $  - 4.710033403111185d-15*tz09
     $  - 3.882792115280683d-14*tz10
     $  - 1.400675065247251d-14*tz11
     $  + 8.921304086059804d-15*tz12
     $  + 6.537207221441735d-15*tz13
     $  + 7.462671173177074d-16*tz14
     $  - 1.192079059779979d-16*tz15

      CsH4YZ(2,3,0,1,16) = 
     $  - 6.272144858066114d-16
     $  + 1.017246603621712d-15*tz01
     $  + 1.017496841118302d-15*tz02
     $  - 1.176497208740688d-15*tz03
     $  - 3.641363791739350d-15*tz04
     $  - 9.419085968430168d-16*tz05
     $  + 6.775730854795112d-15*tz06
     $  + 6.301038459439455d-15*tz07
     $  - 5.184647624334443d-15*tz08
     $  - 9.327153283464290d-15*tz09
     $  - 1.705827381289315d-16*tz10
     $  + 5.827764139994003d-15*tz11
     $  + 2.341935347507506d-15*tz12
     $  - 1.191992738992531d-15*tz13
     $  - 9.289882597889350d-16*tz14
     $  - 1.054394854373170d-16*tz15

      CsH4YZ(2,3,0,1,17) = 
     $  - 1.409745753330975d-16*tz02
     $  - 2.314325648309780d-16*tz03
     $  + 7.524216692833102d-16*tz05
     $  + 4.088778408628634d-16*tz06
     $  - 1.150439279354901d-15*tz07
     $  - 1.315084036831944d-15*tz08
     $  + 6.977618842084990d-16*tz09
     $  + 1.600902576293454d-15*tz10
     $  + 1.583651283696437d-16*tz11
     $  - 8.852517779533524d-16*tz12
     $  - 3.883623703424095d-16*tz13
     $  + 1.623712860216815d-16*tz14
     $  + 1.345321388984406d-16*tz15

      CsH4YZ(2,3,0,1,18) = 
     $  - 1.513777763702748d-16*tz06
     $  - 1.155420982490689d-16*tz07
     $  + 1.924248662363397d-16*tz08
     $  + 2.629752251428324d-16*tz09
     $  - 2.739590058056893d-16*tz11
     $  + 1.359263707015031d-16*tz13

      CsH4YZ(2,3,1,1,02) = 
     $  - 1.190423870137048d-02
     $  + 2.165441659832087d-03*tz01
     $  + 1.155069337265245d-02*tz02
     $  - 2.108831635769548d-03*tz03
     $  + 3.444991544917579d-04*tz04
     $  - 5.516067018631117d-05*tz05
     $  + 8.813066502743328d-06*tz06
     $  - 1.411711405345883d-06*tz07
     $  + 2.270060462372328d-07*tz08
     $  - 3.664996417110229d-08*tz09
     $  + 5.939724777839782d-09*tz10
     $  - 9.660047982795468d-10*tz11
     $  + 1.576046066525754d-10*tz12
     $  - 2.578693998789101d-11*tz13
     $  + 4.230090855871458d-12*tz14
     $  - 6.955207077081232d-13*tz15
     $  + 1.146002814038349d-13*tz16
     $  - 1.891863416092739d-14*tz17
     $  + 3.128634478027920d-15*tz18
     $  - 5.180659355233429d-16*tz19

      CsH4YZ(2,3,1,1,03) = 
     $  + 1.449060928012249d-02
     $  - 1.707697985887450d-02*tz01
     $  + 3.007299497671930d-03*tz02
     $  - 4.883125091218116d-04*tz03
     $  + 7.815715634124096d-05*tz04
     $  - 1.250095805242814d-05*tz05
     $  + 2.005429949926640d-06*tz06
     $  - 3.229683501684381d-07*tz07
     $  + 5.221815575625490d-08*tz08
     $  - 8.474007581280796d-09*tz09
     $  + 1.379823577549619d-09*tz10
     $  - 2.253643516335099d-10*tz11
     $  + 3.690992414394798d-11*tz12
     $  - 6.060103280793012d-12*tz13
     $  + 9.972225614387205d-13*tz14
     $  - 1.644326696874475d-13*tz15
     $  + 2.716365431849836d-14*tz16
     $  - 4.494888896774925d-15*tz17
     $  + 7.449455872309748d-16*tz18
     $  - 1.235999514822470d-16*tz19

      CsH4YZ(2,3,1,1,04) = 
     $  - 3.292408554295862d-03
     $  + 3.560440132414953d-03*tz01
     $  - 4.585196039494675d-05*tz02
     $  - 2.577451530052900d-04*tz03
     $  + 4.126774304503472d-05*tz04
     $  - 6.619659063845424d-06*tz05
     $  + 1.065637821231770d-06*tz06
     $  - 1.722114727724377d-07*tz07
     $  + 2.793353806819034d-08*tz08
     $  - 4.546470667793429d-09*tz09
     $  + 7.422829214764415d-10*tz10
     $  - 1.215289373152883d-10*tz11
     $  + 1.994738036235758d-11*tz12
     $  - 3.281573358356390d-12*tz13
     $  + 5.409726086982198d-13*tz14
     $  - 8.934772117208579d-14*tz15
     $  + 1.478206329011155d-14*tz16
     $  - 2.449422398822009d-15*tz17
     $  + 4.064617376086977d-16*tz18

      CsH4YZ(2,3,1,1,05) = 
     $  + 3.577397006378389d-04
     $  - 2.718893576552964d-04*tz01
     $  - 1.363381201871340d-04*tz02
     $  + 2.884944360784197d-05*tz03
     $  + 2.511972152810227d-05*tz04
     $  - 4.043616020983527d-06*tz05
     $  + 6.533640204710944d-07*tz06
     $  - 1.059602016657372d-07*tz07
     $  + 1.724323530801606d-08*tz08
     $  - 2.814810091187523d-09*tz09
     $  + 4.607898774304006d-10*tz10
     $  - 7.562413912203536d-11*tz11
     $  + 1.243984166004118d-11*tz12
     $  - 2.050557801894424d-12*tz13
     $  + 3.386485872332349d-13*tz14
     $  - 5.602397822458682d-14*tz15
     $  + 9.282871688935788d-15*tz16
     $  - 1.540334404726899d-15*tz17
     $  + 2.559341450662738d-16*tz18

      CsH4YZ(2,3,1,1,06) = 
     $  - 1.341485712937504d-05
     $  - 7.348437276291336d-06*tz01
     $  + 1.718448137189174d-05*tz02
     $  + 1.017202217488092d-05*tz03
     $  - 4.227559439755106d-06*tz04
     $  - 2.749061958584833d-06*tz05
     $  + 4.457697012459165d-07*tz06
     $  - 7.253137601055980d-08*tz07
     $  + 1.183863446070134d-08*tz08
     $  - 1.937795050147011d-09*tz09
     $  + 3.179980792450480d-10*tz10
     $  - 5.230509902151642d-11*tz11
     $  + 8.621274268027608d-12*tz12
     $  - 1.423715919660435d-12*tz13
     $  + 2.355192447049431d-13*tz14
     $  - 3.902259348243178d-14*tz15
     $  + 6.474941922208933d-15*tz16
     $  - 1.075801718849724d-15*tz17
     $  + 1.789644446831323d-16*tz18

      CsH4YZ(2,3,1,1,07) = 
     $  - 1.110835477808637d-06
     $  + 2.313040954574701d-06*tz01
     $  + 7.239140446314768d-07*tz02
     $  - 1.731177181750094d-06*tz03
     $  - 1.052578535076791d-06*tz04
     $  + 5.765108137209179d-07*tz05
     $  + 3.269863851111122d-07*tz06
     $  - 5.336411138339033d-08*tz07
     $  + 8.733861921748463d-09*tz08
     $  - 1.433113515084153d-09*tz09
     $  + 2.357023207388995d-10*tz10
     $  - 3.884725353683688d-11*tz11
     $  + 6.414838571001840d-12*tz12
     $  - 1.061123957767455d-12*tz13
     $  + 1.758071206910781d-13*tz14
     $  - 2.917019748775602d-14*tz15
     $  + 4.846456650346267d-15*tz16
     $  - 8.062007904069584d-16*tz17
     $  + 1.342648100298927d-16*tz18

      CsH4YZ(2,3,1,1,08) = 
     $  + 9.911526236524062d-08
     $  + 1.969366536122879d-08*tz01
     $  - 2.457165396444708d-07*tz02
     $  - 1.072643983139085d-07*tz03
     $  + 2.192413818123825d-07*tz04
     $  + 1.300514431552611d-07*tz05
     $  - 7.961080752044190d-08*tz06
     $  - 4.133421637037541d-08*tz07
     $  + 6.781738080206917d-09*tz08
     $  - 1.115289107997729d-09*tz09
     $  + 1.838029421559825d-10*tz10
     $  - 3.034946055632249d-11*tz11
     $  + 5.020053378945450d-12*tz12
     $  - 8.316850608751697d-13*tz13
     $  + 1.379891244727802d-13*tz14
     $  - 2.292530066264876d-14*tz15
     $  + 3.813504762009557d-15*tz16
     $  - 6.350809711027790d-16*tz17
     $  + 1.058765245816998d-16*tz18

      CsH4YZ(2,3,1,1,09) = 
     $  + 7.263715482199698d-09
     $  - 2.256002737414099d-08*tz01
     $  - 3.822060992925014d-09*tz02
     $  + 3.474620596614489d-08*tz03
     $  + 1.734175267422654d-08*tz04
     $  - 3.115440798682235d-08*tz05
     $  - 1.778787718722047d-08*tz06
     $  + 1.127402855041347d-08*tz07
     $  + 5.472754884056285d-09*tz08
     $  - 9.018602205283485d-10*tz09
     $  + 1.489054516421722d-10*tz10
     $  - 2.462888289756900d-11*tz11
     $  + 4.080142526496085d-12*tz12
     $  - 6.769305635617150d-13*tz13
     $  + 1.124603840495794d-13*tz14
     $  - 1.870664511993283d-14*tz15
     $  + 3.115248494744372d-15*tz16
     $  - 5.193384139237091d-16*tz17

      CsH4YZ(2,3,1,1,10) = 
     $  - 8.684303951812224d-10
     $  + 5.355885022080484d-12*tz01
     $  + 3.258890767713239d-09*tz02
     $  + 1.086104799647646d-09*tz03
     $  - 5.595501109395166d-09*tz04
     $  - 2.907529318728856d-09*tz05
     $  + 4.714500522534722d-09*tz06
     $  + 2.588700722304163d-09*tz07
     $  - 1.637256906214159d-09*tz08
     $  - 7.514598763482323d-10*tz09
     $  + 1.242843505885874d-10*tz10
     $  - 2.058860577004627d-11*tz11
     $  + 3.415689715432322d-12*tz12
     $  - 5.674386429736368d-13*tz13
     $  + 9.438487435185654d-14*tz14
     $  - 1.571766002111375d-14*tz15
     $  + 2.620227584888002d-15*tz16
     $  - 4.372399561650705d-16*tz17

      CsH4YZ(2,3,1,1,11) = 
     $  - 6.183990634422480d-11
     $  + 2.361977341301152d-10*tz01
     $  + 1.999986236692527d-11*tz02
     $  - 5.969247794567086d-10*tz03
     $  - 2.531059416667007d-10*tz04
     $  + 9.559765089165472d-10*tz05
     $  + 4.986469308592427d-10*tz06
     $  - 7.409532286319153d-10*tz07
     $  - 3.923210326896610d-10*tz08
     $  + 2.431874515515247d-10*tz09
     $  + 1.062525353993665d-10*tz10
     $  - 1.762679303965367d-11*tz11
     $  + 2.928186516253526d-12*tz12
     $  - 4.870451193239210d-13*tz13
     $  + 8.110422109643934d-14*tz14
     $  - 1.352024836992167d-14*tz15
     $  + 2.256108973814619d-15*tz16
     $  - 3.768225246156444d-16*tz17

      CsH4YZ(2,3,1,1,12) = 
     $  + 8.258736619424082d-12
     $  - 7.451683051489755d-13*tz01
     $  - 4.309303283812516d-11*tz02
     $  - 1.206107382878804d-11*tz03
     $  + 1.178573361713120d-10*tz04
     $  + 5.433992273206745d-11*tz05
     $  - 1.682151246775313d-10*tz06
     $  - 8.686047702464819d-11*tz07
     $  + 1.193957777160438d-10*tz08
     $  + 6.115558486819834d-11*tz09
     $  - 3.683658275270379d-11*tz10
     $  - 1.539033720363601d-11*tz11
     $  + 2.559796609086816d-12*tz12
     $  - 4.262538638570410d-13*tz13
     $  + 7.105582993376535d-14*tz14
     $  - 1.185677034564927d-14*tz15
     $  + 1.980335679102677d-15*tz16
     $  - 3.310448729346960d-16*tz17

      CsH4YZ(2,3,1,1,13) = 
     $  + 6.059319893024794d-13
     $  - 2.590642985844534d-12*tz01
     $  - 1.678309695364648d-13*tz02
     $  + 9.735471690229527d-12*tz03
     $  + 3.704834364385137d-12*tz04
     $  - 2.372100879276105d-11*tz05
     $  - 1.125403492887674d-11*tz06
     $  + 3.008048718865824d-11*tz07
     $  + 1.529808003449091d-11*tz08
     $  - 1.958426803035976d-11*tz09
     $  - 9.731938191568310d-12*tz10
     $  + 5.675181481403780d-12*tz11
     $  + 2.274700565163892d-12*tz12
     $  - 3.791801955348073d-13*tz13
     $  + 6.327085838482096d-14*tz14
     $  - 1.056740763717844d-14*tz15
     $  + 1.766497587806217d-15*tz16
     $  - 2.955356384272701d-16*tz17

      CsH4YZ(2,3,1,1,14) = 
     $  - 8.241125261067356d-14
     $  + 5.764330467409486d-15*tz01
     $  + 5.752622720108517d-13*tz02
     $  + 1.587520208931988d-13*tz03
     $  - 2.276588832063030d-12*tz04
     $  - 9.575694123610028d-13*tz05
     $  + 4.787043609075542d-12*tz06
     $  + 2.287382265839491d-12*tz07
     $  - 5.430601523667602d-12*tz08
     $  - 2.715863268289181d-12*tz09
     $  + 3.255807576925210d-12*tz10
     $  + 1.573372892876771d-12*tz11
     $  - 8.872772985408440d-13*tz12
     $  - 3.420150132486214d-13*tz13
     $  + 5.712186063277015d-14*tz14
     $  - 9.548610311443847d-15*tz15
     $  + 1.597475887144011d-15*tz16
     $  - 2.674602794982942d-16*tz17

      CsH4YZ(2,3,1,1,15) = 
     $  - 6.453528819516857d-15
     $  + 2.935636669809936d-14*tz01
     $  + 3.235842182390993d-15*tz02
     $  - 1.550528596975358d-13*tz03
     $  - 5.683656781156050d-14*tz04
     $  + 5.268871484131434d-13*tz05
     $  + 2.301099273750903d-13*tz06
     $  - 9.636288223167566d-13*tz07
     $  - 4.595581276923710d-13*tz08
     $  + 9.863620087530330d-13*tz09
     $  + 4.849860274370994d-13*tz10
     $  - 5.470402053625326d-13*tz11
     $  - 2.575816488613976d-13*tz12
     $  + 1.405057470695094d-13*tz13
     $  + 5.218750009741754d-14*tz14
     $  - 8.730812511228815d-15*tz15
     $  + 1.461765385492037d-15*tz16
     $  - 2.449130721351335d-16*tz17

      CsH4YZ(2,3,1,1,16) = 
     $  + 8.481082124043085d-16
     $  - 7.754346682877199d-15*tz02
     $  - 2.465684698704610d-15*tz03
     $  + 4.183395122531081d-14*tz04
     $  + 1.664404456676854d-14*tz05
     $  - 1.198286258981182d-13*tz06
     $  - 5.310997302809295d-14*tz07
     $  + 1.931803520730020d-13*tz08
     $  + 9.158018513488878d-14*tz09
     $  - 1.798834913120531d-13*tz10
     $  - 8.699164250807200d-14*tz11
     $  + 9.271421268417985d-14*tz12
     $  + 4.260384311744354d-14*tz13
     $  - 2.250067092569805d-14*tz14
     $  - 8.065828607819888d-15*tz15
     $  + 1.351393804571539d-15*tz16
     $  - 2.265730266109206d-16*tz17

      CsH4YZ(2,3,1,1,17) = 
     $  - 3.403876989064030d-16*tz01
     $  + 2.433939238664249d-15*tz03
     $  + 9.224738966859186d-16*tz04
     $  - 1.091953992890411d-14*tz05
     $  - 4.475649841310487d-15*tz06
     $  + 2.679455670528754d-14*tz07
     $  + 1.193612847232700d-14*tz08
     $  - 3.855459699721106d-14*tz09
     $  - 1.813546118075660d-14*tz10
     $  + 3.290033877191897d-14*tz11
     $  + 1.565734758897241d-14*tz12
     $  - 1.582816323905598d-14*tz13
     $  - 7.107236195823188d-15*tz14
     $  + 3.639042441966952d-15*tz15
     $  + 1.260690764898443d-15*tz16
     $  - 2.115002610584920d-16*tz17

      CsH4YZ(2,3,1,1,18) = 
     $  + 1.052517483301987d-16*tz02
     $  - 7.430666619530045d-16*tz04
     $  - 2.909338268312017d-16*tz05
     $  + 2.758089880278705d-15*tz06
     $  + 1.144936085427669d-15*tz07
     $  - 5.902046459839996d-15*tz08
     $  - 2.630603661360921d-15*tz09
     $  + 7.661333900034126d-15*tz10
     $  + 3.572878815595071d-15*tz11
     $  - 6.030358311744589d-15*tz12
     $  - 2.825810534383330d-15*tz13
     $  + 2.719025136684441d-15*tz14
     $  + 1.194311663493668d-15*tz15
     $  - 5.937222813956515d-16*tz16
     $  - 1.990108154642708d-16*tz17

      CsH4YZ(2,3,1,1,19) = 
     $  + 2.155469517096877d-16*tz05
     $  - 6.772796952731661d-16*tz07
     $  - 2.828561259205094d-16*tz08
     $  + 1.283137803864207d-15*tz09
     $  + 5.709071144048492d-16*tz10
     $  - 1.516275413767605d-15*tz11
     $  - 7.008273468445557d-16*tz12
     $  + 1.107162275906370d-15*tz13
     $  + 5.111287154135675d-16*tz14
     $  - 4.696091578935223d-16*tz15
     $  - 2.019614936710457d-16*tz16

      CsH4YZ(2,3,1,1,20) = 
     $  + 1.624060336131467d-16*tz08
     $  - 2.758045667063081d-16*tz10
     $  - 1.223457980520182d-16*tz11
     $  + 2.989862375256672d-16*tz12
     $  + 1.369508950836737d-16*tz13
     $  - 2.035460044899392d-16*tz14

      CsH4YZ(2,3,2,1,02) = 
     $  + 1.333073497628959d-01
     $  - 1.400032631829942d-01*tz01
     $  + 7.253353242031792d-03*tz02
     $  - 6.143287774808880d-04*tz03
     $  + 6.338248737786133d-05*tz04
     $  - 7.289577399661965d-06*tz05
     $  + 8.986062868912105d-07*tz06
     $  - 1.162682283585841d-07*tz07
     $  + 1.559203273921266d-08*tz08
     $  - 2.149485928643684d-09*tz09
     $  + 3.029044047790532d-10*tz10
     $  - 4.345634543982877d-11*tz11
     $  + 6.327994328142610d-12*tz12
     $  - 9.331280310075600d-13*tz13
     $  + 1.390893263767670d-13*tz14
     $  - 2.092656273275074d-14*tz15
     $  + 3.174271090057370d-15*tz16
     $  - 4.849744112210558d-16*tz17

      CsH4YZ(2,3,2,1,03) = 
     $  + 7.832464351520881d-02
     $  - 8.706636793313607d-03*tz01
     $  + 7.672497392038493d-04*tz02
     $  - 8.121592712588424d-05*tz03
     $  + 9.510892528807266d-06*tz04
     $  - 1.188345309591607d-06*tz05
     $  + 1.553754900379418d-07*tz06
     $  - 2.101195737547093d-08*tz07
     $  + 2.916653071612885d-09*tz08
     $  - 4.133815069510640d-10*tz09
     $  + 5.959587096289248d-11*tz10
     $  - 8.714639736443247d-12*tz11
     $  + 1.289753981489805d-12*tz12
     $  - 1.928625360241860d-13*tz13
     $  + 2.909915571683530d-14*tz14
     $  - 4.425081093544337d-15*tz15
     $  + 6.776071482363678d-16*tz16
     $  - 1.043801514938981d-16*tz17

      CsH4YZ(2,3,2,1,04) = 
     $  + 2.807887371866309d-03
     $  + 3.513460027911162d-03*tz01
     $  + 3.006847372268924d-04*tz02
     $  - 3.391096905168217d-05*tz03
     $  + 4.147737342964838d-06*tz04
     $  - 5.350608108073969d-07*tz05
     $  + 7.169800547402239d-08*tz06
     $  - 9.887113989654414d-09*tz07
     $  + 1.394451385765799d-09*tz08
     $  - 2.002745278903196d-10*tz09
     $  + 2.919854210213987d-11*tz10
     $  - 4.310938645337076d-12*tz11
     $  + 6.433596054538678d-13*tz12
     $  - 9.691046460769989d-14*tz13
     $  + 1.471667921516946d-14*tz14
     $  - 2.250865306797525d-15*tz15
     $  + 3.464542216228534d-16*tz16

      CsH4YZ(2,3,2,1,05) = 
     $  + 7.266992819264492d-04
     $  - 2.717544973016801d-04*tz01
     $  - 1.108435419120737d-04*tz02
     $  - 1.635352717102204d-05*tz03
     $  + 2.086330006281695d-06*tz04
     $  - 2.776187448528524d-07*tz05
     $  + 3.810284923001206d-08*tz06
     $  - 5.355893167027676d-09*tz07
     $  + 7.673218222567926d-10*tz08
     $  - 1.116593043189113d-10*tz09
     $  + 1.646148695281368d-11*tz10
     $  - 2.453838199532845d-12*tz11
     $  + 3.692796662354912d-13*tz12
     $  - 5.603522615069514d-14*tz13
     $  + 8.564988468596643d-15*tz14
     $  - 1.317619258786501d-15*tz15
     $  + 2.038701900166209d-16*tz16

      CsH4YZ(2,3,2,1,06) = 
     $  - 9.030593047038908d-06
     $  - 8.189662015118202d-06*tz01
     $  + 9.738831770818728d-06*tz02
     $  + 6.433082723792105d-06*tz03
     $  + 1.191090854497887d-06*tz04
     $  - 1.626990858373985d-07*tz05
     $  + 2.279401595617781d-08*tz06
     $  - 3.257781040274069d-09*tz07
     $  + 4.732118464838786d-10*tz08
     $  - 6.966689617547030d-11*tz09
     $  + 1.037358831057305d-11*tz10
     $  - 1.559764333361441d-12*tz11
     $  + 2.365138675904984d-13*tz12
     $  - 3.613006936829165d-14*tz13
     $  + 5.555502521617968d-15*tz14
     $  - 8.592273603925119d-16*tz15
     $  + 1.335876582031364d-16*tz16

      CsH4YZ(2,3,2,1,07) = 
     $  - 2.329267661537104d-06
     $  + 2.060329884508802d-06*tz01
     $  + 4.928583573988798d-07*tz02
     $  - 5.194941607736663d-07*tz03
     $  - 4.961735750824208d-07*tz04
     $  - 1.040388242083680d-07*tz05
     $  + 1.483097938219179d-08*tz06
     $  - 2.150142189917925d-09*tz07
     $  + 3.160815201030150d-10*tz08
     $  - 4.701129558733528d-11*tz09
     $  + 7.062108375919669d-12*tz10
     $  - 1.070064789829614d-12*tz11
     $  + 1.633647166633438d-13*tz12
     $  - 2.510693380617918d-14*tz13
     $  + 3.881476433611766d-15*tz14
     $  - 6.032505140327408d-16*tz15

      CsH4YZ(2,3,2,1,08) = 
     $  + 5.601197123106691d-08
     $  + 3.546097982658401d-08*tz01
     $  - 1.212698259046903d-07*tz02
     $  - 5.798939076777337d-08*tz03
     $  + 3.337242862162977d-08*tz04
     $  + 4.546403305538717d-08*tz05
     $  + 1.026136599169409d-08*tz06
     $  - 1.506135591486225d-09*tz07
     $  + 2.237406660050437d-10*tz08
     $  - 3.357866574533820d-11*tz09
     $  + 5.084004205274165d-12*tz10
     $  - 7.756785307678884d-13*tz11
     $  + 1.191494237326461d-13*tz12
     $  - 1.841224749385312d-14*tz13
     $  + 2.860558240934211d-15*tz14
     $  - 4.465689198911255d-16*tz15

      CsH4YZ(2,3,2,1,09) = 
     $  + 1.521983618218963d-08
     $  - 1.811678644342925d-08*tz01
     $  - 1.643694254501653d-09*tz02
     $  + 9.849923794527240d-09*tz03
     $  + 7.561177709609781d-09*tz04
     $  - 2.303503481850460d-09*tz05
     $  - 4.679861374179828d-09*tz06
     $  - 1.104434465162827d-09*tz07
     $  + 1.655882000911496d-10*tz08
     $  - 2.505116218889195d-11*tz09
     $  + 3.819651656572896d-12*tz10
     $  - 5.864122634064593d-13*tz11
     $  + 9.057856066005775d-14*tz12
     $  - 1.406718601024001d-14*tz13
     $  + 2.195378975702429d-15*tz14
     $  - 3.441329354856017d-16*tz15

      CsH4YZ(2,3,2,1,10) = 
     $  - 4.260952285917031d-10
     $  - 1.870877114474984d-10*tz01
     $  + 1.339858499484455d-09*tz02
     $  + 4.630586834767797d-10*tz03
     $  - 9.525107654445691d-10*tz04
     $  - 1.023952384042123d-09*tz05
     $  + 1.525926800520020d-10*tz06
     $  + 5.239366589646716d-10*tz07
     $  + 1.269752194713084d-10*tz08
     $  - 1.934758217036450d-11*tz09
     $  + 2.968719580171645d-12*tz10
     $  - 4.583472466901807d-13*tz11
     $  + 7.115592431171179d-14*tz12
     $  - 1.110128021439258d-14*tz13
     $  + 1.739691804093874d-15*tz14
     $  - 2.737333635171543d-16*tz15

      CsH4YZ(2,3,2,1,11) = 
     $  - 1.250256015430151d-10
     $  + 1.721078343194460d-10*tz01
     $  - 4.857512525335820d-12*tz02
     $  - 1.395907613604116d-10*tz03
     $  - 9.818562802601951d-11*tz04
     $  + 1.007554446220728d-10*tz05
     $  + 1.426589749201724d-10*tz06
     $  - 7.307070663979575d-12*tz07
     $  - 6.250712974169387d-11*tz08
     $  - 1.537152011435664d-11*tz09
     $  + 2.372142260931559d-12*tz10
     $  - 3.681190363404066d-13*tz11
     $  + 5.741263400822066d-14*tz12
     $  - 8.994661853624523d-15*tz13
     $  + 1.414939745645686d-15*tz14
     $  - 2.234121894365647d-16*tz15

      CsH4YZ(2,3,2,1,12) = 
     $  + 3.603461143496154d-12
     $  + 1.161125017050100d-12*tz01
     $  - 1.455951620752218d-11*tz02
     $  - 3.762345825931766d-12*tz03
     $  + 1.758716352666291d-11*tz04
     $  + 1.827589444373604d-11*tz05
     $  - 1.114846537583602d-11*tz06
     $  - 2.037194161330984d-11*tz07
     $  - 3.016488504885969d-13*tz08
     $  + 7.838007222060809d-12*tz09
     $  + 1.939761740831899d-12*tz10
     $  - 3.024279900129946d-13*tz11
     $  + 4.736695454197044d-14*tz12
     $  - 7.449415692063697d-15*tz13
     $  + 1.175984836131425d-15*tz14
     $  - 1.862825403706270d-16*tz15

      CsH4YZ(2,3,2,1,13) = 
     $  + 1.154875972798500d-12
     $  - 1.723661267979712d-12*tz01
     $  + 1.962279951912818d-13*tz02
     $  + 1.827761039228631d-12*tz03
     $  + 1.300768053816226d-12*tz04
     $  - 2.401329523973546d-12*tz05
     $  - 3.233734720070133d-12*tz06
     $  + 1.253361161028827d-12*tz07
     $  + 2.973048616476251d-12*tz08
     $  + 1.836098765040288d-13*tz09
     $  - 1.023012400175302d-12*tz10
     $  - 2.532829385373018d-13*tz11
     $  + 3.982398646912725d-14*tz12
     $  - 6.285373072267458d-15*tz13
     $  + 9.954583882041807d-16*tz14
     $  - 1.581592292468168d-16*tz15

      CsH4YZ(2,3,2,1,14) = 
     $  - 3.259787972019847d-14
     $  - 8.928047297355507d-15*tz01
     $  + 1.594149906812823d-13*tz02
     $  + 3.619032764633074d-14*tz03
     $  - 2.853004362397196d-13*tz04
     $  - 3.112128400552635d-13*tz05
     $  + 3.390709883426982d-13*tz06
     $  + 5.603280563665410d-13*tz07
     $  - 1.394485973332981d-13*tz08
     $  - 4.422474174665146d-13*tz09
     $  - 4.264214564681342d-14*tz10
     $  + 1.379925604613590d-13*tz11
     $  + 3.403068939290186d-14*tz12
     $  - 5.388660803162089d-15*tz13
     $  + 8.560189096017991d-16*tz14
     $  - 1.363845892905507d-16*tz15

      CsH4YZ(2,3,2,1,15) = 
     $  - 1.147325431714815d-14
     $  + 1.796939637784163d-14*tz01
     $  - 2.999767443173970d-15*tz02
     $  - 2.340959198492547d-14*tz03
     $  - 1.856647572240922d-14*tz04
     $  + 4.768471490765082d-14*tz05
     $  + 6.623453701376291d-14*tz06
     $  - 4.850530658038900d-14*tz07
     $  - 9.627574085028797d-14*tz08
     $  + 1.482075443877198d-14*tz09
     $  + 6.689825884289554d-14*tz10
     $  + 8.210173720893403d-15*tz11
     $  - 1.913323051806297d-14*tz12
     $  - 4.684606620207377d-15*tz13
     $  + 7.462671173177074d-16*tz14
     $  - 1.192079059779979d-16*tz15

      CsH4YZ(2,3,2,1,16) = 
     $  + 3.091905078951723d-16
     $  - 1.772175283105981d-15*tz02
     $  - 4.599850994561131d-16*tz03
     $  + 4.402893624558563d-15*tz04
     $  + 5.306139541182968d-15*tz05
     $  - 8.097106705761819d-15*tz06
     $  - 1.336896623305046d-14*tz07
     $  + 6.952058727104884d-15*tz08
     $  + 1.649813333307366d-14*tz09
     $  - 1.401326738842335d-15*tz10
     $  - 1.027019500189549d-14*tz11
     $  - 1.469640559272865d-15*tz12
     $  + 2.715603152295998d-15*tz13
     $  + 6.584725710470305d-16*tz14
     $  - 1.054394854373170d-16*tz15

      CsH4YZ(2,3,2,1,17) = 
     $  + 1.198426903656030d-16
     $  - 1.933805134074031d-16*tz01
     $  + 2.996322849088126d-16*tz03
     $  + 2.842019063759471d-16*tz04
     $  - 8.725466507432096d-16*tz05
     $  - 1.297643588040270d-15*tz06
     $  + 1.374167852361485d-15*tz07
     $  + 2.621803680951953d-15*tz08
     $  - 9.901005009818281d-16*tz09
     $  - 2.827673754565798d-15*tz10
     $  + 1.597386669324432d-15*tz12
     $  + 2.541641967086180d-16*tz13
     $  - 3.932400047709063d-16*tz14

      CsH4YZ(2,3,2,1,18) = 
     $  + 1.723929022744847d-16*tz06
     $  + 2.956065951671399d-16*tz07
     $  - 2.317657858732942d-16*tz08
     $  - 5.052342027137783d-16*tz09
     $  + 1.389352290184161d-16*tz10
     $  + 4.854286720715009d-16*tz11
     $  - 2.513454960490815d-16*tz13

      CsH4YZ(2,3,3,1,02) = 
     $  + 1.214031110615254d-01
     $  - 1.378378215231621d-01*tz01
     $  + 1.880404661468424d-02*tz02
     $  - 2.723160413250436d-03*tz03
     $  + 4.078816418696192d-04*tz04
     $  - 6.245024758597314d-05*tz05
     $  + 9.711672789634549d-06*tz06
     $  - 1.527979633704467d-06*tz07
     $  + 2.425980789764564d-07*tz08
     $  - 3.879945009974598d-08*tz09
     $  + 6.242629182630277d-09*tz10
     $  - 1.009461143719375d-09*tz11
     $  + 1.639326009928332d-10*tz12
     $  - 2.672006801889857d-11*tz13
     $  + 4.369180195335546d-12*tz14
     $  - 7.164472704408739d-13*tz15
     $  + 1.177745670572273d-13*tz16
     $  - 1.940360857214845d-14*tz17
     $  + 3.203206014185449d-15*tz18
     $  - 5.295955227642772d-16*tz19

      CsH4YZ(2,3,3,1,03) = 
     $  - 2.467171247748191d-02
     $  + 2.624684786058574d-02*tz01
     $  - 1.689035805023855d-03*tz02
     $  + 1.205995894474120d-04*tz03
     $  - 6.625871739455305d-06*tz04
     $  - 2.123890991706944d-07*tz05
     $  + 1.767421765834663d-07*tz06
     $  - 4.565940678963267d-08*tz07
     $  + 9.562677483411004d-09*tz08
     $  - 1.838938126348757d-09*tz09
     $  + 3.383584740349736d-10*tz10
     $  - 6.066521007274742d-11*tz11
     $  + 1.069971555709401d-11*tz12
     $  - 1.866543835830643d-12*tz13
     $  + 3.231418997241640d-13*tz14
     $  - 5.563932957718339d-14*tz15
     $  + 9.542062178636613d-15*tz16
     $  - 1.631607867932839d-15*tz17
     $  + 2.783736313231933d-16*tz18

      CsH4YZ(2,3,3,1,04) = 
     $  + 3.296202604898025d-03
     $  - 3.109648987999290d-03*tz01
     $  - 1.872608736635778d-04*tz02
     $  - 2.142390966339690d-06*tz03
     $  + 3.642829638265075d-06*tz04
     $  - 9.618280944334502d-07*tz05
     $  + 2.011459245910544d-07*tz06
     $  - 3.846822014449187d-08*tz07
     $  + 7.035131636278705d-09*tz08
     $  - 1.254100652917583d-09*tz09
     $  + 2.200355313903377d-10*tz10
     $  - 3.820536179279902d-11*tz11
     $  + 6.586635800349198d-12*tz12
     $  - 1.129873895204795d-12*tz13
     $  + 1.931240341716519d-13*tz14
     $  - 3.292343738189128d-14*tz15
     $  + 5.601906265081265d-15*tz16
     $  - 9.517903754829778d-16*tz17
     $  + 1.615429190401674d-16*tz18

      CsH4YZ(2,3,3,1,05) = 
     $  - 2.757529771760167d-04
     $  + 1.943160407157849d-04*tz01
     $  + 6.977902254103558d-05*tz02
     $  + 1.189742799105931d-05*tz03
     $  - 8.356211402845971d-08*tz04
     $  - 2.057102089334607d-07*tz05
     $  + 6.105806954293847d-08*tz06
     $  - 1.357679824543151d-08*tz07
     $  + 2.709552215564064d-09*tz08
     $  - 5.120098495363024d-10*tz09
     $  + 9.371626015147196d-11*tz10
     $  - 1.680882550968259d-11*tz11
     $  + 2.973789110520401d-12*tz12
     $  - 5.210680954542983d-13*tz13
     $  + 9.066411804053473d-14*tz14
     $  - 1.569315095206405d-14*tz15
     $  + 2.705605823821118d-15*tz16
     $  - 4.650383676454797d-16*tz17

      CsH4YZ(2,3,3,1,06) = 
     $  + 9.933069656006460d-06
     $  + 1.606193315077124d-06*tz01
     $  - 7.176870764160799d-06*tz02
     $  - 3.527160844108697d-06*tz03
     $  - 8.517412652124760d-07*tz04
     $  + 3.736280167505842d-09*tz05
     $  + 1.693573873961832d-08*tz06
     $  - 5.129765905285960d-09*tz07
     $  + 1.166752836129496d-09*tz08
     $  - 2.376238670120197d-10*tz09
     $  + 4.570787113296548d-11*tz10
     $  - 8.497618359983367d-12*tz11
     $  + 1.545221608051209d-12*tz12
     $  - 2.767368043615502d-13*tz13
     $  + 4.902227130459212d-14*tz14
     $  - 8.613940260384288d-15*tz15
     $  + 1.504302929477770d-15*tz16
     $  - 2.614592371488288d-16*tz17

      CsH4YZ(2,3,3,1,07) = 
     $  + 5.451107731128132d-07
     $  - 1.200066399418598d-06*tz01
     $  - 2.561414834828067d-08*tz02
     $  + 3.564445797115075d-07*tz03
     $  + 2.525487663131372d-07*tz04
     $  + 7.287653223180971d-08*tz05
     $  - 8.703843557767750d-11*tz06
     $  - 1.612239296854830d-09*tz07
     $  + 4.934159263656185d-10*tz08
     $  - 1.139193545247387d-10*tz09
     $  + 2.353544280135334d-11*tz10
     $  - 4.586346855726670d-12*tz11
     $  + 8.626797693228279d-13*tz12
     $  - 1.585297575751721d-13*tz13
     $  + 2.866224578459441d-14*tz14
     $  - 5.121238045005935d-15*tz15
     $  + 9.069496972341929d-16*tz16
     $  - 1.595266317300862d-16*tz17

      CsH4YZ(2,3,3,1,08) = 
     $  - 6.267485509984557d-08
     $  + 3.120708742610104d-08*tz01
     $  + 7.002814407038736d-08*tz02
     $  + 1.201025758102776d-08*tz03
     $  - 2.179588286660687d-08*tz04
     $  - 2.176631266509858d-08*tz05
     $  - 7.132760268483718d-09*tz06
     $  - 1.399600133816231d-12*tz07
     $  + 1.675929648711953d-10*tz08
     $  - 5.188598379862451d-11*tz09
     $  + 1.213045250447894d-11*tz10
     $  - 2.535335734335792d-12*tz11
     $  + 4.992717085102608d-13*tz12
     $  - 9.480609600177257d-14*tz13
     $  + 1.757217891322208d-14*tz14
     $  - 3.201952772836595d-15*tz15
     $  + 5.761960318563417d-16*tz16
     $  - 1.027116548027895d-16*tz17

      CsH4YZ(2,3,3,1,09) = 
     $  - 2.289419567663074d-09
     $  + 8.793950718685661d-09*tz01
     $  - 3.122431324297041d-09*tz02
     $  - 5.437619659088588d-09*tz03
     $  - 2.263172197510258d-09*tz04
     $  + 1.438395157949764d-09*tz05
     $  + 2.126250948581426d-09*tz06
     $  + 7.676765991599954d-10*tz07
     $  + 3.101455250888038d-13*tz08
     $  - 1.864325622688869d-11*tz09
     $  + 5.839685765215869d-12*tz10
     $  - 1.379931356049682d-12*tz11
     $  + 2.911993594807085d-13*tz12
     $  - 5.784354105950012d-14*tz13
     $  + 1.107036368551069d-14*tz14
     $  - 2.066582432794391d-15*tz15
     $  + 3.790295664400248d-16*tz16

      CsH4YZ(2,3,3,1,10) = 
     $  + 5.069332738535181d-10
     $  - 4.999652377298229d-10*tz01
     $  - 6.074391941905415d-10*tz02
     $  + 1.342021481489987d-10*tz03
     $  + 5.193741896945755d-10*tz04
     $  + 3.500424544211104d-10*tz05
     $  - 8.869159519035121d-11*tz06
     $  - 2.274266186796878d-10*tz07
     $  - 8.867521833689652d-11*tz08
     $  + 1.505824884492693d-14*tz09
     $  + 2.187950673784317d-12*tz10
     $  - 6.932848260592424d-13*tz11
     $  + 1.653552762227708d-13*tz12
     $  - 3.517939730877402d-14*tz13
     $  + 7.039194972686284d-15*tz14
     $  - 1.356137073723059d-15*tz15
     $  + 2.546907542914151d-16*tz16

      CsH4YZ(2,3,3,1,11) = 
     $  + 1.202200172132670d-11
     $  - 7.087463460031317d-11*tz01
     $  + 5.294654793326105d-11*tz02
     $  + 6.325415583305910d-11*tz03
     $  + 9.032541440050273d-12*tz04
     $  - 5.464645182640095d-11*tz05
     $  - 5.163335177071155d-11*tz06
     $  + 3.240889886992980d-12*tz07
     $  + 2.605100222995019d-11*tz08
     $  + 1.082040146599558d-11*tz09
     $  - 1.396644275769310d-14*tz10
     $  - 2.680705581960986d-13*tz11
     $  + 8.591590450050890d-14*tz12
     $  - 2.066199987156660d-14*tz13
     $  + 4.426939392893354d-15*tz14
     $  - 8.913744741711565d-16*tz15
     $  + 1.727029963729258d-16*tz16

      CsH4YZ(2,3,3,1,12) = 
     $  - 4.664691062953746d-12
     $  + 6.173858719789080d-12*tz01
     $  + 5.155675137152461d-12*tz02
     $  - 3.663884242609646d-12*tz03
     $  - 8.371344114037277d-12*tz04
     $  - 4.085210704529062d-12*tz05
     $  + 5.985853145727252d-12*tz06
     $  + 7.562241022055891d-12*tz07
     $  + 4.064250790432808d-13*tz08
     $  - 3.148561738832019d-12*tz09
     $  - 1.379407323919506d-12*tz10
     $  + 3.856441725858846d-15*tz11
     $  + 3.402495739424112d-14*tz12
     $  - 1.102911359358564d-14*tz13
     $  + 2.672363955802009d-15*tz14
     $  - 5.761348217690670d-16*tz15
     $  + 1.166406821557316d-16*tz16

      CsH4YZ(2,3,3,1,13) = 
     $  - 6.746101533800082d-14
     $  + 6.130508295394736d-13*tz01
     $  - 6.925328391820404d-13*tz02
     $  - 7.026511019348504d-13*tz03
     $  + 9.139020190321395d-14*tz04
     $  + 1.190575720120857d-12*tz05
     $  + 9.152275190585088d-13*tz06
     $  - 6.595407612954156d-13*tz07
     $  - 1.114356436190379d-12*tz08
     $  - 1.491871685707850d-13*tz09
     $  + 3.973538272712000d-13*tz10
     $  + 1.822452333262101d-13*tz11
     $  - 8.326414188683311d-16*tz12
     $  - 4.447757700313809d-15*tz13
     $  + 1.458128152416018d-15*tz14
     $  - 3.557542169741580d-16*tz15

      CsH4YZ(2,3,3,1,14) = 
     $  + 4.647520043807212d-14
     $  - 7.190129965262648d-14*tz01
     $  - 4.435939423197498d-14*tz02
     $  + 5.739394146085983d-14*tz03
     $  + 1.249808759749888d-13*tz04
     $  + 4.326743497467894d-14*tz05
     $  - 1.722556026562223d-13*tz06
     $  - 1.761238946632261d-13*tz07
     $  + 7.062963263644412d-14*tz08
     $  + 1.659701441374518d-13*tz09
     $  + 3.209220913562738d-14*tz10
     $  - 5.196554791809725d-14*tz11
     $  - 2.480379872990276d-14*tz12
     $  + 1.618926235557437d-16*tz13
     $  + 5.960864209968255d-16*tz14
     $  - 1.976483416033587d-16*tz15

      CsH4YZ(2,3,3,1,15) = 
     $  + 3.584057937483053d-16
     $  - 5.628247470324740d-15*tz01
     $  + 8.319309328644067d-15*tz02
     $  + 8.114137430268555d-15*tz03
     $  - 2.890965646127426d-15*tz04
     $  - 2.268338467190540d-14*tz05
     $  - 1.490607983209829d-14*tz06
     $  + 2.492342040012317d-14*tz07
     $  + 3.198420954509956d-14*tz08
     $  - 6.947045838181294d-15*tz09
     $  - 2.501632726780898d-14*tz10
     $  - 6.001664302762291d-15*tz11
     $  + 7.002270453403864d-15*tz12
     $  + 3.461537526177094d-15*tz13

      CsH4YZ(2,3,3,1,16) = 
     $  - 4.886264902320844d-16
     $  + 8.259341059610698d-16*tz01
     $  + 3.996518076911161d-16*tz02
     $  - 7.626925812226485d-16*tz03
     $  - 1.885760809782187d-15*tz04
     $  - 4.669623090242972d-16*tz05
     $  + 4.034305185412774d-15*tz06
     $  + 3.611573353428433d-15*tz07
     $  - 3.575785641780999d-15*tz08
     $  - 5.656724009886162d-15*tz09
     $  + 5.372616058973323d-16*tz10
     $  + 3.815786779293908d-15*tz11
     $  + 1.059857712544290d-15*tz12
     $  - 9.678921453839288d-16*tz13
     $  - 4.935513582332084d-16*tz14

      CsH4YZ(2,3,3,1,17) = 
     $  - 1.015502620348495d-16*tz03
     $  + 4.158321078274309d-16*tz05
     $  + 2.397451465119497d-16*tz06
     $  - 7.025893780528199d-16*tz07
     $  - 7.787201751752199d-16*tz08
     $  + 5.047910023873583d-16*tz09
     $  + 9.876066144769684d-16*tz10
     $  - 5.886525113796243d-16*tz12
     $  - 1.821419237770972d-16*tz13
     $  + 1.367615194939982d-16*tz14

      CsH4YZ(2,3,3,1,18) = 
     $  + 1.201329863672655d-16*tz08
     $  + 1.587677728921011d-16*tz09
     $  - 1.713505382296559d-16*tz11

      CsH4YZ(3,0,2,1,01) = 
     $  + 1.273065662995036d-01
     $  - 1.327372371952402d-01*tz01
     $  + 5.806883908044782d-03*tz02
     $  - 4.088567769137036d-04*tz03
     $  + 3.587858437135829d-05*tz04
     $  - 3.584987499767074d-06*tz05
     $  + 3.905504930632609d-07*tz06
     $  - 4.526743105842968d-08*tz07
     $  + 5.497976370437005d-09*tz08
     $  - 6.926545689584077d-10*tz09
     $  + 8.987294282706789d-11*tz10
     $  - 1.194747785211964d-11*tz11
     $  + 1.620886414121046d-12*tz12
     $  - 2.237376462301967d-13*tz13
     $  + 3.134687381252438d-14*tz14
     $  - 4.449173746791960d-15*tz15
     $  + 6.387248764388301d-16*tz16

      CsH4YZ(3,0,2,1,02) = 
     $  + 1.093062945470712d-01
     $  - 1.012326374765271d-02*tz01
     $  + 7.397941940233271d-04*tz02
     $  - 6.665050246110976d-05*tz03
     $  + 6.791976269447868d-06*tz04
     $  - 7.513266348234561d-07*tz05
     $  + 8.815824984963153d-08*tz06
     $  - 1.081560447074210d-08*tz07
     $  + 1.374105679472354d-09*tz08
     $  - 1.795714439035255d-10*tz09
     $  + 2.401906174616972d-11*tz10
     $  - 3.276093877306557d-12*tz11
     $  + 4.543427585483132d-13*tz12
     $  - 6.392131252078025d-14*tz13
     $  + 9.106304756006494d-15*tz14
     $  - 1.311664890578963d-15*tz15
     $  + 1.907494433270133d-16*tz16

      CsH4YZ(3,0,2,1,03) = 
     $  - 1.357345563151723d-02
     $  + 1.325332220445077d-02*tz01
     $  + 3.507582976534546d-04*tz02
     $  - 3.388365332112763d-05*tz03
     $  + 3.631305253659021d-06*tz04
     $  - 4.173788643455778d-07*tz05
     $  + 5.047605709168359d-08*tz06
     $  - 6.346184435851801d-09*tz07
     $  + 8.228073683987259d-10*tz08
     $  - 1.093827988837288d-10*tz09
     $  + 1.484658051769730d-11*tz10
     $  - 2.050839287983851d-12*tz11
     $  + 2.875899886445340d-13*tz12
     $  - 4.085870192892602d-14*tz13
     $  + 5.871636794798653d-15*tz14
     $  - 8.523589413151205d-16*tz15
     $  + 1.248268023428893d-16*tz16

      CsH4YZ(3,0,2,1,04) = 
     $  + 2.704552436461841d-03
     $  - 3.048338747019639d-03*tz01
     $  - 3.720693343036551d-04*tz02
     $  - 1.842513801584339d-05*tz03
     $  + 2.080073818998195d-06*tz04
     $  - 2.486460580614136d-07*tz05
     $  + 3.101140756773701d-08*tz06
     $  - 3.997482614346752d-09*tz07
     $  + 5.291164026993376d-10*tz08
     $  - 7.157767827126881d-11*tz09
     $  + 9.861462894610948d-12*tz10
     $  - 1.379967284145536d-12*tz11
     $  + 1.957199005725591d-13*tz12
     $  - 2.808645160421712d-14*tz13
     $  + 4.072358536124006d-15*tz14
     $  - 5.959132348612418d-16*tz15

      CsH4YZ(3,0,2,1,05) = 
     $  - 4.132340449830087d-04
     $  + 2.925462402749152d-04*tz01
     $  + 9.965870746040675d-05*tz02
     $  + 1.986957067167765d-05*tz03
     $  + 1.302386975034805d-06*tz04
     $  - 1.611092568548864d-07*tz05
     $  + 2.064747290986095d-08*tz06
     $  - 2.721296003957739d-09*tz07
     $  + 3.669373065803339d-10*tz08
     $  - 5.042630788217608d-11*tz09
     $  + 7.042254868008773d-12*tz10
     $  - 9.971778530049491d-13*tz11
     $  + 1.429079455880325d-13*tz12
     $  - 2.069795570531959d-14*tz13
     $  + 3.025942584448219d-15*tz14
     $  - 4.460900500310808d-16*tz15

      CsH4YZ(3,0,2,1,06) = 
     $  + 2.404873482179306d-05
     $  + 1.612542682128598d-06*tz01
     $  - 1.084404271515537d-05*tz02
     $  - 5.584809775071732d-06*tz03
     $  - 1.443593620395716d-06*tz04
     $  - 1.112739483536861d-07*tz05
     $  + 1.459773210009954d-08*tz06
     $  - 1.961491965111134d-09*tz07
     $  + 2.688327765166790d-10*tz08
     $  - 3.746395271507054d-11*tz09
     $  + 5.295810339552420d-12*tz10
     $  - 7.578976473927535d-13*tz11
     $  + 1.096429744540048d-13*tz12
     $  - 1.601389211986744d-14*tz13
     $  + 2.358866862345614d-15*tz14
     $  - 3.501244983266531d-16*tz15

      CsH4YZ(3,0,2,1,07) = 
     $  + 1.769209100977347d-06
     $  - 2.746752914314007d-06*tz01
     $  - 1.602713613988563d-07*tz02
     $  + 5.912991499191283d-07*tz03
     $  + 4.107228956709798d-07*tz04
     $  + 1.262962704944038d-07*tz05
     $  + 1.079197944688567d-08*tz06
     $  - 1.474872795511250d-09*tz07
     $  + 2.050754075839295d-10*tz08
     $  - 2.893716106512897d-11*tz09
     $  + 4.135267605604393d-12*tz10
     $  - 5.975232773097087d-13*tz11
     $  + 8.718431147720169d-14*tz12
     $  - 1.283164579122158d-14*tz13
     $  + 1.903225444839961d-15*tz14
     $  - 2.842701717354453d-16*tz15

      CsH4YZ(3,0,2,1,08) = 
     $  - 2.533695558373868d-07
     $  + 7.401092411783780d-08*tz01
     $  + 1.537003816614315d-07*tz02
     $  + 2.768768995760403d-08*tz03
     $  - 3.837476299279902d-08*tz04
     $  - 3.571001643370096d-08*tz05
     $  - 1.252011074021333d-08*tz06
     $  - 1.146317188004188d-09*tz07
     $  + 1.614492844836720d-10*tz08
     $  - 2.303716660949728d-11*tz09
     $  + 3.324632397593919d-12*tz10
     $  - 4.845980710831031d-13*tz11
     $  + 7.126096866961458d-14*tz12
     $  - 1.056195780599969d-14*tz13
     $  + 1.576568509535187d-15*tz14
     $  - 2.368462995663025d-16*tz15

      CsH4YZ(3,0,2,1,09) = 
     $  - 1.452293400036869d-08
     $  + 2.961590590716737d-08*tz01
     $  - 6.324535300976272d-09*tz02
     $  - 1.214810629614820d-08*tz03
     $  - 4.230592174559031d-09*tz04
     $  + 2.657791296738349d-09*tz05
     $  + 3.480565399859097d-09*tz06
     $  + 1.357913716965928d-09*tz07
     $  + 1.303998025034517d-10*tz08
     $  - 1.879472815902716d-11*tz09
     $  + 2.736622006167891d-12*tz10
     $  - 4.020690576132845d-13*tz11
     $  + 5.954819222949143d-14*tz12
     $  - 8.883045732169858d-15*tz13
     $  + 1.333754381308579d-15*tz14
     $  - 2.014441136112982d-16*tz15

      CsH4YZ(3,0,2,1,10) = 
     $  + 2.960092306447696d-09
     $  - 1.543335323752688d-09*tz01
     $  - 1.839220372279825d-09*tz02
     $  + 3.484093020755469d-10*tz03
     $  + 1.150194143315039d-09*tz04
     $  + 6.142604324532467d-10*tz05
     $  - 1.750655996245917d-10*tz06
     $  - 3.686246723767522d-10*tz07
     $  - 1.576138715485235d-10*tz08
     $  - 1.564933842350183d-11*tz09
     $  + 2.297135735540164d-12*tz10
     $  - 3.399551547052252d-13*tz11
     $  + 5.067942808608348d-14*tz12
     $  - 7.605093627715894d-15*tz13
     $  + 1.148083491535772d-15*tz14
     $  - 1.742635027760287d-16*tz15

      CsH4YZ(3,0,2,1,11) = 
     $  + 1.421814159394695d-10
     $  - 3.368347374216873d-10*tz01
     $  + 1.566534770229938d-10*tz02
     $  + 1.747485563131408d-10*tz03
     $  + 7.984564145705455d-13*tz04
     $  - 1.199660815068984d-10*tz05
     $  - 8.818777498217554d-11*tz06
     $  + 8.026445170228286d-12*tz07
     $  + 4.158347834147053d-11*tz08
     $  + 1.929024057826229d-11*tz09
     $  + 1.960534083568657d-12*tz10
     $  - 2.920751394063248d-13*tz11
     $  + 4.380457141001997d-14*tz12
     $  - 6.609588318660501d-15*tz13
     $  + 1.002821480165904d-15*tz14
     $  - 1.529186728363901d-16*tz15

      CsH4YZ(3,0,2,1,12) = 
     $  - 3.594365438845826d-11
     $  + 2.467759343891448d-11*tz01
     $  + 2.056779633985995d-11*tz02
     $  - 1.311059522253674d-11*tz03
     $  - 2.100432992532354d-11*tz04
     $  - 5.206682278514854d-12*tz05
     $  + 1.314715014302216d-11*tz06
     $  + 1.271454229892731d-11*tz07
     $  + 4.491477279277065d-13*tz08
     $  - 4.928962332717560d-12*tz09
     $  - 2.463582985259498d-12*tz10
     $  - 2.544264179001318d-13*tz11
     $  + 3.837064786650410d-14*tz12
     $  - 5.819146631804370d-15*tz13
     $  + 8.870198364425224d-16*tz14
     $  - 1.358428054527898d-16*tz15

      CsH4YZ(3,0,2,1,13) = 
     $  - 1.516264404949331d-12
     $  + 3.951677551041798d-12*tz01
     $  - 2.727502914002625d-12*tz02
     $  - 2.191922321990929d-12*tz03
     $  + 9.256541282306551d-13*tz04
     $  + 2.807779540731264d-12*tz05
     $  + 1.338777746106607d-12*tz06
     $  - 1.466696420993946d-12*tz07
     $  - 1.851591549436217d-12*tz08
     $  - 2.332025032552595d-13*tz09
     $  + 6.079681844788205d-13*tz10
     $  + 3.258121256512331d-13*tz11
     $  + 3.400463818160751d-14*tz12
     $  - 5.181369456436925d-15*tz13
     $  + 7.932372448234540d-16*tz14
     $  - 1.219686639517312d-16*tz15

      CsH4YZ(3,0,2,1,14) = 
     $  + 4.460837353685009d-13
     $  - 3.624393447833274d-13*tz01
     $  - 2.202803141239338d-13*tz02
     $  + 2.656564966394132d-13*tz03
     $  + 3.248148775297918d-13*tz04
     $  - 1.935541157778464d-14*tz05
     $  - 3.926204107174687d-13*tz06
     $  - 2.704198312324169d-13*tz07
     $  + 1.618297314691318d-13*tz08
     $  + 2.728609570686257d-13*tz09
     $  + 5.337713342380203d-14*tz10
     $  - 7.747961162223195d-14*tz11
     $  - 4.436488179885884d-14*tz12
     $  - 4.659688961955047d-15*tz13
     $  + 7.162638672273084d-16*tz14
     $  - 1.105470029905576d-16*tz15

      CsH4YZ(3,0,2,1,15) = 
     $  + 1.707079493652499d-14
     $  - 4.744804297150675d-14*tz01
     $  + 4.205532621220471d-14*tz02
     $  + 2.551360446005419d-14*tz03
     $  - 2.515325161067090d-14*tz04
     $  - 5.380114589914471d-14*tz05
     $  - 1.329663415150107d-14*tz06
     $  + 5.587753337778605d-14*tz07
     $  + 5.021475337253908d-14*tz08
     $  - 1.698111296557464d-14*tz09
     $  - 4.069748515200387d-14*tz10
     $  - 1.025789268963158d-14*tz11
     $  + 1.014632139468706d-14*tz12
     $  + 6.192328121284795d-15*tz13
     $  + 6.523581988999464d-16*tz14
     $  - 1.010382658973881d-16*tz15

      CsH4YZ(3,0,2,1,16) = 
     $  - 5.622153110015828d-15
     $  + 5.125675980370743d-15*tz01
     $  + 2.276139444611839d-15*tz02
     $  - 4.472137343739208d-15*tz03
     $  - 4.692333893276378d-15*tz04
     $  + 1.899403336365932d-15*tz05
     $  + 9.147808752778068d-15*tz06
     $  + 4.475596727298151d-15*tz07
     $  - 7.979808115771912d-15*tz08
     $  - 8.978109736076326d-15*tz09
     $  + 1.561408353280778d-15*tz10
     $  + 6.140451284295558d-15*tz11
     $  + 1.839941591321430d-15*tz12
     $  - 1.359558609425607d-15*tz13
     $  - 8.828559958388207d-16*tz14

      CsH4YZ(3,0,2,1,17) = 
     $  - 1.998173429133510d-16
     $  + 5.806942269643845d-16*tz01
     $  - 6.118950944361113d-16*tz02
     $  - 2.840740168149007d-16*tz03
     $  + 4.977031211771211d-16*tz04
     $  + 9.503824680575896d-16*tz05
     $  - 1.556672604802700d-15*tz07
     $  - 1.080785648205655d-15*tz08
     $  + 1.132308416582046d-15*tz09
     $  + 1.575047739485346d-15*tz10
     $  - 9.364833633385459d-16*tz12
     $  - 3.194197126939972d-16*tz13
     $  + 1.857737482106004d-16*tz14
     $  + 1.282118166299730d-16*tz15

      CsH4YZ(3,0,2,1,18) = 
     $  - 1.912681514452426d-16*tz06
     $  + 2.630828741255088d-16*tz08
     $  + 2.326353813867395d-16*tz09
     $  - 1.581075694051902d-16*tz10
     $  - 2.735848622234801d-16*tz11
     $  + 1.442437230594953d-16*tz13

      CsH4YZ(3,1,2,1,01) = 
     $  - 8.312390164085496d-02
     $  + 7.958714627060465d-02*tz01
     $  + 4.607839305444176d-03*tz02
     $  - 1.263829634806922d-03*tz03
     $  + 2.244763010141865d-04*tz04
     $  - 3.682645829702310d-05*tz05
     $  + 5.908773762658756d-06*tz06
     $  - 9.425512945652090d-07*tz07
     $  + 1.503498500260840d-07*tz08
     $  - 2.403558566204450d-08*tz09
     $  + 3.854027766963367d-09*tz10
     $  - 6.199832972794246d-10*tz11
     $  + 1.000537646547268d-10*tz12
     $  - 1.619606556008391d-11*tz13
     $  + 2.629200553514573d-12*tz14
     $  - 4.279438270624319d-13*tz15
     $  + 6.982513845218902d-14*tz16
     $  - 1.141863265606858d-14*tz17
     $  + 1.871215859015513d-15*tz18
     $  - 3.071463543342324d-16*tz19

      CsH4YZ(3,1,2,1,02) = 
     $  - 4.909081861738244d-02
     $  - 1.198907916595697d-02*tz01
     $  + 2.779076160771858d-03*tz02
     $  - 4.808321137530084d-04*tz03
     $  + 7.833251894454272d-05*tz04
     $  - 1.254737565735680d-05*tz05
     $  + 2.001919815705640d-06*tz06
     $  - 3.196246180425962d-07*tz07
     $  + 5.115636222748259d-08*tz08
     $  - 8.212881451474873d-09*tz09
     $  + 1.322779168654704d-09*tz10
     $  - 2.137185362585043d-10*tz11
     $  + 3.463284990769968d-11*tz12
     $  - 5.627810516661778d-12*tz13
     $  + 9.168695996606220d-13*tz14
     $  - 1.497294888163580d-13*tz15
     $  + 2.450526344008236d-14*tz16
     $  - 4.018718907389539d-15*tz17
     $  + 6.602847902074643d-16*tz18
     $  - 1.086430649488813d-16*tz19

      CsH4YZ(3,1,2,1,03) = 
     $  + 7.107838269538700d-03
     $  - 1.107385598602697d-02*tz01
     $  + 2.065013913413750d-03*tz02
     $  - 3.422639018858091d-04*tz03
     $  + 5.511423091648126d-05*tz04
     $  - 8.808210716558764d-06*tz05
     $  + 1.406863692733615d-06*tz06
     $  - 2.251464613167660d-07*tz07
     $  + 3.613520144714737d-08*tz08
     $  - 5.817859792260768d-09*tz09
     $  + 9.396256636782586d-10*tz10
     $  - 1.522102850011035d-10*tz11
     $  + 2.472571541237322d-11*tz12
     $  - 4.027007368345540d-12*tz13
     $  + 6.574450398581998d-13*tz14
     $  - 1.075721007391972d-13*tz15
     $  + 1.763717954333517d-14*tz16
     $  - 2.897192650299450d-15*tz17
     $  + 4.767479763109585d-16*tz18

      CsH4YZ(3,1,2,1,04) = 
     $  - 3.000856183276658d-03
     $  + 3.692738237227398d-03*tz01
     $  - 3.894061189187867d-05*tz02
     $  - 2.606620477227930d-04*tz03
     $  + 4.174035301396449d-05*tz04
     $  - 6.671209688584024d-06*tz05
     $  + 1.067799631733472d-06*tz06
     $  - 1.713734350549108d-07*tz07
     $  + 2.758876188721575d-08*tz08
     $  - 4.455219558508757d-09*tz09
     $  + 7.216096957744252d-10*tz10
     $  - 1.172071860370195d-10*tz11
     $  + 1.908706985939302d-11*tz12
     $  - 3.115821747226533d-12*tz13
     $  + 5.097695654402891d-13*tz14
     $  - 8.357352329401248d-14*tz15
     $  + 1.372742814331795d-14*tz16
     $  - 2.258762375614756d-15*tz17
     $  + 3.722740303082215d-16*tz18

      CsH4YZ(3,1,2,1,05) = 
     $  + 5.071550241113861d-04
     $  - 3.660626444263576d-04*tz01
     $  - 1.717584840502292d-04*tz02
     $  + 3.667784750160301d-05*tz03
     $  + 3.367337208146557d-05*tz04
     $  - 5.391093822693859d-06*tz05
     $  + 8.652448640050770d-07*tz06
     $  - 1.392830916167362d-07*tz07
     $  + 2.249016496371364d-08*tz08
     $  - 3.642339641909363d-09*tz09
     $  + 5.915461011799969d-10*tz10
     $  - 9.632375886033758d-11*tz11
     $  + 1.572282713303405d-11*tz12
     $  - 2.572171124657192d-12*tz13
     $  + 4.216639091717303d-13*tz14
     $  - 6.925672066252999d-14*tz15
     $  + 1.139526266316600d-14*tz16
     $  - 1.877994708381998d-15*tz17
     $  + 3.099752176012219d-16*tz18

      CsH4YZ(3,1,2,1,06) = 
     $  - 2.117021940860573d-05
     $  - 1.328481516861506d-05*tz01
     $  + 2.436554942849153d-05*tz02
     $  + 1.603897179378694d-05*tz03
     $  - 6.401454590185129d-06*tz04
     $  - 4.551884139413588d-06*tz05
     $  + 7.327003354641856d-07*tz06
     $  - 1.182985061634425d-07*tz07
     $  + 1.915672614308037d-08*tz08
     $  - 3.110890715002499d-09*tz09
     $  + 5.065101903719113d-10*tz10
     $  - 8.267002068017388d-11*tz11
     $  + 1.352334098533836d-11*tz12
     $  - 2.216772621589082d-12*tz13
     $  + 3.640749383998071d-13*tz14
     $  - 5.990048050249258d-14*tz15
     $  + 9.871501103077303d-15*tz16
     $  - 1.629277837909742d-15*tz17
     $  + 2.692934130689243d-16*tz18

      CsH4YZ(3,1,2,1,07) = 
     $  - 3.140493994929004d-06
     $  + 4.212429868336118d-06*tz01
     $  + 1.723597921141369d-06*tz02
     $  - 2.811601207477797d-06*tz03
     $  - 2.060893970400291d-06*tz04
     $  + 1.001315080346616d-06*tz05
     $  + 6.400507518833161d-07*tz06
     $  - 1.036362692937721d-07*tz07
     $  + 1.682793196537455d-08*tz08
     $  - 2.739634408174030d-09*tz09
     $  + 4.471104174345534d-10*tz10
     $  - 7.313362596534932d-11*tz11
     $  + 1.198738561383511d-11*tz12
     $  - 1.968646971015667d-12*tz13
     $  + 3.238801322243333d-13*tz14
     $  - 5.337240411309811d-14*tz15
     $  + 8.808739326489588d-15*tz16
     $  - 1.455879854087552d-15*tz17
     $  + 2.409438408689563d-16*tz18

      CsH4YZ(3,1,2,1,08) = 
     $  + 2.125944647394128d-07
     $  + 1.245857191817464d-07*tz01
     $  - 4.628810287858333d-07*tz02
     $  - 3.050434050624159d-07*tz03
     $  + 4.057612836778253d-07*tz04
     $  + 3.057209325925489d-07*tz05
     $  - 1.550857896342940d-07*tz06
     $  - 9.292487123030940d-08*tz07
     $  + 1.512698553516178d-08*tz08
     $  - 2.468520921837946d-09*tz09
     $  + 4.037433146819933d-10*tz10
     $  - 6.617329504127296d-11*tz11
     $  + 1.086675050752064d-11*tz12
     $  - 1.787692751229309d-12*tz13
     $  + 2.945811250537991d-13*tz14
     $  - 4.861651076557616d-14*tz15
     $  + 8.034940183262295d-15*tz16
     $  - 1.329706791933642d-15*tz17
     $  + 2.203286578147567d-16*tz18

      CsH4YZ(3,1,2,1,09) = 
     $  + 3.430609403707442d-08
     $  - 5.402445134601677d-08*tz01
     $  - 2.496794655069997d-08*tz02
     $  + 7.220247370517757d-08*tz03
     $  + 5.683472334848924d-08*tz04
     $  - 6.447593389301756d-08*tz05
     $  - 4.853628225258606d-08*tz06
     $  + 2.426417040230734d-08*tz07
     $  + 1.384521575206775d-08*tz08
     $  - 2.264301883954147d-09*tz09
     $  + 3.710920211908637d-10*tz10
     $  - 6.093571997337809d-11*tz11
     $  + 1.002401466865494d-11*tz12
     $  - 1.651707326974098d-12*tz13
     $  + 2.725799254427691d-13*tz14
     $  - 4.504811674419304d-14*tz15
     $  + 7.454847412300694d-15*tz16
     $  - 1.235203011516943d-15*tz17
     $  + 2.049020300425204d-16*tz18

      CsH4YZ(3,1,2,1,10) = 
     $  - 2.274871805206693d-09
     $  - 1.767832130646553d-09*tz01
     $  + 7.821124193271818d-09*tz02
     $  + 6.201800537833844d-09*tz03
     $  - 1.278433168452600d-08*tz04
     $  - 1.071712601734108d-08*tz05
     $  + 1.073081165937962d-08*tz06
     $  + 7.984807155437244d-09*tz07
     $  - 3.850546170003474d-09*tz08
     $  - 2.107059845156518d-09*tz09
     $  + 3.459715775965458d-10*tz10
     $  - 5.690973577404869d-11*tz11
     $  + 9.376854904768122d-12*tz12
     $  - 1.547389593501476d-12*tz13
     $  + 2.557210002439996d-13*tz14
     $  - 4.231693572171294d-14*tz15
     $  + 7.011392129341016d-15*tz16
     $  - 1.163048989140375d-15*tz17
     $  + 1.931387370438183d-16*tz18

      CsH4YZ(3,1,2,1,11) = 
     $  - 4.231883608693213d-10
     $  + 7.070275301309552d-10*tz01
     $  + 4.496129485052039d-10*tz02
     $  - 1.529499798308795d-09*tz03
     $  - 1.437280828975913d-09*tz04
     $  + 2.374508102278855d-09*tz05
     $  + 2.028171016859045d-09*tz06
     $  - 1.831936238886654d-09*tz07
     $  - 1.341420869347287d-09*tz08
     $  + 6.197150007755126d-10*tz09
     $  + 3.263602685072063d-10*tz10
     $  - 5.377086063269658d-11*tz11
     $  + 8.873011506617062d-12*tz12
     $  - 1.466295609070487d-12*tz13
     $  + 2.426359273219863d-13*tz14
     $  - 4.020059008104398d-14*tz15
     $  + 6.668344643500404d-15*tz16
     $  - 1.107328403809121d-15*tz17
     $  + 1.840705467430826d-16*tz18

      CsH4YZ(3,1,2,1,12) = 
     $  + 2.405277024453321d-11
     $  + 2.960650016794046d-11*tz01
     $  - 1.257457985779814d-10*tz02
     $  - 1.325075616483705d-10*tz03
     $  + 3.232815369683795d-10*tz04
     $  + 3.185117439260069d-10*tz05
     $  - 4.496716158879682d-10*tz06
     $  - 3.841592198188292d-10*tz07
     $  + 3.177149192250251d-10*tz08
     $  + 2.284625145370321d-10*tz09
     $  - 1.010247776497774d-10*tz10
     $  - 5.130353678725885d-11*tz11
     $  + 8.477766615391155d-12*tz12
     $  - 1.402815695495217d-12*tz13
     $  + 2.324156047112303d-13*tz14
     $  - 3.855134718384386d-14*tz15
     $  + 6.401637143413037d-15*tz16
     $  - 1.064111144399183d-15*tz17
     $  + 1.770543138413725d-16*tz18

      CsH4YZ(3,1,2,1,13) = 
     $  + 5.444694036372687d-12
     $  - 9.240041826415703d-12*tz01
     $  - 8.705239715480480d-12*tz02
     $  + 2.962231839622484d-11*tz03
     $  + 3.500293465368614d-11*tz04
     $  - 6.926727204982322d-11*tz05
     $  - 6.872339023949607d-11*tz06
     $  + 8.586090575154389d-11*tz07
     $  + 7.274746821193181d-11*tz08
     $  - 5.569939895954373d-11*tz09
     $  - 3.929123004858454d-11*tz10
     $  + 1.665676038710276d-11*tz11
     $  + 8.167199236995959d-12*tz12
     $  - 1.353084774383299d-12*tz13
     $  + 2.244336809680185d-13*tz14
     $  - 3.726742789328100d-14*tz15
     $  + 6.194689480003387d-15*tz16
     $  - 1.030690013141460d-15*tz17
     $  + 1.716470600542845d-16*tz18

      CsH4YZ(3,1,2,1,14) = 
     $  - 2.435335835664096d-13
     $  - 5.109304001810659d-13*tz01
     $  + 1.956496538491734d-12*tz02
     $  + 2.825980888273252d-12*tz03
     $  - 7.281498812938582d-12*tz04
     $  - 8.678903705364228d-12*tz05
     $  + 1.479699847918139d-11*tz06
     $  + 1.456477281928890d-11*tz07
     $  - 1.645098167243348d-11*tz08
     $  - 1.376637099737634d-11*tz09
     $  + 9.842672834105839d-12*tz10
     $  + 6.807465859964297d-12*tz11
     $  - 2.773797461975286d-12*tz12
     $  - 1.314359825960298d-12*tz13
     $  + 2.182456776952428d-13*tz14
     $  - 3.627661448808984d-14*tz15
     $  + 6.035738662933535d-15*tz16
     $  - 1.005144631763006d-15*tz17
     $  + 1.675346034649719d-16*tz18

      CsH4YZ(3,1,2,1,15) = 
     $  - 7.115416705277736d-14
     $  + 1.196485961708773d-13*tz01
     $  + 1.675136903791148d-13*tz02
     $  - 5.418708688778629d-13*tz03
     $  - 8.180195666604024d-13*tz04
     $  + 1.768132637890609d-12*tz05
     $  + 2.069922705753936d-12*tz06
     $  - 3.138245925163535d-12*tz07
     $  - 3.046169130891961d-12*tz08
     $  + 3.155933533040802d-12*tz09
     $  + 2.602783515314903d-12*tz10
     $  - 1.750064700721375d-12*tz11
     $  - 1.186393796565747d-12*tz12
     $  + 4.659564699015170d-13*tz13
     $  + 2.135284826899959d-13*tz14
     $  - 3.552641850110502d-14*tz15
     $  + 5.916237048237069d-15*tz16
     $  - 9.860798634850962d-16*tz17
     $  + 1.644887487768778d-16*tz18

      CsH4YZ(3,1,2,1,16) = 
     $  + 2.273642271384792d-15
     $  + 8.653211405903745d-15*tz01
     $  - 2.963286952102444d-14*tz02
     $  - 5.852987856819933d-14*tz03
     $  + 1.520046230627866d-13*tz04
     $  + 2.213480423780890d-13*tz05
     $  - 4.206651203113159d-13*tz06
     $  - 4.804350132105467d-13*tz07
     $  + 6.603193376283557d-13*tz08
     $  + 6.304864893182166d-13*tz09
     $  - 6.055432567622763d-13*tz10
     $  - 4.916513343503371d-13*tz11
     $  + 3.127202992609541d-13*tz12
     $  + 2.077671558293745d-13*tz13
     $  - 7.887532071099637d-14*tz14
     $  - 3.497793836268807d-14*tz15
     $  + 5.829851643745283d-15*tz16
     $  - 9.724623973858081d-16*tz17
     $  + 1.623405427116318d-16*tz18

      CsH4YZ(3,1,2,1,17) = 
     $  + 9.336436932430318d-16
     $  - 1.527901237200604d-15*tz01
     $  - 3.121909892720447d-15*tz02
     $  + 9.495149732540291d-15*tz03
     $  + 1.828438417204348d-14*tz04
     $  - 4.124687212205022d-14*tz05
     $  - 5.732764309414879d-14*tz06
     $  + 9.813451946948797d-14*tz07
     $  + 1.092173434423552d-13*tz08
     $  - 1.378855681255614d-13*tz09
     $  - 1.293848587151260d-13*tz10
     $  + 1.161505425086072d-13*tz11
     $  + 9.278765057161234d-14*tz12
     $  - 5.611131603415081d-14*tz13
     $  - 3.653467751578555d-14*tz14
     $  + 1.344214760909739d-14*tz15
     $  + 5.771817019900223d-15*tz16
     $  - 9.635150769902295d-16*tz17
     $  + 1.609629798492217d-16*tz18

      CsH4YZ(3,1,2,1,18) = 
     $  - 1.421801545064914d-16*tz01
     $  + 4.377463992111004d-16*tz02
     $  + 1.168011652737888d-15*tz03
     $  - 2.999104264301510d-15*tz04
     $  - 5.334946858435355d-15*tz05
     $  + 1.081269847854350d-14*tz06
     $  + 1.437544649385846d-14*tz07
     $  - 2.249875515229414d-14*tz08
     $  - 2.441628852103975d-14*tz09
     $  + 2.859285819579442d-14*tz10
     $  + 2.636114660201568d-14*tz11
     $  - 2.226642794651970d-14*tz12
     $  - 1.749698319905844d-14*tz13
     $  + 1.010328618247492d-14*tz14
     $  + 6.447181941959615d-15*tz15
     $  - 2.304581916846756d-15*tz16
     $  - 9.586465193345879d-16*tz17
     $  + 1.602594841912674d-16*tz18

      CsH4YZ(3,1,2,1,19) = 
     $  - 1.605584140932769d-16*tz03
     $  - 3.914964312601947d-16*tz04
     $  + 8.999544635592434d-16*tz05
     $  + 1.485455330793385d-15*tz06
     $  - 2.751044458964849d-15*tz07
     $  - 3.513549163138856d-15*tz08
     $  + 5.080795427723168d-15*tz09
     $  + 5.382917154596370d-15*tz10
     $  - 5.892073173274266d-15*tz11
     $  - 5.337829586960533d-15*tz12
     $  + 4.265680264677548d-15*tz13
     $  + 3.296914968515144d-15*tz14
     $  - 1.824646867363860d-15*tz15
     $  - 1.141245535269862d-15*tz16
     $  + 3.972220256627389d-16*tz17
     $  + 1.601560390450695d-16*tz18

      CsH4YZ(3,1,2,1,20) = 
     $  + 1.224172164732697d-16*tz05
     $  - 2.577869436900672d-16*tz06
     $  - 3.989888327974804d-16*tz07
     $  + 6.824509398754663d-16*tz08
     $  + 8.407652557303037d-16*tz09
     $  - 1.132393802432795d-15*tz10
     $  - 1.172754768457335d-15*tz11
     $  + 1.207351265245379d-15*tz12
     $  + 1.075076916819205d-15*tz13
     $  - 8.166183169025205d-16*tz14
     $  - 6.207991370818088d-16*tz15
     $  + 3.303948576875456d-16*tz16
     $  + 2.025793185788282d-16*tz17

      CsH4YZ(3,1,2,1,21) = 
     $  + 1.040703140211776d-16*tz08
     $  - 1.656846247379211d-16*tz09
     $  - 1.976096312165452d-16*tz10
     $  + 2.495099782774549d-16*tz11
     $  + 2.529022772643081d-16*tz12
     $  - 2.461487147341284d-16*tz13
     $  - 2.155108066819551d-16*tz14
     $  + 1.562296670557104d-16*tz15
     $  + 1.168241921065541d-16*tz16

      CsH4YZ(3,2,0,1,01) = 
     $  + 6.020444085478720d-01
     $  - 6.160375277474026d-01*tz01
     $  + 1.474376828707498d-02*tz02
     $  - 8.070705627247529d-04*tz03
     $  + 6.151714033174060d-05*tz04
     $  - 5.612444433761705d-06*tz05
     $  + 5.735552966205363d-07*tz06
     $  - 6.339053324105759d-08*tz07
     $  + 7.420497143754529d-09*tz08
     $  - 9.077326980244161d-10*tz09
     $  + 1.149759746323152d-10*tz10
     $  - 1.498057909320297d-11*tz11
     $  + 1.998079340982270d-12*tz12
     $  - 2.718027017262366d-13*tz13
     $  + 3.760111977634564d-14*tz14
     $  - 5.277866024032510d-15*tz15
     $  + 7.502846848249696d-16*tz16
     $  - 1.078363764301760d-16*tz17

      CsH4YZ(3,2,0,1,02) = 
     $  + 4.861682705512370d-01
     $  - 2.485232872002744d-02*tz01
     $  + 1.424407958684748d-03*tz02
     $  - 1.121396279545238d-04*tz03
     $  + 1.047714131142832d-05*tz04
     $  - 1.090365165943708d-06*tz05
     $  + 1.222556711016272d-07*tz06
     $  - 1.447912119040051d-08*tz07
     $  + 1.788371140732048d-09*tz08
     $  - 2.283668067467777d-10*tz09
     $  + 2.996151976856058d-11*tz10
     $  - 4.020194793017882d-12*tz11
     $  + 5.497384255189893d-13*tz12
     $  - 7.640112783555872d-14*tz13
     $  + 1.076784757790120d-14*tz14
     $  - 1.536311700321714d-15*tz15
     $  + 2.215339035604322d-16*tz16

      CsH4YZ(3,2,0,1,03) = 
     $  + 4.415909315189595d-02
     $  + 2.557278199040926d-02*tz01
     $  + 6.299800607204542d-04*tz02
     $  - 5.422113687279137d-05*tz03
     $  + 5.392929955958635d-06*tz04
     $  - 5.879332863853465d-07*tz05
     $  + 6.833361136009506d-08*tz06
     $  - 8.328514770885716d-09*tz07
     $  + 1.053110527855259d-09*tz08
     $  - 1.371349952866436d-10*tz09
     $  + 1.829283795104014d-11*tz10
     $  - 2.489710734369538d-12*tz11
     $  + 3.446902425268296d-13*tz12
     $  - 4.842645651750558d-14*tz13
     $  + 6.890915900613910d-15*tz14
     $  - 9.916034175974418d-16*tz15
     $  + 1.440868098088382d-16*tz16

      CsH4YZ(3,2,0,1,04) = 
     $  + 1.096321852120669d-02
     $  - 3.816918532174081d-03*tz01
     $  - 5.292175223563768d-04*tz02
     $  - 2.795146292931814d-05*tz03
     $  + 2.969086469797591d-06*tz04
     $  - 3.396183284379876d-07*tz05
     $  + 4.095548653760603d-08*tz06
     $  - 5.140191985813856d-09*tz07
     $  + 6.657082066132614d-10*tz08
     $  - 8.843558725083255d-11*tz09
     $  + 1.199799139412594d-11*tz10
     $  - 1.656884099688336d-12*tz11
     $  + 2.323074809686865d-13*tz12
     $  - 3.300181882107334d-14*tz13
     $  + 4.742428907482930d-15*tz14
     $  - 6.884457819389681d-16*tz15
     $  + 1.008260616907509d-16*tz16

      CsH4YZ(3,2,0,1,05) = 
     $  - 4.999741533958800d-06
     $  + 1.939993480629919d-04*tz01
     $  + 1.137567841286993d-04*tz02
     $  + 2.522855890632894d-05*tz03
     $  + 1.795387433012060d-06*tz04
     $  - 2.141305383614454d-07*tz05
     $  + 2.667243132809725d-08*tz06
     $  - 3.435658671109291d-09*tz07
     $  + 4.545674146089814d-10*tz08
     $  - 6.148026949448219d-11*tz09
     $  + 8.469664648120266d-12*tz10
     $  - 1.185213040934353d-12*tz11
     $  + 1.681082164014584d-13*tz12
     $  - 2.412640106793718d-14*tz13
     $  + 3.498592172238048d-15*tz14
     $  - 5.120215191111785d-16*tz15

      CsH4YZ(3,2,0,1,06) = 
     $  + 5.788943884596748d-06
     $  + 1.102501744265173d-05*tz01
     $  - 8.887241715543602d-06*tz02
     $  - 6.071542934801842d-06*tz03
     $  - 1.727054462405425d-06*tz04
     $  - 1.444730393395333d-07*tz05
     $  + 1.849880954906428d-08*tz06
     $  - 2.436864550571695d-09*tz07
     $  + 3.284949148198792d-10*tz08
     $  - 4.513773201850358d-11*tz09
     $  + 6.303500282228638d-12*tz10
     $  - 8.926010989599293d-13*tz11
     $  + 1.279304289628634d-13*tz12
     $  - 1.853059426823017d-14*tz13
     $  + 2.709408684078545d-15*tz14
     $  - 3.994780885265753d-16*tz15

      CsH4YZ(3,2,0,1,07) = 
     $  + 1.176204667749464d-07
     $  - 1.601910125976959d-06*tz01
     $  - 4.880510997893664d-07*tz02
     $  + 4.979748025112563d-07*tz03
     $  + 4.339565589446688d-07*tz04
     $  + 1.457202106337089d-07*tz05
     $  + 1.344838287742467d-08*tz06
     $  - 1.806448758791405d-09*tz07
     $  + 2.475429019340848d-10*tz08
     $  - 3.449519292752125d-11*tz09
     $  + 4.876238967982473d-12*tz10
     $  - 6.978962722182206d-13*tz11
     $  + 1.009722609026937d-13*tz12
     $  - 1.474917156596526d-14*tz13
     $  + 2.172845905829191d-15*tz14
     $  - 3.225565641950913d-16*tz15

      CsH4YZ(3,2,0,1,08) = 
     $  - 3.930867609306166d-08
     $  - 5.075195614045127d-08*tz01
     $  + 1.301684348708931d-07*tz02
     $  + 4.387226909094388d-08*tz03
     $  - 3.163165944843693d-08*tz04
     $  - 3.702998569249233d-08*tz05
     $  - 1.410097311355096d-08*tz06
     $  - 1.386411913078248d-09*tz07
     $  + 1.927612365210961d-10*tz08
     $  - 2.719973149910577d-11*tz09
     $  + 3.887201483313532d-12*tz10
     $  - 5.617297141368830d-13*tz11
     $  + 8.197082521604508d-14*tz12
     $  - 1.206583870712752d-14*tz13
     $  + 1.789876203552690d-15*tz14
     $  - 2.673766267934322d-16*tz15

      CsH4YZ(3,2,0,1,09) = 
     $  - 1.348248061139289d-09
     $  + 1.552688826743726d-08*tz01
     $  - 3.037186695612792d-10*tz02
     $  - 1.092226671575137d-08*tz03
     $  - 5.276487324676499d-09*tz04
     $  + 2.050396620068861d-09*tz05
     $  + 3.561652105205794d-09*tz06
     $  + 1.503351830470251d-09*tz07
     $  + 1.541657480833098d-10*tz08
     $  - 2.199896192860240d-11*tz09
     $  + 3.175074056483031d-12*tz10
     $  - 4.628487574405045d-13*tz11
     $  + 6.807108174757947d-14*tz12
     $  - 1.009049846962041d-14*tz13
     $  + 1.506399957134637d-15*tz14
     $  - 2.263362340969565d-16*tz15

      CsH4YZ(3,2,0,1,10) = 
     $  + 3.763138765416784d-10
     $  + 1.533675180640697d-10*tz01
     $  - 1.579764050807854d-09*tz02
     $  - 9.096200018788290d-12*tz03
     $  + 1.038549912115370d-09*tz04
     $  + 6.948533758815829d-10*tz05
     $  - 1.126059371522752d-10*tz06
     $  - 3.734730013581444d-10*tz07
     $  - 1.722796320216397d-10*tz08
     $  - 1.817357844989841d-11*tz09
     $  + 2.646458116803017d-12*tz10
     $  - 3.888692233247669d-13*tz11
     $  + 5.760078139593158d-14*tz12
     $  - 8.593710499195598d-15*tz13
     $  + 1.290495602401089d-15*tz14
     $  - 1.949366168691591d-16*tz15

      CsH4YZ(3,2,0,1,11) = 
     $  + 1.456926684609676d-11
     $  - 1.603903107742355d-10*tz01
     $  + 5.662582648266704d-11*tz02
     $  + 1.623988753084732d-10*tz03
     $  + 2.816215149141149d-11*tz04
     $  - 1.071184497677535d-10*tz05
     $  - 9.516102170594192d-11*tz06
     $  + 1.009796435932024d-12*tz07
     $  + 4.180121734363753d-11*tz08
     $  + 2.087828495863177d-11*tz09
     $  + 2.244147785528732d-12*tz10
     $  - 3.321565171408207d-13*tz11
     $  + 4.952354144492364d-14*tz12
     $  - 7.432663788033826d-15*tz13
     $  + 1.122208274167715d-15*tz14
     $  - 1.703593537309492d-16*tz15

      CsH4YZ(3,2,0,1,12) = 
     $  - 4.131645021708138d-12
     $  + 1.409186192376155d-12*tz01
     $  + 1.806068801908923d-11*tz02
     $  - 6.433867088983911d-12*tz03
     $  - 1.953767047910348d-11*tz04
     $  - 7.713261405965985d-12*tz05
     $  + 1.151061281228567d-11*tz06
     $  + 1.336546641199855d-11*tz07
     $  + 1.289409201611381d-12*tz08
     $  - 4.923037724765624d-12*tz09
     $  - 2.645602474819747d-12*tz10
     $  - 2.877895969697436d-13*tz11
     $  + 4.316778717311643d-14*tz12
     $  - 6.514406617587762d-15*tz13
     $  + 9.885160656763700d-16*tz14
     $  - 1.507576467797129d-16*tz15

      CsH4YZ(3,2,0,1,13) = 
     $  - 1.617740127020654d-13
     $  + 1.720901405564770d-12*tz01
     $  - 1.161509394718712d-12*tz02
     $  - 2.122379204375798d-12*tz03
     $  + 3.485664992790680d-13*tz04
     $  + 2.583599351504053d-12*tz05
     $  + 1.597323966177953d-12*tz06
     $  - 1.246259888598442d-12*tz07
     $  - 1.915253236064916d-12*tz08
     $  - 3.388621963280355d-13*tz09
     $  + 6.039400462299580d-13*tz10
     $  + 3.476664518397342d-13*tz11
     $  + 3.808223008469753d-14*tz12
     $  - 5.776200820802987d-15*tz13
     $  + 8.805928977244675d-16*tz14
     $  - 1.348764523608308d-16*tz15

      CsH4YZ(3,2,0,1,14) = 
     $  + 4.818284688102868d-14
     $  - 4.278256494096832d-14*tz01
     $  - 2.015822452409525d-13*tz02
     $  + 1.524100334645889d-13*tz03
     $  + 3.097047105631454d-13*tz04
     $  + 4.117528531368368d-14*tz05
     $  - 3.570810984915174d-13*tz06
     $  - 2.992585414880494d-13*tz07
     $  + 1.310000039409583d-13*tz08
     $  + 2.792232923192992d-13*tz09
     $  + 6.718934304539601d-14*tz10
     $  - 7.660126800293868d-14*tz11
     $  - 4.709211774201812d-14*tz12
     $  - 5.174404517965336d-15*tz13
     $  + 7.922764013790592d-16*tz14
     $  - 1.218367409312238d-16*tz15

      CsH4YZ(3,2,0,1,15) = 
     $  + 1.877573830401791d-15
     $  - 1.902833016551669d-14*tz01
     $  + 1.848097268919137d-14*tz02
     $  + 2.626511296231520d-14*tz03
     $  - 1.436706476511462d-14*tz04
     $  - 5.029588205945573d-14*tz05
     $  - 2.047787891073953d-14*tz06
     $  + 5.019634283702332d-14*tz07
     $  + 5.360338265213784d-14*tz08
     $  - 1.254486237680399d-14*tz09
     $  - 4.132988924500911d-14*tz10
     $  - 1.212171984114598d-14*tz11
     $  + 9.988642946359591d-15*tz12
     $  + 6.543937876836821d-15*tz13
     $  + 7.191531285420730d-16*tz14
     $  - 1.110069831948548d-16*tz15

      CsH4YZ(3,2,0,1,16) = 
     $  - 5.800885552241476d-16
     $  + 7.371899910071333d-16*tz01
     $  + 2.237689589069920d-15*tz02
     $  - 2.667910539722146d-15*tz03
     $  - 4.594398480326233d-15*tz04
     $  + 6.258859684267726d-16*tz05
     $  + 8.445662685242727d-15*tz06
     $  + 5.395111490485594d-15*tz07
     $  - 7.065034978541616d-15*tz08
     $  - 9.390206889051917d-15*tz09
     $  + 9.082554092232018d-16*tz10
     $  + 6.200527241813642d-15*tz11
     $  + 2.098209884750944d-15*tz12
     $  - 1.333205114193992d-15*tz13
     $  - 9.294682801835176d-16*tz14
     $  - 1.019234843344899d-16*tz15

      CsH4YZ(3,2,0,1,17) = 
     $  + 2.160653537501075d-16*tz01
     $  - 2.665033258602623d-16*tz02
     $  - 3.198835109215317d-16*tz03
     $  + 3.109405043048653d-16*tz04
     $  + 8.983013051837873d-16*tz05
     $  + 1.903611670595294d-16*tz06
     $  - 1.424878122123364d-15*tz07
     $  - 1.204353069845232d-15*tz08
     $  + 9.838107298646948d-16*tz09
     $  + 1.626244641230200d-15*tz10
     $  - 9.414896851023055d-16*tz12
     $  - 3.560237463850839d-16*tz13
     $  + 1.815056596152786d-16*tz14
     $  + 1.345418157509826d-16*tz15

      CsH4YZ(3,2,0,1,18) = 
     $  - 1.779537852151604d-16*tz06
     $  + 2.390085907340287d-16*tz08
     $  + 2.497718093245762d-16*tz09
     $  - 1.337989645675762d-16*tz10
     $  - 2.800123148007301d-16*tz11
     $  + 1.445043621747622d-16*tz13

      CsH4YZ(3,2,1,1,01) = 
     $  - 3.825566833648107d-01
     $  + 3.766483820689516d-01*tz01
     $  + 7.320897840183295d-03*tz02
     $  - 1.615755740190618d-03*tz03
     $  + 2.312296507410159d-04*tz04
     $  - 3.197994587886251d-05*tz05
     $  + 4.462703123009096d-06*tz06
     $  - 6.327606903009151d-07*tz07
     $  + 9.115415963230806d-08*tz08
     $  - 1.332040570736401d-08*tz09
     $  + 1.971012560476424d-09*tz10
     $  - 2.948413724669833d-10*tz11
     $  + 4.452561538415427d-11*tz12
     $  - 6.780224578854703d-12*tz13
     $  + 1.040070118853092d-12*tz14
     $  - 1.605849929464754d-13*tz15
     $  + 2.493826031876405d-14*tz16
     $  - 3.892998866809264d-15*tz17
     $  + 6.105783916302248d-16*tz18

      CsH4YZ(3,2,1,1,02) = 
     $  - 2.670318169364078d-01
     $  - 1.710134107996818d-02*tz01
     $  + 3.293742929324543d-03*tz02
     $  - 4.674915389558473d-04*tz03
     $  + 6.487375887307868d-05*tz04
     $  - 9.100454901834060d-06*tz05
     $  + 1.297069748883168d-06*tz06
     $  - 1.877392804459798d-07*tz07
     $  + 2.754984746013060d-08*tz08
     $  - 4.091653922803412d-09*tz09
     $  + 6.140689464879458d-10*tz10
     $  - 9.300281094962462d-11*tz11
     $  + 1.419873407553530d-11*tz12
     $  - 2.183080343861999d-12*tz13
     $  + 3.377634413850125d-13*tz14
     $  - 5.255178868337837d-14*tz15
     $  + 8.217581972860454d-15*tz16
     $  - 1.290845299629729d-15*tz17
     $  + 2.035561837258841d-16*tz18

      CsH4YZ(3,2,1,1,03) = 
     $  - 4.135229418333422d-02
     $  - 1.321052149871184d-02*tz01
     $  + 2.085170248171630d-03*tz02
     $  - 2.926415463492366d-04*tz03
     $  + 4.100432501001933d-05*tz04
     $  - 5.825740444840474d-06*tz05
     $  + 8.405453879031585d-07*tz06
     $  - 1.230034998606951d-07*tz07
     $  + 1.822543157301137d-08*tz08
     $  - 2.729868067811719d-09*tz09
     $  + 4.127645484562124d-10*tz10
     $  - 6.292863236727417d-11*tz11
     $  + 9.663852827151001d-12*tz12
     $  - 1.493647325763577d-12*tz13
     $  + 2.321867016135949d-13*tz14
     $  - 3.627938292870445d-14*tz15
     $  + 5.694983346531995d-15*tz16
     $  - 8.977379049797309d-16*tz17
     $  + 1.420218618947231d-16*tz18

      CsH4YZ(3,2,1,1,04) = 
     $  - 9.325749013182718d-03
     $  + 3.519753452185100d-03*tz01
     $  - 2.471587938941681d-04*tz02
     $  - 1.968546374158772d-04*tz03
     $  + 2.793414540652221d-05*tz04
     $  - 4.023208853545191d-06*tz05
     $  + 5.877841283483982d-07*tz06
     $  - 8.697207301717000d-08*tz07
     $  + 1.301222366013488d-08*tz08
     $  - 1.965661387947622d-09*tz09
     $  + 2.994486261942738d-10*tz10
     $  - 4.595670163192814d-11*tz11
     $  + 7.099315037241179d-12*tz12
     $  - 1.103092665655940d-12*tz13
     $  + 1.722939901790241d-13*tz14
     $  - 2.703738122250075d-14*tz15
     $  + 4.260873992265517d-15*tz16
     $  - 6.740762047391475d-16*tz17
     $  + 1.069884264252334d-16*tz18

      CsH4YZ(3,2,1,1,05) = 
     $  - 1.773513840818879d-05
     $  - 3.467316218822529d-04*tz01
     $  - 1.083581377324155d-04*tz02
     $  + 4.295755965754242d-05*tz03
     $  + 2.048401551852664d-05*tz04
     $  - 2.988220038517671d-06*tz05
     $  + 4.415902754319246d-07*tz06
     $  - 6.599847596689368d-08*tz07
     $  + 9.961331104715765d-09*tz08
     $  - 1.516438705368173d-09*tz09
     $  + 2.325943031130688d-10*tz10
     $  - 3.591345360184911d-11*tz11
     $  + 5.577986171431406d-12*tz12
     $  - 8.709392683410244d-13*tz13
     $  + 1.366334104591151d-13*tz14
     $  - 2.152713338904562d-14*tz15
     $  + 3.404881864458102d-15*tz16
     $  - 5.404562097493795d-16*tz17

      CsH4YZ(3,2,1,1,06) = 
     $  - 2.258823744353018d-05
     $  - 3.936084319722108d-06*tz01
     $  + 2.295378105214773d-05*tz02
     $  + 7.167924810056312d-06*tz03
     $  - 5.408968075250765d-06*tz04
     $  - 2.336833888359362d-06*tz05
     $  + 3.488719358117405d-07*tz06
     $  - 5.260896395880691d-08*tz07
     $  + 8.002914481186008d-09*tz08
     $  - 1.226760428767483d-09*tz09
     $  + 1.893215636329461d-10*tz10
     $  - 2.939262419343120d-11*tz11
     $  + 4.587705845356830d-12*tz12
     $  - 7.195064482578969d-13*tz13
     $  + 1.133321214725072d-13*tz14
     $  - 1.792156279992433d-14*tz15
     $  + 2.844146691327121d-15*tz16
     $  - 4.528420679946311d-16*tz17

      CsH4YZ(3,2,1,1,07) = 
     $  + 3.649665560068978d-08
     $  + 3.500939641749857d-06*tz01
     $  + 6.292688174018705d-09*tz02
     $  - 2.208909191699579d-06*tz03
     $  - 7.272540790086131d-07*tz04
     $  + 6.892427575192162d-07*tz05
     $  + 2.858237248326124d-07*tz06
     $  - 4.344677554020273d-08*tz07
     $  + 6.655775870208746d-09*tz08
     $  - 1.026633003683276d-09*tz09
     $  + 1.593187387575503d-10*tz10
     $  - 2.485805757736978d-11*tz11
     $  + 3.897386596491030d-12*tz12
     $  - 6.137317081578098d-13*tz13
     $  + 9.702961270537436d-14*tz14
     $  - 1.539559488311632d-14*tz15
     $  + 2.450873655189946d-15*tz16
     $  - 3.913428698146874d-16*tz17

      CsH4YZ(3,2,1,1,08) = 
     $  + 2.277269996321564d-07
     $  - 5.805389811758599d-08*tz01
     $  - 3.448718265372289d-07*tz02
     $  - 1.782014491944286d-08*tz03
     $  + 2.665842122585049d-07*tz04
     $  + 9.256446932730403d-08*tz05
     $  - 9.139506658766354d-08*tz06
     $  - 3.686514884211234d-08*tz07
     $  + 5.683336416050846d-09*tz08
     $  - 8.815851881359456d-10*tz09
     $  + 1.375006118135272d-10*tz10
     $  - 2.155141069096490d-11*tz11
     $  + 3.392857868002901d-12*tz12
     $  - 5.362814387118862d-13*tz13
     $  + 8.507461030501263d-14*tz14
     $  - 1.354099719929644d-14*tz15
     $  + 2.161849357125890d-15*tz16
     $  - 3.461119291857512d-16*tz17

      CsH4YZ(3,2,1,1,09) = 
     $  + 7.400476450909161d-10
     $  - 3.918747684900537d-08*tz01
     $  + 1.021209437464967d-08*tz02
     $  + 4.395585143072898d-08*tz03
     $  + 6.025888489228959d-09*tz04
     $  - 3.649983555553842d-08*tz05
     $  - 1.321078659662685d-08*tz06
     $  + 1.259179984120113d-08*tz07
     $  + 4.955835991867702d-09*tz08
     $  - 7.726699688342457d-10*tz09
     $  + 1.210674455801909d-10*tz10
     $  - 1.905461997027266d-11*tz11
     $  + 3.011114276112574d-12*tz12
     $  - 4.775822851639296d-13*tz13
     $  + 7.600187048290855d-14*tz14
     $  - 1.213204490618964d-14*tz15
     $  + 1.942096650890531d-15*tz16
     $  - 3.117007856633636d-16*tz17

      CsH4YZ(3,2,1,1,10) = 
     $  - 2.569260083872851d-09
     $  + 1.325676817580274d-09*tz01
     $  + 4.668455430897983d-09*tz02
     $  - 8.560351676048359d-10*tz03
     $  - 6.650483013175796d-09*tz04
     $  - 1.395053841900083d-09*tz05
     $  + 5.374041549635668d-09*tz06
     $  + 2.002475436272648d-09*tz07
     $  - 1.792489268454994d-09*tz08
     $  - 6.886280244247503d-10*tz09
     $  + 1.083524000528409d-10*tz10
     $  - 1.711842291231429d-11*tz11
     $  + 2.714546286738386d-12*tz12
     $  - 4.319143830822656d-13*tz13
     $  + 6.893529973914658d-14*tz14
     $  - 1.103369443803639d-14*tz15
     $  + 1.770680130849982d-15*tz16
     $  - 2.848470757713551d-16*tz17

      CsH4YZ(3,2,1,1,11) = 
     $  - 2.009887417303375d-11
     $  + 4.562656888712953d-10*tz01
     $  - 2.126771049648252d-10*tz02
     $  - 7.393555052212541d-10*tz03
     $  + 2.561755959460956d-11*tz04
     $  + 1.095359624964077d-09*tz05
     $  + 2.865283417264071d-10*tz06
     $  - 8.275038641913523d-10*tz07
     $  - 3.142472066022072d-10*tz08
     $  + 2.622447874163010d-10*tz09
     $  + 9.830423418994452d-11*tz10
     $  - 1.558536478941047d-11*tz11
     $  + 2.479366437830144d-12*tz12
     $  - 3.956563312821100d-13*tz13
     $  + 6.331976901789251d-14*tz14
     $  - 1.016032098006448d-14*tz15
     $  + 1.634318726267314d-15*tz16
     $  - 2.634803490234563d-16*tz17

      CsH4YZ(3,2,1,1,12) = 
     $  + 3.022609628579784d-11
     $  - 2.050706563682559d-11*tz01
     $  - 6.143307844884333d-11*tz02
     $  + 2.308947168662903d-11*tz03
     $  + 1.365772751186801d-10*tz04
     $  + 1.190997310184887d-11*tz05
     $  - 1.882780263031996d-10*tz06
     $  - 5.597793240066373d-11*tz07
     $  + 1.312699977477606d-10*tz08
     $  + 5.041775951164280d-11*tz09
     $  - 3.925584031615176d-11*tz10
     $  - 1.435135267477984d-11*tz11
     $  + 2.289827718154262d-12*tz12
     $  - 3.664082095129598d-13*tz13
     $  + 5.878707113365213d-14*tz14
     $  - 9.455091357210167d-15*tz15
     $  + 1.524192216783775d-15*tz16
     $  - 2.462249592401045d-16*tz17

      CsH4YZ(3,2,1,1,13) = 
     $  + 3.629985303626816d-13
     $  - 5.435574294865730d-12*tz01
     $  + 3.328735908096496d-12*tz02
     $  + 1.172630399635275d-11*tz03
     $  - 1.950493004735496d-12*tz04
     $  - 2.660610477730021d-11*tz05
     $  - 4.505168468257568d-12*tz06
     $  + 3.312244702550485d-11*tz07
     $  + 1.067006933291948d-11*tz08
     $  - 2.126847565273054d-11*tz09
     $  - 8.215009572086451d-12*tz10
     $  + 5.990891640660616d-12*tz11
     $  + 2.135113768161126d-12*tz12
     $  - 3.425196971174161d-13*tz13
     $  + 5.508377597094817d-14*tz14
     $  - 8.878859716613992d-15*tz15
     $  + 1.434221789983930d-15*tz16
     $  - 2.321323776300728d-16*tz17

      CsH4YZ(3,2,1,1,14) = 
     $  - 3.637593388491064d-13
     $  + 2.833983746405266d-13*tz01
     $  + 8.042079878866813d-13*tz02
     $  - 4.104787915475925d-13*tz03
     $  - 2.578427964768044d-12*tz04
     $  + 1.489685420939958d-14*tz05
     $  + 5.267904551460166d-12*tz06
     $  + 1.181599280921326d-12*tz07
     $  - 5.907415112349873d-12*tz08
     $  - 2.006386098176671d-12*tz09
     $  + 3.500930968426946d-12*tz10
     $  + 1.354179675286889d-12*tz11
     $  - 9.294381234605727d-13*tz12
     $  - 3.228208360076918d-13*tz13
     $  + 5.203005368887414d-14*tz14
     $  - 8.403828091682247d-15*tz15
     $  + 1.360088221855580d-15*tz16
     $  - 2.205283176865395d-16*tz17

      CsH4YZ(3,2,1,1,15) = 
     $  - 5.753408647611590d-15
     $  + 6.582374358019510d-14*tz01
     $  - 4.621874768465590d-14*tz02
     $  - 1.818555026642563d-13*tz03
     $  + 4.507541766749839d-14*tz04
     $  + 5.811000492363148d-13*tz05
     $  + 5.598187699603387d-14*tz06
     $  - 1.046900131765341d-12*tz07
     $  - 2.746789852916530d-13*tz08
     $  + 1.062807121748100d-12*tz09
     $  + 3.742191477421296d-13*tz10
     $  - 5.834575139726687d-13*tz11
     $  - 2.252923658468341d-13*tz12
     $  + 1.462466917649275d-13*tz13
     $  + 4.949500526433305d-14*tz14
     $  - 8.009747227143058d-15*tz15
     $  + 1.298643342039983d-15*tz16
     $  - 2.109210802068167d-16*tz17

      CsH4YZ(3,2,1,1,16) = 
     $  + 4.446754169287100d-15
     $  - 3.733347849231713d-15*tz01
     $  - 1.056503302867072d-14*tz02
     $  + 6.113578808854599d-15*tz03
     $  + 4.658281251222203d-14*tz04
     $  - 3.040359801197063d-15*tz05
     $  - 1.301850991198710d-13*tz06
     $  - 2.123893541776661d-14*tz07
     $  + 2.078922387234017d-13*tz08
     $  + 6.022477620442564d-14*tz09
     $  - 1.923399174548335d-13*tz10
     $  - 6.943704329444616d-14*tz11
     $  + 9.821641010027822d-14*tz12
     $  + 3.776766718103388d-14*tz13
     $  - 2.329545778309797d-14*tz14
     $  - 7.681545422799430d-15*tz15
     $  + 1.247542615778424d-15*tz16
     $  - 2.029443216638911d-16*tz17

      CsH4YZ(3,2,1,1,17) = 
     $  - 8.070246274866193d-16*tz01
     $  + 5.980793655449133d-16*tz02
     $  + 2.795349343170265d-15*tz03
     $  - 7.780753024361770d-16*tz04
     $  - 1.190824703252276d-14*tz05
     $  - 5.324329737175213d-16*tz06
     $  + 2.881410541497418d-14*tz07
     $  + 6.034259116088874d-15*tz08
     $  - 4.118351934703209d-14*tz09
     $  - 1.276126391752959d-14*tz10
     $  + 3.495499656041954d-14*tz11
     $  + 1.284013929901906d-14*tz12
     $  - 1.667122937428079d-14*tz13
     $  - 6.372351192110868d-15*tz14
     $  + 3.750644260170166d-15*tz15
     $  + 1.205007723141265d-15*tz16
     $  - 1.963200630436022d-16*tz17

      CsH4YZ(3,2,1,1,18) = 
     $  + 1.396235968343355d-16*tz02
     $  - 8.187804550444313d-16*tz04
     $  + 2.971289083924520d-15*tz06
     $  + 3.441759327598649d-16*tz07
     $  - 6.298351708242341d-15*tz08
     $  - 1.530885977579483d-15*tz09
     $  + 8.134371674116203d-15*tz10
     $  + 2.644284767630979d-15*tz11
     $  - 6.372522898088468d-15*tz12
     $  - 2.368849699191290d-15*tz13
     $  + 2.849751340730361d-15*tz14
     $  + 1.081217412114933d-15*tz15
     $  - 6.095849350337699d-16*tz16
     $  - 1.908340615888175d-16*tz17

      CsH4YZ(3,2,1,1,19) = 
     $  + 2.335788193320361d-16*tz05
     $  - 7.233228330040544d-16*tz07
     $  - 1.197488024312504d-16*tz08
     $  + 1.360986489177899d-15*tz09
     $  + 3.653146877068425d-16*tz10
     $  - 1.601773571465442d-15*tz11
     $  - 5.393445482918494d-16*tz12
     $  + 1.164584476535860d-15*tz13
     $  + 4.363233487513489d-16*tz14
     $  - 4.900888576453073d-16*tz15
     $  - 1.843610306391172d-16*tz16

      CsH4YZ(3,2,1,1,20) = 
     $  + 1.722905833812942d-16*tz08
     $  - 2.910802263335085d-16*tz10
     $  + 3.144861343388155d-16*tz12
     $  + 1.087210797709080d-16*tz13
     $  - 2.132444231605358d-16*tz14

      CsH4YZ(3,2,2,1,01) = 
     $  + 3.646754874236878d-01
     $  - 3.743873824713214d-01*tz01
     $  + 1.027532609755988d-02*tz02
     $  - 6.079636698192282d-04*tz03
     $  + 4.869786235154945d-05*tz04
     $  - 4.598715966764390d-06*tz05
     $  + 4.820528948418986d-07*tz06
     $  - 5.432898214974363d-08*tz07
     $  + 6.459236757095767d-09*tz08
     $  - 8.001936334914119d-10*tz09
     $  + 1.024244587296915d-10*tz10
     $  - 1.346402847266131d-11*tz11
     $  + 1.809482877551658d-12*tz12
     $  - 2.477701739782166d-13*tz13
     $  + 3.447399679443501d-14*tz14
     $  - 4.863519885412235d-15*tz15
     $  + 6.945047806318999d-16*tz16
     $  - 1.002223223502946d-16*tz17

      CsH4YZ(3,2,2,1,02) = 
     $  + 2.977372825491541d-01
     $  - 1.748779623384008d-02*tz01
     $  + 1.082101076354037d-03*tz02
     $  - 8.939506520781682d-05*tz03
     $  + 8.634558790438094d-06*tz04
     $  - 9.208459003835820d-07*tz05
     $  + 1.052069604756293d-07*tz06
     $  - 1.264736283057130d-08*tz07
     $  + 1.581238410102201d-09*tz08
     $  - 2.039691253251516d-10*tz09
     $  + 2.699029075736515d-11*tz10
     $  - 3.648144335162220d-12*tz11
     $  + 5.020405920336513d-13*tz12
     $  - 7.016122017816948d-14*tz13
     $  + 9.937076166953850d-15*tz14
     $  - 1.423988295450339d-15*tz15
     $  + 2.061416734437228d-16*tz16

      CsH4YZ(3,2,2,1,03) = 
     $  + 9.439438126018936d-02
     $  - 6.954135402569981d-03*tz01
     $  + 4.903691791869544d-04*tz02
     $  - 4.405239509695950d-05*tz03
     $  + 4.512117604808828d-06*tz04
     $  - 5.026560753654622d-07*tz05
     $  + 5.940483422588933d-08*tz06
     $  - 7.337349603368758d-09*tz07
     $  + 9.379589481269924d-10*tz08
     $  - 1.232588970851862d-10*tz09
     $  + 1.656970923436872d-11*tz10
     $  - 2.270275011176695d-12*tz11
     $  + 3.161401155856818d-13*tz12
     $  - 4.464257922321580d-14*tz13
     $  + 6.381276347706281d-15*tz14
     $  - 9.219811794562812d-16*tz15
     $  + 1.344568060758638d-16*tz16

      CsH4YZ(3,2,2,1,04) = 
     $  + 8.893822002271765d-03
     $  + 3.708484641653139d-03*tz01
     $  + 2.360020794824839d-04*tz02
     $  - 2.318830047258077d-05*tz03
     $  + 2.524580144397893d-06*tz04
     $  - 2.941321932497006d-07*tz05
     $  + 3.598344705267152d-08*tz06
     $  - 4.568837300080304d-09*tz07
     $  + 5.974123046562995d-10*tz08
     $  - 8.000663276105068d-11*tz09
     $  + 1.092972714436844d-11*tz10
     $  - 1.518425691916936d-12*tz11
     $  + 2.140136907706228d-13*tz12
     $  - 3.054413521264523d-14*tz13
     $  + 4.407393721803468d-15*tz14
     $  - 6.421795084001050d-16*tz15

      CsH4YZ(3,2,2,1,05) = 
     $  + 1.418232960257141d-03
     $  - 3.798580041708901d-04*tz01
     $  - 1.198852717835719d-04*tz02
     $  - 1.349982437115294d-05*tz03
     $  + 1.548887204023433d-06*tz04
     $  - 1.876198976081659d-07*tz05
     $  + 2.365995211897910d-08*tz06
     $  - 3.078477337533515d-09*tz07
     $  + 4.107523605946576d-10*tz08
     $  - 5.595328868832913d-11*tz09
     $  + 7.755959758064519d-12*tz10
     $  - 1.091195446969651d-12*tz11
     $  + 1.555080809947455d-13*tz12
     $  - 2.241217838662839d-14*tz13
     $  + 3.262267378343134d-15*tz14
     $  - 4.790557845711296d-16*tz15

      CsH4YZ(3,2,2,1,06) = 
     $  - 4.296506029495524d-06
     $  - 1.363685500841061d-05*tz01
     $  + 1.395237383689660d-05*tz02
     $  + 6.917680883832743d-06*tz03
     $  + 1.021783121075014d-06*tz04
     $  - 1.278734938466097d-07*tz05
     $  + 1.654827082458191d-08*tz06
     $  - 2.199178257841414d-09*tz07
     $  + 2.986638456682791d-10*tz08
     $  - 4.130084236678706d-11*tz09
     $  + 5.799655310890529d-12*tz10
     $  - 8.252493731763414d-13*tz11
     $  + 1.187867017084341d-13*tz12
     $  - 1.727224319404880d-14*tz13
     $  + 2.534137773212080d-15*tz14
     $  - 3.748012934266142d-16*tz15

      CsH4YZ(3,2,2,1,07) = 
     $  - 6.822729832387106d-06
     $  + 3.895124307840007d-06*tz01
     $  + 5.586687080219600d-07*tz02
     $  - 7.727108228134767d-07*tz03
     $  - 5.294613002626276d-07*tz04
     $  - 9.159635298539954d-08*tz05
     $  + 1.212018116215517d-08*tz06
     $  - 1.640660777151327d-09*tz07
     $  + 2.263091547590071d-10*tz08
     $  - 3.171617699632511d-11*tz09
     $  + 4.505753286793433d-12*tz10
     $  - 6.477097747639646d-13*tz11
     $  + 9.407828618994771d-14*tz12
     $  - 1.379040867859342d-14*tz13
     $  + 2.038035675334576d-15*tz14
     $  - 3.034133679652683d-16*tz15

      CsH4YZ(3,2,2,1,08) = 
     $  - 5.761092196717160d-10
     $  + 1.185962531974067d-07*tz01
     $  - 2.114577482970807d-07*tz02
     $  - 5.816233851135780d-08*tz03
     $  + 5.256543921967593d-08*tz04
     $  + 4.815037357073134d-08*tz05
     $  + 9.205131643604813d-09*tz06
     $  - 1.266364550541218d-09*tz07
     $  + 1.771052605023841d-10*tz08
     $  - 2.511844905430152d-11*tz09
     $  + 3.605916940453726d-12*tz10
     $  - 5.231638926099931d-13*tz11
     $  + 7.661589694282983d-14*tz12
     $  - 1.131389825656361d-14*tz13
     $  + 1.683222356543939d-15*tz14
     $  - 2.521114631798674d-16*tz15

      CsH4YZ(3,2,2,1,09) = 
     $  + 5.944400544133661d-08
     $  - 4.425180245136375d-08*tz01
     $  - 5.978130327266947d-10*tz02
     $  + 1.604434864928419d-08*tz03
     $  + 7.491522722947068d-09*tz04
     $  - 4.000088919421321d-09*tz05
     $  - 4.922268836400165d-09*tz06
     $  - 1.004956915397086d-09*tz07
     $  + 1.422827752933807d-10*tz08
     $  - 2.039684504381478d-11*tz09
     $  + 2.955848031325461d-12*tz10
     $  - 4.324589075268945d-13*tz11
     $  + 6.380963698853545d-14*tz12
     $  - 9.486772100895134d-15*tz13
     $  + 1.420077169221608d-15*tz14
     $  - 2.138901738541274d-16*tz15

      CsH4YZ(3,2,2,1,10) = 
     $  + 3.148857221429442d-10
     $  - 1.332657366264871d-09*tz01
     $  + 2.807714798535107d-09*tz02
     $  + 2.423453721946389d-10*tz03
     $  - 1.494046404040672d-09*tz04
     $  - 1.020136524663075d-09*tz05
     $  + 3.198283427986778d-10*tz06
     $  + 5.477097180821098d-10*tz07
     $  + 1.168374441917158d-10*tz08
     $  - 1.691145843670012d-11*tz09
     $  + 2.471796926171591d-12*tz10
     $  - 3.644121890149960d-13*tz11
     $  + 5.414010474100753d-14*tz12
     $  - 8.099402063455746d-15*tz13
     $  + 1.219289546968431d-15*tz14
     $  - 1.846000598225939d-16*tz15

      CsH4YZ(3,2,2,1,11) = 
     $  - 6.062049552357585d-10
     $  + 5.208960469325449d-10*tz01
     $  - 4.563636784057660d-11*tz02
     $  - 2.469575666850833d-10*tz03
     $  - 7.464751751899343d-11*tz04
     $  + 1.564440814968938d-10*tz05
     $  + 1.430278807888582d-10*tz06
     $  - 2.518869282930949d-11*tz07
     $  - 6.499386111550619d-11*tz08
     $  - 1.427231299770795d-11*tz09
     $  + 2.102340934548695d-12*tz10
     $  - 3.121158282735727d-13*tz11
     $  + 4.666405642747181d-14*tz12
     $  - 7.021126053347163d-15*tz13
     $  + 1.062514877166810d-15*tz14
     $  - 1.616390132836697d-16*tz15

      CsH4YZ(3,2,2,1,12) = 
     $  - 6.360979791444392d-12
     $  + 1.658679959105871d-11*tz01
     $  - 3.583102953132789d-11*tz02
     $  + 2.732524390003042d-12*tz03
     $  + 2.796098632527551d-11*tz04
     $  + 1.587804374412229d-11*tz05
     $  - 1.753282987192184d-11*tz06
     $  - 2.052032899984366d-11*tz07
     $  + 1.734871021078921d-12*tz08
     $  + 8.111827317924280d-12*tz09
     $  + 1.814466434002663d-12*tz10
     $  - 2.711080074349377d-13*tz11
     $  + 4.076921751981026d-14*tz12
     $  - 6.166776624696066d-15*tz13
     $  + 9.377679510594462d-16*tz14
     $  - 1.433002261162513d-16*tz15

      CsH4YZ(3,2,2,1,13) = 
     $  + 6.649147726353312d-12
     $  - 6.258745306807457d-12*tz01
     $  + 1.038148078633358d-12*tz02
     $  + 3.410099097590292d-12*tz03
     $  + 5.700125661806400d-13*tz04
     $  - 3.617960810639615d-12*tz05
     $  - 2.985990360821655d-12*tz06
     $  + 2.042181740177689d-12*tz07
     $  + 3.003984653391161d-12*tz08
     $  - 6.031096288256418d-14*tz09
     $  - 1.054448405778897d-12*tz10
     $  - 2.384001387555042d-13*tz11
     $  + 3.604343413315252d-14*tz12
     $  - 5.478785138619956d-15*tz13
     $  + 8.369150712739608d-16*tz14
     $  - 1.284225581562810d-16*tz15

      CsH4YZ(3,2,2,1,14) = 
     $  + 1.021483712251096d-13
     $  - 2.160710114849286d-13*tz01
     $  + 4.511507596691282d-13*tz02
     $  - 9.445664172343355d-14*tz03
     $  - 4.495881764100833d-13*tz04
     $  - 2.279369324783201d-13*tz05
     $  + 4.997301097926555d-13*tz06
     $  + 5.343647500733506d-13*tz07
     $  - 2.422887192278660d-13*tz08
     $  - 4.477061313257410d-13*tz09
     $  - 1.219999276830259d-14*tz10
     $  + 1.417282542720893d-13*tz11
     $  + 3.220076301879146d-14*tz12
     $  - 4.917046739960192d-15*tz13
     $  + 7.542701343031838d-16*tz14
     $  - 1.161918719608907d-16*tz15

      CsH4YZ(3,2,2,1,15) = 
     $  - 7.617303998736882d-14
     $  + 7.640649916277997d-14*tz01
     $  - 1.723690938916646d-14*tz02
     $  - 4.473828737007642d-14*tz03
     $  - 3.066284471411657d-15*tz04
     $  + 6.946319323021505d-14*tz05
     $  + 5.627519958415198d-14*tz06
     $  - 7.129136305780052d-14*tz07
     $  - 9.355758284229867d-14*tz08
     $  + 2.877803605158920d-14*tz09
     $  + 6.779801317133947d-14*tz10
     $  + 4.278646364099408d-15*tz11
     $  - 1.959016869669218d-14*tz12
     $  - 4.452901776815122d-15*tz13
     $  + 6.857556637210097d-16*tz14
     $  - 1.060226245461214d-16*tz15

      CsH4YZ(3,2,2,1,16) = 
     $  - 1.513230450452423d-15
     $  + 2.875333943697542d-15*tz01
     $  - 5.669495213673044d-15*tz02
     $  + 1.762912830254195d-15*tz03
     $  + 6.735391686040489d-15*tz04
     $  + 3.378428064426204d-15*tz05
     $  - 1.139641005103449d-14*tz06
     $  - 1.212713617280633d-14*tz07
     $  + 1.033207518992429d-14*tz08
     $  + 1.622124736819594d-14*tz09
     $  - 3.355690537546472d-15*tz10
     $  - 1.041383197427839d-14*tz11
     $  - 9.470094436572296d-16*tz12
     $  + 2.772875473590192d-15*tz13
     $  + 6.283833317969307d-16*tz14

      CsH4YZ(3,2,2,1,17) = 
     $  + 8.990161549059668d-16
     $  - 9.454211025114703d-16*tz01
     $  + 2.559457304865315d-16*tz02
     $  + 5.723797548218858d-16*tz03
     $  - 1.221822027703494d-15*tz05
     $  - 1.040017297391120d-15*tz06
     $  + 1.908378080256723d-15*tz07
     $  + 2.462398685012263d-15*tz08
     $  - 1.506781527597281d-15*tz09
     $  - 2.801569507304301d-15*tz10
     $  + 3.729063138264573d-16*tz11
     $  + 1.619965248500739d-15*tz12
     $  + 1.829506078707484d-16*tz13
     $  - 4.005692268469139d-16*tz14

      CsH4YZ(3,2,2,1,18) = 
     $  + 2.330588132132902d-16*tz06
     $  + 2.592159828199794d-16*tz07
     $  - 3.212722528234368d-16*tz08
     $  - 4.844204666616207d-16*tz09
     $  + 2.196315882395363d-16*tz10
     $  + 4.834401778613589d-16*tz11
     $  - 2.548696527797162d-16*tz13

      CsH4YZ(3,2,3,1,01) = 
     $  + 1.874368001799659d-01
     $  - 2.083428956341240d-01*tz01
     $  + 2.357274352803091d-02*tz02
     $  - 3.031303696017906d-03*tz03
     $  + 4.167019720338215d-04*tz04
     $  - 5.969390194114389d-05*tz05
     $  + 8.795186668765136d-06*tz06
     $  - 1.322630175277970d-06*tz07
     $  + 2.020165954034241d-07*tz08
     $  - 3.123580323958862d-08*tz09
     $  + 4.877695782261253d-09*tz10
     $  - 7.679308427086443d-10*tz11
     $  + 1.217322986245425d-10*tz12
     $  - 1.940993890481859d-11*tz13
     $  + 3.110479609433400d-12*tz14
     $  - 5.006491836972863d-13*tz15
     $  + 8.089325868828035d-14*tz16
     $  - 1.311505610061754d-14*tz17
     $  + 2.132812065202462d-15*tz18
     $  - 3.476963791684015d-16*tz19

      CsH4YZ(3,2,3,1,02) = 
     $  + 1.999586677141848d-01
     $  - 4.676679204056912d-02*tz01
     $  + 6.130592103954588d-03*tz02
     $  - 8.535074593864284d-04*tz03
     $  + 1.233783588451278d-04*tz04
     $  - 1.830221635455968d-05*tz05
     $  + 2.766926775707441d-06*tz06
     $  - 4.244172000028847d-07*tz07
     $  + 6.585288721541328d-08*tz08
     $  - 1.031346753305577d-08*tz09
     $  + 1.627752710214829d-09*tz10
     $  - 2.585815607425403d-10*tz11
     $  + 4.130667547289741d-11*tz12
     $  - 6.630238115474429d-12*tz13
     $  + 1.068709263423931d-12*tz14
     $  - 1.728997989560287d-13*tz15
     $  + 2.806427030744853d-14*tz16
     $  - 4.568600127449212d-15*tz17
     $  + 7.456926887283465d-16*tz18
     $  - 1.219687581188112d-16*tz19

      CsH4YZ(3,2,3,1,03) = 
     $  - 5.303563353220506d-03
     $  + 2.374297977051178d-02*tz01
     $  - 1.593249452626160d-03*tz02
     $  + 1.359785584894596d-04*tz03
     $  - 1.257677069210971d-05*tz04
     $  + 1.172124430770878d-06*tz05
     $  - 1.019875795701576d-07*tz06
     $  + 6.910813008753989d-09*tz07
     $  - 6.923929264080967d-12*tz08
     $  - 1.338738300696486d-10*tz09
     $  + 3.874247529394727d-11*tz10
     $  - 8.509665048697383d-12*tz11
     $  + 1.680667959016454d-12*tz12
     $  - 3.143556342890368d-13*tz13
     $  + 5.694831901318388d-14*tz14
     $  - 1.010859082517513d-14*tz15
     $  + 1.769852980545826d-15*tz16
     $  - 3.069139370421732d-16*tz17

      CsH4YZ(3,2,3,1,04) = 
     $  + 4.447884269748241d-03
     $  - 3.763614722000802d-03*tz01
     $  - 1.232865916323074d-04*tz02
     $  - 5.025507690124197d-06*tz03
     $  + 3.026318171065751d-06*tz04
     $  - 7.086383966453556d-07*tz05
     $  + 1.382171355301620d-07*tz06
     $  - 2.508424682931039d-08*tz07
     $  + 4.393462493882903d-09*tz08
     $  - 7.545417322928708d-10*tz09
     $  + 1.280917695465162d-10*tz10
     $  - 2.159052959268192d-11*tz11
     $  + 3.622965291392709d-12*tz12
     $  - 6.062392285120675d-13*tz13
     $  + 1.012670941577766d-13*tz14
     $  - 1.689843358111544d-14*tz15
     $  + 2.818283451401850d-15*tz16
     $  - 4.699300821657306d-16*tz17

      CsH4YZ(3,2,3,1,05) = 
     $  - 3.907338249443925d-04
     $  + 2.283376595993223d-04*tz01
     $  + 8.074629418239298d-05*tz02
     $  + 1.070672942091418d-05*tz03
     $  - 1.462037128491456d-07*tz04
     $  - 1.460080015078552d-07*tz05
     $  + 4.304177230855169d-08*tz06
     $  - 9.312970859786086d-09*tz07
     $  + 1.804865168756002d-09*tz08
     $  - 3.314814989074214d-10*tz09
     $  + 5.905948128755925d-11*tz10
     $  - 1.032777656264652d-11*tz11
     $  + 1.784206880218403d-12*tz12
     $  - 3.057137061142205d-13*tz13
     $  + 5.208426083665121d-14*tz14
     $  - 8.837827693056078d-15*tz15
     $  + 1.495295712794732d-15*tz16
     $  - 2.524697659381429d-16*tz17

      CsH4YZ(3,2,3,1,06) = 
     $  - 6.917185487137669d-07
     $  + 8.794665901395428d-06*tz01
     $  - 9.461413543787049d-06*tz02
     $  - 3.947932798730864d-06*tz03
     $  - 8.050275671209203d-07*tz04
     $  + 1.073942653760954d-08*tz05
     $  + 1.255961433652938d-08*tz06
     $  - 3.822391804328012d-09*tz07
     $  + 8.516281268917967d-10*tz08
     $  - 1.692372725122753d-10*tz09
     $  + 3.175730148405555d-11*tz10
     $  - 5.764244315158439d-12*tz11
     $  + 1.024472517953496d-12*tz12
     $  - 1.795297390185279d-13*tz13
     $  + 3.115333937503006d-14*tz14
     $  - 5.367955590412725d-15*tz15
     $  + 9.201552465416178d-16*tz16
     $  - 1.571227133831975d-16*tz17

      CsH4YZ(3,2,3,1,07) = 
     $  + 1.496262882732139d-06
     $  - 1.723153558555097d-06*tz01
     $  - 1.260208961514710d-07*tz02
     $  + 4.747552102307695d-07*tz03
     $  + 2.822341606783984d-07*tz04
     $  + 6.944295638154545d-08*tz05
     $  - 6.083177365721292d-10*tz06
     $  - 1.269940187756012d-09*tz07
     $  + 3.876169413198042d-10*tz08
     $  - 8.768681726032095d-11*tz09
     $  + 1.770296546444723d-11*tz10
     $  - 3.371171563995465d-12*tz11
     $  + 6.201202138217988d-13*tz12
     $  - 1.115486080671391d-13*tz13
     $  + 1.976159865264984d-14*tz14
     $  - 3.463098538786999d-15*tz15
     $  + 6.020720823354857d-16*tz16
     $  - 1.040506151754566d-16*tz17

      CsH4YZ(3,2,3,1,08) = 
     $  + 4.768720575380788d-08
     $  - 4.273579894490252d-08*tz01
     $  + 1.124266283607089d-07*tz02
     $  + 1.306864636436587d-08*tz03
     $  - 3.009361947832236d-08*tz04
     $  - 2.418942451376656d-08*tz05
     $  - 6.842977657125150d-09*tz06
     $  + 4.333781083035341d-11*tz07
     $  + 1.372125321906449d-10*tz08
     $  - 4.225831442547117d-11*tz09
     $  + 9.691724260521494d-12*tz10
     $  - 1.982881380430685d-12*tz11
     $  + 3.822491202110653d-13*tz12
     $  - 7.109958967712852d-14*tz13
     $  + 1.291903485557555d-14*tz14
     $  - 2.309733500891742d-15*tz15
     $  + 4.081515644946728d-16*tz16

      CsH4YZ(3,2,3,1,09) = 
     $  - 1.039452501401962d-08
     $  + 1.562162713533531d-08*tz01
     $  - 3.169306691563977d-09*tz02
     $  - 8.108647503436358d-09*tz03
     $  - 2.302604442793281d-09*tz04
     $  + 2.147857813340534d-09*tz05
     $  + 2.347334660288505d-09*tz06
     $  + 7.403579684319108d-10*tz07
     $  - 3.901140380613530d-12*tz08
     $  - 1.569019205979866d-11*tz09
     $  + 4.884467625178465d-12*tz10
     $  - 1.133751730619855d-12*tz11
     $  + 2.345794935383539d-13*tz12
     $  - 4.568705091004513d-14*tz13
     $  + 8.577584033185293d-15*tz14
     $  - 1.571876861729925d-15*tz15
     $  + 2.832157915348651d-16*tz16

      CsH4YZ(3,2,3,1,10) = 
     $  - 6.955096364557379d-10
     $  + 2.485509601751359d-10*tz01
     $  - 1.216317815690713d-09*tz02
     $  + 2.627072693711204d-10*tz03
     $  + 7.378297969891740d-10*tz04
     $  + 3.569810211788062d-10*tz05
     $  - 1.565898873131985d-10*tz06
     $  - 2.496225960655116d-10*tz07
     $  - 8.583638326754427d-11*tz08
     $  + 4.347356942998976d-13*tz09
     $  + 1.880861337945719d-12*tz10
     $  - 5.920448648371885d-13*tz11
     $  + 1.388699887157737d-13*tz12
     $  - 2.900892198019517d-14*tz13
     $  + 5.699138793846130d-15*tz14
     $  - 1.078504454069273d-15*tz15
     $  + 1.990757071368668d-16*tz16

      CsH4YZ(3,2,3,1,11) = 
     $  + 7.867882786574775d-11
     $  - 1.528369202665996d-10*tz01
     $  + 7.093582899071886d-11*tz02
     $  + 1.040546550516081d-10*tz03
     $  - 4.564769317983659d-12*tz04
     $  - 7.642743218798292d-11*tz05
     $  - 5.300061366275765d-11*tz06
     $  + 1.031102718567862d-11*tz07
     $  + 2.844801415056693d-11*tz08
     $  + 1.050292141753352d-11*tz09
     $  - 5.774977345659456d-14*tz10
     $  - 2.343551767249670d-13*tz11
     $  + 7.459984916239757d-14*tz12
     $  - 1.766213221315266d-14*tz13
     $  + 3.720183588215906d-15*tz14
     $  - 7.363669457556048d-16*tz15
     $  + 1.403031204526891d-16*tz16

      CsH4YZ(3,2,3,1,12) = 
     $  + 8.673158678898379d-12
     $  - 1.443124130310252d-12*tz01
     $  + 1.289752735665331d-11*tz02
     $  - 7.421393345803268d-12*tz03
     $  - 1.201725669337365d-11*tz04
     $  - 2.744738611472084d-12*tz05
     $  + 8.432998644690991d-12*tz06
     $  + 7.805550822621868d-12*tz07
     $  - 3.806543093872587d-13*tz08
     $  - 3.422619734487817d-12*tz09
     $  - 1.341831216728720d-12*tz10
     $  + 8.595991394229308d-15*tz11
     $  + 3.015458530263545d-14*tz12
     $  - 9.707702106558908d-15*tz13
     $  + 2.317856720870795d-15*tz14
     $  - 4.917873391353587d-16*tz15

      CsH4YZ(3,2,3,1,13) = 
     $  - 5.619665208305939d-13
     $  + 1.558320838764542d-12*tz01
     $  - 1.102531525091971d-12*tz02
     $  - 1.211471687510291d-12*tz03
     $  + 5.125340985551087d-13*tz04
     $  + 1.608145058566208d-12*tz05
     $  + 7.826541892350122d-13*tz06
     $  - 9.574903273590284d-13*tz07
     $  - 1.154691667275444d-12*tz08
     $  - 5.673068163570863d-14*tz09
     $  + 4.301669582853381d-13*tz10
     $  + 1.775917091987673d-13*tz11
     $  - 1.360759989667386d-15*tz12
     $  - 3.986650624955765d-15*tz13
     $  + 1.298076700963331d-15*tz14
     $  - 3.123443146961902d-16*tz15

      CsH4YZ(3,2,3,1,14) = 
     $  - 1.016145872967538d-13
     $  + 6.310170378395208d-15*tz01
     $  - 1.368446847351423d-13*tz02
     $  + 1.298740995638225d-13*tz03
     $  + 1.723562132768031d-13*tz04
     $  - 4.279984246932703d-15*tz05
     $  - 2.273146208060740d-13*tz06
     $  - 1.631776221595609d-13*tz07
     $  + 1.090335097313555d-13*tz08
     $  + 1.724146803621814d-13*tz09
     $  + 2.074163826707939d-14*tz10
     $  - 5.604676487888945d-14*tz11
     $  - 2.420577652385457d-14*tz12
     $  + 2.220465499367194d-16*tz13
     $  + 5.394018416196385d-16*tz14
     $  - 1.776523242403906d-16*tz15

      CsH4YZ(3,2,3,1,15) = 
     $  + 3.113454614317661d-15
     $  - 1.633936626052902d-14*tz01
     $  + 1.524017175303896d-14*tz02
     $  + 1.363784835738376d-14*tz03
     $  - 1.155461402808282d-14*tz04
     $  - 2.893933156022404d-14*tz05
     $  - 9.265680945236018d-15*tz06
     $  + 3.277498219561939d-14*tz07
     $  + 3.078901926913462d-14*tz08
     $  - 1.210982632740749d-14*tz09
     $  - 2.602827643546897d-14*tz10
     $  - 4.555781039529792d-15*tz11
     $  + 7.526361515619157d-15*tz12
     $  + 3.382287605249416d-15*tz13

      CsH4YZ(3,2,3,1,16) = 
     $  + 1.137334187206481d-15
     $  + 1.469582182587392d-15*tz02
     $  - 1.936709101497944d-15*tz03
     $  - 2.394332199600855d-15*tz04
     $  + 6.203657081694162d-16*tz05
     $  + 5.016291952658721d-15*tz06
     $  + 2.916996082141543d-15*tz07
     $  - 4.747795424124843d-15*tz08
     $  - 5.563536873313674d-15*tz09
     $  + 1.254190223345551d-15*tz10
     $  + 3.973598667833500d-15*tz11
     $  + 8.698077896477177d-16*tz12
     $  - 1.037047755088224d-15*tz13
     $  - 4.827733565286763d-16*tz14

      CsH4YZ(3,2,3,1,17) = 
     $  + 1.751769261845425d-16*tz01
     $  - 1.997431225319550d-16*tz02
     $  - 1.545270116692188d-16*tz03
     $  + 1.973224254239013d-16*tz04
     $  + 4.958945292295298d-16*tz05
     $  - 8.676691225940313d-16*tz07
     $  - 6.913848372237617d-16*tz08
     $  + 6.848545988417689d-16*tz09
     $  + 9.842750903445205d-16*tz10
     $  - 1.084923695968406d-16*tz11
     $  - 6.132341502114395d-16*tz12
     $  - 1.564786616321818d-16*tz13
     $  + 1.461028877104458d-16*tz14

      CsH4YZ(3,2,3,1,18) = 
     $  - 1.021664657210451d-16*tz06
     $  + 1.486371082187174d-16*tz08
     $  + 1.477366050225939d-16*tz09
     $  - 1.722581281218534d-16*tz11

      CsH4YZ(3,3,2,1,01) = 
     $  + 7.114697789062527d-01
     $  - 7.465137653450127d-01*tz01
     $  + 3.814687613286294d-02*tz02
     $  - 3.439646749648303d-03*tz03
     $  + 3.773232377956643d-04*tz04
     $  - 4.577609377915571d-05*tz05
     $  + 5.908861807534797d-06*tz06
     $  - 7.957476367501820d-07*tz07
     $  + 1.105318699036007d-07*tz08
     $  - 1.572098660787567d-08*tz09
     $  + 2.278285936671066d-09*tz10
     $  - 3.352334579245192d-10*tz11
     $  + 4.995406402270318d-11*tz12
     $  - 7.523535143443474d-12*tz13
     $  + 1.143492115603305d-12*tz14
     $  - 1.751755999961793d-13*tz15
     $  + 2.702178174564039d-14*tz16
     $  - 4.193721209384903d-15*tz17
     $  + 6.543975512248639d-16*tz18
     $  - 1.025855047276436d-16*tz19

      CsH4YZ(3,3,2,1,02) = 
     $  + 1.795741068873234d-01
     $  + 1.052650805747602d-01*tz01
     $  - 3.782306040460532d-03*tz02
     $  + 2.035157514051272d-04*tz03
     $  - 1.064862224869735d-05*tz04
     $  + 1.799660350436386d-07*tz05
     $  + 9.258238326715448d-08*tz06
     $  - 2.536531107717461d-08*tz07
     $  + 5.034917664851520d-09*tz08
     $  - 9.005493422775555d-10*tz09
     $  + 1.537519302122150d-10*tz10
     $  - 2.563233951310657d-11*tz11
     $  + 4.217845686656154d-12*tz12
     $  - 6.889857684655028d-13*tz13
     $  + 1.120888194994833d-13*tz14
     $  - 1.819679887757805d-14*tz15
     $  + 2.951422644192513d-15*tz16
     $  - 4.786412997988865d-16*tz17

      CsH4YZ(3,3,2,1,03) = 
     $  - 5.632233965050983d-02
     $  + 3.613722260259321d-03*tz01
     $  - 7.472088841955095d-08*tz02
     $  - 3.168708853783135d-05*tz03
     $  + 7.092440287516257d-06*tz04
     $  - 1.270704220090195d-06*tz05
     $  + 2.121126117317850d-07*tz06
     $  - 3.440070069027863d-08*tz07
     $  + 5.508717723899042d-09*tz08
     $  - 8.773762754337386d-10*tz09
     $  + 1.394915329531317d-10*tz10
     $  - 2.217998854426127d-11*tz11
     $  + 3.530765417535380d-12*tz12
     $  - 5.629974463592060d-13*tz13
     $  + 8.994842935232308d-14*tz14
     $  - 1.440067550999847d-14*tz15
     $  + 2.310386260343208d-15*tz16
     $  - 3.714442191976975d-16*tz17

      CsH4YZ(3,3,2,1,04) = 
     $  - 3.709456019259694d-03
     $  - 3.070412985477329d-03*tz01
     $  - 1.830351987280532d-04*tz02
     $  + 3.816840193809267d-06*tz03
     $  + 1.415606796595855d-06*tz04
     $  - 4.052300281087182d-07*tz05
     $  + 8.178148390568235d-08*tz06
     $  - 1.478681678415371d-08*tz07
     $  + 2.546034566538930d-09*tz08
     $  - 4.274892277194459d-10*tz09
     $  + 7.078240084973627d-11*tz10
     $  - 1.162641307497170d-11*tz11
     $  + 1.900913986608072d-12*tz12
     $  - 3.100021172068856d-13*tz13
     $  + 5.049031933806656d-14*tz14
     $  - 8.219576259696630d-15*tz15
     $  + 1.338191186667859d-15*tz16
     $  - 2.179577929947725d-16*tz17

      CsH4YZ(3,3,2,1,05) = 
     $  - 8.594672461084673d-04
     $  + 3.562453207131536d-04*tz01
     $  + 8.117427792964341d-05*tz02
     $  + 1.034107296114630d-05*tz03
     $  - 1.958957106740252d-07*tz04
     $  - 1.119401379112045d-07*tz05
     $  + 3.293161743543641d-08*tz06
     $  - 6.929879766274240d-09*tz07
     $  + 1.301266311211692d-09*tz08
     $  - 2.315182163194145d-10*tz09
     $  + 3.999041764481433d-11*tz10
     $  - 6.786938486696779d-12*tz11
     $  + 1.139203513163357d-12*tz12
     $  - 1.898629366485859d-13*tz13
     $  + 3.149627671926057d-14*tz14
     $  - 5.209034582305235d-15*tz15
     $  + 8.598096622438971d-16*tz16
     $  - 1.417507196021727d-16*tz17

      CsH4YZ(3,3,2,1,06) = 
     $  + 2.073460779834993d-05
     $  + 1.823628705085265d-07*tz01
     $  - 1.206348993760818d-05*tz02
     $  - 4.275371724716506d-06*tz03
     $  - 7.515742918625314d-07*tz04
     $  + 1.071332830093470d-08*tz05
     $  + 1.051611686640133d-08*tz06
     $  - 3.111583851769640d-09*tz07
     $  + 6.716274581114310d-10*tz08
     $  - 1.293797069088428d-10*tz09
     $  + 2.356210785723613d-11*tz10
     $  - 4.155783533963328d-12*tz11
     $  + 7.185719067009481d-13*tz12
     $  - 1.226455450609610d-13*tz13
     $  + 2.074990898399717d-14*tz14
     $  - 3.489262849515383d-15*tz15
     $  + 5.842320583773655d-16*tz16

      CsH4YZ(3,3,2,1,07) = 
     $  + 3.576176425168228d-06
     $  - 3.243292346685135d-06*tz01
     $  + 2.139384816894718d-07*tz02
     $  + 6.292563555108531d-07*tz03
     $  + 3.045245559484858d-07*tz04
     $  + 6.588402072929183d-08*tz05
     $  - 7.203052093464643d-10*tz06
     $  - 1.078011411692400d-09*tz07
     $  + 3.225039529352446d-10*tz08
     $  - 7.105811257153823d-11*tz09
     $  + 1.396678761658075d-11*tz10
     $  - 2.590934310616831d-12*tz11
     $  + 4.646599041114102d-13*tz12
     $  - 8.156165778759915d-14*tz13
     $  + 1.411165334561913d-14*tz14
     $  - 2.417179395944781d-15*tz15
     $  + 4.110696652085932d-16*tz16

      CsH4YZ(3,3,2,1,08) = 
     $  - 2.065144481152782d-07
     $  + 9.436612988822979d-08*tz01
     $  + 1.466743607893741d-07*tz02
     $  - 3.516787111477484d-09*tz03
     $  - 4.134579736078079d-08*tz04
     $  - 2.605112278840252d-08*tz05
     $  - 6.549855028239847d-09*tz06
     $  + 5.806141082567387d-11*tz07
     $  + 1.184137482099456d-10*tz08
     $  - 3.587071449279700d-11*tz09
     $  + 8.038964445508206d-12*tz10
     $  - 1.606057479604332d-12*tz11
     $  + 3.024252052123455d-13*tz12
     $  - 5.498045547835517d-14*tz13
     $  + 9.770946981039033d-15*tz14
     $  - 1.709749055786444d-15*tz15
     $  + 2.959010128149029d-16*tz16

      CsH4YZ(3,3,2,1,09) = 
     $  - 2.830592336244204d-08
     $  + 3.350992355128495d-08*tz01
     $  - 1.125144547644656d-08*tz02
     $  - 1.014887982543152d-08*tz03
     $  - 1.188818328138330d-09*tz04
     $  + 3.078198658282425d-09*tz05
     $  + 2.525714386223684d-09*tz06
     $  + 7.124460806905141d-10*tz07
     $  - 5.528638886604302d-12*tz08
     $  - 1.374794322853323d-11*tz09
     $  + 4.217507324281795d-12*tz10
     $  - 9.588140301039641d-13*tz11
     $  + 1.941567455443072d-13*tz12
     $  - 3.701589143817819d-14*tz13
     $  + 6.806081516852161d-15*tz14
     $  - 1.222160111813034d-15*tz15
     $  + 2.158999939472080d-16*tz16

      CsH4YZ(3,3,2,1,10) = 
     $  + 2.346963753622300d-09
     $  - 1.743618510200215d-09*tz01
     $  - 1.488049479331725d-09*tz02
     $  + 7.720905405887946d-10*tz03
     $  + 9.184704203889756d-10*tz04
     $  + 2.742872993854657d-10*tz05
     $  - 2.432303612523692d-10*tz06
     $  - 2.681178552588815d-10*tz07
     $  - 8.293616152688610d-11*tz08
     $  + 6.150005067198605d-13*tz09
     $  + 1.669090181434756d-12*tz10
     $  - 5.184686973597033d-13*tz11
     $  + 1.193333970107412d-13*tz12
     $  - 2.444179963930714d-14*tz13
     $  + 4.708786805448650d-15*tz14
     $  - 8.741427502811815d-16*tz15
     $  + 1.583575493040437d-16*tz16

      CsH4YZ(3,3,2,1,11) = 
     $  + 2.706855064549761d-10
     $  - 3.672103558279661d-10*tz01
     $  + 2.088736812341578d-10*tz02
     $  + 1.127601095367304d-10*tz03
     $  - 4.551535167403140d-11*tz04
     $  - 9.560736176079171d-11*tz05
     $  - 4.657309391645367d-11*tz06
     $  + 1.914231729916762d-11*tz07
     $  + 3.048975332167877d-11*tz08
     $  + 1.018018962968249d-11*tz09
     $  - 7.797740309461331d-14*tz10
     $  - 2.102092333883978d-13*tz11
     $  + 6.610683365041945d-14*tz12
     $  - 1.538160631116678d-14*tz13
     $  + 3.181560830421296d-15*tz14
     $  - 6.184621237017489d-16*tz15
     $  + 1.157606436961539d-16*tz16

      CsH4YZ(3,3,2,1,12) = 
     $  - 2.811810992427253d-11
     $  + 2.614344779839622d-11*tz01
     $  + 1.353710997936579d-11*tz02
     $  - 1.692587578074973d-11*tz03
     $  - 1.306130698061809d-11*tz04
     $  + 1.022261930608977d-12*tz05
     $  + 1.071564126170790d-11*tz06
     $  + 7.303843936669822d-12*tz07
     $  - 1.345746073108147d-12*tz08
     $  - 3.659929821141388d-12*tz09
     $  - 1.303865587750977d-12*tz10
     $  + 1.089482841057886d-14*tz11
     $  + 2.729668507334986d-14*tz12
     $  - 8.689496955282940d-15*tz13
     $  + 2.041599964627711d-15*tz14
     $  - 4.259373491820484d-16*tz15

      CsH4YZ(3,3,2,1,13) = 
     $  - 2.843872524197392d-12
     $  + 4.186610199195209d-12*tz01
     $  - 3.177057649320196d-12*tz02
     $  - 9.747838326378110d-13*tz03
     $  + 1.354662889275112d-12*tz04
     $  + 1.794964015144594d-12*tz05
     $  + 4.042835639159258d-13*tz06
     $  - 1.248205192480664d-12*tz07
     $  - 1.118294581151370d-12*tz08
     $  + 5.480183560733880d-14*tz09
     $  + 4.589460649295539d-13*tz10
     $  + 1.729186434794761d-13*tz11
     $  - 1.624779435798908d-15*tz12
     $  - 3.637478131002611d-15*tz13
     $  + 1.172018902195010d-15*tz14
     $  - 2.778086410738669d-16*tz15

      CsH4YZ(3,3,2,1,14) = 
     $  + 3.467200679793042d-13
     $  - 3.667612021006729d-13*tz01
     $  - 1.079231587620421d-13*tz02
     $  + 2.758745709119505d-13*tz03
     $  + 1.526915553692793d-13*tz04
     $  - 9.310243953147534d-14*tz05
     $  - 2.606328884541372d-13*tz06
     $  - 1.228479201774391d-13*tz07
     $  + 1.477331544587686d-13*tz08
     $  + 1.704678537293751d-13*tz09
     $  + 7.245762133924781d-15*tz10
     $  - 5.966302992645061d-14*tz11
     $  - 2.360865941233464d-14*tz12
     $  + 2.525581108930103d-16*tz13
     $  + 4.955620069840437d-16*tz14
     $  - 1.616090664607437d-16*tz15

      CsH4YZ(3,3,2,1,15) = 
     $  + 3.167477750354847d-14
     $  - 4.923318668596612d-14*tz01
     $  + 4.461711901243440d-14*tz02
     $  + 5.565463865605713d-15*tz03
     $  - 2.521735029874204d-14*tz04
     $  - 2.857487659948641d-14*tz05
     $  + 1.208650438245598d-15*tz06
     $  + 3.848860703472894d-14*tz07
     $  + 2.632855342941541d-14*tz08
     $  - 1.743045414209637d-14*tz09
     $  - 2.607605117459189d-14*tz10
     $  - 2.858701797247704d-15*tz11
     $  + 7.994650621043663d-15*tz12
     $  + 3.303559937640409d-15*tz13

      CsH4YZ(3,3,2,1,16) = 
     $  - 4.356918578724357d-15
     $  + 5.007149877950823d-15*tz01
     $  + 6.529234178327373d-16*tz02
     $  - 3.958611253870163d-15*tz03
     $  - 1.607532340273387d-15*tz04
     $  + 2.235661155625373d-15*tz05
     $  + 5.286865646378851d-15*tz06
     $  + 1.592917632025760d-15*tz07
     $  - 5.704052951991901d-15*tz08
     $  - 5.060939650877017d-15*tz09
     $  + 2.004500256293557d-15*tz10
     $  + 4.015931831416059d-15*tz11
     $  + 6.493223767745888d-16*tz12
     $  - 1.099274110032703d-15*tz13
     $  - 4.721138419437139d-16*tz14

      CsH4YZ(3,3,2,1,17) = 
     $  - 3.673841718343751d-16
     $  + 5.936957230999020d-16*tz01
     $  - 6.030219181023192d-16*tz02
     $  + 3.910253916588398d-16*tz04
     $  + 4.382016617493177d-16*tz05
     $  - 1.272339596845610d-16*tz06
     $  - 9.512125477913475d-16*tz07
     $  - 5.167964287267233d-16*tz08
     $  + 8.430250172893565d-16*tz09
     $  + 9.275847362123655d-16*tz10
     $  - 2.165400739271498d-16*tz11
     $  - 6.234875628746463d-16*tz12
     $  - 1.270180977593782d-16*tz13
     $  + 1.545584382831642d-16*tz14

      CsH4YZ(3,3,2,1,18) = 
     $  - 1.012583153273736d-16*tz06
     $  + 1.673569412193438d-16*tz08
     $  + 1.241323250921922d-16*tz09
     $  - 1.236586252214920d-16*tz10
     $  - 1.659929151217622d-16*tz11
      endif

      return
      end


      subroutine fillcoeff2dhpl321u(iflag,n,CsH2YZ,CsH3YZ,CsH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CsH2YZ,CsH3YZ,CsH4YZ
      real(dp):: 
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
      dimension CsH2YZ(0:3,0:3,22),CsH3YZ(0:3,0:3,0:3,22),
     $          CsH4YZ(0:3,0:3,0:3,0:3,22)
      common /tpu/  
     $     tu01,tu02,tu03,tu04,tu05,tu06,tu07,tu08,tu09,tu10,
     $     tu11,tu12,tu13,tu14,tu15,tu16,tu17,tu18,tu19,tu20,tu21
!$omp threadprivate(/tpu/)

      if (n.eq.2) return

* 2001-04-20:17:54:34.chpl 
* <- HPLexpand.321a 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.3) then 
      CsH3YZ(3,2,1,01) = CsH3YZ(3,2,1,01)
     $  - 5.730452067718910d-01
     $  + 1.661814777672792d-01*tu01
     $  - 1.044414367125278d-02*tu02
     $  + 3.265692532985893d-04*tu03
     $  - 2.745267103560234d-06*tu04
     $  - 1.397343974346948d-07*tu05
     $  + 3.087096008857071d-09*tu06
     $  + 8.746527446236782d-11*tu07
     $  - 3.260099425948416d-12*tu08
     $  - 5.389892458878845d-14*tu09
     $  + 3.253464769413142d-15*tu10

      CsH3YZ(3,2,1,02) = CsH3YZ(3,2,1,02) 
     $  - 2.242658972813053d-01
     $  - 1.287695690983419d-02*tu01
     $  + 4.572644909599990d-04*tu02
     $  + 5.090061736944926d-06*tu03
     $  - 7.304922384557462d-07*tu04
     $  + 4.475932044807181d-09*tu05
     $  + 8.953625710015902d-10*tu06
     $  - 1.600849655683721d-11*tu07
     $  - 9.257309908375396d-13*tu08
     $  + 2.759514561497775d-14*tu09
     $  + 8.335720699011057d-16*tu10

      CsH3YZ(3,2,1,03) = CsH3YZ(3,2,1,03) 
     $  - 3.406321757046882d-02
     $  - 7.837496781905766d-04*tu01
     $  + 7.431212620798203d-05*tu02
     $  - 2.115109548610854d-06*tu03
     $  - 7.853175963076921d-08*tu04
     $  + 5.840545530292171d-09*tu05
     $  + 1.910495532935015d-11*tu06
     $  - 9.628191171685238d-12*tu07
     $  + 1.064589631588044d-13*tu08
     $  + 1.251889518330679d-14*tu09
     $  - 2.849772767209937d-16*tu10

      CsH3YZ(3,2,1,04) = CsH3YZ(3,2,1,04) 
     $  - 2.652720788571377d-03
     $  + 9.496651850643880d-05*tu01
     $  + 5.969617689196941d-07*tu02
     $  - 4.910766031139717d-07*tu03
     $  + 1.799768745591990d-08*tu04
     $  + 7.050022731372880d-10*tu05
     $  - 5.566721429544227d-11*tu06
     $  - 3.300386578361598d-13*tu07
     $  + 1.069739881867316d-13*tu08
     $  - 9.654987437925191d-16*tu09
     $  - 1.607828991834248d-16*tu10

      CsH3YZ(3,2,1,05) = CsH3YZ(3,2,1,05) 
     $  - 1.614883739645722d-05
     $  + 1.330953066187282d-05*tu01
     $  - 1.293424368977583d-06*tu02
     $  + 1.444170564996309d-08*tu03
     $  + 4.335753998463972d-09*tu04
     $  - 1.869476255029723d-10*tu05
     $  - 6.734044371523343d-12*tu06
     $  + 5.895467673182037d-13*tu07
     $  + 3.679734495188505d-15*tu08
     $  - 1.236342987046082d-15*tu09

      CsH3YZ(3,2,1,06) = CsH3YZ(3,2,1,06) 
     $  + 1.348460902369024d-05
     $  - 1.020419092514457d-06*tu01
     $  - 7.194742732938961d-08*tu02
     $  + 1.327684874743251d-08*tu03
     $  - 2.909272395092941d-10*tu04
     $  - 4.071597362896584d-11*tu05
     $  + 2.106338898654203d-12*tu06
     $  + 6.456374326909308d-14*tu07
     $  - 6.621361545003610d-15*tu08

      CsH3YZ(3,2,1,07) = CsH3YZ(3,2,1,07) 
     $  + 2.217275163719011d-07
     $  - 2.134423002377549d-07*tu01
     $  + 2.162798852822383d-08*tu02
     $  + 2.238476971723253d-10*tu03
     $  - 1.408489934170025d-10*tu04
     $  + 4.447356303861687d-12*tu05
     $  + 3.986558188863833d-13*tu06
     $  - 2.485472978116631d-14*tu07
     $  - 6.143198926983901d-16*tu08

      CsH3YZ(3,2,1,08) = CsH3YZ(3,2,1,08) 
     $  - 1.232125449409686d-07
     $  + 1.198635995978051d-08*tu01
     $  + 1.886127458903912d-09*tu02
     $  - 2.958248097161391d-10*tu03
     $  + 2.840022157879987d-12*tu04
     $  + 1.500607637121474d-12*tu05
     $  - 6.258638810701177d-14*tu06
     $  - 3.954611794035953d-15*tu07
     $  + 3.010584964903262d-16*tu08

      CsH3YZ(3,2,1,09) = CsH3YZ(3,2,1,09) 
     $  - 2.883863592689598d-09
     $  + 3.281168055838671d-09*tu01
     $  - 3.538977947659625d-10*tu02
     $  - 1.213863960978383d-11*tu03
     $  + 3.731290714561006d-12*tu04
     $  - 8.796475344926534d-14*tu05
     $  - 1.603084817814503d-14*tu06
     $  + 8.507618018691330d-16*tu07

      CsH3YZ(3,2,1,10) = CsH3YZ(3,2,1,10) 
     $  + 1.316678217767298d-09
     $  - 1.505650541005464d-10*tu01
     $  - 3.874048431683700d-11*tu02
     $  + 6.031337449851309d-12*tu03
     $  + 1.084308179106283d-14*tu04
     $  - 4.497839518959847d-14*tu05
     $  + 1.619155315463806d-15*tu06
     $  + 1.701212723949293d-16*tu07

      CsH3YZ(3,2,1,11) = CsH3YZ(3,2,1,11) 
     $  + 3.619320864929777d-11
     $  - 4.905754016544664d-11*tu01
     $  + 5.754684925130282d-12*tu02
     $  + 3.346709630378127d-13*tu03
     $  - 8.831512205225837d-14*tu04
     $  + 1.535341800594292d-15*tu05
     $  + 5.265921978752235d-16*tu06

      CsH3YZ(3,2,1,12) = CsH3YZ(3,2,1,12) 
     $  - 1.508910422469319d-11
     $  + 2.019735987993291d-12*tu01
     $  + 7.125475311230200d-13*tu02
     $  - 1.169019272147976d-13*tu03
     $  - 1.381997808233664d-15*tu04
     $  + 1.190213352832828d-15*tu05

      CsH3YZ(3,2,1,13) = CsH3YZ(3,2,1,13) 
     $  - 4.434869225452515d-13
     $  + 7.194315297906376d-13*tu01
     $  - 9.363047083541147d-14*tu02
     $  - 7.438614724843276d-15*tu03
     $  + 1.944107685938390d-15*tu04

      CsH3YZ(3,2,1,14) = CsH3YZ(3,2,1,14) 
     $  + 1.796826617212934d-13
     $  - 2.870370164924148d-14*tu01
     $  - 1.226981453426696d-14*tu02
     $  + 2.192213588349093d-15*tu03

      CsH3YZ(3,2,1,15) = CsH3YZ(3,2,1,15) 
     $  + 5.343857667708181d-15
     $  - 1.039551882201847d-14*tu01
     $  + 1.526051792190987d-15*tu02
     $  + 1.470959364221025d-16*tu03

      CsH3YZ(3,2,1,16) = CsH3YZ(3,2,1,16) 
     $  - 2.189781331282027d-15
     $  + 4.256539645833520d-16*tu01
     $  + 2.018293176126689d-16*tu02

      CsH3YZ(3,2,1,17) = CsH3YZ(3,2,1,17) 
     $  + 1.483583074362551d-16*tu01
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CsH4YZ(0,3,2,1,01) = CsH4YZ(0,3,2,1,01) 
     $  - 5.730452067718910d-01
     $  + 1.661814777672792d-01*tu01
     $  - 1.044414367125278d-02*tu02
     $  + 3.265692532985893d-04*tu03
     $  - 2.745267103560234d-06*tu04
     $  - 1.397343974346948d-07*tu05
     $  + 3.087096008857071d-09*tu06
     $  + 8.746527446236782d-11*tu07
     $  - 3.260099425948416d-12*tu08
     $  - 5.389892458878845d-14*tu09
     $  + 3.253464769413142d-15*tu10

      CsH4YZ(0,3,2,1,02) = CsH4YZ(0,3,2,1,02) 
     $  - 4.686972370923115d-03
     $  - 3.759750553628195d-02*tu01
     $  + 2.186909183839897d-03*tu02
     $  - 5.868670412501303d-05*tu03
     $  + 1.494914626896708d-07*tu04
     $  + 2.843816554140888d-08*tu05
     $  - 1.311492161599057d-10*tu06
     $  - 2.440398724011257d-11*tu07
     $  + 1.484031469465581d-13*tu08
     $  + 2.390362116788671d-14*tu09
     $  - 1.932386093144114d-16*tu10

      CsH4YZ(0,3,2,1,03) = CsH4YZ(0,3,2,1,03) 
     $  + 7.724999947529431d-03
     $  + 3.944955311112819d-03*tu01
     $  - 1.955770974973306d-04*tu02
     $  + 3.761350349468724d-06*tu03
     $  + 2.223947810358320d-08*tu04
     $  - 7.959929554206149d-10*tu05
     $  - 5.731612779369032d-11*tu06
     $  + 1.583099258507367d-13*tu07
     $  + 1.002636413771832d-13*tu08
     $  - 1.185988374697760d-16*tu09
     $  - 1.483535472892224d-16*tu10

      CsH4YZ(0,3,2,1,04) = CsH4YZ(0,3,2,1,04) 
     $  - 9.786953367668925d-05
     $  - 5.368367683014316d-05*tu01
     $  - 1.458953136330903d-06*tu02
     $  + 1.351715303835980d-07*tu03
     $  + 3.301318409961313d-09*tu04
     $  - 3.188482465304848d-10*tu05
     $  - 5.215363007062259d-12*tu06
     $  + 6.325336888610150d-13*tu07
     $  + 5.914559201667583d-15*tu08
     $  - 1.091640496707620d-15*tu09

      CsH4YZ(0,3,2,1,05) = CsH4YZ(0,3,2,1,05) 
     $  - 7.325307168314762d-05
     $  - 2.641402993303477d-05*tu01
     $  + 1.311161491591539d-06*tu02
     $  - 8.812441432882592d-09*tu03
     $  - 9.776277645158146d-10*tu04
     $  - 2.322823771487077d-11*tu05
     $  + 2.736014356411163d-12*tu06
     $  + 4.471058292765862d-14*tu07
     $  - 6.002517259670105d-15*tu08

      CsH4YZ(0,3,2,1,06) = CsH4YZ(0,3,2,1,06) 
     $  - 1.041919087620495d-06
     $  + 6.831469031315885d-07*tu01
     $  + 4.002049740856819d-08*tu02
     $  - 2.899141696502405d-09*tu03
     $  - 1.253627154099013d-10*tu04
     $  + 1.007823419393096d-11*tu05
     $  + 2.714558313408015d-13*tu06
     $  - 2.749197878687565d-14*tu07
     $  - 4.300068178025986d-16*tu08

      CsH4YZ(0,3,2,1,07) = CsH4YZ(0,3,2,1,07) 
     $  + 8.577904795669726d-07
     $  + 3.022267019007356d-07*tu01
     $  - 1.516386300683354d-08*tu02
     $  - 1.752128625183497d-10*tu03
     $  + 2.712012247643774d-11*tu04
     $  + 1.080517012513025d-12*tu05
     $  - 9.958658930292380d-14*tu06
     $  - 2.383601474797183d-15*tu07
     $  + 2.821724431123429d-16*tu08

      CsH4YZ(0,3,2,1,08) = CsH4YZ(0,3,2,1,08) 
     $  + 3.394410031145648d-08
     $  - 9.165751199479892d-09*tu01
     $  - 8.959994241914910d-10*tu02
     $  + 6.104343283397081d-11*tu03
     $  + 3.726588027743482d-12*tu04
     $  - 2.948980558996664d-13*tu05
     $  - 1.045185838281554d-14*tu06
     $  + 1.055394421078547d-15*tu07

      CsH4YZ(0,3,2,1,09) = CsH4YZ(0,3,2,1,09) 
     $  - 1.024899456061644d-08
     $  - 3.937176123600539d-09*tu01
     $  + 2.025327746004821d-10*tu02
     $  + 7.051657872375742d-12*tu03
     $  - 6.751514653105742d-13*tu04
     $  - 3.490313211599034d-14*tu05
     $  + 3.215808238713811d-15*tu06

      CsH4YZ(0,3,2,1,10) = CsH4YZ(0,3,2,1,10) 
     $  - 6.354554312558739d-10
     $  + 1.253977767035130d-10*tu01
     $  + 1.796687374194959d-11*tu02
     $  - 1.242954136105294d-12*tu03
     $  - 9.610306167134365d-14*tu04
     $  + 8.002099438419337d-15*tu05
     $  + 3.369613042231594d-16*tu06

      CsH4YZ(0,3,2,1,11) = CsH4YZ(0,3,2,1,11) 
     $  + 1.236663772691993d-10
     $  + 5.396164922206500d-11*tu01
     $  - 2.895693420521096d-12*tu02
     $  - 1.773150925171894d-13*tu03
     $  + 1.574636157140134d-14*tu04
     $  + 9.615026126334737d-16*tu05

      CsH4YZ(0,3,2,1,12) = CsH4YZ(0,3,2,1,12) 
     $  + 1.020537281697551d-11
     $  - 1.742768449040839d-12*tu01
     $  - 3.354633954757997d-13*tu02
     $  + 2.459050633566423d-14*tu03
     $  + 2.244645106204945d-15*tu04
     $  - 2.038393072154295d-16*tu05

      CsH4YZ(0,3,2,1,13) = CsH4YZ(0,3,2,1,13) 
     $  - 1.510307338276937d-12
     $  - 7.569438761990922d-13*tu01
     $  + 4.315093415385588d-14*tu02
     $  + 3.814692752477483d-15*tu03
     $  - 3.501028727576685d-16*tu04

      CsH4YZ(0,3,2,1,14) = CsH4YZ(0,3,2,1,14) 
     $  - 1.532831397444281d-13
     $  + 2.460989700926523d-14*tu01
     $  + 5.955033198582561d-15*tu02
     $  - 4.750752541212174d-16*tu03

      CsH4YZ(0,3,2,1,15) = CsH4YZ(0,3,2,1,15) 
     $  + 1.869870339764732d-14
     $  + 1.073460562246621d-14*tu01
     $  - 6.612270073050465d-16*tu02

      CsH4YZ(0,3,2,1,16) = CsH4YZ(0,3,2,1,16) 
     $  + 2.224139540764735d-15
     $  - 3.529399837488516d-16*tu01
     $  - 1.017734839837405d-16*tu02

      CsH4YZ(0,3,2,1,17) = CsH4YZ(0,3,2,1,17) 
     $  - 2.347008425828115d-16
     $  - 1.529515319632095d-16*tu01

      CsH4YZ(2,3,2,1,02) = CsH4YZ(2,3,2,1,02) 
     $  - 2.148919525394591d-01
     $  + 6.231805416272971d-02*tu01
     $  - 3.916553876719795d-03*tu02
     $  + 1.224634699869709d-04*tu03
     $  - 1.029475163835087d-06*tu04
     $  - 5.240039903801058d-08*tu05
     $  + 1.157661003321401d-09*tu06
     $  + 3.279947792338793d-11*tu07
     $  - 1.222537284730656d-12*tu08
     $  - 2.021209672079567d-14*tu09
     $  + 1.220049288529928d-15*tu10

      CsH4YZ(2,3,2,1,03) = CsH4YZ(2,3,2,1,03) 
     $  - 5.606647432032633d-02
     $  - 3.219239227458547d-03*tu01
     $  + 1.143161227399997d-04*tu02
     $  + 1.272515434236231d-06*tu03
     $  - 1.826230596139365d-07*tu04
     $  + 1.118983011201795d-09*tu05
     $  + 2.238406427503975d-10*tu06
     $  - 4.002124139209304d-12*tu07
     $  - 2.314327477093849d-13*tu08
     $  + 6.898786403744439d-15*tu09
     $  + 2.083930174752764d-16*tu10

      CsH4YZ(2,3,2,1,04) = CsH4YZ(2,3,2,1,04) 
     $  - 6.386853294462904d-03
     $  - 1.469530646607331d-04*tu01
     $  + 1.393352366399663d-05*tu02
     $  - 3.965830403645351d-07*tu03
     $  - 1.472470493076922d-08*tu04
     $  + 1.095102286929782d-09*tu05
     $  + 3.582179124253153d-12*tu06
     $  - 1.805285844690982d-12*tu07
     $  + 1.996105559227582d-14*tu08
     $  + 2.347292846870023d-15*tu09

      CsH4YZ(2,3,2,1,05) = CsH4YZ(2,3,2,1,05) 
     $  - 3.979081182857066d-04
     $  + 1.424497777596582d-05*tu01
     $  + 8.954426533795412d-08*tu02
     $  - 7.366149046709575d-08*tu03
     $  + 2.699653118387985d-09*tu04
     $  + 1.057503409705932d-10*tu05
     $  - 8.350082144316341d-12*tu06
     $  - 4.950579867542397d-14*tu07
     $  + 1.604609822800974d-14*tu08
     $  - 1.448248115688778d-16*tu09

      CsH4YZ(2,3,2,1,06) = CsH4YZ(2,3,2,1,06) 
     $  - 2.018604674557153d-06
     $  + 1.663691332734102d-06*tu01
     $  - 1.616780461221979d-07*tu02
     $  + 1.805213206245387d-09*tu03
     $  + 5.419692498079965d-10*tu04
     $  - 2.336845318787154d-11*tu05
     $  - 8.417555464404179d-13*tu06
     $  + 7.369334591477546d-14*tu07
     $  + 4.599668118985631d-16*tu08
     $  - 1.545428733807603d-16*tu09

      CsH4YZ(2,3,2,1,07) = CsH4YZ(2,3,2,1,07) 
     $  + 1.444779538252526d-06
     $  - 1.093306170551204d-07*tu01
     $  - 7.708652928148887d-09*tu02
     $  + 1.422519508653484d-09*tu03
     $  - 3.117077566171008d-11*tu04
     $  - 4.362425745960626d-12*tu05
     $  + 2.256791677129503d-13*tu06
     $  + 6.917543921688544d-15*tu07
     $  - 7.094315941075296d-16*tu08

      CsH4YZ(2,3,2,1,08) = CsH4YZ(2,3,2,1,08) 
     $  + 2.078695465986572d-08
     $  - 2.001021564728953d-08*tu01
     $  + 2.027623924520978d-09*tu02
     $  + 2.098572160990042d-11*tu03
     $  - 1.320459313284770d-11*tu04
     $  + 4.169396534845568d-13*tu05
     $  + 3.737398301909493d-14*tu06
     $  - 2.330130917813863d-15*tu07

      CsH4YZ(2,3,2,1,09) = CsH4YZ(2,3,2,1,09) 
     $  - 1.026771207841406d-08
     $  + 9.988633299816917d-10*tu01
     $  + 1.571772882419774d-10*tu02
     $  - 2.465206747635693d-11*tu03
     $  + 2.366685131480772d-13*tu04
     $  + 1.250506364212245d-13*tu05
     $  - 5.215532345510606d-15*tu06
     $  - 3.295509845570613d-16*tu07

      CsH4YZ(2,3,2,1,10) = CsH4YZ(2,3,2,1,10) 
     $  - 2.162897694517120d-10
     $  + 2.460876041879153d-10*tu01
     $  - 2.654233460743417d-11*tu02
     $  - 9.103979707234786d-13*tu03
     $  + 2.798468035994978d-13*tu04
     $  - 6.597356503841793d-15*tu05
     $  - 1.202313610484962d-15*tu06

      CsH4YZ(2,3,2,1,11) = CsH4YZ(2,3,2,1,11) 
     $  + 8.977351484778524d-11
     $  - 1.026579914319065d-11*tu01
     $  - 2.641396657941596d-12*tu02
     $  + 4.112275534182049d-13*tu03
     $  + 7.393010448753358d-16*tu04
     $  - 3.066708754155271d-15*tu05
     $  + 1.103969584065336d-16*tu06

      CsH4YZ(2,3,2,1,12) = CsH4YZ(2,3,2,1,12) 
     $  + 2.262075540584994d-12
     $  - 3.066096260333017d-12*tu01
     $  + 3.596678078270318d-13*tu02
     $  + 2.091693519486350d-14*tu03
     $  - 5.519695124724337d-15*tu04

      CsH4YZ(2,3,2,1,13) = CsH4YZ(2,3,2,1,13) 
     $  - 8.705252437264963d-13
     $  + 1.165232300875585d-13*tu01
     $  + 4.110851142038084d-14*tu02
     $  - 6.744341947421058d-15*tu03

      CsH4YZ(2,3,2,1,14) = CsH4YZ(2,3,2,1,14) 
     $  - 2.375822798464596d-14
     $  + 3.854097482697543d-14*tu01
     $  - 5.015918066233222d-15*tu02
     $  - 3.984972065733643d-16*tu03
     $  + 1.041486333990489d-16*tu04

      CsH4YZ(2,3,2,1,15) = CsH4YZ(2,3,2,1,15) 
     $  + 8.984133111778971d-15
     $  - 1.435185033987517d-15*tu01
     $  - 6.134906861376395d-16*tu02
     $  + 1.096107095425985d-16*tu03

      CsH4YZ(2,3,2,1,16) = CsH4YZ(2,3,2,1,16) 
     $  + 2.504933323998402d-16
     $  - 4.872899368701776d-16*tu01

      CsH4YZ(3,0,2,1,01) = CsH4YZ(3,0,2,1,01) 
     $  - 2.199725824145097d-01
     $  + 4.506796579003980d-02*tu01
     $  - 1.290450580325436d-03*tu02
     $  - 2.942794445309582d-05*tu03
     $  + 2.136014739117857d-06*tu04
     $  + 1.226688065023328d-08*tu05
     $  - 2.764815370774007d-09*tu06
     $  + 9.586171373485343d-12*tu07
     $  + 3.105594719889437d-12*tu08
     $  - 3.243751275352189d-14*tu09
     $  - 3.178192073732222d-15*tu10

      CsH4YZ(3,0,2,1,02) = CsH4YZ(3,0,2,1,02) 
     $  - 5.703995207342779d-02
     $  - 7.054595275659064d-03*tu01
     $  + 6.471953361114486d-05*tu02
     $  + 1.035556390404608d-05*tu03
     $  - 1.699845247267622d-07*tu04
     $  - 1.564636140342675d-08*tu05
     $  + 3.383741429751016d-10*tu06
     $  + 2.107690058408739d-11*tu07
     $  - 5.668942582149701d-13*tu08
     $  - 2.563386433009145d-14*tu09
     $  + 8.432014414601681d-16*tu10

      CsH4YZ(3,0,2,1,03) = CsH4YZ(3,0,2,1,03) 
     $  - 2.781228222147158d-03
     $  + 1.028408894053246d-04*tu01
     $  + 2.474155832936472d-05*tu02
     $  - 5.463910015135061d-07*tu03
     $  - 6.288370590195379d-08*tu04
     $  + 1.648058163377465d-09*tu05
     $  + 1.196748060072228d-10*tu06
     $  - 3.717337776409067d-12*tu07
     $  - 1.887811905822976d-13*tu08
     $  + 6.976275857880045d-15*tu09
     $  + 2.592306418716999d-16*tu10

      CsH4YZ(3,0,2,1,04) = CsH4YZ(3,0,2,1,04) 
     $  + 4.526053814052626d-04
     $  + 8.506081014740374d-05*tu01
     $  - 1.586763571508414d-06*tu02
     $  - 2.468662450605818d-07*tu03
     $  + 6.889172554622956d-09*tu04
     $  + 6.064020414718588d-10*tu05
     $  - 2.044367091206949d-11*tu06
     $  - 1.196001644766714d-12*tu07
     $  + 4.782414244837850d-14*tu08
     $  + 1.987329385854981d-15*tu09

      CsH4YZ(3,0,2,1,05) = CsH4YZ(3,0,2,1,05) 
     $  + 4.573332718925202d-05
     $  - 2.626749715067638d-06*tu01
     $  - 6.204493393161241d-07*tu02
     $  + 2.033524125896952d-08*tu03
     $  + 2.260043630401199d-09*tu04
     $  - 8.507098793013698d-11*tu05
     $  - 5.859391951944484d-12*tu06
     $  + 2.559477695507292d-13*tu07
     $  + 1.206631132455866d-14*tu08
     $  - 6.207903565174157d-16*tu09

      CsH4YZ(3,0,2,1,06) = CsH4YZ(3,0,2,1,06) 
     $  - 5.298563050927859d-06
     $  - 1.416019841387887d-06*tu01
     $  + 4.424954354658422d-08*tu02
     $  + 6.671876493265400d-09*tu03
     $  - 2.732204250999883d-10*tu04
     $  - 2.289647480421188d-11*tu05
     $  + 1.085009141785679d-12*tu06
     $  + 5.918241972381937d-14*tu07
     $  - 3.271043996489253d-15*tu08
     $  - 1.233705150398566d-16*tu09

      CsH4YZ(3,0,2,1,07) = CsH4YZ(3,0,2,1,07) 
     $  - 7.046952273108648d-07
     $  + 6.056986043508021d-08*tu01
     $  + 1.397515606056261d-08*tu02
     $  - 6.510500202788806d-10*tu03
     $  - 6.931709417924345d-11*tu04
     $  + 3.606653894684700d-12*tu05
     $  + 2.329661785723346d-13*tu06
     $  - 1.389016882199503d-14*tu07
     $  - 5.979674896938497d-16*tu08

      CsH4YZ(3,0,2,1,08) = CsH4YZ(3,0,2,1,08) 
     $  + 6.759365005017764d-08
     $  + 2.384535439393983d-08*tu01
     $  - 1.130844894416571d-09*tu02
     $  - 1.642102893617303d-10*tu03
     $  + 9.349134834739109d-12*tu04
     $  + 7.385472442780393d-13*tu05
     $  - 4.757742692504475d-14*tu06
     $  - 2.383142963439054d-15*tu07
     $  + 1.786032040571063d-16*tu08

      CsH4YZ(3,0,2,1,09) = CsH4YZ(3,0,2,1,09) 
     $  + 1.033981515391473d-08
     $  - 1.280318672744259d-09*tu01
     $  - 2.868795025244317d-10*tu02
     $  + 1.838512749550543d-11*tu03
     $  + 1.857149534563081d-12*tu04
     $  - 1.304092614867209d-13*tu05
     $  - 7.784678630965965d-15*tu06
     $  + 6.246213675346344d-16*tu07

      CsH4YZ(3,0,2,1,10) = CsH4YZ(3,0,2,1,10) 
     $  - 9.012761751828782d-10
     $  - 3.910027944017708d-10*tu01
     $  + 2.648291000803682d-11*tu02
     $  + 3.673727244015980d-12*tu03
     $  - 2.823846645188488d-13*tu04
     $  - 2.075246685728325d-14*tu05
     $  + 1.789696709238333d-15*tu06

      CsH4YZ(3,0,2,1,11) = CsH4YZ(3,0,2,1,11) 
     $  - 1.467998293785967d-10
     $  + 2.527867326944279d-11*tu01
     $  + 5.463763069950389d-12*tu02
     $  - 4.700740740768691d-13*tu03
     $  - 4.451796410808912d-14*tu04
     $  + 4.159564183686741d-15*tu05
     $  + 2.250868576503743d-16*tu06

      CsH4YZ(3,0,2,1,12) = CsH4YZ(3,0,2,1,12) 
     $  + 1.236552624634069d-11
     $  + 6.210334339814087d-12*tu01
     $  - 5.766509602168771d-13*tu02
     $  - 7.566061187464601d-14*tu03
     $  + 7.715286101289560d-15*tu04
     $  + 5.185048864154814d-16*tu05

      CsH4YZ(3,0,2,1,13) = CsH4YZ(3,0,2,1,13) 
     $  + 2.037065775484562d-12
     $  - 4.728989000169130d-13*tu01
     $  - 9.782097192415088d-14*tu02
     $  + 1.110050323948588d-14*tu03
     $  + 9.709648405701522d-16*tu04
     $  - 1.199842160209790d-16*tu05

      CsH4YZ(3,0,2,1,14) = CsH4YZ(3,0,2,1,14) 
     $  - 1.726668779865371d-13
     $  - 9.565636823156438d-14*tu01
     $  + 1.182946526578400d-14*tu02
     $  + 1.450029815263563d-15*tu03
     $  - 1.943097962169125d-16*tu04

      CsH4YZ(3,0,2,1,15) = CsH4YZ(3,0,2,1,15) 
     $  - 2.779367460924056d-14
     $  + 8.469939799473746d-15*tu01
     $  + 1.661025041964427d-15*tu02
     $  - 2.456764648534167d-16*tu03

      CsH4YZ(3,0,2,1,16) = CsH4YZ(3,0,2,1,16) 
     $  + 2.433159732701680d-15
     $  + 1.431977743924565d-15*tu01
     $  - 2.309965990766241d-16*tu02

      CsH4YZ(3,0,2,1,17) = CsH4YZ(3,0,2,1,17) 
     $  + 3.742194291051212d-16
     $  - 1.463547264715035d-16*tu01

      CsH4YZ(3,1,2,1,01) = CsH4YZ(3,1,2,1,01) 
     $  + 1.261030852329149d-01
     $  - 2.935560827573696d-02*tu01
     $  + 1.121210773821729d-03*tu02
     $  + 7.141485436072964d-06*tu03
     $  - 1.698981017162980d-06*tu04
     $  + 1.987017428070063d-08*tu05
     $  + 1.913706582187447d-09*tu06
     $  - 5.024492795034062d-11*tu07
     $  - 1.718168716712093d-12*tu08
     $  + 7.785514633653744d-14*tu09
     $  + 1.182023909844557d-15*tu10

      CsH4YZ(3,1,2,1,02) = CsH4YZ(3,1,2,1,02) 
     $  + 3.780797467254387d-02
     $  + 3.989511654728157d-03*tu01
     $  - 7.533424213514652d-05*tu02
     $  - 5.254702861238747d-06*tu03
     $  + 1.923864761069694d-07*tu04
     $  + 6.712302611706390d-09*tu05
     $  - 3.693979356256496d-10*tu06
     $  - 6.863841383467530d-12*tu07
     $  + 5.915018663497566d-13*tu08
     $  + 4.883466047256836d-15*tu09
     $  - 8.320521458318316d-16*tu10

      CsH4YZ(3,1,2,1,03) = CsH4YZ(3,1,2,1,03) 
     $  + 2.844939460537531d-03
     $  + 4.724240143609775d-05*tu01
     $  - 1.446835307153640d-05*tu02
     $  + 3.326287475928553d-07*tu03
     $  + 3.308411217662602d-08*tu04
     $  - 1.298236608303544d-09*tu05
     $  - 5.367946599425687d-11*tu06
     $  + 3.111204127152439d-12*tu07
     $  + 6.542729208252646d-14*tu08
     $  - 5.859239475919675d-15*tu09

      CsH4YZ(3,1,2,1,04) = CsH4YZ(3,1,2,1,04) 
     $  - 2.204787248926545d-04
     $  - 5.019454263347941d-05*tu01
     $  + 8.196085257568539d-07*tu02
     $  + 1.386407006319711d-07*tu03
     $  - 4.639272932236299d-09*tu04
     $  - 3.074805446990732d-10*tu05
     $  + 1.506863128422404d-11*tu06
     $  + 5.138447090349691d-13*tu07
     $  - 3.625228051243038d-14*tu08
     $  - 6.439746053869776d-16*tu09

      CsH4YZ(3,1,2,1,05) = CsH4YZ(3,1,2,1,05) 
     $  - 4.723991741268071d-05
     $  - 1.151764660997894d-06*tu01
     $  + 3.830716847494350d-07*tu02
     $  - 8.936388260950035d-09*tu03
     $  - 1.288852423127110d-09*tu04
     $  + 5.212213179106500d-11*tu05
     $  + 2.966205739944165d-12*tu06
     $  - 1.727912442480014d-13*tu07
     $  - 5.067028193624741d-15*tu08
     $  + 4.317362890311838d-16*tu09

      CsH4YZ(3,1,2,1,06) = CsH4YZ(3,1,2,1,06) 
     $  + 1.421088512529125d-06
     $  + 8.785131781981877d-07*tu01
     $  - 1.224390509417771d-08*tu02
     $  - 4.068739314934884d-09*tu03
     $  + 1.386161787104659d-10*tu04
     $  + 1.292153698779566d-11*tu05
     $  - 6.531686779801990d-13*tu06
     $  - 2.937605441677387d-14*tu07
     $  + 2.102126834588395d-15*tu08

      CsH4YZ(3,1,2,1,07) = CsH4YZ(3,1,2,1,07) 
     $  + 7.366142396662339d-07
     $  + 2.525235876474311d-08*tu01
     $  - 9.187898220613220d-09*tu02
     $  + 2.238068491583702d-10*tu03
     $  + 4.301565623916564d-11*tu04
     $  - 1.864991654480581d-12*tu05
     $  - 1.317490116277702d-13*tu06
     $  + 8.167887918472599d-15*tu07
     $  + 2.916884576233356d-16*tu08

      CsH4YZ(3,1,2,1,08) = CsH4YZ(3,1,2,1,08) 
     $  - 1.864281701988772d-09
     $  - 1.565984007552728d-08*tu01
     $  + 1.877456584392548d-10*tu02
     $  + 1.088063302301853d-10*tu03
     $  - 3.967851069826056d-12*tu04
     $  - 4.614907408389199d-13*tu05
     $  + 2.525182406336548d-14*tu06
     $  + 1.344867693890754d-15*tu07
     $  - 1.035722445667870d-16*tu08

      CsH4YZ(3,1,2,1,09) = CsH4YZ(3,1,2,1,09) 
     $  - 1.095584261742589d-08
     $  - 5.061692503450027d-10*tu01
     $  + 2.025246640090794d-10*tu02
     $  - 5.315798521312208d-12*tu03
     $  - 1.263441102702909d-12*tu04
     $  + 6.032004283697377d-14*tu05
     $  + 4.937055233411141d-15*tu06
     $  - 3.361747550286291d-16*tu07

      CsH4YZ(3,1,2,1,10) = CsH4YZ(3,1,2,1,10) 
     $  - 1.879982420169719d-10
     $  + 2.735289161456816d-10*tu01
     $  - 2.934496802664654d-12*tu02
     $  - 2.661289587871552d-12*tu03
     $  + 1.071477544691570d-13*tu04
     $  + 1.445540563888094d-14*tu05
     $  - 8.780834299861081d-16*tu06

      CsH4YZ(3,1,2,1,11) = CsH4YZ(3,1,2,1,11) 
     $  + 1.578436983574210d-10
     $  + 9.459878216060535d-12*tu01
     $  - 4.177463884411066d-12*tu02
     $  + 1.217098212960167d-13*tu03
     $  + 3.354894814283926d-14*tu04
     $  - 1.796187566262762d-15*tu05
     $  - 1.620919957264831d-16*tu06

      CsH4YZ(3,1,2,1,12) = CsH4YZ(3,1,2,1,12) 
     $  + 5.202626055524607d-12
     $  - 4.656103045417717d-12*tu01
     $  + 4.823471905163935d-14*tu02
     $  + 6.047522484741615d-14*tu03
     $  - 2.744067901650856d-15*tu04
     $  - 4.080038168414261d-16*tu05

      CsH4YZ(3,1,2,1,13) = CsH4YZ(3,1,2,1,13) 
     $  - 2.223650296346142d-12
     $  - 1.673235348141138d-13*tu01
     $  + 8.175138079531641d-14*tu02
     $  - 2.709236815360397d-15*tu03
     $  - 8.217717251921495d-16*tu04

      CsH4YZ(3,1,2,1,14) = CsH4YZ(3,1,2,1,14) 
     $  - 1.036244022950701d-13
     $  + 7.733666251284874d-14*tu01
     $  - 8.547881823759483d-16*tu02
     $  - 1.294126531377693d-15*tu03

      CsH4YZ(3,1,2,1,15) = CsH4YZ(3,1,2,1,15) 
     $  + 3.079408158046738d-14
     $  + 2.831581478069942d-15*tu01
     $  - 1.533122031159268d-15*tu02

      CsH4YZ(3,1,2,1,16) = CsH4YZ(3,1,2,1,16) 
     $  + 1.816853376054296d-15
     $  - 1.256637435870418d-15*tu01

      CsH4YZ(3,1,2,1,17) = CsH4YZ(3,1,2,1,17) 
     $  - 4.204438756340348d-16

      CsH4YZ(3,2,0,1,01) = CsH4YZ(3,2,0,1,01) 
     $  - 9.277829968057344d-01
     $  + 2.852544300966231d-01*tu01
     $  - 1.987711772106606d-02*tu02
     $  + 7.707323786098988d-04*tu03
     $  - 1.526723741283461d-05*tu04
     $  + 1.156159152867487d-08*tu05
     $  + 5.649487112708124d-09*tu06
     $  - 1.928582475116397d-11*tu07
     $  - 3.878367154157066d-12*tu08
     $  + 2.951032851077271d-14*tu09
     $  + 2.984237394547123d-15*tu10

      CsH4YZ(3,2,0,1,02) = CsH4YZ(3,2,0,1,02) 
     $  - 3.899629739348479d-01
     $  - 1.731248688898612d-02*tu01
     $  + 7.079557413837057d-04*tu02
     $  + 3.933025354006358d-07*tu03
     $  - 9.141287197268237d-07*tu04
     $  + 1.533145206854596d-08*tu05
     $  + 8.881898248366597d-10*tu06
     $  - 2.919476801769561d-11*tu07
     $  - 6.921692532704932d-13*tu08
     $  + 3.941642865360639d-14*tu09
     $  + 4.021675257200717d-16*tu10

      CsH4YZ(3,2,0,1,03) = CsH4YZ(3,2,0,1,03) 
     $  - 6.760729467241214d-02
     $  - 1.336225505466387d-03*tu01
     $  + 1.159912705498864d-04*tu02
     $  - 3.595739596819166d-06*tu03
     $  - 8.068539132982819d-08*tu04
     $  + 8.217123832197713d-09*tu05
     $  - 4.817153011353016d-11*tu06
     $  - 1.137990183595021d-11*tu07
     $  + 2.408958872642249d-13*tu08
     $  + 1.225979145595431d-14*tu09
     $  - 4.601157250004655d-16*tu10

      CsH4YZ(3,2,0,1,04) = CsH4YZ(3,2,0,1,04) 
     $  - 7.060339153556292d-03
     $  + 8.470644776711373d-05*tu01
     $  + 3.187310943607975d-06*tu02
     $  - 7.196117122905235d-07*tu03
     $  + 2.628136513807527d-08*tu04
     $  + 8.079956802547693d-10*tu05
     $  - 7.596691301932372d-11*tu06
     $  + 1.172881618416542d-13*tu07
     $  + 1.301817639077824d-13*tu08
     $  - 2.245798592209586d-15*tu09
     $  - 1.708479525405052d-16*tu10

      CsH4YZ(3,2,0,1,05) = CsH4YZ(3,2,0,1,05) 
     $  - 3.648874569219032d-04
     $  + 1.942841184479618d-05*tu01
     $  - 1.634291220746264d-06*tu02
     $  + 1.007506859050707d-08*tu03
     $  + 5.932276531425246d-09*tu04
     $  - 2.580590432676791d-10*tu05
     $  - 7.709390126319079d-12*tu06
     $  + 7.822924556674425d-13*tu07
     $  - 1.509859532579108d-15*tu09

      CsH4YZ(3,2,0,1,06) = CsH4YZ(3,2,0,1,06) 
     $  + 4.467179398054359d-06
     $  - 4.995750619816842d-07*tu01
     $  - 1.423844640966288d-07*tu02
     $  + 1.759377948547793d-08*tu03
     $  - 3.145859056372787d-10*tu04
     $  - 5.420625775619458d-11*tu05
     $  + 2.799899900364339d-12*tu06
     $  + 7.344928273117295d-14*tu07
     $  - 8.582712921898413d-15*tu08

      CsH4YZ(3,2,0,1,07) = CsH4YZ(3,2,0,1,07) 
     $  + 1.317046898072150d-06
     $  - 2.727652118552393d-07*tu01
     $  + 2.229753070063253d-08*tu02
     $  + 7.130265332533371d-10*tu03
     $  - 1.857519791281629d-10*tu04
     $  + 5.198097453407761d-12*tu05
     $  + 5.194978618337393d-13*tu06
     $  - 3.221605794587313d-14*tu07
     $  - 6.878683892743355d-16*tu08

      CsH4YZ(3,2,0,1,08) = CsH4YZ(3,2,0,1,08) 
     $  - 4.686313693984350d-08
     $  + 1.917366836108189d-09*tu01
     $  + 3.110046652722197d-09*tu02
     $  - 3.489624956248934d-10*tu03
     $  - 3.216648116142443d-13*tu04
     $  + 1.972680675036576d-12*tu05
     $  - 7.494132924969691d-14*tu06
     $  - 5.082694853953426d-15*tu07
     $  + 3.832980602650022d-16*tu08

      CsH4YZ(3,2,0,1,09) = CsH4YZ(3,2,0,1,09) 
     $  - 1.007208824971394d-08
     $  + 3.762578909128106d-09*tu01
     $  - 3.037380926281004d-10*tu02
     $  - 2.542240186079176d-11*tu03
     $  + 4.604615142918588d-12*tu04
     $  - 7.343519404192997d-14*tu05
     $  - 2.100373599728456d-14*tu06
     $  + 1.028163212558223d-15*tu07

      CsH4YZ(3,2,0,1,10) = CsH4YZ(3,2,0,1,10) 
     $  + 5.209005671446503d-10
     $  + 1.041426423126336d-11*tu01
     $  - 5.719322506594313d-11*tu02
     $  + 6.406736175120714d-12*tu03
     $  + 1.358385780596087d-13*tu04
     $  - 5.684743931092405d-14*tu05
     $  + 1.659848948873926d-15*tu06
     $  + 2.226907342928835d-16*tu07

      CsH4YZ(3,2,0,1,11) = CsH4YZ(3,2,0,1,11) 
     $  + 9.511020395984656d-11
     $  - 5.165970688116321d-11*tu01
     $  + 4.244976391176177d-12*tu02
     $  + 6.123925906769650d-13*tu03
     $  - 1.020543863328420d-13*tu04
     $  + 5.149422357841360d-16*tu05
     $  + 6.756901456101606d-16*tu06

      CsH4YZ(3,2,0,1,12) = CsH4YZ(3,2,0,1,12) 
     $  - 6.041282866063956d-12
     $  - 3.708393327466954d-13*tu01
     $  + 9.728532882515789d-13*tu02
     $  - 1.138103980005352d-13*tu03
     $  - 4.726805374294948d-15*tu04
     $  + 1.441140782594631d-15*tu05

      CsH4YZ(3,2,0,1,13) = CsH4YZ(3,2,0,1,13) 
     $  - 9.790051227347082d-13
     $  + 7.093257441303905d-13*tu01
     $  - 6.174683651766772d-14*tu02
     $  - 1.262077451668067d-14*tu03
     $  + 2.120673429718873d-15*tu04

      CsH4YZ(3,2,0,1,14) = CsH4YZ(3,2,0,1,14) 
     $  + 7.217284680093552d-14
     $  + 5.648372620043285d-15*tu01
     $  - 1.581826681838508d-14*tu02
     $  + 1.992807167371420d-15*tu03
     $  + 1.196066089734382d-16*tu04

      CsH4YZ(3,2,0,1,15) = CsH4YZ(3,2,0,1,15) 
     $  + 1.051215348299913d-14
     $  - 9.743101040610472d-15*tu01
     $  + 9.363173684773688d-16*tu02
     $  + 2.381036910744371d-16*tu03

      CsH4YZ(3,2,0,1,16) = CsH4YZ(3,2,0,1,16) 
     $  - 8.784534368866572d-16
     $  + 2.494154108287365d-16*tu02

      CsH4YZ(3,2,0,1,17) = CsH4YZ(3,2,0,1,17) 
     $  - 1.155424894621144d-16
     $  + 1.336885968983981d-16*tu01

      CsH4YZ(3,2,1,1,01) = CsH4YZ(3,2,1,1,01) 
     $  + 5.519791121767504d-01
     $  - 1.831402382280320d-01*tu01
     $  + 1.421839306380919d-02*tu02
     $  - 6.453306760967352d-04*tu03
     $  + 1.679103090961561d-05*tu04
     $  - 1.404531544719457d-07*tu05
     $  - 5.490360712844313d-09*tu06
     $  + 1.410816434559276d-10*tu07
     $  + 2.554615225954279d-12*tu08
     $  - 1.362378531202994d-13*tu09
     $  - 8.446058799329203d-16*tu10
     $  + 1.244722316679652d-16*tu11

      CsH4YZ(3,2,1,1,02) = CsH4YZ(3,2,1,1,02) 
     $  + 2.540809182850166d-01
     $  + 7.566737406146581d-03*tu01
     $  - 4.140972296718717d-04*tu02
     $  + 6.708872081398627d-06*tu03
     $  + 3.574085510339904d-07*tu04
     $  - 1.791900461720257d-08*tu05
     $  - 8.033011171174320d-11*tu06
     $  + 2.406596235760500d-11*tu07
     $  - 3.012774972489441d-13*tu08
     $  - 2.428294064337410d-14*tu09
     $  + 6.819616541975611d-16*tu10

      CsH4YZ(3,2,1,1,03) = CsH4YZ(3,2,1,1,03) 
     $  + 5.012269567580497d-02
     $  + 8.991407067929753d-04*tu01
     $  - 6.759787185463921d-05*tu02
     $  + 2.326830150591861d-06*tu03
     $  + 1.134170337952992d-08*tu04
     $  - 4.125609714146611d-09*tu05
     $  + 1.028517394871058d-10*tu06
     $  + 3.610607949325748d-12*tu07
     $  - 2.286232520154228d-13*tu08
     $  - 6.319097642506705d-16*tu09
     $  + 3.234354690793395d-16*tu10

      CsH4YZ(3,2,1,1,04) = CsH4YZ(3,2,1,1,04) 
     $  + 6.282794529831127d-03
     $  + 7.272490408553873d-06*tu01
     $  - 3.778240252203175d-06*tu02
     $  + 3.681997120994451d-07*tu03
     $  - 1.353268612253906d-08*tu04
     $  - 2.022354393274812d-10*tu05
     $  + 3.503815029782369d-11*tu06
     $  - 6.880337849028700d-13*tu07
     $  - 4.414004822372526d-14*tu08
     $  + 2.162643071802337d-15*tu09

      CsH4YZ(3,2,1,1,05) = CsH4YZ(3,2,1,1,05) 
     $  + 4.458486964626628d-04
     $  - 1.102745272694798d-05*tu01
     $  + 6.389433520050618d-07*tu02
     $  + 4.955488085448205d-09*tu03
     $  - 2.694367833057702d-09*tu04
     $  + 1.206962429769276d-10*tu05
     $  + 1.886768606488653d-12*tu06
     $  - 3.372141563376588d-13*tu07
     $  + 5.845831864362911d-15*tu08
     $  + 5.112003526624617d-16*tu09

      CsH4YZ(3,2,1,1,06) = CsH4YZ(3,2,1,1,06) 
     $  + 4.650933789238394d-06
     $  - 7.538101189414301d-07*tu01
     $  + 1.152056444834797d-07*tu02
     $  - 7.517027465008906d-09*tu03
     $  + 5.406434436065667d-11*tu04
     $  + 2.313039177654620d-11*tu05
     $  - 1.196840754115016d-12*tu06
     $  - 1.713626256550025d-14*tu07
     $  + 3.462481095095869d-15*tu08

      CsH4YZ(3,2,1,1,07) = CsH4YZ(3,2,1,1,07) 
     $  - 1.960240224238020d-06
     $  + 1.231486548960522d-07*tu01
     $  - 3.020262029910788d-09*tu02
     $  - 7.760618440150173d-10*tu03
     $  + 7.841808350259907d-11*tu04
     $  - 1.421105876971631d-12*tu05
     $  - 2.099451991966842d-13*tu06
     $  + 1.288651867811482d-14*tu07
     $  + 1.439135149052918d-16*tu08

      CsH4YZ(3,2,1,1,08) = CsH4YZ(3,2,1,1,08) 
     $  - 6.643981113847728d-08
     $  + 1.624124611678536d-08*tu01
     $  - 2.166893795790615d-09*tu02
     $  + 1.048095624321847d-10*tu03
     $  + 4.868365659207830d-12*tu04
     $  - 8.264214952063285d-13*tu05
     $  + 2.267447801804280d-14*tu06
     $  + 1.977652237421136d-15*tu07
     $  - 1.453829212841816d-16*tu08

      CsH4YZ(3,2,1,1,09) = CsH4YZ(3,2,1,1,09) 
     $  + 1.602250348024761d-08
     $  - 1.255897854350691d-09*tu01
     $  - 5.269651415779466d-11*tu02
     $  + 2.249746227517441d-11*tu03
     $  - 1.621624337458259d-12*tu04
     $  - 1.898487913642601d-14*tu05
     $  + 8.733068457618187d-15*tu06
     $  - 3.230079328400582d-16*tu07

      CsH4YZ(3,2,1,1,10) = CsH4YZ(3,2,1,1,10) 
     $  + 9.140096262850323d-10
     $  - 2.794084627199096d-10*tu01
     $  + 3.525191430638141d-11*tu02
     $  - 1.029258571053037d-12*tu03
     $  - 2.064043487456482d-13*tu04
     $  + 2.157391604963591d-14*tu05
     $  - 1.384085881452467d-16*tu06

      CsH4YZ(3,2,1,1,11) = CsH4YZ(3,2,1,1,11) 
     $  - 1.551571126528103d-10
     $  + 1.149318449740586d-11*tu01
     $  + 2.137264429187892d-12*tu02
     $  - 4.954180913642716d-13*tu03
     $  + 2.786578005182607d-14*tu04
     $  + 1.615461341562365d-15*tu05
     $  - 2.683642245255989d-16*tu06

      CsH4YZ(3,2,1,1,12) = CsH4YZ(3,2,1,1,12) 
     $  - 1.231320397447126d-11
     $  + 4.405356022148912d-12*tu01
     $  - 5.369176818322591d-13*tu02
     $  + 2.618076953135338d-15*tu03
     $  + 5.763268866691254d-15*tu04
     $  - 4.780406884902810d-16*tu05

      CsH4YZ(3,2,1,1,13) = CsH4YZ(3,2,1,1,13) 
     $  + 1.626790942299511d-12
     $  - 8.681644533842856d-14*tu01
     $  - 4.987989581441160d-14*tu02
     $  + 9.681825554591079d-15*tu03
     $  - 4.137106253970753d-16*tu04

      CsH4YZ(3,2,1,1,14) = CsH4YZ(3,2,1,1,14) 
     $  + 1.641500236169173d-13
     $  - 6.661152046916308d-14*tu01
     $  + 7.914477968214279d-15*tu02
     $  + 2.030578713853421d-16*tu03
     $  - 1.357974538396067d-16*tu04

      CsH4YZ(3,2,1,1,15) = CsH4YZ(3,2,1,1,15) 
     $  - 1.790180288296836d-14
     $  + 3.437701273795653d-16*tu01
     $  + 9.798130619042722d-16*tu02
     $  - 1.773275427399195d-16*tu03

      CsH4YZ(3,2,1,1,16) = CsH4YZ(3,2,1,1,16) 
     $  - 2.178190471368679d-15
     $  + 9.850836066453443d-16*tu01
     $  - 1.149145755306796d-16*tu02

      CsH4YZ(3,2,1,1,17) = CsH4YZ(3,2,1,1,17) 
     $  + 2.037997093526042d-16

      CsH4YZ(3,2,2,1,01) = CsH4YZ(3,2,2,1,01) 
     $  - 5.081817573100070d-01
     $  + 2.190983187974331d-01*tu01
     $  - 2.148761089003609d-02*tu02
     $  + 1.194687594133402d-03*tu03
     $  - 3.724187005513009d-05*tu04
     $  + 3.975309912369289d-07*tu05
     $  + 1.358641039659882d-08*tu06
     $  - 4.135452769901042d-10*tu07
     $  - 6.891073243595210d-12*tu08
     $  + 4.116407436036476d-13*tu09
     $  + 2.727772318983860d-15*tu10
     $  - 3.897333008074184d-16*tu11

      CsH4YZ(3,2,2,1,02) = CsH4YZ(3,2,2,1,02) 
     $  - 3.152398288467741d-01
     $  + 1.666869946205442d-03*tu01
     $  + 4.696164140207647d-04*tu02
     $  - 3.087548268364259d-05*tu03
     $  + 1.188759135161464d-08*tu04
     $  + 5.407192318648852d-08*tu05
     $  - 9.855736326765317d-10*tu06
     $  - 6.408001963920680d-11*tu07
     $  + 2.162896003232295d-12*tu08
     $  + 5.906625760660736d-14*tu09
     $  - 3.291253929772923d-15*tu10

      CsH4YZ(3,2,2,1,03) = CsH4YZ(3,2,2,1,03) 
     $  - 7.997956413285563d-02
     $  - 1.464037504317862d-03*tu01
     $  + 6.206078668466650d-05*tu02
     $  - 2.838359853442911d-06*tu03
     $  + 1.095439870931922d-07*tu04
     $  + 2.511647146864654d-09*tu05
     $  - 3.800584943550103d-10*tu06
     $  + 4.265415764382036d-12*tu07
     $  + 6.337550593643413d-13*tu08
     $  - 1.734033802108234d-14*tu09
     $  - 7.694213241092529d-16*tu10

      CsH4YZ(3,2,2,1,04) = CsH4YZ(3,2,2,1,04) 
     $  - 1.224064981734235d-02
     $  - 1.915847291401790d-04*tu01
     $  + 9.734173384509965d-06*tu02
     $  - 1.381211789472302d-07*tu03
     $  + 7.427520674918976d-09*tu04
     $  - 8.076914037441845d-10*tu05
     $  - 8.703866962077864d-12*tu06
     $  + 3.142814837633670d-12*tu07
     $  - 4.558576168498519d-14*tu08
     $  - 6.229460009052408d-15*tu09
     $  + 1.841347698751821d-16*tu10

      CsH4YZ(3,2,2,1,05) = CsH4YZ(3,2,2,1,05) 
     $  - 1.018391959655337d-03
     $  + 1.628069605111540d-05*tu01
     $  + 3.379105040472035d-07*tu02
     $  - 4.527677410582120d-08*tu03
     $  - 1.092166626533383d-10*tu04
     $  - 1.895869773296674d-11*tu05
     $  + 7.908476258797603d-12*tu06
     $  - 2.659775620519527d-14*tu07
     $  - 2.952821673799212d-14*tu08
     $  + 5.970488311091388d-16*tu09

      CsH4YZ(3,2,2,1,06) = CsH4YZ(3,2,2,1,06) 
     $  - 1.312141771484168d-05
     $  + 4.177025260193862d-06*tu01
     $  - 2.474820129003945d-07*tu02
     $  - 1.714573583716617d-09*tu03
     $  + 4.256132976552970d-10*tu04
     $  + 7.711557230767845d-12*tu05
     $  - 2.943867798049679d-13*tu06
     $  - 8.122291475849154d-14*tu07
     $  + 1.466313633100057d-15*tu08
     $  + 2.923896953440555d-16*tu09

      CsH4YZ(3,2,2,1,07) = CsH4YZ(3,2,2,1,07) 
     $  + 5.268369001952776d-06
     $  - 1.572506132679975d-07*tu01
     $  - 2.721190067146284d-08*tu02
     $  + 2.266209019070538d-09*tu03
     $  + 1.981482279956818d-11*tu04
     $  - 4.541433717204536d-12*tu05
     $  - 1.226284428321048d-13*tu06
     $  + 7.446670724839283d-15*tu07
     $  + 8.540953300193366d-16*tu08

      CsH4YZ(3,2,2,1,08) = CsH4YZ(3,2,2,1,08) 
     $  + 1.785889203050977d-07
     $  - 7.448645958040409d-08*tu01
     $  + 4.521479629066233d-09*tu02
     $  + 1.987440976900800d-10*tu03
     $  - 2.363256700074069d-11*tu04
     $  - 2.319634336905209d-13*tu05
     $  + 5.449532212371533d-14*tu06
     $  + 1.532881106840555d-15*tu07
     $  - 1.257524996138215d-16*tu08

      CsH4YZ(3,2,2,1,09) = CsH4YZ(3,2,2,1,09) 
     $  - 4.833523734494348d-08
     $  + 1.285433326668506d-09*tu01
     $  + 7.487071376261555d-10*tu02
     $  - 6.045511246005236d-11*tu03
     $  - 1.596891664978813d-12*tu04
     $  + 2.607990930178303d-13*tu05
     $  + 2.589209300137905d-15*tu06
     $  - 6.837607591242093d-16*tu07

      CsH4YZ(3,2,2,1,10) = CsH4YZ(3,2,2,1,10) 
     $  - 2.361653270816863d-09
     $  + 1.218813918330246d-09*tu01
     $  - 7.504542786122866d-11*tu02
     $  - 6.992687705970239d-12*tu03
     $  + 7.657591327014716d-13*tu04
     $  + 1.287815802275625d-14*tu05
     $  - 3.034319891628101d-15*tu06

      CsH4YZ(3,2,2,1,11) = CsH4YZ(3,2,2,1,11) 
     $  + 5.171609633771731d-10
     $  - 8.098519880512515d-12*tu01
     $  - 1.612053274870332d-11*tu02
     $  + 1.343935757482687d-12*tu03
     $  + 6.297501027855547d-14*tu04
     $  - 9.520890448394645d-15*tu05

      CsH4YZ(3,2,2,1,12) = CsH4YZ(3,2,2,1,12) 
     $  + 3.061536854315748d-11
     $  - 1.905556307347521d-11*tu01
     $  + 1.214550717495686d-12*tu02
     $  + 1.816167790420794d-13*tu03
     $  - 2.027828730843628d-14*tu04
     $  - 5.255235916147325d-16*tu05
     $  + 1.185264028777155d-16*tu06

      CsH4YZ(3,2,2,1,13) = CsH4YZ(3,2,2,1,13) 
     $  - 5.938433479648387d-12
     $  + 2.402249190180826d-14*tu01
     $  + 3.092074958850534d-13*tu02
     $  - 2.752520538657844d-14*tu03
     $  - 1.850506394097799d-15*tu04
     $  + 2.839103174034079d-16*tu05

      CsH4YZ(3,2,2,1,14) = CsH4YZ(3,2,2,1,14) 
     $  - 3.919560396539304d-13
     $  + 2.898425570403281d-13*tu01
     $  - 1.975463233376934d-14*tu02
     $  - 4.052010582671422d-15*tu03
     $  + 4.820955585846166d-16*tu04

      CsH4YZ(3,2,2,1,15) = CsH4YZ(3,2,2,1,15) 
     $  + 7.098692980997300d-14
     $  + 2.418509248719357d-16*tu01
     $  - 5.536781540450149d-15*tu02
     $  + 5.395554440740667d-16*tu03

      CsH4YZ(3,2,2,1,16) = CsH4YZ(3,2,2,1,16) 
     $  + 4.979865136835140d-15
     $  - 4.326556810258433d-15*tu01
     $  + 3.263249443370151d-16*tu02

      CsH4YZ(3,2,2,1,17) = CsH4YZ(3,2,2,1,17) 
     $  - 8.700760483816719d-16

      CsH4YZ(3,2,3,1,01) = CsH4YZ(3,2,3,1,01) 
     $  - 3.264836604527753d-01
     $  + 9.098625564241893d-02*tu01
     $  - 5.270800003187597d-03*tu02
     $  + 1.307538588943996d-04*tu03
     $  + 8.480363785652211d-07*tu04
     $  - 1.252247178816988d-07*tu05
     $  + 1.032003725221262d-09*tu06
     $  + 1.055031026158454d-10*tu07
     $  - 2.233379457932485d-12*tu08
     $  - 7.669638348337453d-14*tu09
     $  + 2.910467104735362d-15*tu10

      CsH4YZ(3,2,3,1,02) = CsH4YZ(3,2,3,1,02) 
     $  - 1.216402065117825d-01
     $  - 8.148746026166077d-03*tu01
     $  + 2.672828386223765d-04*tu02
     $  + 4.846537610156824d-06*tu03
     $  - 4.852675905583420d-07*tu04
     $  + 5.584844512827592d-10*tu05
     $  + 6.635005581295158d-10*tu06
     $  - 8.775619396120549d-12*tu07
     $  - 7.506672670538974d-13*tu08
     $  + 1.845857256889314d-14*tu09
     $  + 7.262988923510231d-16*tu10

      CsH4YZ(3,2,3,1,03) = CsH4YZ(3,2,3,1,03) 
     $  - 1.655782957100109d-02
     $  - 4.284543601694562d-04*tu01
     $  + 4.291959281690082d-05*tu02
     $  - 1.143877731307203d-06*tu03
     $  - 5.594003442243236d-08*tu04
     $  + 3.605238558102647d-09*tu05
     $  + 3.105357568663122d-11*tu06
     $  - 6.564799879433114d-12*tu07
     $  + 4.457549891173628d-14*tu08
     $  + 9.268760007058660d-15*tu09
     $  - 1.703717266212323d-16*tu10

      CsH4YZ(3,2,3,1,04) = CsH4YZ(3,2,3,1,04) 
     $  - 8.848518118803216d-04
     $  + 7.124407864657772d-05*tu01
     $  - 2.383747714795038d-07*tu02
     $  - 2.956548145424669d-07*tu03
     $  + 1.093652103008422d-08*tu04
     $  + 4.779141939279220d-10*tu05
     $  - 3.508190848459871d-11*tu06
     $  - 3.426960193286640d-13*tu07
     $  + 7.171774724628113d-14*tu08
     $  - 4.023617523130575d-16*tu09
     $  - 1.145849065374780d-16*tu10

      CsH4YZ(3,2,3,1,05) = CsH4YZ(3,2,3,1,05) 
     $  + 6.362321813246216d-05
     $  + 8.176036907197210d-06*tu01
     $  - 8.459824565400605d-07*tu02
     $  + 1.124899936295833d-08*tu03
     $  + 2.725575181677253d-09*tu04
     $  - 1.168602049014917d-10*tu05
     $  - 4.592963505286999d-12*tu06
     $  + 3.772952564785448d-13*tu07
     $  + 3.522109916972395d-15*tu08
     $  - 8.266494678456494d-16*tu09

      CsH4YZ(3,2,3,1,06) = CsH4YZ(3,2,3,1,06) 
     $  + 1.015512297057613d-05
     $  - 8.789805601860726d-07*tu01
     $  - 3.380008971242650d-08*tu02
     $  + 8.455897290790503d-09*tu03
     $  - 2.027256244324524d-10*tu04
     $  - 2.588692419925859d-11*tu05
     $  + 1.342662160908872d-12*tu06
     $  + 4.405965180124763d-14*tu07
     $  - 4.285079734127786d-15*tu08

      CsH4YZ(3,2,3,1,07) = CsH4YZ(3,2,3,1,07) 
     $  - 3.439023216359090d-07
     $  - 1.443460879589328d-07*tu01
     $  + 1.559085875599192d-08*tu02
     $  + 4.206183859420114d-11*tu03
     $  - 9.030689748452932d-11*tu04
     $  + 3.007795046754511d-12*tu05
     $  + 2.565478103779445d-13*tu06
     $  - 1.605438085350049d-14*tu07
     $  - 4.217040007145269d-16*tu08

      CsH4YZ(3,2,3,1,08) = CsH4YZ(3,2,3,1,08) 
     $  - 1.218468563069893d-07
     $  + 1.128601839367150d-08*tu01
     $  + 1.072322174998144d-09*tu02
     $  - 2.005312525738013d-10*tu03
     $  + 2.791404585361763d-12*tu04
     $  + 9.633343808423274d-13*tu05
     $  - 4.188252223010062d-14*tu06
     $  - 2.562477477586306d-15*tu07
     $  + 1.961580535258963d-16*tu08

      CsH4YZ(3,2,3,1,09) = CsH4YZ(3,2,3,1,09) 
     $  + 1.107435637194070d-09
     $  + 2.394003290548718d-09*tu01
     $  - 2.730221999676070d-10*tu02
     $  - 5.635405613950365d-12*tu03
     $  + 2.481500435297507d-12*tu04
     $  - 6.651186211926804d-14*tu05
     $  - 1.031657395144213d-14*tu06
     $  + 5.672653702661477d-16*tu07

      CsH4YZ(3,2,3,1,10) = CsH4YZ(3,2,3,1,10) 
     $  + 1.468778799221939d-09
     $  - 1.457179637751511d-10*tu01
     $  - 2.446129170939831d-11*tu02
     $  + 4.298138217048605d-12*tu03
     $  - 2.087668154824036d-14*tu04
     $  - 2.956902034655432d-14*tu05
     $  + 1.150124153087202d-15*tu06
     $  + 1.095902400430333d-16*tu07

      CsH4YZ(3,2,3,1,11) = CsH4YZ(3,2,3,1,11) 
     $  + 1.468423431388798d-11
     $  - 3.792725810970967d-11*tu01
     $  + 4.620959098826826d-12*tu02
     $  + 1.827623494535203d-13*tu03
     $  - 6.079402146236506d-14*tu04
     $  + 1.333566685160456d-15*tu05
     $  + 3.437888731947343d-16*tu06

      CsH4YZ(3,2,3,1,12) = CsH4YZ(3,2,3,1,12) 
     $  - 1.761824549621730d-11
     $  + 1.886257789921783d-12*tu01
     $  + 4.847716109748084d-13*tu02
     $  - 8.634737189796970d-14*tu03
     $  - 2.809049498132572d-16*tu04
     $  + 8.021967525537223d-16*tu05

      CsH4YZ(3,2,3,1,13) = CsH4YZ(3,2,3,1,13) 
     $  - 4.648297885557433d-13
     $  + 5.805374224120937d-13*tu01
     $  - 7.644126538842720d-14*tu02
     $  - 4.457618474847300d-15*tu03
     $  + 1.376573220158692d-15*tu04

      CsH4YZ(3,2,3,1,14) = CsH4YZ(3,2,3,1,14) 
     $  + 2.094400677945609d-13
     $  - 2.454865950641471d-14*tu01
     $  - 8.842323664461787d-15*tu02
     $  + 1.657439208368570d-15*tu03

      CsH4YZ(3,2,3,1,15) = CsH4YZ(3,2,3,1,15) 
     $  + 8.714511203335929d-15
     $  - 8.649256858451025d-15*tu01
     $  + 1.244307578475668d-15*tu02

      CsH4YZ(3,2,3,1,16) = CsH4YZ(3,2,3,1,16) 
     $  - 2.459672777194201d-15
     $  + 3.223582400129837d-16*tu01
     $  + 1.523625914593801d-16*tu02

      CsH4YZ(3,2,3,1,17) = CsH4YZ(3,2,3,1,17) 
     $  - 1.394687594665408d-16
     $  + 1.260406013649129d-16*tu01

      CsH4YZ(3,3,2,1,01) = CsH4YZ(3,3,2,1,01) 
     $  - 1.103958224353500d+00
     $  + 3.662804764560640d-01*tu01
     $  - 2.843678612761839d-02*tu02
     $  + 1.290661352193470d-03*tu03
     $  - 3.358206181923123d-05*tu04
     $  + 2.809063089438914d-07*tu05
     $  + 1.098072142568862d-08*tu06
     $  - 2.821632869118552d-10*tu07
     $  - 5.109230451908559d-12*tu08
     $  + 2.724757062405989d-13*tu09
     $  + 1.689211759865840d-15*tu10
     $  - 2.489444633359305d-16*tu11

      CsH4YZ(3,3,2,1,02) = CsH4YZ(3,3,2,1,02) 
     $  - 6.900398674926873d-02
     $  - 6.457457206518869d-02*tu01
     $  + 4.287483845103540d-03*tu02
     $  - 1.409712758867131d-04*tu03
     $  + 1.045150300222853d-06*tu04
     $  + 8.376247622760855d-08*tu05
     $  - 1.892363350899505d-09*tu06
     $  - 6.492290608176073d-11*tu07
     $  + 2.750823270066084d-12*tu08
     $  + 4.118283239256611d-14*tu09
     $  - 3.417544666826156d-15*tu10

      CsH4YZ(3,3,2,1,03) = CsH4YZ(3,3,2,1,03) 
     $  + 3.494914648458278d-02
     $  + 4.206201944887735d-03*tu01
     $  - 1.477466297126942d-04*tu02
     $  - 3.389769129621789d-06*tu03
     $  + 3.690488221079987d-07*tu04
     $  - 2.188073383947727d-09*tu05
     $  - 5.701218760938333d-10*tu06
     $  + 1.322425706769031d-11*tu07
     $  + 6.447071808567165d-13*tu08
     $  - 2.786270285207550d-14*tu09
     $  - 5.319945492901031d-16*tu10

      CsH4YZ(3,3,2,1,04) = CsH4YZ(3,3,2,1,04) 
     $  + 6.823516297487944d-03
     $  + 2.058570904339669d-04*tu01
     $  - 2.497737556889801d-05*tu02
     $  + 9.173076549192457d-07*tu03
     $  + 2.922599238369895d-08*tu04
     $  - 3.113206081723019d-09*tu05
     $  + 1.782088061154055d-11*tu06
     $  + 5.716776511237621d-12*tu07
     $  - 1.405287197135242d-13*tu08
     $  - 7.465377882175450d-15*tu09
     $  + 3.407451956488989d-16*tu10

      CsH4YZ(3,3,2,1,05) = CsH4YZ(3,3,2,1,05) 
     $  + 5.358946449354916d-04
     $  - 3.210116132362447d-05*tu01
     $  + 3.220415056818347d-07*tu02
     $  + 1.689423032488700d-07*tu03
     $  - 9.472923901814717d-09*tu04
     $  - 1.847380115155641d-10*tu05
     $  + 3.090704105990408d-11*tu06
     $  - 3.171992572469964d-13*tu07
     $  - 6.072058880047816d-14*tu08
     $  + 1.768670079015443d-15*tu09

      CsH4YZ(3,3,2,1,06) = CsH4YZ(3,3,2,1,06) 
     $  - 3.302314695692972d-06
     $  - 2.866218095876031d-06*tu01
     $  + 4.239044542265679d-07*tu02
     $  - 1.364216836161613d-08*tu03
     $  - 1.344246194246456d-09*tu04
     $  + 1.070297521522977d-10*tu05
     $  + 9.185258553123845d-13*tu06
     $  - 3.300558900395688d-13*tu07
     $  + 5.282470168670464d-15*tu08
     $  + 6.666256415464267d-16*tu09

      CsH4YZ(3,3,2,1,07) = CsH4YZ(3,3,2,1,07) 
     $  - 2.924073507089058d-06
     $  + 4.168899443616945d-07*tu01
     $  + 3.507662788218251d-09*tu02
     $  - 4.293722251195467d-09*tu03
     $  + 2.261443227382951d-10*tu04
     $  + 1.039240016732763d-11*tu05
     $  - 1.240634428336177d-12*tu06
     $  + 1.274864772588004d-15*tu07
     $  + 3.647449807627184d-15*tu08

      CsH4YZ(3,3,2,1,08) = CsH4YZ(3,3,2,1,08) 
     $  + 4.904275749736814d-08
     $  + 4.133668477367700d-08*tu01
     $  - 7.526950868710852d-09*tu02
     $  + 2.239259567815297d-10*tu03
     $  + 4.170224646160578d-11*tu04
     $  - 3.187468213708523d-12*tu05
     $  - 7.098842077652034d-14*tu06
     $  + 1.452384976297652d-14*tu07
     $  - 1.427142340691597d-16*tu08

      CsH4YZ(3,3,2,1,09) = CsH4YZ(3,3,2,1,09) 
     $  + 2.786851661193777d-08
     $  - 6.101062128783235d-09*tu01
     $  - 1.214875335473143d-10*tu02
     $  + 9.653139069270182d-11*tu03
     $  - 5.201399815518443d-12*tu04
     $  - 3.769475322909781d-13*tu05
     $  + 4.215452651733698d-14*tu06
     $  + 2.921505823971058d-16*tu07
     $  - 1.704377007571768d-16*tu08

      CsH4YZ(3,3,2,1,10) = CsH4YZ(3,3,2,1,10) 
     $  - 7.275481453213889d-10
     $  - 5.914107874070430d-10*tu01
     $  + 1.333254979012196d-10*tu02
     $  - 3.991327303729965d-12*tu03
     $  - 1.096217629947484d-12*tu04
     $  + 8.679886689430781d-14*tu05
     $  + 3.009716424481228d-15*tu06
     $  - 5.388102063128265d-16*tu07

      CsH4YZ(3,3,2,1,11) = CsH4YZ(3,3,2,1,11) 
     $  - 3.067576389498712d-10
     $  + 9.272191746898272d-11*tu01
     $  + 2.317297325300030d-12*tu02
     $  - 2.020224924753556d-12*tu03
     $  + 1.169718383359022d-13*tu04
     $  + 1.119686588341522d-14*tu05
     $  - 1.281468784617925d-15*tu06

      CsH4YZ(3,3,2,1,12) = CsH4YZ(3,3,2,1,12) 
     $  + 1.067821735679313d-11
     $  + 8.303292833419615d-12*tu01
     $  - 2.326247277946035d-12*tu02
     $  + 7.676117641740818d-14*tu03
     $  + 2.598162918345012d-14*tu04
     $  - 2.219620413442271d-15*tu05

      CsH4YZ(3,3,2,1,13) = CsH4YZ(3,3,2,1,13) 
     $  + 3.584291387146712d-12
     $  - 1.422691840630811d-12*tu01
     $  - 3.468257564529789d-14*tu02
     $  + 4.009598497714948d-14*tu03
     $  - 2.585544744155824d-15*tu04
     $  - 2.929702511348479d-16*tu05

      CsH4YZ(3,3,2,1,14) = CsH4YZ(3,3,2,1,14) 
     $  - 1.546762825611956d-13
     $  - 1.139188734502114d-13*tu01
     $  + 3.982397204554370d-14*tu02
     $  - 1.555243500456162d-15*tu03
     $  - 5.710972945921437d-16*tu04

      CsH4YZ(3,3,2,1,15) = CsH4YZ(3,3,2,1,15) 
     $  - 4.313010448319136d-14
     $  + 2.177877119139418d-14*tu01
     $  + 4.073846826994494d-16*tu02
     $  - 7.622705955203326d-16*tu03

      CsH4YZ(3,3,2,1,16) = CsH4YZ(3,3,2,1,16) 
     $  + 2.210438028074893d-15
     $  + 1.522964357000962d-15*tu01
     $  - 6.682852651925779d-16*tu02

      CsH4YZ(3,3,2,1,17) = CsH4YZ(3,3,2,1,17) 
     $  + 5.272446121626678d-16
     $  - 3.308275183034576d-16*tu01
      endif

      return
      end

      subroutine fillcoeff2dhpl321e(iflag,n,CbH1YZ,CbH2YZ,CbH3YZ,
     $                              CbH4YZ)
      implicit none
      include 'src/Inc/types.f'
      integer iflag,n
      real(dp):: CbH1YZ,CbH2YZ,CbH3YZ,CbH4YZ
      real(dp):: 
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      real(dp):: HY1,HY2,HY3,HY4,HZ1,HZ2,HZ3,HZ4,HYZ1,HYZ2,HYZ3,HYZ4
      dimension CbH1YZ(3:3,22),CbH2YZ(0:3,0:3,22),
     $          CbH3YZ(0:3,0:3,0:3,22),CbH4YZ(0:3,0:3,0:3,0:3,22)
      common /zm/  
     $     zm01,zm02,zm03,zm04,zm05,zm06,zm07,zm08,zm09,zm10,
     $     zm11,zm12,zm13,zm14,zm15,zm16,zm17,zm18,zm19,zm20,zm21
      common /HPL2/ 
     $ HY1(-1:1),HY2(-1:1,-1:1),HY3(-1:1,-1:1,-1:1), 
     $           HY4(-1:1,-1:1,-1:1,-1:1), 
     $ HZ1(-1:1),HZ2(-1:1,-1:1),HZ3(-1:1,-1:1,-1:1), 
     $           HZ4(-1:1,-1:1,-1:1,-1:1), 
     $ HYZ1(0:3),HYZ2(0:3,0:3),HYZ3(0:3,0:3,0:3), 
     $           HYZ4(0:3,0:3,0:3,0:3) 
!$omp threadprivate(/zm/,/HPL2/)

* 2001-04-20:17:05:41.chpl 
* <- HPLexpand.321enew 
* produced by form-to-fortr for gehrt@pcth78 

      if (n.eq.2) return

      if (n.eq.3) then
      CbH3YZ(3,2,1,01) = 
     $  - HZ2(0,1)*CbH1YZ(3,1)

      CbH3YZ(3,2,1,02) = 
     $  - 2.812500000000000d-01
     $  + 2.812500000000000d-01*zm01
     $  - HZ2(0,1)*CbH1YZ(3,2)

      CbH3YZ(3,2,1,03) = 
     $  - 3.515625000000000d-02
     $  + 2.109375000000000d-01*zm01
     $  - 1.757812500000000d-01*zm02
     $  - HZ2(0,1)*CbH1YZ(3,3)

      CbH3YZ(3,2,1,04) = 
     $  - 2.197265625000000d-03
     $  - 1.318359375000000d-02*zm01
     $  - 4.614257812500000d-02*zm02
     $  + 6.152343750000000d-02*zm03
     $  - HZ2(0,1)*CbH1YZ(3,4)

      CbH3YZ(3,2,1,05) = 
     $  + 3.295898437500000d-03*zm01
     $  + 2.570800781250000d-02*zm02
     $  - 1.384277343750000d-02*zm03
     $  - 1.516113281250000d-02*zm04
     $  - HZ2(0,1)*CbH1YZ(3,5)

      CbH3YZ(3,2,1,06) = 
     $  + 8.239746093750000d-06
     $  - 4.531860351562500d-04*zm01
     $  - 8.981323242187500d-03*zm02
     $  - 7.498168945312500d-03*zm03
     $  + 1.404876708984375d-02*zm04
     $  + 2.875671386718750d-03*zm05
     $  - HZ2(0,1)*CbH1YZ(3,6)

      CbH3YZ(3,2,1,07) = 
     $  + 4.325866699218750d-05*zm01
     $  + 2.299771990094866d-03*zm02
     $  + 7.583509172712053d-03*zm03
     $  - 3.610774448939732d-03*zm04
     $  - 5.875233241489955d-03*zm05
     $  - 4.405321393694196d-04*zm06
     $  - HZ2(0,1)*CbH1YZ(3,7)

      CbH3YZ(3,2,1,08) = 
     $  - 5.911807624661192d-08
     $  - 3.641673496791294d-06*zm01
     $  - 4.716794831412179d-04*zm02
     $  - 3.581254822867257d-03*zm03
     $  - 2.087754862649100d-03*zm04
     $  + 4.366697583879743d-03*zm05
     $  + 1.721435614994594d-03*zm06
     $  + 5.625676135627590d-05*zm07
     $  - HZ2(0,1)*CbH1YZ(3,8)

      CbH3YZ(3,2,1,09) = 
     $  + 3.428848422303491d-07*zm01
     $  + 8.218594959803989d-05*zm02
     $  + 1.212263447897774d-03*zm03
     $  + 2.542964049748011d-03*zm04
     $  - 1.123539039066859d-03*zm05
     $  - 2.307248456137520d-03*zm06
     $  - 4.008323805672781d-04*zm07
     $  - 6.136456314398317d-06*zm08
     $  - HZ2(0,1)*CbH1YZ(3,9)

      CbH3YZ(3,2,1,10) = 
     $  + 5.172831671578543d-10
     $  - 3.037191167169687d-08*zm01
     $  - 1.263914911114439d-05*zm02
     $  - 3.307239583560398d-04*zm03
     $  - 1.441305054085595d-03*zm04
     $  - 6.844308078289031d-04*zm05
     $  + 1.527448219912392d-03*zm06
     $  + 8.625117455210004d-04*zm07
     $  + 7.858499367626345d-05*zm08
     $  + 5.838649005306010d-07*zm09
     $  - HZ2(0,1)*CbH1YZ(3,10)

      CbH3YZ(3,2,1,11) = 
     $  + 1.718119090917159d-09*zm01
     $  + 1.751680355622728d-06*zm02
     $  + 7.688475108544769d-05*zm03
     $  + 5.894576183387211d-04*zm04
     $  + 9.308077250788738d-04*zm05
     $  - 3.915364963861254d-04*zm06
     $  - 9.360843660769524d-04*zm07
     $  - 2.578162952345244d-04*zm08
     $  - 1.341700359681320d-05*zm09
     $  - 4.933168334067420d-08*zm10
     $  - HZ2(0,1)*CbH1YZ(3,11)

      CbH3YZ(3,2,1,12) = 
     $  - 5.009844723248847d-12
     $  - 5.227415122658511d-11*zm01
     $  - 2.209707098479118d-07*zm02
     $  - 1.573169545541778d-05*zm03
     $  - 1.951367241720943d-04*zm04
     $  - 5.974516448044738d-04*zm05
     $  - 2.471682943657717d-04*zm06
     $  + 5.768329875858186d-04*zm07
     $  + 4.116172497545184d-04*zm08
     $  + 6.521618621707211d-05*zm09
     $  + 2.039213694635509d-06*zm10
     $  + 3.749539556984234d-09*zm11
     $  - HZ2(0,1)*CbH1YZ(3,12)

      CbH3YZ(3,2,1,13) = 
     $  + 6.484170456090651d-12*zm01
     $  + 2.557176143152683d-08*zm02
     $  + 2.892747234922730d-06*zm03
     $  + 5.516512470120962d-05*zm04
     $  + 2.795002996309540d-04*zm05
     $  + 3.627254599851875d-04*zm06
     $  - 1.473935277061211d-04*zm07
     $  - 3.922599786086082d-04*zm08
     $  - 1.459462221832193d-04*zm09
     $  - 1.442907892104235d-05*zm10
     $  - 2.801440922111172d-07*zm11
     $  - 2.582866736999981d-10*zm12
     $  - HZ2(0,1)*CbH1YZ(3,13)

      CbH3YZ(3,2,1,14) = 
     $  + 5.173265120044837d-14
     $  - 1.019890740039256d-12*zm01
     $  - 2.742417452233713d-09*zm02
     $  - 4.849807601837021d-07*zm03
     $  - 1.374011447354138d-05*zm04
     $  - 1.064936509449355d-04*zm05
     $  - 2.548556682178075d-04*zm06
     $  - 9.543128603327024d-05*zm07
     $  + 2.301696996873983d-04*zm08
     $  + 1.939161165593301d-04*zm09
     $  + 4.403403081113979d-05*zm10
     $  + 2.853407282064314d-06*zm11
     $  + 3.517323774090562d-08*zm12
     $  + 1.623767522471413d-11*zm13
     $  - HZ2(0,1)*CbH1YZ(3,14)

      CbH3YZ(3,2,1,15) = 
     $  + 1.478168116164896d-14*zm01
     $  + 2.750375245510758d-10*zm02
     $  + 7.493485961403926d-08*zm03
     $  + 3.077758353000338d-06*zm04
     $  + 3.478336898895461d-05*zm05
     $  + 1.318358381685796d-04*zm06
     $  + 1.481225986713430d-04*zm07
     $  - 5.866692976331474d-05*zm08
     $  - 1.690925436304935d-04*zm09
     $  - 7.795303932341878d-05*zm10
     $  - 1.166607918331336d-05*zm11
     $  - 5.121110751452167d-07*zm12
     $  - 4.070173378130343d-09*zm13
     $  - 9.447341079147163d-13*zm14
     $  - HZ2(0,1)*CbH1YZ(3,15)

      CbH3YZ(3,2,1,16) = 
     $  - 5.588656410775645d-16
     $  + 7.690093311711172d-15*zm01
     $  - 2.585985568493176d-11*zm02
     $  - 1.075989947304598d-08*zm03
     $  - 6.291000839688175d-07*zm04
     $  - 1.003476093173988d-05*zm05
     $  - 5.591686160388019d-05*zm06
     $  - 1.114795597945977d-04*zm07
     $  - 3.870678341227419d-05*zm08
     $  + 9.574690994528782d-05*zm09
     $  + 9.135635487514957d-05*zm10
     $  + 2.681889154144599d-05*zm11
     $  + 2.770893424817942d-06*zm12
     $  + 8.436471192224349d-08*zm13
     $  + 4.370284851385175d-10*zm14
     $  + 5.154966464251066d-14*zm15
     $  - HZ2(0,1)*CbH1YZ(3,16)

      CbH3YZ(3,2,1,17) = 
     $  + 5.522457112631734d-17*zm01
     $  + 2.276979943481168d-12*zm02
     $  + 1.445265321077352d-09*zm03
     $  + 1.186219366228185d-07*zm04
     $  + 2.608061783595541d-06*zm05
     $  + 2.042368728305803d-05*zm06
     $  + 6.235282400218208d-05*zm07
     $  + 6.272076379025541d-05*zm08
     $  - 2.435593774453221d-05*zm09
     $  - 7.466877612491439d-05*zm10
     $  - 4.045372407302695d-05*zm11
     $  - 8.135230177144451d-06*zm12
     $  - 5.988254286515277d-07*zm13
     $  - 1.286900304342740d-08*zm14
     $  - 4.378416592380813d-11*zm15
     $  - 2.591253530637954d-15*zm16
     $  - HZ2(0,1)*CbH1YZ(3,17)

      CbH3YZ(3,2,1,18) = 
     $  + 6.245109440736378d-18
     $  - 1.084678842894483d-16*zm01
     $  - 1.889498991700165d-13*zm02
     $  - 1.825469547429476d-10*zm03
     $  - 2.080769533574444d-08*zm04
     $  - 6.193538102015513d-07*zm05
     $  - 6.609920471599398d-06*zm06
     $  - 2.878321074905508d-05*zm07
     $  - 4.982638956556446d-05*zm08
     $  - 1.630505122125097d-05*zm09
     $  + 4.114887739063278d-05*zm10
     $  + 4.325880827900330d-05*zm11
     $  + 1.541715976770251d-05*zm12
     $  + 2.219171215554570d-06*zm13
     $  + 1.190651784343113d-07*zm14
     $  + 1.830305626135966d-09*zm15
     $  + 4.111944632048071d-12*zm16
     $  + 1.158128374442430d-16*zm17
     $  - HZ2(0,1)*CbH1YZ(3,18)

      CbH3YZ(3,2,1,19) = 
     $  - 1.556875912028877d-19*zm01
     $  + 1.499047586796077d-14*zm02
     $  + 2.177545428459566d-11*zm03
     $  + 3.418252181629776d-09*zm04
     $  + 1.358258609619798d-07*zm05
     $  + 1.931676468209736d-06*zm06
     $  + 1.149597375621517d-05*zm07
     $  + 2.966578641323106d-05*zm08
     $  + 2.733720837289668d-05*zm09
     $  - 1.045039393085799d-05*zm10
     $  - 3.365412496028597d-05*zm11
     $  - 2.068912090275732d-05*zm12
     $  - 5.201920999866998d-06*zm13
     $  - 5.521361871690784d-07*zm14
     $  - 2.196948645228668d-08*zm15
     $  - 2.440833261718289d-10*zm16
     $  - 3.634198951260414d-13*zm17
     $  - 5.162465243299312d-18*zm18
     $  - HZ2(0,1)*CbH1YZ(3,19)

      CbH3YZ(3,2,1,20) = 
     $  - 7.165321710867487d-20
     $  + 1.367249409734930d-18*zm01
     $  - 1.136456763537555d-15*zm02
     $  - 2.462300426430547d-12*zm03
     $  - 5.287748946416486d-10*zm04
     $  - 2.773615350924479d-08*zm05
     $  - 5.166927940845775d-07*zm06
     $  - 4.082420976716709d-06*zm07
     $  - 1.466742386782678d-05*zm08
     $  - 2.268540537813743d-05*zm09
     $  - 7.077561941885397d-06*zm10
     $  + 1.815434102269993d-05*zm11
     $  + 2.062830255416229d-05*zm12
     $  + 8.561330474497993d-06*zm13
     $  + 1.583339313469099d-06*zm14
     $  + 1.266403812895935d-07*zm15
     $  + 3.787911454004556d-09*zm16
     $  + 3.066258372872891d-11*zm17
     $  + 3.033343840245557d-14*zm18
     $  + 2.887081027081440d-19*zm19
     $  - HZ2(0,1)*CbH1YZ(3,20)
      endif
      if (n.eq.3) return

      if (n.eq.4) then
      CbH4YZ(0,3,2,1,01) = 
     $  - HZ2(0,1)*CbH2YZ(0,3,1)

      CbH4YZ(0,3,2,1,02) = 
     $  - 1.406250000000000d-01
     $  + 1.406250000000000d-01*zm01
     $  - HZ2(0,1)*CbH2YZ(0,3,2)

      CbH4YZ(0,3,2,1,03) = 
     $  + 2.343750000000000d-02
     $  - 2.343750000000000d-02*zm02
     $  - HZ2(0,1)*CbH2YZ(0,3,3)

      CbH4YZ(0,3,2,1,04) = 
     $  - 5.493164062500000d-04
     $  - 1.647949218750000d-02*zm01
     $  + 1.483154296875000d-02*zm02
     $  + 2.197265625000000d-03*zm03
     $  - HZ2(0,1)*CbH2YZ(0,3,4)

      CbH4YZ(0,3,2,1,05) = 
     $  - 1.647949218750000d-04
     $  + 4.119873046875000d-03*zm01
     $  + 1.680908203125000d-03*zm02
     $  - 5.570068359375000d-03*zm03
     $  - 6.591796875000000d-05*zm04
     $  - HZ2(0,1)*CbH2YZ(0,3,5)

      CbH4YZ(0,3,2,1,06) = 
     $  + 4.806518554687500d-06
     $  - 4.531860351562500d-04*zm01
     $  - 2.863311767578125d-03*zm02
     $  + 2.114868164062500d-03*zm03
     $  + 1.205062866210937d-03*zm04
     $  - 8.239746093750000d-06*zm05
     $  - HZ2(0,1)*CbH2YZ(0,3,6)

      CbH4YZ(0,3,2,1,07) = 
     $  + 1.765659877232142d-06
     $  + 3.089904785156250d-05*zm01
     $  + 9.906613096898915d-04*zm02
     $  + 4.565491968271683d-04*zm03
     $  - 1.292715267259247d-03*zm04
     $  - 1.877905397999043d-04*zm05
     $  + 6.305928132971938d-07*zm06
     $  - HZ2(0,1)*CbH2YZ(0,3,7)

      CbH4YZ(0,3,2,1,08) = 
     $  - 4.532385845573580d-08
     $  - 3.545113972255161d-06*zm01
     $  - 2.159829650606427d-04*zm02
     $  - 6.390861102512904d-04*zm03
     $  + 4.000323159354073d-04*zm04
     $  + 4.349178927285330d-04*zm05
     $  + 2.365018640245710d-05*zm06
     $  + 5.911807624661192d-08*zm07
     $  - HZ2(0,1)*CbH2YZ(0,3,8)

      CbH4YZ(0,3,2,1,09) = 
     $  - 2.052710980785136d-08
     $  + 5.276288305010114d-07*zm01
     $  + 3.826171159744262d-05*zm02
     $  + 2.879784362656729d-04*zm03
     $  + 1.208201050758361d-04*zm04
     $  - 3.368030701364789d-04*zm05
     $  - 1.081340227808271d-04*zm06
     $  - 2.623364633443404d-06*zm07
     $  - 6.897108895438058d-09*zm08
     $  - HZ2(0,1)*CbH2YZ(0,3,9)

      CbH4YZ(0,3,2,1,10) = 
     $  + 4.470804516150026d-10
     $  - 3.199765876847870d-08*zm01
     $  - 6.113511111055101d-06*zm02
     $  - 8.904771081038883d-05*zm03
     $  - 1.722977118832724d-04*zm04
     $  + 9.575097646032060d-05*zm05
     $  + 1.494661260928426d-04*zm06
     $  + 2.201002927459016d-05*zm07
     $  + 2.638698384469869d-07*zm08
     $  - 5.172831671578543d-10*zm09
     $  - HZ2(0,1)*CbH2YZ(0,3,10)

      CbH4YZ(0,3,2,1,11) = 
     $  + 2.468851479617032d-10
     $  - 9.976175366615762d-10*zm01
     $  + 8.760268303119104d-07*zm02
     $  + 2.222108691402984d-05*zm03
     $  + 9.346627308563752d-05*zm04
     $  + 3.521610791173707d-05*zm05
     $  - 9.894382729334701d-05*zm06
     $  - 4.895445199544764d-05*zm07
     $  - 3.857570359956797d-06*zm08
     $  - 2.297451809176829d-08*zm09
     $  + 8.015751557198156d-11*zm10
     $  - HZ2(0,1)*CbH2YZ(0,3,11)

      CbH4YZ(0,3,2,1,12) = 
     $  - 4.589972022633706d-12
     $  - 2.975847765609815d-11*zm01
     $  - 1.099910335451129d-07*zm02
     $  - 4.766690592732494d-06*zm03
     $  - 3.575344862682478d-05*zm04
     $  - 5.292969144118780d-05*zm05
     $  + 2.706988406663929d-05*zm06
     $  + 5.278331090666782d-05*zm07
     $  + 1.310708180565822d-05*zm08
     $  + 5.979306884852958d-07*zm09
     $  + 1.643565444514183d-09*zm10
     $  + 5.009844723248847d-12*zm11
     $  - HZ2(0,1)*CbH2YZ(0,3,12)

      CbH4YZ(0,3,2,1,13) = 
     $  - 3.034809784275744d-12
     $  + 4.593669765167532d-11*zm01
     $  + 1.248975982599424d-08*zm02
     $  + 9.042990489163168d-07*zm03
     $  + 1.104255857096220d-05*zm04
     $  + 3.268915151179146d-05*zm05
     $  + 1.128691665995143d-05*zm06
     $  - 3.201335339124048d-05*zm07
     $  - 2.082495918355235d-05*zm08
     $  - 3.013773129324387d-06*zm09
     $  - 8.326543840762331d-08*zm10
     $  - 1.063451342932808d-10*zm11
     $  - 9.656761557417030d-13*zm12
     $  - HZ2(0,1)*CbH2YZ(0,3,13)

      CbH4YZ(0,3,2,1,14) = 
     $  + 4.886700372748551d-14
     $  - 1.316033908220691d-12*zm01
     $  - 1.361428194821000d-09*zm02
     $  - 1.545551603021491d-07*zm03
     $  - 2.918034060893446d-06*zm04
     $  - 1.447315331306119d-05*zm05
     $  - 1.781495348945288d-05*zm06
     $  + 8.581547612124585d-06*zm07
     $  + 1.942719567749690d-05*zm08
     $  + 6.730206355412833d-06*zm09
     $  + 6.125523658169625d-07*zm10
     $  + 1.054858142794552d-08*zm11
     $  + 8.178524813760189d-12*zm12
     $  - 5.173265120044837d-14*zm13
     $  - HZ2(0,1)*CbH2YZ(0,3,14)

      CbH4YZ(0,3,2,1,15) = 
     $  + 3.793727754699547d-14
     $  - 5.542774820432832d-13*zm01
     $  + 1.413263182429965d-10*zm02
     $  + 2.418518636928907d-08*zm03
     $  + 6.811775303537934d-07*zm04
     $  + 5.202968019895033d-06*zm05
     $  + 1.209323042219264d-05*zm06
     $  + 3.902029586522876d-06*zm07
     $  - 1.115298483569070d-05*zm08
     $  - 8.766839586450588d-06*zm09
     $  - 1.870507053332525d-06*zm10
     $  - 1.121694765676041d-07*zm11
     $  - 1.230007420507223d-09*zm12
     $  - 6.077722149955795d-13*zm13
     $  + 1.192246700965470d-14*zm14
     $  - HZ2(0,1)*CbH2YZ(0,3,15)

      CbH4YZ(0,3,2,1,16) = 
     $  - 5.370677276043950d-16
     $  + 1.153281900422219d-14*zm01
     $  - 1.301307079115108d-11*zm02
     $  - 3.507563856307911d-09*zm03
     $  - 1.434129265089482d-07*zm04
     $  - 1.604127373996741d-06*zm05
     $  - 5.964722200801278d-06*zm06
     $  - 6.411155110530426d-06*zm07
     $  + 2.952242514076860d-06*zm08
     $  + 7.440898492893315d-06*zm09
     $  + 3.254801411307492d-06*zm10
     $  + 4.601020143521302d-07*zm11
     $  + 1.876058398773758d-08*zm12
     $  + 1.331452164022211d-10*zm13
     $  + 1.537593949977636d-14*zm14
     $  + 5.588656410775645d-16*zm15
     $  - HZ2(0,1)*CbH2YZ(0,3,16)

      CbH4YZ(0,3,2,1,17) = 
     $  - 4.807888235740812d-16
     $  + 8.228634571885698d-15*zm01
     $  + 1.072577334668818d-12*zm02
     $  + 4.746618753204123d-10*zm03
     $  + 2.763629637928727d-08*zm04
     $  + 4.375331504148017d-07*zm05
     $  + 2.405282038534544d-06*zm06
     $  + 4.676479769896664d-06*zm07
     $  + 1.430766337375696d-06*zm08
     $  - 4.114062353567955d-06*zm09
     $  - 3.716694763945723d-06*zm10
     $  - 1.042410849860475d-06*zm11
     $  - 1.020963739751888d-07*zm12
     $  - 2.895566238157549d-09*zm13
     $  - 1.342828597056507d-11*zm14
     $  + 1.221858518385773d-15*zm15
     $  - 1.498826265776730d-16*zm16
     $  - HZ2(0,1)*CbH2YZ(0,3,17)

      CbH4YZ(0,3,2,1,18) = 
     $  + 6.064381388442552d-18
     $  - 1.582982095944186d-16*zm01
     $  - 9.312237027856816d-14*zm02
     $  - 6.019953297640497d-11*zm03
     $  - 4.927627376995874d-09*zm04
     $  - 1.077317719293178d-07*zm05
     $  - 8.352571885773428d-07*zm06
     $  - 2.507287319420754d-06*zm07
     $  - 2.428462742731598d-06*zm08
     $  + 1.079653756504655d-06*zm09
     $  + 2.951375329698941d-06*zm10
     $  + 1.535419309186893d-06*zm11
     $  + 2.961425242436253d-07*zm12
     $  + 2.071906712500121d-08*zm13
     $  + 4.156911585159197d-10*zm14
     $  + 1.264768638156404d-12*zm15
     $  + 1.635635007898742d-16*zm16
     $  - 6.245109440736378d-18*zm17
     $  - HZ2(0,1)*CbH2YZ(0,3,18)

      CbH4YZ(0,3,2,1,19) = 
     $  + 6.162936948095109d-18
     $  - 1.172514896050099d-16*zm01
     $  + 8.549250511262111d-15*zm02
     $  + 7.197145896921452d-12*zm03
     $  + 8.195925864217425d-10*zm04
     $  + 2.429036956092763d-08*zm05
     $  + 2.572747339147350d-07*zm06
     $  + 1.106638360978021d-06*zm07
     $  + 1.874521641077567d-06*zm08
     $  + 5.495595308817704d-07*zm09
     $  - 1.587684310033966d-06*zm10
     $  - 1.596163599412202d-06*zm11
     $  - 5.492938576502281d-07*zm12
     $  - 7.602860288472057d-08*zm13
     $  - 3.885098921482762d-09*zm14
     $  - 5.585359256219224d-11*zm15
     $  - 1.120529501696083d-13*zm16
     $  - 3.730039358413974d-17*zm17
     $  + 1.910752456231329d-18*zm18
     $  - HZ2(0,1)*CbH2YZ(0,3,19)

      CbH4YZ(2,3,2,1,01) = 
     $  - HZ2(0,1)*CbH2YZ(2,3,1)

      CbH4YZ(2,3,2,1,02) = 
     $  - HZ2(0,1)*CbH2YZ(2,3,2)

      CbH4YZ(2,3,2,1,03) = 
     $  - 7.031250000000000d-02
     $  + 1.406250000000000d-01*zm01
     $  - 7.031250000000000d-02*zm02
     $  - HZ2(0,1)*CbH2YZ(2,3,3)

      CbH4YZ(2,3,2,1,04) = 
     $  - 6.591796875000000d-03
     $  + 4.614257812500000d-02*zm01
     $  - 7.250976562500000d-02*zm02
     $  + 3.295898437500000d-02*zm03
     $  - HZ2(0,1)*CbH2YZ(2,3,4)

      CbH4YZ(2,3,2,1,05) = 
     $  - 3.295898437500000d-04
     $  - 1.647949218750000d-03*zm01
     $  - 4.943847656250000d-03*zm02
     $  + 1.614990234375000d-02*zm03
     $  - 9.228515625000000d-03*zm04
     $  - HZ2(0,1)*CbH2YZ(2,3,5)

      CbH4YZ(2,3,2,1,06) = 
     $  + 4.119873046875000d-04*zm01
     $  + 2.801513671875000d-03*zm02
     $  - 4.943847656250000d-03*zm03
     $  - 1.647949218750000d-04*zm04
     $  + 1.895141601562500d-03*zm05
     $  - HZ2(0,1)*CbH2YZ(2,3,6)

      CbH4YZ(2,3,2,1,07) = 
     $  + 8.828299386160714d-07
     $  - 4.943847656250000d-05*zm01
     $  - 9.137289864676339d-04*zm02
     $  + 1.589093889508928d-04*zm03
     $  + 2.308600289481026d-03*zm04
     $  - 1.197117396763392d-03*zm05
     $  - 3.081076485770089d-04*zm06
     $  - HZ2(0,1)*CbH2YZ(2,3,7)

      CbH4YZ(2,3,2,1,08) = 
     $  + 4.055500030517578d-06*zm01
     $  + 2.115481240408761d-04*zm02
     $  + 4.953503608703613d-04*zm03
     $  - 1.049464089529854d-03*zm04
     $  - 2.122930118015834d-04*zm05
     $  + 5.095032283238002d-04*zm06
     $  + 4.129988806588309d-05*zm07
     $  - HZ2(0,1)*CbH2YZ(2,3,8)

      CbH4YZ(2,3,2,1,09) = 
     $  - 4.926506353884327d-09
     $  - 2.985462850453902d-07*zm01
     $  - 3.900315080370221d-05*zm02
     $  - 2.591312783105032d-04*zm03
     $  + 1.244583300181797d-04*zm04
     $  + 5.378710372107369d-04*zm05
     $  - 2.204384974070957d-04*zm06
     $  - 1.387649044698598d-04*zm07
     $  - 4.688063446356325d-06*zm08
     $  - HZ2(0,1)*CbH2YZ(2,3,9)

      CbH4YZ(2,3,2,1,10) = 
     $  + 2.571636316727618d-08*zm01
     $  + 6.138229856685716d-06*zm02
     $  + 8.475581237248011d-05*zm03
     $  + 9.980254513876778d-05*zm04
     $  - 2.749877316611153d-04*zm05
     $  - 8.877820628029959d-05*zm06
     $  + 1.429812056677682d-04*zm07
     $  + 2.960219431896598d-05*zm08
     $  + 4.602342235798738d-07*zm09
     $  - HZ2(0,1)*CbH2YZ(2,3,10)

      CbH4YZ(2,3,2,1,11) = 
     $  + 3.526930685167188d-11
     $  - 2.106081466285549d-09*zm01
     $  - 8.596893545095022d-07*zm02
     $  - 2.168760063033378d-05*zm03
     $  - 7.572143834519695d-05*zm04
     $  + 5.160506224477445d-05*zm05
     $  + 1.508099337096338d-04*zm06
     $  - 4.533657779941311d-05*zm07
     $  - 5.344955126214115d-05*zm08
     $  - 5.318258780163603d-06*zm09
     $  - 3.980897049072279d-08*zm10
     $  - HZ2(0,1)*CbH2YZ(2,3,11)

      CbH4YZ(2,3,2,1,12) = 
     $  + 1.073824431823224d-10*zm01
     $  + 1.093726397832382d-07*zm02
     $  + 4.695816920614060d-06*zm03
     $  + 3.203580420332959d-05*zm04
     $  + 2.133438167125954d-05*zm05
     $  - 8.264651384156245d-05*zm06
     $  - 3.403424185567668d-05*zm07
     $  + 4.239175442765174d-05*zm08
     $  + 1.527495572735695d-05*zm09
     $  + 8.354794945920332d-07*zm10
     $  + 3.083230208792137d-09*zm11
     $  - HZ2(0,1)*CbH2YZ(2,3,12)

      CbH4YZ(2,3,2,1,13) = 
     $  - 2.890295032643565d-13
     $  - 2.726786913654015d-12*zm01
     $  - 1.274529436711645d-08*zm02
     $  - 8.948495045521079d-07*zm03
     $  - 1.035029011826980d-05*zm04
     $  - 2.321047619032958d-05*zm05
     $  + 2.020865483300204d-05*zm06
     $  + 4.753853549720713d-05*zm07
     $  - 9.531677182575011d-06*zm08
     $  - 1.998467674254497d-05*zm09
     $  - 3.644825337832881d-06*zm10
     $  - 1.174306243314533d-07*zm11
     $  - 2.163195898260135d-10*zm12
     $  - HZ2(0,1)*CbH2YZ(2,3,13)

      CbH4YZ(2,3,2,1,14) = 
     $  + 3.473662744334277d-13*zm01
     $  + 1.369568424700218d-09*zm02
     $  + 1.535986860798858d-07*zm03
     $  + 2.800305935693940d-06*zm04
     $  + 1.201795579980773d-05*zm05
     $  + 4.458490733262506d-06*zm06
     $  - 2.732780291203439d-05*zm07
     $  - 1.311784558406181d-05*zm08
     $  + 1.319537980850297d-05*zm09
     $  + 7.045561246188054d-06*zm10
     $  + 7.579786515445303d-07*zm11
     $  + 1.499388243950449d-08*zm12
     $  + 1.383678609107132d-11*zm13
     $  - HZ2(0,1)*CbH2YZ(2,3,14)

      CbH4YZ(2,3,2,1,15) = 
     $  + 2.586632560022418d-15
     $  - 5.358116956198525d-14*zm01
     $  - 1.370698780746837d-10*zm02
     $  - 2.411191713657342d-08*zm03
     $  - 6.627566856678839d-07*zm04
     $  - 4.637676823569709d-06*zm05
     $  - 7.418100863643597d-06*zm06
     $  + 7.971219109226863d-06*zm07
     $  + 1.628004928603342d-05*zm08
     $  - 1.812679156403406d-06*zm09
     $  - 7.494104287409519d-06*zm10
     $  - 2.059031176453774d-06*zm11
     $  - 1.409117022161704d-07*zm12
     $  - 1.757850003284045d-09*zm13
     $  - 8.118837612357066d-13*zm14
     $  - HZ2(0,1)*CbH2YZ(2,3,15)

      CbH4YZ(2,3,2,1,16) = 
     $  + 6.928913044522954d-16*zm01
     $  + 1.289169107202722d-11*zm02
     $  + 3.499679160444759d-09*zm03
     $  + 1.407573512524827d-07*zm04
     $  + 1.486200498560356d-06*zm05
     $  + 4.549334492794921d-06*zm06
     $  + 7.634418985670374d-07*zm07
     $  - 9.693259145374584d-06*zm08
     $  - 5.176200650024007d-06*zm09
     $  + 4.272164264394129d-06*zm10
     $  + 3.107201256567441d-06*zm11
     $  + 5.228422550703818d-07*zm12
     $  + 2.381441727033217d-08*zm13
     $  + 1.907450926885513d-10*zm14
     $  + 4.428441130850232d-14*zm15
     $  - HZ2(0,1)*CbH2YZ(2,3,16)

      CbH4YZ(2,3,2,1,17) = 
     $  - 2.465583710636314d-17
     $  + 3.639246596818560d-16*zm01
     $  - 1.141215254922506d-12*zm02
     $  - 4.735605713541642d-10*zm03
     $  - 2.727971402187227d-08*zm04
     $  - 4.149556256369589d-07*zm05
     $  - 2.024210323770895d-06*zm06
     $  - 2.451295508414011d-06*zm07
     $  + 3.210563663926040d-06*zm08
     $  + 5.931780589304206d-06*zm09
     $  - 1.937009589766874d-07*zm10
     $  - 2.847241029428098d-06*zm11
     $  - 1.060941093380649d-06*zm12
     $  - 1.185233255689278d-07*zm13
     $  - 3.702691916342866d-09*zm14
     $  - 1.927839421208272d-11*zm15
     $  - 2.274249910698999d-15*zm16
     $  - HZ2(0,1)*CbH2YZ(2,3,17)

      CbH4YZ(2,3,2,1,18) = 
     $  + 2.301023796929889d-18*zm01
     $  + 9.487186328791842d-14*zm02
     $  + 6.012451421391130d-11*zm03
     $  + 4.882361304239214d-09*zm04
     $  + 1.037266602905301d-07*zm05
     $  + 7.423177291442706d-07*zm06
     $  + 1.747047363296835d-06*zm07
     $  + 1.533082450305537d-08*zm08
     $  - 3.628195897282817d-06*zm09
     $  - 2.096368265849257d-06*zm10
     $  + 1.425627168828643d-06*zm11
     $  + 1.346603912328437d-06*zm12
     $  + 3.140168645205384d-07*zm13
     $  + 2.441485106700417d-08*zm14
     $  + 5.343841198959832d-10*zm15
     $  + 1.824232277928229d-12*zm16
     $  + 1.079688971099147d-16*zm17
     $  - HZ2(0,1)*CbH2YZ(2,3,18)

      CbH4YZ(2,3,2,1,19) = 
     $  + 2.465174779238043d-19
     $  - 4.528144489349395d-18*zm01
     $  - 7.454267024436594d-15*zm02
     $  - 7.198342296464905d-12*zm03
     $  - 8.141505939869014d-10*zm04
     $  - 2.362682032365027d-08*zm05
     $  - 2.364697366341255d-07*zm06
     $  - 8.752614583206190d-07*zm07
     $  - 8.306517953885281d-07*zm08
     $  + 1.323210724117637d-06*zm09
     $  + 2.267918234679621d-06*zm10
     $  + 8.328674559357331d-08*zm11
     $  - 1.099012441235557d-06*zm12
     $  - 5.209732323216295d-07*zm13
     $  - 8.289892251790497d-08*zm14
     $  - 4.627692347691135d-09*zm15
     $  - 7.208659269094414d-11*zm16
     $  - 1.623090323372616d-13*zm17
     $  - 4.571559372799067d-18*zm18
     $  - HZ2(0,1)*CbH2YZ(2,3,19)

      CbH4YZ(3,0,2,1,01) = 
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,1)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,1)
     $  - HZ3(0,1,1)*CbH1YZ(3,1)
     $  - HZ3(1,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,0,2,1,02) = 
     $  - 9.986229369893631d-02
     $  + 9.986229369893631d-02*zm01
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,2)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,2)
     $  - HZ3(0,1,1)*CbH1YZ(3,2)
     $  - HZ3(1,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,0,2,1,03) = 
     $  + 4.993114684946815d-02*zm01
     $  - 4.993114684946815d-02*zm02
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,3)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,3)
     $  - HZ3(0,1,1)*CbH1YZ(3,3)
     $  - HZ3(1,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,0,2,1,04) = 
     $  + 7.324218750000000d-04
     $  - 1.624040067641291d-02*zm01
     $  + 2.197265625000000d-03*zm02
     $  + 1.331071317641291d-02*zm03
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,4)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,4)
     $  - HZ3(0,1,1)*CbH1YZ(3,4)
     $  - HZ3(1,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,0,2,1,05) = 
     $  + 2.369173885282583d-03*zm01
     $  + 9.744240405847751d-03*zm02
     $  - 9.744240405847751d-03*zm03
     $  - 2.369173885282583d-03*zm04
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,5)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,5)
     $  - HZ3(0,1,1)*CbH1YZ(3,5)
     $  - HZ3(1,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,0,2,1,06) = 
     $  - 7.690429687500000d-06
     $  - 2.630667203566537d-04*zm01
     $  - 4.740042203566537d-03*zm02
     $  + 7.690429687500000d-05*zm03
     $  + 4.624685758254037d-03*zm04
     $  + 3.092092984816537d-04*zm05
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,6)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,6)
     $  - HZ3(0,1,1)*CbH1YZ(3,6)
     $  - HZ3(1,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,0,2,1,07) = 
     $  + 3.193949351014119d-05*zm01
     $  + 1.356435858958886d-03*zm02
     $  + 2.541792212503862d-03*zm03
     $  - 2.541792212503862d-03*zm04
     $  - 1.356435858958886d-03*zm05
     $  - 3.193949351014119d-05*zm06
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,7)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,7)
     $  - HZ3(0,1,1)*CbH1YZ(3,7)
     $  - HZ3(1,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,0,2,1,08) = 
     $  + 8.558045323319059d-08
     $  - 3.564586541173167d-06*zm01
     $  - 2.986829772746865d-04*zm02
     $  - 1.614806262102931d-03*zm03
     $  + 2.995315863161670d-06*zm04
     $  + 1.610013756721872d-03*zm05
     $  + 3.010792299652158d-04*zm06
     $  + 2.879942915307643d-06*zm07
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,8)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,8)
     $  - HZ3(0,1,1)*CbH1YZ(3,8)
     $  - HZ3(1,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,0,2,1,09) = 
     $  + 2.474479099804969d-07*zm01
     $  + 5.517448807020464d-05*zm02
     $  + 6.308236025828144d-04*zm03
     $  + 7.921484164380826d-04*zm04
     $  - 7.921484164380826d-04*zm05
     $  - 6.308236025828144d-04*zm06
     $  - 5.517448807020464d-05*zm07
     $  - 2.474479099804969d-07*zm08
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,9)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,9)
     $  - HZ3(0,1,1)*CbH1YZ(3,9)
     $  - HZ3(1,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,0,2,1,10) = 
     $  - 9.754482580690967d-10
     $  - 9.969020977238580d-09*zm01
     $  - 8.804320862717909d-06*zm02
     $  - 1.885203243840459d-04*zm03
     $  - 5.962863084763391d-04*zm04
     $  + 1.229064805167061d-07*zm05
     $  + 5.960814643421446d-04*zm06
     $  + 1.886373781750142d-04*zm07
     $  + 8.760425691104799d-06*zm08
     $  + 1.972350355792954d-08*zm09
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,10)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,10)
     $  - HZ3(0,1,1)*CbH1YZ(3,10)
     $  - HZ3(1,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,0,2,1,11) = 
     $  + 1.287163586927524d-09*zm01
     $  + 1.238239775523217d-06*zm02
     $  + 4.671386408331616d-05*zm03
     $  + 2.815453217783372d-04*zm04
     $  + 2.764116319866694d-04*zm05
     $  - 2.764116319866694d-04*zm06
     $  - 2.815453217783372d-04*zm07
     $  - 4.671386408331616d-05*zm08
     $  - 1.238239775523217d-06*zm09
     $  - 1.287163586927524d-09*zm10
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,11)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,11)
     $  - HZ3(0,1,1)*CbH1YZ(3,11)
     $  - HZ3(1,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,0,2,1,12) = 
     $  + 1.132268282733228d-11
     $  - 2.036002703013753d-10*zm01
     $  - 1.574401929993939d-07*zm02
     $  - 1.000727395353897d-05*zm03
     $  - 1.023826361245823d-04*zm04
     $  - 2.332166838057299d-04*zm05
     $  + 5.231079466227516d-09*zm06
     $  + 2.332077162409307d-04*zm07
     $  + 1.023882408525818d-04*zm08
     $  + 1.000478296331695d-05*zm09
     $  + 1.581874900659978d-07*zm10
     $  + 6.772807637338792d-11*zm11
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,12)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,12)
     $  - HZ3(0,1,1)*CbH1YZ(3,12)
     $  - HZ3(1,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,0,2,1,13) = 
     $  + 4.404853894478976d-12*zm01
     $  + 1.846634855040465d-08*zm02
     $  + 1.903054506188204d-06*zm03
     $  + 3.097240271634505d-05*zm04
     $  + 1.254581361480611d-04*zm05
     $  + 1.041524983562038d-04*zm06
     $  - 1.041524983562038d-04*zm07
     $  - 1.254581361480611d-04*zm08
     $  - 3.097240271634505d-05*zm09
     $  - 1.903054506188204d-06*zm10
     $  - 1.846634855040465d-08*zm11
     $  - 4.404853894478976d-12*zm12
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,13)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,13)
     $  - HZ3(0,1,1)*CbH1YZ(3,13)
     $  - HZ3(1,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,0,2,1,14) = 
     $  - 1.336239731582691d-13
     $  + 1.489971154763642d-12*zm01
     $  - 2.003175500085616d-09*zm02
     $  - 3.272460555398215d-07*zm03
     $  - 8.120487475488642d-06*zm04
     $  - 5.245841008198113d-05*zm05
     $  - 9.535087055860464d-05*zm06
     $  + 2.292987379395898d-10*zm07
     $  + 9.535046928581325d-05*zm08
     $  + 5.245867759717540d-05*zm09
     $  + 8.120353717891511d-06*zm10
     $  + 3.272946946660511d-07*zm11
     $  + 1.991015718528214d-09*zm12
     $  + 3.807644694521254d-13*zm13
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,14)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,14)
     $  - HZ3(0,1,1)*CbH1YZ(3,14)
     $  - HZ3(1,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,0,2,1,15) = 
     $  + 1.442326737847127d-14*zm01
     $  + 2.000980131524468d-10*zm02
     $  + 5.155860781940167d-08*zm03
     $  + 1.893219136872576d-06*zm04
     $  + 1.835127549572086d-05*zm05
     $  + 5.645600786088533d-05*zm06
     $  + 4.148415315855085d-05*zm07
     $  - 4.148415315855085d-05*zm08
     $  - 5.645600786088533d-05*zm09
     $  - 1.835127549572086d-05*zm10
     $  - 1.893219136872576d-06*zm11
     $  - 5.155860781940167d-08*zm12
     $  - 2.000980131524468d-10*zm13
     $  - 1.442326737847127d-14*zm14
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,15)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,15)
     $  - HZ3(0,1,1)*CbH1YZ(3,15)
     $  - HZ3(1,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,0,2,1,16) = 
     $  + 1.601034013519994d-15
     $  - 2.459652319550422d-14*zm01
     $  - 1.863973770265655d-11*zm02
     $  - 7.514088967760999d-09*zm03
     $  - 3.994869690306850d-07*zm04
     $  - 5.584482512410566d-06*zm05
     $  - 2.622061864072916d-05*zm06
     $  - 4.037850996248656d-05*zm07
     $  + 1.030265387700116d-11*zm08
     $  + 4.037849164665745d-05*zm09
     $  + 2.622063146180954d-05*zm10
     $  + 5.584475519093995d-06*zm11
     $  + 3.994898829125896d-07*zm12
     $  + 7.513192388713428d-09*zm13
     $  + 1.883186178427895d-11*zm14
     $  - 1.020021020815679d-15*zm15
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,16)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,16)
     $  - HZ3(0,1,1)*CbH1YZ(3,16)
     $  - HZ3(1,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,0,2,1,17) = 
     $  + 1.161439269671540d-19*zm01
     $  + 1.662222844210394d-12*zm02
     $  + 1.020567945252679d-09*zm03
     $  + 7.728170476565552d-08*zm04
     $  + 1.514918876153956d-06*zm05
     $  + 1.025098684623170d-05*zm06
     $  + 2.573439716708017d-05*zm07
     $  + 1.723134931541685d-05*zm08
     $  - 1.723134931541685d-05*zm09
     $  - 2.573439716708017d-05*zm10
     $  - 1.025098684623170d-05*zm11
     $  - 1.514918876153956d-06*zm12
     $  - 7.728170476565552d-08*zm13
     $  - 1.020567945252679d-09*zm14
     $  - 1.662222844210394d-12*zm15
     $  - 1.161439269671540d-19*zm16
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,17)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,17)
     $  - HZ3(0,1,1)*CbH1YZ(3,17)
     $  - HZ3(1,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,0,2,1,18) = 
     $  - 1.944607762378269d-17
     $  + 3.297259288486184d-16*zm01
     $  - 1.412444347077270d-13*zm02
     $  - 1.299968675927625d-10*zm03
     $  - 1.384112248539960d-08*zm04
     $  - 3.725880075618466d-07*zm05
     $  - 3.500963826901754d-06*zm06
     $  - 1.297476255708759d-05*zm07
     $  - 1.759165609640555d-05*zm08
     $  + 4.727341470341573d-13*zm09
     $  + 1.759165524548409d-05*zm10
     $  + 1.297476317593956d-05*zm11
     $  + 3.500963465904769d-06*zm12
     $  + 3.725881741758397d-07*zm13
     $  + 1.384106298040208d-08*zm14
     $  + 1.300127355921035d-10*zm15
     $  + 1.382691848312883d-13*zm16
     $  + 2.030346837947012d-17*zm17
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,18)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,18)
     $  - HZ3(0,1,1)*CbH1YZ(3,18)
     $  - HZ3(1,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,0,2,1,19) = 
     $  + 3.420710890813615d-19*zm01
     $  + 1.094173044633044d-14*zm02
     $  + 1.561004151280244d-11*zm03
     $  + 2.312743492471602d-09*zm04
     $  + 8.412008270455884d-08*zm05
     $  + 1.069058630245605d-06*zm06
     $  + 5.539474666217551d-06*zm07
     $  + 1.188560998663745d-05*zm08
     $  + 7.396457109299644d-06*zm09
     $  - 7.396457109299644d-06*zm10
     $  - 1.188560998663745d-05*zm11
     $  - 5.539474666217551d-06*zm12
     $  - 1.069058630245605d-06*zm13
     $  - 8.412008270455884d-08*zm14
     $  - 2.312743492471602d-09*zm15
     $  - 1.561004151280244d-11*zm16
     $  - 1.094173044633044d-14*zm17
     $  - 3.420710890813615d-19*zm18
     $  + 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,19)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,19)
     $  - HZ3(0,1,1)*CbH1YZ(3,19)
     $  - HZ3(1,0,1)*CbH1YZ(3,19)

      CbH4YZ(3,1,2,1,01) = 
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,1)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,1)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,1)
     $  + HZ3(-1,0,1)*CbH1YZ(3,1)
     $  + HZ3(0,-1,1)*CbH1YZ(3,1)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,1,2,1,02) = 
     $  + 5.871414221550108d-02
     $  - 5.871414221550108d-02*zm01
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,2)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,2)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,2)
     $  + HZ3(-1,0,1)*CbH1YZ(3,2)
     $  + HZ3(0,-1,1)*CbH1YZ(3,2)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,1,2,1,03) = 
     $  + 2.195748841508230d-03
     $  - 3.374856879076700d-02*zm01
     $  + 3.155281994925877d-02*zm02
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,3)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,3)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,3)
     $  + HZ3(-1,0,1)*CbH1YZ(3,3)
     $  + HZ3(0,-1,1)*CbH1YZ(3,3)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,1,2,1,04) = 
     $  - 4.282687182121974d-04
     $  + 8.306373680343052d-03*zm01
     $  + 1.185411292060167d-03*zm02
     $  - 9.063516254191021d-03*zm03
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,4)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,4)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,4)
     $  + HZ3(-1,0,1)*CbH1YZ(3,4)
     $  + HZ3(0,-1,1)*CbH1YZ(3,4)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,1,2,1,05) = 
     $  - 3.503933972807269d-05
     $  - 8.836840429668446d-04*zm01
     $  - 6.305658097587809d-03*zm02
     $  + 5.494514184122635d-03*zm03
     $  + 1.729867296160090d-03*zm04
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,5)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,5)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,5)
     $  + HZ3(-1,0,1)*CbH1YZ(3,5)
     $  + HZ3(0,-1,1)*CbH1YZ(3,5)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,1,2,1,06) = 
     $  + 4.372968960426554d-06
     $  + 1.005611954368156d-04*zm01
     $  + 2.542554847024747d-03*zm02
     $  + 5.508336056316494d-04*zm03
     $  - 2.960934192095807d-03*zm04
     $  - 2.373884249578315d-04*zm05
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,6)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,6)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,6)
     $  + HZ3(-1,0,1)*CbH1YZ(3,6)
     $  + HZ3(0,-1,1)*CbH1YZ(3,6)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,1,2,1,07) = 
     $  + 5.247592492389602d-07
     $  - 1.655959723994455d-05*zm01
     $  - 6.574455620525874d-04*zm02
     $  - 1.641371322623760d-03*zm03
     $  + 1.362466393780555d-03*zm04
     $  + 9.277153820019389d-04*zm05
     $  + 2.466994688455853d-05*zm06
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,7)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,7)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,7)
     $  + HZ3(-1,0,1)*CbH1YZ(3,7)
     $  + HZ3(0,-1,1)*CbH1YZ(3,7)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,1,2,1,08) = 
     $  - 4.573267202813310d-08
     $  + 1.615037937936050d-06*zm01
     $  + 1.389437078008454d-04*zm02
     $  + 9.053656518962465d-04*zm03
     $  + 1.601892305372272d-04*zm04
     $  - 9.897020001337704d-04*zm05
     $  - 2.142509674292660d-04*zm06
     $  - 2.114927937190717d-06*zm07
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,8)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,8)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,8)
     $  + HZ3(-1,0,1)*CbH1YZ(3,8)
     $  + HZ3(0,-1,1)*CbH1YZ(3,8)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,1,2,1,09) = 
     $  - 7.502523854489902d-09
     $  - 4.455081902538367d-08*zm01
     $  - 2.510281853485196d-05*zm02
     $  - 3.284245432576146d-04*zm03
     $  - 5.155679784915691d-04*zm04
     $  + 4.159100489125848d-04*zm05
     $  + 4.129485069886791d-04*zm06
     $  + 4.011361038735724d-05*zm07
     $  + 1.752273382943888d-07*zm08
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,9)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,9)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,9)
     $  + HZ3(-1,0,1)*CbH1YZ(3,9)
     $  + HZ3(0,-1,1)*CbH1YZ(3,9)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,1,2,1,10) = 
     $  + 4.661133818847343d-10
     $  + 3.352839565865062d-09*zm01
     $  + 3.886691126864123d-06*zm02
     $  + 9.356634374218653d-05*zm03
     $  + 3.435758529085472d-04*zm04
     $  + 5.091319723937472d-05*zm05
     $  - 3.571339931782672d-04*zm06
     $  - 1.283604374119450d-04*zm07
     $  - 6.436330977297204d-06*zm08
     $  - 1.514240241100968d-08*zm09
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,10)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,10)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,10)
     $  + HZ3(-1,0,1)*CbH1YZ(3,10)
     $  + HZ3(0,-1,1)*CbH1YZ(3,10)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,1,2,1,11) = 
     $  + 1.040872199725930d-10
     $  - 1.595333982037266d-09*zm01
     $  - 5.328634389048266d-07*zm02
     $  - 2.237690600536061d-05*zm03
     $  - 1.520581555718706d-04*zm04
     $  - 1.811863508630481d-04*zm05
     $  + 1.436300848867938d-04*zm06
     $  + 1.789902822507832d-04*zm07
     $  + 3.261833339758500d-05*zm08
     $  + 9.159561689124797d-07*zm09
     $  + 1.110421871637568d-09*zm10
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,11)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,11)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,11)
     $  + HZ3(-1,0,1)*CbH1YZ(3,11)
     $  + HZ3(0,-1,1)*CbH1YZ(3,11)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,1,2,1,12) = 
     $  - 4.491779111918475d-12
     $  + 8.641832044969679d-11*zm01
     $  + 6.765899198155587d-08*zm02
     $  + 4.660332878946111d-06*zm03
     $  + 5.285754273969151d-05*zm04
     $  + 1.367391628005075d-04*zm05
     $  + 1.786757420821831d-05*zm06
     $  - 1.373735245548805d-04*zm07
     $  - 6.758755628458841d-05*zm08
     $  - 7.113446286843074d-06*zm09
     $  - 1.177775569263959d-07*zm10
     $  - 4.886264800367311d-11*zm11
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,12)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,12)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,12)
     $  + HZ3(-1,0,1)*CbH1YZ(3,12)
     $  + HZ3(0,-1,1)*CbH1YZ(3,12)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,1,2,1,13) = 
     $  - 1.415100372575546d-12
     $  + 1.533197289744456d-11*zm01
     $  - 7.982819203371054d-09*zm02
     $  - 8.670700870158052d-07*zm03
     $  - 1.544484023860488d-05*zm04
     $  - 6.936132190871644d-05*zm05
     $  - 6.862497867363831d-05*zm06
     $  + 5.377133688939205d-05*zm07
     $  + 7.815108685391407d-05*zm08
     $  + 2.099876561060059d-05*zm09
     $  + 1.371177789076238d-06*zm10
     $  + 1.381134686609682d-08*zm11
     $  + 1.320457233275733d-12*zm12
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,13)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,13)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,13)
     $  + HZ3(-1,0,1)*CbH1YZ(3,13)
     $  + HZ3(0,-1,1)*CbH1YZ(3,13)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,1,2,1,14) = 
     $  + 3.884586225687804d-14
     $  - 4.338819377798065d-13*zm01
     $  + 8.511235123213074d-10*zm02
     $  + 1.466700249688104d-07*zm03
     $  + 3.938096879701873d-06*zm04
     $  + 2.782077449156978d-05*zm05
     $  + 5.657683505125248d-05*zm06
     $  + 6.753341474714677d-06*zm07
     $  - 5.550174516283685d-05*zm08
     $  - 3.387816551261443d-05*zm09
     $  - 5.616972301478889d-06*zm10
     $  - 2.381958558786150d-07*zm11
     $  - 1.489538543667777d-09*zm12
     $  - 2.793314031649452d-13*zm13
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,14)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,14)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,14)
     $  + HZ3(-1,0,1)*CbH1YZ(3,14)
     $  + HZ3(0,-1,1)*CbH1YZ(3,14)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,1,2,1,15) = 
     $  + 1.895388896407933d-14
     $  - 2.748861547509540d-13*zm01
     $  - 8.313436260505394d-11*zm02
     $  - 2.281139529249211d-08*zm03
     $  - 8.974557017503064d-07*zm04
     $  - 9.420129221283219d-06*zm05
     $  - 3.172870667020158d-05*zm06
     $  - 2.743564200176201d-05*zm07
     $  + 2.131995295277978d-05*zm08
     $  + 3.464874734805066d-05*zm09
     $  + 1.216807675702440d-05*zm10
     $  + 1.330095324190062d-06*zm11
     $  + 3.780647202919056d-08*zm12
     $  + 1.494890735551949d-10*zm13
     $  + 3.743682669834299d-14*zm14
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,15)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,15)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,15)
     $  + HZ3(-1,0,1)*CbH1YZ(3,15)
     $  + HZ3(0,-1,1)*CbH1YZ(3,15)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,1,2,1,16) = 
     $  - 2.585019011430134d-16
     $  + 4.413680527158036d-15*zm01
     $  + 7.937208571556300d-12*zm02
     $  + 3.288951682911295d-09*zm03
     $  + 1.858399407854180d-07*zm04
     $  + 2.790953162656389d-06*zm05
     $  + 1.417940862513318d-05*zm06
     $  + 2.416718976060670d-05*zm07
     $  + 2.695766292608538d-06*zm08
     $  - 2.329719738133547d-05*zm09
     $  - 1.665625833785998d-05*zm10
     $  - 3.779305120282750d-06*zm11
     $  - 2.841396635230826d-07*zm12
     $  - 5.540068239362652d-09*zm13
     $  - 1.410389008395714d-11*zm14
     $  + 2.938516744107520d-16*zm15
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,16)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,16)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,16)
     $  + HZ3(-1,0,1)*CbH1YZ(3,16)
     $  + HZ3(0,-1,1)*CbH1YZ(3,16)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,1,2,1,17) = 
     $  - 2.507868083077688d-16
     $  + 4.062434406541067d-15*zm01
     $  - 7.345039342475623d-13*zm02
     $  - 4.429668363988898d-10*zm03
     $  - 3.539042325651960d-08*zm04
     $  - 7.402782193032896d-07*zm05
     $  - 5.376921093224132d-06*zm06
     $  - 1.463992986804417d-05*zm07
     $  - 1.142780732461082d-05*zm08
     $  + 8.825546247016038d-06*zm09
     $  + 1.561630693457162d-05*zm10
     $  + 6.680520462116383d-06*zm11
     $  + 1.042128689337595d-06*zm12
     $  + 5.551133053889441d-08*zm13
     $  + 7.557112645932001d-10*zm14
     $  + 1.251474147363906d-12*zm15
     $  - 3.516541864928189d-16*zm16
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,17)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,17)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,17)
     $  + HZ3(-1,0,1)*CbH1YZ(3,17)
     $  + HZ3(0,-1,1)*CbH1YZ(3,17)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,1,2,1,18) = 
     $  + 2.527509875765344d-19
     $  - 6.090798150134385d-18*zm01
     $  + 5.871023398112613d-14*zm02
     $  + 5.609100272849113d-11*zm03
     $  + 6.255277503805130d-09*zm04
     $  + 1.786128353896743d-07*zm05
     $  + 1.790327303622743d-06*zm06
     $  + 7.120673667765741d-06*zm07
     $  + 1.059831815222758d-05*zm08
     $  + 1.121633909516441d-06*zm09
     $  - 1.008120038260021d-05*zm10
     $  - 8.136684002372408d-06*zm11
     $  - 2.328121141933477d-06*zm12
     $  - 2.597536822435680d-07*zm13
     $  - 1.002139666133954d-08*zm14
     $  - 9.658605494318441d-11*zm15
     $  - 1.038629419152266d-13*zm16
     $  - 4.218807312974805d-18*zm17
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,18)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,18)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,18)
     $  + HZ3(-1,0,1)*CbH1YZ(3,18)
     $  + HZ3(0,-1,1)*CbH1YZ(3,18)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,1,2,1,19) = 
     $  + 3.299526360655451d-18
     $  - 6.022868643566904d-17*zm01
     $  - 4.110920727395944d-15*zm02
     $  - 6.704565208759499d-12*zm03
     $  - 1.033685444148115d-09*zm04
     $  - 3.966657753722736d-08*zm05
     $  - 5.349713262226985d-07*zm06
     $  - 2.954418484066294d-06*zm07
     $  - 6.825278095445227d-06*zm08
     $  - 4.915909967380221d-06*zm09
     $  + 3.778260354101000d-06*zm10
     $  + 7.148669478886592d-06*zm11
     $  + 3.560617377496491d-06*zm12
     $  + 7.227433928061899d-07*zm13
     $  + 5.929724714523627d-08*zm14
     $  + 1.685360546116476d-09*zm15
     $  + 1.162571799197369d-11*zm16
     $  + 8.124382659896390d-15*zm17
     $  + 4.874141765153969d-18*zm18
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,19)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,19)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,19)
     $  + HZ3(-1,0,1)*CbH1YZ(3,19)
     $  + HZ3(0,-1,1)*CbH1YZ(3,19)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,19)

      CbH4YZ(3,1,2,1,20) = 
     $  + 5.253441013323540d-20
     $  - 9.955030367243574d-19*zm01
     $  + 3.562262682312064d-16*zm02
     $  + 7.586404375682558d-13*zm03
     $  + 1.606432173917253d-10*zm04
     $  + 8.182264222335432d-09*zm05
     $  + 1.456667879905788d-07*zm06
     $  + 1.080422549363746d-06*zm07
     $  + 3.553990330266339d-06*zm08
     $  + 4.750869208431134d-06*zm09
     $  + 4.821018895987318d-07*zm10
     $  - 4.471671967848458d-06*zm11
     $  - 3.973329890351265d-06*zm12
     $  - 1.360293922256660d-06*zm13
     $  - 2.033090127336597d-07*zm14
     $  - 1.252254875839078d-08*zm15
     $  - 2.657660603358471d-10*zm16
     $  - 1.323462411978103d-12*zm17
     $  - 6.147335842640955d-16*zm18
     $  - 6.342095877727846d-20*zm19
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,20)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,20)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,20)
     $  + HZ3(-1,0,1)*CbH1YZ(3,20)
     $  + HZ3(0,-1,1)*CbH1YZ(3,20)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,20)

      CbH4YZ(3,1,2,1,21) = 
     $  - 2.694561786172124d-20
     $  + 5.413296692593885d-19*zm01
     $  - 2.997324342698568d-17*zm02
     $  - 8.161688976041870d-14*zm03
     $  - 2.359104170386100d-11*zm04
     $  - 1.579023289885229d-09*zm05
     $  - 3.655538653820639d-08*zm06
     $  - 3.555665144475647d-07*zm07
     $  - 1.586749169991169d-06*zm08
     $  - 3.215550313032205d-06*zm09
     $  - 2.170343058878851d-06*zm10
     $  + 1.661674585821201d-06*zm11
     $  + 3.319025943401881d-06*zm12
     $  + 1.866118057316256d-06*zm13
     $  + 4.645675914292351d-07*zm14
     $  + 5.247544543957965d-08*zm15
     $  + 2.465866773038189d-09*zm16
     $  + 3.950561044425386d-11*zm17
     $  + 1.430297446713271d-13*zm18
     $  + 4.458380629934038d-17*zm19
     $  - 2.915043921906001d-20*zm20
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,21)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,21)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,21)
     $  + HZ3(-1,0,1)*CbH1YZ(3,21)
     $  + HZ3(0,-1,1)*CbH1YZ(3,21)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,21)

      CbH4YZ(3,1,2,1,22) = 
     $  + 1.434558708650260d-20
     $  - 3.090885630011808d-19*zm01
     $  + 4.864697582747481d-18*zm02
     $  + 8.355668840607055d-15*zm03
     $  + 3.287022123484410d-12*zm04
     $  + 2.867546905684177d-10*zm05
     $  + 8.529447993772539d-09*zm06
     $  + 1.068520836567902d-07*zm07
     $  + 6.268025408372421d-07*zm08
     $  + 1.771027036048657d-06*zm09
     $  + 2.169423292144408d-06*zm10
     $  + 2.127128276184245d-07*zm11
     $  - 2.024587461574754d-06*zm12
     $  - 1.945235851738203d-06*zm13
     $  - 7.686574781082209d-07*zm14
     $  - 1.441478628469160d-07*zm15
     $  - 1.254738898696897d-08*zm16
     $  - 4.556601422777706d-10*zm17
     $  - 5.560245164578124d-12*zm18
     $  - 1.472729596653319d-14*zm19
     $  - 2.392873568622672d-18*zm20
     $  - 3.050744279308798d-20*zm21
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,22)
     $  + 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,22)
     $  + 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,22)
     $  + HZ3(-1,0,1)*CbH1YZ(3,22)
     $  + HZ3(0,-1,1)*CbH1YZ(3,22)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,22)

      CbH4YZ(3,2,0,1,01) = 
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,1)
     $  + HZ3(1,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,2,0,1,02) = 
     $  - 4.626377063010636d-01
     $  + 4.626377063010636d-01*zm01
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,2)
     $  + HZ3(1,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,2,0,1,03) = 
     $  - 7.031250000000000d-02
     $  + 3.719438531505318d-01*zm01
     $  - 3.016313531505318d-01*zm02
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,3)
     $  + HZ3(1,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,2,0,1,04) = 
     $  - 6.591796875000000d-03
     $  - 5.732255573587080d-03*zm01
     $  - 9.887695312500000d-02*zm02
     $  + 1.112010055735870d-01*zm03
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,4)
     $  + HZ3(1,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,2,0,1,05) = 
     $  - 3.295898437500000d-04
     $  + 6.419888614717416d-03*zm01
     $  + 3.705751740665224d-02*zm02
     $  - 1.398622834415224d-02*zm03
     $  - 2.916158783346741d-02*zm04
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,5)
     $  + HZ3(1,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,2,0,1,06) = 
     $  - 5.911202913620962d-04*zm01
     $  - 1.373896170268346d-02*zm02
     $  - 1.359558105468750d-02*zm03
     $  + 2.197870779643346d-02*zm04
     $  + 5.946955252299596d-03*zm05
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,6)
     $  + HZ3(1,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,2,0,1,07) = 
     $  + 8.828299386160714d-07
     $  + 5.746175160704630d-05*zm01
     $  + 3.436153601141559d-03*zm02
     $  + 1.215968047862337d-02*zm03
     $  - 4.483474162356629d-03*zm04
     $  - 1.018185716210696d-02*zm05
     $  - 9.888473368470016d-04*zm06
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,7)
     $  + HZ3(1,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,2,0,1,08) = 
     $  - 5.995200508412291d-06*zm01
     $  - 6.937739454510687d-04*zm02
     $  - 5.535911298022925d-03*zm03
     $  - 3.928041458129882d-03*zm04
     $  + 6.822084164844214d-03*zm05
     $  + 3.203549107194232d-03*zm06
     $  + 1.380886300738419d-04*zm07
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,8)
     $  + HZ3(1,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,2,0,1,09) = 
     $  - 4.926506353884327d-09
     $  + 5.414124388683408d-07*zm01
     $  + 1.206257235053618d-04*zm02
     $  + 1.836257666856571d-03*zm03
     $  + 4.157374575131592d-03*zm04
     $  - 1.412256263658290d-03*zm05
     $  - 3.892258240151074d-03*zm06
     $  - 7.937441396492852d-04*zm07
     $  - 1.653580796738919d-05*zm08
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,9)
     $  + HZ3(1,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,2,0,1,10) = 
     $  - 3.709107907493046d-08*zm01
     $  - 1.856009380696400d-05*zm02
     $  - 4.950125094257093d-04*zm03
     $  - 2.262843459525450d-03*zm04
     $  - 1.304407098463603d-03*zm05
     $  + 2.375532562924455d-03*zm06
     $  + 1.538736824932780d-03*zm07
     $  + 1.648649177213167d-04*zm08
     $  + 1.725946722250016d-06*zm09
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,10)
     $  + HZ3(1,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,2,0,1,11) = 
     $  + 3.526930685167188d-11
     $  + 2.163326274001959d-09*zm01
     $  + 2.565968698992418d-06*zm02
     $  + 1.142397229520166d-04*zm03
     $  + 9.058175406760478d-04*zm04
     $  + 1.540273403724757d-03*zm05
     $  - 4.945900192993036d-04*zm06
     $  - 1.555750099239670d-03*zm07
     $  - 4.826685706060394d-04*zm08
     $  - 2.973091787575288d-05*zm09
     $  - 1.592276266293116d-07*zm10
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,11)
     $  + HZ3(1,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,2,0,1,12) = 
     $  - 1.608525055678969d-10*zm01
     $  - 3.228102875746127d-07*zm02
     $  - 2.326130187219473d-05*zm03
     $  - 2.958499519163456d-04*zm04
     $  - 9.493202472833303d-04*zm05
     $  - 4.747777830011078d-04*zm06
     $  + 8.941898238375436d-04*zm07
     $  + 7.169750740152440d-04*zm08
     $  + 1.275927673672227d-04*zm09
     $  + 4.761446256767010d-06*zm10
     $  + 1.314373628128868d-08*zm11
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,12)
     $  + HZ3(1,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,2,0,1,13) = 
     $  - 2.890295032643565d-13
     $  + 1.282890041706537d-11*zm01
     $  + 3.737524688242454d-08*zm02
     $  + 4.262513734370379d-06*zm03
     $  + 8.287135695070270d-05*zm04
     $  + 4.344633512386406d-04*zm05
     $  + 6.055015322966819d-04*zm06
     $  - 1.867978211895226d-04*zm07
     $  - 6.456679388170598d-04*zm08
     $  - 2.645611513798355d-04*zm09
     $  - 2.942030912542891d-05*zm10
     $  - 6.879406074993246d-07*zm11
     $  - 9.808878026790862d-10*zm12
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,13)
     $  + HZ3(1,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,2,0,1,14) = 
     $  - 4.748079377654854d-13*zm01
     $  - 4.015538246332496d-09*zm02
     $  - 7.128949132759721d-07*zm03
     $  - 2.050563738153119d-05*zm04
     $  - 1.631831654879672d-04*zm05
     $  - 4.087059495214095d-04*zm06
     $  - 1.843416945772191d-04*zm07
     $  + 3.559795107769173d-04*zm08
     $  + 3.325695950831150d-04*zm09
     $  + 8.276242232915158d-05*zm10
     $  + 6.051058240541183d-06*zm11
     $  + 9.070474418523066d-08*zm12
     $  + 6.672054704279475d-11*zm13
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,14)
     $  + HZ3(1,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,2,0,1,15) = 
     $  + 2.586632560022418d-15
     $  - 1.303076797794943d-14*zm01
     $  + 4.022479801153557d-10*zm02
     $  + 1.099683715926604d-07*zm03
     $  + 4.571006300245923d-06*zm04
     $  + 5.276088032185263d-05*zm05
     $  + 2.068202484909049d-04*zm06
     $  + 2.489093394712030d-04*zm07
     $  - 7.452830937409244d-05*zm08
     $  - 2.764569902054207d-04*zm09
     $  - 1.383676613054289d-04*zm10
     $  - 2.268014098188752d-05*zm11
     $  - 1.127730013598524d-06*zm12
     $  - 1.100914818578217d-08*zm13
     $  - 4.164721118071205d-12*zm14
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,15)
     $  + HZ3(1,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,2,0,1,16) = 
     $  - 1.302986435460910d-15*zm01
     $  - 3.770798172512488d-11*zm02
     $  - 1.577210733324713d-08*zm03
     $  - 9.309298191287303d-07*zm04
     $  - 1.510666744076729d-05*zm05
     $  - 8.644416262416670d-05*zm06
     $  - 1.800825536545483d-04*zm07
     $  - 7.508614543179104d-05*zm08
     $  + 1.478290198132883d-04*zm09
     $  + 1.549791735746147d-04*zm10
     $  + 4.910168105914225d-05*zm11
     $  + 5.562483101887694d-06*zm12
     $  + 1.926725144839269d-07*zm13
     $  + 1.238483309442187d-09*zm14
     $  + 2.402936826882039d-13*zm15
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,16)
     $  + HZ3(1,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,2,0,1,17) = 
     $  - 2.465583710636314d-17
     $  + 4.561726922807413d-16*zm01
     $  + 3.325355920973646d-12*zm02
     $  + 2.116669937591706d-09*zm03
     $  + 1.750496622569546d-07*zm04
     $  + 3.903643544030586d-06*zm05
     $  + 3.123923297572401d-05*zm06
     $  + 9.855156542039270d-05*zm07
     $  + 1.059479937613599d-04*zm08
     $  - 3.099711429577687d-05*zm09
     $  - 1.214719459737684d-04*zm10
     $  - 7.074988501295887d-05*zm11
     $  - 1.533043912267977d-05*zm12
     $  - 1.239643220553403d-06*zm13
     $  - 3.044786695646866d-08*zm14
     $  - 1.298538852061222d-10*zm15
     $  - 1.291020513091890d-14*zm16
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,17)
     $  + HZ3(1,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,0,1,18) = 
     $  - 6.900928026265356d-19*zm01
     $  - 2.774462580928227d-13*zm02
     $  - 2.671810955693785d-10*zm03
     $  - 3.064063491354709d-08*zm04
     $  - 9.228521736679497d-07*zm05
     $  - 1.002850974437493d-05*zm06
     $  - 4.482821904430459d-05*zm07
     $  - 8.096317648533024d-05*zm08
     $  - 3.173477569687646d-05*zm09
     $  + 6.344867111936290d-05*zm10
     $  + 7.279505342156396d-05*zm11
     $  + 2.770849218448095d-05*zm12
     $  + 4.297832954774841d-06*zm13
     $  + 2.538966385952101d-07*zm14
     $  + 4.482169687842110d-09*zm15
     $  + 1.274889719469710d-11*zm16
     $  + 6.473425832697350d-16*zm17
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,18)
     $  + HZ3(1,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,2,0,1,19) = 
     $  + 2.465174779238043d-19
     $  - 4.443727443001002d-18*zm01
     $  + 2.194300524410708d-14*zm02
     $  + 3.185816047106869d-11*zm03
     $  + 5.025318711349793d-09*zm04
     $  + 2.016622190409474d-07*zm05
     $  + 2.912102776814102d-06*zm06
     $  + 1.771058983337543d-05*zm07
     $  + 4.717939571478136d-05*zm08
     $  + 4.637209167546056d-05*zm09
     $  - 1.331894891915871d-05*zm10
     $  - 5.453935820755451d-05*zm11
     $  - 3.578727017722454d-05*zm12
     $  - 9.588808040310344d-06*zm13
     $  - 1.097680907556623d-06*zm14
     $  - 4.821382653239926d-08*zm15
     $  - 6.181631872402092d-10*zm16
     $  - 1.176728474298430d-12*zm17
     $  - 3.018173262179648d-17*zm18
     $  - 1.644934066848226d+00*HZ1(1)*CbH1YZ(3,19)
     $  + HZ3(1,0,1)*CbH1YZ(3,19)

      CbH4YZ(3,2,1,1,01) = 
     $  + HZ3(0,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,2,1,1,02) = 
     $  + 2.812500000000000d-01
     $  - 2.812500000000000d-01*zm01
     $  + HZ3(0,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,2,1,1,03) = 
     $  + 5.273437500000000d-02
     $  - 2.460937500000000d-01*zm01
     $  + 1.933593750000000d-01*zm02
     $  + HZ3(0,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,2,1,1,04) = 
     $  + 6.225585937500000d-03
     $  - 8.789062500000000d-03*zm01
     $  + 7.800292968750000d-02*zm02
     $  - 7.543945312500000d-02*zm03
     $  + HZ3(0,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,2,1,1,05) = 
     $  + 4.119873046875000d-04
     $  - 4.394531250000000d-03*zm01
     $  - 1.895141601562500d-02*zm02
     $  + 1.977539062500000d-03*zm03
     $  + 2.095642089843750d-02*zm04
     $  + HZ3(0,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,2,1,1,06) = 
     $  + 3.845214843750000d-06
     $  + 2.725982666015625d-04*zm01
     $  + 7.385559082031250d-03*zm02
     $  + 9.540252685546875d-03*zm03
     $  - 1.268371582031250d-02*zm04
     $  - 4.518539428710937d-03*zm05
     $  + HZ3(0,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,2,1,1,07) = 
     $  - 1.544952392578125d-06
     $  - 2.152633666992187d-05*zm01
     $  - 1.800384521484375d-03*zm02
     $  - 7.163134983607700d-03*zm03
     $  + 1.542818886893136d-03*zm04
     $  + 6.653079986572265d-03*zm05
     $  + 7.906919206891741d-04*zm06
     $  + HZ3(0,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,2,1,1,08) = 
     $  - 4.279022661659529d-08
     $  + 3.331500656750737d-06*zm01
     $  + 3.528375674267204d-04*zm02
     $  + 3.075663240588441d-03*zm03
     $  + 2.817074863278135d-03*zm04
     $  - 3.890160565473595d-03*zm05
     $  - 2.243319350845959d-03*zm06
     $  - 1.153844654038765d-04*zm07
     $  + HZ3(0,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,2,1,1,09) = 
     $  + 1.108463929623973d-08
     $  - 3.427018577086334d-07*zm01
     $  - 6.074273951199589d-05*zm02
     $  - 9.852619572561614d-04*zm03
     $  - 2.518213281826097d-03*zm04
     $  + 5.072749694999383d-04*zm05
     $  + 2.458401652014985d-03*zm06
     $  + 5.845584540951008d-04*zm07
     $  + 1.431452020264227d-05*zm08
     $  + HZ3(0,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,2,1,1,10) = 
     $  + 4.877241290345483d-10
     $  + 1.383099064187475d-08*zm01
     $  + 9.360306473237084d-06*zm02
     $  + 2.596968560194482d-04*zm03
     $  + 1.289813077145693d-03*zm04
     $  + 9.447627176131520d-04*zm05
     $  - 1.342687693359900d-03*zm06
     $  - 1.033094752382258d-03*zm07
     $  - 1.263305192612350d-04*zm08
     $  - 1.534310962907079d-06*zm09
     $  + HZ3(0,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,2,1,1,11) = 
     $  - 9.699059384209769d-11
     $  - 2.703860228332764d-10*zm01
     $  - 1.291328015853990d-06*zm02
     $  - 5.907555446660616d-05*zm03
     $  - 4.982831584161807d-04*zm04
     $  - 9.479753609857278d-04*zm05
     $  + 1.805170249933435d-04*zm06
     $  + 9.634113384186115d-04*zm07
     $  + 3.390256461112992d-04*zm08
     $  + 2.352743564672687d-05*zm09
     $  + 1.443240910041815d-07*zm10
     $  + HZ3(0,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,2,1,1,12) = 
     $  - 5.661341413666143d-12
     $  + 1.458549152438495d-10*zm01
     $  + 1.613810019114502d-07*zm02
     $  + 1.191345261496620d-05*zm03
     $  + 1.589119955375222d-04*zm04
     $  + 5.510848988010222d-04*zm05
     $  + 3.459697432748405d-04*zm06
     $  - 5.023731936500287d-04*zm07
     $  - 4.688718646614452d-04*zm08
     $  - 9.291469827553249d-05*zm09
     $  - 3.869757012274848d-06*zm10
     $  - 1.209782455513981d-08*zm11
     $  + HZ3(0,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,2,1,1,13) = 
     $  + 9.393458856091589d-13
     $  - 1.605649656979671d-11*zm01
     $  - 1.865446559703880d-08*zm02
     $  - 2.168003112279123d-06*zm03
     $  - 4.376496297418292d-05*zm04
     $  - 2.435694275664058d-04*zm05
     $  - 3.767797702786712d-04*zm06
     $  + 6.886227151125393d-05*zm07
     $  + 3.945970345797244d-04*zm08
     $  + 1.801923121639104d-04*zm09
     $  + 2.207647648719872d-05*zm10
     $  + 5.718235618469732d-07*zm11
     $  + 9.152103523543046d-10*zm12
     $  + HZ3(0,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,2,1,1,14) = 
     $  + 6.681198657913456d-14
     $  - 6.625736796245896d-13*zm01
     $  + 2.014573477276632d-09*zm02
     $  + 3.607726892630700d-07*zm03
     $  + 1.069465296277285d-05*zm04
     $  + 8.929796379694925d-05*zm05
     $  + 2.404680258190597d-04*zm06
     $  + 1.349006684590475d-04*zm07
     $  - 1.991433122314432d-04*zm08
     $  - 2.135880456803431d-04*zm09
     $  - 5.825785066199586d-05*zm10
     $  - 4.657964187921840d-06*zm11
     $  - 7.686201037988780d-08*zm12
     $  - 6.293272409663106d-11*zm13
     $  + HZ3(0,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,2,1,1,15) = 
     $  - 9.699872100084071d-15
     $  + 1.247451353370347d-13*zm01
     $  - 2.019603577340168d-10*zm02
     $  - 5.545823759499373d-08*zm03
     $  - 2.361614319230415d-06*zm04
     $  - 2.835880373515686d-05*zm05
     $  - 1.176492031303077d-04*zm06
     $  - 1.561451944092198d-04*zm07
     $  + 2.767049750153122d-05*zm08
     $  + 1.673661186291012d-04*zm09
     $  + 9.222173286947825d-05*zm10
     $  + 1.641515550706661d-05*zm11
     $  + 8.874813158310633d-07*zm12
     $  + 9.485904993906272d-09*zm13
     $  + 3.948820093967175d-12*zm14
     $  + HZ3(0,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,2,1,1,16) = 
     $  - 8.005170067599970d-16
     $  + 1.298054945284806d-14*zm01
     $  + 1.876855814809380d-11*zm02
     $  + 7.935140873991568d-09*zm03
     $  + 4.775094223415808d-07*zm04
     $  + 8.008574335530035d-06*zm05
     $  + 4.801249404602347d-05*zm06
     $  + 1.070482095516029d-04*zm07
     $  + 5.512125300177551d-05*zm08
     $  - 8.243697161540978d-05*zm09
     $  - 9.822687353363462d-05*zm10
     $  - 3.373258082519473d-05*zm11
     $  - 4.123897723683016d-06*zm12
     $  - 1.545876559875512d-07*zm13
     $  - 1.082696391220340d-09*zm14
     $  - 2.285848362064537d-13*zm15
     $  + HZ3(0,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,2,1,1,17) = 
     $  + 1.047873077020433d-16
     $  - 1.708853231428238d-15*zm01
     $  - 1.651725572105252d-12*zm02
     $  - 1.063031134939355d-09*zm03
     $  - 8.929191606068496d-08*zm04
     $  - 2.047151259022046d-06*zm05
     $  - 1.703841281815931d-05*zm06
     $  - 5.670791012194338d-05*zm07
     $  - 6.687721055314744d-05*zm08
     $  + 1.157011367383068d-05*zm09
     $  + 7.301799112985010d-05*zm10
     $  + 4.640188124026401d-05*zm11
     $  + 1.080743702975477d-05*zm12
     $  + 9.386524639979976d-07*zm13
     $  + 2.485084034044671d-08*zm14
     $  + 1.149622957614133d-10*zm15
     $  + 1.246366234515635d-14*zm16
     $  + HZ3(0,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,1,1,18) = 
     $  + 9.723038811891347d-18
     $  - 1.683137847815903d-16*zm01
     $  + 1.401076674652707d-13*zm02
     $  + 1.339994828479047d-10*zm03
     $  + 1.556233003470604d-08*zm04
     $  + 4.797988206889173d-07*zm05
     $  + 5.391848195366329d-06*zm06
     $  + 2.520006644159026d-05*zm07
     $  + 4.851873397165021d-05*zm08
     $  + 2.335352625130401d-05*zm09
     $  - 3.529569807897629d-05*zm10
     $  - 4.567564210686438d-05*zm11
     $  - 1.868976060115335d-05*zm12
     $  - 3.098954079719678d-06*zm13
     $  - 1.958894537804074d-07*zm14
     $  - 3.714416259882810d-09*zm15
     $  - 1.141267537684674d-11*zm16
     $  - 6.370112003033208d-16*zm17
     $  + HZ3(0,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,2,1,1,19) = 
     $  - 1.170958020138070d-18
     $  + 2.105925208386484d-17*zm01
     $  - 1.113162243787890d-14*zm02
     $  - 1.596335191097907d-11*zm03
     $  - 2.543761188511240d-09*zm04
     $  - 1.041187170435296d-07*zm05
     $  - 1.547677429028011d-06*zm06
     $  - 9.778127485385190d-06*zm07
     $  - 2.740001900043942d-05*zm08
     $  - 2.941561042506023d-05*zm09
     $  + 4.992244421397104d-06*zm10
     $  + 3.260315234771001d-05*zm11
     $  + 2.318153011935026d-05*zm12
     $  + 6.625347105202969d-06*zm13
     $  + 8.074892319594474d-07*zm14
     $  + 3.782918138132477d-08*zm15
     $  + 5.193218160466049d-10*zm16
     $  + 1.063762836829101d-12*zm17
     $  + 2.854578031200260d-17*zm18
     $  + HZ3(0,0,1)*CbH1YZ(3,19)

      CbH4YZ(3,2,1,1,20) = 
     $  - 1.195305817510443d-19
     $  + 2.315666689473286d-18*zm01
     $  + 8.002111199202107d-16*zm02
     $  + 1.803651467268058d-12*zm03
     $  + 3.919583056584110d-10*zm04
     $  + 2.108131096091210d-08*zm05
     $  + 4.079832145221261d-07*zm06
     $  + 3.393689085145540d-06*zm07
     $  + 1.305126207537887d-05*zm08
     $  + 2.234274335767725d-05*zm09
     $  + 1.018390755431530d-05*zm10
     $  - 1.552557259902471d-05*zm11
     $  - 2.147466395005425d-05*zm12
     $  - 1.009085040072618d-05*zm13
     $  - 2.109706984334292d-06*zm14
     $  - 1.933882520749314d-07*zm15
     $  - 6.809826154612563d-09*zm16
     $  - 6.825495740543571d-11*zm17
     $  - 9.343199831217305d-14*zm18
     $  - 1.144528626572777d-18*zm19
     $  + HZ3(0,0,1)*CbH1YZ(3,20)

      CbH4YZ(3,2,1,1,21) = 
     $  + 1.343497820787653d-20
     $  - 2.685820024275691d-19*zm01
     $  - 5.612780619517056d-17*zm02
     $  - 1.938186965439525d-13*zm03
     $  - 5.717699197908885d-11*zm04
     $  - 4.007381015072215d-09*zm05
     $  - 9.970911037410608d-08*zm06
     $  - 1.070867193663376d-06*zm07
     $  - 5.450374129973300d-06*zm08
     $  - 1.329883068052327d-05*zm09
     $  - 1.322480920665138d-05*zm10
     $  + 2.209452180037513d-06*zm11
     $  + 1.484215935221288d-05*zm12
     $  + 1.155738104767411d-05*zm13
     $  + 3.883327895530075d-06*zm14
     $  + 6.122287498130906d-07*zm15
     $  + 4.294781649859432d-08*zm16
     $  + 1.149555354267954d-09*zm17
     $  + 8.468189406740687d-12*zm18
     $  + 7.757559195818219d-15*zm19
     $  + 6.462200425362374d-20*zm20
     $  + HZ3(0,0,1)*CbH1YZ(3,21)

      CbH4YZ(3,2,2,1,01) = 
     $  - HZ3(0,0,1)*CbH1YZ(3,1)
     $  + HZ3(0,1,1)*CbH1YZ(3,1)

      CbH4YZ(3,2,2,1,02) = 
     $  - 2.812500000000000d-01
     $  + 2.812500000000000d-01*zm01
     $  - HZ3(0,0,1)*CbH1YZ(3,2)
     $  + HZ3(0,1,1)*CbH1YZ(3,2)

      CbH4YZ(3,2,2,1,03) = 
     $  - 8.789062500000000d-02
     $  + 3.164062500000000d-01*zm01
     $  - 2.285156250000000d-01*zm02
     $  - HZ3(0,0,1)*CbH1YZ(3,3)
     $  + HZ3(0,1,1)*CbH1YZ(3,3)

      CbH4YZ(3,2,2,1,04) = 
     $  - 1.281738281250000d-02
     $  + 4.833984375000000d-02*zm01
     $  - 1.373291015625000d-01*zm02
     $  + 1.018066406250000d-01*zm03
     $  - HZ3(0,0,1)*CbH1YZ(3,4)
     $  + HZ3(0,1,1)*CbH1YZ(3,4)

      CbH4YZ(3,2,2,1,05) = 
     $  - 9.063720703125000d-04
     $  + 4.394531250000000d-03*zm01
     $  + 1.005249023437500d-02*zm02
     $  + 1.779785156250000d-02*zm03
     $  - 3.133850097656250d-02*zm04
     $  - HZ3(0,0,1)*CbH1YZ(3,5)
     $  + HZ3(0,1,1)*CbH1YZ(3,5)

      CbH4YZ(3,2,2,1,06) = 
     $  - 3.845214843750000d-06
     $  + 3.639221191406250d-05*zm01
     $  - 3.677673339843750d-03*zm02
     $  - 1.510208129882812d-02*zm03
     $  + 1.144775390625000d-02*zm04
     $  + 7.299453735351562d-03*zm05
     $  - HZ3(0,0,1)*CbH1YZ(3,6)
     $  + HZ3(0,1,1)*CbH1YZ(3,6)

      CbH4YZ(3,2,2,1,07) = 
     $  + 3.752027239118303d-06
     $  - 2.482223510742187d-05*zm01
     $  + 6.085641043526785d-04*zm02
     $  + 6.787932259695870d-03*zm03
     $  + 2.396809714181082d-03*zm04
     $  - 8.420946938650948d-03*zm05
     $  - 1.351288931710379d-03*zm06
     $  - HZ3(0,0,1)*CbH1YZ(3,7)
     $  + HZ3(0,1,1)*CbH1YZ(3,7)

      CbH4YZ(3,2,2,1,08) = 
     $  + 4.279022661659529d-08
     $  - 4.347149206667530d-07*zm01
     $  - 6.605577955440599d-05*zm02
     $  - 2.076272161639466d-03*zm03
     $  - 4.526178447567686d-03*zm04
     $  + 3.238383774854698d-03*zm05
     $  + 3.225329715378430d-03*zm06
     $  + 2.051848232224800d-04*zm07
     $  - HZ3(0,0,1)*CbH1YZ(3,8)
     $  + HZ3(0,1,1)*CbH1YZ(3,8)

      CbH4YZ(3,2,2,1,09) = 
     $  - 2.832741153483488d-08
     $  + 2.392452242770625d-07*zm01
     $  + 6.428006960421192d-06*zm02
     $  + 4.887046023290984d-04*zm03
     $  + 2.605116853908616d-03*zm04
     $  + 6.825797715965582d-04*zm05
     $  - 2.852123113311066d-03*zm06
     $  - 9.047567345658127d-04*zm07
     $  - 2.616030473055714d-05*zm08
     $  - HZ3(0,0,1)*CbH1YZ(3,9)
     $  + HZ3(0,1,1)*CbH1YZ(3,9)

      CbH4YZ(3,2,2,1,10) = 
     $  - 4.877241290345483d-10
     $  + 5.567128126544785d-09*zm01
     $  - 7.165047500293386d-07*zm02
     $  - 9.555773424858949d-05*zm03
     $  - 1.010309463435289d-03*zm04
     $  - 1.530003960856369d-03*zm05
     $  + 1.056449052813101d-03*zm06
     $  + 1.371910054038982d-03*zm07
     $  + 2.053662143712836d-04*zm08
     $  + 2.857262662913291d-06*zm09
     $  - HZ3(0,0,1)*CbH1YZ(3,10)
     $  + HZ3(0,1,1)*CbH1YZ(3,10)

      CbH4YZ(3,2,2,1,11) = 
     $  + 2.557024746746211d-10
     $  - 2.639331792429654d-09*zm01
     $  + 6.977557304633444d-08*zm02
     $  + 1.627307636448567d-05*zm03
     $  + 3.074866127941570d-04*zm04
     $  + 1.027241047245756d-03*zm05
     $  + 2.294505763059423d-04*zm06
     $  - 1.052938594237826d-03*zm07
     $  - 4.876945036445420d-04*zm08
     $  - 3.961314928890451d-05*zm09
     $  - 2.724574827963055d-07*zm10
     $  - HZ3(0,0,1)*CbH1YZ(3,11)
     $  + HZ3(0,1,1)*CbH1YZ(3,11)

      CbH4YZ(3,2,2,1,12) = 
     $  + 5.661341413666143d-12
     $  - 7.311196986227625d-11*zm01
     $  - 3.892525160344140d-09*zm02
     $  - 2.471782019164898d-06*zm03
     $  - 7.850339857163878d-05*zm04
     $  - 4.707177653140062d-04*zm05
     $  - 5.631366779028801d-04*zm06
     $  + 3.797220408716118d-04*zm07
     $  + 5.872522791165099d-04*zm08
     $  + 1.411465444960577d-04*zm09
     $  + 6.689637289991535d-06*zm10
     $  + 2.308200930775738d-08*zm11
     $  - HZ3(0,0,1)*CbH1YZ(3,12)
     $  + HZ3(0,1,1)*CbH1YZ(3,12)

      CbH4YZ(3,2,2,1,13) = 
     $  - 2.529008153563120d-12
     $  + 3.093573539784579d-11*zm01
     $  + 1.218013582872133d-10*zm02
     $  + 3.384410962787868d-07*zm03
     $  + 1.745112925550101d-05*zm04
     $  + 1.707867953125157d-04*zm05
     $  + 4.206852801831281d-04*zm06
     $  + 8.436350501196468d-05*zm07
     $  - 4.126650287740522d-04*zm08
     $  - 2.450433823955508d-04*zm09
     $  - 3.490606582997905d-05*zm10
     $  - 1.009063729491190d-06*zm11
     $  - 1.760338400489348d-09*zm12
     $  - HZ3(0,0,1)*CbH1YZ(3,13)
     $  + HZ3(0,1,1)*CbH1YZ(3,13)

      CbH4YZ(3,2,2,1,14) = 
     $  - 6.681198657913456d-14
     $  + 9.725578218756121d-13*zm01
     $  - 2.500668541178460d-11*zm02
     $  - 4.217709897341637d-08*zm03
     $  - 3.457834031298709d-06*zm04
     $  - 5.225231582089939d-05*zm05
     $  - 2.173472573771650d-04*zm06
     $  - 2.203605400986615d-04*zm07
     $  + 1.460999566315992d-04*zm08
     $  + 2.557358643607390d-04*zm09
     $  + 8.389077593894449d-05*zm10
     $  + 7.595539628558286d-06*zm11
     $  + 1.378902712359045d-07*zm12
     $  + 1.216968607776463d-10*zm13
     $  - HZ3(0,0,1)*CbH1YZ(3,14)
     $  + HZ3(0,1,1)*CbH1YZ(3,14)

      CbH4YZ(3,2,2,1,15) = 
     $  + 2.651298374022979d-14
     $  - 3.756279054116259d-13*zm01
     $  + 3.535079934248129d-12*zm02
     $  + 4.847733372919136d-09*zm03
     $  + 6.206175205727474d-07*zm04
     $  + 1.396425538789709d-05*zm05
     $  + 8.967135303913309d-05*zm06
     $  + 1.780420905979058d-04*zm07
     $  + 3.300097002804961d-05*zm08
     $  - 1.691853578209976d-04*zm09
     $  - 1.199462618072849d-04*zm10
     $  - 2.467210640243975d-05*zm11
     $  - 1.483155326075263d-06*zm12
     $  - 1.724846878278420d-08*zm13
     $  - 7.667315896451850d-12*zm14
     $  - HZ3(0,0,1)*CbH1YZ(3,15)
     $  + HZ3(0,1,1)*CbH1YZ(3,15)

      CbH4YZ(3,2,2,1,16) = 
     $  + 8.005170067599970d-16
     $  - 1.304213872757906d-14*zm01
     $  + 4.204498350001485d-14*zm02
     $  - 5.192263209665652d-10*zm03
     $  - 1.021114788653270d-07*zm04
     $  - 3.334573053412244d-06*zm05
     $  - 3.137270087317456d-05*zm06
     $  - 1.006835650377740d-04*zm07
     $  - 9.027762387618937d-05*zm08
     $  + 5.910340338628354d-05*zm09
     $  + 1.134808155644795d-04*zm10
     $  + 4.651158589734795d-05*zm11
     $  + 6.409720186248765d-06*zm12
     $  + 2.635772610900133d-07*zm13
     $  + 1.990774002434554d-09*zm14
     $  + 4.464808469519728d-13*zm15
     $  - HZ3(0,0,1)*CbH1YZ(3,16)
     $  + HZ3(0,1,1)*CbH1YZ(3,16)

      CbH4YZ(3,2,2,1,17) = 
     $  - 2.897060859997668d-16
     $  + 4.670270858077006d-15*zm01
     $  - 3.240204886828242d-14*zm02
     $  + 5.185552197367953d-11*zm03
     $  + 1.554438115944470d-08*zm04
     $  + 7.228913568815971d-07*zm05
     $  + 9.625018632522235d-06*zm06
     $  + 4.583776777835726d-05*zm07
     $  + 7.745228858266554d-05*zm08
     $  + 1.351812258970166d-05*zm09
     $  - 7.184763024870172d-05*zm10
     $  - 5.820477186160145d-05*zm11
     $  - 1.557695309043060d-05*zm12
     $  - 1.499032466674934d-06*zm13
     $  - 4.308408611961880d-08*zm14
     $  - 2.133707792339073d-10*zm15
     $  - 2.448066576062319d-14*zm16
     $  - HZ3(0,0,1)*CbH1YZ(3,17)
     $  + HZ3(0,1,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,2,1,18) = 
     $  - 9.723038811891347d-18
     $  + 1.759055182987792d-16*zm01
     $  - 1.632309595262303d-15*zm02
     $  - 4.820485381573270d-12*zm03
     $  - 2.205232705171440d-09*zm04
     $  - 1.439562808369556d-07*zm05
     $  - 2.646071014822295d-06*zm06
     $  - 1.779721772337861d-05*zm07
     $  - 4.700136933321485d-05*zm08
     $  - 3.832580352976973d-05*zm09
     $  + 2.484676787208188d-05*zm10
     $  + 5.125710972308961d-05*zm11
     $  + 2.485982615307433d-05*zm12
     $  + 4.626441110208354d-06*zm13
     $  + 3.199306597656102d-07*zm14
     $  + 6.531066356214215d-09*zm15
     $  + 2.135085975101185d-11*zm16
     $  + 1.243387549260757d-15*zm17
     $  - HZ3(0,0,1)*CbH1YZ(3,18)
     $  + HZ3(0,1,1)*CbH1YZ(3,18)

      CbH4YZ(3,2,2,1,19) = 
     $  + 3.266356582490408d-18
     $  - 5.907609989767490d-17*zm01
     $  + 5.101316231991939d-16*zm02
     $  + 4.218537490660721d-13*zm03
     $  + 2.932213617123261d-10*zm04
     $  + 2.657384938508275d-08*zm05
     $  + 6.618708800243264d-07*zm06
     $  + 6.084317956562586d-06*zm07
     $  + 2.313505129989947d-05*zm08
     $  + 3.447828249042050d-05*zm09
     $  + 5.738672764267042d-06*zm10
     $  - 3.138448884893807d-05*zm11
     $  - 2.821784551460869d-05*zm12
     $  - 9.218174645052959d-06*zm13
     $  - 1.240666705617159d-06*zm14
     $  - 6.296097411910345d-08*zm15
     $  - 9.241922193868030d-10*zm16
     $  - 2.003618305742543d-12*zm17
     $  - 5.511353722512998d-17*zm18
     $  - HZ3(0,0,1)*CbH1YZ(3,19)
     $  + HZ3(0,1,1)*CbH1YZ(3,19)

      CbH4YZ(3,2,3,1,01) = 
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,1)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,1)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,1)
     $  + HZ3(-1,0,1)*CbH1YZ(3,1)
     $  - HZ3(0,-1,1)*CbH1YZ(3,1)

      CbH4YZ(3,2,3,1,02) = 
     $  - 1.585764359144373d-01
     $  + 1.585764359144373d-01*zm01
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,2)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,2)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,2)
     $  + HZ3(-1,0,1)*CbH1YZ(3,2)
     $  - HZ3(0,-1,1)*CbH1YZ(3,2)

      CbH4YZ(3,2,3,1,03) = 
     $  - 1.697064569097638d-02
     $  + 1.132295093391714d-01*zm01
     $  - 9.625886364819508d-02*zm02
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,3)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,3)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,3)
     $  + HZ3(-1,0,1)*CbH1YZ(3,3)
     $  - HZ3(0,-1,1)*CbH1YZ(3,3)

      CbH4YZ(3,2,3,1,04) = 
     $  - 5.583920096136407d-04
     $  - 1.107864707045261d-02*zm01
     $  - 2.076715243118935d-02*zm02
     $  + 3.240419151125561d-02*zm03
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,4)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,4)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,4)
     $  + HZ3(-1,0,1)*CbH1YZ(3,4)
     $  - HZ3(0,-1,1)*CbH1YZ(3,4)

      CbH4YZ(3,2,3,1,05) = 
     $  + 7.120010101627783d-05
     $  + 1.646400601444359d-03*zm01
     $  + 1.566577822942603d-02*zm02
     $  - 9.795785106688947d-03*zm03
     $  - 7.587593825197723d-03*zm04
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,5)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,5)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,5)
     $  + HZ3(-1,0,1)*CbH1YZ(3,5)
     $  - HZ3(0,-1,1)*CbH1YZ(3,5)

      CbH4YZ(3,2,3,1,06) = 
     $  + 6.091240652461769d-06
     $  - 2.798996087662689d-04*zm01
     $  - 5.327132843578969d-03*zm02
     $  - 3.907658360775991d-03*zm03
     $  + 8.160999008740839d-03*zm04
     $  + 1.347600563727928d-03*zm05
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,6)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,6)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,6)
     $  + HZ3(-1,0,1)*CbH1YZ(3,6)
     $  - HZ3(0,-1,1)*CbH1YZ(3,6)

      CbH4YZ(3,2,3,1,07) = 
     $  - 4.719572846933055d-07
     $  + 2.936634502994450d-05*zm01
     $  + 1.368986269773548d-03*zm02
     $  + 4.362163874278674d-03*zm03
     $  - 2.384868876981235d-03*zm04
     $  - 3.185004195255211d-03*zm05
     $  - 1.901714595610269d-04*zm06
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,7)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,7)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,7)
     $  + HZ3(-1,0,1)*CbH1YZ(3,7)
     $  - HZ3(0,-1,1)*CbH1YZ(3,7)

      CbH4YZ(3,2,3,1,08) = 
     $  - 6.946698286231185d-08
     $  - 1.943534194186698d-06*zm01
     $  - 2.846464621757431d-04*zm02
     $  - 2.100559133455173d-03*zm03
     $  - 1.065491797881444d-03*zm04
     $  + 2.547612627103039d-03*zm05
     $  + 8.830667888941946d-04*zm06
     $  + 2.203097869217619d-05*zm07
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,8)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,8)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,8)
     $  + HZ3(-1,0,1)*CbH1YZ(3,8)
     $  - HZ3(0,-1,1)*CbH1YZ(3,8)

      CbH4YZ(3,2,3,1,09) = 
     $  + 3.133550552456125d-09
     $  + 1.797034170981660d-07*zm01
     $  + 4.965841034299987d-05*zm02
     $  + 7.192262850330762d-04*zm03
     $  + 1.441154859123901d-03*zm04
     $  - 7.350764719418927d-04*zm05
     $  - 1.277904360083363d-03*zm06
     $  - 1.950797338582714d-04*zm07
     $  - 2.161825584100806d-06*zm08
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,9)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,9)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,9)
     $  + HZ3(-1,0,1)*CbH1YZ(3,9)
     $  - HZ3(0,-1,1)*CbH1YZ(3,9)

      CbH4YZ(3,2,3,1,10) = 
     $  + 7.959690687685520d-10
     $  - 2.301773485269171d-08*zm01
     $  - 7.609129895803690d-06*zm02
     $  - 1.975605709056943d-04*zm03
     $  - 8.376716071728645d-04*zm04
     $  - 3.451432827076208d-04*zm05
     $  + 8.932947655310216d-04*zm06
     $  + 4.582154494259108d-04*zm07
     $  + 3.631043293203172d-05*zm08
     $  + 1.861645588030683d-07*zm09
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,10)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,10)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,10)
     $  + HZ3(-1,0,1)*CbH1YZ(3,10)
     $  - HZ3(0,-1,1)*CbH1YZ(3,10)

      CbH4YZ(3,2,3,1,11) = 
     $  - 1.586947406402992d-11
     $  + 1.211046490189629d-09*zm01
     $  + 1.057186184871644d-06*zm02
     $  + 4.609677974694026d-05*zm03
     $  + 3.469453997115964d-04*zm04
     $  + 5.230507361292517d-04*zm05
     $  - 2.554042310252630d-04*zm06
     $  - 5.240326612366202d-04*zm07
     $  - 1.318093580835628d-04*zm08
     $  - 5.890517564559682d-06*zm09
     $  - 1.452903967032361d-08*zm10
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,11)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,11)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,11)
     $  + HZ3(-1,0,1)*CbH1YZ(3,11)
     $  - HZ3(0,-1,1)*CbH1YZ(3,11)

      CbH4YZ(3,2,3,1,12) = 
     $  - 9.068568553433812d-12
     $  + 3.855528704386090d-11*zm01
     $  - 1.338886033157843d-07*zm02
     $  - 9.454348465415038d-06*zm03
     $  - 1.157311312662778d-04*zm04
     $  - 3.446536111885839d-04*zm05
     $  - 1.237393632177370d-04*zm06
     $  + 3.379493977203636d-04*zm07
     $  + 2.227735533621230d-04*zm08
     $  + 3.213658030720866d-05*zm09
     $  + 8.517471158690143d-07*zm10
     $  + 1.034749046744160d-09*zm11
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,12)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,12)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,12)
     $  + HZ3(-1,0,1)*CbH1YZ(3,12)
     $  - HZ3(0,-1,1)*CbH1YZ(3,12)

      CbH4YZ(3,2,3,1,13) = 
     $  - 2.022425316817777d-14
     $  + 4.137185724161208d-12*zm01
     $  + 1.545322765967001d-08*zm02
     $  + 1.741678121006342d-06*zm03
     $  + 3.288087511107485d-05*zm04
     $  + 1.633979667157250d-04*zm05
     $  + 2.025837145599618d-04*zm06
     $  - 9.597491678741890d-05*zm07
     $  - 2.210549080341235d-04*zm08
     $  - 7.661817421866135d-05*zm09
     $  - 6.860163302202033d-06*zm10
     $  - 1.114640437721569d-07*zm11
     $  - 6.546621126575985d-11*zm12
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,13)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,13)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,13)
     $  + HZ3(-1,0,1)*CbH1YZ(3,13)
     $  - HZ3(0,-1,1)*CbH1YZ(3,13)

      CbH4YZ(3,2,3,1,14) = 
     $  + 1.021345828919518d-13
     $  - 1.579555620108746d-12*zm01
     $  - 1.651506819930104d-09*zm02
     $  - 2.923682282879783d-07*zm03
     $  - 8.218394428455920d-06*zm04
     $  - 6.277700493755120d-05*zm05
     $  - 1.461541439666999d-04*zm06
     $  - 4.753252252404842d-05*zm07
     $  + 1.350227974283333d-04*zm08
     $  + 1.061714107566429d-04*zm09
     $  + 2.245856090917418d-05*zm10
     $  + 1.309970553747776d-06*zm11
     $  + 1.334378505980821d-08*zm12
     $  + 3.636326313935896d-12*zm13
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,14)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,14)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,14)
     $  + HZ3(-1,0,1)*CbH1YZ(3,14)
     $  - HZ3(0,-1,1)*CbH1YZ(3,14)

      CbH4YZ(3,2,3,1,15) = 
     $  + 2.420143748858902d-15
     $  - 2.593851375634192d-14*zm01
     $  + 1.664810874212738d-10*zm02
     $  + 4.520690627664371d-08*zm03
     $  + 1.845585816992888d-06*zm04
     $  + 2.062198349736462d-05*zm05
     $  + 7.664332355443607d-05*zm06
     $  + 8.234044417309769d-05*zm07
     $  - 3.815148304485649d-05*zm08
     $  - 9.572542377107983d-05*zm09
     $  - 4.160507361028286d-05*zm10
     $  - 5.786082213966209d-06*zm11
     $  - 2.271742828426580d-07*zm12
     $  - 1.473283960665536d-09*zm13
     $  - 1.987482557069437d-13*zm14
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,15)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,15)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,15)
     $  + HZ3(-1,0,1)*CbH1YZ(3,15)
     $  - HZ3(0,-1,1)*CbH1YZ(3,15)

      CbH4YZ(3,2,3,1,16) = 
     $  - 1.130375818363913d-15
     $  + 1.698698071695413d-14*zm01
     $  - 1.572135576180107d-11*zm02
     $  - 6.494775549559501d-09*zm03
     $  - 3.779489265333210d-07*zm04
     $  - 5.974015529831774d-06*zm05
     $  - 3.279231610828874d-05*zm06
     $  - 6.362904834717872d-05*zm07
     $  - 1.920528324575288d-05*zm08
     $  + 5.622077946979588d-05*zm09
     $  + 5.041570731276379d-05*zm10
     $  + 1.397497762562961d-05*zm11
     $  + 1.337323758940561d-06*zm12
     $  + 3.618339276735464d-08*zm13
     $  + 1.510663025756628d-10*zm14
     $  + 1.243437998301508d-14*zm15
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,16)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,16)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,16)
     $  + HZ3(-1,0,1)*CbH1YZ(3,16)
     $  - HZ3(0,-1,1)*CbH1YZ(3,16)

      CbH4YZ(3,2,3,1,17) = 
     $  - 5.228120170057312d-17
     $  + 8.942937410008348d-16*zm01
     $  + 1.369741191158303d-12*zm02
     $  + 8.728213118645191d-10*zm03
     $  + 7.136532753525181d-08*zm04
     $  + 1.557501591127326d-06*zm05
     $  + 1.205128120509072d-05*zm06
     $  + 3.608153479397884d-05*zm07
     $  + 3.473722305767622d-05*zm08
     $  - 1.582320967662229d-05*zm09
     $  - 4.241133252687029d-05*zm10
     $  - 2.183595642683282d-05*zm11
     $  - 4.142541570760088d-06*zm12
     $  - 2.813862458055071d-07*zm13
     $  - 5.339247058153782d-09*zm14
     $  - 1.447271306964074d-11*zm15
     $  - 6.412101869818217d-16*zm16
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,17)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,17)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,17)
     $  + HZ3(-1,0,1)*CbH1YZ(3,17)
     $  - HZ3(0,-1,1)*CbH1YZ(3,17)

      CbH4YZ(3,2,3,1,18) = 
     $  + 1.221706108953050d-17
     $  - 2.137111177361339d-16*zm01
     $  - 1.129935337438080d-13*zm02
     $  - 1.102777779932175d-10*zm03
     $  - 1.253172515790844d-08*zm04
     $  - 3.707550984049253d-07*zm05
     $  - 3.918147862914472d-06*zm06
     $  - 1.680525500365466d-05*zm07
     $  - 2.832946911528492d-05*zm08
     $  - 8.065900291563469d-06*zm09
     $  + 2.417950074317844d-05*zm10
     $  + 2.401023708141069d-05*zm11
     $  + 8.152267637693257d-06*zm12
     $  + 1.104931505164430d-06*zm13
     $  + 5.449624839037510d-08*zm14
     $  + 7.349704400699740d-10*zm15
     $  + 1.301673458826710d-12*zm16
     $  + 2.650926801287483d-18*zm17
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,18)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,18)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,18)
     $  + HZ3(-1,0,1)*CbH1YZ(3,18)
     $  - HZ3(0,-1,1)*CbH1YZ(3,18)

      CbH4YZ(3,2,3,1,19) = 
     $  + 8.942194027780485d-19
     $  - 1.661490786280717d-17*zm01
     $  + 9.205919438837334d-15*zm02
     $  + 1.315586273881008d-11*zm03
     $  + 2.060388245627683d-09*zm04
     $  + 8.145942974134217d-08*zm05
     $  + 1.149070073470637d-06*zm06
     $  + 6.755109773579409d-06*zm07
     $  + 1.709992259399164d-05*zm08
     $  + 1.509509257750027d-05*zm09
     $  - 6.783982960735008d-06*zm10
     $  - 1.916358139293328d-05*zm11
     $  - 1.125937592207032d-05*zm12
     $  - 2.696967932753522d-06*zm13
     $  - 2.689264300863751d-07*zm14
     $  - 9.798341464897816d-09*zm15
     $  - 9.491126307360852d-11*zm16
     $  - 1.102765638656800d-13*zm17
     $  + 1.176147200653516d-18*zm18
     $  + 8.224670334241132d-01*HZ1(-1)*CbH1YZ(3,19)
     $  - 6.931471805599453d-01*HZ2(0,-1)*CbH1YZ(3,19)
     $  - 6.931471805599453d-01*HZ2(0,1)*CbH1YZ(3,19)
     $  + HZ3(-1,0,1)*CbH1YZ(3,19)
     $  - HZ3(0,-1,1)*CbH1YZ(3,19)

      CbH4YZ(3,3,2,1,01) = 
     $  - HZ2(0,1)*CbH2YZ(3,3,1)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,1)

      CbH4YZ(3,3,2,1,02) = 
     $  - 2.812500000000000d-01
     $  + 2.812500000000000d-01*zm01
     $  - HZ2(0,1)*CbH2YZ(3,3,2)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,2)

      CbH4YZ(3,3,2,1,03) = 
     $  + 5.273437500000000d-02
     $  - 3.515625000000000d-02*zm01
     $  - 1.757812500000000d-02*zm02
     $  - HZ2(0,1)*CbH2YZ(3,3,3)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,3)

      CbH4YZ(3,3,2,1,04) = 
     $  + 6.958007812500000d-03
     $  - 5.712890625000000d-02*zm01
     $  + 8.020019531250000d-02*zm02
     $  - 3.002929687500000d-02*zm03
     $  - HZ2(0,1)*CbH2YZ(3,3,4)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,4)

      CbH4YZ(3,3,2,1,05) = 
     $  + 4.119873046875000d-04
     $  + 2.746582031250000d-03*zm01
     $  + 4.119873046875000d-03*zm02
     $  - 2.208251953125000d-02*zm03
     $  + 1.480407714843750d-02*zm04
     $  - HZ2(0,1)*CbH2YZ(3,3,5)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,5)

      CbH4YZ(3,3,2,1,06) = 
     $  - 3.845214843750000d-06
     $  - 5.815887451171875d-04*zm01
     $  - 2.441711425781250d-03*zm02
     $  + 1.097671508789062d-02*zm03
     $  - 3.878173828125000d-03*zm04
     $  - 4.071395874023437d-03*zm05
     $  - HZ2(0,1)*CbH2YZ(3,3,6)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,6)

      CbH4YZ(3,3,2,1,07) = 
     $  - 1.544952392578125d-06
     $  + 7.405471801757812d-05*zm01
     $  + 8.116149902343750d-04*zm02
     $  - 3.748643057686941d-03*zm03
     $  - 3.001695360456194d-03*zm04
     $  + 5.060749053955078d-03*zm05
     $  + 8.054646083286830d-04*zm06
     $  - HZ2(0,1)*CbH2YZ(3,3,7)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,7)

      CbH4YZ(3,3,2,1,08) = 
     $  + 4.279022661659529d-08
     $  - 6.228286392834721d-06*zm01
     $  - 1.929349947948845d-04*zm02
     $  + 9.926654124746517d-04*zm03
     $  + 3.617927736165572d-03*zm04
     $  - 2.117359224630861d-03*zm05
     $  - 2.168899151743674d-03*zm06
     $  - 1.252142813045846d-04*zm07
     $  - HZ2(0,1)*CbH2YZ(3,3,8)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,8)

      CbH4YZ(3,3,2,1,09) = 
     $  + 1.108463929623973d-08
     $  + 4.018199339552453d-07*zm01
     $  + 3.609936942859571d-05*zm02
     $  - 2.187387541848786d-04*zm03
     $  - 2.132162390923013d-03*zm04
     $  - 5.478299698051141d-04*zm05
     $  + 2.211251440096874d-03*zm06
     $  + 6.349351345276346d-04*zm07
     $  + 1.603226628664978d-05*zm08
     $  - HZ2(0,1)*CbH2YZ(3,3,9)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,9)

      CbH4YZ(3,3,2,1,10) = 
     $  - 4.877241290345483d-10
     $  - 3.322910941029429d-08*zm01
     $  - 5.673555443307748d-06*zm02
     $  + 4.223874223475553d-05*zm03
     $  + 9.145891396975030d-04*zm04
     $  + 1.326524212743554d-03*zm05
     $  - 1.019149655407788d-03*zm06
     $  - 1.110079242380297d-03*zm07
     $  - 1.466708450379941d-04*zm08
     $  - 1.745079572885446d-06*zm09
     $  - HZ2(0,1)*CbH2YZ(3,3,10)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,10)

      CbH4YZ(3,3,2,1,11) = 
     $  - 9.699059384209769d-11
     $  + 3.568066213464597d-09*zm01
     $  + 7.896926645026610d-07*zm02
     $  - 7.430573268686148d-06*zm03
     $  - 3.188688726401118d-04*zm04
     $  - 1.017029452278988d-03*zm05
     $  - 9.235014160870399d-05*zm06
     $  + 1.005621264957872d-03*zm07
     $  + 4.006361026879179d-04*zm08
     $  + 2.846306721797608d-05*zm09
     $  + 1.654411926016452d-07*zm10
     $  - HZ2(0,1)*CbH2YZ(3,3,11)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,11)

      CbH4YZ(3,3,2,1,12) = 
     $  + 5.661341413666143d-12
     $  - 2.185978606254227d-10*zm01
     $  - 1.006302506684963d-07*zm02
     $  + 1.219030649333838d-06*zm03
     $  + 9.516523105902179d-05*zm04
     $  + 5.377611867920821d-04*zm05
     $  + 5.212815842828180d-04*zm06
     $  - 4.759180312922598d-04*zm07
     $  - 5.581564062570436d-04*zm08
     $  - 1.164307797710389d-04*zm09
     $  - 4.807055522527961d-06*zm10
     $  - 1.391675319774827d-08*zm11
     $  - HZ2(0,1)*CbH2YZ(3,3,12)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,12)

      CbH4YZ(3,3,2,1,13) = 
     $  + 9.393458856091589d-13
     $  - 2.580125800688999d-12*zm01
     $  + 1.181190516328899d-08*zm02
     $  - 1.883014172677448d-07*zm03
     $  - 2.505189945980467d-05*zm04
     $  - 2.263796400960219d-04*zm05
     $  - 4.698713103841627d-04*zm06
     $  - 5.098928703844925d-06*zm07
     $  + 4.657448394741948d-04*zm08
     $  + 2.313528286570415d-04*zm09
     $  + 2.875691082508574d-05*zm10
     $  + 7.226390241293041d-07*zm11
     $  + 1.051816267248827d-09*zm12
     $  - HZ2(0,1)*CbH2YZ(3,3,13)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,13)

      CbH4YZ(3,3,2,1,14) = 
     $  - 6.681198657913456d-14
     $  + 3.525895373735672d-13*zm01
     $  - 1.269113950147979d-09*zm02
     $  + 2.750726126339226d-08*zm03
     $  + 5.931449789489233d-06*zm04
     $  + 8.062564267854913d-05*zm05
     $  + 2.831598338682840d-04*zm06
     $  + 2.152460018640584d-04*zm07
     $  - 2.216398776104049d-04*zm08
     $  - 2.784556666945000d-04*zm09
     $  - 7.856449049562645d-05*zm10
     $  - 6.230860386006273d-06*zm11
     $  - 9.819942878934513d-08*zm12
     $  - 7.201814449735226d-11*zm13
     $  - HZ2(0,1)*CbH2YZ(3,3,14)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,14)

      CbH4YZ(3,3,2,1,15) = 
     $  - 9.699872100084071d-15
     $  + 1.649371231378928d-13*zm01
     $  + 1.257995234543974d-10*zm02
     $  - 3.821539148307632d-09*zm03
     $  - 1.280603277889825d-06*zm04
     $  - 2.513442576167816d-05*zm05
     $  - 1.349879711160625d-04*zm06
     $  - 2.168630918659378d-04*zm07
     $  + 7.814175267254051d-06*zm08
     $  + 2.186359242161344d-04*zm09
     $  + 1.276982145939476d-04*zm10
     $  + 2.289990741361553d-05*zm11
     $  + 1.209363527243961d-06*zm12
     $  + 1.219807681978839d-08*zm13
     $  + 4.510940522083749d-12*zm14
     $  - HZ2(0,1)*CbH2YZ(3,3,15)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,15)

      CbH4YZ(3,3,2,1,16) = 
     $  + 8.005170067599970d-16
     $  - 1.291896017811707d-14*zm01
     $  - 1.181789102519920d-11*zm02
     $  + 5.074714401843443d-10*zm03
     $  + 2.547395632963047d-07*zm04
     $  + 7.010779698440539d-06*zm05
     $  + 5.434627266864274d-05*zm06
     $  + 1.433523303969164d-04*zm07
     $  + 9.226037020845674d-05*zm08
     $  - 1.038758141502446d-04*zm09
     $  - 1.385918675233515d-04*zm10
     $  - 4.864883551895846d-05*zm11
     $  - 5.893677875139880d-06*zm12
     $  - 2.133959277538075d-07*zm13
     $  - 1.396919665413250d-09*zm14
     $  - 2.620697339840643d-13*zm15
     $  - HZ2(0,1)*CbH2YZ(3,3,16)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,16)

      CbH4YZ(3,3,2,1,17) = 
     $  + 1.047873077020433d-16
     $  - 1.671713626028703d-15*zm01
     $  + 1.063878879265575d-12*zm02
     $  - 6.442603639866973d-11*zm03
     $  - 4.706884442896506d-08*zm04
     $  - 1.776987407748449d-06*zm05
     $  - 1.914179467366295d-05*zm06
     $  - 7.490500091312231d-05*zm07
     $  - 1.008848692867805d-04*zm08
     $  + 6.746984767105749d-06*zm09
     $  + 1.038392139143930d-04*zm10
     $  + 6.872518104326354d-05*zm11
     $  + 1.604447708554909d-05*zm12
     $  + 1.365168608208796d-06*zm13
     $  + 3.461056430834006d-08*zm14
     $  + 1.484924575746069d-10*zm15
     $  + 1.418158272125980d-14*zm16
     $  - HZ2(0,1)*CbH2YZ(3,3,17)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,17)

      CbH4YZ(3,3,2,1,18) = 
     $  - 9.723038811891347d-18
     $  + 1.607220512644015d-16*zm01
     $  - 8.886443079986273d-14*zm02
     $  + 7.798904032401082d-12*zm03
     $  + 8.131418341252538d-09*zm04
     $  + 4.140305675158336d-07*zm05
     $  + 6.032117436194094d-06*zm06
     $  + 3.310928484809912d-05*zm07
     $  + 7.150632365744002d-05*zm08
     $  + 4.071774696805702d-05*zm09
     $  - 4.913302305837794d-05*zm10
     $  - 6.903040592302711d-05*zm11
     $  - 2.863111200553566d-05*zm12
     $  - 4.699326901812393d-06*zm13
     $  - 2.885563158515630d-07*zm14
     $  - 5.203666930620028d-09*zm15
     $  - 1.473360260323531d-11*zm16
     $  - 7.000525987177085d-16*zm17
     $  - HZ2(0,1)*CbH2YZ(3,3,18)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,18)

      CbH4YZ(3,3,2,1,19) = 
     $  - 1.170958020138070d-18
     $  + 2.164142781049748d-17*zm01
     $  + 6.728603425349032d-15*zm02
     $  - 9.006437075904762d-13*zm03
     $  - 1.320462850280406d-09*zm04
     $  - 8.947413230980534d-08*zm05
     $  - 1.727561889563743d-06*zm06
     $  - 1.282717527480928d-05*zm07
     $  - 4.012861696491306d-05*zm08
     $  - 4.742148561077396d-05*zm09
     $  + 4.209758455322313d-06*zm10
     $  + 4.983637889185606d-05*zm11
     $  + 3.644619868793107d-05*zm12
     $  + 1.040234256757332d-05*zm13
     $  + 1.243983662150740d-06*zm14
     $  + 5.624143561293425d-08*zm15
     $  + 7.301577657405322d-10*zm16
     $  + 1.370870832928365d-12*zm17
     $  + 3.174609085805230d-17*zm18
     $  - HZ2(0,1)*CbH2YZ(3,3,19)
     $  - 2.000000000000000d+00*HZ3(0,0,1)*CbH1YZ(3,19)
      endif

      return
      end

